Type: Package
Package: INLAvaan
Title: Approximate Bayesian Latent Variable Analysis
Version: 0.2.5
Authors@R: c(
    person("Haziq", "Jamil", , "haziq.jamil@gmail.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0003-3298-1010")),
    person("Håvard", "Rue", role = "ctb",
           comment = c(ORCID = "0000-0002-0222-1881", "Statistical and computational methodology")),
    person("Alvin", "Bong", role = "ctb",
           comment = "Initial site build")
  )
Description: Implements approximate Bayesian inference for Structural
    Equation Models (SEM) using a custom adaptation of the Integrated
    Nested Laplace Approximation (Rue et al., 2009)
    <doi:10.1111/j.1467-9868.2008.00700.x> as described in Jamil and Rue
    (2026a) <doi:10.48550/arXiv.2603.25690>. Provides a computationally
    efficient alternative to Markov Chain Monte Carlo (MCMC) for Bayesian
    estimation, allowing users to fit latent variable models using the
    'lavaan' syntax. See also the companion paper on implementation and
    workflows, Jamil and Rue (2026b) <doi:10.48550/arXiv.2604.00671>.
License: GPL (>= 3)
URL: https://inlavaan.haziqj.ml/, https://github.com/haziqj/INLAvaan
BugReports: https://github.com/haziqj/INLAvaan/issues
Depends: R (>= 3.5)
Imports: cli, graphics, lavaan (>= 0.6-19), methods, parallel, stats,
        utils
Suggests: blavaan, ggplot2, knitr, lme4, numDeriv, qrng, quarto, sn,
        testthat (>= 3.0.0), ucminf
VignetteBuilder: quarto
Config/Needs/website: rmarkdown
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-06-11 08:47:58 UTC; haziqj
Author: Haziq Jamil [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-3298-1010>),
  Håvard Rue [ctb] (ORCID: <https://orcid.org/0000-0002-0222-1881>,
    Statistical and computational methodology),
  Alvin Bong [ctb] (Initial site build)
Maintainer: Haziq Jamil <haziq.jamil@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-11 09:10:02 UTC
Built: R 4.5.2; ; 2026-06-11 10:50:05 UTC; unix

Type: Package
Package: INLAvaan
Title: Approximate Bayesian Latent Variable Analysis
Version: 0.3.1
Authors@R: c(
    person("Haziq", "Jamil", , "haziq.jamil@gmail.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0003-3298-1010")),
    person("Håvard", "Rue", role = "ctb",
           comment = c(ORCID = "0000-0002-0222-1881", "Statistical and computational methodology")),
    person("Alvin", "Bong", role = "ctb",
           comment = "Initial site build")
  )
Description: Implements approximate Bayesian inference for Structural
    Equation Models (SEM) using a custom adaptation of the Integrated
    Nested Laplace Approximation (Rue et al., 2009)
    <doi:10.1111/j.1467-9868.2008.00700.x> as described in Jamil and Rue
    (2026a) <doi:10.48550/arXiv.2603.25690>. Provides a computationally
    efficient alternative to Markov Chain Monte Carlo (MCMC) for Bayesian
    estimation, allowing users to fit latent variable models using the
    'lavaan' syntax. See also the companion paper on implementation and
    workflows, Jamil and Rue (2026b) <doi:10.48550/arXiv.2604.00671>.
License: GPL (>= 3)
URL: https://inlavaan.haziqj.ml/, https://github.com/haziqj/INLAvaan
BugReports: https://github.com/haziqj/INLAvaan/issues
Depends: R (>= 3.5)
Imports: cli, graphics, lavaan (>= 0.6-19), methods, parallel, stats,
        utils
Suggests: blavaan, ggplot2, knitr, lme4, numDeriv, qrng, quarto, sn,
        testthat (>= 3.0.0), ucminf
VignetteBuilder: quarto
Config/Needs/website: rmarkdown
Config/roxygen2/markdown: TRUE
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Encoding: UTF-8
NeedsCompilation: yes
Packaged: 2026-07-21 16:03:06 UTC; haziqj
Author: Haziq Jamil [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-3298-1010>),
  Håvard Rue [ctb] (ORCID: <https://orcid.org/0000-0002-0222-1881>,
    Statistical and computational methodology),
  Alvin Bong [ctb] (Initial site build)
Maintainer: Haziq Jamil <haziq.jamil@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-21 22:30:09 UTC
Built: R 4.6.0; ; 2026-07-22 03:28:49 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(compare,inlavaan_internal)
S3method(deviance,INLAvaan)
S3method(loo,INLAvaan)
S3method(loo,inlavaan_internal)
S3method(plot,inlavaan_internal)
S3method(predict,inlavaan_internal)
S3method(print,bfit_indices)
S3method(print,compare.inlavaan_internal)
S3method(print,diagnostics.INLAvaan)
S3method(print,diagnostics.INLAvaan.param)
S3method(print,fitmeasures.inlavaan_internal)
S3method(print,inlavaan_deviance)
S3method(print,inlavaan_internal)
S3method(print,inlavaan_logLik)
S3method(print,inlavaan_loo)
S3method(print,inlavaan_waic)
S3method(print,predict.inlavaan_internal)
S3method(print,summary.inlavaan_internal)
S3method(print,summary.predict.inlavaan_internal)
S3method(print,timing.INLAvaan)
S3method(sampling,inlavaan_internal)
S3method(summary,bfit_indices)
S3method(summary,inlavaan_internal)
S3method(summary,predict.inlavaan_internal)
S3method(waic,INLAvaan)
S3method(waic,inlavaan_internal)
export(acfa)
export(add_loo)
export(agrowth)
export(as_fun_string)
export(asem)
export(bfit_indices)
export(compare)
export(dbeta_box)
export(diagnostics)
export(dsnorm)
export(fit_skew_normal)
export(fit_skew_normal_samp)
export(get_inlavaan_internal)
export(inlavaan)
export(is_same_function)
export(loo)
export(priors_for)
export(qsnorm_fast)
export(sampling)
export(standardisedSolution)
export(standardisedsolution)
export(standardizedSolution)
export(standardizedsolution)
export(timing)
export(waic)
exportClasses(INLAvaan)
exportMethods(AIC)
exportMethods(BIC)
exportMethods(anova)
exportMethods(coef)
exportMethods(compare)
exportMethods(diagnostics)
exportMethods(fitMeasures)
exportMethods(fitmeasures)
exportMethods(fitted)
exportMethods(fitted.values)
exportMethods(logLik)
exportMethods(nobs)
exportMethods(plot)
exportMethods(predict)
exportMethods(resid)
exportMethods(residuals)
exportMethods(sampling)
exportMethods(show)
exportMethods(simulate)
exportMethods(summary)
exportMethods(timing)
exportMethods(update)
exportMethods(vcov)
import(methods)
importClassesFrom(lavaan,lavaan)
importFrom(cli,cli_abort)
importFrom(cli,cli_alert_info)
importFrom(cli,cli_alert_warning)
importFrom(cli,cli_inform)
importFrom(cli,cli_progress_bar)
importFrom(cli,cli_progress_done)
importFrom(cli,cli_progress_step)
importFrom(cli,cli_progress_update)
importFrom(cli,cli_rule)
importFrom(cli,cli_warn)
importFrom(cli,col_grey)
importFrom(cli,symbol)
importFrom(grDevices,adjustcolor)
importFrom(grDevices,recordPlot)
importFrom(graphics,abline)
importFrom(graphics,barplot)
importFrom(graphics,grid)
importFrom(graphics,layout)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,mtext)
importFrom(graphics,par)
importFrom(graphics,plot.new)
importFrom(graphics,polygon)
importFrom(stats,aggregate)
importFrom(stats,anova)
importFrom(stats,approx)
importFrom(stats,cov)
importFrom(stats,cov2cor)
importFrom(stats,dbeta)
importFrom(stats,density)
importFrom(stats,deviance)
importFrom(stats,dgamma)
importFrom(stats,dnorm)
importFrom(stats,fitted)
importFrom(stats,fitted.values)
importFrom(stats,logLik)
importFrom(stats,median)
importFrom(stats,nlminb)
importFrom(stats,optimize)
importFrom(stats,pnorm)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,reshape)
importFrom(stats,resid)
importFrom(stats,residuals)
importFrom(stats,rgamma)
importFrom(stats,rnorm)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,simulate)
importFrom(stats,splinefun)
importFrom(stats,uniroot)
importFrom(stats,update)
importFrom(stats,var)
importFrom(utils,capture.output)
importFrom(utils,head)
importFrom(utils,packageVersion)
importFrom(utils,tail)
importMethodsFrom(lavaan,fitMeasures)
importMethodsFrom(lavaan,fitmeasures)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# INLAvaan <a href="https://inlavaan.haziqj.ml"><img src="man/figures/logo.png" align="right" height="139" alt="INLAvaan website" /></a>

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![R-CMD-check](https://github.com/haziqj/INLAvaan/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/haziqj/INLAvaan/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/haziqj/INLAvaan/branch/main/graph/badge.svg)](https://app.codecov.io/gh/haziqj/INLAvaan?branch=main)
[![CRAN_Status_Badge_version_ago](http://www.r-pkg.org/badges/version-ago/INLAvaan)](https://cran.r-project.org/package=INLAvaan)
[![Dependencies](https://tinyverse.netlify.app/badge/INLAvaan)](https://cran.r-project.org/package=INLAvaan)
[![CRAN
Downloads](http://cranlogs.r-pkg.org/badges/grand-total/INLAvaan)](https://cran.r-project.org/package=INLAvaan)
![GitHub Repo
stars](https://img.shields.io/github/stars/haziqj/inlavaan)
<!-- [![GitHub Repo stars](https://img.shields.io/github/stars/haziqj/inlavaan)](https://github.com/haziqj/INLAvaan/stargazers) -->
<!-- badges: end -->

> Efficient approximate Bayesian inference for Structural Equation
> Models.

While Markov Chain Monte Carlo (MCMC) methods remain the gold standard
for exact Bayesian inference, they can be prohibitively slow for
iterative model development. `{INLAvaan}` offers a rapid alternative for
<u>**la**</u>tent <u>**va**</u>riable <u>**an**</u>alysis, delivering
Bayesian results at (or near) the speed of frequentist estimators. It
achieves this through a custom, ground-up implementation of the
[Integrated Nested Laplace Approximation
(INLA)](https://www.r-inla.org), engineered specifically for the
[lavaan](https://lavaan.ugent.be) modelling framework.

## A familiar interface

`{INLAvaan}` is designed to fit seamlessly into your existing workflow.
If you are familiar with the [(b)lavaan
syntax](https://lavaan.ugent.be/tutorial/syntax1.html), you can begin
using `{INLAvaan}` immediately.

As a first impression of the package, consider the canonical example of
SEM applied to the Industrialisation and Political Democracy data set of
Bollen (1989)[^1]:

``` r
library(INLAvaan)
mod_poldem <- "
  # Latent variable definitions
     ind60 =~ x1 + x2 + x3
     dem60 =~ y1 + y2 + y3
     dem65 =~ y5 + y6 + y7 + y8

  # Latent regressions
    dem60 ~ ind60
    dem65 ~ ind60 + dem60

  # Residual correlations
    y1 ~~ y5
    y2 ~~ y4 + y6
    y3 ~~ y7
    y4 ~~ y8
    y6 ~~ y8
  
  # Fixed loading
    dem60 =~ 1.5*y4
  
  # Custom priors on latent variances
    ind60 ~~ prior('gamma(1, 1)')*ind60
    dem60 ~~ prior('gamma(2, 1)')*dem60
    dem65 ~~ prior('gamma(1,.5)')*dem65
"
utils::data("PoliticalDemocracy", package = "lavaan")

fit <- asem(model = mod_poldem, data = PoliticalDemocracy)
#> ℹ Mode finding and Hessian computation.
#> ✔ Posterior mode and Hessian. [90ms]
#> 
#> ℹ Performing VB correction.
#> ✔ VB correction; mean |δ| = 0.157σ. [103ms]
#> 
#> ⠙ Fitting 0/30 skew-normal marginals.
#> ✔ Fit 30/30 skew-normal marginals. [951ms]
#> 
#> ℹ Adjusting copula correlations (NORTA).
#> ✔ Adjust copula correlations (NORTA). [100ms]
#> 
#> ⠙ Posterior sampling and summarising.
#> ⠹ Computing fit indices (PPP/DIC).
#> ✔ Summarise 1000 posterior draws. [503ms]
#> 
#> ℹ Fit measures: PPP, DIC, LOO, WAIC.

summary(fit)
#> INLAvaan 0.3.1 ended normally after 71 iterations
#> 
#>   Estimator                                      BAYES
#>   Optimization method                           NLMINB
#>   Number of model parameters                        30
#> 
#>   Number of observations                            75
#> 
#> Model Test (User Model):
#> 
#>    Marginal log-likelihood                   -1659.573 
#>    PPP (Chi-square)                              0.464 
#> 
#> Information Criteria:
#> 
#>    Deviance (DIC)                             3175.519 
#>    Effective parameters (pD)                    29.787 
#> 
#> Parameter Estimates:
#> 
#>    Marginalisation method                     SKEWNORM
#>    VB correction                                  TRUE
#> 
#> Latent Variables:
#>                    Estimate       SD     2.5%    97.5%     NMAD    Prior       
#>   ind60 =~                                                                     
#>     x1                1.000                                                    
#>     x2                2.217    0.147    1.948    2.527    0.007    normal(0,10)
#>     x3                1.838    0.155    1.547    2.156    0.004    normal(0,10)
#>   dem60 =~                                                                     
#>     y1                1.000                                                    
#>     y2                1.444    0.170    1.116    1.782    0.001    normal(0,10)
#>     y3                1.168    0.157    0.866    1.481    0.001    normal(0,10)
#>   dem65 =~                                                                     
#>     y5                1.000                                                    
#>     y6                1.260    0.187    0.925    1.659    0.007    normal(0,10)
#>     y7                1.355    0.174    1.046    1.729    0.008    normal(0,10)
#>     y8                1.367    0.177    1.054    1.750    0.008    normal(0,10)
#>   dem60 =~                                                                     
#>     y4                1.500                                                    
#> 
#> Regressions:
#>                    Estimate       SD     2.5%    97.5%     NMAD    Prior       
#>   dem60 ~                                                                      
#>     ind60             1.373    0.349    0.698    2.067    0.001    normal(0,10)
#>   dem65 ~                                                                      
#>     ind60             0.516    0.235    0.063    0.985    0.001    normal(0,10)
#>     dem60             0.882    0.103    0.688    1.091    0.011    normal(0,10)
#> 
#> Covariances:
#>                    Estimate       SD     2.5%    97.5%     NMAD    Prior       
#>  .y1 ~~                                                                        
#>    .y5                0.836    0.401    0.111    1.687    0.005       beta(1,1)
#>  .y2 ~~                                                                        
#>    .y4                1.102    0.729   -0.201    2.662    0.004       beta(1,1)
#>    .y6                2.221    0.760    0.809    3.795    0.011       beta(1,1)
#>  .y3 ~~                                                                        
#>    .y7                1.032    0.670   -0.186    2.447    0.005       beta(1,1)
#>  .y8 ~~                                                                        
#>    .y4                0.256    0.461   -0.646    1.159    0.003       beta(1,1)
#>  .y6 ~~                                                                        
#>    .y8                1.297    0.596    0.244    2.583    0.005       beta(1,1)
#> 
#> Variances:
#>                    Estimate       SD     2.5%    97.5%     NMAD    Prior       
#>     ind60             0.464    0.090    0.314    0.667    0.003      gamma(1,1)
#>    .dem60             3.169    0.615    2.135    4.534    0.001      gamma(2,1)
#>    .dem65             0.331    0.194    0.059    0.787    0.037     gamma(1,.5)
#>    .x1                0.090    0.022    0.053    0.137    0.006 gamma(1,.5)[sd]
#>    .x2                0.133    0.067    0.031    0.284    0.031 gamma(1,.5)[sd]
#>    .x3                0.509    0.101    0.342    0.735    0.003 gamma(1,.5)[sd]
#>    .y1                2.355    0.507    1.513    3.489    0.004 gamma(1,.5)[sd]
#>    .y2                7.652    1.461    5.189   10.894    0.003 gamma(1,.5)[sd]
#>    .y3                5.596    1.094    3.782    8.054    0.002 gamma(1,.5)[sd]
#>    .y5                2.677    0.561    1.744    3.934    0.005 gamma(1,.5)[sd]
#>    .y6                5.229    0.977    3.581    7.397    0.002 gamma(1,.5)[sd]
#>    .y7                3.678    0.801    2.337    5.463    0.007 gamma(1,.5)[sd]
#>    .y8                3.292    0.746    2.018    4.931    0.007 gamma(1,.5)[sd]
#>    .y4                2.951    0.776    1.628    4.656    0.012 gamma(1,.5)[sd]
```

## Validation against MCMC

Computation speed is valuable only when accuracy is preserved. Our
method yields posterior distributions that are visually and numerically
comparable to those obtained via MCMC (e.g., via `{blavaan}`/Stan), but
at a fraction of the computational cost.

The figure below illustrates the posterior density overlap for the
example above. The percentages refer to the one minus the
[Jensen-Shannon
distance](https://en.wikipedia.org/wiki/Jensen–Shannon_divergence),
which gives a measure of similarity between two probability
distributions.

``` r
# install.packages("blavaan")
library(blavaan)
fit_blav <- bsem(model = mod_poldem, data = PoliticalDemocracy, seed = 2026)
res <- INLAvaan:::compare_mcmc(fit_blav, INLAvaan = fit)
print(res$p_compare)
```

<img src="man/figures/README-fig-compare-poldem-1.png" alt="" width="100%" />

## Installation

Install the CRAN version of `{INLAvaan}` using:

``` r
install.packages("INLAvaan")
```

Alternatively, install the development version of `{INLAvaan}` from
GitHub using:

``` r
# install.packages("pak")
pak::pak("haziqj/INLAvaan")
```

*Optionally*[^2], you may wish to install
[INLA](https://www.r-inla.org). Following the official instructions
given [here](https://www.r-inla.org/download/), install the package by
running this command in R:

``` r
install.packages(
  "INLA",
  repos = c(getOption("repos"), 
            INLA = "https://inla.r-inla-download.org/R/stable"), 
  dep = TRUE
)
```

## Citation

There are two papers related to `{INLAvaan}` and its underlying
methodology. To cite `{INLAvaan}` in publications, consider citing both.

To cite the methodological contribution exclusively, please use:

> Jamil, H., & Rue, H. (2026). *Approximate Bayesian inference for
> structural equation models using integrated nested Laplace
> approximations* (2603.25690 \[stat.ME\]). arXiv.
> <https://doi.org/10.48550/arXiv.2603.25690>

To cite the software implementation and workflows, please use:

> Jamil, H., & Rue, H. (2026). *Implementation and workflows for
> INLA-based approximate Bayesian structural equation modelling*
> (2604.00671 \[stat.CO\]). arXiv.
> <https://doi.org/10.48550/arXiv.2604.00671>

BibTeX entries for LaTeX users:

    @Misc{jamil2026approximate,
      title = {Approximate Bayesian inference for structural equation models using integrated nested Laplace approximations},
      author = {Haziq Jamil and Håvard Rue},
      year = {2026},
      number = {2603.25690 [stat.ME]},
      eprint = {2603.25690},
      primaryclass = {stat.ME},
      publisher = {arXiv},
      doi = {10.48550/arXiv.2603.25690},
      archiveprefix = {arXiv},
      copyright = {Creative Commons Attribution Non Commercial Share Alike 4.0 International},
    }

    @Misc{jamil2026implementation,
      title = {Implementation and workflows for INLA-based approximate Bayesian structural equation modelling},
      author = {Haziq Jamil and Håvard Rue},
      year = {2026},
      number = {2604.00671 [stat.CO]},
      eprint = {2604.00671},
      primaryclass = {stat.CO},
      publisher = {arXiv},
      doi = {10.48550/arXiv.2604.00671},
      archiveprefix = {arXiv},
      copyright = {Creative Commons Attribution Non Commercial Share Alike 4.0 International},
    }

## License

The `{INLAvaan}` package is licensed under the
[GPL-3](https://www.gnu.org/licenses/gpl-3.0.en.html).

``` plaintext
INLAvaan: Bayesian Latent Variable Analysis using INLA
Copyright (C) 2026 Haziq Jamil

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
```

<!-- By using this package, you agree to comply with both licenses:  -->

<!-- the GPL-3 license for the software and the CC BY 4.0 license for the data. -->

[^1]: Bollen, K. A. (1989). *Structural equations with latent variables*
    (pp. xiv, 514). John Wiley & Sons.
    <https://doi.org/10.1002/9781118619179>

[^2]: R-INLA dependency has been removed temporarily from v0.2.0.

<!-- README.md is generated from README.Rmd. Please edit that file -->

# INLAvaan <a href="https://inlavaan.haziqj.ml"><img src="man/figures/logo.png" align="right" height="139" alt="INLAvaan website" /></a>

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![R-CMD-check](https://github.com/haziqj/INLAvaan/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/haziqj/INLAvaan/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/haziqj/INLAvaan/branch/main/graph/badge.svg)](https://app.codecov.io/gh/haziqj/INLAvaan?branch=main)
[![CRAN_Status_Badge_version_ago](http://www.r-pkg.org/badges/version-ago/INLAvaan)](https://cran.r-project.org/package=INLAvaan)
[![Dependencies](https://tinyverse.netlify.app/badge/INLAvaan)](https://cran.r-project.org/package=INLAvaan)
[![CRAN
Downloads](http://cranlogs.r-pkg.org/badges/grand-total/INLAvaan)](https://cran.r-project.org/package=INLAvaan)
[![GitHub Repo
stars](https://img.shields.io/github/stars/haziqj/inlavaan)](https://github.com/haziqj/INLAvaan/stargazers)
<!-- badges: end -->

> Efficient approximate Bayesian inference for Structural Equation
> Models.

While Markov Chain Monte Carlo (MCMC) methods remain the gold standard
for exact Bayesian inference, they can be prohibitively slow for
iterative model development. `{INLAvaan}` offers a rapid alternative for
<u>**la**</u>tent <u>**va**</u>riable <u>**an**</u>alysis, delivering
Bayesian results at (or near) the speed of frequentist estimators. It
achieves this through a custom, ground-up implementation of the
[Integrated Nested Laplace Approximation
(INLA)](https://www.r-inla.org), engineered specifically for the
[lavaan](https://lavaan.ugent.be) modelling framework.

## A familiar interface

`{INLAvaan}` is designed to fit seamlessly into your existing workflow.
If you are familiar with the [(b)lavaan
syntax](https://lavaan.ugent.be/tutorial/syntax1.html), you can begin
using `{INLAvaan}` immediately.

As a first impression of the package, consider the canonical example of
SEM applied to the Industrialisation and Political Democracy data set of
Bollen (1989)[^1]:

``` r
library(INLAvaan)
mod_poldem <- "
  # Latent variable definitions
     ind60 =~ x1 + x2 + x3
     dem60 =~ y1 + y2 + y3
     dem65 =~ y5 + y6 + y7 + y8

  # Latent regressions
    dem60 ~ ind60
    dem65 ~ ind60 + dem60

  # Residual correlations
    y1 ~~ y5
    y2 ~~ y4 + y6
    y3 ~~ y7
    y4 ~~ y8
    y6 ~~ y8
  
  # Fixed loading
    dem60 =~ 1.5*y4
  
  # Custom priors on latent variances
    ind60 ~~ prior('gamma(1, 1)')*ind60
    dem60 ~~ prior('gamma(2, 1)')*dem60
    dem65 ~~ prior('gamma(1,.5)')*dem65
"
utils::data("PoliticalDemocracy", package = "lavaan")

fit <- asem(model = mod_poldem, data = PoliticalDemocracy)
#> ℹ Finding posterior mode.
#> ✔ Finding posterior mode. [35ms]
#> 
#> ℹ Computing the Hessian.
#> ✔ Computing the Hessian. [33ms]
#> 
#> ℹ Performing VB correction.
#> ✔ VB correction; mean |δ| = 0.156σ. [67ms]
#> 
#> ⠙ Fitting 0/30 skew-normal marginals.
#> ✔ Fitting 30/30 skew-normal marginals. [484ms]
#> 
#> ℹ Adjusting copula correlations (NORTA).
#> ✔ Adjusting copula correlations (NORTA). [111ms]
#> 
#> ⠙ Posterior sampling and summarising.
#> ✔ Posterior sampling and summarising. [269ms]
#> 

summary(fit)
#> INLAvaan 0.2.5 ended normally after 80 iterations
#> 
#>   Estimator                                      BAYES
#>   Optimization method                           NLMINB
#>   Number of model parameters                        30
#> 
#>   Number of observations                            75
#> 
#> Model Test (User Model):
#> 
#>    Marginal log-likelihood                   -1651.234 
#>    PPP (Chi-square)                              0.486 
#> 
#> Information Criteria:
#> 
#>    Deviance (DIC)                             3157.281 
#>    Effective parameters (pD)                    29.234 
#> 
#> Parameter Estimates:
#> 
#>    Marginalisation method                     SKEWNORM
#>    VB correction                                  TRUE
#> 
#> Latent Variables:
#>                    Estimate       SD     2.5%    97.5%     NMAD    Prior       
#>   ind60 =~                                                                     
#>     x1                1.000                                                    
#>     x2                2.216    0.146    1.950    2.524    0.006    normal(0,10)
#>     x3                1.838    0.154    1.548    2.154    0.004    normal(0,10)
#>   dem60 =~                                                                     
#>     y1                1.000                                                    
#>     y2                1.443    0.169    1.117    1.779    0.001    normal(0,10)
#>     y3                1.168    0.156    0.868    1.478    0.001    normal(0,10)
#>   dem65 =~                                                                     
#>     y5                1.000                                                    
#>     y6                1.260    0.186    0.926    1.656    0.007    normal(0,10)
#>     y7                1.355    0.173    1.048    1.726    0.008    normal(0,10)
#>     y8                1.367    0.176    1.056    1.746    0.008    normal(0,10)
#>   dem60 =~                                                                     
#>     y4                1.500                                                    
#> 
#> Regressions:
#>                    Estimate       SD     2.5%    97.5%     NMAD    Prior       
#>   dem60 ~                                                                      
#>     ind60             1.373    0.347    0.702    2.063    0.001    normal(0,10)
#>   dem65 ~                                                                      
#>     ind60             0.516    0.233    0.066    0.982    0.001    normal(0,10)
#>     dem60             0.882    0.102    0.689    1.090    0.011    normal(0,10)
#> 
#> Covariances:
#>                    Estimate       SD     2.5%    97.5%     NMAD    Prior       
#>  .y1 ~~                                                                        
#>    .y5                0.331    0.378    0.100    1.582    0.005       beta(1,1)
#>  .y2 ~~                                                                        
#>    .y4                0.217    0.736   -0.231    2.657    0.004       beta(1,1)
#>    .y6                0.347    0.717    0.891    3.707    0.011       beta(1,1)
#>  .y3 ~~                                                                        
#>    .y7                0.225    0.649   -0.104    2.444    0.005       beta(1,1)
#>  .y8 ~~                                                                        
#>    .y4                0.070    0.445   -0.575    1.172    0.003       beta(1,1)
#>  .y6 ~~                                                                        
#>    .y8                0.307    0.584    0.228    2.523    0.005       beta(1,1)
#> 
#> Variances:
#>                    Estimate       SD     2.5%    97.5%     NMAD    Prior       
#>     ind60             0.458    0.089    0.311    0.657    0.003      gamma(1,1)
#>    .dem60             3.132    0.604    2.116    4.475    0.001      gamma(2,1)
#>    .dem65             0.325    0.190    0.058    0.771    0.037     gamma(1,.5)
#>    .x1                0.088    0.021    0.053    0.135    0.006 gamma(1,.5)[sd]
#>    .x2                0.131    0.066    0.031    0.279    0.031 gamma(1,.5)[sd]
#>    .x3                0.501    0.099    0.338    0.723    0.003 gamma(1,.5)[sd]
#>    .y1                2.321    0.496    1.496    3.430    0.004 gamma(1,.5)[sd]
#>    .y2                7.547    1.432    5.131   10.720    0.003 gamma(1,.5)[sd]
#>    .y3                5.518    1.071    3.739    7.922    0.002 gamma(1,.5)[sd]
#>    .y5                2.639    0.550    1.725    3.869    0.005 gamma(1,.5)[sd]
#>    .y6                5.158    0.957    3.542    7.279    0.002 gamma(1,.5)[sd]
#>    .y7                3.627    0.784    2.312    5.374    0.006 gamma(1,.5)[sd]
#>    .y8                3.247    0.731    1.997    4.852    0.007 gamma(1,.5)[sd]
#>    .y4                2.910    0.761    1.613    4.580    0.011 gamma(1,.5)[sd]
```

## Validation against MCMC

Computation speed is valuable only when accuracy is preserved. Our
method yields posterior distributions that are visually and numerically
comparable to those obtained via MCMC (e.g., via `{blavaan}`/Stan), but
at a fraction of the computational cost.

The figure below illustrates the posterior density overlap for the
example above. The percentages refer to the one minus the
[Jensen-Shannon
distance](https://en.wikipedia.org/wiki/Jensen–Shannon_divergence),
which gives a measure of similarity between two probability
distributions.

``` r
# install.packages("blavaan")
library(blavaan)
fit_blav <- bsem(model = mod_poldem, data = PoliticalDemocracy, seed = 2026)
res <- INLAvaan:::compare_mcmc(fit_blav, INLAvaan = fit)
print(res$p_compare)
```

<img src="man/figures/README-fig-compare-poldem-1.png" alt="" width="100%" />

## Installation

Install the CRAN version of `{INLAvaan}` using:

``` r
install.packages("INLAvaan")
```

Alternatively, install the development version of `{INLAvaan}` from
GitHub using:

``` r
# install.packages("pak")
pak::pak("haziqj/INLAvaan")
```

*Optionally*[^2], you may wish to install
[INLA](https://www.r-inla.org). Following the official instructions
given [here](https://www.r-inla.org/download/), install the package by
running this command in R:

``` r
install.packages(
  "INLA",
  repos = c(getOption("repos"), 
            INLA = "https://inla.r-inla-download.org/R/stable"), 
  dep = TRUE
)
```

## Citation

There are two papers related to `{INLAvaan}` and its underlying
methodology. To cite `{INLAvaan}` in publications, consider citing both.

To cite the methodological contribution exclusively, please use:

> Jamil, H., & Rue, H. (2026). *Approximate Bayesian inference for
> structural equation models using integrated nested Laplace
> approximations* (2603.25690 \[stat.ME\]). arXiv.
> <https://doi.org/10.48550/arXiv.2603.25690>

To cite the software implementation and workflows, please use:

> Jamil, H., & Rue, H. (2026). *Implementation and workflows for
> INLA-based approximate Bayesian structural equation modelling*
> (2604.00671 \[stat.CO\]). arXiv.
> <https://doi.org/10.48550/arXiv.2604.00671>

BibTeX entries for LaTeX users:

    @Misc{jamil2026approximate,
      title = {Approximate Bayesian inference for structural equation models using integrated nested Laplace approximations},
      author = {Haziq Jamil and Håvard Rue},
      year = {2026},
      number = {2603.25690 [stat.ME]},
      eprint = {2603.25690},
      primaryclass = {stat.ME},
      publisher = {arXiv},
      doi = {10.48550/arXiv.2603.25690},
      archiveprefix = {arXiv},
      copyright = {Creative Commons Attribution Non Commercial Share Alike 4.0 International},
    }

    @Misc{jamil2026implementation,
      title = {Implementation and workflows for INLA-based approximate Bayesian structural equation modelling},
      author = {Haziq Jamil and Håvard Rue},
      year = {2026},
      number = {2604.00671 [stat.CO]},
      eprint = {2604.00671},
      primaryclass = {stat.CO},
      publisher = {arXiv},
      doi = {10.48550/arXiv.2604.00671},
      archiveprefix = {arXiv},
      copyright = {Creative Commons Attribution Non Commercial Share Alike 4.0 International},
    }

## License

The `{INLAvaan}` package is licensed under the
[GPL-3](https://www.gnu.org/licenses/gpl-3.0.en.html).

``` plaintext
INLAvaan: Bayesian Latent Variable Analysis using INLA
Copyright (C) 2026 Haziq Jamil

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
```

<!-- By using this package, you agree to comply with both licenses:  -->

<!-- the GPL-3 license for the software and the CC BY 4.0 license for the data. -->

[^1]: Bollen, K. A. (1989). *Structural equations with latent variables*
    (pp. xiv, 514). John Wiley & Sons.
    <https://doi.org/10.1002/9781118619179>

[^2]: R-INLA dependency has been removed temporarily from v0.2.0.

HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"
utils::data("HolzingerSwineford1939", package = "lavaan")

# Fit a CFA model with standardised latent variables
fit <- acfa(
  HS.model,
  data = HolzingerSwineford1939,
  test = "none",
  nsamp = 10,
  vb_correction = FALSE,
  verbose = FALSE
)
standardisedsolution(fit, nsamp = 10, se = FALSE, ci = FALSE)

\donttest{
# Model comparison on multigroup analysis (measurement invariance)
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")

# Configural invariance
fit1 <- acfa(HS.model, data = HolzingerSwineford1939, group = "school")

# Weak invariance
fit2 <- acfa(
  HS.model,
  data = HolzingerSwineford1939,
  group = "school",
  group.equal = "loadings"
)

# Strong invariance
fit3 <- acfa(
  HS.model,
  data = HolzingerSwineford1939,
  group = "school",
  group.equal = c("intercepts", "loadings")
)

# Compare models (fit1 = configural = baseline, always first argument)
compare(fit1, fit2, fit3)

# With extra fit measures
compare(fit1, fit2, fit.measures = c("BRMSEA", "BMc"))

# With incremental indices (baseline = fit1, passed to fitMeasures())
compare(fit1, fit2, fit3, fit.measures = c("BCFI", "BTLI"))
}

# The famous Holzinger and Swineford (1939) example
hs_mod <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
fit <- acfa(hs_mod, data = lavaan::HolzingerSwineford1939)

# The famous Holzinger and Swineford (1939) example
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")

# Fit a CFA model with standardised latent variables
fit <- acfa(HS.model, data = HolzingerSwineford1939, std.lv = TRUE, nsamp = 100)
summary(fit)

## Validation: CFA model against MCMC (blavaan)
## Extracted from tests/testthat/test-cfa.R
## Requires: future, blavaan

dat <- lavaan::HolzingerSwineford1939
mod <- "
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
    speed   =~ x7 + x8 + x9
  "

testthat::skip_on_ci()
testthat::skip_on_cran()
testthat::skip_if_not(interactive())
testthat::skip_if_not_installed("future")
library(blavaan)
future::plan("multisession", workers = future::availableCores() - 2)

fit_blav <- bcfa(mod, dat, bcontrol = list(cores = 3))
fit_inl1 <- acfa(mod, dat, marginal_method = "skewnorm", debug = TRUE)
fit_inl2 <- acfa(mod, dat, marginal_method = "asymgaus", debug = TRUE)
fit_inl3 <- acfa(mod, dat, marginal_method = "sampling", debug = TRUE)

res <- compare_mcmc(
  fit_blav,
  "skewnorm" = fit_inl1,
  "asymgaus" = fit_inl2,
  "sampling" = fit_inl3
)
print(res$p_compare)
print(res$p_errors)
print(res$metrics_df, n = 1000)

# Fit a SN curve to gamma log-density
logdens <- function(x) dgamma(x, shape = 3, rate = 1, log = TRUE)

x_grid <- seq(0.1, 8, length.out = 21)
y_log  <- sapply(x_grid, logdens)
y_log  <- y_log - max(y_log)  # normalise to have maximum at zero

res <- fit_skew_normal(x_grid, y_log, temp = 10)
unlist(res)

# Compare truth vs skew-normal approximation
x_fine <- seq(0.1, 8, length.out = 200)
y_true <- exp(logdens(x_fine))
y_sn   <- dsnorm(x_fine, xi = res$xi, omega = res$omega, alpha = res$alpha)

plot(x_fine, y_true, type = "n", xlab = "x", ylab = "Density", bty = "n")
polygon(c(x_fine, rev(x_fine)), c(y_true, rep(0, length(x_fine))),
        col = adjustcolor("#131516", 0.25), border = NA)
lines(x_fine, y_sn, col = "#00A6AA", lwd = 2)
legend("topright", legend = c("Truth", "SN Approx."),
       fill = c(adjustcolor("#131516", 0.25), NA), border = NA,
       col = c(NA, "#00A6AA"), lwd = c(NA, 2), lty = c(NA, 1),
       bty = "n")

testthat::skip()

## ----- 1. Define the Log-PDF -------------------------------------------------
# p(x, y) = p(x) * p(y|x)
# p(x)    = Normal(0, 1)
# p(y|x)  = Normal(0, sd = sqrt(exp(x)))

lp_joint <- function(theta) {
  x <- theta[1]
  y <- theta[2]
  lp_x <- dnorm(x, mean = 0, sd = 1, log = TRUE)
  lp_y.x <- dnorm(y, mean = 0, sd = sqrt(exp(x)), log = TRUE)
  lp_x + lp_y.x
}

## ----- 2. Find the Joint Mode (Optimization) ---------------------------------
opt <- optim(
  c(0, 0),
  function(.theta) -1 * lp_joint(.theta),
  hessian = TRUE
)
joint_mode <- opt$par

## ----- 3. Compare the Modes --------------------------------------------------
cat(
  "=== Results ===\n",
  sprintf(
    "Marginal Mode of X:  %0.4f (By definition, since X ~ N(0,1))\n",
    0
  ),
  sprintf(
    "Joint Mode of X:    %0.4f (Calculated via optimization)\n",
    joint_mode[1]
  )
)

## ----- 4. Plot ---------------------------------------------------------------
library(ggplot2)
library(dplyr)
library(tidyr)

grid_size <- 0.1
grid <- expand.grid(
  x = seq(-2, 2, by = grid_size),
  y = seq(-2, 2, by = grid_size)
)
grid$z <- with(grid, mapply(function(x, y) lp_joint(c(x, y)), x, y))

ggplot(grid, aes(x, y)) +
  geom_contour_filled(aes(z = z), bins = 15, show.legend = FALSE) +
  # Marginal mode
  annotate("point", x = 0, y = 0, color = "deepskyblue", size = 3) +
  annotate(
    "text",
    x = 0,
    y = 0.4,
    label = "Marginal Mode (0,0)",
    color = "deepskyblue",
    fontface = "bold",
    vjust = 0
  ) +
  # Joint Mode (-0.5, 0) - Peak of density
  annotate("point", x = -0.5, y = 0, color = "red", shape = 17, size = 3) +
  annotate(
    "text",
    x = -0.5,
    y = -0.4,
    label = "Joint Mode (-0.5, 0)",
    color = "red",
    fontface = "bold",
    vjust = 1
  ) +
  scale_x_continuous(expand = c(0, 0)) +
  scale_y_continuous(expand = c(0, 0)) +
  labs(
    x = "X (Controls Variance)",
    y = "Y"
  ) +
  theme_minimal()

## ----- 5. Skew normal marginal fit -------------------------------------------
library(INLAvaan)
library(numDeriv)
sn_fit_cor <- !TRUE
sn_fit_logthresh <- -6
sn_fit_temp <- 1

theta_star <- opt$par
H_neg <- opt$hessian
Sigma_theta <- solve(0.5 * (H_neg + t(H_neg)))
m <- length(theta_star)
pars_list <- setNames(as.list(1:m), paste0("theta[", 1:m, "]"))

# For whitening transformation: z = L^{-1}(theta - theta*)
L <- t(chol(Sigma_theta))
L_inv <- solve(L)

approx_data <- matrix(NA, nrow = m, ncol = 4)
colnames(approx_data) <- c("xi", "omega", "alpha", "logC")

approx_data <-
  do.call(
    what = "rbind",
    lapply(
      pars_list,
      function(j) {
        mv <- seq(-4, 4, length = 31)
        tt <- theta_star[j] + mv * sqrt(Sigma_theta[j, j])
        yy <- numeric(length(mv))
        for (k in seq_along(mv)) {
          if (isTRUE(sn_fit_cor)) {
            # Fit in decoupled Z-space (conditional modes are 0)
            tt[k] <- mv[k]
            yy[k] <- lp_joint(theta_star + L[, j] * mv[k])
          } else {
            # Evaluate lp_joint at theta_j with others fixed at the
            # conditional mode
            theta_new <- rep(NA, length(theta_star))
            theta_new[j] <- tt[k]
            theta_new[-j] <- theta_star[-j] +
              Sigma_theta[-j, j] /
                Sigma_theta[j, j] *
                (tt[k] - theta_star[j])
            yy[k] <- lp_joint(theta_new)
          }
        }
        fit_sn <- fit_skew_normal(
          tt,
          yy - max(yy),
          threshold_log_drop = sn_fit_logthresh,
          temp = sn_fit_temp
        )
        if (isTRUE(sn_fit_cor)) {
          # Adjust back to theta space
          fit_sn$xi <- theta_star[j] + fit_sn$xi * sqrt(Sigma_theta[j, j])
          fit_sn$omega <- fit_sn$omega * sqrt(Sigma_theta[j, j])
        }
        unlist(fit_sn)
      }
    )
  )

# Compare marginal of x and skew normal fit of x. There is a shift because the
# skew normal is fitted at the joint mode, which is biased.
tibble(
  x = seq(-4, 4, length = 100),
  truth = dnorm(x, 0, 1),
  sn_fit = INLAvaan::dsnorm(
    x,
    xi = approx_data["theta[1]", "xi"],
    omega = approx_data["theta[1]", "omega"],
    alpha = approx_data["theta[1]", "alpha"]
  )
) |>
  pivot_longer(
    cols = c("truth", "sn_fit"),
    names_to = "type",
    values_to = "density"
  ) |>
  ggplot(aes(
    x,
    density,
    color = type,
    fill = type,
    alpha = type,
    linewidth = type
  )) +
  geom_area(position = "identity") +
  geom_line(alpha = 1) +
  scale_alpha_manual(values = c(0, 0.3)) +
  scale_linewidth_manual(values = c(0.8, 0)) +
  scale_colour_manual(values = c("truth" = "black", "sn_fit" = "#00A6AA")) +
  scale_fill_manual(values = c("truth" = "black", "sn_fit" = "#00A6AA")) +
  labs(
    title = "Marginal of X: True vs Skew Normal Approximation",
    y = "Density"
  ) +
  theme_minimal()

## ----- 6. The fix ------------------------------------------------------------

# When we "feel out" and explore the lp_joint(), for each value of x we move,
# the value of y is set at the joint mode (in whitened space this is 0).
# However, the volume of y-values that contribute to the marginal density of x
# changes with x. To correct for this, we apply a Laplace approximation
# correction to account for the curvature (2nd derivative) of the log-pdf wrt y
# at the joint mode.

marginal_objective <- function(x) {
  # A. Find y* for this x (Profile Likelihood)
  y_star <- if (TRUE) {
    # Cheap case: Just set at joint mode
    0
  } else {
    # General case: optimize y for this x
    optim(0, function(y) -1 * lp_joint(c(x, y)))$par
  }

  # B. Get the Joint Density Height at this point
  joint_val <- lp_joint(c(x, y_star))

  # C. Calculate Geometry (Curvature/Hessian) wrt Y
  # We calculate the 2nd derivative of the PDF wrt y at y_star
  H_yy <- hessian(function(y) lp_joint(c(x, y)), y_star)

  # D. Apply Laplace Correction
  # Correction = -0.5 * log(determinant of negative Hessian)
  # Note: H_yy is negative at the peak, so we take -H_yy
  volume_correction <- -0.5 * log(det(-H_yy))

  return(joint_val + volume_correction)
}

# Optimize Corrected (Approximate Marginal)
opt_marginal <- optimize(
  marginal_objective,
  interval = c(-3, 3),
  maximum = TRUE
)

cat(
  sprintf("Joint Mode (Uncorrected): %0.4f\n", opt$par[1]),
  sprintf("Recovered Marginal Mode:  %0.4f\n", opt_marginal$maximum)
)

## ----- 7. Corrected skew normal fit ------------------------------------------

approx_data <-
  do.call(
    what = "rbind",
    lapply(
      pars_list,
      function(j) {
        mv <- seq(-4, 4, length = 31)
        tt <- theta_star[j] + mv * sqrt(Sigma_theta[j, j])
        yy <- numeric(length(mv))
        for (k in seq_along(mv)) {
          if (isTRUE(sn_fit_cor)) {
            yy[k] <- marginal_objective(tt[k])
          } else {
            yy[k] <- marginal_objective(tt[k])
          }
        }
        fit_sn <- fit_skew_normal(
          tt,
          yy - max(yy),
          threshold_log_drop = sn_fit_logthresh,
          temp = sn_fit_temp
        )
        if (isTRUE(sn_fit_cor)) {
          # Adjust back to theta space
          # fit_sn$xi <- theta_star[j] + fit_sn$xi * sqrt(Sigma_theta[j, j])
          # fit_sn$omega <- fit_sn$omega * sqrt(Sigma_theta[j, j])
        }
        unlist(fit_sn)
      }
    )
  )

tibble(
  x = seq(-4, 4, length = 100),
  truth = dnorm(x, 0, 1),
  sn_fit = INLAvaan::dsnorm(
    x,
    xi = approx_data["theta[1]", "xi"],
    omega = approx_data["theta[1]", "omega"],
    alpha = approx_data["theta[1]", "alpha"]
  )
) |>
  pivot_longer(
    cols = c("truth", "sn_fit"),
    names_to = "type",
    values_to = "density"
  ) |>
  ggplot(aes(
    x,
    density,
    color = type,
    fill = type,
    alpha = type,
    linewidth = type
  )) +
  geom_area(position = "identity") +
  geom_line(alpha = 1) +
  scale_alpha_manual(values = c(0, 0.3)) +
  scale_linewidth_manual(values = c(0.8, 0)) +
  scale_colour_manual(values = c("truth" = "black", "sn_fit" = "#00A6AA")) +
  scale_fill_manual(values = c("truth" = "black", "sn_fit" = "#00A6AA")) +
  labs(
    title = "Marginal of X: True vs Skew Normal Approximation",
    y = "Density"
  ) +
  theme_minimal()

## ---- Cheaper fix ------------------------------------------------------------
approx_data <-
  do.call(
    what = "rbind",
    lapply(
      pars_list,
      function(j) {
        # 1. Grid in Z-space (Standard Normal Scale)
        mv <- seq(-4, 4, length = 31)

        # 2. Calculate SLA Correction Term (Gamma_1)
        # We estimate d(H_zz)/dz_j via finite difference of the Hessian in Z-space
        delta <- 0.01

        # Direction in Theta space corresponding to Z_j
        dir_vec <- L[, j]

        # Hessian at z* + delta (mapped to theta)
        H_plus_theta <- -1 *
          numDeriv::hessian(lp_joint, theta_star + dir_vec * delta)
        H_plus_z <- t(L) %*% H_plus_theta %*% L

        # Hessian at z* - delta (mapped to theta)
        H_minus_theta <- -1 *
          numDeriv::hessian(lp_joint, theta_star - dir_vec * delta)
        H_minus_z <- t(L) %*% H_minus_theta %*% L

        # 3rd derivative diagonal: d^3(log pi) / dz_j dz_k^2
        d_curvature_dz <- diag(H_plus_z - H_minus_z) / (2 * delta)

        # Rue et al (2009) Eq 21 correction: -0.5 * sum_{k!=j} (change in curvature)
        gamma_1 <- -0.5 * sum(d_curvature_dz[-j])

        # 3. Sweep axis and apply correction
        yy <- numeric(length(mv))
        for (k in seq_along(mv)) {
          # Map Z point back to Theta to evaluate Joint Log-PDF
          theta_val <- theta_star + dir_vec * mv[k]

          # SLA: Joint Density + Linear Location Correction
          yy[k] <- lp_joint(theta_val) + gamma_1 * mv[k]
        }

        # 4. Fit Skew Normal to the corrected curve
        fit_sn <- fit_skew_normal(
          mv,
          yy - max(yy),
          threshold_log_drop = sn_fit_logthresh,
          temp = sn_fit_temp
        )

        # 5. Transform parameters back to Theta space
        # Scaling Z-space (mean 0, sd 1) to marginal theta (mean theta*, sd sigma_jj)
        marginal_scale <- sqrt(Sigma_theta[j, j])
        fit_sn$xi <- theta_star[j] + fit_sn$xi * marginal_scale
        fit_sn$omega <- fit_sn$omega * marginal_scale

        unlist(fit_sn)
      }
    )
  )


tibble(
  x = seq(-4, 4, length = 100),
  truth = dnorm(x, 0, 1),
  sn_fit = INLAvaan::dsnorm(
    x,
    xi = approx_data["theta[1]", "xi"],
    omega = approx_data["theta[1]", "omega"],
    alpha = approx_data["theta[1]", "alpha"]
  )
) |>
  pivot_longer(
    cols = c("truth", "sn_fit"),
    names_to = "type",
    values_to = "density"
  ) |>
  ggplot(aes(
    x,
    density,
    color = type,
    fill = type,
    alpha = type,
    linewidth = type
  )) +
  geom_area(position = "identity") +
  geom_line(alpha = 1) +
  scale_alpha_manual(values = c(0, 0.3)) +
  scale_linewidth_manual(values = c(0.8, 0)) +
  scale_colour_manual(values = c("truth" = "black", "sn_fit" = "#00A6AA")) +
  scale_fill_manual(values = c("truth" = "black", "sn_fit" = "#00A6AA")) +
  labs(
    title = "Marginal of X: True vs Skew Normal Approximation",
    y = "Density"
  ) +
  theme_minimal()

## Validation: SEM model against MCMC (blavaan)
## Extracted from tests/testthat/test-sem.R
## Requires: future, blavaan

mod <- "
  ind60 =~ x1 + x2 + x3
  dem60 =~ y1 + y2 + y3 + y4
  dem65 =~ y5 + y6 + y7 + y8

  dem60 ~ ind60
  dem65 ~ ind60 + dem60

  y1 ~~ y5
  y2 ~~ y4 + y6
  y3 ~~ y7
  y4 ~~ y8
  y6 ~~ y8
"
dat <- lavaan::PoliticalDemocracy

testthat::skip_on_ci()
testthat::skip_on_cran()
testthat::skip_if_not(interactive())
testthat::skip_if_not_installed("future")
library(blavaan)
future::plan("multisession", workers = future::availableCores() - 2)

fit_blav <- bsem(
  mod,
  dat,
  bcontrol = list(cores = 3),
  burnin = 1000,
  sample = 2000
)


fit_inl1 <- asem(
  mod,
  dat,
  marginal_method = "skewnorm",
  debug = TRUE,
  test = "none"
)
fit_inl2 <- asem(
  mod,
  dat,
  marginal_method = "skewnorm",
  marginal_correction = "none",
  debug = TRUE,
  test = "none"
)
fit_inl3 <- asem(
  mod,
  dat,
  marginal_method = "skewnorm",
  # marginal_correction = "none",
  vb_correction = FALSE,
  debug = TRUE,
  test = "none"
)

res <- compare_mcmc(
  fit_blav,
  "skewnorm" = fit_inl1,
  # "sn_nocor" = fit_inl2,
  # "sn_novb" = fit_inl3,
  show_error = !FALSE
)
print(res$p_compare)
# print(res$p_errors)
# print(res$metrics_df, n = 1000)

## Validation: Growth model against MCMC (blavaan)
## Extracted from tests/testthat/test-growth.R
## Requires: future, blavaan

mod <- "
  # intercept and slope with fixed coefficients
  i =~ 1*t1 + 1*t2 + 1*t3 + 1*t4
  s =~ 0*t1 + 1*t2 + 2*t3 + 3*t4

  # regressions
  i ~ x1 + x2
  s ~ x1 + x2

  # time-varying covariates
  t1 ~ c1
  t2 ~ c2
  t3 ~ c3
  t4 ~ c4
"
dat <- lavaan::Demo.growth

testthat::skip_on_ci()
testthat::skip_on_cran()
testthat::skip_if_not(interactive())
testthat::skip_if_not_installed("future")
library(blavaan)
future::plan("multisession", workers = future::availableCores() - 2)

fit_blav <- bgrowth(
  mod,
  dat,
  bcontrol = list(cores = 3),
  burnin = 1000,
  sample = 2000
)
fit_inl1 <- agrowth(
  mod,
  dat,
  marginal_method = "skewnorm",
  debug = TRUE,
  test = "none"
)
fit_inl2 <- agrowth(
  mod,
  dat,
  marginal_method = "asymgaus",
  debug = TRUE,
  test = "none"
)
fit_inl3 <- agrowth(
  mod,
  dat,
  marginal_method = "sampling",
  debug = TRUE,
  test = "none"
)

res <- compare_mcmc(
  fit_blav,
  "skewnorm" = fit_inl1,
  "asymgaus" = fit_inl2,
  "sampling" = fit_inl3
)
print(res$p_compare)
print(res$p_errors)
print(res$metrics_df, n = 1000)

# Linear growth model with a time-varying covariate
mod <- "
  # Intercept and slope with fixed coefficients
    i =~ 1*t1 + 1*t2 + 1*t3 + 1*t4
    s =~ 0*t1 + 1*t2 + 2*t3 + 3*t4

  # (Latent) regressions
    i ~ x1 + x2
    s ~ x1 + x2

  # Time-varying covariates
    t1 ~ c1
    t2 ~ c2
    t3 ~ c3
    t4 ~ c4
"
utils::data("Demo.growth", package = "lavaan")
str(Demo.growth)

fit <- agrowth(mod, data = Demo.growth, nsamp = 100)
summary(fit)

# The Holzinger and Swineford (1939) example
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")

fit <- inlavaan(
  HS.model,
  data = HolzingerSwineford1939,
  auto.var = TRUE,
  auto.fix.first = TRUE,
  auto.cov.lv.x = TRUE
)
summary(fit)

\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, meanstructure = TRUE)

# Leave-one-subject-out (LOSO) from the single fit -- no refitting
res <- loo(fit)
res
head(res$per_unit)

# Score a submodel without refitting: condition the Laplace summary on the
# visual ~~ speed covariance being zero, then evaluate at that summary
int <- get_inlavaan_internal(fit)
theta <- int$theta_star
Sigma <- int$Sigma_theta
p <- which(names(coef(fit)) == "visual~~speed")
theta_c <- theta - Sigma[, p] * (theta[p] / Sigma[p, p])
Sigma_c <- Sigma - tcrossprod(Sigma[, p]) / Sigma[p, p]
loo(fit, theta = theta_c, Sigma = Sigma_c)

# Two-level models are scored per cluster (LOCO) automatically
utils::data("Demo.twolevel", package = "lavaan")
model2l <- "
  level: 1
    fw =~ y1 + y2 + y3
    fw ~ x1 + x2 + x3
  level: 2
    fb =~ y1 + y2 + y3
    fb ~ w1 + w2
"
fit2l <- asem(model2l, Demo.twolevel, cluster = "cluster",
              meanstructure = TRUE, fixed.x = FALSE)
loo(fit2l)
}

utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa("visual =~ x1 + x2 + x3", HolzingerSwineford1939)

# Posterior samples of lavaan-side parameters
samps <- sampling(fit, nsamp = 500)
head(samps)

# Compare copula vs Gaussian sampling
s_cop <- sampling(fit, nsamp = 500, samp_copula = TRUE)
s_gaus <- sampling(fit, nsamp = 500, samp_copula = FALSE)

# Prior predictive samples
y_prior <- sampling(fit, type = "observed", nsamp = 500, prior = TRUE)

# The industrialization and Political Democracy Example from Bollen (1989), page
# 332
mod <- "
  # Latent variable definitions
     ind60 =~ x1 + x2 + x3
     dem60 =~ y1 + y2 + y3 + y4
     dem65 =~ y5 + y6 + y7 + y8

  # (Latent) regressions
    dem60 ~ ind60
    dem65 ~ ind60 + dem60

  # Residual correlations
    y1 ~~ y5
    y2 ~~ y4 + y6
    y3 ~~ y7
    y4 ~~ y8
    y6 ~~ y8
"
fit <- asem(mod, lavaan::PoliticalDemocracy)

utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa("visual =~ x1 + x2 + x3", HolzingerSwineford1939)

# Simulate one replicate dataset from the posterior
sims <- simulate(fit, nsim = 1)
head(sims[[1]])                    # data frame
attr(sims[[1]], "truth")           # true lavaan-side (x-space) parameters
attr(sims[[1]], "truth_theta")     # corresponding unconstrained (theta-space) parameters

# Simulate from the prior (e.g., for SBC)
sims_prior <- simulate(fit, nsim = 5, prior = TRUE)
lapply(sims_prior, nrow)

## Validation: Binary/ordered CFA model against MCMC (blavaan)
## Extracted from tests/testthat/test-binary.R
## Requires: future, blavaan

# Simulate binary data
set.seed(141)
n <- 100
truval <- c(0.8, 0.7, 0.6, 0.5, 0.4, -1.43, -0.55, -0.13, -0.72, -1.13)
dat <- lavaan::simulateData(
  "eta =~ 0.8*y1 + 0.7*5y2 + 0.6*y3 + 0.5*y4 + 0.4*y5
   y1 | -1.43*t1
   y2 | -0.55*t1
   y3 | -0.13*t1
   y4 | -0.72*t1
   y5 | -1.13*t1",
  ordered = TRUE,
  sample.nobs = n
)
mod <- "eta  =~ y1 + y2 + y3 + y4 + y5"

testthat::skip_on_ci()
testthat::skip_on_cran()
testthat::skip_if_not(interactive())
testthat::skip_if_not_installed("future")
library(blavaan)
future::plan("multisession", workers = future::availableCores() - 2)

fit_blav <- bcfa(
  mod,
  dat,
  ordered = TRUE,
  std.lv = TRUE,
  # burnin = 500,
  # sample = 1000,
  # n.chains = 1
  bcontrol = list(cores = 3)
)
fit_inl1 <- acfa(mod, dat, ordered = TRUE, std.lv = TRUE, debug = TRUE)

# Compare
res <- compare_mcmc(fit_blav, INLAvaan = fit_inl1, truth = truval)
print(res$p_compare)

# The industrialization and Political Democracy Example from Bollen (1989), page
# 332
model <- "
  # Latent variable definitions
     ind60 =~ x1 + x2 + x3
     dem60 =~ y1 + a*y2 + b*y3 + c*y4
     dem65 =~ y5 + a*y6 + b*y7 + c*y8

  # (Latent) regressions
    dem60 ~ ind60
    dem65 ~ ind60 + dem60

  # Residual correlations
    y1 ~~ y5
    y2 ~~ y4 + y6
    y3 ~~ y7
    y4 ~~ y8
    y6 ~~ y8
"
utils::data("PoliticalDemocracy", package = "lavaan")

fit <- asem(model, PoliticalDemocracy, test = "none")
summary(fit)

## Validation: Equality constraints model against MCMC (blavaan)
## Extracted from tests/testthat/test-equality_constraints.R
## Requires: future, blavaan

mod <- "
  # intercept with coefficients fixed to 1
  i =~  1*Day0 + 1*Day1 + 1*Day2 + 1*Day3 + 1*Day4 +
        1*Day5 + 1*Day6 + 1*Day7 + 1*Day8 + 1*Day9

  # slope with coefficients fixed to 0:9 (number of days)
  s =~  0*Day0 + 1*Day1 + 2*Day2 + 3*Day3 + 4*Day4 +
        5*Day5 + 6*Day6 + 7*Day7 + 8*Day8 + 9*Day9

  i ~~ i
  i ~ 1

  s ~~ s
  s ~ 1

  i ~~ s

  # fix intercepts
  Day0 ~ 0*1
  Day1 ~ 0*1
  Day2 ~ 0*1
  Day3 ~ 0*1
  Day4 ~ 0*1
  Day5 ~ 0*1
  Day6 ~ 0*1
  Day7 ~ 0*1
  Day8 ~ 0*1
  Day9 ~ 0*1

  # apply equality constraints
  Day0 ~~ v*Day0
  Day1 ~~ v*Day1
  Day2 ~~ v*Day2
  Day3 ~~ v*Day3
  Day4 ~~ v*Day4
  Day5 ~~ v*Day5
  Day6 ~~ v*Day6
  Day7 ~~ v*Day7
  Day8 ~~ v*Day8
  Day9 ~~ v*Day9
  "
dat <- reshape(
  lme4::sleepstudy,
  timevar = "Days",
  idvar = "Subject",
  direction = "wide"
)
names(dat) <- sub("^Reaction\\.(.*)$", "Day\\1", names(dat))

testthat::skip_on_ci()
testthat::skip_on_cran()
testthat::skip_if_not(interactive())
testthat::skip_if_not_installed("future")
library(blavaan)
future::plan("multisession", workers = future::availableCores() - 2)

fit_blav <- bgrowth(
  mod,
  dat,
  bcontrol = list(cores = 3),
  burnin = 1000,
  sample = 2000
)
fit_inl1 <- agrowth(
  mod,
  dat,
  marginal_method = "skewnorm",
  debug = TRUE,
  test = "none"
)
fit_inl2 <- agrowth(
  mod,
  dat,
  marginal_method = "asymgaus",
  debug = TRUE,
  test = "none"
)
fit_inl3 <- agrowth(
  mod,
  dat,
  marginal_method = "sampling",
  debug = TRUE,
  test = "none"
)

res <- compare_mcmc(
  fit_blav,
  "skewnorm" = fit_inl1,
  "asymgaus" = fit_inl2,
  "sampling" = fit_inl3
)
print(res$p_compare)
print(res$p_errors)
print(res$metrics_df, n = 1000)

prepare_priors_for_optim <- function(pt) {
  # 1. Filter for relevant rows: Free parameters that are not duplicates
  # We only calculate gradients for unique free parameters.
  idx <- which(
    !is.na(pt$prior) & pt$prior != "" & pt$free > 0 & !duplicated(pt$free)
  )

  # Initialize storage vectors (length of RELEVANT priors only)
  n <- length(idx)

  # A. Transformation Codes (0=Identity, 1=Exp/Log, 2=Tanh/Atanh)
  # Derived from your partable_transform_funcs logic
  trans_type <- integer(n) # Default 0 (Identity)

  # Check the 'mat' column for transformation types
  # (theta_var/psi_var -> Exp; theta_cor/psi_cor -> Tanh)
  mat_vals <- pt$mat[idx]
  trans_type[grepl("theta_var|psi_var", mat_vals)] <- 1L
  trans_type[grepl("theta_cor|theta_cov|psi_cor|psi_cov", mat_vals)] <- 2L

  # B. Prior Codes and Hyperparameters
  prior_type <- integer(n) # 0=None, 1=Normal, 2=Gamma, 3=Beta
  p1 <- numeric(n) # mu / shape / a
  p2 <- numeric(n) # sd / rate / b
  is_sd_prior <- logical(n) # Special flag for [sd] qualifier
  is_prec_prior <- logical(n) # Special flag for [prec] qualifier

  raw_priors <- pt$prior[idx]

  # Detect [prec] qualifier (valid for any distribution)
  is_prec_prior <- grepl("\\[prec\\]", raw_priors)

  # Strip ALL qualifiers before numeric parsing
  clean_priors <- gsub("\\[sd\\]|\\[prec\\]", "", raw_priors)

  # Parse Normal
  is_norm <- grepl("normal", clean_priors)
  if (any(is_norm)) {
    prior_type[is_norm] <- 1L
    # Parse string "normal(0,1)" -> c(0, 1)
    # Using efficient vectorized parsing approach
    clean <- gsub("normal\\(|\\)", "", clean_priors[is_norm])
    splits <- strsplit(clean, ",")
    p1[is_norm] <- as.numeric(sapply(splits, `[`, 1))
    p2[is_norm] <- as.numeric(sapply(splits, `[`, 2))
  }

  # Parse Gamma
  is_gamma <- grepl("gamma", clean_priors)
  if (any(is_gamma)) {
    prior_type[is_gamma] <- 2L
    is_sd <- grepl("\\[sd\\]", raw_priors[is_gamma]) & !is_prec_prior[is_gamma]
    is_sd_prior[is_gamma] <- is_sd

    clean <- gsub("gamma\\(|\\)", "", clean_priors[is_gamma])
    splits <- strsplit(clean, ",")
    p1[is_gamma] <- as.numeric(sapply(splits, `[`, 1))
    p2[is_gamma] <- as.numeric(sapply(splits, `[`, 2))
  }

  # Parse Beta
  is_beta <- grepl("beta", clean_priors)
  if (any(is_beta)) {
    prior_type[is_beta] <- 3L
    clean <- gsub("beta\\(|\\)", "", clean_priors[is_beta])
    splits <- strsplit(clean, ",")
    p1[is_beta] <- as.numeric(sapply(splits, `[`, 1))
    p2[is_beta] <- as.numeric(sapply(splits, `[`, 2))
  }

  # Return a lightweight list optimized for the gradient function
  list(
    parname = pt$names[idx], # Just for names(theta)
    idx_in_pt = idx, # Indices in original pt
    free_id = pt$free[idx], # Which theta index this corresponds to
    trans_type = trans_type, # 0=Id, 1=Exp, 2=Tanh
    prior_type = prior_type, # 1=Norm, 2=Gam, 3=Beta
    p1 = p1,
    p2 = p2,
    is_sd_prior = is_sd_prior,
    is_prec_prior = is_prec_prior,
    prior_names = raw_priors # Just for names(grad)
  )
}

prior_logdens_vectorized <- function(theta, cache, debug = FALSE) {
  # 1. Map theta to relevant parameters (vectorized)
  th <- theta[cache$free_id]
  n <- length(th)

  # 2. Apply Transformations & Jacobian of Transformation
  # Initialize vectors
  xval <- th
  dx_dth <- rep(1, n)

  # A. Exp (Log-link) - e.g., Variances
  idx_exp <- which(cache$trans_type == 1L)
  if (length(idx_exp) > 0) {
    ex <- exp(th[idx_exp])
    xval[idx_exp] <- ex
    dx_dth[idx_exp] <- ex
  }

  # B. Tanh (Atanh-link) - e.g., Correlations
  idx_tanh <- which(cache$trans_type == 2L)
  if (length(idx_tanh) > 0) {
    # Replicating your safe_tanh logic roughly, or use standard tanh
    # Note: d/dx tanh(x) = 1 - tanh^2(x)
    eps <- 1e-6
    safe_scale <- (1 - eps)

    t_val <- tanh(th[idx_tanh])
    xval[idx_tanh] <- safe_scale * t_val
    dx_dth[idx_tanh] <- safe_scale * (1 - t_val^2)
  }

  # 3. Calculate Log-Densities (Vectorized by Type)
  lp <- numeric(n)

  # Type 1: Normal
  idx_norm <- which(cache$prior_type == 1L)
  if (length(idx_norm) > 0) {
    # dnorm is fully vectorized
    lp[idx_norm] <- dnorm(
      xval[idx_norm],
      mean = cache$p1[idx_norm],
      sd = cache$p2[idx_norm],
      log = TRUE
    )
  }

  # Type 2: Gamma
  idx_gam <- which(cache$prior_type == 2L)
  if (length(idx_gam) > 0) {
    # Separate standard Gamma vs Gamma[sd]
    # We can do this with masking to keep it vectorized
    sub_sd <- cache$is_sd_prior[idx_gam]

    # 1. Standard Gamma Logic (applied to all first)
    # Note: Gamma check for x > 0 is handled by dgamma returning -Inf/NaN if invalid
    lp[idx_gam] <- dgamma(
      xval[idx_gam],
      shape = cache$p1[idx_gam],
      rate = cache$p2[idx_gam],
      log = TRUE
    )

    # 2. Fix up Gamma[sd] cases
    if (any(sub_sd)) {
      # Indices of the [sd] items within the global vectors
      global_sd_idx <- idx_gam[sub_sd]

      # Get current xval (which is variance)
      vars <- xval[global_sd_idx]
      sds <- sqrt(vars)

      # Re-calculate density for s = sqrt(x)
      lp[global_sd_idx] <- dgamma(
        sds,
        shape = cache$p1[global_sd_idx],
        rate = cache$p2[global_sd_idx],
        log = TRUE
      )

      # Apply Jacobian correction for [sd]:
      # The gradient code divided by 2*sqrt(x).
      # Here we adjust the Jacobian term dx_dth directly.
      # Original dx_dth is d(xval)/d(theta).
      # We need d(sd)/d(theta) = d(sd)/d(xval) * d(xval)/d(theta)
      #                        = (1 / 2*sqrt(xval)) * dx_dth
      dx_dth[global_sd_idx] <- dx_dth[global_sd_idx] / (2 * sds)
    }
  }

  # Type 3: Beta (Boxed on -1, 1)
  idx_beta <- which(cache$prior_type == 3L)
  if (length(idx_beta) > 0) {
    # Inline dbeta_box logic for speed:
    # dbeta_box(x, a, b, min=-1, max=1) = dbeta((x - min)/(max-min), a, b) / (max-min)
    # Scaled to [0,1]: y = (x + 1) / 2
    # Jacobian of scale: 1/2

    xv <- xval[idx_beta]
    # Shift to [0,1]
    y <- (xv + 1) / 2

    # Calculate density
    # log(dbeta(y) * 1/2) = log(dbeta(y)) - log(2)
    val <- dbeta(
      y,
      shape1 = cache$p1[idx_beta],
      shape2 = cache$p2[idx_beta],
      log = TRUE
    ) -
      0.69314718 # log(2)

    lp[idx_beta] <- val
  }

  # 4. Handle [prec] qualifier: prior is on precision (1/x) rather than x
  # Re-evaluate lp at 1/xval and apply Jacobian |d(1/x)/dx| = 1/x^2
  idx_prec <- which(cache$is_prec_prior %||% logical(n))
  if (length(idx_prec) > 0) { # nocov start
    ip_norm  <- idx_prec[cache$prior_type[idx_prec] == 1L]
    ip_gamma <- idx_prec[cache$prior_type[idx_prec] == 2L]
    ip_beta  <- idx_prec[cache$prior_type[idx_prec] == 3L]

    if (length(ip_norm) > 0) {
      lp[ip_norm] <- dnorm(
        1 / xval[ip_norm], cache$p1[ip_norm], cache$p2[ip_norm], log = TRUE
      )
    }
    if (length(ip_gamma) > 0) {
      lp[ip_gamma] <- dgamma(
        1 / xval[ip_gamma], cache$p1[ip_gamma], cache$p2[ip_gamma], log = TRUE
      )
    }
    if (length(ip_beta) > 0) {
      y <- (1 / xval[ip_beta] + 1) / 2
      lp[ip_beta] <- dbeta(
        y, shape1 = cache$p1[ip_beta], shape2 = cache$p2[ip_beta], log = TRUE
      ) - 0.69314718
    }

    # Jacobian: |d(1/x)/dx| = 1/x^2 — multiply dx_dth by 1/x^2
    dx_dth[idx_prec] <- dx_dth[idx_prec] / (xval[idx_prec]^2)
  } # nocov end

  # 5. Final Summation
  # log|J| + log(dens)
  ljcb <- log(abs(dx_dth))

  if (debug) {
    # Reconstruct names for debugging if needed
    names(lp) <- cache$prior_names
    return(list(theta = theta, lp = lp, ljcb = ljcb))
  }

  total_log_dens <- sum(lp + ljcb)

  # Safety check (vectorized finite check is faster than is.finite on sum if sum is NA)
  if (!is.finite(total_log_dens)) {
    return(-1e40)
  }

  return(total_log_dens)
}

prior_grad_vectorized <- function(theta, cache) {
  # 'cache' is the object returned by prepare_priors_for_optim()

  # 1. Map theta to relevant parameters
  th <- theta[cache$free_id]
  n <- length(th)

  # 2. Apply Transformations (Vectorized)
  # Calculate xval, dx_dth, ddx_dth2 based on trans_type
  xval <- th # Default Identity
  dx_dth <- rep(1, n)
  ddx_dth2 <- rep(0, n)

  # Type 1: Exp (Log-link) -> for Variances
  idx_exp <- which(cache$trans_type == 1L)
  if (length(idx_exp) > 0) {
    ex <- exp(th[idx_exp])
    xval[idx_exp] <- ex
    dx_dth[idx_exp] <- ex
    ddx_dth2[idx_exp] <- ex
  }

  # Type 2: Tanh (Atanh-link) -> for Correlations
  idx_tanh <- which(cache$trans_type == 2L)
  if (length(idx_tanh) > 0) {
    # Using your safe_tanh logic: (1 - eps) * tanh(x)
    # Note: Your original code applied safe_tanh on the INVERSE (ginv).
    # Since theta is unrestricted, theta -> tanh(theta) maps R -> (-1, 1)

    # Calculate tanh(theta)
    th_t <- th[idx_tanh]
    t_val <- tanh(th_t)

    # Apply safe scaling (eps hardcoded here or passed in env)
    eps <- 1e-6
    safe_scale <- (1 - eps)

    xval[idx_tanh] <- safe_scale * t_val

    # Derivatives
    # d/dx tanh(x) = 1 - tanh^2(x)
    dt_dth <- 1 - t_val^2
    dx_dth[idx_tanh] <- safe_scale * dt_dth

    # Second derivative: -2 * tanh * (1-tanh^2) * scale
    # Note: Your original code had: -2 * safe_tanh(x) * (1 - safe_tanh(x)^2)
    # But strictly: d2/dth2 (scale*tanh) = scale * (-2*tanh * sech^2)
    # We will stick to the strict derivative of your ginv:
    ddx_dth2[idx_tanh] <- safe_scale * (-2 * t_val * dt_dth)
  }

  # 3. Calculate Prior Gradients (dlp_dx)
  dlp_dx <- numeric(n)
  jac_extra <- numeric(n) # Defaults to 0

  # A. Normal (Type 1)
  idx_norm <- which(cache$prior_type == 1L)
  if (length(idx_norm) > 0) {
    # -(x - mu) / sd^2
    dlp_dx[idx_norm] <- -(xval[idx_norm] - cache$p1[idx_norm]) /
      (cache$p2[idx_norm]^2)
  }

  # B. Gamma (Type 2)
  idx_gam <- which(cache$prior_type == 2L)
  if (length(idx_gam) > 0) {
    # Split into standard Gamma vs Gamma[sd]
    # We use logical masking within the gamma group
    sub_sd <- cache$is_sd_prior[idx_gam]

    # Parameters for all gammas
    shape <- cache$p1[idx_gam]
    rate <- cache$p2[idx_gam]
    xv <- xval[idx_gam]

    # Calculation for Standard Gamma
    # (shape - 1)/x - rate
    # We calculate this for everyone first
    res_gam <- (shape - 1) / xv - rate

    # Correction for Gamma[sd]
    if (any(sub_sd)) {
      # Indices within the gamma subset that are [sd]
      # Re-implementing:
      # s = sqrt(x); dlp_ds = (shape-1)/s - rate; dlp_dx = dlp_ds / (2s)
      # jac_extra = -dx_dth / (2x)

      # We just overwrite the values for the [sd] rows
      xv_sd <- xv[sub_sd]
      s_val <- sqrt(xv_sd)
      dlp_ds <- (shape[sub_sd] - 1) / s_val - rate[sub_sd]

      # Update dlp_dx for these specific rows
      res_gam[sub_sd] <- dlp_ds / (2 * s_val)

      # Add Jacobian extra
      # We need the dx_dth specific to these rows
      dx_dth_gam <- dx_dth[idx_gam]
      jac_extra[idx_gam[sub_sd]] <- -dx_dth_gam[sub_sd] / (2 * xv_sd)
    }

    dlp_dx[idx_gam] <- res_gam
  }

  # C. Beta (Type 3)
  idx_beta <- which(cache$prior_type == 3L)
  if (length(idx_beta) > 0) {
    a <- cache$p1[idx_beta]
    b <- cache$p2[idx_beta]
    xv <- xval[idx_beta]
    # d/dx log p_X(x) = d/dy log dbeta(y,a,b) * dy/dx
    #   = [(a-1)/y - (b-1)/(1-y)] * 1/2, with y = (x+1)/2
    #   = [(a-1)/((x+1)/2) - (b-1)/((1-x)/2)] * 1/2
    #   = (a-1)/(x+1) - (b-1)/(1-x)
    dlp_dx[idx_beta] <- (a - 1) / (xv + 1) - (b - 1) / (1 - xv)
  }

  # Handle [prec] qualifier: prior is on precision (1/x)
  # Chain rule: dlp_dx = (d log p(tau) / d tau)|_{tau=1/x} * (-1/x^2)
  # Jacobian extra: d/d(theta) log|d(1/x)/dx| = -2 * x'(theta) / x
  idx_prec <- which(cache$is_prec_prior %||% logical(n))
  if (length(idx_prec) > 0) { # nocov start
    pv <- 1 / xval[idx_prec] # precision values

    ip_norm  <- which(cache$prior_type[idx_prec] == 1L)
    ip_gamma <- which(cache$prior_type[idx_prec] == 2L)
    ip_beta  <- which(cache$prior_type[idx_prec] == 3L)

    raw_dlp_dprec <- numeric(length(idx_prec))

    if (length(ip_norm) > 0) {
      raw_dlp_dprec[ip_norm] <- -(pv[ip_norm] - cache$p1[idx_prec[ip_norm]]) /
        (cache$p2[idx_prec[ip_norm]]^2)
    }
    if (length(ip_gamma) > 0) {
      raw_dlp_dprec[ip_gamma] <- (cache$p1[idx_prec[ip_gamma]] - 1) / pv[ip_gamma] -
        cache$p2[idx_prec[ip_gamma]]
    }
    if (length(ip_beta) > 0) {
      pv_b <- pv[ip_beta]
      a_b <- cache$p1[idx_prec[ip_beta]]
      b_b <- cache$p2[idx_prec[ip_beta]]
      raw_dlp_dprec[ip_beta] <- (a_b - 1) / (pv_b + 1) - (b_b - 1) / (1 - pv_b)
    }

    # Chain: dlp_dx = dlp_dprec * d(1/x)/dx = dlp_dprec * (-1/x^2)
    dlp_dx[idx_prec] <- raw_dlp_dprec * (-1 / xval[idx_prec]^2)

    # Jacobian: d/d(theta) log(1/x^2) = -2 * x'/x
    jac_extra[idx_prec] <- -2 * dx_dth[idx_prec] / xval[idx_prec]
  } # nocov end

  # 4. Chain Rule Assembly
  grad <- dlp_dx * dx_dth + ddx_dth2 / dx_dth + jac_extra

  names(grad) <- cache$prior_names
  return(grad)
}

dat <- lavaan::HolzingerSwineford1939
mod <- "
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
    speed   =~ x7 + x8 + x9
  "
NSAMP <- 3

test_that("Setting dp argument", {
  DP <- priors_for(theta = "gamma(1,1)", psi = "gamma(1,1)")
  fit <- asem(
    mod,
    dat,
    dp = DP,
    verbose = FALSE,
    nsamp = NSAMP
  )
  pt <- lavaan::partable(fit)
  where_theta_var <- grep("theta_var", pt$mat)
  expect_true(all(pt$prior[where_theta_var] == "gamma(1,1)"))
  where_psi_var <- grep("psi_var", pt$mat)
  expect_true(all(pt$prior[where_psi_var] == "gamma(1,1)"))
})

test_that("Straight from textual model", {
  mod <- "
    visual  =~ x1 + prior('normal(1,2)')*x2 + x3
    textual =~ x4 + x5 + prior('normal(3,1.5)')*x6
    speed   =~ x7 + x8 + x9

    x1 ~~ prior('gamma(3,3)')*x1
  "

  expect_no_error({
    fit <- asem(
      mod,
      dat,
      verbose = FALSE,
      nsamp = NSAMP
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  pt <- lavaan::partable(fit)
  expect_true(pt$prior[grep("visual=~x2", pt$names)] == "normal(1,2)")
  expect_true(pt$prior[grep("textual=~x6", pt$names)] == "normal(3,1.5)")
  expect_true(pt$prior[grep("x1~~x1", pt$names)] == "gamma(3,3)")

  fit_inlv <- get_inlavaan_internal(fit)
  pt <- fit_inlv$partable
  idx <- which(
    pt$names %in% c("visual=~x2", "textual=~x6", "x1~~x1", "textual~~textual")
  )
  short_pt <- lapply(pt, function(x) x[idx])
  short_pt$free <- 1:4

  res <- prior_logdens_vectorized(c(2, 4, 1, 1), prepare_priors_for_optim(short_pt), debug = TRUE)

  # Check log densities
  expect_equal(
    as.numeric(res$lp),
    c(
      dnorm(2, 1, 2, log = TRUE),
      dnorm(4, 3, 1.5, log = TRUE),
      dgamma(exp(1), 3, 3, log = TRUE),
      dgamma(exp(1 / 2), 1, 0.5, log = TRUE)
    )
  )

  # Check jacobians
  expect_equal(
    as.numeric(res$ljcb),
    c(
      0, # log(abs(1))
      0, # log(abs(1))
      log(abs(exp(1))), # t = log x  ---> log(abs(d/dt e^t))
      log(abs(0.5 * exp(1 / 2))) # t = log(sqrt(x)) ---> log(abs(0.5 exp^{t/2}))
    )
  )
})

## ---- TESTS FOR NEW VECTORISED LOG PRIOR AND GRADIENT ------------------------

get_test_objects <- function(model, data, group = NULL, group.equal = "") {
  # 1. Fit standard lavaan to get data structures
  lav_fit <- lavaan::cfa(
    model,
    data = data,
    group = group,
    group.equal = group.equal,
    se = "none", # Speed up
    do.fit = FALSE # We only need the setup
  )

  # 2. Run inlavaanify (assuming it is exported or available in namespace)
  pt <- inlavaanify_partable(
    pt = lavaan::parTable(lav_fit),
    lavdata = lav_fit@Data,
    lavoptions = lav_fit@Options
  )

  # 3. Create the Cache
  cache <- prepare_priors_for_optim(pt)

  # 4. Generate Random Theta (Validation Vector)
  # Size is the maximum integer in the 'free' column
  n_theta <- max(pt$free, na.rm = TRUE)
  set.seed(42) # Reproducibility
  theta <- rnorm(n_theta)

  list(pt = pt, cache = cache, theta = theta)
}

test_that("Vectorized vs Old: Standard Single-Group CFA", {
  # Setup
  dat <- lavaan::HolzingerSwineford1939
  mod <- "
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
  "
  obj <- get_test_objects(mod, dat)

  # --- LOG DENSITY TEST ---
  new_dens <- prior_logdens_vectorized(obj$theta, obj$cache)
  expect_true(is.finite(new_dens), info = "Log-density should be finite")

  # --- GRADIENT TEST (against numerical finite differences) ---
  new_grad <- prior_grad_vectorized(obj$theta, obj$cache)
  h <- 1e-5
  num_grad <- sapply(seq_along(obj$theta), function(i) {
    tp <- obj$theta; tp[i] <- tp[i] + h
    tm <- obj$theta; tm[i] <- tm[i] - h
    (prior_logdens_vectorized(tp, obj$cache) - prior_logdens_vectorized(tm, obj$cache)) / (2 * h)
  })
  # Accumulate cache-indexed grad back to full theta-length vector
  full_grad <- numeric(length(obj$theta))
  full_grad[obj$cache$free_id] <- new_grad

  expect_equal(
    full_grad,
    num_grad,
    tolerance = 1e-4,
    info = "Gradient should match finite differences in single-group models"
  )
})

test_that("Vectorized vs Old: Multi-Group with Equality Constraints", {
  # This is the critical test for the indexing bug
  dat <- lavaan::HolzingerSwineford1939
  mod <- "
    visual  =~ x1 + x2
    textual =~ x4 + x5
  "

  # Create object with equality constraints
  obj <- get_test_objects(
    mod,
    dat,
    group = "school",
    group.equal = "loadings"
  )

  # --- LOG DENSITY TEST ---
  new_dens <- prior_logdens_vectorized(obj$theta, obj$cache)
  expect_true(is.finite(new_dens), info = "Log-density should be finite")

  # --- GRADIENT TEST (against numerical finite differences) ---
  new_grad <- prior_grad_vectorized(obj$theta, obj$cache)
  h <- 1e-5
  num_grad <- sapply(seq_along(obj$theta), function(i) {
    tp <- obj$theta; tp[i] <- tp[i] + h
    tm <- obj$theta; tm[i] <- tm[i] - h
    (prior_logdens_vectorized(tp, obj$cache) - prior_logdens_vectorized(tm, obj$cache)) / (2 * h)
  })
  # Accumulate cache-indexed grad back to full theta-length vector
  full_grad <- numeric(length(obj$theta))
  full_grad[obj$cache$free_id] <- new_grad

  expect_equal(
    full_grad,
    num_grad,
    tolerance = 1e-4,
    info = "Gradient should match finite differences even with equality constraints"
  )
})

test_that("Vectorized version handles different Prior Types correctly", {
  # Mock a pt with Normal, Gamma, and Beta to ensure coverage
  # We construct a fake cache manually to test the math isolated from the parser

  # 3 params: 1=Normal, 2=Gamma[sd], 3=Beta
  theta <- c(0.5, 0.5, 0.0) # Arbitrary values

  cache <- list(
    free_id = c(1L, 2L, 3L),
    # Exp, Identity, Identity
    trans_type = c(1L, 0L, 0L),
    # Normal, Gamma, Beta
    prior_type = c(1L, 2L, 3L),
    p1 = c(0, 1, 1), # mu, shape, alpha
    p2 = c(1, 0.5, 1), # sd, rate, beta
    is_sd_prior = c(FALSE, TRUE, FALSE),
    prior_names = c("norm", "gam", "beta")
  )

  # Calculate New
  res_grad <- prior_grad_vectorized(theta, cache)

  # We expect a result vector of length 3
  expect_equal(length(res_grad), 3)
  expect_true(all(is.finite(res_grad)))
})

test_that("Beta prior gradient with tanh transform matches numerical", {
  # Regression test: non-trivial Beta(5,5) on correlation with tanh transform
  # Previously the gradient was off by a factor of 2
  cache <- list(
    free_id = 1L,
    trans_type = 2L,   # tanh (correlation)
    prior_type = 3L,   # beta
    p1 = 5,            # beta(5,5)
    p2 = 5,
    is_sd_prior = FALSE,
    prior_names = "rho"
  )

  h <- 1e-7
  for (th in c(-0.8, -0.3, 0.0, 0.3, 0.8)) {
    ag <- as.numeric(prior_grad_vectorized(th, cache))
    ng <- (prior_logdens_vectorized(th + h, cache) -
           prior_logdens_vectorized(th - h, cache)) / (2 * h)
    expect_equal(ag, ng, tolerance = 1e-5,
      info = paste0("Beta(5,5) tanh gradient at theta=", th))
  }

  # Also test asymmetric Beta(2,8)
  cache$p1 <- 2
  cache$p2 <- 8
  for (th in c(-0.5, 0.0, 0.5)) {
    ag <- as.numeric(prior_grad_vectorized(th, cache))
    ng <- (prior_logdens_vectorized(th + h, cache) -
           prior_logdens_vectorized(th - h, cache)) / (2 * h)
    expect_equal(ag, ng, tolerance = 1e-5,
      info = paste0("Beta(2,8) tanh gradient at theta=", th))
  }
})

test_that("dbeta_box with log = TRUE returns log-density", {
  x <- seq(0, 100, length.out = 20)
  ld  <- dbeta_box(x, shape1 = 2, shape2 = 5, a = 0, b = 100, log = TRUE)
  nld <- dbeta_box(x, shape1 = 2, shape2 = 5, a = 0, b = 100, log = FALSE)
  expect_equal(ld, log(nld))
  expect_true(all(is.finite(ld[x > 0 & x < 100])))
})

test_that("dbeta_box stops when b <= a", {
  expect_error(dbeta_box(0.5, shape1 = 1, shape2 = 1, a = 1, b = 0),
               "Require finite scalars with b > a")
  expect_error(dbeta_box(0.5, shape1 = 1, shape2 = 1, a = 0, b = 0),
               "Require finite scalars with b > a")
})

set.seed(1)
gen_data_twofac <- function(n) {
  mod <- "
    fx =~ 1*x1 + 0.7*x2 + 0.6*x3
    fy =~ 1*y1 + 0.7*y2 + 0.6*y3
    fy ~ 0.25*fx

    x1 ~~ 0.25*x1
    x2 ~~ 0.1225*x2
    x3 ~~ 0.09*x3

    y1 ~~ 0.25*y1
    y2 ~~ 0.1225*y2
    y3 ~~ 0.09*y3
  "
  lavaan::simulateData(model = mod, sample.nobs = n)
}

dat <- gen_data_twofac(250)
mod <- "
    fx =~ x1 + a*x2 + b*x3
    fy =~ y1 + a*y2 + b*y3
    x1 ~~ c*x1
    y1 ~~ c*y1
    fy ~ fx
  "

# meanstructure = TRUE so both sides maximise the same likelihood (without
# one, INLAvaan marginalises the saturated means and its mode shifts from
# the ML estimates by the n/(n-1) recalibration)
fit_lav <- lavaan::sem(mod, dat, meanstructure = TRUE)
fit_inl <- get_inlavaan_internal(asem(
  mod,
  dat,
  meanstructure = TRUE,
  add_priors = FALSE,
  vb_correction = FALSE,
  test = "none",
  verbose = FALSE
))
pt <- fit_inl$partable


test_that("INLAvaan produces same results as lavaan", {
  expect_equal(
    as.numeric(coef(fit_lav)),
    fit_inl$theta_star_trans[pt$free[pt$free > 0]],
    tolerance = 1e-4
  )
  expect_no_error(out <- capture.output(summary(fit_inl)))
})

test_that("Parameter Transformations (pars_to_x) work correctly", {
  m_unpacked <- sum(pt$free > 0)
  n_free <- sum(pt$free > 0 & !duplicated(pt$free))
  theta <- rnorm(m_unpacked)
  x_out <- pars_to_x(theta, pt)

  # 1. Check dimensions (should return packed vector)
  expect_equal(length(x_out), n_free)

  ptidxprior <- which(!is.na(pt$prior))
  thidxprior <- which(!duplicated(pt$free[pt$free > 0]))

  # 2. Check Variance Transformation (Log -> Exp)
  ptidxvar <- which(grepl("theta_var|psi_var", pt$mat))
  xidxvar <- pt$free[ptidxvar]
  xidxvar <- xidxvar[!duplicated(xidxvar)]
  x_val <- x_out[xidxvar]

  thidxvar <- pt$free[pt$free > 0]
  thidxvar <- thidxvar %in% xidxvar & !duplicated(thidxvar)
  th_val <- theta[thidxvar]

  expect_equal(log(x_val), th_val, tolerance = 1e-8)
})

test_that("Prior Densities match manual calculation", {
  n_free <- sum(pt$free > 0 & !duplicated(pt$free))
  theta <- rep(0.5, n_free)

  # --- Manual Verification for specific params ---
  # 1. Normal Prior on loading (id 2, free=1) -> "normal(0,10)"
  # theta = 0.5. Identity transform.
  lp_normal <- dnorm(0.5, mean = 0, sd = 10, log = TRUE)

  # 2. Gamma[sd] on variance (id 8, free=5) -> "gamma(1,.5)[sd]"
  # theta = 0.5. Transform: x = exp(0.5) ≈ 1.648.
  # Prior is on SD: s = sqrt(x) ≈ 1.284
  # No Jacobian adjustment needed
  # xval = exp(0.5)
  # dx_dth = exp(0.5)
  # lp = dgamma(sqrt(xval), 1, 0.5, log=TRUE)
  # ljcb = log(abs(dx_dth / (2*sqrt(xval))))  <-- From your prior_logdens code
  val_x <- exp(0.5)
  val_s <- sqrt(val_x)
  lp_gamma <- dgamma(val_s, shape = 1, rate = 0.5, log = TRUE)
  jac_gamma <- log(abs(exp(0.5) / (2 * sqrt(exp(0.5)))))

  # Sum manual components (assuming all others are 0 or ignored for this
  # specific test logic)
  pt_mini <- pt
  pt_mini$prior[-c(2)] <- NA # Keep only one normal prior

  lp_mini_pkg <- prior_logdens_vectorized(theta, prepare_priors_for_optim(pt_mini))

  expect_equal(lp_mini_pkg, lp_normal, tolerance = 1e-6)
})

test_that("Gradients are correct (Finite Difference Check)", {
  suppressMessages(
    tmp <- capture.output(fit <- asem(mod, dat, test = "none", debug = TRUE))
  )
  test_df <- read.table(text = tmp, skip = 1)[, -1]
  colnames(test_df) <- c("fd", "analytic", "diff")

  expect_equal(
    as.numeric(test_df$fd),
    as.numeric(test_df$diff),
    tolerance = 1e-2
  )
})

.onLoad <- function(libname, pkgname) {
  # Resolve unexported lavaan internals once per session; see
  # R/lavaan-unexported.R for why this cannot happen at install time.
  resolve_lavaan_internals(asNamespace(pkgname))

  # Resolve which spelling the installed lavaan uses for every renamed
  # argument INLAvaan sets, once per session; see R/lavaan-argnames.R for why
  # this cannot happen at install time either. Must precede the fitMeasures
  # registration below, which consumes the resolved map.
  resolve_lavaan_argnames(asNamespace(pkgname))

  # Register fitMeasures()/fitmeasures() against whichever lavaan generic is
  # active this session; see R/method-fitmeasures.R for why this cannot
  # happen at install time either.
  register_fitmeasures_methods(asNamespace(pkgname))
}

# nocov start
# .onLoad_qrng_warmup <- function(libname, pkgname) {
#   # This warms up the qrng/spacefillr connection quietly
#   if (requireNamespace("qrng", quietly = TRUE)) {
#     suppressPackageStartupMessages({
#       loadNamespace("qrng")
#       if (requireNamespace("spacefillr", quietly = TRUE)) {
#         loadNamespace("spacefillr")
#       }
#     })
#   }
#
#   invisible()
# }
# nocov end

# Gauss-Hermite quadrature nodes and weights
.gauss_hermite <- function(n) {
  i <- seq_len(n - 1)
  b <- sqrt(i / 2)
  cm <- diag(0, n)
  cm[cbind(i, i + 1)] <- b
  cm[cbind(i + 1, i)] <- b
  ev <- eigen(cm, symmetric = TRUE)
  ord <- order(ev$values)
  list(nodes = ev$values[ord], weights = sqrt(pi) * ev$vectors[1, ord]^2)
}

# Adjust the Laplace correlation matrix R so that the SN copula samples
# have the correct (Laplace) Pearson correlations after the qsn transform.
norta_adjust_R <- function(R, approx_data, use_spline = TRUE) {
  m <- nrow(R)
  alphas <- approx_data[, "alpha"]

  # Skip entirely if all marginals are near-Gaussian
  if (all(abs(alphas) < 0.01)) {
    return(R)
  }

  # Gauss-Hermite setup (9 nodes: 81-point 2D rule, plenty for tol 1e-6)
  gh <- .gauss_hermite(9)
  n <- length(gh$nodes)
  z1 <- sqrt(2) * gh$nodes # n-vector of GH z-nodes
  ww <- outer(gh$weights, gh$weights) # n x n weight matrix

  ## ---- Precompute per-marginal quantile functions + standardised moments ----
  eps <- 1e-10
  u1 <- pmax(pmin(pnorm(z1), 1 - eps), eps)

  q_funs <- vector("list", m) # quantile function per marginal
  q1_all <- matrix(NA_real_, m, n)
  mu_sn <- sd_sn <- numeric(m)

  for (j in seq_len(m)) {
    delta <- alphas[j] / sqrt(1 + alphas[j]^2)
    mu_sn[j] <- delta * sqrt(2 / pi)
    sd_sn[j] <- sqrt(1 - 2 * delta^2 / pi)

    if (abs(alphas[j]) >= 0.01) {
      aj <- alphas[j]
      if (use_spline) {
        u_grid <- seq(eps, 1 - eps, length.out = 2000)
        q_vals <- qsnorm_fast(u_grid, xi = 0, omega = 1, alpha = aj)
        q_funs[[j]] <- splinefun(u_grid, q_vals, method = "monoH.FC")
      } else {
        q_funs[[j]] <- function(u, .a = aj) {
          qsnorm_fast(u, xi = 0, omega = 1, alpha = .a)
        }
      }
      q1_all[j, ] <- q_funs[[j]](u1)
    } else {
      q1_all[j, ] <- qnorm(u1)
    }
  }

  ## ---- Pairwise root-finding ------------------------------------------------
  R_star <- R

  for (j in seq_len(m - 1)) {
    q1_j <- q1_all[j, ] # precomputed n-vector
    q1_j_mat <- matrix(q1_j, n, n) # q1_j_mat[i, l] = q1_j[i]

    for (k in (j + 1):m) {
      rho_target <- R[j, k]

      # Skip near-zero correlations or near-Gaussian pair
      if (abs(rho_target) < 1e-10) {
        next
      }
      if (abs(alphas[j]) < 0.01 && abs(alphas[k]) < 0.01) {
        next
      }

      qfk <- if (!is.null(q_funs[[k]])) q_funs[[k]] else qnorm

      f <- function(r_star) {
        s <- sqrt(max(0, 1 - r_star^2))
        z2_mat <- outer(r_star * z1, s * z1, "+") # n x n
        u2 <- pnorm(as.vector(z2_mat))
        u2 <- pmax(pmin(u2, 1 - eps), eps)
        q2_mat <- matrix(qfk(u2), n, n)
        cross <- sum(ww * q1_j_mat * q2_mat) / pi
        (cross - mu_sn[j] * mu_sn[k]) / (sd_sn[j] * sd_sn[k]) - rho_target
      }

      sol <- tryCatch(
        uniroot(f, interval = c(-0.999, 0.999), tol = 1e-6),
        error = function(e) NULL
      )

      if (!is.null(sol)) {
        R_star[j, k] <- sol$root
        R_star[k, j] <- sol$root
      }
    }
  }

  ## ---- Project to nearest positive-definite correlation matrix if needed ----
  ev <- eigen(R_star, symmetric = TRUE, only.values = TRUE)
  if (any(ev$values < 1e-8)) { # nocov start
    ev_full <- eigen(R_star, symmetric = TRUE)
    ev_full$values <- pmax(ev_full$values, 1e-8)
    R_star <- ev_full$vectors %*% diag(ev_full$values) %*% t(ev_full$vectors)
    R_star <- cov2cor(R_star)
  } # nocov end

  R_star
}

test_that("norta_adjust_R returns R unchanged when all marginals near-Gaussian", {
  R <- matrix(c(1, 0.5, 0.5, 1), 2, 2)
  # all alphas near zero -> early return
  approx_data <- matrix(c(0, 0, 0, 0), nrow = 2,
                        dimnames = list(NULL, c("alpha", "other")))
  result <- INLAvaan:::norta_adjust_R(R, approx_data)
  expect_equal(result, R)
})

test_that("norta_adjust_R handles near-zero off-diagonal correlation", {
  R <- matrix(c(1, 1e-12, 1e-12, 1), 2, 2)
  approx_data <- matrix(c(2, -2, 0, 0), nrow = 2,
                        dimnames = list(NULL, c("alpha", "other")))
  result <- INLAvaan:::norta_adjust_R(R, approx_data)
  expect_equal(dim(result), c(2, 2))
  expect_equal(diag(result), c(1, 1))
})

test_that("norta_adjust_R with use_spline = FALSE produces a valid matrix", {
  R <- matrix(c(1, 0.5, 0.5, 1), 2, 2)
  approx_data <- matrix(c(1.5, -1.5, 0, 0), nrow = 2,
                        dimnames = list(NULL, c("alpha", "other")))
  result <- INLAvaan:::norta_adjust_R(R, approx_data, use_spline = FALSE)
  expect_equal(dim(result), c(2, 2))
  expect_equal(diag(result), c(1, 1))
  # symmetric
  expect_equal(result[1, 2], result[2, 1])
})

test_that("norta_adjust_R skips near-Gaussian marginal in one pair", {
  # j=1 has alpha=0 (near-Gaussian), k=2 has large alpha -> pair is skipped
  R <- matrix(c(1, 0.6, 0.6, 1), 2, 2)
  approx_data <- matrix(c(0.005, 2.0, 0, 0), nrow = 2,
                        dimnames = list(NULL, c("alpha", "other")))
  result <- INLAvaan:::norta_adjust_R(R, approx_data)
  expect_equal(dim(result), c(2, 2))
})

#' @importFrom stats anova
#' @name anova
#' @rdname compare
#' @aliases anova,INLAvaan-method
#' @export
setMethod("anova", "INLAvaan", function(object, ...) {
  cli_abort(
    "{.fn anova} performs a classical likelihood-ratio test, which has no
     direct Bayesian analogue under the Laplace/posterior-mean approximation
     used here. Use {.fn compare} instead, which reports Bayes factors,
     DIC/pD, and (optionally) LOO/WAIC comparisons."
  )
})

sobol_owen <- function(n, d) {
  N <- nrow(SobolOwen)
  D <- ncol(SobolOwen)

  # If within stored table dimensions, use it directly
  if (n <= N && d <= D) {
    return(SobolOwen[seq_len(n), seq_len(d), drop = FALSE])
  }

  # Otherwise, fall back to qrng if available
  # nocov start
  if (requireNamespace("qrng", quietly = TRUE)) {
    return(qrng::sobol(n = n, d = d, randomize = "Owen"))
  }

  # No fallback available — error with guidance
  cli_abort(c(
    "Requested Sobol sequence ({n} x {d}) exceeds stored table ({N} x {D}).",
    "i" = "Install the {.pkg qrng} package to generate larger sequences on the fly."
  ))
  # nocov end
}

test_that("sobol_owen returns correct dimensions from stored table", {
  res <- sobol_owen(10, 5)
  expect_true(is.matrix(res))
  expect_equal(dim(res), c(10, 5))
})

test_that("sobol_owen returns values in [0, 1]", {
  res <- sobol_owen(50, 10)
  expect_true(all(res >= 0 & res <= 1))
})

test_that("sobol_owen with d = 1 returns a matrix (not a vector)", {
  res <- sobol_owen(5, 1)
  expect_true(is.matrix(res))
  expect_equal(ncol(res), 1)
})

test_that("sobol_owen boundary: full stored table dimensions work", {
  res <- sobol_owen(100, 300)
  expect_equal(dim(res), c(100, 300))
})

# test_that("sobol_owen errors when exceeding stored table and qrng unavailable", {
#   skip_if(requireNamespace("qrng", quietly = TRUE),
#           "qrng is installed; cannot test the error path")
#   expect_error(sobol_owen(200, 5))
# })
#
# test_that("sobol_owen falls back to qrng for large requests", {
#   skip_if_not_installed("qrng")
#   res <- sobol_owen(200, 5)
#   expect_true(is.matrix(res))
#   expect_equal(dim(res), c(200, 5))
# })

#' Class For Representing a (Fitted) Latent Variable Model
#'
#' This is a class that extends the [lavaan-class] class. Several S4 methods are
#' available.
#'
#' @slot external A list containing an `inlavaan_internal` object.
#'
#' @param object An object of class \code{INLAvaan}.
#' @param ... Additional arguments passed to methods.
#'
#' @seealso [lavaan-class], [inlavaan()], [acfa()], [asem()], [agrowth()]
#'
#' @examples
#' \donttest{
#' HS.model <- "
#'   visual  =~ x1 + x2 + x3
#'   textual =~ x4 + x5 + x6
#'   speed   =~ x7 + x8 + x9
#' "
#' utils::data("HolzingerSwineford1939", package = "lavaan")
#' fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
#'             test = "none", verbose = FALSE)
#'
#' # Print basic info
#' fit
#'
#' # Detailed summary
#' summary(fit)
#'
#' # Extract coefficients
#' coef(fit)
#' }
#'
#' @name INLAvaan-class
#' @rdname INLAvaan-class
#'
#' @importClassesFrom lavaan lavaan
#' @export
setClass(
  Class = "INLAvaan",
  contains = "lavaan"
)

#' @exportS3Method summary inlavaan_internal
summary.inlavaan_internal <- function(object, ...) {
  structure(
    list(summary = object$summary),
    class = "summary.inlavaan_internal"
  )
}

#' @exportS3Method print summary.inlavaan_internal
print.summary.inlavaan_internal <- function(x, digits = 3, ...) {
  summ <- x$summary
  which_numeric <- sapply(summ, is.numeric)
  summ[, which_numeric] <- round(summ[, which_numeric], digits)
  print(summ)
  invisible(x)
}

# Match rows of one parameter table (e.g. parameterEstimates() output) to
# rows of another (e.g. standardisedsolution() output) by lhs/op/rhs and
# group, because the two need not align row-for-row: parameterEstimates()
# keeps user equality constraints and r-square rows that
# standardizedSolution() drops, and the latter keeps system equality
# constraints the former drops. Duplicate keys (the same parameter in
# different blocks of a multilevel model, where lavaan omits the block
# column) are matched by order of occurrence, both tables being in partable
# order. Rows with no counterpart get NA.
match_partable_rows <- function(a, b) {
  key <- function(df) {
    k <- paste(df$lhs, df$op, df$rhs, if (is.null(df$group)) 1L else df$group)
    paste(k, stats::ave(seq_along(k), k, FUN = seq_along))
  }
  match(key(a), key(b))
}

# Format a numeric column for the estimates table: nd decimal places, blanks
# for missing entries.
format_pe_col <- function(x, nd) {
  out <- rep("", length(x))
  ok <- !is.na(x)
  out[ok] <- formatC(x[ok], digits = nd, format = "f")
  out
}

summary_inlavaan <- function(
  object,
  header = TRUE,
  fit.measures = TRUE,
  estimates = TRUE,
  ci = TRUE,
  standardized = FALSE,
  standardised = standardized,
  rsquare = FALSE,
  postmedian = FALSE,
  postmode = FALSE,
  nmad = TRUE,
  kld = FALSE,
  vb_shift = FALSE,
  priors = TRUE,
  nd = 3L,
  ...
) {
  vb_correction <- all(!is.na(object@external$inlavaan_internal$vb$correction))

  ## ----- Header --------------------------------------------------------------
  if (isTRUE(header)) {
    show_inlavaan(object)
  }

  ## ----- Fit measures --------------------------------------------------------
  if (isTRUE(fit.measures) & length(object@Fit@test$ppp) > 0) {
    # DIC
    dic_list <- object@external$inlavaan_internal$DIC
    cat(
      "\nInformation Criteria:\n\n",
      sprintf("  %-38s", "Deviance (DIC)"),
      sprintf("  %10.3f", dic_list$dic),
      "\n",
      sprintf("  %-38s", "Effective parameters (pD)"),
      sprintf("  %10.3f", dic_list$pD),
      "\n"
    )
  }

  # Everything below prints the parameter estimates table
  if (!isTRUE(estimates)) {
    return(invisible(NULL))
  }

  marg_method <- object@external$inlavaan_internal$marginal_method
  # if (marg_method == "skewnorm")
  #   marg_method <- "Skew Normal"
  # else if (marg_method == "asymgaus")
  #   marg_method <- "Two-piece Gaussian"
  # else if (marg_method == "marggaus")
  #   marg_method <- "Marginal Gaussian"
  # else if (marg_method == "sampling")
  #   marg_method <- "Sampling"

  PE <- call_lavaan(
    "parameterEstimates",
    object,
    se = FALSE, # create our own
    zstat = FALSE,
    ci = TRUE,
    standardized = FALSE,
    rsquare = rsquare,
    remove.eq = FALSE,
    remove.system.eq = TRUE,
    remove.ineq = FALSE,
    remove.def = FALSE,
    header = TRUE,
    output = "text"
  )
  if (is.null(PE$block)) {
    PE$block <- 1
    PE$block[PE$op == ":="] <- 0
  }

  # # If PML, remove intercepts when not estimated
  # if (object@Model@estimator == "PML") {
  #   browser()
  # }

  # Now need to put information into PE from pt and summary
  pt <- object@ParTable
  ptfreeidx <- which(pt$free > 0)
  ptdefidx <- which(pt$op == ":=")
  ptdeltaidx <- which(pt$op == "~*~")
  ptidx <- c(ptfreeidx, ptdefidx, ptdeltaidx)
  summ <- object@external$inlavaan_internal$summary
  peidx <- match(
    paste0(
      pt$lhs[ptidx],
      pt$op[ptidx],
      pt$rhs[ptidx],
      pt$block[ptidx]
    ),
    paste0(PE$lhs, PE$op, PE$rhs, PE$block)
  )
  summidx <- match(pt$free[pt$free > 0], seq_len(nrow(summ)))
  if (length(ptdefidx) > 0 | length(ptdeltaidx) > 0) {
    # FIXME: I think this should be ok, since pt$free always in increasing
    # order
    summidx <- seq_len(nrow(summ))
  }

  char.format <- paste("%", max(8, nd + 5), "s", sep = "")
  PE$SD <- ""
  PE$SD[peidx] <- formatC(summ$SD[summidx], digits = nd, format = "f")

  if (isTRUE(ci)) {
    PE$`2.5%` <- ""
    PE$`2.5%`[peidx] <- formatC(summ$`2.5%`[summidx], digits = nd, format = "f")
  }

  if (isTRUE(postmedian)) {
    PE$`50%` <- ""
    PE$`50%`[peidx] <- formatC(summ$`50%`[summidx], digits = nd, format = "f")
  }

  if (isTRUE(ci)) {
    PE$`97.5%` <- ""
    PE$`97.5%`[peidx] <- formatC(
      summ$`97.5%`[summidx],
      digits = nd,
      format = "f"
    )
  }

  if (isTRUE(postmode)) {
    PE$Mode <- ""
    PE$Mode[peidx] <- formatC(summ$Mode[summidx], digits = nd, format = "f")
  }

  # Standardised solution?
  if (isTRUE(standardised)) {
    stdlv <- standardisedsolution(object, type = "std.lv", ...)
    stdall <- standardisedsolution(object, type = "std.all", ...)
    stdidx <- match_partable_rows(PE, stdlv)
    PE$std.lv <- format_pe_col(stdlv$est.std[stdidx], nd)
    PE$std.all <- format_pe_col(stdall$est.std[stdidx], nd)
  }

  # NMAD (skewnorm marginal fit quality)
  if (isTRUE(nmad)) {
    # nocov start
    nmad_vals <- tryCatch(
      object@external$inlavaan_internal$approx_data[, "nmad"],
      error = function(e) NULL
    )
    if (!is.null(nmad_vals) && !all(is.na(nmad_vals))) {
      PE$NMAD <- ""
      PE$NMAD[peidx] <- formatC(nmad_vals[summidx], digits = nd, format = "f")
      PE$NMAD[peidx][grepl("NA", PE$NMAD[peidx])] <- ""
    }
  } # nocov end

  # KLD and VB shift in units of posterior SD (opt-in)
  if (isTRUE(vb_correction)) {
    # nocov start
    if (isTRUE(kld)) {
      PE$KLD <- ""
      PE$KLD[peidx] <- formatC(summ$kld[summidx], digits = nd, format = "f")
      PE$KLD[peidx][grepl("NA", PE$KLD[peidx])] <- ""
    }

    if (isTRUE(vb_shift)) {
      PE$VBshift <- ""
      PE$VBshift[peidx] <- formatC(
        summ$vb_shift_sigma[summidx],
        digits = nd,
        format = "f"
      )
      PE$VBshift[peidx][grepl("NA", PE$VBshift[peidx])] <- ""
    }
  } # nocov end

  if (isTRUE(priors)) {
    # nocov start
    PE$Prior <- ""
    PE$Prior[peidx] <- summ$Prior[summidx]
    PE$Prior[peidx][is.na(PE$Prior[peidx])] <- ""
  } # nocov end

  # If PML, intercepts shown regardless
  if (object@Model@estimator == "PML") {
    int_idx <- PE$est == 0 & pt$op == "~1"
    PE <- PE[!int_idx, ]
  }

  # Repair credible intervals labels
  garb <- capture.output(PE)
  garb <- gsub("X2.5.", " 2.5%", garb)
  garb <- gsub(" X50.", "  50%", garb)
  garb <- gsub("X97.5.", " 97.5%", garb)

  # Add Parameter Estimates section
  idxpehead <- grep("Parameter Estimates", garb)
  idxparameterization <- grep("Parameterization", garb)
  if (length(idxparameterization) > 0) {
    garb <- garb[-idxparameterization]
  }

  newgarb <- garb[1:(idxpehead + 1)]
  if (length(idxparameterization) > 0) {
    PARAMZ <- gsub(
      "^([a-z])",
      "\\U\\1",
      object@Options$parameterization,
      perl = TRUE
    )
    newgarb <- c(
      newgarb,
      capture.output(cat(
        "\n",
        sprintf("  %-38s", "Parameterization"),
        sprintf("  %10s", PARAMZ)
      ))[2]
    )
  }
  newgarb <- c(
    newgarb,
    capture.output(cat(
      "\n",
      sprintf("  %-38s", "Marginalisation method"),
      sprintf("  %10s", toupper(marg_method))
    ))[2],
    capture.output(cat(
      "\n",
      sprintf("  %-38s", "VB correction"),
      sprintf("  %10s", toupper(vb_correction))
    ))[2],
    garb[(idxpehead + 2):length(garb)]
  )

  # Print
  cat(paste0(newgarb, collapse = "\n"))
  cat("\n")
}

#' @param header Logical; if TRUE, print model fit information header.
#' @param fit.measures Logical; if TRUE, print fit measures (DIC and PPP).
#' @param estimates Logical; if TRUE, print parameter estimates table.
#' @param ci Logical; if TRUE (default), include 95% credible intervals
#'   (2.5% and 97.5% posterior quantiles) in the estimates table.
#' @param standardized Logical; if TRUE, include standardized estimates.
#' @param standardised Alias of `standardized`; either spelling is accepted
#'   (the alias wins if both are supplied).
#' @param rsquare Logical; if TRUE, include R-square values for the observed
#'   dependent variables, computed at the posterior mean point estimates.
#' @param postmedian Logical; if TRUE, include posterior median in estimates.
#' @param postmode Logical; if TRUE, include posterior mode in estimates.
#' @param nmad Logical; if TRUE (default), include the NMAD column for
#'   skew-normal marginal fit quality.
#' @param kld Logical; if FALSE (default), omit the per-parameter KLD column.
#'   Set to TRUE to show it.
#' @param vb_shift Logical; if FALSE (default), omit the VB shift column
#'   (shift in units of posterior SD). Set to TRUE to show it.
#' @param priors Logical; if TRUE, include prior information in estimates.
#' @param nd Integer; number of decimal places to print for numeric values.
#'
#' @name summary
#' @rdname INLAvaan-class
#' @aliases summary,INLAvaan-method
#' @export
setMethod("summary", "INLAvaan", summary_inlavaan)

#' Update and re-fit an INLAvaan model
#'
#' Re-fit an [INLAvaan] model with modified arguments, in the style of
#' [stats::update()]. This is convenient for prior-sensitivity analyses
#' (vary `dp` or the `prior()` modifiers), iterative respecification (swap the
#' `model` or `add` syntax), and swapping approximation settings (e.g.
#' `marginal_method`, `nsamp`) without retyping the whole call.
#'
#' @details
#' The method edits the recorded [inlavaan()] call, replacing `model` (or
#' extending it via `add`) and overriding any argument supplied through `...`,
#' then re-evaluates it. Because [INLAvaan] extends the \pkg{lavaan} class, a
#' dedicated method is required: without it, `update()` would dispatch to
#' \pkg{lavaan}'s method and silently return a frequentist \pkg{lavaan} fit
#' instead of a Bayesian one.
#'
#' When the parameter structure is preserved (no change to `model`, `add`, or
#' `model.type`), the previous posterior mode is passed as the optimiser's
#' starting value (a warm start), so mode-finding is typically much faster than
#' a cold fit. This makes prior sweeps particularly cheap. The warm start only
#' changes the optimiser's starting point, not the fitted result. Supply
#' `start` explicitly to override it.
#'
#' Fits produced by INLAvaan versions that did not record their call cannot be
#' updated; re-fit once with [inlavaan()] (or [acfa()], [asem()], [agrowth()])
#' and `update()` will work on the new object.
#'
#' @param object An object of class [INLAvaan].
#' @param model Optional replacement model, in the same form accepted by
#'   [inlavaan()] (lavaan model syntax or a parameter table).
#' @param add Optional lavaan syntax appended to the original model. Requires
#'   the original model to have been specified as a syntax string, and is
#'   ignored when `model` is supplied.
#' @param ... Further arguments passed to [inlavaan()], overriding the values
#'   in the original call (e.g. `dp`, `marginal_method`, `nsamp`, `data`).
#' @param evaluate Logical. If `TRUE` (default) the updated call is evaluated
#'   and the re-fitted [INLAvaan] object is returned; if `FALSE` the updated,
#'   unevaluated call is returned for inspection.
#'
#' @returns A re-fitted [INLAvaan] object, or the updated call when
#'   `evaluate = FALSE`.
#'
#' @seealso [inlavaan()], [compare()]
#'
#' @importFrom stats update
#' @name update
#' @aliases update,INLAvaan-method
#' @export
#'
#' @examples
#' \dontrun{
#' model <- "visual =~ x1 + x2 + x3"
#' fit <- acfa(model, data = lavaan::HolzingerSwineford1939)
#'
#' # Prior sensitivity: refit under a tighter loading prior (warm-started)
#' fit2 <- update(fit, dp = priors_for(lambda = "normal(0,1)"))
#'
#' # Respecify: add a residual covariance
#' fit3 <- update(fit, add = "x1 ~~ x2")
#' compare(fit, fit3)
#'
#' # Swap approximation settings only
#' fit4 <- update(fit, marginal_method = "sampling", nsamp = 4000)
#' }
setMethod(
  "update",
  "INLAvaan",
  function(object, model, add, ..., evaluate = TRUE) {
    internal <- get_inlavaan_internal(object)
    cl <- internal$call
    if (is.null(cl)) {
      cli_abort(c(
        "This fit was produced by an INLAvaan version that did not record its
         call, so it cannot be updated.",
        "i" = "Re-fit once with {.fn inlavaan} (or {.fn acfa}, {.fn asem},
               {.fn agrowth}) and {.fn update} will work on the new object."
      ))
    }

    # Resolve the caller's environment once: symbols in the recorded call
    # (`model`, `data`) and the re-fit must evaluate there, not in the
    # auto-generated `.local` wrapper frame that `setMethod()` inserts.
    caller <- parent.frame(2L)

    # Replace or extend the model
    if (!missing(model)) {
      cl$model <- model
    } else if (!missing(add)) {
      orig <- cl$model
      if (is.symbol(orig) || is.call(orig)) {
        orig <- eval(orig, caller)
      }
      if (!is.character(orig)) {
        cli_abort(c(
          "{.arg add} requires the original model to be lavaan syntax.",
          "i" = "Supply a full replacement via {.arg model} instead."
        ))
      }
      cl$model <- paste(paste(orig, collapse = "\n"), add, sep = "\n")
    }

    # Override any inlavaan() argument passed through ..., preserving the
    # supplied expressions. Inside an S4 method `match.call()` sees the
    # `.local` wrapper, so match against this method's formals and the
    # original generic call one frame up.
    outer_call <- match.call(
      definition = sys.function(),
      call = sys.call(-1L),
      expand.dots = FALSE
    )
    extras <- outer_call[["..."]]
    for (nm in names(extras)) cl[[nm]] <- extras[[nm]]

    if (!evaluate) {
      return(cl)
    }

    # Warm start (Tier A): reuse the previous posterior mode as the optimiser's
    # starting value whenever the parameter structure is preserved. Harmless
    # when only downstream settings changed (the optimiser simply begins at the
    # mode); skipped when the model structure changed or `start` was set.
    structure_changed <- !missing(model) ||
      !missing(add) ||
      "model.type" %in% names(extras)
    if (!structure_changed && is.null(cl$start)) {
      cl$start <- internal$theta_star_novbc
    }

    eval(cl, caller)
  }
)

#' Density of a Beta distribution on a bounded interval
#'
#' @inheritParams stats::dbeta
#' @param x A numeric vector of quantiles.
#' @param a The lower bound of the interval.
#' @param b The upper bound of the interval.
#' @param log Logical; if TRUE, probabilities p are given as log(p).
#'
#' @returns A numeric vector of density values.
#' @export
#'
#' @examples
#' # Beta(2,5) on (0,100)
#' x <- seq(0, 100, length.out = 100)
#' y <- dbeta_box(x, shape1 = 2, shape2 = 5, a = 0, b = 100)
#' plot(x, y, type = "l", main = "Beta(2,5) on (0,100)")
#'
#' # Beta(1,1) i.e. uniform on (-1, 1)
#' x <- seq(-1, 1, length.out = 100)
#' y <- dbeta_box(x, shape1 = 1, shape2 = 1, a = -1, b = 1)
#' plot(x, y, type = "l", main = "Beta(1,1) on (-1,1)")
dbeta_box <- function(x, shape1, shape2, a, b, log = FALSE) {
  # basic checks
  if (
    !is.numeric(a) ||
      !is.numeric(b) ||
      length(a) != 1 ||
      length(b) != 1 ||
      !is.finite(a) ||
      !is.finite(b) ||
      b <= a
  ) {
    stop("Require finite scalars with b > a.")
  }

  # transform to (0,1)
  u <- (x - a) / (b - a)
  inside <- (x >= a) & (x <= b)

  # init output
  out <- if (log) rep(-Inf, length(x)) else rep(0, length(x))

  # evaluate where inside support
  if (any(inside, na.rm = TRUE)) {
    if (log) {
      out[inside] <- -log(b - a) + dbeta(u[inside], shape1, shape2, log = TRUE)
    } else {
      out[inside] <- dbeta(u[inside], shape1, shape2) / (b - a)
    }
  }

  # propagate NAs from x
  out[is.na(x)] <- NA_real_
  out
}

inlav_model_loglik <- function(
  x,
  lavmodel,
  lavsamplestats,
  lavdata,
  lavoptions,
  lavcache
) {
  lavmodel_x <- lavaan::lav_model_set_parameters(lavmodel, x)
  lavimplied <- lavaan::lav_model_implied(lavmodel_x)
  Sigma <- lavimplied$cov[[1]]

  out <- -1e40
  if (!is_bad_cov(Sigma)) {
    if (lavmodel@estimator == "ML") {
      # Multivariate normal log-likelihood
      # no lavh1 argument: it does not exist in lavaan < 0.6-20 and
      # defaults to NULL in later versions
      out <- lavaan___lav_model_loglik(
        lavdata = lavdata,
        lavsamplestats = lavsamplestats,
        lavimplied = lavimplied,
        lavmodel = lavmodel,
        lavoptions = lavoptions
      )$loglik
      if (is.na(out)) out <- -1e40
      if (out != -1e40 && marginalised_means_active(lavmodel)) {
        out <- out + marginalised_means_loglik_corr(lavimplied, lavsamplestats)
      }
    } else if (lavmodel@estimator == "PML") {
      # Pairwise log-likelihood
      no_ord <- length(lavdata@ordered)
      kappa <- 1 / sqrt(no_ord) # scaling factor for PML
      fx <- lavaan___lav_model_objective(
        lavmodel = lavmodel_x,
        lavsamplestats = lavsamplestats,
        lavdata = lavdata,
        lavcache = lavcache
      )
      logl <- sum(attr(fx, "logl.group"))
      if (is.na(logl)) {
        return(-1e40) # nocov
      }
      out <- kappa * logl
    }
  }

  out
}

inlav_model_grad <- function(
  x,
  lavmodel,
  lavsamplestats,
  lavdata,
  lavcache
) {
  lavmodel_x <- lavaan::lav_model_set_parameters(lavmodel, x)

  # Gradient of fit function F_ML (not loglik yet)
  grad_F <- lavaan___lav_model_gradient(
    lavmodel = lavmodel_x,
    lavsamplestats = lavsamplestats,
    lavdata = lavdata,
    lavcache = lavcache
  )

  out <-
    if (lavmodel@estimator == "ML") {
      -1 * lavsamplestats@ntotal * grad_F
    } else if (lavmodel@estimator == "PML") {
      # nocov start
      no_ord <- length(lavdata@ordered)
      kappa <- 1 / sqrt(no_ord) # scaling factor for PML
      -1 * kappa * grad_F
    } else {
      0 * x
    } # nocov end

  if (lavmodel@estimator == "ML" && marginalised_means_active(lavmodel)) {
    out <- out + marginalised_means_grad_corr(lavmodel_x)
  }

  out
}

# Without a mean structure, lavaan's log-likelihood profiles the saturated
# means at the sample means -- a frequentist device with no Bayesian
# counterpart. The coherent reading assigns the saturated means flat priors
# and integrates them out; the integral is closed form and equals the
# profiled log-likelihood plus, per group,
#   (1/2) log|Sigma| + (p/2) log(2*pi/n).
# The corrections below apply this at the loglik and gradient level so the
# whole posterior (mode, Hessian, marginals, samples) is built from the
# marginalised likelihood.
marginalised_means_active <- function(lavmodel) {
  !isTRUE(lavmodel@meanstructure) && !isTRUE(lavmodel@conditional.x)
}

marginalised_means_loglik_corr <- function(lavimplied, lavsamplestats) {
  corr <- 0
  for (g in seq_len(lavsamplestats@ngroups)) {
    Sigma_g <- lavimplied$cov[[g]]
    corr <- corr +
      0.5 * as.numeric(determinant(Sigma_g, logarithm = TRUE)$modulus) +
      0.5 * ncol(Sigma_g) * log(2 * pi / lavsamplestats@nobs[[g]])
  }
  corr
}

# d corr / dx_j = (1/2) tr(Sigma^{-1} dSigma/dx_j), assembled from the same
# Delta matrices (d vech(Sigma) / dx) the LOO machinery uses; off-diagonal
# vech elements are doubled to undo the half-vectorisation.
marginalised_means_grad_corr <- function(lavmodel_x) {
  lavimplied <- lavaan::lav_model_implied(lavmodel_x)
  Delta <- lavaan___lav_model_delta(lavmodel_x, lavmodel_x@GLIST)
  out <- 0
  for (g in seq_along(Delta)) {
    Sigma_inv <- tryCatch(
      chol2inv(chol(lavimplied$cov[[g]])),
      error = function(e) NULL # nocov
    )
    if (is.null(Sigma_inv)) {
      return(0) # nocov -- loglik is -1e40 here; gradient is moot
    }
    W <- 2 * Sigma_inv
    diag(W) <- diag(Sigma_inv)
    out <- out +
      0.5 * as.numeric(crossprod(Delta[[g]], lavaan::lav_matrix_vech(W)))
  }
  out
}

# The closed-form correction inlav_model_loglik() adds to lavaan's profiled
# loglik when the model has no mean structure (saturated means with flat
# priors, marginalised analytically), summed over groups
marg_corr <- function(fit) {
  imp <- lavaan::lav_model_implied(fit@Model)
  nobs <- unlist(fit@SampleStats@nobs)
  sum(vapply(seq_along(nobs), function(g) {
    S <- imp$cov[[g]]
    0.5 * as.numeric(determinant(S, logarithm = TRUE)$modulus) +
      0.5 * ncol(S) * log(2 * pi / nobs[g])
  }, numeric(1)))
}

test_that("Standard MVN loglik", {
  mod <- "
    # Latent variable definitions
    ind60 =~ x1 + x2 + x3
    dem60 =~ y1 + y2 + y3 + y4
    dem65 =~ y5 + y6 + y7 + y8

    # Latent regressions
    dem60 ~ ind60
    dem65 ~ ind60 + dem60

    # Residual correlations
    y1 ~~ y5
    y2 ~~ y4 + y6
    y3 ~~ y7
    y4 ~~ y8
    y6 ~~ y8
  "
  dat <- lavaan::PoliticalDemocracy
  fit <- lavaan::sem(mod, dat)

  # no mean structure: lavaan profiles the saturated means, INLAvaan
  # marginalises them under flat priors -- the closed-form correction
  # separates the two
  target <- as.numeric(lavaan::logLik(fit)) + marg_corr(fit)
  output <- inlav_model_loglik(
    coef(fit),
    fit@Model,
    fit@SampleStats,
    fit@Data,
    fit@Options
  )
  expect_equal(output, target, tolerance = 1e-5)
})

test_that("Standard multigroup likelihood", {
  mod <- "
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
    speed   =~ x7 + x8 + x9
  "
  dat <- lavaan::HolzingerSwineford1939
  fit <- lavaan::cfa(mod, dat, group = "school")

  target <- as.numeric(lavaan::logLik(fit))
  output <- inlav_model_loglik(
    coef(fit),
    fit@Model,
    fit@SampleStats,
    fit@Data,
    fit@Options
  )
  expect_equal(output, target, tolerance = 1e-5)
})

test_that("Multilevel no missing", {
  mod <- "
    level: 1
      fw =~ y1 + y2 + y3
      fw ~ x1 + x2 + x3
    level: 2
      fb =~ y1 + y2 + y3
      fb ~ w1 + w2
  "
  dat <- lavaan::Demo.twolevel
  fit <- lavaan::sem(mod, dat, cluster = "cluster")

  target <- as.numeric(lavaan::logLik(fit))
  output <- inlav_model_loglik(
    coef(fit),
    fit@Model,
    fit@SampleStats,
    fit@Data,
    fit@Options
  )
  expect_equal(output, target, tolerance = 1e-5)
})

test_that("Missing data", {
  mod <- "
    # Latent variable definitions
    ind60 =~ x1 + x2 + x3
    dem60 =~ y1 + y2 + y3 + y4
    dem65 =~ y5 + y6 + y7 + y8

    # Latent regressions
    dem60 ~ ind60
    dem65 ~ ind60 + dem60

    # Residual correlations
    y1 ~~ y5
    y2 ~~ y4 + y6
    y3 ~~ y7
    y4 ~~ y8
    y6 ~~ y8
  "
  set.seed(9619)
  mis <- matrix(
    rbinom(prod(dim(lavaan::PoliticalDemocracy)), 1, .95),
    nrow(lavaan::PoliticalDemocracy),
    ncol(lavaan::PoliticalDemocracy)
  )
  dat <- lavaan::PoliticalDemocracy * mis
  dat[dat == 0] <- NA

  # Complete cases
  suppressWarnings(fit <- lavaan::sem(mod, dat))
  target <- as.numeric(lavaan::logLik(fit)) + marg_corr(fit)
  output <- inlav_model_loglik(
    coef(fit),
    fit@Model,
    fit@SampleStats,
    fit@Data,
    fit@Options
  )
  expect_equal(output, target, tolerance = 1e-5)

  # FIML
  suppressWarnings(fit <- lavaan::sem(mod, dat, missing = "ML"))
  target <- as.numeric(lavaan::logLik(fit))
  output <- inlav_model_loglik(
    coef(fit),
    fit@Model,
    fit@SampleStats,
    fit@Data,
    fit@Options
  )
  expect_equal(output, target, tolerance = 1e-5)
})

# ---------------------------------------------------------------------------
# Shared helper: summarise a numeric vector of posterior samples into the
# standard (Mean, SD, 2.5%, 50%, 97.5%, Mode) list + pdf_data table.
# ---------------------------------------------------------------------------
summarise_samples <- function(y) {
  Ex <- mean(y)
  SDx <- stats::sd(y)
  qq <- stats::quantile(y, probs = c(0.025, 0.25, 0.5, 0.75, 0.975))
  dens <- stats::density(y)
  xmax <- dens$x[which.max(dens$y)]
  res <- c(Ex, SDx, qq, xmax)
  names(res) <- c("Mean", "SD", "2.5%", "25%", "50%", "75%", "97.5%", "Mode")
  list(
    summary = res,
    pdf_data = data.frame(x = dens$x, y = dens$y)
  )
}

# ---------------------------------------------------------------------------
# Core sampling function: draw from the Gaussian-copula posterior.
# Returns a list with:
#   $theta_samp  - nsamp x m matrix in the internal (theta) parameterisation
#   $x_samp      - nsamp x p matrix in the lavaan (x) parameterisation
# ---------------------------------------------------------------------------
sample_params <- function(
  theta_star,
  Sigma_theta,
  method,
  approx_data,
  pt,
  lavmodel,
  nsamp = 1000,
  R_star = NULL # NORTA-adjusted correlation matrix for SN copula
) {
  R <- cov2cor(Sigma_theta)
  R_use <- if (method == "skewnorm" && !is.null(R_star)) R_star else R
  Lt <- chol(R_use)
  z_raw <- matrix(rnorm(nsamp * ncol(R)), nrow = nsamp)
  z <- z_raw %*% Lt
  u <- apply(z, 2, pnorm)
  if (!is.matrix(u)) {
    u <- matrix(u, nrow = 1)
  } # nocov – nsamp == 1 edge case

  # Use marginals to get theta
  if (method == "sampling") {
    D <- diag(sqrt(diag(Sigma_theta)))
    theta <- t(D %*% qnorm(t(u)) + theta_star)
  } else {
    if (method == "skewnorm") {
      theta <- do.call(
        "cbind",
        lapply(seq_len(ncol(u)), function(j) {
          xi <- approx_data[j, "xi"]
          omega <- approx_data[j, "omega"]
          alpha <- approx_data[j, "alpha"]
          qsnorm_fast(u[, j], xi = xi, omega = omega, alpha = alpha)
        })
      )
    } else if (method == "asymgaus") {
      theta <- do.call(
        "cbind",
        lapply(seq_len(ncol(u)), function(j) {
          tt <- theta_star[j] +
            seq(-4, 4, length = 100) * sqrt(Sigma_theta[j, j])
          yy <- marg_lp(
            tt,
            j = j,
            theta_star = theta_star,
            Sigma_theta = Sigma_theta,
            sigma_asym = approx_data
          )
          yy <- yy - max(yy) # stabilise
          fj_lp <- stats::splinefun(tt, yy)
          fj <- function(par) exp(fj_lp(par)) # unnormalised
          dt <- diff(tt)
          ft <- fj(tt)
          fmid <- (head(ft, -1) + tail(ft, -1)) / 2
          ymid <- (head(tt, -1) + tail(tt, -1)) / 2
          C <- sum(fmid * dt)
          ft <- ft / C
          Ft <- c(0, cumsum(fmid * dt))
          Ft <- Ft / tail(Ft, 1)
          qfj <- splinefun(Ft, tt, method = "monoH.FC")
          qfj(u[, j])
        })
      )
    } else if (method == "marggaus") {
      theta <- do.call(
        "cbind",
        lapply(seq_len(ncol(u)), function(j) {
          mu_j <- theta_star[j]
          sd_j <- sqrt(Sigma_theta[j, j])
          qnorm(u[, j], mean = mu_j, sd = sd_j)
        })
      )
    }
  }

  if (lavmodel@ceq.simple.only) {
    K <- lavmodel@ceq.simple.K
    theta <- t(apply(theta, 1, function(pars) as.numeric(K %*% pars)))
  }

  x <- t(apply(theta, 1, pars_to_x, pt = pt))

  list(theta_samp = theta, x_samp = x)
}

# 1. Get correlation matrix R = cov2cor(Sigma_theta)
# 2. Sample z ~ N(0, R)
# 3. Compute u = Phi^{-1}(z)
# 4. For each j, compute theta_j = F^{-1}_j(u_j) where F_j is the marginal posterior CDF for parameter j
# 5. Get x = pars_to_x(theta)
# 6. Compute implied covariance matrix Sigma(x)
# 7. Sample Srep ~ Wishart(n - 1, Sigma(x))
# 7. Repeat 1-6 and get samples of Sigma, and Srep
#
# Define F(S, Sigma) = log |Sigma| + trace(S Sigma^{-1}) - log |S| - p
# and the test statistic is T(S, Sigma) = (n - 1) / 2 * F(S, Sigma)
# Then ppp = P(T(Srep, Sigma) >= T(S, Sigma))

get_ppp <- function(
  x_samp,
  lavmodel,
  lavsamplestats,
  lavdata,
  lavpartable = NULL,
  cli_env = NULL
) {
  n_blocks <- lavmodel@nblocks
  n_levels <- lavdata@nlevels

  # Pre-compute per-block observed variable names (needed for alignment)
  ov_names_block <- vector("list", n_blocks)
  for (b in seq_len(n_blocks)) {
    g <- ceiling(b / n_levels)
    ov_all <- lavdata@ov.names[[g]]
    if (!is.null(lavpartable)) {
      ov_names_block[[b]] <- unique(
        lavpartable$lhs[
          lavpartable$block == b &
            lavpartable$op == "~~" &
            lavpartable$lhs == lavpartable$rhs &
            lavpartable$lhs %in% ov_all
        ]
      )
    }
  }

  # Pre-compute per-block observed stats (S, n, logdet_S) — invariant across
  # samples, so we avoid redundant extraction and determinant calls in the loop.
  block_obs <- vector("list", n_blocks)
  for (b in seq_len(n_blocks)) {
    g <- ceiling(b / n_levels)
    l <- (b - 1) %% n_levels + 1

    if (n_levels > 1) {
      # nocov start
      cluster_stats <- lavsamplestats@YLp[[g]][[2]]
      ov_all <- lavdata@ov.names[[g]]
      if (l == 1) {
        n <- lavdata@nobs[[g]]
        S <- cluster_stats$Sigma.W
      } else {
        n <- lavdata@Lp[[g]]$nclusters[[l]]
        S <- cluster_stats$Sigma.B
      }
      rownames(S) <- colnames(S) <- ov_all
      keep <- rowSums(S != 0) > 0
      S <- S[keep, keep, drop = FALSE]
      # nocov end
    } else {
      n <- lavsamplestats@nobs[[g]]
      # rescale the divisor-n sample covariance to its unbiased (n - 1)
      # form: the replicates below are Wishart(n - 1, Sigma) / (n - 1)
      # objects, so the observed discrepancy must be on the same scale
      S <- lavsamplestats@cov[[g]] * n / (n - 1)
    }

    logdet_S <- as.numeric(determinant(S, logarithm = TRUE)$modulus)
    block_obs[[b]] <- list(S = S, n = n, logdet_S = logdet_S)
  }

  res <- vector("numeric", length = nrow(x_samp))
  for (i in seq_len(nrow(x_samp))) {
    if (!is.null(cli_env)) {
      cli_progress_update(.envir = cli_env) # nocov
    }

    xx <- x_samp[i, ]
    lavmodel_x <- lavaan::lav_model_set_parameters(lavmodel, xx)
    lavimplied <- lavaan::lav_model_implied(lavmodel_x)

    Tobs <- 0
    Trep <- 0

    for (b in seq_len(n_blocks)) {
      S <- block_obs[[b]]$S
      n <- block_obs[[b]]$n
      logdet_S <- block_obs[[b]]$logdet_S

      Sigma <- lavimplied$cov[[b]]

      # Align dimensions for multilevel models
      if (n_levels > 1 && length(ov_names_block[[b]]) == nrow(Sigma)) {
        # nocov start
        rownames(Sigma) <- colnames(Sigma) <- ov_names_block[[b]]
        shared <- intersect(rownames(S), rownames(Sigma))
        if (length(shared) > 0) {
          S <- S[shared, shared, drop = FALSE]
          Sigma <- Sigma[shared, shared, drop = FALSE]
          logdet_S <- as.numeric(determinant(S, logarithm = TRUE)$modulus)
        }
        # nocov end
      }

      if (!is_bad_cov(Sigma)) {
        p <- nrow(Sigma)

        # Cholesky of Sigma — shared for logdet, inverse, and both discrepancies
        L <- chol(Sigma)
        logdet_Sigma <- 2 * sum(log(diag(L)))
        Sigma_inv <- chol2inv(L)

        # Wishart draw for replicated data
        W <- matrix(stats::rWishart(1, df = n - 1, Sigma = Sigma)[,, 1], p, p)
        Srep <- W / (n - 1)
        logdet_Srep <- as.numeric(determinant(Srep, logarithm = TRUE)$modulus)

        # F(S, Sigma) = log|Sigma| + tr(Sigma^{-1} S) - log|S| - p
        Tobs <- Tobs + logdet_Sigma + sum(Sigma_inv * S) - logdet_S - p
        Trep <- Trep + logdet_Sigma + sum(Sigma_inv * Srep) - logdet_Srep - p
      }
    }

    res[i] <- as.numeric(Trep >= Tobs)
  }

  mean(res)
}

sample_covariances <- function(x_samp, pt) {
  pt_cov_rows <- grep("cov", pt$mat)
  pt_cov_free_rows <- pt_cov_rows[pt$free[pt_cov_rows] > 0]
  idxcov <- pt$free[pt_cov_free_rows]

  cov_samp <- x_samp[, idxcov, drop = FALSE]
  colnames(cov_samp) <- pt$names[pt_cov_free_rows]

  apply(cov_samp, 2, summarise_samples)
}

get_defpars <- function(x_samp, pt) {
  pt_def_rows <- which(pt$op == ":=")
  param_map <- setNames(pt$free[pt$free > 0], pt$label[pt$free > 0])
  param_map <- param_map[names(param_map) != ""]

  def_funs <- lapply(pt_def_rows, function(i) {
    expr <- parse(text = pt$rhs[i])

    function(theta) {
      env_vals <- as.list(theta[param_map])
      names(env_vals) <- names(param_map)
      eval(expr, envir = env_vals)
    }
  })

  # Apply each function in def_funs to every row of x_samp
  def_samp <- sapply(def_funs, function(fn) {
    apply(x_samp, 1, fn)
  })
  colnames(def_samp) <- pt$names[pt_def_rows]

  apply(def_samp, 2, summarise_samples)
}

get_defpars_fit_sn <- function(x_samp, pt) { # nocov start
  pt_def_rows <- which(pt$op == ":=")
  param_map <- setNames(pt$free[pt$free > 0], pt$label[pt$free > 0])
  param_map <- param_map[names(param_map) != ""]

  def_funs <- lapply(pt_def_rows, function(i) {
    expr <- parse(text = pt$rhs[i])

    function(theta) {
      env_vals <- as.list(theta[param_map])
      names(env_vals) <- names(param_map)
      eval(expr, envir = env_vals)
    }
  })

  def_samp <- sapply(def_funs, function(fn) {
    apply(x_samp, 1, fn)
  })
  colnames(def_samp) <- pt$names[pt_def_rows]

  sn_params <- apply(def_samp, 2, fit_skew_normal_samp)
  sn_params <- do.call("rbind", lapply(sn_params, unlist))
  sn_params <- cbind(sn_params, logC = 0, k = 0, rmse = 0, nmad = 0, gamma1 = 0)

  apply(sn_params, 1, function(y) {
    xi <- y["xi"]
    omega <- y["omega"]
    alpha <- y["alpha"]
    delta <- alpha / sqrt(1 + alpha^2)

    Ex <- xi + omega * delta * sqrt(2 / pi)
    Vx <- omega^2 * (1 - 2 * delta^2 / pi)
    SDx <- sqrt(Vx)
    qq <- qsnorm_fast(
      c(0.025, 0.25, 0.5, 0.75, 0.975),
      xi = xi,
      omega = omega,
      alpha = alpha
    )

    x <- seq(Ex - 4 * SDx, Ex + 4 * SDx, length.out = 200)
    fx <- dsnorm(x, xi = xi, omega = omega, alpha = alpha)

    xmax <- optimize(
      function(x) dsnorm(x, xi = xi, omega = omega, alpha = alpha),
      interval = range(x),
      maximum = TRUE
    )$maximum

    res <- c(Ex, SDx, qq, xmax)
    names(res) <- c("Mean", "SD", "2.5%", "25%", "50%", "75%", "97.5%", "Mode")

    list(
      summary = res,
      pdf_data = data.frame(x = x, y = fx),
      sn_params = c(
        xi = xi,
        omega = omega,
        alpha = alpha,
        logC = NA_real_,
        k = NA_real_,
        rmse = NA_real_,
        nmad = NA_real_,
        gamma1 = NA_real_
      )
    )
  })
} # nocov end

sample_covariances_fit_sn <- function(x_samp, pt) {
  pt_cov_rows <- grep("cov", pt$mat)
  pt_cov_free_rows <- pt_cov_rows[pt$free[pt_cov_rows] > 0]
  idxcov <- pt$free[pt_cov_free_rows]

  cov_samp <- x_samp[, idxcov, drop = FALSE]
  colnames(cov_samp) <- pt$names[pt_cov_free_rows]

  sn_params <- apply(cov_samp, 2, fit_skew_normal_samp)
  sn_params <- do.call("rbind", lapply(sn_params, unlist))

  # FIXME: Repeated code in post_marg_skewnorm
  apply(sn_params, 1, function(y) {
    xi <- y["xi"]
    omega <- y["omega"]
    alpha <- y["alpha"]
    delta <- alpha / sqrt(1 + alpha^2)

    Ex <- xi + omega * delta * sqrt(2 / pi)
    Vx <- omega^2 * (1 - 2 * delta^2 / pi)
    SDx <- sqrt(Vx)
    qq <- qsnorm_fast(
      c(0.025, 0.25, 0.5, 0.75, 0.975),
      xi = xi,
      omega = omega,
      alpha = alpha
    )

    x <- seq(Ex - 4 * SDx, Ex + 4 * SDx, length.out = 200)
    fx <- dsnorm(x, xi = xi, omega = omega, alpha = alpha)

    xmax <- optimize(
      function(x) dsnorm(x, xi = xi, omega = omega, alpha = alpha),
      interval = range(x),
      maximum = TRUE
    )$maximum

    res <- res <- c(Ex, SDx, qq, xmax)
    names(res) <- c("Mean", "SD", "2.5%", "25%", "50%", "75%", "97.5%", "Mode")

    list(
      summary = res,
      pdf_data = data.frame(x = x, y = fx),
      sn_params = c(
        xi = xi,
        omega = omega,
        alpha = alpha,
        logC = NA_real_,
        k = NA_real_,
        rmse = NA_real_,
        nmad = NA_real_,
        gamma1 = NA_real_
      )
    )
  })
}

get_dic <- function(
  x_samp,
  theta_star,
  pt,
  lavmodel,
  loglik,
  cli_env = NULL
) {
  if (lavmodel@ceq.simple.only) {
    # nocov start
    xhat <- pars_to_x(as.numeric(lavmodel@ceq.simple.K %*% theta_star), pt)
  } else {
    # nocov end
    xhat <- pars_to_x(theta_star, pt)
  }

  Dhat <- -2 * loglik(xhat)
  Dbar_samp <- sapply(seq_len(nrow(x_samp)), function(i) {
    if (!is.null(cli_env)) {
      cli_progress_update(.envir = cli_env)
    }
    xx <- x_samp[i, ]
    -2 * loglik(xx)
  })
  Dbar <- mean(Dbar_samp[Dbar_samp < 1e40])
  pD <- Dbar - Dhat

  list(dic = Dbar + pD, Dbar = Dbar, Dhat = Dhat, pD = pD)
}


get_thetaparamerization_deltas <- function(x_samp, lavmodel) {
  delta_samp <- apply(x_samp, 1, function(xx) {
    lavmodel_x <- lavaan::lav_model_set_parameters(lavmodel, xx)
    lavimplied <- lavaan::lav_model_implied(lavmodel_x, delta = FALSE)
    Sigma_list <- lavimplied$cov
    ngroups <- length(Sigma_list)
    delta_list <- vector("list", ngroups)
    for (g in seq_len(ngroups)) {
      delta_list[[g]] <- sqrt(diag(Sigma_list[[g]]))
    }
    unlist(delta_list)
  })

  apply(delta_samp, 1, summarise_samples)
}

#' Model Residuals for INLAvaan Models
#'
#' Extract the difference between the observed and model-implied (fitted)
#' sample statistics from a fitted \code{INLAvaan} model. As in \pkg{lavaan}
#' and \pkg{blavaan}, residuals are computed at the parameter estimates --
#' here the posterior means -- not as a posterior distribution over residuals.
#'
#' @param object An object of class [INLAvaan].
#' @param type Character. \code{"raw"} (default) returns the unscaled
#'   difference between the observed and model-implied covariance matrix (and
#'   mean vector, when a mean structure is present). \code{"cor"} (or
#'   \code{"cor.bollen"}) first rescales both matrices to a correlation
#'   matrix. \code{"cor.bentler"} rescales both by the observed variances
#'   (the basis of the SRMR). \code{"normalized"} and \code{"standardized"}
#'   divide the raw residuals by their asymptotic standard errors.
#'   \code{"casewise"} (aliases \code{"case"}, \code{"obs"},
#'   \code{"observations"}, \code{"ov"}) returns observed-minus-fitted values
#'   for each observation.
#' @param labels Logical. Attach variable names to the output. Default
#'   \code{TRUE}.
#' @param ... Currently unused.
#'
#' @returns For moment-based \code{type}s, a list with elements
#'   \code{type}, \code{cov}, and (when relevant) \code{mean}. For
#'   \code{type = "casewise"}, a numeric matrix of observed-minus-fitted
#'   values.
#'
#' @details
#' This delegates to \pkg{lavaan}'s own \code{residuals()} machinery, so the
#' return structure matches lavaan exactly. Because INLAvaan stores the
#' posterior means as the point estimates of the fitted object, the residuals
#' are the observed statistics minus the posterior-mean model-implied
#' statistics (mirroring \pkg{blavaan}, which likewise inherits lavaan's
#' \code{residuals()} without overriding it).
#'
#' @seealso [fitted()], [predict()], [fitMeasures()][lavaan::fitMeasures]
#'
#' @examples
#' \donttest{
#' HS.model <- "
#'   visual  =~ x1 + x2 + x3
#'   textual =~ x4 + x5 + x6
#'   speed   =~ x7 + x8 + x9
#' "
#' utils::data("HolzingerSwineford1939", package = "lavaan")
#' fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
#'             test = "none", verbose = FALSE)
#'
#' # Raw residual covariance matrix (posterior means)
#' residuals(fit)
#'
#' # SRMR-basis residuals
#' residuals(fit, type = "cor.bentler")
#'
#' # Casewise observed-minus-fitted values
#' head(residuals(fit, type = "casewise"))
#' }
#'
#' @importFrom stats residuals
#' @name residuals
#' @aliases residuals,INLAvaan-method
#' @export
setMethod("residuals", "INLAvaan", function(object, type = "raw", labels = TRUE, ...) {
  # Delegate to lavaan's implementation so the output structure (moments,
  # casewise) stays identical; the posterior means already live in the object.
  lavaan::residuals(as(object, "lavaan"), type = type, labels = labels, ...)
})

#' @importFrom stats resid
#' @rdname residuals
#' @aliases resid,INLAvaan-method
#' @export
setMethod("resid", "INLAvaan", function(object, type = "raw", ...) {
  lavaan::resid(as(object, "lavaan"), type = type, ...)
})

cli_messages <- c(
  "Laplace-ing through p dimensions",
  "Summoning Bayesian spirits",
  "Casting statistical spells",
  "Conjuring INLA magic",
  "Channeling Laplace's wizardry",
  "Harnessing the power of priors",
  "Diving into the probability pool",
  "Navigating the seas of stochasticity"
)

# Moore-Penrose pseudoinverse (replaces MASS::ginv, uses only base svd)
ginv_base <- function(X, tol = sqrt(.Machine$double.eps)) {
  s <- svd(X)
  pos <- s$d > max(tol * s$d[1], 0)
  if (all(pos)) {
    s$v %*% (1 / s$d * t(s$u))
  } else if (!any(pos)) { # nocov start
    array(0, dim(X)[2:1])
  } else {
    s$v[, pos, drop = FALSE] %*%
      ((1 / s$d[pos]) * t(s$u[, pos, drop = FALSE]))
  } # nocov end
}

#' Helper function to check if two functions are the same
#'
#' @param f,g Functions to compare.
#'
#' @returns Logical.
#' @export
#'
#' @examples
#' f1 <- function(x) { x^2 + 1 }
#' f2 <- function(x) { x^2 + 1 }
#' is_same_function(f1, f2)  # TRUE
is_same_function <- function(f, g) {
  identical(deparse(body(f)), deparse(body(g)))
}

#' Convert function to single string
#'
#' @param f Function to convert.
#' @returns A single character vector representing the function.
#' @export
#'
#' @examples
#' f <- function(x) { x^2 + 1 }
#' as_fun_string(f)
as_fun_string <- function(f) {
  gsub("\\s+", " ", paste(deparse(f), collapse = " "))
}

# Check if matrix is a bad covariance (not PD, or contains NA/NaN/Inf)
is_bad_cov <- function(mat) {
  if (any(!is.finite(mat))) return(TRUE)
  tryCatch({ chol(mat); FALSE }, error = function(e) TRUE)
}

# Nearest PD via eigenvalue clamping
make_pd <- function(X, tol = 1e-8) {
  e <- eigen(X, symmetric = TRUE)
  vals <- pmax(e$values, tol * max(e$values))
  e$vectors %*% (vals * t(e$vectors))
}

#' Extract the Internal INLAvaan Object
#'
#' Returns the `inlavaan_internal` list stored inside a fitted [INLAvaan]
#' object, optionally extracting a single named element.
#'
#' @param object An object of class [INLAvaan].
#' @param what Character. Name of the element to extract from the internal
#'   list. If missing, the entire list is returned. Common elements include
#'   `"coefficients"`, `"summary"`, `"Sigma_theta"`, `"vcov_x"`,
#'   `"theta_star"`, `"approx_data"`, `"pdf_data"`, `"partable"`,
#'   `"marginal_method"`, `"nsamp"`, `"mloglik"`, `"DIC"`, `"ppp"`,
#'   `"vb"`, `"opt"`, `"timing"`, `"visual_debug"`.
#'
#' @returns The full `inlavaan_internal` list, or the named element when
#'   `what` is supplied.
#'
#' @seealso [INLAvaan-class], [diagnostics()], [timing()]
#'
#' @examples
#' \donttest{
#' HS.model <- "
#'   visual  =~ x1 + x2 + x3
#'   textual =~ x4 + x5 + x6
#'   speed   =~ x7 + x8 + x9
#' "
#' utils::data("HolzingerSwineford1939", package = "lavaan")
#' fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
#'             test = "none", verbose = FALSE)
#'
#' # Full internal object
#' int <- get_inlavaan_internal(fit)
#' names(int)
#'
#' # Extract a specific element
#' get_inlavaan_internal(fit, "coefficients")
#' }
#'
#' @export
get_inlavaan_internal <- function(object, what) {
  if (!inherits(object, "INLAvaan")) {
    cli_abort("Object must be of class {.cls INLAvaan}.")
  }
  int <- object@external$inlavaan_internal
  if (missing(what)) {
    return(int)
  }
  if (!what %in% names(int)) { # nocov start
    cli_abort(c(
      "Element {.val {what}} not found in the internal list.",
      "i" = "Available: {.val {names(int)}}."
    ))
  } # nocov end
  int[[what]]
}

# Helper function to add timing information. Adapted by Haziq Jamil. Original
# author: Luc De Wilde (lavaan).
add_timing <- function(timing, part) {
  timenow <- proc.time()[3]
  timing[[part]] <- (timenow - timing$start.time)
  timing$start.time <- timenow

  timing
}
is_lavaan <- function(object) { # nocov start
  is(object, "lavaan") & attr(class(object), "package") == "lavaan"
}

is_blavaan <- function(object) {
  is(object, "blavaan") & attr(class(object), "package") == "blavaan"
} # nocov end

is_INLAvaan <- function(object) {
  is(object, "INLAvaan") & attr(class(object), "package") == "INLAvaan"
}

is_inlavaan <- function(object) {
  is(object, "INLAvaan") & attr(class(object), "package") == "INLAvaan"
}

# Two-level (cluster) model check from a lavData object
is_multilevel <- function(lavdata) {
  lavdata@nlevels > 1L
}

dmode <- function(x, na.rm = TRUE) {
  if (na.rm) {
    x <- x[!is.na(x)]
  }
  if (length(x) == 0) {
    return(NA)
  }
  if (length(x) == 1) {
    return(x)
  }
  if (stats::sd(x) < .Machine$double.eps^0.5) {
    return(mean(x))
  }
  d <- stats::density(x)
  d$x[which.max(d$y)]
}

# ---------------------------------------------------------------------------
# Parallel or serial lapply with cli progress (chunked for parallel)
# ---------------------------------------------------------------------------
run_parallel_or_serial <- function(m, FUN, cores = 1L, verbose = FALSE,
                                   msg_serial = NULL, msg_parallel = NULL,
                                   msg_done = NULL) {
  if (cores > 1L) { # nocov start
    # Parallel: process in chunks of `cores` for progress feedback
    if (verbose) {
      msg <- if (!is.null(msg_parallel)) msg_parallel
             else "Processing {m} items ({cores} cores)."
      done <- 0L
      cli_progress_step(
        msg,
        msg_done = if (is.null(msg_done)) msg else msg_done,
        spinner = TRUE
      )
    }
    chunk_ids <- split(seq_len(m), ceiling(seq_len(m) / cores))
    results <- vector("list", m)
    for (ch in chunk_ids) {
      results[ch] <- parallel::mclapply(ch, FUN, mc.cores = cores)
      if (verbose) {
        done <- max(ch)
        cli_progress_update()
      }
    }
  } else { # nocov end
    # Serial with per-item progress
    if (verbose) {
      j <- 0L
      msg <- if (!is.null(msg_serial)) msg_serial
             else "Processing {j}/{m} item{?s}."
      cli_progress_step(
        msg,
        msg_done = if (is.null(msg_done)) msg else msg_done,
        spinner = TRUE
      )
    }
    results <- vector("list", m)
    for (j in seq_len(m)) {
      results[[j]] <- FUN(j)
      if (verbose) cli_progress_update()
    }
  }
  results
}

dat <- lavaan::HolzingerSwineford1939
mod <- "
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
    speed   =~ x7 + x8 + x9
  "
fit_lav <- lavaan::cfa(mod, dat)
NSAMP <- 3
STDLV <- FALSE

test_that("Method: skewnorm", {
  expect_no_error({
    fit <- acfa(
      mod,
      dat,
      marginal_method = "skewnorm",
      verbose = FALSE,
      nsamp = NSAMP,
      std.lv = STDLV
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
  expect_equal(coef(fit), coef(fit_lav), tolerance = 0.1)
  expect_equal(fit@optim$dx, rep(0, length(coef(fit))), tolerance = 1e-3)
})

test_that("Method: asymgaus", {
  expect_no_error({
    fit <- acfa(
      mod,
      dat,
      marginal_method = "asymgaus",
      verbose = FALSE,
      nsamp = NSAMP,
      std.lv = STDLV
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
  expect_equal(coef(fit), coef(fit_lav), tolerance = 0.1)
})

test_that("Method: marggaus", {
  expect_no_error({
    fit <- acfa(
      mod,
      dat,
      marginal_method = "marggaus",
      verbose = FALSE,
      nsamp = NSAMP,
      std.lv = STDLV
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
  expect_equal(coef(fit), coef(fit_lav), tolerance = 0.1)
})

test_that("Method: sampling", {
  expect_no_error({
    fit <- acfa(
      mod,
      dat,
      marginal_method = "sampling",
      verbose = FALSE,
      # pure-sampling summaries are means of the draws, so a handful of
      # draws is too seed-sensitive for the coef comparison below
      nsamp = 100,
      std.lv = STDLV
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
  expect_equal(coef(fit), coef(fit_lav), tolerance = 0.1)
})

test_that("Gradients are correct (Finite Difference Check)", {
  suppressMessages(
    tmp <- capture.output(fit <- acfa(mod, dat, test = "none", debug = TRUE))
  )
  test_df <- read.table(text = tmp, skip = 1)[, -1]
  colnames(test_df) <- c("fd", "analytic", "diff")

  expect_equal(
    as.numeric(test_df$fd),
    as.numeric(test_df$diff),
    tolerance = 1e-3
  )
  expect_equal(
    as.numeric(test_df$diff),
    rep(0, nrow(test_df)),
    tolerance = 1e-3
  )
})

test_that("Covariance coefficients are on the covariance scale", {
  # Regression test: coef() used to return the posterior-mean correlation
  # (tanh of theta) for covariance parameters, while summary() showed the
  # sample-based covariance. Invisible when factor sds are near 1, so use a
  # subsample where the two scales differ clearly.
  set.seed(20260612)
  dat_sub <- dat[sample(nrow(dat), 120L), ]
  fit <- acfa(mod, dat_sub, verbose = FALSE, nsamp = NSAMP, test = "none")
  co <- coef(fit)
  summ <- get_inlavaan_internal(fit)$summary

  cov_names <- c("visual~~textual", "visual~~speed", "textual~~speed")
  expect_equal(
    unname(co[cov_names]),
    summ[cov_names, "Mean"],
    tolerance = 1e-10
  )
  # posterior-mean covariances must respect the positive-definiteness bound
  for (nm in cov_names) {
    lv <- strsplit(nm, "~~", fixed = TRUE)[[1]]
    bound <- sqrt(co[paste0(lv[1], "~~", lv[1])] * co[paste0(lv[2], "~~", lv[2])])
    expect_lt(abs(co[nm]), bound)
  }
})

test_that("is_same_function compares function bodies correctly", {
  f1 <- function(x) {
    x + 1
  }
  f2 <- function(x) {
    x + 1
  }
  f3 <- function(x) {
    x + 2
  }

  # Identical bodies should return TRUE
  expect_true(is_same_function(f1, f2))

  # Different bodies should return FALSE
  expect_false(is_same_function(f1, f3))

  # Edge case: It checks body only, ignores argument names
  # (Documenting this behavior is important)
  g1 <- function(a) {
    return(1)
  }
  g2 <- function(b) {
    return(1)
  }
  expect_true(is_same_function(g1, g2))
})

test_that("as_fun_string flattens functions to single string", {
  f <- function(x) {
    x^2 + 1
  }

  res <- as_fun_string(f)

  expect_type(res, "character")
  expect_length(res, 1)
  # Check that newlines/spaces were normalized
  expect_match(res, "function \\(x\\)")
  expect_match(res, "x\\^2 \\+ 1")
})

test_that("is_bad_cov identifies problematic matrices", {
  # 1. Positive Definite (Good) -> Should return FALSE
  mat_pd <- matrix(c(2, 0.5, 0.5, 2), 2, 2)
  expect_false(is_bad_cov(mat_pd))

  # 2. Negative Definite (Bad) -> Should return TRUE
  mat_nd <- matrix(c(-2, 0.5, 0.5, -2), 2, 2)
  expect_true(is_bad_cov(mat_nd))

  # 3. NA/Inf/NaN (Bad) -> Should return TRUE
  mat_na <- mat_pd
  mat_na[1, 1] <- NA
  expect_true(is_bad_cov(mat_na))

  mat_inf <- mat_pd
  mat_inf[1, 1] <- Inf
  expect_true(is_bad_cov(mat_inf))
})

test_that("make_pd corrects non-PD matrices", {
  # Case 1: Already PD -> Should remain unchanged
  mat_pd <- matrix(c(2, 1, 1, 2), 2, 2)
  expect_equal(make_pd(mat_pd), mat_pd)

  # Case 2: Non-PD -> Should become PD
  # Create a matrix with negative eigenvalues
  mat_bad <- matrix(c(1, 2, 2, 1), 2, 2) # Eigenvalues: 3, -1

  mat_fixed <- make_pd(mat_bad)

  # Check that it is now PD (all eigenvalues > 0)
  eigs <- eigen(mat_fixed)$values
  expect_true(all(eigs + 1e-9 > 0))

  # Check that it is still symmetric
  expect_true(isSymmetric(mat_fixed))
})

test_that("is_lavaan / is_blavaan identify objects correctly", {
  mod <- "visual =~ x1 + x2 + x3"
  dat <- lavaan::HolzingerSwineford1939
  lav_fit <- lavaan::cfa(mod, dat)

  expect_true(is_lavaan(lav_fit))
  expect_false(isTRUE(is_lavaan(list())))
  expect_false(is_blavaan(lav_fit))
})

test_that("dmode handles edge cases", {
  expect_true(is.na(dmode(numeric(0))))
  expect_equal(dmode(42), 42)
  # Constant vector: returns mean
  expect_equal(dmode(rep(5, 10)), 5)
})

test_that("get_inlavaan_internal errors for non-INLAvaan", {
  expect_error(get_inlavaan_internal(list()), "Object must be of class")
})

dat <- lavaan::HolzingerSwineford1939
mod <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"
fit <- acfa(mod, dat, marginal_method = "marggaus", vb_correction = FALSE, verbose = FALSE, nsamp = 3)

test_that("deviance(type = 'mean') returns posterior mean deviance with pD/DIC attrs", {
  dev <- deviance(fit)
  int <- get_inlavaan_internal(fit)
  expect_s3_class(dev, "inlavaan_deviance")
  expect_equal(as.numeric(dev), int$DIC$Dbar)
  expect_equal(attr(dev, "pD"), int$DIC$pD)
  expect_equal(attr(dev, "DIC"), int$DIC$dic)
  expect_no_error(capture.output(print(dev)))
})

test_that("deviance(type = 'plugin') matches -2 * logLik(type = 'plugin')", {
  dev_plugin <- deviance(fit, type = "plugin")
  expect_equal(as.numeric(dev_plugin), as.numeric(-2 * logLik(fit, type = "plugin")))
})

test_that("deviance() errors without DIC components", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  expect_error(deviance(fit), "DIC")
})

test_that("plain lavaan fits are unaffected by the INLAvaan deviance() method", {
  fit_lav <- lavaan::cfa(mod, dat)
  expect_error(deviance(fit_lav))
})

dat <- lavaan::HolzingerSwineford1939
mod <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"
fit <- acfa(
  mod,
  dat,
  verbose = FALSE,
  nsamp = 3,
  test = "none",
  vb_correction = FALSE,
  marginal_method = "marggaus"
)

# ---- diagnostics() ----

test_that("diagnostics(type = 'global') returns named numeric vector", {
  dg <- diagnostics(fit)
  expect_s3_class(dg, "diagnostics.INLAvaan")
  expect_type(dg, "double")
  expected_names <- c(
    "npar",
    "nsamp",
    "converged",
    "iterations",
    "grad_inf",
    "grad_inf_rel",
    "grad_l2",
    "mode_shift_max",
    "hess_cond",
    "vb_applied",
    "vb_kld_global",
    "kld_max",
    "kld_mean",
    "nmad_max",
    "nmad_mean"
  )
  expect_named(dg, expected_names)
  expect_equal(unname(dg["npar"]), length(coef(fit)))
  expect_equal(unname(dg["converged"]), 1)
})

test_that("diagnostics(type = 'param') returns data frame", {
  dp <- diagnostics(fit, type = "param")
  expect_s3_class(dp, "diagnostics.INLAvaan.param")
  expect_s3_class(dp, "data.frame")
  expect_equal(nrow(dp), length(coef(fit)))
  expect_true(all(
    c("names", "grad", "grad_num", "grad_diff", "mode_shift_sigma") %in%
      names(dp)
  ))
})

# ---- fit-time diagnostics warnings ----

test_that("a healthy fit passes warn_fit_diagnostics() silently", {
  int <- INLAvaan:::get_inlavaan_internal(fit)
  expect_no_warning(INLAvaan:::warn_fit_diagnostics(int))
})

test_that("an unconverged optimiser triggers the fit-time warning", {
  w <- expect_warning(
    acfa(
      mod,
      dat,
      verbose = FALSE,
      nsamp = 3,
      test = "none",
      vb_correction = FALSE,
      marginal_method = "marggaus",
      std.lv = TRUE, # keeps the Hessian PD at the early stopping point
      control = list(iter.max = 2)
    ),
    class = "inlavaan_diagnostics_warning"
  )
  expect_match(conditionMessage(w), "did not converge")
  expect_match(conditionMessage(w), "posterior mode is not zero")
})

test_that("warn_fit_diagnostics() flags high NMAD marginals by name", {
  int <- INLAvaan:::get_inlavaan_internal(fit)
  m <- length(coef(fit))
  int$approx_data <- matrix(
    0,
    nrow = m,
    ncol = 1,
    dimnames = list(names(coef(fit)), "nmad")
  )
  int$approx_data[1:4, "nmad"] <- c(0.5, 0.3, 0.2, 0.15)
  w <- expect_warning(
    INLAvaan:::warn_fit_diagnostics(int),
    class = "inlavaan_diagnostics_warning"
  )
  expect_match(conditionMessage(w), "NMAD")
  expect_match(conditionMessage(w), names(coef(fit))[1], fixed = TRUE)
  expect_match(conditionMessage(w), "1 other") # top 3 shown, 1 elided
})

test_that("warn_fit_diagnostics() flags large VB shifts", {
  int <- INLAvaan:::get_inlavaan_internal(fit)
  m <- length(coef(fit))
  se <- sqrt(diag(int$Sigma_theta))
  int$vb$correction <- 2 * se
  int$vb$kld <- rep(2, m)
  int$vb$kld_global <- 1
  w <- expect_warning(
    INLAvaan:::warn_fit_diagnostics(int),
    class = "inlavaan_diagnostics_warning"
  )
  expect_match(conditionMessage(w), "VB correction")
})

test_that("warn_fit_diagnostics() flags a near-singular Hessian", {
  int <- INLAvaan:::get_inlavaan_internal(fit)
  m <- length(coef(fit))
  # zero out the gradient so only the condition-number check fires
  int$opt$dx_analytic <- int$opt$dx <- rep(0, m)
  int$Sigma_theta[1, 1] <- 1e12
  w <- expect_warning(
    INLAvaan:::warn_fit_diagnostics(int),
    class = "inlavaan_diagnostics_warning"
  )
  expect_match(conditionMessage(w), "near-singular")
})

test_that("print.diagnostics.INLAvaan works", {
  dg <- diagnostics(fit)
  expect_output(print(dg))
})

test_that("print.diagnostics.INLAvaan.param works", {
  dp <- diagnostics(fit, type = "param")
  expect_output(print(dp))
})

# ---- timing() ----

test_that("timing(what = 'total') returns total time", {
  tt <- timing(fit)
  expect_s3_class(tt, "timing.INLAvaan")
  expect_type(tt, "double")
  expect_named(tt, "total")
  expect_true(tt > 0)
})

test_that("timing(what = 'all') returns all segments", {
  tt <- timing(fit, what = "all")
  expect_s3_class(tt, "timing.INLAvaan")
  expect_true(length(tt) > 1)
  expect_true("total" %in% names(tt))
})

test_that("timing(what = ...) selects specific segments", {
  tt <- timing(fit, what = c("optim", "total"))
  expect_named(tt, c("optim", "total"))
})

test_that("timing rejects unknown segments", {
  expect_error(timing(fit, what = "nonexistent"), "Unknown")
})

test_that("print.timing.INLAvaan works", {
  tt <- timing(fit, what = "all")
  expect_output(print(tt))
})

mod <- "
  ind60 =~ x1 + x2 + x3
  dem60 =~ y1 + y2 + y3 + y4
  dem65 =~ y5 + y6 + y7 + y8

  dem60 ~ ind60
  dem65 ~ ind60 + dem60

  y1 ~~ y5
  y2 ~~ y4 + y6
  y3 ~~ y7
  y4 ~~ y8
  y6 ~~ y8
"
dat <- lavaan::PoliticalDemocracy
fit_lav <- lavaan::cfa(mod, dat)
NSAMP <- 3

test_that("Method: skewnorm", {
  expect_no_error({
    fit <- asem(
      mod,
      dat,
      marginal_method = "skewnorm",
      marginal_correction = "none",
      vb_correction = FALSE,
      test = "none",
      verbose = FALSE,
      nsamp = NSAMP
    )
  })

  # Summary
  expect_no_error(
    out <- capture.output(summary(fit, postmedian = TRUE, postmode = TRUE))
  )
  expect_no_error(out <- capture.output(summary(fit, rsquare = TRUE)))
  expect_no_error({
    tmp <- get_inlavaan_internal(fit)
    out <- capture.output(print(tmp))
    out <- capture.output(print(summary(tmp)))
  })

  expect_s4_class(fit, "INLAvaan")
  expect_equal(fit@optim$dx, rep(0, length(coef(fit))), tolerance = 1e-3)
})

test_that("Method: asymgaus", {
  expect_no_error({
    fit <- asem(
      mod,
      dat,
      marginal_method = "asymgaus",
      marginal_correction = "none",
      vb_correction = FALSE,
      test = "none",
      verbose = FALSE,
      nsamp = NSAMP
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
})

test_that("Method: marggaus", {
  expect_no_error({
    fit <- asem(
      mod,
      dat,
      marginal_method = "marggaus",
      vb_correction = FALSE,
      test = "none",
      verbose = FALSE,
      nsamp = NSAMP
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
})

test_that("Method: sampling", {
  expect_no_error({
    fit <- asem(
      mod,
      dat,
      marginal_method = "sampling",
      vb_correction = FALSE,
      test = "none",
      verbose = FALSE,
      nsamp = NSAMP
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
})

test_that("Gradients are correct (Finite Difference Check)", {
  suppressMessages(
    tmp <- capture.output(fit <- asem(mod, dat, test = "none", debug = TRUE))
  )
  test_df <- read.table(text = tmp, skip = 1)[, -1]
  colnames(test_df) <- c("fd", "analytic", "diff")

  expect_equal(
    as.numeric(test_df$fd),
    as.numeric(test_df$diff),
    tolerance = 1e-3
  )
  expect_equal(
    as.numeric(test_df$diff),
    rep(0, nrow(test_df)),
    tolerance = 1e-3
  )
})

dat <- lavaan::HolzingerSwineford1939
mod <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"

test_that("logLik(type = 'marginal') returns the marginal log-likelihood", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  ll <- logLik(fit)
  expect_s3_class(ll, "inlavaan_logLik")
  expect_equal(as.numeric(ll), get_inlavaan_internal(fit)$mloglik)
  expect_no_error(capture.output(print(ll)))
})

test_that("logLik(type = 'plugin') is AIC/BIC-compatible", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "standard")
  ll <- logLik(fit, type = "plugin")
  expect_s3_class(ll, "logLik")
  expect_equal(attr(ll, "df"), length(coef(fit)))
  expect_equal(attr(ll, "nobs"), nobs(fit))
  expect_no_error(AIC(ll))
  expect_no_error(BIC(ll))
})

test_that("logLik(type = 'plugin') errors without DIC components", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  expect_error(logLik(fit, type = "plugin"), "DIC")
})

test_that("plain lavaan fits are unaffected by the INLAvaan logLik() method", {
  fit_lav <- lavaan::cfa(mod, dat)
  expect_s3_class(logLik(fit_lav), "logLik")
})

test_that("AIC()/BIC() on an INLAvaan fit redirect instead of computing", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "standard")
  expect_error(AIC(fit), "plugin")
  expect_error(BIC(fit), "plugin")
})

test_that("AIC()/BIC() still work via the logLik(type = 'plugin') escape hatch", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "standard")
  ll <- logLik(fit, type = "plugin")
  expect_true(is.finite(AIC(ll)))
  expect_true(is.finite(BIC(ll)))
  expect_equal(AIC(ll), -2 * as.numeric(ll) + 2 * attr(ll, "df"))
})

test_that("plain lavaan fits are unaffected by the INLAvaan AIC()/BIC() methods", {
  fit_lav <- lavaan::cfa(mod, dat)
  expect_no_error(AIC(fit_lav))
  expect_no_error(BIC(fit_lav))
})

dat <- lavaan::HolzingerSwineford1939
mod <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"

test_that("update() records the call and returns an INLAvaan fit", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  expect_false(is.null(get_inlavaan_internal(fit, "call")))

  fit2 <- update(fit, verbose = FALSE)
  expect_s4_class(fit2, "INLAvaan")
  # A no-op update reproduces the (deterministic) posterior mode
  expect_equal(
    get_inlavaan_internal(fit2, "theta_star_novbc"),
    get_inlavaan_internal(fit, "theta_star_novbc"),
    tolerance = 1e-3
  )
})

test_that("update() overrides arguments and preserves expressions", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")

  cl <- update(fit, dp = priors_for(lambda = "normal(0,1)"), evaluate = FALSE)
  expect_true(is.call(cl))
  # The dot argument is kept as its literal expression, not a `..1` placeholder
  expect_equal(cl$dp, quote(priors_for(lambda = "normal(0,1)")))
  # warm-start `start` never leaks into the returned call
  expect_false("start" %in% names(as.list(cl)))
})

test_that("update() applies a changed prior", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  fit_tight <- update(
    fit,
    dp = priors_for(lambda = "normal(0,0.3)"),
    verbose = FALSE
  )
  # A tight loading prior shrinks loadings towards zero
  expect_lt(coef(fit_tight)["visual=~x2"], coef(fit)["visual=~x2"])
})

test_that("update(add=) extends the model structure", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  # nsamp = 3 can trip the marginal-fit diagnostic; irrelevant to structure here
  fit_add <- suppressWarnings(update(fit, add = "x1 ~~ x2", verbose = FALSE))
  expect_s4_class(fit_add, "INLAvaan")
  expect_equal(length(coef(fit_add)), length(coef(fit)) + 1L)
})

test_that("warm start reaches the same posterior mode as a cold fit", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  dp2 <- priors_for(lambda = "normal(0,0.3)")
  fit_warm <- update(fit, dp = dp2, verbose = FALSE)
  fit_cold <- acfa(mod, dat, dp = dp2, verbose = FALSE, nsamp = 3, test = "none")
  # Compare the deterministic optimiser target (the mode), not sample-based
  # summaries which carry Monte Carlo noise at nsamp = 3
  expect_equal(
    get_inlavaan_internal(fit_warm, "theta_star_novbc"),
    get_inlavaan_internal(fit_cold, "theta_star_novbc"),
    tolerance = 1e-3
  )
})

test_that("update() errors on a call-less fit", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  fit@external$inlavaan_internal$call <- NULL
  expect_error(update(fit), "did not record its call")
})

test_that("start of wrong length is rejected", {
  expect_error(
    acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none", start = 1:3),
    "free parameter"
  )
})

dat <- lavaan::HolzingerSwineford1939
mod <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"

# Fit once, reuse (fast defaults)
fit <- acfa(
  mod,
  dat,
  verbose = FALSE,
  nsamp = 5,
  vb_correction = FALSE,
  test = "none",
  marginal_method = "marggaus"
)

test_that("sampling() returns matrix for type = 'lavaan'", {
  s <- sampling(fit, type = "lavaan", nsamp = 10)
  expect_true(is.matrix(s))
  expect_equal(nrow(s), 10)
  expect_true(all(!is.na(colnames(s))))
})

test_that("sampling() returns matrix for type = 'theta'", {
  s <- sampling(fit, type = "theta", nsamp = 10)
  expect_true(is.matrix(s))
  expect_equal(nrow(s), 10)
  expect_equal(ncol(s), ncol(sampling(fit, type = "lavaan", nsamp = 10)))
})

test_that("sampling() with samp_copula = FALSE returns matrix", {
  s <- sampling(fit, type = "lavaan", nsamp = 10, samp_copula = FALSE)
  expect_true(is.matrix(s))
  expect_equal(nrow(s), 10)
})

test_that("sampling() type = 'latent' returns nsamp x nlv matrix", {
  s <- sampling(fit, type = "latent", nsamp = 8)
  expect_true(is.matrix(s))
  expect_equal(nrow(s), 8)
  expect_equal(ncol(s), 2) # visual, textual
  expect_equal(colnames(s), c("visual", "textual"))
})

test_that("sampling() type = 'observed' returns nsamp x nobs_vars matrix", {
  s <- sampling(fit, type = "observed", nsamp = 8)
  expect_true(is.matrix(s))
  expect_equal(nrow(s), 8)
  expect_equal(ncol(s), 6) # x1..x6
  expect_equal(colnames(s), paste0("x", 1:6))
})

test_that("sampling() type = 'all' returns named list of matrices", {
  s <- sampling(fit, type = "all", nsamp = 8)
  expect_true(is.list(s))
  expect_named(s, c("lavaan", "theta", "latent", "observed", "implied"))
  expect_equal(nrow(s$lavaan), 8)
  expect_equal(nrow(s$theta), 8)
  expect_equal(nrow(s$latent), 8)
  expect_equal(nrow(s$observed), 8)
  expect_equal(ncol(s$latent), 2)
  expect_equal(ncol(s$observed), 6)
  expect_length(s$implied, 8)
})

test_that("sampling() with prior = TRUE draws from priors", {
  s <- sampling(fit, type = "lavaan", nsamp = 10, prior = TRUE)
  expect_true(is.matrix(s))
  expect_equal(nrow(s), 10)
  sp <- sampling(fit, type = "lavaan", nsamp = 10)
  expect_equal(colnames(s), colnames(sp))
})

test_that("sampling() prior = TRUE with type = 'latent' works", {
  s <- sampling(fit, type = "latent", nsamp = 8, prior = TRUE)
  expect_true(is.matrix(s))
  expect_equal(nrow(s), 8)
  expect_equal(ncol(s), 2)
})

test_that("sampling() prior = TRUE with type = 'observed' works", {
  s <- sampling(fit, type = "observed", nsamp = 8, prior = TRUE)
  expect_true(is.matrix(s))
  expect_equal(nrow(s), 8)
  expect_equal(ncol(s), 6)
})

test_that("sampling() prior = TRUE with type = 'all' works", {
  s <- sampling(fit, type = "all", nsamp = 8, prior = TRUE)
  expect_true(is.list(s))
  expect_named(s, c("lavaan", "theta", "latent", "observed", "implied"))
  expect_equal(nrow(s$lavaan), 8)
  expect_equal(nrow(s$theta), 8)
  expect_equal(nrow(s$latent), 8)
  expect_equal(nrow(s$observed), 8)
  expect_length(s$implied, 8)
})

test_that("sampling() type = 'implied' returns list of covariance matrices", {
  s <- sampling(fit, type = "implied", nsamp = 8)
  expect_true(is.list(s))
  expect_length(s, 8)
  expect_true(is.matrix(s[[1]]$cov))
  expect_equal(nrow(s[[1]]$cov), 6)
  expect_equal(ncol(s[[1]]$cov), 6)
  expect_true(isSymmetric(s[[1]]$cov))
  expect_null(s[[1]]$mean)
})

test_that("sampling() type = 'implied' prior = TRUE works", {
  s <- sampling(fit, type = "implied", nsamp = 8, prior = TRUE)
  expect_true(is.list(s))
  expect_length(s, 8)
  expect_true(is.matrix(s[[1]]$cov))
  expect_equal(nrow(s[[1]]$cov), 6)
})

test_that("sampling.inlavaan_internal S3 dispatch works", {
  int <- INLAvaan:::get_inlavaan_internal(fit)
  s <- INLAvaan:::sampling.inlavaan_internal(int, type = "lavaan", nsamp = 5)
  expect_true(is.matrix(s))
  expect_equal(nrow(s), 5)
})

test_that("sampling() works with a single latent variable (nlv = 1)", {
  # Regression test: with one latent variable the generative draws were built
  # with t(vapply(., numeric(1))), yielding a 1 x nsamp row matrix and a
  # dimnames crash. They must come back as nsamp x 1 / nsamp x nobs matrices.
  fit1 <- acfa(
    "visual =~ x1 + x2 + x3",
    dat,
    verbose = FALSE,
    nsamp = 5,
    vb_correction = FALSE,
    test = "none",
    marginal_method = "marggaus"
  )

  lat <- sampling(fit1, type = "latent", nsamp = 8)
  expect_equal(dim(lat), c(8L, 1L))
  expect_equal(colnames(lat), "visual")

  obs <- sampling(fit1, type = "observed", nsamp = 8, silent = TRUE)
  expect_equal(dim(obs), c(8L, 3L))
  expect_equal(colnames(obs), paste0("x", 1:3))

  all_s <- sampling(fit1, type = "all", nsamp = 8, silent = TRUE)
  expect_equal(ncol(all_s$latent), 1L)
  expect_equal(ncol(all_s$observed), 3L)
})

test_that("Observed posterior draws are centred without a mean structure", {
  # Regression test: with meanstructure = FALSE the generative draws were
  # centred at zero; they must live on the data scale (saturated means).
  mod_hs <- "
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
  "
  hs <- lavaan::HolzingerSwineford1939
  fit <- acfa(
    mod_hs,
    hs,
    meanstructure = FALSE,
    verbose = FALSE,
    nsamp = 200,
    vb_correction = FALSE,
    test = "none"
  )
  yrep <- sampling(fit, type = "observed", nsamp = 200, silent = TRUE)
  ybar <- colMeans(hs[, colnames(yrep)])
  expect_lt(max(abs(colMeans(yrep) - ybar)), 0.5)
})

#' Specify priors for a SEM
#'
#' Specify priors for a SEM, similar to how [blavaan::dpriors()] works.
#'
#' This function provides a convenient way to specify prior distributions for
#' different types of parameters in a structural equation model (SEM). It uses a
#' registry of default priors for common lavaan parameter types (e.g., loadings,
#' regressions, residuals, etc.) and allows users to override these defaults by
#' passing named arguments.
#'
#' The parameter names, and default settings, are:
#' \itemize{
#'   \item \code{nu = "normal(0,32)"}: Observed variable intercepts
#'   \item \code{alpha = "normal(0,10)"}: Latent variable intercepts
#'   \item \code{lambda = "normal(0,10)"}: Factor loadings
#'   \item \code{beta = "normal(0,10)"}: Regression coefficients
#'   \item \code{theta = "gamma(1,.5)[sd]"}: Residual precisions
#'   \item \code{psi = "gamma(1,.5)[sd]"}: Latent variable precisions
#'   \item \code{rho = "beta(1,1)"}: Correlations (both latent and observed)
#'   \item \code{tau = "normal(0,1.5)"}: Thresholds for ordinal variables
#' }
#'
#' Note that the normal distributions are parameterised using standard
#' deviations, and not variances. For example, \code{normal(0,10)} means a
#' normal distribution with mean 0 and standard deviation 10 (not variance 10).
#'
#' @section Scale qualifiers:
#' For variance parameters (\code{theta}, \code{psi}), the prior distribution
#' can be placed on a transformed scale by appending a qualifier:
#' \itemize{
#'   \item \code{[sd]}: Prior is on the standard deviation \eqn{\sigma}.
#'     Example: \code{"gamma(1,0.5)[sd]"} places a Gamma(1, 0.5) prior on
#'     \eqn{\sigma = \sqrt{\text{variance}}}.
#'   \item \code{[prec]}: Prior is on the precision \eqn{\tau = 1/\sigma^2}.
#'     Example: \code{"gamma(1,1)[prec]"} places a Gamma(1, 1) prior on
#'     \eqn{\tau = 1/\text{variance}}. This is the parameterisation used by
#'     blavaan and corresponds to an Inverse-Gamma prior on the variance.
#' }
#' The necessary Jacobian adjustment is applied automatically in both cases.
#'
#' @param ... Named arguments specifying prior distributions for lavaan
#'   parameter types.
#'
#' @returns A named character vector of prior specifications, where names
#'   correspond to lavaan parameter types (e.g., "lambda", "beta", "theta",
#'   etc.) and values are character strings specifying the prior distribution
#'   (e.g., \code{"normal(0,10)"}, \code{"gamma(1,0.5)[sd]"},
#'   \code{"gamma(1,1)[prec]"}, etc.).
#'
#' @seealso [inlavaan()], [acfa()], [asem()], [agrowth()]
#'
#' @export
#'
#' @examples
#' priors_for(nu = "normal(0,10)", lambda = "normal(0,1)", rho = "beta(3,3)")
#'
#' # Precision-scale prior for residual variances (blavaan-style)
#' priors_for(theta = "gamma(1,1)[prec]")
priors_for <- function(...) {
  userspec <- list(...)

  # Ensure the user didn't just pass strings without names
  if (length(userspec) > 0 && is.null(names(userspec))) {
    cli_abort(
      "priors_for ERROR: All arguments must be named (e.g., `lambda = 'normal(0,1)'`)"
    )
  }

  # The prior distribution dictionary
  # fmt: skip
  out <- list(
    nu     = "normal(0,32)",     # Intercepts
    alpha  = "normal(0,10)",     # Latent intercepts
    lambda = "normal(0,10)",     # Loadings
    beta   = "normal(0,10)",     # Regressions
    theta  = "gamma(1,.5)[sd]",  # Residual precisions
    psi    = "gamma(1,.5)[sd]",  # Latent precisions
    rho    = "beta(1,1)",        # Correlations
    tau    = "normal(0,1.5)"     # Thresholds
  )

  # Merge: user specifications overwrite the registry
  if (length(userspec) > 0) {
    for (name in names(userspec)) {
      out[[name]] <- userspec[[name]]
    }
  }

  unlist(out)
}

test_that("priors_for returns named character vector with all defaults", {
  res <- priors_for()
  expect_type(res, "character")
  expect_named(res)
  expect_true(all(c("nu", "alpha", "lambda", "beta", "theta", "psi", "rho",
                     "tau") %in% names(res)))
})

test_that("priors_for default values are correct", {
  res <- priors_for()
  expect_equal(res[["nu"]],     "normal(0,32)")
  expect_equal(res[["lambda"]], "normal(0,10)")
  expect_equal(res[["theta"]],  "gamma(1,.5)[sd]")
  expect_equal(res[["rho"]],    "beta(1,1)")
})

test_that("priors_for overrides specific defaults", {
  res <- priors_for(lambda = "normal(0,1)", rho = "beta(3,3)")
  expect_equal(res[["lambda"]], "normal(0,1)")
  expect_equal(res[["rho"]],    "beta(3,3)")
  # Non-overridden defaults should remain

  expect_equal(res[["nu"]],     "normal(0,32)")
})

test_that("priors_for allows adding new (non-default) names", {
  res <- priors_for(custom = "normal(0,5)")
  expect_equal(res[["custom"]], "normal(0,5)")
})

test_that("priors_for errors on unnamed arguments", {
  expect_error(priors_for("normal(0,1)"), "named")
})

#' Standardised solution of a latent variable model
#'
#' @inheritParams lavaan::standardizedSolution
#' @inheritParams INLAvaan-class
#' @param object An object of class [INLAvaan].
#' @param cov.std Logical. If `TRUE`, the (residual) observed covariances
#'   are scaled by the square root of the `Theta` diagonal elements, and the
#'   (residual) latent covariances are scaled by the square root of the
#'   `Psi` diagonal elements. If `FALSE`, the (residual) observed
#'   covariances are scaled by the square root of the diagonal elements of
#'   the model-implied observed covariance matrix, and the (residual)
#'   latent covariances are scaled similarly using the model-implied
#'   covariance matrix of the latent variables. Documented explicitly here
#'   (rather than inherited) because lavaan >= 0.7-1 renamed this and the
#'   next three arguments to snake_case.
#' @param remove.eq Logical. If `TRUE`, filter the output by removing all
#'   rows containing equality constraints, if any.
#' @param remove.ineq Logical. If `TRUE`, filter the output by removing all
#'   rows containing inequality constraints, if any.
#' @param remove.def Logical. If `TRUE`, filter the output by removing all
#'   rows containing parameter definitions, if any.
#' @param nsamp The number of samples to draw from the approximate posterior
#'   distribution for the calculation of standardised estimates.
#' @param ... Additional arguments sent to `lavaan::standardizedSolution()`.
#'
#' @returns A `data.frame` containing standardised model parameters.
#'
#' @seealso [summary()], [coef()], [vcov()]
#'
#' @export
#' @name standardisedsolution
#' @rdname standardisedsolution
#' @example inst/examples/ex-stdsoln.R
standardisedsolution <- function(
  object,
  type = "std.all",
  se = TRUE,
  ci = TRUE,
  level = 0.95,
  postmedian = FALSE,
  postmode = FALSE,
  cov.std = TRUE,
  remove.eq = TRUE,
  remove.ineq = TRUE,
  remove.def = FALSE,
  nsamp = 250,
  ...
) {
  if (is_lavaan(object)) {
    return(lavaan::standardizedSolution(object))
  } else if (is_blavaan(object)) {
    return(blavaan::standardizedPosterior(object))
  }

  fit_inlv <- get_inlavaan_internal(object)
  pt <- fit_inlv$partable

  samp <- with(
    fit_inlv,
    sample_params(
      theta_star = theta_star,
      Sigma_theta = Sigma_theta,
      method = marginal_method,
      approx_data = approx_data,
      pt = partable,
      lavmodel = lavmodel,
      nsamp = nsamp
    )
  )
  x_samp <- samp$x_samp

  xstd_samp <- vector("list", nrow(x_samp))
  for (i in seq_len(nrow(x_samp))) {
    xi <- x_samp[i, ]
    lavmodel <- lavaan::lav_model_set_parameters(object@Model, xi)

    esti <- pt$est
    esti[pt$free > 0] <- xi[pt$free[pt$free > 0]]
    if (any(pt$op == ":=")) {
      pt_def_rows <- which(pt$op == ":=")
      def_names <- pt$names[pt_def_rows]
      esti[pt_def_rows] <- fit_inlv$summary[def_names, "Mean"]
    }
    xstd_samp[[i]] <- call_lavaan(
      "standardizedSolution",
      object = object,
      est = esti,
      GLIST = lavmodel@GLIST,
      type = type,
      cov.std = cov.std,
      remove.eq = remove.eq,
      remove.ineq = remove.ineq,
      remove.def = remove.def,
      ...
    )$est.std
  }
  xstd_samp <- do.call("rbind", xstd_samp)

  res <- list(
    mean = apply(xstd_samp, 2, mean),
    sd = apply(xstd_samp, 2, sd),
    ci_lower = apply(xstd_samp, 2, quantile, probs = (1 - level) / 2),
    ci_upper = apply(xstd_samp, 2, quantile, probs = 1 - (1 - level) / 2),
    median = apply(xstd_samp, 2, median),
    mode = apply(xstd_samp, 2, dmode)
  )

  out <- call_lavaan(
    "standardizedSolution",
    object = object,
    est = esti,
    type = type,
    cov.std = cov.std,
    remove.eq = remove.eq,
    remove.ineq = remove.ineq,
    remove.def = remove.def,
    ...
  )
  out$est.std <- res$mean
  if (isTRUE(se)) {
    out$se <- res$sd
  }
  if (isTRUE(ci)) {
    out$ci.lower <- res$ci_lower
    out$ci.upper <- res$ci_upper
  }
  if (isTRUE(postmedian)) {
    out$median <- res$median
  }
  if (isTRUE(postmode)) {
    out$mode <- res$mode
  }
  out
}

#' @name standardisedsolution
#' @rdname standardisedsolution
#' @export
standardisedSolution <- function(
  object,
  type = "std.all",
  se = TRUE,
  ci = TRUE,
  level = 0.95,
  postmedian = FALSE,
  postmode = FALSE,
  cov.std = TRUE,
  remove.eq = TRUE,
  remove.ineq = TRUE,
  remove.def = FALSE,
  nsamp = 250,
  ...
) {
  standardisedsolution(
    object = object,
    type = type,
    se = se,
    ci = ci,
    level = level,
    postmedian = postmedian,
    postmode = postmode,
    cov.std = cov.std,
    remove.eq = remove.eq,
    remove.ineq = remove.ineq,
    remove.def = remove.def,
    nsamp = nsamp,
    ...
  )
}

#' @name standardizedsolution
#' @rdname standardisedsolution
#' @export
standardizedsolution <- function(
  object,
  type = "std.all",
  se = TRUE,
  ci = TRUE,
  level = 0.95,
  postmedian = FALSE,
  postmode = FALSE,
  cov.std = TRUE,
  remove.eq = TRUE,
  remove.ineq = TRUE,
  remove.def = FALSE,
  nsamp = 250,
  ...
) {
  standardisedsolution(
    object = object,
    type = type,
    se = se,
    ci = ci,
    level = level,
    postmedian = postmedian,
    postmode = postmode,
    cov.std = cov.std,
    remove.eq = remove.eq,
    remove.ineq = remove.ineq,
    remove.def = remove.def,
    nsamp = nsamp,
    ...
  )
}

#' @name standardizedsolution
#' @rdname standardisedsolution
#' @export
standardizedSolution <- function(
  object,
  type = "std.all",
  se = TRUE,
  ci = TRUE,
  level = 0.95,
  postmedian = FALSE,
  postmode = FALSE,
  cov.std = TRUE,
  remove.eq = TRUE,
  remove.ineq = TRUE,
  remove.def = FALSE,
  nsamp = 250,
  ...
) {
  standardisedsolution(
    object = object,
    type = type,
    se = se,
    ci = ci,
    level = level,
    postmedian = postmedian,
    postmode = postmode,
    cov.std = cov.std,
    remove.eq = remove.eq,
    remove.ineq = remove.ineq,
    remove.def = remove.def,
    nsamp = nsamp,
    ...
  )
}

test_that("Standardised solution", {
  HS.model <- "
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
    speed   =~ x7 + x8 + x9
  "
  utils::data("HolzingerSwineford1939", package = "lavaan")
  fit <- acfa(
    HS.model,
    data = HolzingerSwineford1939,
    marginal_correction = FALSE,
    vb_correction = FALSE,
    verbose = FALSE,
    nsamp = 3
  )
  expect_no_error(out <- standardisedsolution(fit))
  expect_no_error(out <- standardisedSolution(fit))
  expect_no_error(out <- standardizedsolution(fit))
  expect_no_error(out <- standardizedSolution(fit))

  idx <- which(
    paste0(out$lhs, out$op, out$rhs) %in%
      c("visual~~visual", "textual~~textual", "speed~~speed")
  )

  expect_equal(out$est.std[idx], rep(1, length(idx)))
  expect_s3_class(out, "data.frame")
})

test_that("standardisedsolution with postmedian and postmode", {
  mod <- "visual =~ x1 + x2 + x3; textual =~ x4 + x5 + x6"
  dat <- lavaan::HolzingerSwineford1939
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3)
  out <- standardisedsolution(fit, postmedian = TRUE, postmode = TRUE)
  expect_true("median" %in% colnames(out))
  expect_true("mode" %in% colnames(out))
})

test_that("standardisedsolution dispatches to lavaan for lavaan objects", {
  mod <- "visual =~ x1 + x2 + x3; textual =~ x4 + x5 + x6"
  dat <- lavaan::HolzingerSwineford1939
  lav_fit <- lavaan::cfa(mod, dat)
  out <- standardisedsolution(lav_fit)
  expect_s3_class(out, "data.frame")
})

test_that("standardisedsolution handles := defined parameters", {
  mod <- "
    visual  =~ x1 + l2*x2 + x3
    textual =~ x4 + l5*x5 + x6
    ratio := l2 / l5
  "
  dat <- lavaan::HolzingerSwineford1939
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3)
  out <- standardisedsolution(fit)
  expect_s3_class(out, "data.frame")
})

#' Model-Implied Moments for INLAvaan Models
#'
#' Extract the model-implied (fitted) sample statistics from a fitted
#' \code{INLAvaan} model. As in \pkg{lavaan} and \pkg{blavaan}, the moments are
#' the model-implied covariance matrix (and mean vector, when a mean structure
#' is present) evaluated at the parameter estimates -- here the posterior means.
#'
#' @param object An object of class [INLAvaan].
#' @param type Character. \code{"moments"} (default) returns the model-implied
#'   variance-covariance matrix and, when relevant, the mean vector (plus
#'   thresholds for ordinal data). \code{"casewise"} (aliases \code{"obs"},
#'   \code{"ov"}) returns the model-predicted values for each observation.
#' @param labels Logical. Attach variable names to the output. Default
#'   \code{TRUE}.
#' @param ... Currently unused.
#'
#' @returns For \code{type = "moments"}, a list (or list of lists, for
#'   multiple groups) with elements such as \code{cov}, \code{mean}, and
#'   \code{th}. For \code{type = "casewise"}, a numeric matrix of predicted
#'   observed-variable values.
#'
#' @details
#' This delegates to \pkg{lavaan}'s own \code{fitted()} machinery, so the return
#' structure matches lavaan exactly. Because INLAvaan stores the posterior means
#' as the point estimates of the fitted object, the implied moments are the
#' posterior-mean model-implied moments (mirroring \pkg{blavaan}).
#'
#' @seealso [predict()], [coef()], [fitMeasures()][lavaan::fitMeasures]
#'
#' @examples
#' \donttest{
#' HS.model <- "
#'   visual  =~ x1 + x2 + x3
#'   textual =~ x4 + x5 + x6
#'   speed   =~ x7 + x8 + x9
#' "
#' utils::data("HolzingerSwineford1939", package = "lavaan")
#' fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
#'             test = "none", verbose = FALSE)
#'
#' # Model-implied covariance matrix (posterior means)
#' fitted(fit)
#'
#' # Casewise model-predicted observed values
#' head(fitted(fit, type = "ov"))
#' }
#'
#' @importFrom stats fitted
#' @name fitted
#' @aliases fitted,INLAvaan-method
#' @export
setMethod("fitted", "INLAvaan", function(object, type = "moments", labels = TRUE, ...) {
  # Delegate to lavaan's implementation so the output structure (moments,
  # casewise) stays identical; the posterior means already live in the object.
  lavaan::fitted(as(object, "lavaan"), type = type, labels = labels, ...)
})

#' @importFrom stats fitted.values
#' @rdname fitted
#' @aliases fitted.values,INLAvaan-method
#' @export
setMethod("fitted.values", "INLAvaan", function(object, type = "moments", labels = TRUE, ...) {
  lavaan::fitted.values(as(object, "lavaan"), type = type, labels = labels, ...)
})

# Internal machinery for Taylor-approximation leave-one-out cross-validation.
# The user-facing interface lives in R/method-loo.R.
#
# For a unit u (a subject for LOSO, a cluster for LOCO) with log-likelihood
# contribution l_u(theta), score s_u and Hessian H_u at the posterior summary
# (theta, Sigma), the case-deletion predictive density is approximated by a
# Taylor expansion of l_u around theta:
#
#   log_cpo_1 = l_u - 1/2 s_u' Sigma s_u
#   log_cpo_2 = l_u - 1/2 s_u' (Sigma^-1 + H_u)^-1 s_u + 1/2 log|I + Sigma H_u|
#
# and the full-posterior pointwise predictive density (used for p_loo) has the
# sign-flipped Gaussian-integral form
#
#   lpd_1 = l_u + 1/2 s_u' Sigma s_u
#   lpd_2 = l_u + 1/2 s_u' (Sigma^-1 - H_u)^-1 s_u - 1/2 log|I - Sigma H_u|
#
# Everything is evaluated from a single fit: no refitting and no sampling.

# Model-implied moments from a lavaan model whose parameters are already set.
# Single-level: a list over groups, each holding that group's (mu, Sigma) --
# lavaan's implied blocks are group-indexed there. Two-level: the flat
# within/between view of the (single) group, whose blocks are the levels.
loo_implied_moments <- function(lavmodel_x, two_level = FALSE) {
  imp <- lavaan::lav_model_implied(lavmodel_x)
  if (two_level) {
    pw <- ncol(imp$cov[[1L]])
    pb <- ncol(imp$cov[[2L]])
    list(
      mu_w = if (is.null(imp$mean[[1L]])) {
        rep(0, pw)
      } else {
        as.numeric(imp$mean[[1L]])
      },
      Sigma_w = imp$cov[[1L]],
      mu_b = if (is.null(imp$mean[[2L]])) {
        rep(0, pb)
      } else {
        as.numeric(imp$mean[[2L]])
      },
      Sigma_b = imp$cov[[2L]]
    )
  } else {
    lapply(seq_along(imp$cov), function(g) {
      p <- ncol(imp$cov[[g]])
      list(
        mu = if (is.null(imp$mean[[g]])) {
          rep(0, p)
        } else {
          as.numeric(imp$mean[[g]])
        },
        Sigma = imp$cov[[g]]
      )
    })
  }
}

# Derivative pieces that are constant across units at a given theta: build
# once, reuse for every unit. `theta` is in the packed space (as stored in
# `theta_star`); equality constraints are unpacked with ceq.simple.K before
# the transformation and scores are repacked in loo_chain_rule().
loo_grad_cache <- function(theta, lavmodel, pt, two_level = FALSE) {
  K <- NULL
  if (isTRUE(lavmodel@ceq.simple.only)) {
    K <- lavmodel@ceq.simple.K
  }
  theta_unp <- if (is.null(K)) theta else as.numeric(K %*% theta)
  x <- pars_to_x(theta_unp, pt)
  lavmodel_x <- lavaan::lav_model_set_parameters(lavmodel, x)
  jcb_mat <- attr(x, "jcb_mat")
  if (!is.null(jcb_mat)) {
    jcb_mat <- rbind(jcb_mat) # a single row drops to a vector upstream
  }
  list(
    x = x,
    mom = loo_implied_moments(lavmodel_x, two_level),
    Delta = lavaan___lav_model_delta(lavmodel_x, lavmodel_x@GLIST),
    jcb_vec = as.numeric(mapply(
      function(f, th) f(th),
      pt$ginv_prime[pt$free > 0],
      theta_unp
    )),
    sd1sd2 = attr(x, "sd1sd2"),
    jcb_mat = jcb_mat,
    K = K
  )
}

# Chain rule for likelihood gradients, vectorised over units: an n x q matrix
# of gradients w.r.t. lavaan-x (unpacked columns) becomes an n x m matrix
# w.r.t. packed theta. Mirrors the Jacobian adjustment in joint_lp_grad().
loo_chain_rule <- function(gll_x, cache) {
  n <- nrow(gll_x)
  G <- gll_x * rep(cache$jcb_vec * cache$sd1sd2, each = n)
  jm <- cache$jcb_mat
  if (!is.null(jm) && nrow(jm) > 0L) {
    for (k in seq_len(nrow(jm))) {
      G[, jm[k, 1L]] <- G[, jm[k, 1L]] + jm[k, 3L] * gll_x[, jm[k, 2L]]
    }
  }
  if (!is.null(cache$K)) {
    G <- G %*% cache$K
  }
  G
}

# ---- LOSO (single level): per-row log-likelihoods and theta-scores ---------

# Data view for the unit kernels, one entry per group (groups are
# independent, so every transformation is blockwise). With an estimated
# mean structure the units are iid within group and Y is used as-is.
# Without one the fit uses the marginalised likelihood (saturated means,
# flat priors, integrated out), under which units are exchangeable within
# their group and the coherent unit contribution is the case-deletion
# conditional
#   l_i = log phi(y_i; ybar_{g,-i}, c_g * Sigma_g),  c_g = n_g / (n_g - 1),
# which equals log phi(ytilde_i; 0, Sigma_g) - (p_g/2) log c_g with
# ytilde_i = sqrt(c_g) (y_i - ybar_g). Transforming the data once lets
# every iid kernel (casewise logliks, mvn scores, chain rule,
# second-order) apply verbatim -- the zero implied mean is then exact, and
# the theta-free constant shifts l_i without touching the scores. The
# transformation always uses the group's full-data ybar_g and n_g,
# regardless of any `units` subset. For the conditional (fixed.x) flavour
# the same marginalisation applies blockwise -- flat priors on
# (mu_y, mu_x) reparameterise with unit Jacobian to the regression
# intercept and mu_x, so the conditional contribution is the difference of
# two exchangeable conditionals. The subtracted x-block term evaluated on
# the transformed data carries its own theta-free constant, returned as
# `lx_const` (per group; x_idx gives each group's frozen exogenous
# columns).
loso_data_view <- function(lavmodel, lavdata, x_idx = NULL) {
  G <- lavdata@ngroups
  Y <- lavdata@X[seq_len(G)]
  if (isTRUE(lavmodel@meanstructure)) {
    return(list(Y = Y, l_const = numeric(G), lx_const = numeric(G)))
  }
  l_const <- lx_const <- numeric(G)
  for (g in seq_len(G)) {
    n <- nrow(Y[[g]])
    cc <- n / (n - 1)
    l_const[g] <- -0.5 * ncol(Y[[g]]) * log(cc)
    lx_const[g] <- 0.5 * length(x_idx[[g]]) * log(cc)
    Y[[g]] <- sqrt(cc) * sweep(Y[[g]], 2L, colMeans(Y[[g]]), "-")
  }
  list(Y = Y, l_const = l_const, lx_const = lx_const)
}

# Resolve the requested units of a single-level fit against its
# group-stacked rows. Units are identified by case index (the row number
# of the analysed dataset, as recorded in lavdata@case.idx), so a unit
# keeps its identity across fits that assign or order groups differently.
# Returns, in the requested order, each unit's id and group, plus the
# per-group gather/scatter maps: pos[[g]] indexes rows within group g's
# data block, rows[[g]] their positions in the output vectors.
loso_resolve_units <- function(lavdata, units) {
  G <- lavdata@ngroups
  ids_g <- lavdata@case.idx[seq_len(G)]
  ids <- unlist(ids_g)
  grp <- rep.int(seq_len(G), lengths(ids_g))
  if (is.null(units)) {
    sel <- seq_along(ids)
  } else {
    units <- as.integer(units)
    sel <- match(units, ids)
    if (length(units) == 0L || anyNA(sel) || anyDuplicated(units) > 0L) {
      cli_abort(
        "{.arg units} must be distinct case numbers of the analysed data
         (the row numbers recorded in the fit)."
      )
    }
  }
  pos_in_group <- sequence(lengths(ids_g))
  grp_sel <- grp[sel]
  list(
    ids = ids[sel],
    grp = grp_sel,
    rows = lapply(seq_len(G), function(g) which(grp_sel == g)),
    pos = lapply(seq_len(G), function(g) pos_in_group[sel][grp_sel == g]),
    n = length(sel)
  )
}

# ---- FIML (missing data): observed-data casewise kernels --------------------
#
# Under FIML the fitted likelihood is the observed-data likelihood, so each
# unit's contribution is Gaussian on the entries it actually has,
#   l_i = log N(y_i,obs ; mu[o_i], Sigma[o_i, o_i]),
# scoring the held-out unit on its observed entries with its full row removed
# from the conditioning set. Rows sharing a missing pattern share the observed
# set o_i, so the complete-data kernels evaluate verbatim on the subsetted
# moments; scores scatter back into full (mu, vech Sigma) coordinates (zero
# where unobserved, since l_i does not depend on the unobserved blocks). FIML
# forces a mean structure, so the saturated-means transformation never
# coincides with missingness and the iid kernels apply as-is; and lavaan
# rejects fixed.x with incomplete data, so only the joint flavour arises here.

# Rows grouped by missing pattern: each entry lists the row positions (within
# Y) sharing a pattern and that pattern's observed column indices.
fiml_patterns <- function(Y) {
  obs <- !is.na(Y)
  key <- apply(obs, 1L, paste0, collapse = "")
  lapply(
    split(seq_len(nrow(Y)), key),
    function(rows) list(rows = rows, o = which(obs[rows[1L], ]))
  )
}

# Column-major lower-triangle (lavaan vech) index of full-matrix entry (a, b)
vech_idx <- function(a, b, p) {
  lo <- pmin(a, b)
  hi <- pmax(a, b)
  (lo - 1L) * p - (lo - 1L) * lo / 2L + hi
}

# Vectorised per-row multivariate normal log-likelihood at fixed moments
mvn_loglik_rows <- function(Y, mu, Sigma) {
  p <- length(mu)
  Sigma_inv <- chol2inv(chol(Sigma))
  log_det <- as.numeric(determinant(Sigma, logarithm = TRUE)$modulus)
  D <- sweep(Y, 2L, mu, "-")
  quad <- rowSums((D %*% Sigma_inv) * D)
  as.numeric(-0.5 * (p * log(2 * pi) + log_det + quad))
}

# Per-row log-likelihood at fixed moments. Complete rows use the full moments;
# rows with missing entries are scored on their observed subset (the FIML
# observed-data contribution), grouping rows by pattern so each subset is
# factorised once.
loso_loglik_all <- function(Y, mom) {
  if (!anyNA(Y)) {
    return(mvn_loglik_rows(Y, mom$mu, mom$Sigma))
  }
  out <- numeric(nrow(Y))
  for (pat in fiml_patterns(Y)) {
    o <- pat$o
    out[pat$rows] <- mvn_loglik_rows(
      Y[pat$rows, o, drop = FALSE],
      mom$mu[o],
      mom$Sigma[o, o, drop = FALSE]
    )
  }
  out
}

# All-rows theta-scores at one theta from a single shared build; row i is
# the gradient of row i's log-likelihood w.r.t. packed theta. Y holds rows
# of one group; its Delta block has columns for all free parameters, so
# the result needs no cross-group assembly
# Complete-data (mu, vech Sigma)-coordinate scores, one row per observation;
# a lone row is duplicated because lavaan drops a single-row matrix internally
mvn_scores_rows <- function(Y, mu, Sigma) {
  n <- nrow(Y)
  Yc <- if (n == 1L) rbind(Y, Y) else Y
  sc <- lavaan___lav_mvnorm_scores_mu_vech_sigma(Yc, NULL, mu, Sigma)
  if (n == 1L) sc[1L, , drop = FALSE] else sc
}

# (mu, vech Sigma)-coordinate scores with missing entries: per pattern, the
# complete-data scores on the observed subset scattered into full coordinates
# (zero where unobserved)
fiml_scores_mu_vech <- function(Y, mom) {
  p <- length(mom$mu)
  SC <- matrix(0, nrow(Y), p + p * (p + 1L) / 2L)
  for (pat in fiml_patterns(Y)) {
    o <- pat$o
    po <- length(o)
    sck <- mvn_scores_rows(
      Y[pat$rows, o, drop = FALSE],
      mom$mu[o],
      mom$Sigma[o, o, drop = FALSE]
    )
    pr <- which(lower.tri(diag(po), diag = TRUE), arr.ind = TRUE)
    cols <- c(o, p + vech_idx(o[pr[, 1L]], o[pr[, 2L]], p))
    SC[pat$rows, cols] <- sck
  }
  SC
}

loso_scores_theta <- function(
  theta,
  Y,
  lavmodel,
  pt,
  cache = NULL,
  group = 1L
) {
  if (is.null(cache)) {
    cache <- loo_grad_cache(theta, lavmodel, pt, two_level = FALSE)
  }
  mom <- cache$mom[[group]]
  Delta_g <- cache$Delta[[group]]
  scores <- if (anyNA(Y)) {
    fiml_scores_mu_vech(Y, mom)
  } else {
    mvn_scores_rows(Y, mom$mu, mom$Sigma)
  }
  # Without an estimated mean structure Delta has only the vech(Sigma) rows,
  # so drop the leading Mu columns of the scores
  n_sig <- nrow(Delta_g)
  n_tot <- ncol(scores)
  if (n_tot > n_sig) {
    scores <- scores[, (n_tot - n_sig + 1L):n_tot, drop = FALSE]
  }
  loo_chain_rule(scores %*% Delta_g, cache)
}

# Stacked per-unit log-likelihoods over the resolved units (uv), scattered
# back into the requested order; includes the exchangeability constant of
# each unit's group (zero with a mean structure)
loso_loglik_units <- function(uv, dv, mom) {
  out <- numeric(uv$n)
  for (g in seq_along(dv$Y)) {
    rows <- uv$rows[[g]]
    if (length(rows) == 0L) {
      next
    }
    Yg <- dv$Y[[g]][uv$pos[[g]], , drop = FALSE]
    out[rows] <- loso_loglik_all(Yg, mom[[g]]) + dv$l_const[g]
  }
  out
}

# Stacked per-unit theta-scores over the resolved units at one theta; the
# per-theta build (cache) is shared across groups
loso_scores_units <- function(theta, uv, dv, lavmodel, pt, cache = NULL) {
  if (is.null(cache)) {
    cache <- loo_grad_cache(theta, lavmodel, pt, two_level = FALSE)
  }
  out <- matrix(0, uv$n, length(theta))
  for (g in seq_along(dv$Y)) {
    rows <- uv$rows[[g]]
    if (length(rows) == 0L) {
      next
    }
    Yg <- dv$Y[[g]][uv$pos[[g]], , drop = FALSE]
    out[rows, ] <- loso_scores_theta(theta, Yg, lavmodel, pt, cache, group = g)
  }
  out
}

# ---- LOCO (two level): per-cluster sufficient stats, loglik, theta-scores --

loco_suff_stats <- function(lavdata) {
  Lp <- lavdata@Lp[[1L]]
  cluster_idx <- Lp$cluster.idx[[2L]]
  X <- lavdata@X[[1L]]

  J <- Lp$nclusters[[2L]]
  p <- ncol(X)
  y_idx <- c(Lp$both.idx[[2L]], Lp$within.idx[[2L]])
  z_idx <- Lp$between.idx[[2L]]
  zy_idx <- c(z_idx, y_idx)
  d <- length(zy_idx)

  n_j <- integer(J)
  S <- vector("list", J)
  mean_d <- vector("list", J)
  ybar <- matrix(0, J, p)
  for (j in seq_len(J)) {
    Yj <- X[cluster_idx == j, , drop = FALSE]
    n_j[j] <- nrow(Yj)
    ybar[j, ] <- colMeans(Yj)
    Yc <- sweep(Yj, 2L, ybar[j, ], "-")
    S[[j]] <- crossprod(Yc)
    mean_d[[j]] <- ybar[j, zy_idx]
  }
  list(
    J = J,
    n_j = n_j,
    S = S,
    mean_d = mean_d,
    Lp = Lp,
    p = p,
    d = d,
    ybar = ybar,
    zy_idx = zy_idx,
    cluster_idx = cluster_idx
  )
}

# Single-cluster sample statistics in the (YLp, Lp) shapes that lavaan's
# two-level loglik/gradient kernels expect
loco_stats_build <- function(nj, S_j, mean_d_j, css) {
  S_w_j <- if (nj > 1L) S_j / (nj - 1L) else matrix(0, css$p, css$p)
  YLp_j <- list(
    NULL,
    list(
      Sigma.W = S_w_j,
      cov.d = list(matrix(0, css$d, css$d)),
      mean.d = list(mean_d_j),
      loglik.x = 0
    )
  )
  Lp_j <- css$Lp
  Lp_j$nclusters[[1L]] <- nj
  Lp_j$nclusters[[2L]] <- 1L
  Lp_j$cluster.size[[2L]] <- nj
  Lp_j$cluster.sizes[[2L]] <- nj
  Lp_j$ncluster.sizes[[2L]] <- 1L
  Lp_j$cluster.size.ns[[2L]] <- 1L
  list(YLp = YLp_j, Lp = Lp_j)
}

loco_unit_stats <- function(j, css) {
  loco_stats_build(css$n_j[j], css$S[[j]], css$mean_d[[j]], css)
}

loco_loglik_us <- function(us, mom) {
  as.numeric(lavaan___lav_mvnorm_cluster_loglik_samplestats_2l(
    us$YLp,
    us$Lp,
    mom$mu_w,
    mom$Sigma_w,
    mom$mu_b,
    mom$Sigma_b,
    "eigen",
    TRUE, # log2pi
    FALSE # minus.two
  ))
}

loco_loglik_one <- function(j, css, mom) {
  loco_loglik_us(loco_unit_stats(j, css), mom)
}

loco_grad_x_us <- function(us, mom, Delta) {
  DX <- lavaan___lav_mvnorm_cluster_dlogl_2l_samplestats(
    us$YLp,
    us$Lp,
    mom$mu_w,
    mom$Sigma_w,
    mom$mu_b,
    mom$Sigma_b
  )
  # dlogl is the derivative of -2 * loglik w.r.t. the stacked moments
  -0.5 * as.numeric(DX %*% Delta[[1L]])
}

loco_grad_x_one <- function(j, css, mom, Delta) {
  loco_grad_x_us(loco_unit_stats(j, css), mom, Delta)
}

# All-cluster theta-scores at one theta; the moment/Delta build is shared,
# the per-cluster gradient kernel does not vectorise so loops over units
loco_scores_theta <- function(theta, css, lavmodel, pt, units, cache = NULL) {
  if (is.null(cache)) {
    cache <- loo_grad_cache(theta, lavmodel, pt, two_level = TRUE)
  }
  q <- ncol(cache$Delta[[1L]])
  G_x <- t(vapply(
    units,
    function(j) loco_grad_x_one(j, css, cache$mom, cache$Delta),
    numeric(q)
  ))
  loo_chain_rule(G_x, cache)
}

# ---- LOCO under missing data (two-level FIML) ------------------------------
#
# With incomplete data a cluster's sufficient statistics (n_j, ybar_j, S_j) no
# longer determine its observed-data likelihood -- rows in a cluster have
# different observed patterns -- so the samplestats kernels above do not
# apply. lavaan exposes raw-data kernels for exactly this case (the same ones
# it fits two-level FIML with), which we reuse per cluster: the same kernel
# swap as the complete-data path, fed each cluster's raw rows and missing
# pattern object instead of its sufficient statistics. Because LOCO deletes a
# whole cluster, no downdating is needed. FIML forces joint scoring (lavaan
# rejects two-level + conditional.x), so only the joint flavour arises.

# Single-cluster missing-kernel inputs from a level-1 data block Y1j (raw rows
# of one cluster, possibly with a row removed for the per-row override), its
# level-2 row Y2j, the full-data Lp template and the between-only columns. The
# missing-pattern object is (re)built on Y1j.
#
# Rows fully missing on the within variables are dropped first: they
# contribute nothing to the cluster's marginal likelihood (a zero-observed
# pattern integrates to 1), so the loglik is unchanged, but lavaan keeps such
# rows in two-level FIML and its analytic gradient kernel mishandles the
# zero-observed pattern (the loglik kernel is fine). Dropping them keeps both
# kernels exact.
loco_missing_build_cj <- function(Y1j, Y2j, Lp, between_idx) {
  within_cols <- if (length(between_idx) > 0L) {
    seq_len(ncol(Y1j))[-between_idx]
  } else {
    seq_len(ncol(Y1j))
  }
  obs_row <- rowSums(!is.na(Y1j[, within_cols, drop = FALSE])) > 0L
  if (!all(obs_row)) {
    Y1j <- Y1j[obs_row, , drop = FALSE]
  }
  nj <- nrow(Y1j)
  Lpj <- Lp
  Lpj$nclusters[[1L]] <- nj
  Lpj$nclusters[[2L]] <- 1L
  Lpj$cluster.size[[2L]] <- nj
  Lpj$cluster.sizes[[2L]] <- nj
  Lpj$ncluster.sizes[[2L]] <- 1L
  Lpj$cluster.size.ns[[2L]] <- 1L
  Lpj$cluster.idx[[2L]] <- rep.int(1L, nj)
  Ywj <- if (length(between_idx) > 0L) {
    Y1j[, -between_idx, drop = FALSE]
  } else {
    Y1j
  }
  Mpj <- lavaan___lav_data_missing_patterns(Ywj, FALSE, FALSE, Lpj)
  list(Y1 = Y1j, Y2 = Y2j, Lp = Lpj, Mp = Mpj)
}

# Theta-independent inputs for the missing-data kernels, built once and reused
# at every theta: per-cluster single-cluster inputs (Y1, Y2, Lp, Mp), plus the
# data and maps needed by the per-row override (raw X, cluster index, and each
# row's position within its cluster). The implied moments are passed in
# lavaan's flat (mu_w, Sigma_w, mu_b, Sigma_b) view, which is cache$mom.
loco_missing_info <- function(int) {
  Lp <- int$lavdata@Lp[[1L]]
  X <- int$lavdata@X[[1L]]
  Y2 <- int$lavsamplestats@YLp[[1L]][[2L]]$Y2
  cl <- Lp$cluster.idx[[2L]]
  J <- Lp$nclusters[[2L]]
  between_idx <- Lp$between.idx[[2L]]
  within_cols <- if (length(between_idx) > 0L) {
    seq_len(ncol(X))[-between_idx]
  } else {
    seq_len(ncol(X))
  }
  row_observed <- rowSums(!is.na(X[, within_cols, drop = FALSE])) > 0L
  rows_by_cluster <- lapply(seq_len(J), function(j) which(cl == j))
  clusters <- lapply(seq_len(J), function(j) {
    loco_missing_build_cj(
      X[rows_by_cluster[[j]], , drop = FALSE],
      Y2[j, , drop = FALSE],
      Lp,
      between_idx
    )
  })
  list(
    clusters = clusters,
    J = J,
    n_j = tabulate(cl, J),
    # observed (not fully-missing) rows per cluster: the effective unit count
    # the per-row override needs for its singleton guard
    n_obs = vapply(
      rows_by_cluster,
      function(r) sum(row_observed[r]),
      integer(1)
    ),
    X = X,
    rows_by_cluster = rows_by_cluster,
    Lp = Lp,
    cl = cl,
    between_idx = between_idx
  )
}

loco_missing_loglik_cj <- function(cj, mom) {
  as.numeric(lavaan___lav_mvnorm_cluster_missing_loglik_samplestats_2l(
    cj$Y1,
    cj$Y2,
    cj$Lp,
    cj$Mp,
    mom$mu_w,
    mom$Sigma_w,
    mom$mu_b,
    mom$Sigma_b,
    "eigen", # Sinv.method
    TRUE, # log2pi
    0, # loglik.x
    FALSE # minus.two
  ))
}

loco_missing_loglik_one <- function(j, minfo, mom) {
  loco_missing_loglik_cj(minfo$clusters[[j]], mom)
}

loco_missing_grad_cj <- function(cj, mom, Delta) {
  DX <- lavaan___lav_mvnorm_cluster_missing_dlogl_2l_samplestats(
    cj$Y1,
    cj$Y2,
    cj$Lp,
    cj$Mp,
    mom$mu_w,
    mom$Sigma_w,
    mom$mu_b,
    mom$Sigma_b
  )
  # dlogl is the derivative of -2 * loglik w.r.t. the stacked moments
  -0.5 * as.numeric(DX %*% Delta[[1L]])
}

loco_missing_grad_x_one <- function(j, minfo, mom, Delta) {
  loco_missing_grad_cj(minfo$clusters[[j]], mom, Delta)
}

# All-cluster theta-scores at one theta (missing-data analogue of
# loco_scores_theta); the moment/Delta build is shared across clusters
loco_missing_scores_theta <- function(
  theta,
  minfo,
  lavmodel,
  pt,
  units,
  cache = NULL
) {
  if (is.null(cache)) {
    cache <- loo_grad_cache(theta, lavmodel, pt, two_level = TRUE)
  }
  q <- ncol(cache$Delta[[1L]])
  G_x <- t(vapply(
    units,
    function(j) loco_missing_grad_x_one(j, minfo, cache$mom, cache$Delta),
    numeric(q)
  ))
  loo_chain_rule(G_x, cache)
}

# ---- Per-row deletion under two-level missing data (LOSO override) ----------
#
# The leave-one-unit-out contribution of row i is the conditional density of
# its observed entries given the rest of its cluster (Merkle, Furr &
# Rabe-Hesketh, 2019), l_i = ll_j(full cluster) - ll_j(cluster minus row i),
# both observed-data cluster marginals from the missing kernel. "Minus row i"
# drops the raw row and rebuilds the pattern object (no sufficient-statistic
# downdating); a singleton cluster's only row is the cluster, l_i = ll_j(full).

# Cluster of row i with that row removed: rebuild from the cluster's original
# rows minus i (loco_missing_build_cj then drops any fully-missing rows). If i
# is itself fully missing, the observed data is unchanged and l_i collapses to
# zero automatically.
loco_missing_minus_row <- function(i, minfo) {
  j <- minfo$cl[i]
  keep <- setdiff(minfo$rows_by_cluster[[j]], i)
  loco_missing_build_cj(
    minfo$X[keep, , drop = FALSE],
    minfo$clusters[[j]]$Y2,
    minfo$Lp,
    minfo$between_idx
  )
}

loso2l_missing_loglik_all <- function(units, minfo, mom) {
  cl <- minfo$cl[units]
  need <- unique(cl)
  ll_full <- vapply(
    need,
    function(j) loco_missing_loglik_one(j, minfo, mom),
    numeric(1)
  )
  vapply(
    seq_along(units),
    function(u) {
      ll_j <- ll_full[match(cl[u], need)]
      # a cluster with <= 1 observed row has nothing to condition on
      if (minfo$n_obs[cl[u]] <= 1L) {
        return(ll_j)
      }
      ll_j -
        loco_missing_loglik_cj(loco_missing_minus_row(units[u], minfo), mom)
    },
    numeric(1)
  )
}

loso2l_missing_scores_theta <- function(
  theta,
  minfo,
  lavmodel,
  pt,
  units,
  cache = NULL
) {
  if (is.null(cache)) {
    cache <- loo_grad_cache(theta, lavmodel, pt, two_level = TRUE)
  }
  q <- ncol(cache$Delta[[1L]])
  cl <- minfo$cl[units]
  need <- unique(cl)
  g_full <- vapply(
    need,
    function(j) loco_missing_grad_x_one(j, minfo, cache$mom, cache$Delta),
    numeric(q)
  )
  G_x <- t(vapply(
    seq_along(units),
    function(u) {
      g_j <- g_full[, match(cl[u], need)]
      if (minfo$n_obs[cl[u]] <= 1L) {
        return(g_j)
      }
      g_j -
        loco_missing_grad_cj(
          loco_missing_minus_row(units[u], minfo),
          cache$mom,
          cache$Delta
        )
    },
    numeric(q)
  ))
  loo_chain_rule(G_x, cache)
}

# ---- Row deletion on two-level models (LOSO override) -----------------------
#
# Removing row i from cluster j changes the dataset likelihood by
#   l_i(theta) = ll_j(theta; full cluster) - ll_j(theta; cluster minus row i),
# i.e. the conditional density of the row given the remaining rows in its
# cluster. Both terms use the cluster kernels above, with the cluster's
# sufficient statistics downdated for the removed row. A singleton cluster's
# only row is the cluster itself: l_i = ll_j(full).

# Sufficient statistics of cluster `j = cluster_idx[i]` with row i removed
loso2l_minus_stats <- function(i, css, X) {
  j <- css$cluster_idx[i]
  n <- css$n_j[j]
  y_i <- X[i, ]
  ybar <- css$ybar[j, ]
  dvec <- y_i - ybar
  S_minus <- css$S[[j]] - (n / (n - 1)) * tcrossprod(dvec)
  ybar_minus <- (n * ybar - y_i) / (n - 1)
  loco_stats_build(n - 1L, S_minus, ybar_minus[css$zy_idx], css)
}

loso2l_loglik_all <- function(units, css, X, mom) {
  cl <- css$cluster_idx[units]
  need <- unique(cl)
  ll_full <- vapply(
    need,
    function(j) loco_loglik_one(j, css, mom),
    numeric(1)
  )
  vapply(
    seq_along(units),
    function(u) {
      i <- units[u]
      j <- cl[u]
      ll_j <- ll_full[match(j, need)]
      if (css$n_j[j] == 1L) {
        return(ll_j)
      }
      ll_j - loco_loglik_us(loso2l_minus_stats(i, css, X), mom)
    },
    numeric(1)
  )
}

loso2l_scores_theta <- function(
  theta,
  css,
  X,
  lavmodel,
  pt,
  units,
  cache = NULL
) {
  if (is.null(cache)) {
    cache <- loo_grad_cache(theta, lavmodel, pt, two_level = TRUE)
  }
  q <- ncol(cache$Delta[[1L]])
  cl <- css$cluster_idx[units]
  need <- unique(cl)
  g_full <- vapply(
    need,
    function(j) loco_grad_x_one(j, css, cache$mom, cache$Delta),
    numeric(q)
  )
  G_x <- t(vapply(
    seq_along(units),
    function(u) {
      i <- units[u]
      j <- cl[u]
      g_j <- g_full[, match(j, need)]
      if (css$n_j[j] == 1L) {
        return(g_j)
      }
      g_j -
        loco_grad_x_us(loso2l_minus_stats(i, css, X), cache$mom, cache$Delta)
    },
    numeric(q)
  ))
  loo_chain_rule(G_x, cache)
}

# ---- Per-unit Hessians and the Taylor formulas ------------------------------

# Per-unit Hessians of the unit log-likelihoods via central finite differences
# of the analytic scores, batched across units: the perturbed thetas do not
# depend on the unit, so the expensive per-theta build (implied moments,
# Delta, transformation Jacobian) inside score_fn is shared by every unit at
# each of the 2m points. score_fn(theta_act) must return an (n_units x m)
# score matrix from one shared build. Returns an m x m x n_units array of
# symmetrised Hessians.
loo_batched_hessians <- function(
  theta_act,
  score_fn,
  n_units,
  h = 1e-5,
  cores = 1L,
  verbose = FALSE
) {
  m <- length(theta_act)
  one_dir <- function(k) {
    tp <- theta_act
    tm <- theta_act
    tp[k] <- tp[k] + h
    tm[k] <- tm[k] - h
    (score_fn(tp) - score_fn(tm)) / (2 * h)
  }
  # cols[[k]] is n_units x m: column k of every unit's Hessian
  cols <- run_parallel_or_serial(
    m = m,
    FUN = one_dir,
    cores = cores,
    verbose = verbose,
    msg_serial = "Differentiating unit scores {j}/{m} direction{?s}.",
    msg_parallel = "Differentiating unit scores in {m} directions ({cores} cores).",
    msg_done = "Differentiate unit scores in {m} direction{?s}."
  )
  H_arr <- array(0, dim = c(m, m, n_units))
  for (k in seq_len(m)) {
    H_arr[, k, ] <- t(cols[[k]])
  }
  for (u in seq_len(n_units)) {
    H_u <- H_arr[,, u]
    H_arr[,, u] <- 0.5 * (H_u + t(H_u))
  }
  H_arr
}

# First- and second-order Taylor case-deletion (CPO) and full-posterior (LPD)
# predictive densities for one unit. A second-order term is voided (NA) when
# its curvature matrix is not positive definite; `ok` records the CPO status.
taylor_loo_unit <- function(
  l_star,
  s_u,
  H_u,
  S_act,
  S_inv,
  second_order = TRUE
) {
  d <- length(s_u)
  quad1 <- as.numeric(crossprod(s_u, S_act %*% s_u))
  out <- list(
    log_cpo_1 = l_star - 0.5 * quad1,
    lpd_1 = l_star + 0.5 * quad1,
    log_cpo_2 = NA_real_,
    lpd_2 = NA_real_,
    det_term = NA_real_,
    ok = FALSE
  )
  if (second_order && !is.null(H_u)) {
    A_u <- S_inv + H_u
    Ac <- tryCatch(chol(A_u), error = function(e) NULL)
    if (!is.null(Ac)) {
      det_a <- determinant(diag(d) + S_act %*% H_u, logarithm = TRUE)
      if (is.finite(det_a$modulus) && det_a$sign > 0) {
        out$ok <- TRUE
        out$det_term <- 0.5 * as.numeric(det_a$modulus)
        out$log_cpo_2 <- l_star -
          0.5 * as.numeric(crossprod(s_u, chol2inv(Ac) %*% s_u)) +
          out$det_term
      }
    }
    B_u <- S_inv - H_u
    Bc <- tryCatch(chol(B_u), error = function(e) NULL)
    if (!is.null(Bc)) {
      det_b <- determinant(diag(d) - S_act %*% H_u, logarithm = TRUE)
      if (is.finite(det_b$modulus) && det_b$sign > 0) {
        out$lpd_2 <- l_star +
          0.5 * as.numeric(crossprod(s_u, chol2inv(Bc) %*% s_u)) -
          0.5 * as.numeric(det_b$modulus)
      }
    }
  }
  out
}

# Insert a group column (labels when available) after `unit` for multigroup
# fits; single-group tables keep their original shape
add_loo_group_column <- function(per_unit, unit_group, lavdata) {
  if (is.null(unit_group) || lavdata@ngroups == 1L) {
    return(per_unit)
  }
  labs <- lavdata@group.label
  grp <- if (length(labs) == lavdata@ngroups) {
    labs[unit_group]
  } else {
    unit_group # nocov
  }
  cbind(per_unit[1L], group = grp, per_unit[-1L])
}

# Parallelism for loo() is strictly opt-in: NULL means serial (the m > 120
# auto rule used for marginal fitting does not apply here)
resolve_loo_cores <- function(cores) {
  if (is.null(cores)) {
    return(1L)
  }
  cores <- as.integer(cores)
  if (is.na(cores) || cores < 1L) {
    cores <- 1L
  }
  if (cores > 1L && .Platform$OS.type == "windows") {
    # nocov start
    cli_alert_warning(
      "Parallel LOO uses forking, which is not available on Windows.
       Continuing serially."
    )
    cores <- 1L
  } # nocov end
  cores
}

# ---- Workhorse ---------------------------------------------------------------

# The flavour of the score follows the fitted likelihood: a fixed.x fit is
# scored on its own conditional likelihood, a fit with modelled covariates
# on the joint one
loo_flavour <- function(int) {
  if (
    isTRUE(int$lavmodel@fixed.x) &&
      length(unlist(int$lavsamplestats@x.idx)) > 0L
  ) {
    "conditional"
  } else {
    "joint"
  }
}

# The frozen-covariate marginal log-density per unit: the theta-free constant
# separating the joint kernel value of a fixed.x fit from its conditional
# likelihood. (The conditional likelihood itself is exactly invariant to the
# frozen covariate moments, which cancel from the implied conditional
# moments; only this additive constant distinguishes the two.) Evaluated
# blockwise on each unit's own group: the x-block term on the transformed
# data carries its group's theta-free constant (dv$lx_const), folded in
# here so that subtracting the result converts a joint kernel value into
# the conditional likelihood under either mean treatment. For the
# two-level row-deletion override the analogous adjustment is the difference
# of the cluster constant before and after removing the row (zero when all
# covariates are cluster-level, since z_j appears once in both terms).
loso_fixedx_const_units <- function(int, uv, dv, mom) {
  x_idx <- int$lavsamplestats@x.idx
  out <- numeric(uv$n)
  for (g in seq_along(dv$Y)) {
    rows <- uv$rows[[g]]
    xg <- x_idx[[g]]
    if (length(rows) == 0L || length(xg) == 0L) {
      next
    }
    Yg <- dv$Y[[g]][uv$pos[[g]], xg, drop = FALSE]
    out[rows] <- loso_loglik_all(
      Yg,
      list(
        mu = mom[[g]]$mu[xg],
        Sigma = mom[[g]]$Sigma[xg, xg, drop = FALSE]
      )
    ) -
      dv$lx_const[g]
  }
  out
}

# Positions and frozen blocks needed for the two-level covariate marginal:
# z = cluster-level (between-only) covariates, v = within-side covariates
# (with or without a between presence)
loo_fixedx_info_loco <- function(int, css, mom) {
  ovn <- int$lavdata@ov.names[[1L]]
  l1 <- int$lavdata@ov.names.l[[1L]][[1L]]
  l2 <- int$lavdata@ov.names.l[[1L]][[2L]]
  x_idx <- int$lavsamplestats@x.idx[[1L]]
  z_cols <- intersect(x_idx, css$Lp$between.idx[[2L]])
  v_cols <- setdiff(x_idx, z_cols)
  qz <- length(z_cols)
  qv <- length(v_cols)
  v1 <- match(ovn[v_cols], l1)
  z2 <- match(ovn[z_cols], l2)
  v2 <- match(ovn[v_cols], l2)
  Sw_vv <- mom$Sigma_w[v1, v1, drop = FALSE]
  mu_zv <- c(
    mom$mu_b[z2],
    if (qv > 0L && anyNA(v2)) mom$mu_w[v1] else mom$mu_b[v2]
  )
  B0 <- matrix(0, qz + qv, qz + qv)
  if (qz > 0L) {
    B0[seq_len(qz), seq_len(qz)] <- mom$Sigma_b[z2, z2]
  }
  if (qv > 0L && !anyNA(v2)) {
    if (qz > 0L) {
      B0[seq_len(qz), qz + seq_len(qv)] <- mom$Sigma_b[z2, v2]
      B0[qz + seq_len(qv), seq_len(qz)] <- mom$Sigma_b[v2, z2]
    }
    B0[qz + seq_len(qv), qz + seq_len(qv)] <- mom$Sigma_b[v2, v2]
  }
  list(
    z_cols = z_cols,
    v_cols = v_cols,
    qz = qz,
    qv = qv,
    Sw_vv = Sw_vv,
    logdet_Sw_vv = if (qv > 0L) {
      as.numeric(determinant(Sw_vv, logarithm = TRUE)$modulus)
    } else {
      0
    },
    mu_zv = mu_zv,
    B0 = B0
  )
}

# Frozen marginal log-density of one cluster's covariate data from its raw
# sufficient statistics: within deviations of v (n_j - 1 degrees of freedom)
# plus the joint between/mean density of (z_j, vbar_j) with the within
# mean-uncertainty inflation Sw_vv / n_j
loo_fixedx_const_cluster <- function(nj, ybar_j, S_j, info) {
  out <- 0
  B <- info$B0
  if (info$qv > 0L) {
    A_vv <- S_j[info$v_cols, info$v_cols, drop = FALSE]
    out <- out -
      0.5 * (nj - 1) * (info$qv * log(2 * pi) + info$logdet_Sw_vv) -
      0.5 * sum(diag(solve(info$Sw_vv, A_vv))) -
      0.5 * info$qv * log(nj)
    idx <- info$qz + seq_len(info$qv)
    B[idx, idx] <- B[idx, idx] + info$Sw_vv / nj
  }
  obs <- c(ybar_j[info$z_cols], ybar_j[info$v_cols])
  ch <- chol(B)
  d <- backsolve(ch, obs - info$mu_zv, transpose = TRUE)
  out -
    0.5 *
      (length(obs) * log(2 * pi) + 2 * sum(log(diag(ch))) + sum(d^2))
}

loo_fixedx_const_loco <- function(int, css, units, mom) {
  info <- loo_fixedx_info_loco(int, css, mom)
  vapply(
    units,
    function(j) {
      loo_fixedx_const_cluster(css$n_j[j], css$ybar[j, ], css$S[[j]], info)
    },
    numeric(1)
  )
}

# Row-deletion analogue for the two-level LOSO override: the covariate part
# of row i's contribution is c_j(full cluster) - c_j(cluster minus row i),
# via the same mean/scatter downdates; a singleton cluster's only row
# removes the whole cluster, so its covariate part is c_j itself
loo_fixedx_rowdiff_loco <- function(int, css, X, units, mom) {
  info <- loo_fixedx_info_loco(int, css, mom)
  vapply(
    units,
    function(i) {
      j <- css$cluster_idx[i]
      n <- css$n_j[j]
      c_full <- loo_fixedx_const_cluster(n, css$ybar[j, ], css$S[[j]], info)
      if (n == 1L) {
        return(c_full)
      }
      y_i <- X[i, ]
      ybar_m <- (n * css$ybar[j, ] - y_i) / (n - 1)
      S_m <- css$S[[j]] - (n / (n - 1)) * tcrossprod(y_i - css$ybar[j, ])
      c_full - loo_fixedx_const_cluster(n - 1L, ybar_m, S_m, info)
    },
    numeric(1)
  )
}

# Shared validation gate for the casewise log-likelihood machinery
check_loo_model <- function(int, fn = "loo") {
  lavmodel <- int$lavmodel
  lavdata <- int$lavdata
  lavsamplestats <- int$lavsamplestats
  if (lavdata@ngroups > 1L && is_multilevel(lavdata)) {
    cli_abort(
      "{.fn {fn}} does not support multigroup two-level models yet."
    )
  }
  if (lavmodel@estimator != "ML" || isTRUE(lavmodel@categorical)) {
    cli_abort(
      "{.fn {fn}} requires a continuous-data model fitted with the
       {.val ML} estimator."
    )
  }
  # Missing data is supported for single-level (FIML observed-data casewise
  # kernels) and for two-level per-cluster scoring (LOCO, via lavaan's
  # raw-data missing kernels); the two-level per-row override is gated
  # separately in inlav_loo() because its sufficient-statistic downdating
  # breaks under missingness.
  if (isTRUE(lavmodel@conditional.x)) {
    cli_abort("{.fn {fn}} does not support {.code conditional.x = TRUE}.")
  }
  invisible(NULL)
}

inlav_loo <- function(
  int,
  type = c("auto", "loso", "loco"),
  units = NULL,
  second_order = TRUE,
  theta = NULL,
  Sigma = NULL,
  eff_cores = 1L,
  verbose = FALSE,
  max_seconds = Inf
) {
  type <- match.arg(type)
  pt <- int$partable
  lavmodel <- int$lavmodel
  lavdata <- int$lavdata

  check_loo_model(int, fn = "loo")
  flavour <- loo_flavour(int)

  # Resolve LOSO (per-row) vs LOCO (per-cluster)
  two_level <- is_multilevel(lavdata)
  if (type == "auto") {
    type <- if (two_level) "loco" else "loso"
  } else if (type == "loco" && !two_level) {
    cli_abort(
      "Leave-one-cluster-out requires a two-level model, but this model has
       no clusters."
    )
  } else if (type == "loso" && two_level) {
    cli_warn(c(
      "Scoring leave-one-unit-out (the {.emph conditional} predictive) on a
       two-level model, not the default leave-one-cluster-out (the
       {.emph marginal} predictive).",
      "i" = "These target different predictions -- a new observation within an
       observed cluster vs a new cluster -- and are easily conflated
       (Merkle, Furr & Rabe-Hesketh, 2019). This diagnostic is expensive;
       consider subsetting with {.arg units}."
    ))
  }

  # Posterior summary to score at (defaults to the fit's own Laplace summary)
  theta_star <- int$theta_star
  m <- length(theta_star)
  theta_overridden <- !is.null(theta) || !is.null(Sigma)
  if (is.null(theta)) {
    theta <- theta_star
  }
  if (is.null(Sigma)) {
    Sigma <- int$Sigma_theta
  }
  theta <- as.numeric(theta)
  if (length(theta) != m) {
    cli_abort(
      "{.arg theta} must have length {m} (the number of free parameters)."
    )
  }
  if (!is.matrix(Sigma) || !all(dim(Sigma) == c(m, m))) {
    cli_abort("{.arg Sigma} must be a {m} x {m} covariance matrix.")
  }

  # Active (free) subspace: after Gaussian conditioning on a locked parameter
  # the corresponding diagonal of Sigma is exactly zero and Sigma is singular.
  # The second-order formula needs Sigma^{-1}, which exists only on this
  # block; restricting is exact (locked precisions are infinite, zeroing
  # those dimensions) and is a no-op for an unconditioned summary.
  free <- which(diag(Sigma) > .Machine$double.eps)
  S_act <- Sigma[free, free, drop = FALSE]
  S_inv <- chol2inv(chol(S_act))

  if (isTRUE(verbose)) {
    cli_progress_step(
      "Computing unit log-likelihoods and scores.",
      msg_done = "Compute unit log-likelihoods and scores."
    )
  }
  unit_group <- NULL # group of each scored unit; multigroup LOSO only
  if (type == "loso" && two_level && isTRUE(int$lavsamplestats@missing.flag)) {
    # Per-row deletion under FIML: each row's conditional density via the
    # missing kernel, dropping the raw row and rebuilding its pattern object
    minfo <- loco_missing_info(int)
    units <- check_loo_units(units, length(minfo$cl), "rows")
    cache <- loo_grad_cache(theta, lavmodel, pt, two_level = TRUE)
    l_star <- loso2l_missing_loglik_all(units, minfo, cache$mom)
    s_mat <- loso2l_missing_scores_theta(
      theta,
      minfo,
      lavmodel,
      pt,
      units,
      cache
    )
    score_fn <- function(th_act) {
      th <- theta
      th[free] <- th_act
      loso2l_missing_scores_theta(th, minfo, lavmodel, pt, units)[,
        free,
        drop = FALSE
      ]
    }
    nobs <- rep(1L, length(units))
  } else if (type == "loso" && two_level) {
    X <- lavdata@X[[1L]]
    css <- loco_suff_stats(lavdata)
    units <- check_loo_units(units, nrow(X), "rows")
    cache <- loo_grad_cache(theta, lavmodel, pt, two_level = TRUE)
    l_star <- loso2l_loglik_all(units, css, X, cache$mom)
    if (flavour == "conditional") {
      l_star <- l_star - loo_fixedx_rowdiff_loco(int, css, X, units, cache$mom)
    }
    s_mat <- loso2l_scores_theta(theta, css, X, lavmodel, pt, units, cache)
    score_fn <- function(th_act) {
      th <- theta
      th[free] <- th_act
      loso2l_scores_theta(th, css, X, lavmodel, pt, units)[,
        free,
        drop = FALSE
      ]
    }
    nobs <- rep(1L, length(units))
  } else if (type == "loso") {
    dv <- loso_data_view(lavmodel, lavdata, x_idx = int$lavsamplestats@x.idx)
    uv <- loso_resolve_units(lavdata, units)
    cache <- loo_grad_cache(theta, lavmodel, pt, two_level = FALSE)
    l_star <- loso_loglik_units(uv, dv, cache$mom)
    if (flavour == "conditional") {
      l_star <- l_star - loso_fixedx_const_units(int, uv, dv, cache$mom)
    }
    s_mat <- loso_scores_units(theta, uv, dv, lavmodel, pt, cache = cache)
    score_fn <- function(th_act) {
      th <- theta
      th[free] <- th_act
      loso_scores_units(th, uv, dv, lavmodel, pt)[, free, drop = FALSE]
    }
    units <- uv$ids
    unit_group <- uv$grp
    nobs <- rep(1L, uv$n)
  } else if (isTRUE(int$lavsamplestats@missing.flag)) {
    # Two-level FIML: per-cluster observed-data kernels (joint flavour only)
    minfo <- loco_missing_info(int)
    units <- check_loo_units(units, minfo$J, "clusters")
    cache <- loo_grad_cache(theta, lavmodel, pt, two_level = TRUE)
    l_star <- vapply(
      units,
      function(j) loco_missing_loglik_one(j, minfo, cache$mom),
      numeric(1)
    )
    s_mat <- loco_missing_scores_theta(theta, minfo, lavmodel, pt, units, cache)
    score_fn <- function(th_act) {
      th <- theta
      th[free] <- th_act
      loco_missing_scores_theta(th, minfo, lavmodel, pt, units)[,
        free,
        drop = FALSE
      ]
    }
    nobs <- minfo$n_j[units]
  } else {
    css <- loco_suff_stats(lavdata)
    units <- check_loo_units(units, css$J, "clusters")
    cache <- loo_grad_cache(theta, lavmodel, pt, two_level = TRUE)
    l_star <- vapply(
      units,
      function(j) loco_loglik_one(j, css, cache$mom),
      numeric(1)
    )
    if (flavour == "conditional") {
      l_star <- l_star - loo_fixedx_const_loco(int, css, units, cache$mom)
    }
    s_mat <- loco_scores_theta(theta, css, lavmodel, pt, units, cache)
    score_fn <- function(th_act) {
      th <- theta
      th[free] <- th_act
      loco_scores_theta(th, css, lavmodel, pt, units)[, free, drop = FALSE]
    }
    nobs <- css$n_j[units]
  }
  s_mat <- s_mat[, free, drop = FALSE]
  n_units <- length(units)

  H_arr <- NULL
  if (isTRUE(second_order)) {
    # Budget gate (used by the fit-time default path): the Hessian stage
    # costs 2 * m_free score-matrix evaluations, so timing a single
    # evaluation predicts the total before committing to it
    if (is.finite(max_seconds)) {
      t_one <- as.numeric(system.time(score_fn(theta[free]))["elapsed"])
      # floor at clock resolution so a fast evaluation never predicts zero
      t_hat <- 1.2 * 2 * length(free) * max(t_one, 1e-4)
      if (t_hat > max_seconds) {
        cli_abort(
          "Predicted serial LOO cost ({round(t_hat, 1)} s) exceeds the
           {max_seconds} s budget.",
          class = "inlavaan_loo_budget"
        )
      }
    }
    H_arr <- loo_batched_hessians(
      theta[free],
      score_fn,
      n_units = n_units,
      cores = eff_cores,
      verbose = verbose
    )
  }

  raw <- lapply(seq_len(n_units), function(u) {
    taylor_loo_unit(
      l_star[u],
      s_mat[u, ],
      if (isTRUE(second_order)) H_arr[,, u] else NULL,
      S_act,
      S_inv,
      second_order = second_order
    )
  })

  per_unit <- data.frame(
    unit = units,
    nobs = nobs,
    l_star = l_star,
    score_norm = sqrt(rowSums(s_mat^2)),
    lpd_1 = vapply(raw, `[[`, numeric(1), "lpd_1"),
    lpd_2 = vapply(raw, `[[`, numeric(1), "lpd_2"),
    log_cpo_1 = vapply(raw, `[[`, numeric(1), "log_cpo_1"),
    log_cpo_2 = vapply(raw, `[[`, numeric(1), "log_cpo_2"),
    det_term = vapply(raw, `[[`, numeric(1), "det_term"),
    ok = vapply(raw, `[[`, logical(1), "ok")
  )
  per_unit <- add_loo_group_column(per_unit, unit_group, lavdata)

  # loo-style SE of the total ELPD: sqrt(n * var(pointwise)). n is the number
  # of units scored, known a priori: a failed second-order term voids only
  # that unit's term (NA, dropped from the sum/var), not its place in the
  # sample, so both orders scale by the same n.
  elpd_1 <- sum(per_unit$log_cpo_1)
  se_1 <- sqrt(n_units * var(per_unit$log_cpo_1))
  p_loo_1 <- sum(per_unit$lpd_1 - per_unit$log_cpo_1)
  n_ok <- sum(per_unit$ok)
  if (isTRUE(second_order)) {
    elpd_2 <- sum(per_unit$log_cpo_2, na.rm = TRUE)
    se_2 <- sqrt(n_units * var(per_unit$log_cpo_2, na.rm = TRUE))
    p_loo_2 <- sum(per_unit$lpd_2 - per_unit$log_cpo_2, na.rm = TRUE)
    if (n_ok < ceiling(0.9 * n_units)) {
      cli_warn(c(
        "{n_units - n_ok} of {n_units} units fell back to the first-order
         approximation (non-positive-definite curvature).",
        "i" = "The Gaussian posterior summary may be a poor fit."
      ))
    }
  } else {
    elpd_2 <- se_2 <- p_loo_2 <- NA_real_
  }

  use_second <- isTRUE(second_order) && n_ok > 0L
  elpd_loo <- if (use_second) elpd_2 else elpd_1
  se_elpd <- if (use_second) se_2 else se_1
  p_loo <- if (use_second && is.finite(p_loo_2)) p_loo_2 else p_loo_1
  p_diff <- per_unit$lpd_1 - per_unit$log_cpo_1
  if (use_second && is.finite(p_loo_2)) {
    p_diff <- per_unit$lpd_2 - per_unit$log_cpo_2
  }
  estimates <- cbind(
    Estimate = c(elpd_loo, p_loo, -2 * elpd_loo),
    SE = c(se_elpd, sqrt(n_units * var(p_diff, na.rm = TRUE)), 2 * se_elpd)
  )
  rownames(estimates) <- c("elpd_loo", "p_loo", "looic")

  structure(
    list(
      per_unit = per_unit,
      estimates = estimates,
      elpd_1 = elpd_1,
      elpd_2 = elpd_2,
      se_1 = se_1,
      se_2 = se_2,
      p_loo_1 = p_loo_1,
      p_loo_2 = p_loo_2,
      type = type,
      flavour = flavour,
      n_units = n_units,
      n_groups = lavdata@ngroups,
      n_ok = n_ok,
      second_order = isTRUE(second_order),
      theta_overridden = theta_overridden
    ),
    class = "inlavaan_loo"
  )
}

# ---- Sampling-based WAIC -----------------------------------------------------
#
# Unlike the Taylor LOO above, WAIC is computed from an S x n matrix of unit
# log-likelihoods evaluated over posterior draws (one implied-moment build
# per draw, reusing the casewise kernels):
#   lpd_u    = log mean_s p(y_u | theta_s)
#   p_waic_u = var_s log p(y_u | theta_s)
#   elpd_waic = sum_u (lpd_u - p_waic_u),  waic = -2 * elpd_waic

log_mean_exp <- function(v) {
  v <- v[is.finite(v)]
  if (length(v) == 0L) {
    return(NA_real_) # nocov
  }
  a <- max(v)
  a + log(mean(exp(v - a)))
}

inlav_waic <- function(
  int,
  type = c("auto", "loso", "loco"),
  units = NULL,
  nsamp = NULL,
  eff_cores = 1L,
  verbose = FALSE
) {
  type <- match.arg(type)
  check_loo_model(int, fn = "waic")
  nsamp <- nsamp %||% int$nsamp %||% 1000L
  samp <- sample_params(
    theta_star = int$theta_star,
    Sigma_theta = int$Sigma_theta,
    method = int$marginal_method,
    approx_data = int$approx_data,
    pt = int$partable,
    lavmodel = int$lavmodel,
    nsamp = nsamp,
    R_star = int$R_star
  )
  waic_from_draws(
    int,
    samp$x_samp,
    type = type,
    units = units,
    eff_cores = eff_cores,
    verbose = verbose
  )
}

# WAIC from an existing matrix of posterior draws (lavaan-x space). Used by
# inlav_waic() with fresh draws and by the fit-time path with the draws the
# fit already produced, so that fit-time WAIC costs only the casewise pass.
waic_from_draws <- function(
  int,
  x_samp,
  type = c("auto", "loso", "loco"),
  units = NULL,
  eff_cores = 1L,
  verbose = FALSE
) {
  type <- match.arg(type)
  pt <- int$partable
  lavmodel <- int$lavmodel
  lavdata <- int$lavdata
  two_level <- is_multilevel(lavdata)
  two_level_missing <- two_level && isTRUE(int$lavsamplestats@missing.flag)
  nsamp <- nrow(x_samp)

  # Resolve unit type (matches loo): conditional WAIC = leave-one-unit-out,
  # marginal WAIC = leave-one-cluster-out (Merkle, Furr & Rabe-Hesketh 2019)
  if (type == "auto") {
    type <- if (two_level) "loco" else "loso"
  } else if (type == "loco" && !two_level) {
    cli_abort(
      "Leave-one-cluster-out requires a two-level model, but this model has
       no clusters."
    )
  } else if (type == "loso" && two_level) {
    cli_warn(c(
      "Scoring leave-one-unit-out (the {.emph conditional} WAIC) on a
       two-level model, not the default leave-one-cluster-out (the
       {.emph marginal} WAIC).",
      "i" = "These target different predictions -- a new observation within an
       observed cluster vs a new cluster -- and are easily conflated
       (Merkle, Furr & Rabe-Hesketh, 2019)."
    ))
  }
  per_row_2l <- type == "loso" && two_level

  unit_group <- NULL
  if (per_row_2l) {
    X <- lavdata@X[[1L]]
    if (two_level_missing) {
      minfo <- loco_missing_info(int)
    } else {
      css <- loco_suff_stats(lavdata)
    }
    units <- check_loo_units(units, nrow(X), "rows")
    nobs <- rep(1L, length(units))
  } else if (two_level_missing) {
    minfo <- loco_missing_info(int)
    units <- check_loo_units(units, minfo$J, "clusters")
    nobs <- minfo$n_j[units]
  } else if (two_level) {
    css <- loco_suff_stats(lavdata)
    units <- check_loo_units(units, css$J, "clusters")
    nobs <- css$n_j[units]
  } else {
    dv <- loso_data_view(lavmodel, lavdata, x_idx = int$lavsamplestats@x.idx)
    uv <- loso_resolve_units(lavdata, units)
    units <- uv$ids
    unit_group <- uv$grp
    nobs <- rep(1L, uv$n)
  }

  one_draw <- function(s) {
    lavmodel_x <- lavaan::lav_model_set_parameters(lavmodel, x_samp[s, ])
    mom <- loo_implied_moments(lavmodel_x, two_level)
    tryCatch(
      if (per_row_2l && two_level_missing) {
        loso2l_missing_loglik_all(units, minfo, mom)
      } else if (per_row_2l) {
        loso2l_loglik_all(units, css, X, mom)
      } else if (two_level_missing) {
        vapply(
          units,
          function(j) loco_missing_loglik_one(j, minfo, mom),
          numeric(1)
        )
      } else if (two_level) {
        vapply(units, function(j) loco_loglik_one(j, css, mom), numeric(1))
      } else {
        # includes the exchangeability constant; 0 with a mean structure
        loso_loglik_units(uv, dv, mom)
      },
      error = function(e) rep(NA_real_, length(units)) # nocov
    )
  }
  ll_list <- run_parallel_or_serial(
    m = nsamp,
    FUN = one_draw,
    cores = eff_cores,
    verbose = verbose,
    msg_serial = "Evaluating unit log-likelihoods at draw {j}/{m}.",
    msg_parallel = "Evaluating unit log-likelihoods at {m} draws ({cores} cores).",
    msg_done = "Evaluate unit log-likelihoods at {m} draw{?s}."
  )
  ll_mat <- do.call(rbind, ll_list) # nsamp x n_units

  # Score a fixed.x fit on its conditional likelihood: subtract the
  # frozen-covariate marginal, a per-unit constant across draws
  flavour <- loo_flavour(int)
  if (flavour == "conditional") {
    mom0 <- loo_grad_cache(
      int$theta_star,
      lavmodel,
      pt,
      two_level = two_level
    )$mom
    cvec <- if (per_row_2l) {
      loo_fixedx_rowdiff_loco(int, css, X, units, mom0)
    } else if (two_level) {
      loo_fixedx_const_loco(int, css, units, mom0)
    } else {
      loso_fixedx_const_units(int, uv, dv, mom0)
    }
    ll_mat <- sweep(ll_mat, 2L, cvec, "-")
  }

  lpd <- apply(ll_mat, 2L, log_mean_exp)
  p_waic <- apply(ll_mat, 2L, var, na.rm = TRUE)
  elpd_waic_u <- lpd - p_waic
  n_units <- length(units)

  n_high <- sum(p_waic > 0.4, na.rm = TRUE)
  if (n_high > 0L) {
    cli_warn(
      "{n_high} unit{?s} {?has/have} p_waic > 0.4, so the WAIC may be
       unreliable. Consider {.fn loo} instead."
    )
  }

  estimates <- cbind(
    Estimate = c(sum(elpd_waic_u), sum(p_waic), -2 * sum(elpd_waic_u)),
    SE = c(
      sqrt(n_units * var(elpd_waic_u)),
      sqrt(n_units * var(p_waic)),
      2 * sqrt(n_units * var(elpd_waic_u))
    )
  )
  rownames(estimates) <- c("elpd_waic", "p_waic", "waic")

  per_unit <- data.frame(
    unit = units,
    nobs = nobs,
    lpd = lpd,
    p_waic = p_waic,
    elpd_waic = elpd_waic_u
  )
  per_unit <- add_loo_group_column(per_unit, unit_group, lavdata)

  structure(
    list(
      per_unit = per_unit,
      estimates = estimates,
      type = type,
      flavour = flavour,
      n_units = n_units,
      n_groups = lavdata@ngroups,
      nsamp = nsamp
    ),
    class = "inlavaan_waic"
  )
}

check_loo_units <- function(units, n_avail, what) {
  if (is.null(units)) {
    return(seq_len(n_avail))
  }
  units <- as.integer(units)
  if (
    length(units) == 0L ||
      anyNA(units) ||
      any(units < 1L | units > n_avail) ||
      anyDuplicated(units) > 0L
  ) {
    cli_abort(
      "{.arg units} must be distinct integers between 1 and {n_avail}
       (the number of {what})."
    )
  }
  units
}

HS_model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"

# Small deterministic subset of HolzingerSwineford1939 shared by every acfa()
# call in this file, in place of the full 301-row dataset
set.seed(1)
dat <- lavaan::HolzingerSwineford1939[
  sample(nrow(lavaan::HolzingerSwineford1939), 40),
]

fit <- acfa(
  HS_model,
  dat,
  meanstructure = TRUE,
  verbose = FALSE,
  nsamp = 3,
  test = "none",
  vb_correction = FALSE,
  marginal_method = "marggaus",
  marginal_correction = "none"
)
res <- loo(fit)
# Shared across two test_that() blocks that both need a copy of `fit` with
# LOO stored via add_loo() -- historically each block called add_loo(fit)
# separately, but the call is deterministic so it is computed once here
fit_with_loo <- add_loo(fit)

test_that("LOSO matches reference values", {
  # Reference values computed with an independent implementation of the same
  # Taylor LOO formulas on this exact fit
  expect_equal(res$type, "loso")
  expect_equal(res$n_units, 40L)
  expect_equal(res$elpd_1, -488.6660523464, tolerance = 1e-4)
  expect_equal(res$elpd_2, -509.4810089914, tolerance = 1e-4)
  expect_equal(res$se_1, 14.9606326798, tolerance = 1e-4)
  expect_equal(res$se_2, 16.1424977191, tolerance = 1e-4)
  expect_equal(res$p_loo_1, 28.1110170037, tolerance = 1e-4)
  expect_equal(res$p_loo_2, 33.8073861249, tolerance = 1e-4)

  pu <- res$per_unit[c(1L, 20L, 40L), ]
  expect_equal(
    pu$l_star,
    c(-10.1330813639, -10.7290547282, -12.8417007213),
    tolerance = 1e-4
  )
  expect_equal(
    pu$log_cpo_1,
    c(-10.2877928268, -10.8713710491, -13.1844720013),
    tolerance = 1e-4
  )
  expect_equal(
    pu$log_cpo_2,
    c(-10.6530882509, -11.2745912656, -13.5171746493),
    tolerance = 1e-4
  )
  expect_equal(
    pu$det_term,
    c(-0.3503871084, -0.3930981418, -0.2827972374),
    tolerance = 1e-3
  )
})

test_that("loo object structure and internal identities", {
  expect_s3_class(res, "inlavaan_loo")
  expect_named(
    res$per_unit,
    c(
      "unit",
      "nobs",
      "l_star",
      "score_norm",
      "lpd_1",
      "lpd_2",
      "log_cpo_1",
      "log_cpo_2",
      "det_term",
      "ok"
    )
  )
  expect_true(all(res$per_unit$ok))
  expect_true(all(res$per_unit$nobs == 1L))

  # First-order CPO and LPD are symmetric about l_star by construction
  expect_equal(
    res$per_unit$lpd_1 + res$per_unit$log_cpo_1,
    2 * res$per_unit$l_star
  )
  expect_equal(
    unname(res$estimates["elpd_loo", "Estimate"]),
    res$elpd_2
  )
  expect_equal(
    unname(res$estimates["looic", "Estimate"]),
    -2 * res$elpd_2
  )
  expect_equal(
    unname(res$estimates["looic", "SE"]),
    2 * res$se_2
  )

  expect_output(print(res), "leave-one-subject-out")
  expect_output(print(res), "elpd_loo")
})

test_that("sum of unit logliks equals the model loglik at the mode", {
  int <- get_inlavaan_internal(fit)
  x <- INLAvaan:::pars_to_x(int$theta_star, int$partable)
  lm_x <- lavaan::lav_model_set_parameters(int$lavmodel, x)
  opts <- fit@Options
  opts$estimator <- "ML"
  ll <- INLAvaan:::lavaan___lav_model_loglik(
    lavdata = int$lavdata,
    lavsamplestats = int$lavsamplestats,
    lavimplied = lavaan::lav_model_implied(lm_x),
    lavmodel = lm_x,
    lavoptions = opts
  )$loglik
  expect_equal(sum(res$per_unit$l_star), ll, tolerance = 1e-6)
})

test_that("first-order only and unit subsetting", {
  res1 <- loo(fit, second_order = FALSE)
  expect_true(all(is.na(res1$per_unit$log_cpo_2)))
  expect_equal(res1$elpd_1, res$elpd_1)
  expect_equal(
    unname(res1$estimates["elpd_loo", "Estimate"]),
    res1$elpd_1
  )

  res25 <- loo(fit, units = 1:25)
  expect_equal(nrow(res25$per_unit), 25L)
  expect_equal(
    res25$per_unit$log_cpo_2,
    res$per_unit$log_cpo_2[1:25],
    tolerance = 1e-8
  )

  expect_error(loo(fit, units = c(1L, 1L)), "distinct")
  expect_error(loo(fit, units = 0L), "distinct")
})

test_that("theta/Sigma override scores arbitrary summaries", {
  int <- get_inlavaan_internal(fit)
  res_same <- loo(fit, theta = int$theta_star, Sigma = int$Sigma_theta)
  expect_true(res_same$theta_overridden)
  expect_equal(res_same$elpd_2, res$elpd_2)

  res_pert <- loo(fit, theta = int$theta_star * 1.01)
  expect_false(isTRUE(all.equal(res_pert$elpd_2, res$elpd_2)))

  # Conditioning a parameter to zero gives a singular Sigma; the active
  # block restriction handles it
  p <- 1L
  theta_c <- int$theta_star -
    int$Sigma_theta[, p] * (int$theta_star[p] / int$Sigma_theta[p, p])
  Sigma_c <- int$Sigma_theta -
    tcrossprod(int$Sigma_theta[, p]) / int$Sigma_theta[p, p]
  res_cond <- loo(fit, theta = theta_c, Sigma = Sigma_c, units = 1:10)
  expect_true(all(is.finite(res_cond$per_unit$log_cpo_1)))

  expect_error(loo(fit, theta = 1:3), "length")
  expect_error(loo(fit, Sigma = diag(3)), "covariance")
})

test_that("type override and parallel agree with serial", {
  expect_error(loo(fit, type = "loco"), "two-level")

  res_ser <- loo(fit, units = 1:10)
  res_par <- loo(fit, units = 1:10, cores = 2L)
  # Parallel (fork-based mclapply) and serial paths are algorithmically
  # identical but not bit-reproducible: floating-point reduction order in the
  # per-unit inner numerics can differ, so agreement is checked to numerical
  # tolerance rather than exact equality (any real divergence is O(1)).
  expect_equal(
    res_par$per_unit$log_cpo_2,
    res_ser$per_unit$log_cpo_2,
    tolerance = 1e-3
  )
})

test_that("equality constraints (ceq.simple) are handled", {
  hs_eq <- "
    visual  =~ x1 + a*x2 + a*x3
    textual =~ x4 + x5 + x6
    speed   =~ x7 + x8 + x9
  "
  fit_eq <- acfa(
    hs_eq,
    dat,
    meanstructure = TRUE,
    verbose = FALSE,
    nsamp = 3,
    test = "none",
    vb_correction = FALSE,
    marginal_method = "marggaus",
    marginal_correction = "none"
  )
  int <- get_inlavaan_internal(fit_eq)
  expect_true(int$lavmodel@ceq.simple.only)

  res_eq <- loo(fit_eq)
  expect_true(all(res_eq$per_unit$ok))

  # Total loglik consistency in the packed parameter space
  theta_unp <- as.numeric(int$lavmodel@ceq.simple.K %*% int$theta_star)
  x <- INLAvaan:::pars_to_x(theta_unp, int$partable)
  lm_x <- lavaan::lav_model_set_parameters(int$lavmodel, x)
  opts <- fit_eq@Options
  opts$estimator <- "ML"
  ll <- INLAvaan:::lavaan___lav_model_loglik(
    lavdata = int$lavdata,
    lavsamplestats = int$lavsamplestats,
    lavimplied = lavaan::lav_model_implied(lm_x),
    lavmodel = lm_x,
    lavoptions = opts
  )$loglik
  expect_equal(sum(res_eq$per_unit$l_star), ll, tolerance = 1e-6)

  # Analytic unit score matches a numerical derivative in packed theta space
  Y1 <- int$lavdata@X[[1L]][1L, , drop = FALSE]
  s1 <- as.numeric(INLAvaan:::loso_scores_theta(
    int$theta_star,
    Y1,
    int$lavmodel,
    int$partable
  ))
  h <- 1e-6
  g_num <- vapply(
    seq_along(int$theta_star),
    function(k) {
      tp <- tm <- int$theta_star
      tp[k] <- tp[k] + h
      tm[k] <- tm[k] - h
      cp <- INLAvaan:::loo_grad_cache(tp, int$lavmodel, int$partable)
      cm <- INLAvaan:::loo_grad_cache(tm, int$lavmodel, int$partable)
      (INLAvaan:::loso_loglik_all(Y1, cp$mom[[1L]]) -
        INLAvaan:::loso_loglik_all(Y1, cm$mom[[1L]])) /
        (2 * h)
    },
    numeric(1)
  )
  expect_equal(s1, g_num, tolerance = 1e-5)
})

test_that("fit-time LOO via test = 'loo' and add_loo()", {
  fit_loo <- acfa(
    HS_model,
    dat,
    meanstructure = TRUE,
    verbose = FALSE,
    nsamp = 3,
    test = "loo",
    vb_correction = FALSE,
    marginal_method = "marggaus",
    marginal_correction = "none"
  )
  stored <- get_inlavaan_internal(fit_loo, "loo")
  expect_s3_class(stored, "inlavaan_loo")
  expect_equal(stored$elpd_2, res$elpd_2, tolerance = 1e-10)

  # loo() returns the stored result for default arguments only
  expect_identical(loo(fit_loo), stored)
  res_sub <- loo(fit_loo, units = 1:5)
  expect_equal(nrow(res_sub$per_unit), 5L)

  # add_loo() returns an updated copy; the original fit is unchanged
  fit2 <- fit_with_loo
  expect_null(fit@external$inlavaan_internal$loo)
  expect_s3_class(get_inlavaan_internal(fit2, "loo"), "inlavaan_loo")
  expect_equal(get_inlavaan_internal(fit2, "loo")$elpd_2, res$elpd_2)
  expect_identical(loo(fit2), get_inlavaan_internal(fit2, "loo"))
})

test_that("fitMeasures reports LOO measures on request or when stored", {
  # Not stored: bare fitMeasures() excludes the LOO measures
  fm_all <- fitMeasures(fit)
  expect_false(any(
    c("elpd_loo", "se_loo", "p_loo", "looic") %in%
      names(fm_all)
  ))

  # On request by name: computed on demand, agreeing with loo()
  fm <- fitMeasures(fit, c("elpd_loo", "se_loo", "p_loo", "looic"))
  expect_equal(unname(fm["elpd_loo"]), res$elpd_2, tolerance = 1e-10)
  expect_equal(unname(fm["looic"]), -2 * res$elpd_2, tolerance = 1e-10)
  expect_equal(unname(fm["se_loo"]), 2 * res$se_2, tolerance = 1e-10)
  expect_equal(unname(fm["p_loo"]), res$p_loo_2, tolerance = 1e-10)

  # Stored: included in "all" for free
  fit2 <- fit_with_loo
  fm2 <- fitMeasures(fit2)
  expect_true(all(c("elpd_loo", "se_loo", "p_loo", "looic") %in% names(fm2)))
  expect_equal(unname(fm2["elpd_loo"]), res$elpd_2, tolerance = 1e-10)
})

test_that("waic() sanity and agreement with loo()", {
  set.seed(123)
  # A few units genuinely exceed the p_waic reliability threshold here
  expect_warning(w <- waic(fit, nsamp = 100), "p_waic")
  expect_s3_class(w, "inlavaan_waic")
  expect_equal(w$n_units, 40L)
  expect_equal(w$type, "loso")
  expect_true(all(is.finite(w$per_unit$lpd)))
  expect_true(all(w$per_unit$p_waic > 0))
  expect_output(print(w), "WAIC")

  # WAIC and LOO estimate the same quantity; loose agreement on this model
  expect_equal(
    unname(w$estimates["elpd_waic", "Estimate"]),
    res$elpd_2,
    tolerance = 0.005
  )
  expect_equal(
    unname(w$estimates["waic", "Estimate"]),
    -2 * unname(w$estimates["elpd_waic", "Estimate"])
  )

  # fitMeasures computes WAIC on request by name only (the test fit has
  # nsamp = 3, so suppress the small-sample p_waic reliability warning)
  expect_false("waic" %in% names(fitMeasures(fit)))
  suppressWarnings(
    fm <- fitMeasures(fit, c("waic", "p_waic", "se_waic"))
  )
  expect_true(all(c("waic", "p_waic", "se_waic") %in% names(fm)))
})

test_that("single-level FIML is supported (see test-loo-missing.R)", {
  d_miss <- dat
  d_miss[1, "x1"] <- NA
  fit_miss <- acfa(
    HS_model,
    d_miss,
    meanstructure = TRUE,
    missing = "ml",
    verbose = FALSE,
    nsamp = 3,
    test = "none",
    vb_correction = FALSE,
    marginal_method = "marggaus",
    marginal_correction = "none"
  )
  res_miss <- loo(fit_miss)
  expect_s3_class(res_miss, "inlavaan_loo")
  expect_equal(res_miss$flavour, "joint")
})

test_that("test = 'standard' stores LOO and WAIC when supported and cheap", {
  fit_std <- acfa(
    HS_model,
    dat,
    meanstructure = TRUE,
    verbose = FALSE,
    nsamp = 100,
    test = "standard",
    vb_correction = FALSE,
    marginal_method = "marggaus",
    marginal_correction = "none"
  )
  int <- get_inlavaan_internal(fit_std)

  expect_s3_class(int$loo, "inlavaan_loo")
  expect_equal(int$loo$n_units, 40L)
  expect_equal(int$loo$elpd_2, res$elpd_2, tolerance = 1e-10)
  expect_identical(loo(fit_std), int$loo)

  expect_s3_class(int$waic, "inlavaan_waic")
  expect_equal(int$waic$nsamp, 100L)
  expect_identical(waic(fit_std), int$waic)
  # non-default arguments still trigger a fresh computation
  set.seed(1)
  w2 <- suppressWarnings(waic(fit_std, nsamp = 120))
  expect_equal(w2$nsamp, 120L)

  # stored results appear in fitMeasures' "all" for free
  fm <- fitMeasures(fit_std)
  expect_true(all(
    c("elpd_loo", "looic", "waic", "p_waic", "se_waic") %in% names(fm)
  ))
  expect_true(all(c("ppp", "dic", "p_dic") %in% names(fm)))
})

test_that("the fit-time budget gate aborts with its own condition class", {
  int <- get_inlavaan_internal(fit)
  expect_error(
    INLAvaan:::inlav_loo(int, max_seconds = 1e-9),
    class = "inlavaan_loo_budget"
  )
})

test_that("small nsamp skips fit-time WAIC but not LOO", {
  fit_s3 <- acfa(
    HS_model,
    dat,
    meanstructure = TRUE,
    verbose = FALSE,
    nsamp = 3,
    test = "standard",
    vb_correction = FALSE,
    marginal_method = "marggaus",
    marginal_correction = "none"
  )
  int <- get_inlavaan_internal(fit_s3)
  expect_null(int$waic)
  expect_s3_class(int$loo, "inlavaan_loo")
})

# Two-level FIML LOCO (leave-one-cluster-out under missing data). Each cluster
# is scored on its observed-data marginal likelihood via lavaan's raw-data
# missing kernels; since LOCO deletes a whole cluster there is no downdating.
# The flavour is joint (lavaan rejects two-level + conditional.x).

twolevel_model <- "
  level: 1
    fw =~ y1 + y2 + y3
  level: 2
    fb =~ y1 + y2 + y3
"

# Keep the first 30 cluster ids (cluster sizes cycle 5, 10, 15, 20 every 4
# cluster ids in lavaan::Demo.twolevel, so every size is represented several
# times). MCAR holes in y1-y3; seed set immediately before the hole-punching
# loop so the dataset (and the pinned reference values) are reproducible. With
# ncl = 30, cluster 9 ends up with a row fully missing on y1-y3 -- needed by
# the two dedicated regression tests below.
make_miss <- function() {
  d <- lavaan::Demo.twolevel[, c("y1", "y2", "y3", "cluster")]
  d <- d[d$cluster <= 30, ]
  set.seed(20260613)
  for (v in c("y1", "y2", "y3")) {
    d[[v]][runif(nrow(d)) < 0.12] <- NA
  }
  d
}
d <- make_miss()

# a row is fully missing after punching holes; lavaan flags this with a
# (benign) note about the two-level FIML gradient, suppressed here so the
# fixture sets up cleanly (loo()/waic() handle these rows correctly)
fit <- suppressWarnings(asem(
  twolevel_model,
  d,
  cluster = "cluster",
  missing = "ml",
  verbose = FALSE,
  nsamp = 3,
  test = "none",
  vb_correction = FALSE,
  marginal_method = "marggaus",
  marginal_correction = "none"
))
res <- loo(fit)

test_that("the test dataset has the expected missingness", {
  expect_equal(sum(is.na(d[, 1:3])), 129L)
  expect_equal(sum(!complete.cases(d[, 1:3])), 118L)
})

test_that("two-level FIML LOCO matches reference values", {
  # Reference values cross-checked against (i) lavaan's fitted two-level FIML
  # loglik, (ii) an independent dense marginal-covariance kernel, and (iii)
  # finite differences (lab graduated/2026-06-loo-missing/05-*.R). This
  # dataset has fully-missing-within cases, so the fitted mode (and the LOO
  # values pinned to it) shifts slightly between lavaan versions: lavaan <
  # 0.7-1.2707 optimises with the inexact gradient (lavaan PR #581), patched
  # lavaan with the correct one. The values below are pinned on the buggy
  # gradient; skip them on patched lavaan (the version-invariant identities
  # and finite-difference checks below still run there).
  skip_if(
    utils::packageVersion("lavaan") >= "0.7.1.2707",
    "fit mode differs once the lavaan gradient is patched"
  )
  expect_equal(res$type, "loco")
  expect_equal(res$flavour, "joint")
  expect_equal(res$n_units, 30L)
  expect_equal(res$elpd_1, -1622.0173425091, tolerance = 1e-4)
  expect_equal(res$elpd_2, -1631.5418961566, tolerance = 1e-4)
  expect_equal(res$se_1, 146.6123016390, tolerance = 1e-4)
  expect_equal(res$se_2, 147.4268836230, tolerance = 1e-4)
  expect_equal(res$p_loo_1, 14.2027032692, tolerance = 1e-2)
  expect_equal(res$p_loo_2, 16.3011978533, tolerance = 1e-2)

  # first, middle, and last of the 30 clusters
  pu <- res$per_unit[c(1L, 15L, 30L), ]
  expect_equal(pu$nobs, c(5L, 15L, 10L))
  expect_equal(
    pu$l_star,
    c(-16.9345018087, -69.6199644963, -42.7357136336),
    tolerance = 1e-4
  )
  expect_equal(
    pu$log_cpo_1,
    c(-16.9717923529, -69.7909182029, -42.8737272647),
    tolerance = 1e-4
  )
  expect_equal(
    pu$log_cpo_2,
    c(-17.0345397512, -70.1542742739, -43.0387010435),
    tolerance = 1e-4
  )
})

test_that("per-cluster observed-data logliks sum to the fitted FIML loglik", {
  int <- get_inlavaan_internal(fit)
  x <- INLAvaan:::pars_to_x(int$theta_star, int$partable)
  lm_x <- lavaan::lav_model_set_parameters(int$lavmodel, x)
  opts <- fit@Options
  opts$estimator <- "ML"
  ll <- INLAvaan:::lavaan___lav_model_loglik(
    lavdata = int$lavdata,
    lavsamplestats = int$lavsamplestats,
    lavimplied = lavaan::lav_model_implied(lm_x),
    lavmodel = lm_x,
    lavoptions = opts
  )$loglik
  expect_equal(sum(res$per_unit$l_star), ll, tolerance = 1e-6)
})

test_that("analytic per-cluster scores match finite differences", {
  int <- get_inlavaan_internal(fit)
  minfo <- INLAvaan:::loco_missing_info(int)
  js <- c(1L, 15L, 30L) # first, middle, and last of the 30 clusters
  s_an <- INLAvaan:::loco_missing_scores_theta(
    int$theta_star,
    minfo,
    int$lavmodel,
    int$partable,
    js
  )
  h <- 1e-6
  lj <- function(theta, j) {
    cache <- INLAvaan:::loo_grad_cache(
      theta,
      int$lavmodel,
      int$partable,
      two_level = TRUE
    )
    INLAvaan:::loco_missing_loglik_one(j, minfo, cache$mom)
  }
  # vapply stacks the per-parameter columns into a length(js) x m matrix,
  # matching the score matrix returned by loco_missing_scores_theta()
  s_fd <- vapply(
    seq_along(int$theta_star),
    function(k) {
      tp <- tm <- int$theta_star
      tp[k] <- tp[k] + h
      tm[k] <- tm[k] - h
      vapply(js, function(j) (lj(tp, j) - lj(tm, j)) / (2 * h), numeric(1))
    },
    numeric(length(js))
  )
  expect_equal(max(abs(s_an - s_fd)), 0, tolerance = 1e-5)
})

test_that("loo object structure and unit subsetting", {
  expect_s3_class(res, "inlavaan_loo")
  expect_true(all(res$per_unit$ok))
  expect_equal(
    res$per_unit$lpd_1 + res$per_unit$log_cpo_1,
    2 * res$per_unit$l_star
  )
  res5 <- loo(fit, units = 1:5)
  expect_equal(nrow(res5$per_unit), 5L)
  expect_equal(
    res5$per_unit$log_cpo_2,
    res$per_unit$log_cpo_2[1:5],
    tolerance = 1e-8
  )
})

test_that("waic() runs on a two-level FIML fit and agrees loosely with loo()", {
  set.seed(1)
  w <- suppressWarnings(waic(fit, nsamp = 50))
  expect_s3_class(w, "inlavaan_waic")
  expect_equal(w$n_units, 30L)
  expect_equal(w$type, "loco")
  expect_equal(w$flavour, "joint")
  expect_true(all(is.finite(w$per_unit$lpd)))
  expect_equal(
    unname(w$estimates["elpd_waic", "Estimate"]),
    res$elpd_2,
    tolerance = 0.01
  )
})

test_that("the per-row (leave-one-unit-out) override works under missing data", {
  # type = "loso" on a clustered fit warns (conditional vs marginal) then
  # scores the leave-one-unit-out conditional predictive per row
  expect_warning(
    res_row <- loo(fit, type = "loso", units = 1:10),
    "leave-one-unit-out"
  )
  expect_equal(res_row$type, "loso")
  expect_equal(nrow(res_row$per_unit), 10L)
  expect_true(all(res_row$per_unit$nobs == 1L))

  # analytic per-row scores agree with finite differences, including rows in
  # clusters that contain a fully-missing row (lavaan's gradient kernel
  # mishandles zero-observed patterns; INLAvaan drops them before the kernel)
  int <- get_inlavaan_internal(fit)
  minfo <- INLAvaan:::loco_missing_info(int)
  rows <- c(1L, 5L, 200L, minfo$rows_by_cluster[[9L]])
  s_an <- INLAvaan:::loso2l_missing_scores_theta(
    int$theta_star,
    minfo,
    int$lavmodel,
    int$partable,
    rows
  )
  h <- 1e-6
  s_fd <- vapply(
    seq_along(int$theta_star),
    function(k) {
      tp <- tm <- int$theta_star
      tp[k] <- tp[k] + h
      tm[k] <- tm[k] - h
      cp <- INLAvaan:::loo_grad_cache(
        tp,
        int$lavmodel,
        int$partable,
        two_level = TRUE
      )
      cm <- INLAvaan:::loo_grad_cache(
        tm,
        int$lavmodel,
        int$partable,
        two_level = TRUE
      )
      (INLAvaan:::loso2l_missing_loglik_all(rows, minfo, cp$mom) -
        INLAvaan:::loso2l_missing_loglik_all(rows, minfo, cm$mom)) /
        (2 * h)
    },
    numeric(length(rows))
  )
  expect_equal(max(abs(s_an - s_fd)), 0, tolerance = 1e-5)
})

test_that("LOCO scores are correct for clusters with a fully-missing row", {
  # regression guard for the fix: cluster 9 has a row with all within
  # variables missing; its score must match finite differences
  int <- get_inlavaan_internal(fit)
  minfo <- INLAvaan:::loco_missing_info(int)
  expect_true(any(minfo$n_obs < minfo$n_j)) # some cluster has a fully-missing row
  s_an <- INLAvaan:::loco_missing_scores_theta(
    int$theta_star,
    minfo,
    int$lavmodel,
    int$partable,
    9L
  )
  h <- 1e-6
  s_fd <- vapply(
    seq_along(int$theta_star),
    function(k) {
      tp <- tm <- int$theta_star
      tp[k] <- tp[k] + h
      tm[k] <- tm[k] - h
      cp <- INLAvaan:::loo_grad_cache(
        tp,
        int$lavmodel,
        int$partable,
        two_level = TRUE
      )
      cm <- INLAvaan:::loo_grad_cache(
        tm,
        int$lavmodel,
        int$partable,
        two_level = TRUE
      )
      (INLAvaan:::loco_missing_loglik_one(9L, minfo, cp$mom) -
        INLAvaan:::loco_missing_loglik_one(9L, minfo, cm$mom)) /
        (2 * h)
    },
    numeric(1)
  )
  expect_equal(max(abs(as.numeric(s_an) - s_fd)), 0, tolerance = 1e-5)
})

# Conditional-flavour LOO for fixed.x fits: units are scored by the
# predictive density of their outcomes given their covariates, matching the
# (conditional) likelihood the model was fitted with.

HS_model_x <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
  visual ~ ageyr + grade
  textual ~ ageyr + grade
"
# The full HolzingerSwineford1939 has exactly one NA (in "grade"); na.omit it
# first, then subsample down to N = 90 (still comfortably satisfies the
# closed-form Student-t downdated-moments check below, which needs
# n - 1 - q > 0 with q = 2 covariates, and keeps every LOO unit's curvature
# well-behaved -- smaller subsamples occasionally push a left-out unit's
# downdated Hessian to non-positive-definite).
dat_x <- na.omit(
  lavaan::HolzingerSwineford1939[, c(paste0("x", 1:9), "ageyr", "grade")]
)
set.seed(123)
dat_x <- dat_x[sort(sample(nrow(dat_x), 90)), ]

fit_c <- asem(
  HS_model_x,
  dat_x,
  fixed.x = TRUE,
  meanstructure = TRUE,
  verbose = FALSE,
  nsamp = 3,
  test = "none",
  vb_correction = FALSE,
  marginal_method = "marggaus",
  marginal_correction = "none"
)
res_c <- loo(fit_c)

# Shared with the "compare(loo = TRUE) enforces the flavour rules" test below
# (identical model/data/arguments, just fixed.x = FALSE)
fit_j <- asem(
  HS_model_x,
  dat_x,
  fixed.x = FALSE,
  meanstructure = TRUE,
  verbose = FALSE,
  nsamp = 3,
  test = "none",
  vb_correction = FALSE,
  marginal_method = "marggaus",
  marginal_correction = "none"
)

# Total loglik of the fit at the INLAvaan mode (under fixed.x this is the
# conditional loglik)
ll_at_mode <- function(fit) {
  int <- get_inlavaan_internal(fit)
  K <- if (isTRUE(int$lavmodel@ceq.simple.only)) {
    int$lavmodel@ceq.simple.K
  } else {
    NULL
  }
  theta_unp <- if (is.null(K)) {
    int$theta_star
  } else {
    as.numeric(K %*% int$theta_star)
  }
  lm_x <- lavaan::lav_model_set_parameters(
    int$lavmodel,
    INLAvaan:::pars_to_x(theta_unp, int$partable)
  )
  opts <- fit@Options
  opts$estimator <- "ML"
  INLAvaan:::lavaan___lav_model_loglik(
    lavdata = int$lavdata,
    lavsamplestats = int$lavsamplestats,
    lavimplied = lavaan::lav_model_implied(lm_x),
    lavmodel = lm_x,
    lavoptions = opts
  )$loglik
}

test_that("conditional LOSO on a fixed.x fit matches reference values", {
  expect_equal(res_c$flavour, "conditional")
  expect_equal(res_c$type, "loso")
  expect_equal(res_c$n_units, 90L)
  expect_equal(res_c$elpd_1, -1134.0330682985, tolerance = 1e-4)
  expect_equal(res_c$elpd_2, -1165.6523740521, tolerance = 1e-4)
  expect_equal(res_c$se_1, 21.0799322283, tolerance = 1e-3)
  expect_equal(res_c$se_2, 21.5461686423, tolerance = 1e-3)
  expect_equal(res_c$p_loo_2, 58.1103250541, tolerance = 1e-2)

  pu <- res_c$per_unit[c(1L, 45L, 90L), ]
  expect_equal(
    pu$l_star,
    c(-10.6710473345, -13.4890135349, -13.9737268755),
    tolerance = 1e-4
  )
  expect_equal(
    pu$log_cpo_2,
    c(-10.9903065940, -13.8526005859, -14.8047188622),
    tolerance = 1e-4
  )

  expect_output(print(res_c), "conditionally on the exogenous covariates")
})

test_that("conditional unit logliks sum to the fitted likelihood", {
  expect_equal(sum(res_c$per_unit$l_star), ll_at_mode(fit_c), tolerance = 1e-6)
})

test_that("joint and conditional scales differ by the covariate predictive", {
  res_j <- loo(fit_j)
  expect_equal(res_j$flavour, "joint")

  # Closed form: the leave-one-out predictive of the saturated covariate
  # block is multivariate Student-t with downdated moments
  X <- as.matrix(dat_x[, c("ageyr", "grade")])
  n <- nrow(X)
  q <- ncol(X)
  lmvt <- function(x, mu, Lam, nu) {
    k <- length(x)
    ch <- chol(Lam)
    d2 <- sum(backsolve(ch, x - mu, transpose = TRUE)^2)
    lgamma((nu + k) / 2) -
      lgamma(nu / 2) -
      (k / 2) * log(nu * pi) -
      sum(log(diag(ch))) -
      ((nu + k) / 2) * log1p(d2 / nu)
  }
  xbar <- colMeans(X)
  A <- crossprod(sweep(X, 2L, xbar))
  lt <- vapply(
    seq_len(n),
    function(i) {
      m <- n - 1
      xb <- (n * xbar - X[i, ]) / m
      Am <- A - (n / m) * tcrossprod(X[i, ] - xbar)
      nu <- m - q
      lmvt(X[i, ], xb, Am * (m + 1) / (m * nu), nu)
    },
    numeric(1)
  )

  expect_equal(res_j$elpd_2 - res_c$elpd_2, sum(lt), tolerance = 2e-3)
})

twolevel_model_z <- "
  level: 1
    fw =~ y1 + y2 + y3
  level: 2
    fb =~ y1 + y2 + y3
    fb ~ w1 + w2
"
# Demo.twolevel's cluster sizes cycle 5, 10, 15, 20 every 4 cluster ids;
# keeping the first 24 cluster ids gives 300 rows across a representative mix
# of cluster sizes while running in a fraction of the time of the full 200
# clusters / 2500 rows.
d24 <- lavaan::Demo.twolevel[lavaan::Demo.twolevel$cluster %in% 1:24, ]

fit_2c <- asem(
  twolevel_model_z,
  d24,
  cluster = "cluster",
  fixed.x = TRUE,
  meanstructure = TRUE,
  verbose = FALSE,
  nsamp = 3,
  test = "none",
  vb_correction = FALSE,
  marginal_method = "marggaus",
  marginal_correction = "none"
)

test_that("conditional LOCO with cluster-level covariates matches reference values", {
  res_2c <- loo(fit_2c)
  expect_equal(res_2c$flavour, "conditional")
  expect_equal(res_2c$type, "loco")
  expect_equal(res_2c$n_units, 24L)
  expect_equal(res_2c$elpd_1, -1503.5857553604, tolerance = 1e-4)
  expect_equal(res_2c$elpd_2, -1515.1500934774, tolerance = 1e-4)
  expect_equal(res_2c$se_1, 142.9954162751, tolerance = 1e-3)
  expect_equal(res_2c$se_2, 144.0573657962, tolerance = 1e-3)

  pu <- res_2c$per_unit[c(1L, 12L, 24L), ]
  expect_equal(
    pu$l_star,
    c(-23.3368311865, -100.8922738534, -101.0054761098),
    tolerance = 1e-4
  )
  expect_equal(
    pu$log_cpo_2,
    c(-23.4605212734, -102.1991697020, -101.5943866124),
    tolerance = 1e-4
  )

  # The subtracted covariate constant is pinned by the loglik identity
  expect_equal(
    sum(res_2c$per_unit$l_star),
    ll_at_mode(fit_2c),
    tolerance = 1e-6
  )

  # Row-deletion override runs on the conditional flavour (the covariate
  # constant cancels between the full and downdated cluster terms)
  expect_warning(
    res_row <- loo(fit_2c, type = "loso", units = 1:4),
    "leave-one-unit-out"
  )
  expect_equal(res_row$flavour, "conditional")
  expect_true(all(is.finite(res_row$per_unit$log_cpo_1)))
})

test_that("conditional LOCO with within-level covariates matches reference values", {
  twolevel_model_wx <- "
    level: 1
      fw =~ y1 + y2 + y3
      fw ~ x1 + x2 + x3
    level: 2
      fb =~ y1 + y2 + y3
      fb ~ w1 + w2
  "
  fit_wx <- asem(
    twolevel_model_wx,
    d24,
    cluster = "cluster",
    fixed.x = TRUE,
    meanstructure = TRUE,
    verbose = FALSE,
    nsamp = 3,
    test = "none",
    vb_correction = FALSE,
    marginal_method = "marggaus",
    marginal_correction = "none"
  )
  res_wx <- loo(fit_wx)
  expect_equal(res_wx$flavour, "conditional")
  expect_equal(res_wx$n_units, 24L)
  expect_equal(res_wx$elpd_1, -1455.6390218782, tolerance = 1e-4)
  expect_equal(res_wx$elpd_2, -1468.9588402195, tolerance = 1e-4)
  expect_equal(res_wx$se_1, 138.9590837453, tolerance = 1e-3)
  expect_equal(res_wx$se_2, 140.2301868846, tolerance = 1e-3)

  pu <- res_wx$per_unit[c(1L, 12L, 24L), ]
  expect_equal(
    pu$l_star,
    c(-22.5346219951, -101.6347742457, -100.0472924937),
    tolerance = 1e-4
  )
  expect_equal(
    pu$log_cpo_2,
    c(-22.7161458081, -103.2586591522, -100.9165757220),
    tolerance = 1e-4
  )

  # The general covariate constant is pinned by the loglik identity
  expect_equal(
    sum(res_wx$per_unit$l_star),
    ll_at_mode(fit_wx),
    tolerance = 1e-6
  )

  # Row-deletion override: the conditional row contribution equals the
  # conditional cluster loglik difference reconstructed from raw data
  expect_warning(
    res_row <- loo(fit_wx, type = "loso", units = c(2L, 50L)),
    "leave-one-unit-out"
  )
  expect_equal(res_row$flavour, "conditional")
  int <- get_inlavaan_internal(fit_wx)
  css <- INLAvaan:::loco_suff_stats(int$lavdata)
  X <- int$lavdata@X[[1L]]
  cache <- INLAvaan:::loo_grad_cache(
    int$theta_star,
    int$lavmodel,
    int$partable,
    two_level = TRUE
  )
  i <- 2L
  j <- css$cluster_idx[i]
  lcond <- function(cs) {
    INLAvaan:::loco_loglik_one(j, cs, cache$mom) -
      INLAvaan:::loo_fixedx_const_loco(int, cs, j, cache$mom)
  }
  rows <- setdiff(which(css$cluster_idx == j), i)
  Yj <- X[rows, , drop = FALSE]
  css_m <- css
  css_m$n_j[j] <- length(rows)
  css_m$S[[j]] <- crossprod(sweep(Yj, 2L, colMeans(Yj), "-"))
  css_m$ybar[j, ] <- colMeans(Yj)
  css_m$mean_d[[j]] <- colMeans(Yj)[css$zy_idx]
  expect_equal(
    res_row$per_unit$l_star[1L],
    lcond(css) - lcond(css_m),
    tolerance = 1e-8
  )

  # Within-only covariates (no cluster-level z) also run
  twolevel_model_w <- "
    level: 1
      fw =~ y1 + y2 + y3
      fw ~ x1 + x2
    level: 2
      fb =~ y1 + y2 + y3
  "
  fit_w <- asem(
    twolevel_model_w,
    d24,
    cluster = "cluster",
    fixed.x = TRUE,
    meanstructure = TRUE,
    verbose = FALSE,
    nsamp = 3,
    test = "none",
    vb_correction = FALSE,
    marginal_method = "marggaus",
    marginal_correction = "none"
  )
  res_w <- loo(fit_w, units = 1:10)
  expect_equal(res_w$flavour, "conditional")
  expect_true(all(is.finite(res_w$per_unit$log_cpo_2)))
  expect_equal(
    sum(loo(fit_w)$per_unit$l_star),
    ll_at_mode(fit_w),
    tolerance = 1e-6
  )
})

test_that("waic scores fixed.x fits conditionally", {
  set.seed(1)
  w <- suppressWarnings(waic(fit_c, nsamp = 100))
  expect_equal(w$flavour, "conditional")
  expect_true(all(is.finite(w$per_unit$lpd)))
  expect_output(print(w), "conditionally on the exogenous covariates")
})

test_that("compare(loo = TRUE) enforces the flavour rules", {
  # Conditional fits may condition on different covariate sets (covariate
  # selection); the outcome variables match, so the comparison is valid
  HS_model_x1 <- "
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
    speed   =~ x7 + x8 + x9
    visual ~ ageyr
  "
  fit_c1 <- asem(
    HS_model_x1,
    dat_x,
    fixed.x = TRUE,
    meanstructure = TRUE,
    verbose = FALSE,
    nsamp = 3,
    test = "none",
    vb_correction = FALSE,
    marginal_method = "marggaus",
    marginal_correction = "none"
  )
  cmp <- compare(fit_c, fit_c1, loo = TRUE)
  expect_true(all(
    c("ELPD", "SE", "elpd_diff", "se_diff") %in% names(cmp)
  ))
  expect_true(all(is.finite(cmp$se_diff)))

  # Mixing joint and conditional scores is refused
  expect_error(
    compare(fit_c, fit_j, loo = TRUE),
    "mix joint and conditional"
  )
})

test_that("Conditional flavour works without a mean structure", {
  # Regression test: this combination (the default for SEM with exogenous
  # covariates) used to abort. The per-unit contribution is the difference
  # of two exchangeable case-deletion conditionals; the x-block term is a
  # theta-free constant evaluated at the frozen covariate moments.
  mod_x <- "
    visual  =~ x1 + x2 + x3
    visual  ~ grade
  "
  set.seed(123)
  dat_x <- lavaan::HolzingerSwineford1939[, c("x1", "x2", "x3", "grade")]
  dat_x <- dat_x[sort(sample(nrow(dat_x), 90)), ]
  fit <- asem(mod_x, dat_x, verbose = FALSE, test = "none", nsamp = 3)
  expect_false(fit@Model@meanstructure)
  res <- loo(fit)
  int <- get_inlavaan_internal(fit)
  Y <- int$lavdata@X[[1L]]
  n <- nrow(Y)
  cc <- n / (n - 1)
  ybar <- colMeans(Y)
  x_idx <- int$lavsamplestats@x.idx[[1L]]
  Sg <- loo_grad_cache(
    int$theta_star,
    int$lavmodel,
    int$partable
  )$mom[[1L]]$Sigma
  ldmvn <- function(y, mu, S) {
    ch <- chol(S)
    d <- backsolve(ch, y - mu, transpose = TRUE)
    -0.5 * (length(y) * log(2 * pi) + 2 * sum(log(diag(ch))) + sum(d^2))
  }
  ii <- c(1L, ceiling(n / 2), n)
  l_closed <- vapply(
    ii,
    function(i) {
      ybar_m <- (n * ybar - Y[i, ]) / (n - 1)
      ldmvn(Y[i, ], ybar_m, cc * Sg) -
        ldmvn(Y[i, x_idx], ybar_m[x_idx], cc * Sg[x_idx, x_idx, drop = FALSE])
    },
    numeric(1)
  )
  expect_equal(res$per_unit$l_star[ii], l_closed, tolerance = 1e-10)
})

# FIML LOO (single-level, missing data). Under FIML the fitted likelihood is
# the observed-data likelihood, so each unit is scored on the entries it
# actually has, l_i = log N(y_i,obs; mu[o_i], Sigma[o_i, o_i]); the Taylor
# machinery is unchanged save for pattern-aware casewise kernels.

HS_model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"

# A 70-row subsample (bigger than a plain complete-data fixture, since 12%
# MCAR holes across 6 columns still need to yield a genuine mix of complete,
# single-hole, and multi-hole rows plus several distinct FIML patterns).
# MCAR holes in x4-x9 (x1-x3 kept complete so every row has >= 3 observed
# entries); the seed is set immediately before the holes so the dataset is
# reproducible and the pinned reference values are stable
make_miss <- function() {
  set.seed(1)
  d <- lavaan::HolzingerSwineford1939[
    sample(nrow(lavaan::HolzingerSwineford1939), 70),
    paste0("x", 1:9)
  ]
  set.seed(20260613)
  for (v in paste0("x", 4:9)) {
    d[[v]][runif(nrow(d)) < 0.12] <- NA
  }
  d
}
dat <- make_miss()

fit <- acfa(
  HS_model,
  dat,
  meanstructure = TRUE,
  missing = "ml",
  verbose = FALSE,
  nsamp = 3,
  test = "none",
  vb_correction = FALSE,
  marginal_method = "marggaus",
  marginal_correction = "none"
)
res <- loo(fit)

test_that("the test dataset has the expected missingness", {
  expect_equal(sum(is.na(dat)), 55L)
  expect_equal(sum(!complete.cases(dat)), 41L)
  X <- get_inlavaan_internal(fit)$lavdata@X[[1L]]
  expect_equal(nrow(X), 70L) # no rows dropped under FIML
  expect_true(anyNA(X)) # NAs retained, not listwise-deleted
  expect_length(INLAvaan:::fiml_patterns(X), 18L)
})

test_that("FIML LOSO matches reference values", {
  # Reference values from an independent prototype of the same observed-data
  # Taylor LOO formulas on this exact fit (lab 02-package-validation.R),
  # cross-checked below against lavaan's FIML loglik and finite differences
  expect_equal(res$type, "loso")
  expect_equal(res$flavour, "joint")
  expect_equal(res$n_units, 70L)
  expect_equal(res$elpd_1, -810.8096330551, tolerance = 1e-4)
  expect_equal(res$elpd_2, -829.4835098090, tolerance = 1e-4)
  expect_equal(res$se_1, 21.8869310500, tolerance = 1e-4)
  expect_equal(res$se_2, 22.5837971790, tolerance = 1e-4)
  expect_equal(res$p_loo_1, 27.7612891234, tolerance = 1e-2)
  expect_equal(res$p_loo_2, 31.1295585209, tolerance = 1e-2)

  # rows spanning complete (4), one hole (2), and three holes (11)
  pu <- res$per_unit[c(4L, 2L, 11L), ]
  expect_equal(pu$unit, c(4L, 2L, 11L))
  expect_equal(
    pu$l_star,
    c(-10.03278944439, -10.69608891448, -8.96709021099),
    tolerance = 1e-4
  )
  expect_equal(
    pu$log_cpo_1,
    c(-10.09565853301, -10.79854756016, -9.08394011848),
    tolerance = 1e-4
  )
  expect_equal(
    pu$log_cpo_2,
    c(-10.23563590292, -10.94860225615, -9.19913075203),
    tolerance = 1e-4
  )
  expect_equal(
    pu$det_term,
    c(-0.13882114800, -0.14444389100, -0.10621393414),
    tolerance = 1e-3
  )
})

test_that("casewise observed-data logliks sum to the fitted FIML loglik", {
  int <- get_inlavaan_internal(fit)
  x <- INLAvaan:::pars_to_x(int$theta_star, int$partable)
  lm_x <- lavaan::lav_model_set_parameters(int$lavmodel, x)
  opts <- fit@Options
  opts$estimator <- "ML"
  ll <- INLAvaan:::lavaan___lav_model_loglik(
    lavdata = int$lavdata,
    lavsamplestats = int$lavsamplestats,
    lavimplied = lavaan::lav_model_implied(lm_x),
    lavmodel = lm_x,
    lavoptions = opts
  )$loglik
  expect_equal(sum(res$per_unit$l_star), ll, tolerance = 1e-6)
})

test_that("pattern-scattered scores match finite differences on missing rows", {
  int <- get_inlavaan_internal(fit)
  dv <- INLAvaan:::loso_data_view(
    int$lavmodel,
    int$lavdata,
    x_idx = int$lavsamplestats@x.idx
  )
  uv <- INLAvaan:::loso_resolve_units(int$lavdata, NULL)
  X <- int$lavdata@X[[1L]]
  miss_count <- rowSums(is.na(X))
  # complete, one-hole, and a maximally sparse row
  i_chk <- c(
    which(miss_count == 0L)[1L],
    which(miss_count == 1L)[1L],
    which(miss_count == max(miss_count))[1L]
  )
  s_an <- INLAvaan:::loso_scores_units(
    int$theta_star,
    uv,
    dv,
    int$lavmodel,
    int$partable
  )[i_chk, ]
  h <- 1e-6
  g_num <- vapply(
    seq_along(int$theta_star),
    function(k) {
      tp <- tm <- int$theta_star
      tp[k] <- tp[k] + h
      tm[k] <- tm[k] - h
      cp <- INLAvaan:::loo_grad_cache(tp, int$lavmodel, int$partable)
      cm <- INLAvaan:::loo_grad_cache(tm, int$lavmodel, int$partable)
      (INLAvaan:::loso_loglik_units(uv, dv, cp$mom)[i_chk] -
        INLAvaan:::loso_loglik_units(uv, dv, cm$mom)[i_chk]) /
        (2 * h)
    },
    numeric(length(i_chk))
  )
  expect_equal(max(abs(s_an - g_num)), 0, tolerance = 1e-6)
})

test_that("loo object structure and internal identities", {
  expect_s3_class(res, "inlavaan_loo")
  expect_named(
    res$per_unit,
    c(
      "unit",
      "nobs",
      "l_star",
      "score_norm",
      "lpd_1",
      "lpd_2",
      "log_cpo_1",
      "log_cpo_2",
      "det_term",
      "ok"
    )
  )
  expect_true(all(res$per_unit$ok))
  expect_true(all(res$per_unit$nobs == 1L))
  expect_equal(
    res$per_unit$lpd_1 + res$per_unit$log_cpo_1,
    2 * res$per_unit$l_star
  )
  expect_equal(
    unname(res$estimates["elpd_loo", "Estimate"]),
    res$elpd_2
  )

  # unit subsetting addresses case numbers and agrees with the full run
  res10 <- loo(fit, units = 1:10)
  expect_equal(nrow(res10$per_unit), 10L)
  expect_equal(
    res10$per_unit$log_cpo_2,
    res$per_unit$log_cpo_2[1:10],
    tolerance = 1e-8
  )
})

test_that("waic() runs on a FIML fit and agrees loosely with loo()", {
  set.seed(1)
  w <- suppressWarnings(waic(fit, nsamp = 150))
  expect_s3_class(w, "inlavaan_waic")
  expect_equal(w$n_units, 70L)
  expect_equal(w$type, "loso")
  expect_equal(w$flavour, "joint")
  expect_true(all(is.finite(w$per_unit$lpd)))
  # WAIC and LOO estimate the same quantity; loose agreement on this model
  expect_equal(
    unname(w$estimates["elpd_waic", "Estimate"]),
    res$elpd_2,
    tolerance = 0.01
  )
})

# Two-level FIML (per-cluster LOCO) is supported; see test-loo-missing-2l.R
# for the reference-pinned coverage. Only the per-row deletion override
# remains gated under missing data.

# Multigroup LOSO: groups are independent, so every single-group kernel
# applies blockwise with group-indexed moments and constants. Units are
# identified by case index, so they keep their identity across fits that
# stack groups differently. The data are shuffled so the group stack does
# not coincide with the data order, exercising the id bookkeeping.

HS_model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"
# 60 rows per school (120 total, vs. the full 301) keeps both groups
# represented and is small enough to fit fast, while still agreeing with
# waic() closely enough for the tolerance below (smaller per-group samples,
# e.g. 20, push the loo/waic elpd_waic relative gap past it). The row shuffle
# after sampling keeps the group-stacked order different from the sampled
# data's own order, exercising the id bookkeeping (the point of the original
# full-data shuffle).
set.seed(421)
full_hs <- lavaan::HolzingerSwineford1939
dat_mg <- do.call(
  rbind,
  by(full_hs, full_hs$school, function(g) g[sample(nrow(g), 60), ])
)
rownames(dat_mg) <- NULL
dat_mg <- dat_mg[sample(nrow(dat_mg)), ]
rownames(dat_mg) <- NULL

fit_args <- list(
  meanstructure = TRUE,
  verbose = FALSE,
  nsamp = 3,
  test = "none",
  vb_correction = FALSE,
  marginal_method = "marggaus",
  marginal_correction = "none"
)
fit_conf <- do.call(
  acfa,
  c(list(HS_model, dat_mg, group = "school"), fit_args)
)
res_conf <- loo(fit_conf)

# Total loglik of the fit at the INLAvaan mode (conditional under fixed.x)
ll_at_mode <- function(fit) {
  int <- get_inlavaan_internal(fit)
  K <- if (isTRUE(int$lavmodel@ceq.simple.only)) {
    int$lavmodel@ceq.simple.K
  } else {
    NULL
  }
  theta_unp <- if (is.null(K)) {
    int$theta_star
  } else {
    as.numeric(K %*% int$theta_star)
  }
  lm_x <- lavaan::lav_model_set_parameters(
    int$lavmodel,
    INLAvaan:::pars_to_x(theta_unp, int$partable)
  )
  opts <- fit@Options
  opts$estimator <- "ML"
  INLAvaan:::lavaan___lav_model_loglik(
    lavdata = int$lavdata,
    lavsamplestats = int$lavsamplestats,
    lavimplied = lavaan::lav_model_implied(lm_x),
    lavmodel = lm_x,
    lavoptions = opts
  )$loglik
}

test_that("multigroup LOSO structure: case ids, groups, and identities", {
  expect_s3_class(res_conf, "inlavaan_loo")
  expect_equal(res_conf$type, "loso")
  expect_equal(res_conf$n_units, 120L)
  expect_equal(res_conf$n_groups, 2L)
  expect_named(
    res_conf$per_unit,
    c(
      "unit",
      "group",
      "nobs",
      "l_star",
      "score_norm",
      "lpd_1",
      "lpd_2",
      "log_cpo_1",
      "log_cpo_2",
      "det_term",
      "ok"
    )
  )
  expect_true(all(res_conf$per_unit$ok))

  # Units are the case indices of the group-stacked rows, with labels
  int <- get_inlavaan_internal(fit_conf)
  expect_equal(
    res_conf$per_unit$unit,
    unlist(int$lavdata@case.idx)
  )
  expect_equal(
    res_conf$per_unit$group,
    rep(
      int$lavdata@group.label,
      times = vapply(int$lavdata@X, nrow, integer(1))
    )
  )

  # Unit logliks sum to the model loglik at the mode
  expect_equal(
    sum(res_conf$per_unit$l_star),
    ll_at_mode(fit_conf),
    tolerance = 1e-6
  )

  expect_output(print(res_conf), "in 2 groups")
})

test_that("unit subsetting by case id preserves order across groups", {
  ids <- res_conf$per_unit$unit
  pick <- c(ids[50L], ids[3L], ids[30L])
  res_sub <- loo(fit_conf, units = pick)
  expect_equal(res_sub$per_unit$unit, pick)
  expect_equal(
    res_sub$per_unit$log_cpo_2,
    res_conf$per_unit$log_cpo_2[match(pick, ids)],
    tolerance = 1e-8
  )
  expect_error(loo(fit_conf, units = c(1L, 1L)), "distinct")
  expect_error(loo(fit_conf, units = 0L), "distinct")
})

test_that("group-2 unit scores match numerical derivatives", {
  int <- get_inlavaan_internal(fit_conf)
  Y2 <- int$lavdata@X[[2L]][7L, , drop = FALSE]
  s_an <- as.numeric(INLAvaan:::loso_scores_theta(
    int$theta_star,
    Y2,
    int$lavmodel,
    int$partable,
    group = 2L
  ))
  h <- 1e-6
  g_num <- vapply(
    seq_along(int$theta_star),
    function(k) {
      tp <- tm <- int$theta_star
      tp[k] <- tp[k] + h
      tm[k] <- tm[k] - h
      cp <- INLAvaan:::loo_grad_cache(tp, int$lavmodel, int$partable)
      cm <- INLAvaan:::loo_grad_cache(tm, int$lavmodel, int$partable)
      (INLAvaan:::loso_loglik_all(Y2, cp$mom[[2L]]) -
        INLAvaan:::loso_loglik_all(Y2, cm$mom[[2L]])) /
        (2 * h)
    },
    numeric(1)
  )
  expect_equal(s_an, g_num, tolerance = 1e-5)
})

test_that("configural fit equals independent per-group fits", {
  # With no cross-group constraints the posterior factorises over groups,
  # so the multigroup CPOs must reproduce those of separate fits on each
  # group's data (up to independent-optimisation numerics)
  res_sep <- lapply(unique(dat_mg$school), function(s) {
    fit_g <- do.call(
      acfa,
      c(list(HS_model, dat_mg[dat_mg$school == s, ]), fit_args)
    )
    list(ids = which(dat_mg$school == s), loo = loo(fit_g))
  })
  cpo_sep <- unlist(lapply(res_sep, function(r) r$loo$per_unit$log_cpo_2))
  ids_sep <- unlist(lapply(res_sep, function(r) r$ids))
  m <- match(res_conf$per_unit$unit, ids_sep)
  expect_equal(res_conf$per_unit$log_cpo_2, cpo_sep[m], tolerance = 1e-3)
  expect_equal(
    res_conf$elpd_2,
    sum(vapply(res_sep, function(r) r$loo$elpd_2, numeric(1))),
    tolerance = 1e-5
  )
})

test_that("marginalised means: per-group case-deletion conditionals", {
  fit0 <- do.call(
    acfa,
    c(
      list(HS_model, dat_mg, group = "school"),
      modifyList(fit_args, list(meanstructure = FALSE))
    )
  )
  res0 <- loo(fit0)
  expect_true(all(res0$per_unit$ok))

  # Closed form: l_i = log phi(y_i; ybar_{g,-i}, c_g * Sigma_g) with the
  # downdated mean of the unit's own group
  int <- get_inlavaan_internal(fit0)
  mom <- INLAvaan:::loo_grad_cache(
    int$theta_star,
    int$lavmodel,
    int$partable
  )$mom
  ldmvn <- function(y, mu, S) {
    ch <- chol(S)
    d <- backsolve(ch, y - mu, transpose = TRUE)
    -0.5 * (length(y) * log(2 * pi) + 2 * sum(log(diag(ch))) + sum(d^2))
  }
  for (g in 1:2) {
    Y <- int$lavdata@X[[g]]
    n <- nrow(Y)
    cc <- n / (n - 1)
    ybar <- colMeans(Y)
    ids <- int$lavdata@case.idx[[g]]
    ii <- c(1L, n %/% 2L, n)
    l_closed <- vapply(
      ii,
      function(i) {
        ybar_m <- (n * ybar - Y[i, ]) / (n - 1)
        ldmvn(Y[i, ], ybar_m, cc * mom[[g]]$Sigma)
      },
      numeric(1)
    )
    expect_equal(
      res0$per_unit$l_star[match(ids[ii], res0$per_unit$unit)],
      l_closed,
      tolerance = 1e-10
    )
  }
})

test_that("cross-group equality constraints flow through the packed space", {
  fit_metr <- do.call(
    acfa,
    c(
      list(HS_model, dat_mg, group = "school", group.equal = "loadings"),
      fit_args
    )
  )
  int <- get_inlavaan_internal(fit_metr)
  expect_true(int$lavmodel@ceq.simple.only)

  res_metr <- loo(fit_metr)
  expect_true(all(res_metr$per_unit$ok))
  expect_equal(
    sum(res_metr$per_unit$l_star),
    ll_at_mode(fit_metr),
    tolerance = 1e-6
  )

  # The MI ladder comparison: identical units pair the fits pointwise
  cmp <- compare(fit_conf, fit_metr, loo = TRUE)
  expect_true(all(c("ELPD", "SE", "elpd_diff", "se_diff") %in% names(cmp)))
  expect_true(all(is.finite(cmp$se_diff)))
})

test_that("pooled and multigroup fits pair by unit id in compare()", {
  # The pooled fit scores units in data order, the multigroup fit in
  # group-stacked order; pairing must align them by case id
  fit_pool <- do.call(acfa, c(list(HS_model, dat_mg), fit_args))
  res_pool <- loo(fit_pool)
  expect_false(identical(res_pool$per_unit$unit, res_conf$per_unit$unit))
  expect_setequal(res_pool$per_unit$unit, res_conf$per_unit$unit)

  cmp <- compare(fit_pool, fit_conf, loo = TRUE)
  expect_true(all(is.finite(cmp$se_diff)))

  # Paired SE agrees with a manual alignment by case id
  d <- res_pool$per_unit$log_cpo_2[
    match(res_conf$per_unit$unit, res_pool$per_unit$unit)
  ] -
    res_conf$per_unit$log_cpo_2
  se_manual <- sqrt(length(d) * var(d))
  k <- which(cmp$se_diff > 0)
  expect_equal(cmp$se_diff[k], se_manual, tolerance = 1e-3)
})

test_that("conditional flavour scores fixed.x multigroup fits", {
  # The leave-one-out curvature for the conditional predictive (2 exogenous
  # covariates) is only reliably positive-definite for most units once each
  # group has enough rows; 100/group (vs. dat_mg's 60) is the smallest size
  # found to keep the >= 90% "ok" fraction asserted below, so this test uses
  # its own, slightly larger fixture rather than dat_mg.
  mod_x <- "
    visual  =~ x1 + x2 + x3
    visual ~ ageyr + grade
  "
  set.seed(421)
  dat_mg_x <- do.call(
    rbind,
    by(full_hs, full_hs$school, function(g) g[sample(nrow(g), 100), ])
  )
  rownames(dat_mg_x) <- NULL
  dat_mg_x <- dat_mg_x[sample(nrow(dat_mg_x)), ]
  rownames(dat_mg_x) <- NULL

  fit_x <- do.call(
    asem,
    c(list(mod_x, dat_mg_x, group = "school", fixed.x = TRUE), fit_args)
  )
  res_x <- loo(fit_x)
  expect_equal(res_x$flavour, "conditional")
  # a couple of influential rows fall back to first order by design
  expect_gte(res_x$n_ok, ceiling(0.9 * res_x$n_units))
  expect_true(all(is.finite(res_x$per_unit$log_cpo_1)))
  expect_true(all(is.finite(res_x$per_unit$log_cpo_2[res_x$per_unit$ok])))
  # the NA grade row is listwise-deleted, so its case id is not a unit
  expect_equal(res_x$n_units, 199L)
  expect_equal(
    sum(res_x$per_unit$l_star),
    ll_at_mode(fit_x),
    tolerance = 1e-6
  )
})

test_that("waic() supports multigroup fits and agrees with loo()", {
  set.seed(2)
  w <- suppressWarnings(waic(fit_conf, nsamp = 100))
  expect_s3_class(w, "inlavaan_waic")
  expect_equal(w$n_units, 120L)
  expect_equal(w$n_groups, 2L)
  expect_true("group" %in% names(w$per_unit))
  expect_true(all(is.finite(w$per_unit$lpd)))
  expect_output(print(w), "in 2 groups")
  expect_equal(
    unname(w$estimates["elpd_waic", "Estimate"]),
    res_conf$elpd_2,
    tolerance = 0.005
  )
})

twolevel_model <- "
  level: 1
    fw =~ y1 + y2 + y3
    fw ~ x1 + x2 + x3
  level: 2
    fb =~ y1 + y2 + y3
    fb ~ w1 + w2
"

# Shrunk two-level fixture: first 24 cluster ids (6 full cycles of the
# 5/10/15/20 cluster-size pattern that repeats every 4 cluster ids in
# lavaan::Demo.twolevel), 300 rows total. Shared across both the main
# `fit` below and `fit_fx` (the fixed.x = TRUE variant further down).
d_sub <- lavaan::Demo.twolevel[lavaan::Demo.twolevel$cluster %in% 1:24, ]

fit <- asem(
  twolevel_model,
  d_sub,
  cluster = "cluster",
  meanstructure = TRUE,
  fixed.x = FALSE,
  verbose = FALSE,
  nsamp = 3,
  test = "none",
  vb_correction = FALSE,
  marginal_method = "marggaus",
  marginal_correction = "none"
)
res <- loo(fit)

test_that("LOCO matches reference values", {
  # Reference values computed with an independent implementation of the same
  # Taylor LOO formulas on this exact fit
  expect_equal(res$type, "loco")
  expect_equal(res$n_units, 24L)
  expect_equal(res$elpd_1, -2811.3389254026, tolerance = 1e-4)
  expect_equal(res$elpd_2, -2832.9728115262, tolerance = 1e-4)
  expect_equal(res$se_1, 259.1529664233, tolerance = 1e-4)
  expect_equal(res$se_2, 260.9893088931, tolerance = 1e-4)
  # p_loo aggregates small differences of nearly-equal numbers, so it is
  # more sensitive to BLAS/optimiser endpoint noise than the elpd totals
  expect_equal(res$p_loo_1, 30.7792227732, tolerance = 1e-2)
  expect_equal(res$p_loo_2, 35.1657403910, tolerance = 1e-2)

  # cluster ids 1, 6, 8 have sizes 5, 10, 20 respectively under the
  # 5/10/15/20 cycle
  pu <- res$per_unit[c(1L, 6L, 8L), ]
  expect_equal(pu$nobs, c(5L, 10L, 20L))
  expect_equal(
    pu$l_star,
    c(-45.1868700825, -91.3569642108, -177.6891056793),
    tolerance = 1e-4
  )
  expect_equal(
    pu$log_cpo_1,
    c(-45.3396350528, -91.5858133174, -178.4008589303),
    tolerance = 1e-4
  )
  expect_equal(
    pu$log_cpo_2,
    c(-45.5376421501, -92.0469159588, -179.5548931984),
    tolerance = 1e-4
  )
  # det_term values are tiny finite-difference remainders whose relative
  # error is dominated by cross-platform noise; their correctness is pinned
  # through log_cpo_2 above, so only check structure here
  expect_true(all(is.finite(pu$det_term)))
  expect_true(all(pu$det_term < 0))
})

test_that("LOCO structure and internal identities", {
  expect_s3_class(res, "inlavaan_loo")
  expect_true(all(res$per_unit$ok))
  expect_equal(sum(res$per_unit$nobs), nrow(d_sub))
  expect_equal(
    res$per_unit$lpd_1 + res$per_unit$log_cpo_1,
    2 * res$per_unit$l_star
  )
  expect_output(print(res), "leave-one-cluster-out")

  # Sum of cluster logliks equals the model loglik at the mode
  int <- get_inlavaan_internal(fit)
  x <- INLAvaan:::pars_to_x(int$theta_star, int$partable)
  lm_x <- lavaan::lav_model_set_parameters(int$lavmodel, x)
  opts <- fit@Options
  opts$estimator <- "ML"
  ll <- INLAvaan:::lavaan___lav_model_loglik(
    lavdata = int$lavdata,
    lavsamplestats = int$lavsamplestats,
    lavimplied = lavaan::lav_model_implied(lm_x),
    lavmodel = lm_x,
    lavoptions = opts
  )$loglik
  expect_equal(sum(res$per_unit$l_star), ll, tolerance = 1e-6)
})

test_that("LOCO unit subsetting and theta/Sigma override", {
  sub <- c(3L, 7L, 11L)
  res_sub <- loo(fit, units = sub)
  expect_equal(res_sub$per_unit$unit, sub)
  expect_equal(
    res_sub$per_unit$log_cpo_2,
    res$per_unit$log_cpo_2[sub],
    tolerance = 1e-8
  )

  int <- get_inlavaan_internal(fit)
  res_same <- loo(
    fit,
    theta = int$theta_star,
    Sigma = int$Sigma_theta,
    units = sub
  )
  expect_true(res_same$theta_overridden)
  expect_equal(
    res_same$per_unit$log_cpo_2,
    res$per_unit$log_cpo_2[sub],
    tolerance = 1e-8
  )
})

test_that("two-level LOSO override scores row deletions", {
  expect_warning(
    res_row <- loo(fit, type = "loso", units = 1:8),
    "leave-one-unit-out"
  )
  expect_equal(res_row$type, "loso")
  expect_equal(nrow(res_row$per_unit), 8L)
  expect_true(all(res_row$per_unit$ok))
  expect_true(all(res_row$per_unit$nobs == 1L))

  int <- get_inlavaan_internal(fit)
  css <- INLAvaan:::loco_suff_stats(int$lavdata)
  X <- int$lavdata@X[[1L]]
  cache <- INLAvaan:::loo_grad_cache(
    int$theta_star,
    int$lavmodel,
    int$partable,
    two_level = TRUE
  )

  # The downdated sufficient statistics agree with rebuilding the
  # cluster-minus-row statistics from the raw data
  i <- 2L
  j <- css$cluster_idx[i]
  rows <- setdiff(which(css$cluster_idx == j), i)
  Yj <- X[rows, , drop = FALSE]
  us_raw <- INLAvaan:::loco_stats_build(
    length(rows),
    crossprod(sweep(Yj, 2L, colMeans(Yj), "-")),
    colMeans(Yj)[css$zy_idx],
    css
  )
  ll_minus_raw <- INLAvaan:::loco_loglik_us(us_raw, cache$mom)
  ll_full <- INLAvaan:::loco_loglik_one(j, css, cache$mom)
  expect_equal(
    res_row$per_unit$l_star[i],
    ll_full - ll_minus_raw,
    tolerance = 1e-8
  )

  # Analytic row score matches a numerical derivative
  s2 <- as.numeric(INLAvaan:::loso2l_scores_theta(
    int$theta_star,
    css,
    X,
    int$lavmodel,
    int$partable,
    units = i
  ))
  h <- 1e-6
  g_num <- vapply(
    seq_along(int$theta_star),
    function(k) {
      tp <- tm <- int$theta_star
      tp[k] <- tp[k] + h
      tm[k] <- tm[k] - h
      cp <- INLAvaan:::loo_grad_cache(
        tp,
        int$lavmodel,
        int$partable,
        two_level = TRUE
      )
      cm <- INLAvaan:::loo_grad_cache(
        tm,
        int$lavmodel,
        int$partable,
        two_level = TRUE
      )
      (INLAvaan:::loso2l_loglik_all(i, css, X, cp$mom) -
        INLAvaan:::loso2l_loglik_all(i, css, X, cm$mom)) /
        (2 * h)
    },
    numeric(1)
  )
  expect_equal(s2, g_num, tolerance = 1e-5)
})

test_that("fixed.x two-level fits are scored conditionally", {
  fit_fx <- asem(
    twolevel_model,
    d_sub,
    cluster = "cluster",
    meanstructure = TRUE,
    fixed.x = TRUE,
    verbose = FALSE,
    nsamp = 3,
    test = "none",
    vb_correction = FALSE,
    marginal_method = "marggaus",
    marginal_correction = "none"
  )
  res_fx <- loo(fit_fx, units = 1:5)
  expect_equal(res_fx$flavour, "conditional")
  expect_true(all(is.finite(res_fx$per_unit$log_cpo_2)))
})

test_that("waic gains type: conditional (leave-one-unit-out) WAIC", {
  # default is marginal (per-cluster) WAIC
  set.seed(1)
  w_marg <- suppressWarnings(waic(fit, nsamp = 120))
  expect_equal(w_marg$type, "loco")

  # type = "loso" warns and scores the conditional (leave-one-unit-out) WAIC,
  # the same estimand as loo(type = "loso"); the two routes agree loosely
  w_cond <- testthat::capture_warnings(
    w <- waic(fit, type = "loso", units = 1:5, nsamp = 120)
  )
  expect_true(any(grepl("leave-one-unit-out", w_cond)))
  expect_equal(w$type, "loso")
  expect_equal(w$n_units, 5L)
  expect_true(all(w$per_unit$nobs == 1L))

  l <- suppressWarnings(loo(fit, type = "loso", units = 1:5))
  expect_equal(
    unname(w$estimates["elpd_waic", "Estimate"]),
    l$elpd_2,
    tolerance = 0.05
  )
})

#' Convergence and Approximation Diagnostics for INLAvaan Models
#'
#' Extract convergence and approximation-quality diagnostics from a fitted
#' \code{INLAvaan} model.
#'
#' @param object An object of class [INLAvaan].
#' @param type Character. \code{"global"} (default) returns a named numeric
#'   vector of scalar diagnostics. \code{"param"} returns a data frame with one
#'   row per free parameter containing per-parameter diagnostics.
#' @param ... Currently unused.
#'
#' @details
#' \strong{Global diagnostics} (\code{type = "global"}):
#' \describe{
#'   \item{\code{npar}}{Number of free parameters.}
#'   \item{\code{nsamp}}{Number of posterior samples drawn.}
#'   \item{\code{converged}}{1 if the optimiser converged, 0 otherwise.}
#'   \item{\code{iterations}}{Number of optimiser iterations.}
#'   \item{\code{grad_inf}}{L-infinity norm of the analytic gradient at the mode
#'     (max |grad|). Should be ~0 at convergence.}
#'   \item{\code{grad_inf_rel}}{Relative L-infinity norm of the analytic gradient
#'     (max |grad| / (|par| + 1e-6)).}
#'   \item{\code{grad_l2}}{L2 (Euclidean) norm of the analytic gradient at the mode.}
#'   \item{\code{mode_shift_max}}{Maximum, across parameters, of the Newton
#'     step at the reported mode in posterior-SD units
#'     (max |\eqn{\Sigma_\theta} grad| / se). Unlike the raw gradient norms
#'     this is scale-free: it estimates how far the reported mode sits from
#'     the true posterior mode relative to the posterior uncertainty. Should
#'     be ~0 at convergence.}
#'   \item{\code{hess_cond}}{Condition number of the Hessian (precision matrix)
#'     computed from \eqn{\Sigma_\theta}. Large values indicate near-singularity.}
#'   \item{\code{vb_kld_global}}{Global KL divergence from the VB mean correction
#'     (NA if VB correction was not applied).}
#'   \item{\code{vb_applied}}{1 if VB correction was applied, 0 otherwise.}
#'   \item{\code{kld_max}}{Maximum per-parameter KL divergence from the VB correction.}
#'   \item{\code{kld_mean}}{Mean per-parameter KL divergence.}
#'   \item{\code{nmad_max}}{Maximum normalised max-absolute-deviation across
#'     marginals (skew-normal method only; NA otherwise).}
#'   \item{\code{nmad_mean}}{Mean NMAD across marginals.}
#' }
#'
#' \strong{Per-parameter diagnostics} (\code{type = "param"}):
#' A data frame with columns:
#' \describe{
#'   \item{\code{param}}{Parameter name.}
#'   \item{\code{grad}}{Analytic gradient of the negative log-posterior at the
#'     mode. Should be ~0 at convergence.}
#'   \item{\code{grad_num}}{Numerical (finite-difference) gradient at the mode.
#'     Should agree with \code{grad}; large discrepancies indicate a bug in the
#'     analytic gradient.}
#'   \item{\code{grad_diff}}{Difference \code{grad_num - grad}: should be ~0.}
#'   \item{\code{grad_abs}}{Absolute analytic gradient.}
#'   \item{\code{grad_rel}}{Relative analytic gradient |grad| / (|par| + 1e-6).}
#'   \item{\code{mode_shift_sigma}}{Newton step at the reported mode in
#'     posterior-SD units. Should be ~0 at convergence.}
#'   \item{\code{kld}}{Per-parameter KL divergence from the VB correction.}
#'   \item{\code{vb_shift}}{VB correction shift (in original scale).}
#'   \item{\code{vb_shift_sigma}}{VB shift in units of posterior SD.}
#'   \item{\code{nmad}}{Normalised max-absolute-deviation of the skew-normal fit
#'     (NA when not using the skewnorm method).}
#' }
#'
#' \strong{Fit-time warnings}: [inlavaan()] runs these checks once at the end
#' of every fit and emits a single consolidated warning (condition class
#' \code{"inlavaan_diagnostics_warning"}) when any of them look off: the
#' optimiser did not converge, \code{mode_shift_max} exceeds 0.1,
#' any marginal has NMAD above 0.1, the VB correction shifted a posterior
#' mean by more than 1 posterior SD, or the Hessian condition number exceeds
#' 1e8. A healthy fit stays silent. Silence the check with
#' \code{suppressWarnings()}, or selectively by handling the condition class.
#'
#' @returns For \code{type = "global"}, a named numeric vector (class
#'   \code{"diagnostics.INLAvaan"}). For \code{type = "param"}, a data frame
#'   (class \code{c("diagnostics.INLAvaan.param", "data.frame")}).
#'
#' @examples
#' \donttest{
#' HS.model <- "
#'   visual  =~ x1 + x2 + x3
#'   textual =~ x4 + x5 + x6
#'   speed   =~ x7 + x8 + x9
#' "
#' utils::data("HolzingerSwineford1939", package = "lavaan")
#' fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
#'             test = "none", verbose = FALSE)
#'
#' # Global convergence summary
#' diagnostics(fit)
#'
#' # Per-parameter table
#' diagnostics(fit, type = "param")
#' }
#'
#' @seealso [timing()], [fitmeasures()], [plot()]
#'
#' @export
setGeneric("diagnostics", function(object, ...) standardGeneric("diagnostics"))

#' @name diagnostics
#' @rdname diagnostics
#' @aliases diagnostics,INLAvaan-method
#' @export
setMethod(
  "diagnostics",
  "INLAvaan",
  function(object, type = c("global", "param"), ...) {
    type <- match.arg(type)
    dg <- diagnostics_internal(object@external$inlavaan_internal)
    if (type == "global") {
      dg$global
    } else {
      dg$param
    }
  }
)

# Converged flag and iteration count from an optimiser result; mirrors how
# create_lav_from_inlavaan_internal() fills the lavaan @Fit slot.
optim_convergence <- function(opt, optim_method) {
  if (optim_method == "nlminb") {
    list(
      converged = opt$convergence == 0L,
      iterations = as.integer(opt$iterations)
    )
  } else if (optim_method == "optim") {
    # nocov start
    list(
      converged = opt$convergence == 0L,
      iterations = as.integer(opt$counts["function"])
    )
  } else {
    # ucminf
    list(
      converged = opt$convergence == 1L,
      iterations = as.integer(opt$info["neval"])
    )
  } # nocov end
}

# Shared computation behind diagnostics() and the fit-time warning; works on
# the internal list so it can run before the S4 object exists.
diagnostics_internal <- function(int) {
  grad_analytic <- int$opt$dx_analytic
  grad_num <- int$opt$dx
  pars <- int$theta_star
  parnames <- names(int$coefficients)
  m <- length(parnames)

  Sigma_theta <- int$Sigma_theta
  se_laplace <- sqrt(diag(Sigma_theta))
  conv <- optim_convergence(int$opt, int$optim_method)

  # Newton step at the reported mode in posterior-SD units: a scale-free
  # measure of how far a non-zero gradient says the true mode is
  mode_shift_sigma <- abs(as.numeric(Sigma_theta %*% grad_analytic)) /
    se_laplace

  # VB diagnostics
  vb <- int$vb
  vb_applied <- all(!is.na(vb$correction))
  vb_shift <- if (vb_applied) vb$correction else rep(NA_real_, m)
  vb_kld <- if (vb_applied) vb$kld else rep(NA_real_, m)
  vb_shift_sigma <- vb_shift / se_laplace

  # NMAD (skewnorm method only); approx_data may carry extra rows for
  # covariance/defined parameters, so keep the first m (marginal-scan) rows
  nmad <- tryCatch(
    int$approx_data[seq_len(m), "nmad"],
    error = function(e) rep(NA_real_, m)
  )
  if (is.null(nmad) || length(nmad) == 0) {
    nmad <- rep(NA_real_, m)
  }

  # Hessian condition number: kappa(H) = kappa(Sigma_theta)
  eig <- eigen(Sigma_theta, symmetric = TRUE, only.values = TRUE)$values
  hess_cond <- if (length(eig) > 0) max(eig) / min(eig) else NA_real_

  global <- c(
    npar = m,
    nsamp = int$nsamp,
    converged = as.numeric(conv$converged),
    iterations = conv$iterations,
    grad_inf = max(abs(grad_analytic)),
    grad_inf_rel = max(abs(grad_analytic) / (abs(pars) + 1e-06)),
    grad_l2 = sqrt(sum(grad_analytic^2)),
    mode_shift_max = max(mode_shift_sigma),
    hess_cond = hess_cond,
    vb_applied = as.numeric(vb_applied),
    vb_kld_global = if (vb_applied) vb$kld_global else NA_real_,
    kld_max = if (all(is.na(vb_kld))) NA_real_ else max(vb_kld, na.rm = TRUE),
    kld_mean = if (all(is.na(vb_kld))) NA_real_ else mean(vb_kld, na.rm = TRUE),
    nmad_max = if (all(is.na(nmad))) NA_real_ else max(nmad, na.rm = TRUE),
    nmad_mean = if (all(is.na(nmad))) NA_real_ else mean(nmad, na.rm = TRUE)
  )
  class(global) <- c("diagnostics.INLAvaan", "numeric")

  param <- data.frame(
    names = parnames,
    grad = grad_analytic,
    grad_num = grad_num,
    grad_diff = grad_num - grad_analytic,
    grad_abs = abs(grad_analytic),
    grad_rel = abs(grad_analytic) / (abs(pars) + 1e-06),
    mode_shift_sigma = mode_shift_sigma,
    kld = vb_kld,
    vb_shift = vb_shift,
    vb_shift_sigma = vb_shift_sigma,
    nmad = nmad,
    row.names = NULL,
    stringsAsFactors = FALSE
  )
  class(param) <- c("diagnostics.INLAvaan.param", "data.frame")

  list(global = global, param = param)
}

# Fit-time diagnostics check, called once at the end of inlavaan(). Emits a
# single consolidated warning when the checks reported by diagnostics() look
# off. Thresholds are deliberately loose so that a healthy fit stays silent:
# healthy fits sit orders of magnitude below them (mode shifts ~1e-4 SD,
# NMAD < 0.06, VB shifts < 0.6 SD, condition numbers < 1e3).
warn_fit_diagnostics <- function(
  int,
  mode_shift_tol = 0.1,
  nmad_tol = 0.1,
  vb_shift_tol = 1,
  hess_cond_tol = 1e8
) {
  dg <- diagnostics_internal(int)
  glob <- dg$global
  param <- dg$param
  issues <- character(0)

  fmt <- function(x) formatC(x, digits = 3, format = "g")

  if (glob[["converged"]] == 0) {
    msg <- int$opt$message
    issues <- c(
      issues,
      "x" = paste0(
        "The optimiser did not converge",
        if (!is.null(msg) && nzchar(msg)) paste0(": ", msg),
        "."
      )
    )
  }

  if (isTRUE(glob[["mode_shift_max"]] > mode_shift_tol)) {
    worst <- param$names[which.max(param$mode_shift_sigma)]
    issues <- c(
      issues,
      "x" = paste0(
        "The gradient at the posterior mode is not zero (max |grad| = ",
        fmt(glob[["grad_inf"]]),
        "): a Newton step would move {.code ",
        worst,
        "} by ",
        fmt(glob[["mode_shift_max"]]),
        " posterior SDs."
      )
    )
  }

  bad_nmad <- which(!is.na(param$nmad) & param$nmad > nmad_tol)
  if (length(bad_nmad)) {
    bad_nmad <- bad_nmad[order(param$nmad[bad_nmad], decreasing = TRUE)]
    shown <- head(bad_nmad, 3L)
    listed <- paste0(
      "{.code ",
      param$names[shown],
      "} (",
      formatC(param$nmad[shown], digits = 2, format = "f"),
      ")",
      collapse = ", "
    )
    more <- length(bad_nmad) - length(shown)
    issues <- c(
      issues,
      "x" = paste0(
        "The fitted marginal deviates from the scanned posterior (NMAD > ",
        nmad_tol,
        ") for ",
        listed,
        if (more > 0) paste0(" and ", more, " other", if (more > 1) "s"),
        "."
      )
    )
  }

  vbs <- abs(param$vb_shift_sigma)
  if (any(!is.na(vbs) & vbs > vb_shift_tol)) {
    worst <- param$names[which.max(vbs)]
    issues <- c(
      issues,
      "x" = paste0(
        "The VB correction shifted {.code ",
        worst,
        "} by ",
        fmt(max(vbs, na.rm = TRUE)),
        " posterior SDs; the Gaussian approximation at the mode may be
         inaccurate."
      )
    )
  }

  if (isTRUE(glob[["hess_cond"]] > hess_cond_tol)) {
    issues <- c(
      issues,
      "x" = paste0(
        "The Hessian at the mode is near-singular (condition number = ",
        fmt(glob[["hess_cond"]]),
        "); some parameters may be weakly identified."
      )
    )
  }

  if (length(issues)) {
    cli_warn(
      c(
        "Fit diagnostics flagged {length(issues)} potential issue{?s}:",
        issues,
        "i" = "Inspect with {.code diagnostics(fit)} and
               {.code diagnostics(fit, type = \"param\")}."
      ),
      class = "inlavaan_diagnostics_warning"
    )
  }
  invisible(issues)
}

#' @exportS3Method print diagnostics.INLAvaan
print.diagnostics.INLAvaan <- function(x, ...) {
  nm <- names(x)
  formatted <- vapply(
    seq_along(x),
    function(i) {
      val <- x[i]
      name <- nm[i]

      if (
        name %in% c("npar", "nsamp", "converged", "iterations", "vb_applied")
      ) {
        as.character(as.integer(round(val)))
      } else if (
        startsWith(name, "grad_") ||
          name %in% c("hess_cond", "mode_shift_max")
      ) {
        formatC(val, digits = 2, format = "e")
      } else if (is.na(val)) {
        "NA"
      } else {
        formatC(val, digits = 4, format = "f", drop0trailing = FALSE)
      }
    },
    character(1)
  )
  names(formatted) <- nm
  print(formatted, quote = FALSE, right = TRUE)
  invisible(x)
}

#' @exportS3Method print diagnostics.INLAvaan.param
print.diagnostics.INLAvaan.param <- function(x, digits = 4, ...) {
  num_cols <- sapply(x, is.numeric)
  x[, num_cols] <- round(x[, num_cols], digits)
  print.data.frame(x, ...)
  invisible(x)
}

#' Compare Bayesian Models Fitted with INLAvaan
#'
#' Compare two or more Bayesian SEM fitted with INLAvaan, reporting model-fit
#' statistics and (optionally) fit indices side by side.
#'
#' @details
#' The first argument `x` serves as the **baseline** (null) model.
#' All models (including the baseline) appear in the comparison table. The
#' baseline is also passed to [fitMeasures()][lavaan::fitMeasures] when
#' incremental fit indices (BCFI, BTLI, BNFI) are requested via
#' `fit.measures`.
#'
#' The default table always includes:
#'
#'   - **npar**: Number of free parameters.
#'   - **Marg.Loglik**: Approximated marginal log-likelihood.
#'   - **logBF**: Natural-log Bayes factor relative to the best model.
#'   - **DIC** / **pD**: Deviance Information Criterion and effective number
#'     of parameters (when `test != "none"` was used during fitting).
#'
#' Set `fit.measures` to a character vector of measure names (anything
#' returned by [fitMeasures()][lavaan::fitMeasures]) to append extra columns.
#' Use `fit.measures = "all"` to include every available measure.
#'
#' Set `loo = TRUE` to compare models by leave-one-out cross-validation
#' (see [loo()]). This appends **ELPD** / **SE** (the second-order Taylor
#' expected log predictive density and its standard error), **p_loo**, and,
#' against the best-ELPD model, the difference **elpd_diff** with its
#' *paired* standard error **se_diff** computed from the pointwise
#' contributions (the appropriate uncertainty for nested or same-data
#' comparisons). The table is then sorted by descending ELPD. All models
#' must be fitted to the same data with matching units; units are paired
#' by id rather than by row order, so fits that stack groups differently
#' -- a pooled fit against a multigroup fit, or multigroup fits with
#' different group orderings -- still pair up unit by unit. For
#' missing-data (FIML) fits, "the same data" also means the same observed
#' entries: each unit is scored on the entries it has, so comparisons
#' require identical missingness patterns across models. All models
#' must also share the
#' score flavour (see [loo()]): mixing fits with modelled covariates
#' (`fixed.x = FALSE`, joint scores) and fixed covariates
#' (`fixed.x = TRUE`, conditional scores) is refused. Joint scores
#' additionally require identical variable sets across models, while
#' conditional scores require only matching outcome variables -- covariate
#' sets may differ, which is the covariate-selection setting. Stored LOO
#' results (`test = "loo"` or [add_loo()]) are reused.
#'
#' `anova()` is disabled for `INLAvaan` fits -- there is no direct Bayesian
#' analogue of the classical likelihood-ratio test -- and points here instead.
#'
#' @param x An [INLAvaan] (or `inlavaan_internal`) object used as the
#'   **baseline** (null) model. It is included in the comparison table and
#'   passed to [fitMeasures()][lavaan::fitMeasures] for incremental indices.
#' @param y,... One or more [INLAvaan] (or `inlavaan_internal`) objects to
#'   compare against the baseline.
#' @param fit.measures Character vector of additional fit-measure names to
#'   include (e.g. `"BRMSEA"`, `"BCFI"`). Use `"all"` to include every
#'   measure returned by [fitMeasures()][lavaan::fitMeasures]. The default
#'   (`NULL`) shows only the core comparison statistics.
#' @param loo Logical; if `TRUE`, compare models by leave-one-out
#'   cross-validation with paired standard errors (see Details). Defaults to
#'   `FALSE`.
#' @param object An [INLAvaan] object (the `anova()` method, which is disabled
#'   and redirects to `compare()`).
#'
#' @return A data frame of class `compare.inlavaan_internal` containing model
#'   fit statistics, sorted by descending marginal log-likelihood (or by
#'   descending ELPD when `loo = TRUE`).
#'
#' @references <https://lavaan.ugent.be/tutorial/groups.html>
#'
#' @seealso [fitmeasures()], [bfit_indices()]
#'
#' @example inst/examples/ex-model_comparison.R
#' @export
setGeneric("compare", function(x, y, ..., fit.measures = NULL, loo = FALSE) {
  standardGeneric("compare")
})

#' @name compare
#' @rdname compare
#' @aliases compare,INLAvaan-method
#' @export
setMethod(
  "compare",
  "INLAvaan",
  function(x, y, ..., fit.measures = NULL, loo = FALSE) {
    mc <- match.call()
    dots <- list(...)

    # x = baseline, y + unnamed ... = models to compare
    model_objs <- c(list(x, y), dots)
    model_exprs <- c(
      list(mc$x, mc$y),
      as.list(mc)[-1][
        !names(as.list(mc)[-1]) %in%
          c("x", "y", "fit.measures", "loo")
      ]
    )

    modnames <- vapply(model_exprs, deparse, character(1))

    compare_impl(
      models = model_objs,
      modnames = modnames,
      fit.measures = fit.measures,
      baseline = x,
      loo = loo
    )
  }
)

#' @exportS3Method compare inlavaan_internal
compare.inlavaan_internal <- function(
  x,
  y,
  ...,
  fit.measures = NULL,
  loo = FALSE
) {
  mc <- match.call()
  dots <- list(...)

  model_objs <- c(list(x, y), dots)
  model_exprs <- c(
    list(mc$x, mc$y),
    as.list(mc)[-1][
      !names(as.list(mc)[-1]) %in%
        c("x", "y", "fit.measures", "loo")
    ]
  )

  modnames <- vapply(model_exprs, deparse, character(1))

  compare_impl(
    models = model_objs,
    modnames = modnames,
    fit.measures = fit.measures,
    baseline = x,
    loo = loo
  )
}

# ---- Internal workhorse ------------------------------------------------------

compare_impl <- function(
  models,
  modnames,
  fit.measures = NULL,
  baseline = NULL,
  loo = FALSE
) {
  # Normalise to internal objects, keeping originals for fitMeasures()
  originals <- models
  internals <- lapply(models, function(m) {
    if (inherits(m, "INLAvaan")) {
      m@external$inlavaan_internal
    } else if (inherits(m, "inlavaan_internal")) {
      m # nocov
    } else {
      cli_abort(
        # nocov
        "Each model must be an {.cls INLAvaan} or {.cls inlavaan_internal} object."
      )
    }
  })

  nmod <- length(internals)
  npar <- vapply(internals, function(m) length(m$theta_star), integer(1))
  marg_ll <- vapply(internals, function(m) m$mloglik, numeric(1))
  DIC_vec <- vapply(internals, function(m) m$DIC$dic %||% NA_real_, numeric(1))
  pD_vec <- vapply(internals, function(m) m$DIC$pD %||% NA_real_, numeric(1))

  # Marginal likelihoods, Bayes factors, and DIC are only comparable
  # between fits with the same mean treatment: without a mean structure
  # the saturated means carry an improper flat prior whose arbitrary
  # normalisation cancels within that flavour but is orphaned against a
  # fit with modelled (proper-prior) means. LOO comparisons are unaffected
  # (leave-one-out conditionals are proper under both treatments).
  ms_vec <- vapply(
    internals,
    function(m) isTRUE(m$lavmodel@meanstructure),
    logical(1)
  )
  if (length(unique(ms_vec)) > 1L) {
    hint <- if (isTRUE(loo)) {
      c(
        "i" = "Interpret only the ELPD columns; leave-one-out conditionals
         are proper under both treatments."
      )
    } else {
      c(
        "i" = "Use {.code compare(..., loo = TRUE)} for a comparison that
         is valid across mean treatments."
      )
    }
    cli_warn(c(
      "Comparing fits with and without a mean structure: marginal
       log-likelihoods, Bayes factors, and DIC are not comparable across
       the two mean treatments (the flat-prior normalisation of the
       saturated means does not cancel).",
      hint
    ))
  }

  best_ll <- max(marg_ll)
  logBF <- marg_ll - best_ll

  out <- data.frame(
    Model = modnames,
    npar = npar,
    Marg.Loglik = marg_ll,
    logBF = round(logBF, 3),
    stringsAsFactors = FALSE
  )

  # Append DIC/pD if any model has them
  if (!all(is.na(DIC_vec))) {
    out$DIC <- round(DIC_vec, 3)
    out$pD <- round(pD_vec, 3)
  }

  # Append extra fit measures if requested
  if (!is.null(fit.measures)) {
    has_inlavaan <- vapply(originals, function(m) is(m, "INLAvaan"), logical(1))
    if (!all(has_inlavaan)) {
      # nocov start
      cli_warn(
        "Fit measures require {.cls INLAvaan} objects; skipping for {.cls inlavaan_internal} models."
      )
    } else {
      # nocov end
      # baseline (x) is used for incremental indices
      baseline_obj <- if (is(baseline, "INLAvaan")) baseline else NULL

      fm_list <- lapply(originals, function(m) {
        tryCatch(
          fitMeasures(
            m,
            fit.measures = fit.measures,
            baseline.model = baseline_obj
          ),
          error = function(e) NULL
        )
      })
      # Union of all measure names
      all_names <- unique(unlist(lapply(fm_list, names)))
      for (nm in all_names) {
        out[[nm]] <- vapply(
          fm_list,
          function(fm) {
            if (is.null(fm) || is.na(fm[nm])) NA_real_ else round(fm[nm], 4)
          },
          numeric(1)
        )
      }
    }
  }

  # LOO comparison with paired standard errors
  if (isTRUE(loo)) {
    loo_list <- lapply(internals, function(m) {
      if (!is.null(m$loo)) m$loo else loo.inlavaan_internal(m)
    })

    # Paired differences are only meaningful for matching units on the
    # same data. Units are matched by id (case number for LOSO, cluster
    # position for LOCO), not by row order, so fits that stack groups
    # differently -- e.g. a pooled fit against a multigroup fit -- still
    # pair up unit by unit.
    pu1 <- loo_list[[1L]]$per_unit
    align <- lapply(loo_list, function(l) {
      if (
        !identical(l$type, loo_list[[1L]]$type) ||
          nrow(l$per_unit) != nrow(pu1)
      ) {
        return(NULL)
      }
      idx <- match(pu1$unit, l$per_unit$unit)
      if (anyNA(idx) || !identical(l$per_unit$nobs[idx], pu1$nobs)) {
        return(NULL)
      }
      idx
    })
    if (any(vapply(align, is.null, logical(1)))) {
      cli_abort(
        "LOO comparison requires models fitted to the same data, with
         matching units."
      )
    }

    # Joint and conditional scores live on different scales and must not
    # meet in one comparison
    flavs <- vapply(
      loo_list,
      function(l) l$flavour %||% "joint",
      character(1)
    )
    if (length(unique(flavs)) > 1L) {
      cli_abort(c(
        "LOO comparison cannot mix joint and conditional scores.",
        "i" = "Fit all models with the same {.code fixed.x} setting."
      ))
    }

    # Joint scores are densities over all modelled variables, so the
    # variable sets must match; conditional scores are densities over the
    # outcomes only, so covariate sets may differ but the outcomes must match
    score_vars <- lapply(internals, function(m) {
      ov <- sort(unique(unlist(m$lavdata@ov.names)))
      if (flavs[1L] == "conditional") {
        setdiff(ov, unlist(m$lavdata@ov.names.x))
      } else {
        ov
      }
    })
    same_vars <- vapply(
      score_vars,
      identical,
      logical(1),
      y = score_vars[[1L]]
    )
    if (!all(same_vars)) {
      if (flavs[1L] == "conditional") {
        cli_abort(
          "Conditional LOO comparison requires models for the same set of
           outcome variables (the covariate sets may differ)."
        )
      }
      cli_abort(c(
        "Joint LOO comparison requires models for the same set of observed
         variables.",
        "i" = "To drop a covariate's effect, keep the variable in the model
         without the path (e.g. {.code fb ~ w1} plus {.code w2 ~~ w1}), or
         compare {.code fixed.x = TRUE} fits, which are scored
         conditionally."
      ))
    }

    elpd <- vapply(
      loo_list,
      function(l) unname(l$estimates["elpd_loo", "Estimate"]),
      numeric(1)
    )
    best <- which.max(elpd)
    # Headline pointwise contributions (second order when available),
    # aligned to the first model's unit order for pairing
    pw <- lapply(seq_along(loo_list), function(k) {
      l <- loo_list[[k]]
      v <- if (l$second_order && l$n_ok > 0L) {
        l$per_unit$log_cpo_2
      } else {
        l$per_unit$log_cpo_1 # nocov
      }
      v[align[[k]]]
    })
    n_units <- nrow(pu1)

    out$ELPD <- round(elpd, 3)
    out$SE <- round(
      vapply(
        loo_list,
        function(l) unname(l$estimates["elpd_loo", "SE"]),
        numeric(1)
      ),
      3
    )
    out$p_loo <- round(
      vapply(
        loo_list,
        function(l) unname(l$estimates["p_loo", "Estimate"]),
        numeric(1)
      ),
      3
    )
    out$elpd_diff <- round(elpd - elpd[best], 3)
    out$se_diff <- round(
      vapply(
        seq_along(loo_list),
        function(k) {
          if (k == best) {
            return(0)
          }
          sqrt(n_units * var(pw[[k]] - pw[[best]], na.rm = TRUE))
        },
        numeric(1)
      ),
      3
    )
  }

  if (isTRUE(loo)) {
    out <- out[order(-out$ELPD), ]
  } else if (is.null(fit.measures)) {
    out <- out[order(-out$Marg.Loglik), ]
  }

  rownames(out) <- NULL
  attr(out, "fit_measures_used") <- !is.null(fit.measures)
  attr(out, "baseline_name") <- modnames[1]
  attr(out, "loo_used") <- isTRUE(loo)
  class(out) <- c("compare.inlavaan_internal", class(out))
  out
}

#' @exportS3Method print compare.inlavaan_internal
print.compare.inlavaan_internal <- function(x, ...) {
  cat("Bayesian Model Comparison (INLAvaan)\n")
  if (isTRUE(attr(x, "fit_measures_used"))) {
    cat("Baseline model:", attr(x, "baseline_name"), "\n")
  } else if (isTRUE(attr(x, "loo_used"))) {
    cat("Models ordered by ELPD (Taylor LOO)\n")
  } else {
    cat("Models ordered by marginal log-likelihood\n")
  }
  if (isTRUE(attr(x, "loo_used"))) {
    cat("elpd_diff/se_diff are paired differences vs the best model\n")
  }
  cat("\n")
  print.data.frame(x, row.names = FALSE)
  invisible(x)
}

#' Deviance for INLAvaan Models
#'
#' Extract the (Bayesian) deviance of a fitted \code{INLAvaan} model. Unlike
#' \pkg{lavaan}, which has no \code{deviance()} method, this follows the
#' BUGS/JAGS/Stan convention: "deviance" is \eqn{-2} times the log-likelihood,
#' summarised over the posterior.
#'
#' @param object An object of class [INLAvaan].
#' @param type Character. \code{"mean"} (default) returns the posterior mean
#'   deviance \eqn{\bar{D} = E[-2\log p(y \mid \theta)]}, averaged over
#'   posterior draws. \code{"plugin"} returns the deviance evaluated at the
#'   posterior mean point estimate, \eqn{\hat{D} = -2\log p(y \mid \hat\theta)}
#'   (matching \code{-2 * logLik(object, type = "plugin")}). Both require the
#'   model to have been fitted with \code{test != "none"}.
#' @param ... Currently unused.
#'
#' @returns A length-one numeric of class \code{inlavaan_deviance}, with the
#'   effective number of parameters (\code{pD}) and \code{DIC} attached as
#'   attributes.
#'
#' @details
#' \eqn{\bar{D}} and \eqn{\hat{D}} are the two ingredients of the Deviance
#' Information Criterion, \eqn{DIC = \bar{D} + p_D} where
#' \eqn{p_D = \bar{D} - \hat{D}} is the effective number of parameters. Use
#' [compare()] to compare models by DIC (or Bayes factors, or LOO/WAIC)
#' rather than comparing raw deviances directly.
#'
#' @seealso [logLik()], [compare()]
#'
#' @examples
#' \donttest{
#' HS.model <- "
#'   visual  =~ x1 + x2 + x3
#'   textual =~ x4 + x5 + x6
#'   speed   =~ x7 + x8 + x9
#' "
#' utils::data("HolzingerSwineford1939", package = "lavaan")
#' fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
#'             test = "standard", verbose = FALSE)
#'
#' deviance(fit)
#' attr(deviance(fit), "DIC")
#' }
#'
#' @importFrom stats deviance
#' @exportS3Method deviance INLAvaan
#' @name deviance
deviance.INLAvaan <- function(object, type = c("mean", "plugin"), ...) {
  type <- match.arg(type)
  int <- get_inlavaan_internal(object)
  if (is.null(int$DIC)) {
    cli_abort(
      "{.fn deviance} requires DIC components. Refit with
       {.code test != \"none\"}."
    )
  }
  val <- if (type == "mean") int$DIC$Dbar else int$DIC$Dhat
  structure(
    val,
    pD = int$DIC$pD,
    DIC = int$DIC$dic,
    class = "inlavaan_deviance"
  )
}

#' @exportS3Method print inlavaan_deviance
print.inlavaan_deviance <- function(x, digits = 3L, ...) {
  cat(sprintf("Deviance: %s\n", format(round(unclass(x), digits))))
  cat(col_grey(paste0(
    "# ", symbol$info,
    " pD = ", format(round(attr(x, "pD"), digits)),
    ", DIC = ", format(round(attr(x, "DIC"), digits)), "\n"
  )))
  invisible(x)
}

#' @exportS3Method print inlavaan_internal
print.inlavaan_internal <- function(x, digits = 3, ...) {
  print(round(x$coefficients, digits))
  invisible(x)
}

show_inlavaan <- function(object) {
  inlavaan_version <- object@external$inlavaan_internal$version
  lavaan_version <- object@version[1]
  class(object) <- "lavaan"
  garb <- capture.output(tmp <- show(object))
  tmp$test <- NULL
  garb <- capture.output(tmp)
  garb <- gsub("lavaan", "INLAvaan", garb)
  garb <- gsub(lavaan_version, inlavaan_version, garb, fixed = TRUE)
  cat(paste0(garb, collapse = "\n"))
  cat("\n\n")

  ## ----- Print marginal log-likelihood and ppp -------------------------------
  show_ppp <- if (length(object@Fit@test$ppp) == 0) FALSE else TRUE

  cat(
    "Model Test (User Model):\n\n",
    sprintf("  %-38s", "Marginal log-likelihood"),
    sprintf("  %10.3f", object@Fit@test$mloglik$stat),
    "\n"
  )

  if (isTRUE(show_ppp)) {
    cat(
      sprintf("   %-38s", "PPP (Chi-square)"),
      sprintf("  %10.3f", object@Fit@test$ppp$stat),
      "\n"
    )
  }
}

#' @name show
#' @rdname INLAvaan-class
#' @aliases show,INLAvaan-method
#' @export
setMethod("show", "INLAvaan", show_inlavaan)

# Simple central-difference Jacobian of a gradient function.
# Does 2m evaluations vs numDeriv::jacobian's ~4m (Richardson extrapolation).
fast_jacobian <- function(gr_fn, x, h = 1e-5, cols = seq_along(x)) {
  m <- length(x)
  H <- matrix(0, m, m)
  for (j in cols) {
    xp <- xm <- x
    xp[j] <- x[j] + h
    xm[j] <- x[j] - h
    H[, j] <- (gr_fn(xp) - gr_fn(xm)) / (2 * h)
  }
  H
}

# Central-difference Hessian of a scalar function (no analytic gradient needed).
# Does 4m(m-1)/2 + 2m = 2m^2 evaluations of fn.
fast_hessian <- function(fn, x, h = 1e-5) {
  m <- length(x)
  f0 <- fn(x)
  H <- matrix(0, m, m)

  # Diagonal: (f(x+h*ei) - 2*f(x) + f(x-h*ei)) / h^2
  for (i in seq_len(m)) {
    xp <- xm <- x
    xp[i] <- x[i] + h
    xm[i] <- x[i] - h
    H[i, i] <- (fn(xp) - 2 * f0 + fn(xm)) / (h^2)
  }

  # Off-diagonal (symmetric): use the cross-partial formula
  # (f(x+h*ei+h*ej) - f(x+h*ei-h*ej) - f(x-h*ei+h*ej) + f(x-h*ei-h*ej)) / (4h^2)
  if (m > 1) {
    for (i in seq_len(m - 1)) {
      for (j in (i + 1):m) {
        xpp <- xpm <- xmp <- xmm <- x
        xpp[c(i, j)] <- x[c(i, j)] + h
        xmm[c(i, j)] <- x[c(i, j)] - h
        xpm[i] <- x[i] + h; xpm[j] <- x[j] - h
        xmp[i] <- x[i] - h; xmp[j] <- x[j] + h
        H[i, j] <- H[j, i] <-
          (fn(xpp) - fn(xpm) - fn(xmp) + fn(xmm)) / (4 * h^2)
      }
    }
  }

  H
}

# Central-difference gradient of a scalar function.
# Does 2m evaluations of fn.
fast_grad <- function(fn, x, h = 1e-5) {
  m <- length(x)
  g <- numeric(m)
  for (j in seq_len(m)) {
    xp <- xm <- x
    xp[j] <- x[j] + h
    xm[j] <- x[j] - h
    g[j] <- (fn(xp) - fn(xm)) / (2 * h)
  }
  g
}


# Saturated-means fast path: detect a mean block for which the posterior is
# exactly block-diagonal at the mode. With Gaussian likelihood the cross
# information between means and covariance parameters is n dSigma^{-1}/dtheta
# (ybar - mu_hat), which vanishes when the mean structure is saturated
# (mu_hat = ybar): all intercepts free and unconstrained with normal priors,
# no free latent means, fixed intercepts only for frozen exogenous
# covariates. Along those axes the posterior is exactly Gaussian, so the
# Hessian block is analytic (n Sigma^{-1}[free, free] + prior precision) and
# the skew-normal scans are redundant. Returns NULL when the fast path does
# not apply, else the packed indices, prior precisions, and Sigma row
# positions of the free intercepts.
saturated_mean_idx <- function(pt, lavmodel, lavsamplestats, lavdata,
                               ceq.simple) {
  if (!isTRUE(lavmodel@meanstructure)) return(NULL)
  if (lavmodel@ngroups > 1L || lavdata@nlevels > 1L) return(NULL)
  if (isTRUE(ceq.simple)) return(NULL)
  if (!is.null(attr(pt, "gcp_blocks"))) return(NULL)
  if (any(pt$mat == "alpha" & pt$free > 0)) return(NULL)
  # mu must equal nu EXACTLY: any nonzero alpha (frozen exogenous dummies
  # under fixed.x, or latent means fixed at nonzero values) makes the
  # implied mean depend on free loadings/regressions, so the mean/covariance
  # cross-information no longer vanishes at the mode
  a_vals <- unlist(lavmodel@GLIST[names(lavmodel@GLIST) == "alpha"])
  if (length(a_vals) && any(abs(a_vals) > 1e-12)) return(NULL)
  if (length(unlist(lavsamplestats@x.idx)) > 0L) return(NULL)
  nu_rows <- which(pt$mat == "nu")
  if (!length(nu_rows)) return(NULL)
  free_nu <- nu_rows[pt$free[nu_rows] > 0]
  if (!length(free_nu)) return(NULL)
  fixed_nu <- setdiff(nu_rows, free_nu)
  ovn <- lavdata@ov.names[[1L]]
  xn <- ovn[unlist(lavsamplestats@x.idx[[1L]])]
  if (length(fixed_nu) && !all(pt$lhs[fixed_nu] %in% xn)) return(NULL)
  if (!setequal(pt$lhs[free_nu], setdiff(ovn, xn))) return(NULL)
  if (!all(grepl("^normal\\(", pt$prior[free_nu]))) return(NULL)
  idx <- pt$free[free_nu]
  if (any(duplicated(idx)) || any(idx %in% pt$free[-nu_rows])) return(NULL)
  sds <- vapply(pt$prior[free_nu], function(s) {
    as.numeric(strsplit(gsub("normal\\(|\\)", "", s), ",")[[1]][2])
  }, numeric(1))
  ord <- order(idx)
  list(
    idx = idx[ord],
    prec = (1 / sds^2)[ord],
    sigma_pos = match(pt$lhs[free_nu], ovn)[ord]
  )
}

## ---- Unit tests: fast_grad --------------------------------------------------

test_that("fast_grad matches numDeriv::grad on a quadratic", {
  set.seed(42)
  m <- 6
  Q <- crossprod(matrix(rnorm(m^2), m, m)) + diag(m)
  fn <- function(x) 0.5 * as.numeric(t(x) %*% Q %*% x)

  x0 <- rnorm(m)
  g_fast <- fast_grad(fn, x0)
  g_numD <- numDeriv::grad(fn, x0)
  g_true <- as.numeric(Q %*% x0)

  expect_equal(g_fast, g_numD, tolerance = 1e-6)
  expect_equal(g_fast, g_true, tolerance = 1e-4)
})

test_that("fast_grad works for scalar case", {
  fn <- function(x) 2 * x^3
  g <- fast_grad(fn, 2.0)
  expect_equal(as.numeric(g), 24, tolerance = 1e-4) # true: 6x^2 = 24
})

## ---- Unit tests: fast_jacobian vs numDeriv::jacobian ------------------------

# Test 1: Known quadratic function (exact Hessian is constant)
test_that("fast_jacobian matches numDeriv on a quadratic", {
  # f(x) = 0.5 * x^T A x, so grad(f) = A x and Hessian = A
  set.seed(42)
  m <- 10
  Q <- crossprod(matrix(rnorm(m^2), m, m)) + diag(m) # pos-def
  gr_fn <- function(x) as.numeric(Q %*% x)

  x0 <- rnorm(m)
  H_fast <- fast_jacobian(gr_fn, x0)
  H_numD <- numDeriv::jacobian(gr_fn, x0)

  expect_equal(H_fast, H_numD, tolerance = 1e-4)
  # Both should recover Q exactly (up to FD error)
  expect_equal(H_fast, Q, tolerance = 1e-4)
})

# Test 2: Rosenbrock-like function with non-trivial curvature
test_that("fast_jacobian matches numDeriv on a non-linear function", {
  rosenbrock_grad <- function(x) {
    n <- length(x)
    g <- numeric(n)
    for (i in 1:(n - 1)) {
      g[i] <- g[i] - 2 * (1 - x[i]) - 400 * x[i] * (x[i + 1] - x[i]^2)
      g[i + 1] <- g[i + 1] + 200 * (x[i + 1] - x[i]^2)
    }
    g
  }

  x0 <- rep(1.1, 5) # near the minimum at (1,1,...,1)
  H_fast <- fast_jacobian(rosenbrock_grad, x0)
  H_numD <- numDeriv::jacobian(rosenbrock_grad, x0)

  expect_equal(H_fast, H_numD, tolerance = 1e-3)
})

# Test 3: Scalar gradient (m = 1)
test_that("fast_jacobian works for scalar case", {
  gr_fn <- function(x) 6 * x^2 # gradient of 2x^3
  H_fast <- fast_jacobian(gr_fn, 2.0)
  H_numD <- numDeriv::jacobian(gr_fn, 2.0)

  expect_equal(H_fast, H_numD, tolerance = 1e-4)
  expect_equal(as.numeric(H_fast), 24, tolerance = 1e-4) # true Hessian = 12x = 24
})

# Test 4: Symmetry of the resulting Hessian
test_that("fast_jacobian produces a (near-)symmetric matrix", {
  set.seed(123)
  m <- 8
  Q <- crossprod(matrix(rnorm(m^2), m, m)) + diag(m)
  gr_fn <- function(x) as.numeric(Q %*% x)

  x0 <- rnorm(m)
  H_fast <- fast_jacobian(gr_fn, x0)

  expect_equal(H_fast, t(H_fast), tolerance = 1e-6)
})

## ---- Integration test: Hessian on an actual SEM model -----------------------

dat <- lavaan::PoliticalDemocracy
mod <- "
  ind60 =~ x1 + x2 + x3
  dem60 =~ y1 + y2 + y3 + y4
  dem65 =~ y5 + y6 + y7 + y8
  dem60 ~ ind60
  dem65 ~ ind60 + dem60
"

test_that("fast_jacobian Hessian matches numDeriv on the PoliticalDemocracy SEM", {
  # Fit with debug = TRUE to get internal objects, but we need the gradient

  # closure. Instead, build a lightweight version using INLAvaan internals.
  fit <- asem(mod, dat, test = "none", verbose = FALSE, nsamp = 3)

  # Extract the Sigma_theta (covariance) from the fit
  int <- INLAvaan:::get_inlavaan_internal(fit)
  Sigma_fast <- int$Sigma_theta

  # Now refit with numDeriv to get the "reference" Hessian.
  # We do this by accessing the internal gradient function via the same
  # model setup and using numDeriv::jacobian directly.
  #
  # Reconstruct the joint_lp_grad closure from the fitted objects
  pt <- int$partable
  lavmodel <- int$lavmodel
  lavsamplestats <- int$lavsamplestats
  lavdata <- int$lavdata
  lavcache <- fit@Cache
  lavoptions <- fit@Options
  theta_star <- int$theta_star_novbc # mode before VB correction
  ceq.simple <- lavmodel@ceq.simple.only
  ceq.K <- lavmodel@ceq.simple.K
  prior_cache <- INLAvaan:::prepare_priors_for_optim(pt)

  neg_grad <- function(pars) {
    if (isTRUE(ceq.simple)) {
      pars_unpacked <- as.numeric(ceq.K %*% pars)
      x <- INLAvaan:::pars_to_x(pars_unpacked, pt)
      jcb <- mapply(
        function(f, x) f(x),
        pt$ginv_prime[pt$free > 0],
        pars_unpacked
      )
    } else {
      x <- INLAvaan:::pars_to_x(pars, pt)
      jcb <- mapply(function(f, x) f(x), pt$ginv_prime[pt$free > 0], pars)
    }
    gll <- INLAvaan:::inlav_model_grad(
      x,
      lavmodel,
      lavsamplestats,
      lavdata,
      lavcache
    )
    jcb <- diag(jcb, length(jcb))
    sd1sd2 <- attr(x, "sd1sd2")
    jcb <- jcb * sd1sd2
    jcb_mat <- attr(x, "jcb_mat")
    if (!is.null(jcb_mat)) {
      for (k in seq_len(nrow(jcb_mat))) {
        i <- jcb_mat[k, 1]
        j <- jcb_mat[k, 2]
        jcb[i, j] <- jcb_mat[k, 3]
      }
    }
    gll_th <- as.numeric(jcb %*% gll)
    if (isTRUE(ceq.simple)) {
      gll_th <- as.numeric(gll_th %*% ceq.K)
    }
    glp_th <- INLAvaan:::prior_grad_vectorized(pars, prior_cache)
    -1 * as.numeric(gll_th + glp_th)
  }

  H_fast <- INLAvaan:::fast_jacobian(neg_grad, theta_star)
  H_numD <- numDeriv::jacobian(neg_grad, theta_star)

  # The two Hessians should be very close
  expect_equal(H_fast, H_numD, tolerance = 1e-3)

  # And the resulting covariance matrices should be close
  H_sym_fast <- 0.5 * (H_fast + t(H_fast))
  H_sym_numD <- 0.5 * (H_numD + t(H_numD))
  Sigma_numD <- solve(H_sym_numD)
  Sigma_fast2 <- solve(H_sym_fast)
  expect_equal(Sigma_fast2, Sigma_numD, tolerance = 1e-3)

  # The posterior SDs should agree to ~0.1%
  se_fast <- sqrt(diag(Sigma_fast2))
  se_numD <- sqrt(diag(Sigma_numD))
  expect_equal(se_fast, se_numD, tolerance = 1e-3)
})

## ---- Unit tests: fast_hessian -----------------------------------------------

test_that("fast_hessian matches numDeriv::hessian on a quadratic", {
  set.seed(42)
  m <- 6
  Q <- crossprod(matrix(rnorm(m^2), m, m)) + diag(m)
  fn <- function(x) 0.5 * as.numeric(t(x) %*% Q %*% x)

  x0 <- rnorm(m)
  H_fast <- fast_hessian(fn, x0)
  H_numD <- numDeriv::hessian(fn, x0)

  expect_equal(H_fast, H_numD, tolerance = 1e-3)
  # Both should recover the symmetric part of Q
  expect_equal(H_fast, 0.5 * (Q + t(Q)), tolerance = 1e-3)
})

test_that("fast_hessian matches numDeriv on Rosenbrock", {
  rosenbrock <- function(x) {
    sum(100 * (x[-1] - x[-length(x)]^2)^2 + (1 - x[-length(x)])^2)
  }
  x0 <- rep(1.1, 4)
  H_fast <- fast_hessian(rosenbrock, x0)
  H_numD <- numDeriv::hessian(rosenbrock, x0)

  expect_equal(H_fast, H_numD, tolerance = 1e-2)
})

test_that("fast_hessian is symmetric", {
  fn <- function(x) sin(x[1]) * cos(x[2]) + x[3]^3
  H <- fast_hessian(fn, c(1, 2, 3))
  expect_equal(H, t(H), tolerance = 1e-10)
})

test_that("fast_hessian works for scalar case", {
  fn <- function(x) 2 * x^3
  H <- fast_hessian(fn, 2.0)
  expect_equal(as.numeric(H), 24, tolerance = 1e-3) # true: 12x = 24
})

# --- Bayesian fit index helpers -----------------------------------------------

# Saturated log-likelihood (constant for ML; sum across groups)
# Under FIML, uses the pattern-based formula to stay on the same scale as
# inlav_model_loglik (which delegates to lavaan:::lav_model_loglik).
compute_loglik_sat <- function(lavsamplestats, lavdata) {
  ngroups <- lavdata@ngroups
  logl_sat <- 0
  for (g in seq_len(ngroups)) {
    if (lavsamplestats@missing.flag) {
      # nocov start
      # positional: argument names changed in lavaan >= 0.7
      # (Yp/Mu/Sigma/x.idx/x.mean/x.cov -> yp/mu/sigma_1/x_idx/x_mean/x_cov)
      logl_sat <- logl_sat +
        lavaan___lav_mvnorm_missing_loglik_samplestats(
          lavsamplestats@missing[[g]], # Yp
          lavsamplestats@mean[[g]], # Mu
          lavsamplestats@cov[[g]], # Sigma
          lavsamplestats@x.idx[[g]], # x.idx
          lavsamplestats@mean.x[[g]], # x.mean
          lavsamplestats@cov.x[[g]] # x.cov
        )
    } else {
      # nocov end
      # positional: argument names changed in lavaan >= 0.7
      # (sample.mean/sample.cov/sample.nobs/Mu/Sigma/x.idx/x.mean/x.cov ->
      #  sample_mean/sample_cov/sample_nobs/mu/sigma_1/x_idx/x_mean/x_cov)
      logl_sat <- logl_sat +
        lavaan___lav_mvnorm_loglik_samplestats(
          lavsamplestats@mean[[g]], # sample.mean
          lavsamplestats@cov[[g]], # sample.cov
          lavsamplestats@nobs[[g]], # sample.nobs
          lavsamplestats@mean[[g]], # Mu
          lavsamplestats@cov[[g]], # Sigma
          lavsamplestats@x.idx[[g]], # x.idx
          lavsamplestats@mean.x[[g]], # x.mean
          lavsamplestats@cov.x[[g]] # x.cov
        )
    }
  }
  logl_sat
}

# Per-sample deviance chi-square:  chisq_s = 2 * (loglik_sat - loglik(x_s))
# This equals N * F_ML(x_s).
compute_chisq_dev <- function(
  x_samp,
  lavmodel,
  lavsamplestats,
  lavdata,
  lavoptions,
  lavcache
) {
  loglik_sat <- compute_loglik_sat(lavsamplestats, lavdata)
  vapply(
    seq_len(nrow(x_samp)),
    function(i) {
      ll_i <- inlav_model_loglik(
        x_samp[i, ],
        lavmodel,
        lavsamplestats,
        lavdata,
        lavoptions,
        lavcache
      )
      2 * (loglik_sat - ll_i)
    },
    numeric(1)
  )
}

# Number of sample statistics:  sum_g [ p_g(p_g+1)/2 + meanstructure * p_g ]
compute_p_samplestats <- function(nvar, meanstructure) {
  sum(vapply(
    nvar,
    function(nv) {
      nMom <- nv * (nv + 1) / 2
      if (isTRUE(meanstructure)) {
        nMom <- nMom + nv # nocov
      }
      nMom
    },
    numeric(1)
  ))
}

# Absolute fit indices (vectorised over posterior samples) ---------------------
compute_BRMSEA <- function(nonc, df, N, Ngr) {
  sqrt(nonc / (df * N)) * sqrt(Ngr)
}
compute_BGammaHat <- function(nonc, nvar_total, N) {
  nvar_total / (nvar_total + 2 * nonc / N)
}
compute_adjBGammaHat <- function(BGammaHat, p, df) {
  1 - (p / df) * (1 - BGammaHat)
}
compute_BMc <- function(nonc, N) exp(-0.5 * nonc / N)

# Incremental fit indices (vectorised) ----------------------------------------
compute_BCFI <- function(nonc, nonc_null) 1 - nonc / nonc_null
compute_BTLI <- function(adj_dev, df, adj_dev_null, df_null) {
  tli_null <- adj_dev_null / df_null
  (tli_null - adj_dev / df) / (tli_null - 1)
}
compute_BNFI <- function(adj_dev, adj_dev_null) {
  (adj_dev_null - adj_dev) / adj_dev_null
}

# Reconstruct lavoptions suitable for inlav_model_loglik() from the INLAvaan
# object (whose @Options$estimator was changed to "Bayes").
reconstruct_lavoptions <- function(object) {
  opts <- object@Options
  opts$estimator <- object@external$inlavaan_internal$lavmodel@estimator
  opts
}

# Compute rescaled chi-square, adjusted deviance, df, and noncentrality for
# a single model under the chosen rescaling method.
#
# Returns list(chisq, adj_dev, df, nonc, N_adj, pD) where chisq/adj_dev/nonc
# are per-sample vectors and df/N_adj/pD are scalars.
compute_rescaled_quantities <- function(
  object,
  x_samp,
  lavmodel,
  lavsamplestats,
  lavdata,
  lavoptions,
  lavcache,
  p,
  rescale
) {
  N <- lavsamplestats@ntotal
  Ngr <- lavdata@ngroups
  npar <- object@Fit@npar

  chisq_dev <- compute_chisq_dev(
    x_samp,
    lavmodel,
    lavsamplestats,
    lavdata,
    lavoptions,
    lavcache
  )

  if (rescale == "devM") {
    pD <- object@external$inlavaan_internal$DIC$pD
    if (is.null(pD) || pD <= 0 || pD >= p) {
      pD <- npar
    } # nocov
    adj_dev <- chisq_dev - pD # obs - reps
    df <- p - pD
    N_adj <- N
  } else {
    # MCMC: use classical chi-square = (N-1)/N * deviance
    pD <- npar
    chisq_dev <- (N - 1) / N * chisq_dev
    adj_dev <- chisq_dev # reps = 0
    df <- p - npar
    N_adj <- N - Ngr # EQS-style: Min1 = TRUE
  }

  nonc <- pmax(adj_dev - df, 0)
  list(
    chisq = chisq_dev,
    adj_dev = adj_dev,
    df = df,
    nonc = nonc,
    N_adj = N_adj,
    pD = pD
  )
}

# ---------------------------------------------------------------------------
# bfit_indices: compute per-sample Bayesian fit index vectors and return
# an S3 object of class "bfit_indices" with a summary() and print() method.
# ---------------------------------------------------------------------------

#' Bayesian Fit Indices
#'
#' Compute posterior distributions of Bayesian fit indices for an INLAvaan
#' model, analogous to [blavaan::blavFitIndices()].
#'
#' @param object An object of class [INLAvaan].
#' @param baseline.model An optional [INLAvaan] object representing the
#'   baseline (null) model. Required for incremental fit indices (BCFI, BTLI,
#'   BNFI).
#' @param rescale Character string controlling how the Bayesian chi-square
#'   is rescaled. `"devM"` (default) subtracts pD from the deviance at each
#'   sample. `"MCMC"` uses the classical chi-square and classical df at each
#'   sample.
#' @param nsamp Number of posterior samples to draw. Defaults to the value
#'   used when fitting the model.
#' @param samp_copula Logical. When `TRUE` (default), posterior samples are
#'   drawn using the copula method with the fitted marginals. When `FALSE`,
#'   samples are drawn from the Gaussian (Laplace) approximation.
#' @param x An object of class `bfit_indices` (for `print`).
#' @param ... Additional arguments passed to methods.
#'
#' @returns An S3 object of class `"bfit_indices"` containing:
#' \describe{
#'   \item{`indices`}{Named list of numeric vectors (one per posterior sample)
#'     for each computed fit index.}
#'   \item{`details`}{List with `chisq` (per-sample deviance), `df`, `pD`,
#'     `rescale`, and `nsamp`.}
#' }
#' Use [summary()] to obtain a table of posterior summaries (Mean, SD,
#' quantiles, Mode) for each index.
#'
#' @seealso [lavaan::fitMeasures()], [blavaan::blavFitIndices()],
#'   [fitmeasures()], [compare()]
#'
#' @examples
#' \donttest{
#' HS.model <- "
#'   visual  =~ x1 + x2 + x3
#'   textual =~ x4 + x5 + x6
#'   speed   =~ x7 + x8 + x9
#' "
#' utils::data("HolzingerSwineford1939", package = "lavaan")
#' fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
#'             verbose = FALSE)
#'
#' # Absolute fit indices
#' bf <- bfit_indices(fit)
#' bf
#' summary(bf)
#' }
#'
#' @export
bfit_indices <- function(
  object,
  baseline.model = NULL,
  rescale = c("devM", "MCMC"),
  nsamp = NULL,
  samp_copula = TRUE
) {
  rescale <- match.arg(rescale)
  if (!is(object, "INLAvaan")) {
    cli_abort("{.arg object} must be an {.cls INLAvaan} object.")
  }

  int <- object@external$inlavaan_internal
  lavmodel <- int$lavmodel
  lavsamplestats <- int$lavsamplestats
  lavdata <- int$lavdata

  nsamp <- nsamp %||% int$nsamp %||% 500L
  method <- if (isTRUE(samp_copula)) int$marginal_method else "sampling"
  samp <- sample_params(
    theta_star = int$theta_star,
    Sigma_theta = int$Sigma_theta,
    method = method,
    approx_data = int$approx_data,
    pt = int$partable,
    lavmodel = lavmodel,
    nsamp = nsamp,
    R_star = int$R_star
  )
  x_samp <- samp$x_samp

  if (lavmodel@estimator != "ML") {
    # nocov
    cli_abort("Bayesian fit indices are only supported for the ML estimator.")
  }
  if (rescale == "devM" && is.null(int$DIC)) {
    cli_abort(
      "DIC not available. Refit with {.code test != \"none\"}, or use {.code rescale = \"MCMC\"}."
    )
  }

  lavoptions <- reconstruct_lavoptions(object)
  lavcache <- object@Cache
  N <- lavsamplestats@ntotal
  Ngr <- lavdata@ngroups
  nvar <- lavmodel@nvar
  ms <- isTRUE(lavoptions$meanstructure)
  p <- compute_p_samplestats(nvar, ms)

  rq <- compute_rescaled_quantities(
    object,
    x_samp,
    lavmodel,
    lavsamplestats,
    lavdata,
    lavoptions,
    lavcache,
    p,
    rescale
  )

  indices <- list()

  if (rq$df > 0) {
    indices$BRMSEA <- compute_BRMSEA(rq$nonc, rq$df, rq$N_adj, Ngr)
    bgh <- compute_BGammaHat(rq$nonc, sum(nvar), rq$N_adj)
    indices$BGammaHat <- bgh
    indices$adjBGammaHat <- compute_adjBGammaHat(bgh, p, rq$df)
    indices$BMc <- compute_BMc(rq$nonc, rq$N_adj)
  } # else: df == 0, no absolute fit indices (nocov – saturated model)

  # Incremental indices
  if (!is.null(baseline.model)) {
    if (!is(baseline.model, "INLAvaan")) {
      cli_abort("{.arg baseline.model} must be an {.cls INLAvaan} object.")
    }
    bint <- baseline.model@external$inlavaan_internal

    bmethod <- if (isTRUE(samp_copula)) bint$marginal_method else "sampling"
    samp_null <- sample_params(
      theta_star = bint$theta_star,
      Sigma_theta = bint$Sigma_theta,
      method = bmethod,
      approx_data = bint$approx_data,
      pt = bint$partable,
      lavmodel = bint$lavmodel,
      nsamp = nsamp,
      R_star = bint$R_star
    )
    x_samp_null <- samp_null$x_samp
    n_use <- min(nrow(x_samp), nrow(x_samp_null))

    rq_null <- compute_rescaled_quantities(
      baseline.model,
      x_samp_null[seq_len(n_use), , drop = FALSE],
      bint$lavmodel,
      bint$lavsamplestats,
      bint$lavdata,
      reconstruct_lavoptions(baseline.model),
      baseline.model@Cache,
      p,
      rescale
    )

    adj_dev_use <- rq$adj_dev[seq_len(n_use)]
    nonc_use <- rq$nonc[seq_len(n_use)]

    indices$BCFI <- compute_BCFI(nonc_use, rq_null$nonc)
    indices$BTLI <- compute_BTLI(
      adj_dev_use,
      rq$df,
      rq_null$adj_dev,
      rq_null$df
    )
    indices$BNFI <- compute_BNFI(adj_dev_use, rq_null$adj_dev)
  }

  structure(
    list(
      indices = indices,
      details = list(
        chisq = rq$chisq,
        df = rq$df,
        pD = rq$pD,
        rescale = rescale,
        nsamp = nrow(x_samp)
      )
    ),
    class = "bfit_indices"
  )
}

#' @rdname bfit_indices
#' @exportS3Method summary bfit_indices
#' @usage \method{summary}{bfit_indices}(object, ...)
summary.bfit_indices <- function(object, ...) {
  summ_one <- function(x) {
    x <- x[is.finite(x)]
    if (length(x) < 3) {
      # nocov
      return(rep(NA_real_, 8))
    }
    dens <- stats::density(x)
    c(
      Mean = mean(x),
      SD = stats::sd(x),
      `2.5%` = stats::quantile(x, 0.025, names = FALSE),
      `25%` = stats::quantile(x, 0.250, names = FALSE),
      `50%` = stats::quantile(x, 0.500, names = FALSE),
      `75%` = stats::quantile(x, 0.750, names = FALSE),
      `97.5%` = stats::quantile(x, 0.975, names = FALSE),
      Mode = dens$x[which.max(dens$y)]
    )
  }
  tab <- as.data.frame(do.call(rbind, lapply(object$indices, summ_one)))
  attr(tab, "header") <- paste0(
    "Posterior summary of ",
    object$details$rescale,
    "-based Bayesian fit indices (nsamp = ",
    object$details$nsamp,
    "):"
  )
  class(tab) <- c("lavaan.data.frame", "data.frame")
  tab
}

#' @rdname bfit_indices
#' @exportS3Method print bfit_indices
print.bfit_indices <- function(x, ...) {
  tab <- summary(x)
  hdr <- attr(tab, "header")
  if (!is.null(hdr)) {
    cat(hdr, "\n\n")
  }
  eap <- vapply(x$indices, mean, numeric(1), na.rm = TRUE)
  print(round(eap, 3))
  invisible(x)
}

# --- Main fitMeasures function ------------------------------------------------

inlav_fit_measures <- function(
  object,
  fit.measures = "all",
  baseline.model = NULL,
  ...
) {
  dots <- list(...)
  rescale <- match.arg(dots$rescale %||% "devM", c("devM", "MCMC"))

  # Has the model converged?
  if (object@Fit@npar > 0L & !object@optim$converged) {
    # nocov
    cli_alert_warning("Optimiser did not converge.") # nocov
  } # nocov

  out <- vector("numeric")
  out["npar"] <- object@Fit@npar
  out["margloglik"] <- object@external$inlavaan_internal$mloglik

  # If test != "none"
  if (length(object@Fit@test$ppp) > 0) {
    out["ppp"] <- object@external$inlavaan_internal$ppp
    out["dic"] <- object@external$inlavaan_internal$DIC$dic
    out["p_dic"] <- object@external$inlavaan_internal$DIC$pD
  }

  # Validate baseline.model early (before tryCatch)
  if (!is.null(baseline.model) && !is(baseline.model, "INLAvaan")) {
    cli_abort("{.arg baseline.model} must be an {.cls INLAvaan} object.")
  }

  # Bayesian fit indices (BRMSEA, BGammaHat, etc.)
  bfi <- tryCatch(
    bfit_indices(object, baseline.model, rescale),
    error = function(e) NULL
  )
  if (!is.null(bfi)) {
    for (nm in names(bfi$indices)) {
      out[nm] <- mean(bfi$indices[[nm]], na.rm = TRUE)
    }
  }

  # LOO measures: free when stored with the fit (test = "loo" or add_loo());
  # otherwise computed on demand, and only when requested by name -- the LOO
  # computation is fresh work and cannot be cached into `object` (S4 copy
  # semantics), so it never silently inflates a bare fitMeasures() call
  loo_measures <- c("elpd_loo", "se_loo", "p_loo", "looic")
  res_loo <- object@external$inlavaan_internal$loo
  if (
    is.null(res_loo) &&
      !identical(fit.measures, "all") &&
      any(loo_measures %in% fit.measures)
  ) {
    res_loo <- tryCatch(loo(object), error = function(e) NULL)
  }
  if (!is.null(res_loo)) {
    out["elpd_loo"] <- res_loo$estimates["elpd_loo", "Estimate"]
    out["p_loo"] <- res_loo$estimates["p_loo", "Estimate"]
    out["looic"] <- res_loo$estimates["looic", "Estimate"]
    out["se_loo"] <- res_loo$estimates["looic", "SE"]
  }

  # WAIC: free when stored with the fit; otherwise sampling-based, computed
  # on demand when requested by name
  waic_measures <- c("elpd_waic", "se_waic", "p_waic", "waic")
  res_waic <- object@external$inlavaan_internal$waic
  if (
    is.null(res_waic) &&
      !identical(fit.measures, "all") &&
      any(waic_measures %in% fit.measures)
  ) {
    res_waic <- tryCatch(waic(object), error = function(e) NULL)
  }
  if (!is.null(res_waic)) {
    out["elpd_waic"] <- res_waic$estimates["elpd_waic", "Estimate"]
    out["p_waic"] <- res_waic$estimates["p_waic", "Estimate"]
    out["waic"] <- res_waic$estimates["waic", "Estimate"]
    out["se_waic"] <- res_waic$estimates["waic", "SE"]
  }

  # Filter if specific measures requested
  if (!identical(fit.measures, "all")) {
    idx <- which(names(out) %in% fit.measures)
    if (length(idx) == 0L) {
      cli_abort("No matching fit measures found.")
    }
    out <- out[idx]
  }

  class(out) <- c("fitmeasures.inlavaan_internal", "numeric")
  out
}

#' @exportS3Method print fitmeasures.inlavaan_internal
print.fitmeasures.inlavaan_internal <- function(x, ...) {
  nm <- names(x)

  # Apply conditional formatting
  formatted_values <- sapply(seq_along(x), function(i) {
    val <- x[i]
    name <- nm[i]

    if (name == "npar") {
      return(as.character(as.integer(round(val))))
    } else if (startsWith(name, "grad_")) {
      # nocov start
      # Use formatC to force scientific and maintain 3 significant digits
      return(formatC(val, digits = 2, format = "e"))
    } else {
      # nocov end
      # Round to 3 decimal places
      return(formatC(val, digits = 3, format = "f", drop0trailing = FALSE))
    }
  })

  # Set names back onto the formatted character vector
  names(formatted_values) <- nm

  # Print using the standard named vector style without quotes
  print(formatted_values, quote = FALSE, right = TRUE)

  invisible(x)
}

#' Fit Measures for a Latent Variable Model estimated using INLA
#'
#' @param object An object of class [INLAvaan].
#' @param fit.measures If `"all"`, all fit measures available will be returned. If
#'   only a single or a few fit measures are specified by name, only those are
#'   computed and returned. The LOO measures `"elpd_loo"`, `"se_loo"`,
#'   `"p_loo"` and `"looic"` (see [loo()]) are included in `"all"` only when
#'   a LOO result is stored with the fit (`test = "loo"` in [inlavaan()] or
#'   [add_loo()]); otherwise they are computed on demand when requested by
#'   name, and recomputed on every call -- store the result with
#'   `fit <- add_loo(fit)` (or call [loo()] directly) for repeated access.
#' @param baseline.model An optional [INLAvaan] object representing the
#'   baseline (null) model. Required for incremental fit indices (BCFI, BTLI,
#'   BNFI). Must have been fitted with `test != "none"`.
#' @param h1.model Ignored (included for compatibility with the lavaan
#'   generic).
#' @param fm.args Ignored (included for compatibility with the lavaan
#'   generic).
#' @param output Ignored (included for compatibility with the lavaan
#'   generic).
#' @param ... Additional arguments. Currently supports:
#' \describe{
#'   \item{`rescale`}{Character string controlling how the Bayesian chi-square
#'     is computed, following `blavaan::blavFitIndices()`. Options are `"devM"`
#'     (default) which uses the deviance rescaled by `pD` from DIC, or `"MCMC"`
#'     which uses the classical chi-square (`(N-1) * F_ML`) and classical
#'     degrees of freedom (`p - npar`) at each posterior sample.}
#' }
#'
#' @returns A named numeric vector of fit measures.
#'
#' @seealso [bfit_indices()], [compare()], [diagnostics()]
#'
#' @examples
#' \donttest{
#' HS.model <- "
#'   visual  =~ x1 + x2 + x3
#'   textual =~ x4 + x5 + x6
#'   speed   =~ x7 + x8 + x9
#' "
#' utils::data("HolzingerSwineford1939", package = "lavaan")
#' fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
#'             verbose = FALSE)
#'
#' # All available fit measures
#' fitMeasures(fit)
#'
#' # Specific measures
#' fitMeasures(fit, c("npar", "DIC", "pD", "ppp"))
#' }
#'
#' @importMethodsFrom lavaan fitmeasures fitMeasures
#' @name fitmeasures
#' @rdname fitmeasures
#' @aliases fitMeasures,INLAvaan-method
#' @rawNamespace exportMethods(fitMeasures)
NULL

#' @name fitmeasures
#' @rdname fitmeasures
#' @aliases fitmeasures,INLAvaan-method
#' @rawNamespace exportMethods(fitmeasures)
NULL

# lavaan >= 0.7 renamed the fitMeasures()/fitmeasures() generics' arguments
# (fit.measures/baseline.model -> fit_measures/baseline_model). Unlike a
# plain rename, registering setMethod() against the WRONG argument names
# doesn't just fail loudly: S4 dispatch silently drops any value bound to a
# generic-level formal the method doesn't also declare (verified empirically
# -- positional and named calls both default silently), and if the method
# was compiled against a lavaan present at a *different* time than the one
# later loaded, dispatch can also error outright ("could not find symbol").
# So -- like the lavaan___-prefixed internals in lavaan-unexported.R -- these
# methods must be (re)built from whichever lavaan generic is actually active
# in the current session, in .onLoad(), not baked in at build/install time.
# The resolved installed spellings come from the shared lavaan_argnames map
# (see lavaan-argnames.R, resolved earlier in the same .onLoad()), the one
# mechanism INLAvaan uses for every renamed lavaan argument.
#
# INLAvaan's own documented/stable parameter names are fit.measures and
# baseline.model (used internally, e.g. by compare()), regardless of which
# lavaan is loaded. Under lavaan >= 0.7 those don't match the active
# generic's own (renamed) formals, so a caller using our documented names
# falls through into "..." unmatched rather than being dropped; recover it
# from there instead of re-forwarding "..." verbatim (which would otherwise
# also duplicate whatever the generic's own formal already bound).
register_fitmeasures_methods <- function(ns) {
  fm_map <- lavaan_argnames[["fitMeasures"]]
  fit_measures_arg <- fm_map[["fit.measures"]]
  baseline_model_arg <- fm_map[["baseline.model"]]

  for (generic_name in c("fitMeasures", "fitmeasures")) {
    gf <- formals(methods::getGeneric(generic_name))

    method_fn <- function(object, ...) NULL
    formals(method_fn) <- gf
    body(method_fn) <- bquote({
      dots <- list(...)
      fit.measures <- .(as.name(fit_measures_arg))
      baseline.model <- .(as.name(baseline_model_arg))
      if ("fit.measures" %in% names(dots)) {
        fit.measures <- dots[["fit.measures"]]
        dots[["fit.measures"]] <- NULL
      }
      if ("baseline.model" %in% names(dots)) {
        baseline.model <- dots[["baseline.model"]]
        dots[["baseline.model"]] <- NULL
      }
      do.call(
        inlav_fit_measures,
        c(
          list(
            object,
            fit.measures = fit.measures,
            baseline.model = baseline.model
          ),
          dots
        )
      )
    })
    environment(method_fn) <- ns
    methods::setMethod(generic_name, "INLAvaan", method_fn, where = ns)
  }
}

dat <- lavaan::HolzingerSwineford1939
mod <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"

# Fit shared models once (fast defaults)
fit_notest <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none",
                   vb_correction = FALSE, marginal_method = "marggaus")
fit_test   <- acfa(mod, dat, verbose = FALSE, nsamp = 10,
                   vb_correction = FALSE, marginal_method = "marggaus")

null_mod <- "
  x1 ~~ x1
  x2 ~~ x2
  x3 ~~ x3
  x4 ~~ x4
  x5 ~~ x5
  x6 ~~ x6
"
fit_null <- acfa(null_mod, dat, verbose = FALSE, nsamp = 3,
                 vb_correction = FALSE, marginal_method = "marggaus")

test_that("Basic fitMeasures returns expected names", {
  fm <- fitMeasures(fit_notest)
  expect_s3_class(fm, "fitmeasures.inlavaan_internal")
  expect_true("npar" %in% names(fm))
  expect_true("margloglik" %in% names(fm))
})

test_that("Bayesian absolute fit indices are computed with test != 'none'", {
  fm <- fitMeasures(fit_test)
  abs_names <- c("BRMSEA", "BGammaHat", "adjBGammaHat", "BMc")
  for (nm in abs_names) {
    expect_true(nm %in% names(fm), info = paste(nm, "missing"))
  }
  expect_true(fm["BRMSEA"] >= 0)
  expect_true(fm["BGammaHat"] > 0 && fm["BGammaHat"] <= 1)
  expect_true(fm["BMc"] > 0 && fm["BMc"] <= 1)
})

test_that("Incremental indices require baseline.model", {
  fm <- fitMeasures(fit_test)
  inc_names <- c("BCFI", "BTLI", "BNFI")
  for (nm in inc_names) {
    expect_false(nm %in% names(fm), info = paste(nm, "should be absent"))
  }
})

test_that("Incremental indices computed with baseline.model", {
  fm <- fitMeasures(fit_test, baseline.model = fit_null)
  inc_names <- c("BCFI", "BTLI", "BNFI")
  for (nm in inc_names) {
    expect_true(nm %in% names(fm), info = paste(nm, "missing"))
  }
})

test_that("baseline.model must be INLAvaan", {
  expect_error(fitMeasures(fit_test, baseline.model = "not_a_model"))
})

test_that("Selecting specific fit measures works", {
  fm <- fitMeasures(fit_test, fit.measures = c("BRMSEA", "BMc"))
  expect_equal(length(fm), 2)
  expect_named(fm, c("BRMSEA", "BMc"))
})

test_that("rescale = 'MCMC' produces fit indices", {
  fm <- fitMeasures(fit_test, rescale = "MCMC")
  abs_names <- c("BRMSEA", "BGammaHat", "adjBGammaHat", "BMc")
  for (nm in abs_names) {
    expect_true(nm %in% names(fm), info = paste(nm, "missing"))
  }
  expect_true(fm["BRMSEA"] >= 0)
})

test_that("rescale = 'devM' and 'MCMC' give different results", {
  fm_devm <- fitMeasures(fit_test, fit.measures = "BRMSEA", rescale = "devM")
  fm_mcmc <- fitMeasures(fit_test, fit.measures = "BRMSEA", rescale = "MCMC")
  expect_false(identical(fm_devm, fm_mcmc))
})

# --- bfit_indices S3 class tests ---

test_that("bfit_indices returns correct S3 class", {
  bfi <- bfit_indices(fit_test)
  expect_s3_class(bfi, "bfit_indices")
  expect_true(is.list(bfi$indices))
  expect_true(is.list(bfi$details))
})

test_that("bfit_indices stores per-sample vectors", {
  bfi <- bfit_indices(fit_test)
  for (nm in names(bfi$indices)) {
    expect_true(is.numeric(bfi$indices[[nm]]))
    expect_equal(length(bfi$indices[[nm]]), 10)
  }
})

test_that("summary.bfit_indices returns data.frame with correct columns", {
  bfi <- bfit_indices(fit_test)
  tab <- summary(bfi)
  expect_s3_class(tab, "data.frame")
  expect_equal(colnames(tab), c("Mean", "SD", "2.5%", "25%", "50%", "75%", "97.5%", "Mode"))
  expect_equal(nrow(tab), length(bfi$indices))
  expect_equal(rownames(tab), names(bfi$indices))
})

test_that("print.bfit_indices runs without error", {
  bfi <- bfit_indices(fit_test)
  expect_output(print(bfi))
})

test_that("bfit_indices details has expected fields", {
  bfi <- bfit_indices(fit_test)
  expect_true(all(c("chisq", "df", "pD", "rescale", "nsamp") %in%
                    names(bfi$details)))
  expect_equal(bfi$details$nsamp, 10)
  expect_equal(bfi$details$rescale, "devM")
})

test_that("print.fitmeasures.inlavaan_internal formats output", {
  fm <- fitMeasures(fit_test)
  expect_output(print(fm), "npar")
  expect_output(print(fm), "margloglik")
})

test_that("fitMeasures errors on unrecognised measure names", {
  expect_error(fitMeasures(fit_test, fit.measures = "nonexistent_measure"))
})

test_that("bfit_indices errors for non-INLAvaan object", {
  expect_error(bfit_indices("not_a_model"), class = "error")
})

test_that("bfit_indices errors when DIC not available and rescale = devM", {
  expect_error(bfit_indices(fit_notest, rescale = "devM"), "DIC not available")
})

test_that("bfit_indices errors for non-INLAvaan baseline.model", {
  expect_error(bfit_indices(fit_test, baseline.model = "not_a_model"), class = "error")
})

#' Widely Applicable Information Criterion for INLAvaan Models
#'
#' Computes the WAIC of a fitted [INLAvaan] model from unit log-likelihoods
#' evaluated over posterior draws. Single-level models are scored per
#' subject; two-level models are scored per cluster by default, matching the
#' units used by [loo()]. For a two-level model `type = "loso"` instead
#' scores the *conditional* (leave-one-unit-out) WAIC; see Details.
#'
#' @details
#' For each posterior draw \eqn{\theta_s} (drawn with the same copula
#' sampler used for the fit's posterior summaries) and unit \eqn{u}, the
#' log-likelihood \eqn{\log p(y_u \mid \theta_s)} is evaluated; then
#' \eqn{\mathrm{lpd}_u = \log \tfrac{1}{S}\sum_s p(y_u \mid \theta_s)},
#' \eqn{p_{\mathrm{waic},u} = \mathrm{var}_s \log p(y_u \mid \theta_s)}, and
#' \eqn{\mathrm{elpd}_{\mathrm{waic}} = \sum_u (\mathrm{lpd}_u -
#' p_{\mathrm{waic},u})} with \eqn{\mathrm{WAIC} = -2\,
#' \mathrm{elpd}_{\mathrm{waic}}}. Unlike [loo()], this is a sampling-based
#' estimate: results vary with the random draws, and units with
#' \eqn{p_{\mathrm{waic},u} > 0.4} trigger a reliability warning (also
#' annotated when printing). The same model restrictions as [loo()] apply,
#' and so does the flavour rule: fits with `fixed.x = TRUE` are scored
#' conditionally on the exogenous covariates, fits with `fixed.x = FALSE`
#' jointly (see [loo()]).
#'
#' **Marginal vs conditional WAIC (two-level models).** The default
#' per-cluster scoring is the *marginal* WAIC, which corresponds to
#' leave-one-cluster-out cross-validation -- prediction for a *new* cluster.
#' Setting `type = "loso"` scores the *conditional* WAIC, corresponding to
#' leave-one-unit-out -- prediction for a new observation within an
#' *observed* cluster (each row contributes the conditional density of its
#' observed entries given the rest of its cluster). The two answer different
#' questions and are easily conflated (Merkle, Furr & Rabe-Hesketh, 2019);
#' the per-cluster marginal is the usual model-comparison target, so it is
#' the default, and `type = "loso"` warns. This matches `loo(type = "loso")`
#' -- the two compute the same estimand by sampling and by Taylor expansion.
#'
#' Under the default `test = "standard"`, [inlavaan()] computes the WAIC at
#' fit time by reusing the posterior draws the fit already produced (when
#' the model is supported and `nsamp >= 100`), and
#' stores it with the fit: `waic(fit)` then returns the stored result when
#' called with default arguments, and [fitmeasures()] reports `waic`,
#' `p_waic`, `se_waic` as part of `"all"` for free. If the `loo` package is
#' attached it masks this generic, but dispatch on INLAvaan objects
#' continues to work.
#'
#' @param x A fitted [INLAvaan] object (or its `inlavaan_internal` list).
#' @param type Unit type: `"auto"` (default) resolves to per-subject for
#'   single-level models and per-cluster (marginal WAIC) for two-level
#'   models. `"loso"` on a two-level model scores the conditional
#'   (leave-one-unit-out) WAIC instead (with a warning; see Details);
#'   `"loco"` cannot be forced on a model without clusters.
#' @param units Optional integer vector of unit indices to score; defaults
#'   to all units.
#' @param nsamp Number of posterior draws. Defaults to the `nsamp` used when
#'   fitting the model.
#' @param cores Number of cores for evaluating draws. The default `NULL`
#'   runs serially; parallelism must be requested explicitly.
#' @param verbose Logical; print progress (default `FALSE`).
#' @param ... Not used.
#'
#' @returns An object of class `inlavaan_waic`: a list with `per_unit`
#'   (pointwise `lpd`, `p_waic`, `elpd_waic`, with the same `unit`/`group`
#'   identification as [loo()]), `estimates` (matrix with rows
#'   `elpd_waic`, `p_waic`, `waic` and columns `Estimate`, `SE`), `type`,
#'   `n_units`, `n_groups`, and `nsamp`.
#'
#' @seealso [loo()], [fitmeasures()]
#'
#' @examples
#' \donttest{
#' HS.model <- "
#'   visual  =~ x1 + x2 + x3
#'   textual =~ x4 + x5 + x6
#'   speed   =~ x7 + x8 + x9
#' "
#' utils::data("HolzingerSwineford1939", package = "lavaan")
#' fit <- acfa(HS.model, HolzingerSwineford1939, meanstructure = TRUE)
#' waic(fit)
#' }
#'
#' @export
waic <- function(x, ...) {
  UseMethod("waic")
}

#' @rdname waic
#' @exportS3Method waic INLAvaan
waic.INLAvaan <- function(
  x,
  type = c("auto", "loso", "loco"),
  units = NULL,
  nsamp = NULL,
  cores = NULL,
  verbose = FALSE,
  ...
) {
  waic.inlavaan_internal(
    x@external$inlavaan_internal,
    type = type,
    units = units,
    nsamp = nsamp,
    cores = cores,
    verbose = verbose
  )
}

#' @rdname waic
#' @exportS3Method waic inlavaan_internal
waic.inlavaan_internal <- function(
  x,
  type = c("auto", "loso", "loco"),
  units = NULL,
  nsamp = NULL,
  cores = NULL,
  verbose = FALSE,
  ...
) {
  type <- match.arg(type)
  # Reuse the result stored at fit time when no argument deviates from the
  # defaults
  if (type == "auto" && is.null(units) && is.null(nsamp) && !is.null(x$waic)) {
    if (isTRUE(verbose)) {
      cli_alert_info("Returning the WAIC stored with the fit.")
    }
    return(x$waic)
  }
  inlav_waic(
    int = x,
    type = type,
    units = units,
    nsamp = nsamp,
    eff_cores = resolve_loo_cores(cores),
    verbose = verbose
  )
}

#' @exportS3Method print inlavaan_waic
print.inlavaan_waic <- function(x, ...) {
  unit_word <- switch(x$type, loso = "subject", loco = "cluster")
  cat("WAIC (INLAvaan)\n")
  cat(
    "Computed from ",
    x$nsamp,
    " posterior draws and ",
    x$n_units,
    " ",
    unit_word,
    if (x$n_units != 1L) "s",
    if (!is.null(x$n_groups) && x$n_groups > 1L) {
      paste0(" in ", x$n_groups, " groups")
    },
    "\n",
    sep = ""
  )
  if (identical(x$flavour, "conditional")) {
    cat("Scored conditionally on the exogenous covariates (fixed.x fit)\n")
  }
  cat("\n")
  print(round(x$estimates, 1))
  n_high <- sum(x$per_unit$p_waic > 0.4, na.rm = TRUE)
  if (n_high > 0L) {
    cat(
      "\n",
      n_high,
      " unit",
      if (n_high != 1L) "s",
      " with p_waic > 0.4: the WAIC may be unreliable; prefer loo().\n",
      sep = ""
    )
  }
  invisible(x)
}

#' Timing Information for INLAvaan Models
#'
#' Extract wall-clock timings for individual computation stages of a fitted
#' \code{INLAvaan} model.
#'
#' @param object An object of class [INLAvaan].
#' @param what Character vector of timing segment names to return, or
#'   \code{"all"} to return every segment. Defaults to \code{"total"}.
#'   Available segments (depending on model options): \code{"init"},
#'   \code{"optim"}, \code{"vb"}, \code{"loglik"}, \code{"marginals"},
#'   \code{"norta"}, \code{"sampling"}, \code{"covariances"},
#'   \code{"definedpars"}, \code{"deltapars"}, \code{"test"}, \code{"total"}.
#' @param ... Currently unused.
#'
#' @returns A named numeric vector (class \code{c("timing.INLAvaan",
#'   "numeric")}) of elapsed times in seconds. Printing formats short
#'   durations as seconds, longer ones as minutes or hours.
#'
#' @examples
#' \donttest{
#' HS.model <- "
#'   visual  =~ x1 + x2 + x3
#'   textual =~ x4 + x5 + x6
#'   speed   =~ x7 + x8 + x9
#' "
#' utils::data("HolzingerSwineford1939", package = "lavaan")
#' fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
#'             test = "none", verbose = FALSE)
#'
#' # Total elapsed time
#' timing(fit)
#'
#' # All stages
#' timing(fit, what = "all")
#'
#' # Specific stages
#' timing(fit, what = c("optim", "marginals"))
#' }
#'
#' @seealso [diagnostics()], [summary()]
#'
#' @export
setGeneric("timing", function(object, ...) standardGeneric("timing"))

#' @name timing
#' @rdname timing
#' @aliases timing,INLAvaan-method
#' @export
setMethod(
  "timing",
  "INLAvaan",
  function(object, what = "total", ...) {
    t <- object@timing
    available <- names(t)

    if (identical(what, "all")) {
      what <- available
    } else {
      unknown <- setdiff(what, available)
      if (length(unknown) > 0L) {
        cli_abort(c(
          "Unknown timing segment{?s}: {.val {unknown}}.",
          "i" = "Available: {.val {available}}."
        ))
      }
    }

    out <- unlist(t[what])
    names(out) <- what
    class(out) <- c("timing.INLAvaan", "numeric")
    out
  }
)

#' @exportS3Method print timing.INLAvaan
print.timing.INLAvaan <- function(x, ...) {
  fmt_time <- function(s) {
    if (is.na(s)) {
      return("NA")
    }
    if (s < 60) {
      return(sprintf("%.2f s", s))
    }
    if (s < 3600) { # nocov start
      return(sprintf("%.1f min", s / 60))
    }
    sprintf("%.2f hr", s / 3600) # nocov end
  }
  vals <- vapply(x, fmt_time, character(1))
  names(vals) <- names(x)
  print(vals, quote = FALSE, right = TRUE)
  invisible(x)
}

# nocov start
visual_debug <- function(object, params, logscale = FALSE, use_ggplot = TRUE,
                         points = FALSE) {
  if (inherits(object, "INLAvaan")) {
    dat_list <- object@external$inlavaan_internal$visual_debug
  } else if (inherits(object, "inlavaan_internal")) {
    dat_list <- object$visual_debug
  }

  all_names <- names(coef(object))
  if (!missing(params) && !is.null(params)) {
    if (is.numeric(params)) {
      keep_names <- all_names[params]
    } else {
      keep_names <- params
    }
    dat_list <- dat_list[names(dat_list) %in% keep_names]
    all_names <- keep_names
  }

  type_labels <- c(
    Original = "Raw", Corrected = "Corrected",
    SN_Fit = "Skew-normal fit"
  )
  type_cols <- c(
    Original = "gray60", Corrected = "black", SN_Fit = "red2"
  )
  type_lty <- c(Original = 1, Corrected = 1)
  type_lwd <- c(Original = 1.3, Corrected = 1.3)

  use_ggplot <- isTRUE(use_ggplot) &&
    requireNamespace("ggplot2", quietly = TRUE)

  if (use_ggplot) {
    # --- ggplot2 version ---
    plot_df <- do.call(rbind, Map(
      function(nm, df) { df$name <- nm; df },
      names(dat_list), dat_list
    ))
    rownames(plot_df) <- NULL
    plot_df <- reshape(
      plot_df, direction = "long",
      varying = list(c("Original", "Corrected", "SN_Fit")),
      v.names = "value",
      timevar = "type",
      times = c("Original", "Corrected", "SN_Fit"),
      idvar = c("name", "x")
    )
    rownames(plot_df) <- NULL
    plot_df$name <- factor(plot_df$name, levels = all_names)
    plot_df$type <- factor(
      plot_df$type,
      levels = c("Original", "Corrected", "SN_Fit"),
      labels = c("Raw", "Corrected", "Skew-normal fit")
    )
    if (isTRUE(logscale)) {
      plot_df$value[plot_df$value <= 0] <- NA
      plot_df <- plot_df[!is.na(plot_df$value), ]
      plot_df$value <- log10(plot_df$value)
    }

    x <- type <- value <- NULL # no visible binding NOTE
    df_sn <- subset(plot_df, type == "Skew-normal fit")

    p <- ggplot2::ggplot(plot_df, ggplot2::aes(x, value, col = type,
                                                linetype = type,
                                                linewidth = type)) +
      ggplot2::geom_line() +
      ggplot2::facet_wrap(. ~ name) +
      ggplot2::theme_minimal() +
      ggplot2::theme(
        legend.position = "top",
        legend.key.width = grid::unit(1.2, "cm")
      ) +
      ggplot2::labs(col = NULL, x = NULL, y = NULL,
                    linetype = NULL, linewidth = NULL) +
      ggplot2::scale_colour_manual(
        values = c("Raw" = "gray60", "Corrected" = "black",
                   "Skew-normal fit" = "red2")
      ) +
      ggplot2::scale_linetype_manual(
        values = c("Raw" = "solid", "Corrected" = "solid",
                   "Skew-normal fit" = "dashed")
      ) +
      ggplot2::scale_linewidth_manual(
        values = c("Raw" = 0.65, "Corrected" = 0.65, "Skew-normal fit" = 0.45)
      )

    if (isTRUE(points)) {
      p <- p +
        ggplot2::geom_point(ggplot2::aes(shape = type)) +
        ggplot2::scale_shape_manual(
          values = c("Raw" = 16, "Corrected" = 16, "Skew-normal fit" = 4),
          name = NULL
        )
    }

    return(p)
  }

  # --- base R fallback ---
  n_params <- length(dat_list)
  n_cols <- ceiling(sqrt(n_params))
  n_rows <- ceiling(n_params / n_cols)

  # Reserve top row for a horizontal legend
  layout_mat <- matrix(seq_len(n_rows * n_cols), nrow = n_rows, ncol = n_cols,
                       byrow = TRUE)
  layout_mat <- rbind(rep(n_rows * n_cols + 1, n_cols), layout_mat)
  layout(layout_mat, heights = c(0.8, rep(4, n_rows)))
  op <- par(mar = c(2, 2, 2, 1), oma = c(0, 0, 0, 0))
  on.exit(par(op))

  for (nm in names(dat_list)) {
    dd <- dat_list[[nm]]
    xvals <- dd$x
    ycols <- c("Original", "Corrected", "SN_Fit")

    ylim_vals <- unlist(dd[, ycols, drop = FALSE])
    if (isTRUE(logscale)) {
      ylim_vals[ylim_vals <= 0] <- NA
      ylim_vals <- log10(ylim_vals[!is.na(ylim_vals)])
    }

    plot(NULL, xlim = range(xvals), ylim = range(ylim_vals, na.rm = TRUE),
         main = nm, font.main = 1, xlab = "", ylab = "", bty = "n",
         axes = TRUE)
    # grid(col = "lightgray", lty = "solid")

    for (tp in ycols) {
      yvals <- dd[[tp]]
      if (isTRUE(logscale)) {
        yvals[yvals <= 0] <- NA
        yvals <- log10(yvals)
      }
      if (tp == "SN_Fit") {
        lines(xvals, yvals, col = type_cols[tp], lty = 2, lwd = 0.9)
        if (isTRUE(points)) {
          points(xvals, yvals, col = type_cols[tp], pch = 4, cex = 0.8)
        }
      } else {
        lines(xvals, yvals, col = type_cols[tp], lty = type_lty[tp],
              lwd = type_lwd[tp])
        if (isTRUE(points)) {
          points(xvals, yvals, col = type_cols[tp], pch = 16, cex = 0.8)
        }
      }
    }
  }

  # Fill remaining empty panels
  remaining <- n_rows * n_cols - n_params
  for (i in seq_len(remaining)) plot.new()

  # Top legend panel
  par(mar = c(0, 0, 0, 0))
  plot.new()
  if (isTRUE(points)) {
    legend_pch <- c(16L, 16L, 4L)
  } else {
    legend_pch <- c(NA_integer_, NA_integer_, NA_integer_)
  }
  legend(
    "center",
    legend = c("Raw", "Corrected", "Skew-normal fit"),
    col    = c(type_cols["Original"], type_cols["Corrected"], type_cols["SN_Fit"]),
    lty    = c(1L, 1L, 2L),
    pch    = legend_pch,
    lwd    = c(1.3, 1.3, 0.9),
    horiz  = TRUE, bty = "n", cex = 0.9, seg.len = 1.5
  )

  invisible(NULL)
}
# nocov end

#' @exportS3Method plot inlavaan_internal
#' @keywords internal
plot.inlavaan_internal <- function(
  x,
  truth,
  type = c("marg_pdf", "sn_fit", "sn_fit_log"),
  params = "all",
  nrow = NULL,
  ncol = NULL,
  use_ggplot = TRUE,
  points = FALSE,
  ...
) {
  type <- match.arg(type)

  # Dispatch to visual_debug for sn_fit types
  if (type %in% c("sn_fit", "sn_fit_log")) { # nocov start
    logscale <- type == "sn_fit_log"
    if (identical(params, "all")) {
      params <- NULL
    }
    return(visual_debug(
      x,
      params = params,
      logscale = logscale,
      use_ggplot = use_ggplot,
      points = points
    ))
  } # nocov end

  all_names <- names(x$pdf_data)
  postmode <- setNames(x$summary[, "Mode"], rownames(x$summary))

  # Resolve which parameters to plot
  if (identical(params, "all")) {
    param_names <- all_names
  } else {
    bad <- setdiff(params, all_names)
    if (length(bad) > 0) {
      # nocov start
      stop(
        "Unknown parameter(s): ",
        paste(bad, collapse = ", "),
        "\nAvailable: ",
        paste(all_names, collapse = ", ")
      )
    } # nocov end
    param_names <- params
  }
  n_plot <- length(param_names)

  use_ggplot <- isTRUE(use_ggplot) &&
    requireNamespace("ggplot2", quietly = TRUE)

  if (use_ggplot) {
    # --- ggplot2 version (facet_wrap, no extra dependency) ---
    plot_df <- do.call(
      rbind,
      Map(
        function(nm, df) {
          df$name <- nm
          df
        },
        param_names,
        x$pdf_data[param_names]
      )
    )
    rownames(plot_df) <- NULL
    plot_df$name <- factor(plot_df$name, levels = param_names)

    modes <- postmode[param_names]
    if (missing(truth)) {
      vline_df <- data.frame(
        name = factor(param_names, levels = param_names),
        xint = modes
      )
    } else {
      n_truth <- min(length(truth), length(param_names))
      vline_df <- data.frame(
        name = factor(param_names[seq_len(n_truth)], levels = param_names),
        xint = truth[seq_len(n_truth)]
      )
    }

    x <- xint <- NULL # no visible binding NOTE
    p <- ggplot2::ggplot(plot_df, ggplot2::aes(x, y)) +
      ggplot2::geom_line() +
      ggplot2::facet_wrap(~name, scales = "free", nrow = nrow, ncol = ncol) +
      ggplot2::theme_minimal() +
      ggplot2::labs(x = NULL, y = NULL)

    if (missing(truth)) {
      p <- p +
        ggplot2::geom_vline(
          data = vline_df,
          ggplot2::aes(xintercept = xint),
          linetype = "dashed",
          color = "red3"
        )
    } else {
      p <- p +
        ggplot2::geom_vline(
          data = vline_df,
          ggplot2::aes(xintercept = xint),
          linetype = "dotted",
          color = "steelblue3"
        )
    }

    return(p)
  }

  # --- base R fallback ---
  if (!is.null(ncol) && !is.null(nrow)) {
    n_cols <- ncol
    n_rows <- nrow
  } else if (!is.null(ncol)) {
    # nocov start
    n_cols <- ncol
    n_rows <- ceiling(n_plot / n_cols)
  } else if (!is.null(nrow)) {
    n_rows <- nrow
    n_cols <- ceiling(n_plot / n_rows)
  } else {
    # nocov end
    n_cols <- ceiling(sqrt(n_plot))
    n_rows <- ceiling(n_plot / n_cols)
  }

  layout_mat <- matrix(
    seq_len(n_rows * n_cols),
    nrow = n_rows,
    ncol = n_cols,
    byrow = TRUE
  )
  layout_mat <- rbind(rep(n_rows * n_cols + 1, n_cols), layout_mat)
  layout(layout_mat, heights = c(0.8, rep(4, n_rows)))
  op <- par(mar = c(2, 2, 2, 1), oma = c(0, 0, 0, 0))
  on.exit(par(op))

  for (j in seq_len(n_plot)) {
    param <- param_names[j]
    plot_df <- x$pdf_data[[param]]

    plot(
      plot_df$x,
      plot_df$y,
      type = "l",
      main = param,
      font.main = 1,
      xlab = "",
      ylab = "",
      bty = "n",
      axes = TRUE,
      col = "black",
      ...
    )

    if (missing(truth)) {
      abline(v = postmode[param], lty = 2, col = "red3")
    } else {
      abline(v = truth[j], lty = 3, col = "steelblue3")
    }
  }

  remaining <- n_rows * n_cols - n_plot
  for (i in seq_len(remaining)) {
    plot.new()
  }

  par(mar = c(0, 0, 0, 0))
  plot.new()
  if (missing(truth)) {
    legend(
      "center",
      legend = "Mode",
      col = "red3",
      lty = 2,
      horiz = TRUE,
      bty = "n",
      cex = 0.9,
      seg.len = 1.5
    )
  } else {
    legend(
      "center",
      legend = "Truth",
      col = "steelblue3",
      lty = 3,
      horiz = TRUE,
      bty = "n",
      cex = 0.9,
      seg.len = 1.5
    )
  }

  invisible(NULL)
}

#' Plot an INLAvaan Object
#'
#' Generates diagnostic plots for a fitted \code{INLAvaan} model.
#'
#' @param x An object of class [INLAvaan].
#' @param y Not used.
#' @param type Character. One of \code{"marg_pdf"} (default; posterior marginal
#'   densities), \code{"sn_fit"} (skew-normal fit diagnostic on natural scale),
#'   or \code{"sn_fit_log"} (same on log scale).
#' @param params Character vector of parameter names to plot, or \code{"all"}
#'   (default) to plot all free parameters.
#' @param nrow,ncol Integer. Number of rows/columns for the facet grid when
#'   \code{use_ggplot = TRUE}. If \code{NULL} (default), layout is chosen
#'   automatically.
#' @param use_ggplot Logical. When \code{TRUE} (default) and \pkg{ggplot2} is
#'   available, a \pkg{ggplot2} object is returned. Set to \code{FALSE} for a
#'   base-R plot.
#' @param points Logical. When \code{TRUE}, individual grid points are overlaid
#'   on the curves. Defaults to \code{FALSE}.
#' @param ... Additional arguments (currently unused).
#'
#' @examples
#' \donttest{
#' HS.model <- "
#'   visual  =~ x1 + x2 + x3
#'   textual =~ x4 + x5 + x6
#'   speed   =~ x7 + x8 + x9
#' "
#' utils::data("HolzingerSwineford1939", package = "lavaan")
#' fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
#'             test = "none", verbose = FALSE)
#'
#' # Posterior marginal densities (default)
#' plot(fit)
#'
#' # Skew-normal fit diagnostic for a single parameter
#' plot(fit, type = "sn_fit", params = "visual=~x1")
#' }
#'
#' @seealso [diagnostics()], [summary()]
#'
#' @name plot
#' @rdname plot
#' @aliases plot,INLAvaan,ANY-method
#' @export
setMethod(
  "plot",
  "INLAvaan",
  function(
    x,
    y,
    type = c("marg_pdf", "sn_fit", "sn_fit_log"),
    params = "all",
    nrow = NULL,
    ncol = NULL,
    use_ggplot = TRUE,
    points = FALSE,
    ...
  ) {
    plot.inlavaan_internal(
      x@external$inlavaan_internal,
      type = type,
      params = params,
      nrow = nrow,
      ncol = ncol,
      use_ggplot = use_ggplot,
      points = points,
      ...
    )
  }
)

# Silence R CMD check notes for ggplot2 pipelines
utils::globalVariables(c(
  "name",
  "method",
  "value",
  "x",
  "y",
  "JS_percent",
  "metric"
))

# nocov start
compare_mcmc <- function(
  fit_blavaan,
  ...,
  params = NULL,
  show_error = TRUE,
  truth = NULL,
  use_ggplot = TRUE,
  nrow = NULL,
  ncol = NULL
) {
  parnames <- unique(names(coef(fit_blavaan)))

  if (requireNamespace("blavaan", quietly = TRUE) == FALSE) {
    cli_abort("blavaan is required for plotting. Please install it.")
  }

  # MCMC Histograms
  draws <- do.call("rbind", blavaan::blavInspect(fit_blavaan, "mcmc"))
  draws_df <- as.data.frame(draws)
  plot_df_blav <- data.frame(
    name = rep(colnames(draws_df), each = nrow(draws_df)),
    value = unlist(draws_df, use.names = FALSE),
    stringsAsFactors = FALSE
  )
  plot_df_blav$name <- factor(plot_df_blav$name, levels = parnames)

  # INLAvaan Densities
  fit_inlavaan_list <- list(...)
  fit_inlavaan_list <- lapply(fit_inlavaan_list, function(fit) {
    if (inherits(fit, "INLAvaan")) {
      return(fit@external$inlavaan_internal$pdf_data)
    } else if (inherits(fit, "inlavaan_internal")) {
      return(fit$pdf_data)
    } else {
      stop(
        "Unsupported object type. Provide INLAvaan or inlavaan_internal objects."
      )
    }
  })
  inlav_names <- names(fit_inlavaan_list)

  # INLAvaan may estimate fewer parameters than blavaan. In particular, blavaan
  # fits a meanstructure (intercepts) by default whereas INLAvaan does not
  # (meanstructure = FALSE), so coef(fit_blavaan) carries intercepts that have
  # no INLAvaan counterpart. Restrict the comparison to parameters INLAvaan
  # actually produced, otherwise those blavaan-only parameters render as panels
  # with an MCMC density but no INLAvaan curve.
  inlav_available <- unique(unlist(lapply(fit_inlavaan_list, names)))
  parnames <- parnames[parnames %in% inlav_available]
  plot_df_blav <- plot_df_blav[
    as.character(plot_df_blav$name) %in% parnames,
    ,
    drop = FALSE
  ]
  plot_df_blav$name <- factor(
    as.character(plot_df_blav$name),
    levels = parnames
  )

  # Subset parameters if requested
  if (!is.null(params)) {
    bad <- setdiff(params, parnames)
    if (length(bad) > 0) {
      cli_abort(c(
        "Unknown parameter(s): {paste(bad, collapse = ', ')}",
        i = "Available: {paste(parnames, collapse = ', ')}"
      ))
    }
    parnames <- params
    plot_df_blav <- plot_df_blav[
      plot_df_blav$name %in% parnames,
      ,
      drop = FALSE
    ]
    plot_df_blav$name <- factor(
      as.character(plot_df_blav$name),
      levels = parnames
    )
    fit_inlavaan_list <- lapply(fit_inlavaan_list, function(pdfs) {
      pdfs[parnames]
    })
    if (!is.null(truth)) truth <- truth[parnames]
  }

  mycols <- c("#00A6AA", "#F18F00", "#adbf04", "#9C6FAE")
  mycols <- mycols[1:length(inlav_names)]
  names(mycols) <- inlav_names

  # Create plot
  plot_df_parts <- lapply(names(fit_inlavaan_list), function(meth) {
    pdf_list <- fit_inlavaan_list[[meth]]
    part <- do.call(
      rbind,
      Map(
        function(nm, df) {
          df$name <- nm
          df
        },
        names(pdf_list),
        pdf_list
      )
    )
    part$method <- meth
    part
  })
  plot_df <- do.call(rbind, plot_df_parts)
  rownames(plot_df) <- NULL
  plot_df$name <- factor(plot_df$name, levels = parnames)
  plot_df$method <- factor(plot_df$method, levels = inlav_names)
  plot_df <- plot_df[!is.na(plot_df$name), ]

  # Helper function for Integration
  trapz <- function(x, y) {
    if (length(x) < 2) {
      return(0)
    }
    sum((head(y, -1) + tail(y, -1)) * diff(x) / 2, na.rm = TRUE)
  }

  # Align MCMC to Approximation Grid
  plot_df_aligned <- do.call(
    rbind,
    lapply(
      split(plot_df, list(plot_df$name, plot_df$method), drop = TRUE),
      function(data) {
        pnm <- as.character(data$name[1])
        mcmc_vals <- plot_df_blav$value[plot_df_blav$name == pnm]

        # Create KDE of MCMC (gold standard). We constrain the KDE to the range
        # of your approximation to prevent extrapolation
        d <- stats::density(
          mcmc_vals,
          from = min(data$x),
          to = max(data$x),
          n = length(data$x)
        )

        # Interpolate KDE onto exact x-points of your approximation. 'rule = 2'
        # clamps values at the ends if floating point errors occur
        data$f_mcmc <- stats::approx(
          x = d$x,
          y = d$y,
          xout = data$x,
          rule = 2
        )$y
        data
      }
    )
  )
  rownames(plot_df_aligned) <- NULL

  # Calculate Metrics
  metrics_df <- do.call(
    rbind,
    lapply(
      split(
        plot_df_aligned,
        list(plot_df_aligned$name, plot_df_aligned$method),
        drop = TRUE
      ),
      function(chunk) {
        x_val <- chunk$x
        pa <- chunk$y # Approx
        pm <- chunk$f_mcmc # MCMC
        eps <- 1e-12

        # Normalize
        Za <- trapz(x_val, pa)
        pa <- if (!is.na(Za) && Za > 0) pa / Za else pa
        Zm <- trapz(x_val, pm)
        pm <- if (!is.na(Zm) && Zm > 0) pm / Zm else pm

        # --- KL Divergences ---
        kl_fwd <- trapz(x_val, pm * (log(pmax(pm, eps)) - log(pmax(pa, eps))))
        kl_rev <- trapz(x_val, pa * (log(pmax(pa, eps)) - log(pmax(pm, eps))))

        # --- Interpretability Hacks ---
        # 1. Jensen-Shannon % (0 = Identical, 1 = Disjoint)
        m_mix <- 0.5 * (pm + pa)
        kl_pm <- trapz(x_val, pm * (log(pmax(pm, eps)) - log(pmax(m_mix, eps))))
        kl_qm <- trapz(x_val, pa * (log(pmax(pa, eps)) - log(pmax(m_mix, eps))))
        js_val <- 0.5 * kl_pm + 0.5 * kl_qm
        js_pct <- js_val / log(2)

        # 2. Gaussian Bias Equivalent
        bias_equiv <- sqrt(2 * abs(kl_fwd))

        # --- L1 (Total Variation) ---
        l1 <- trapz(x_val, abs(pm - pa))

        data.frame(
          name = as.character(chunk$name[1]),
          method = as.character(chunk$method[1]),
          L1 = l1,
          L1_percent = l1 / 2,
          js = js_val,
          JS_percent = js_pct,
          KL_fwd = kl_fwd,
          KL_rev = kl_rev,
          Bias_SDs = bias_equiv,
          stringsAsFactors = FALSE
        )
      }
    )
  )
  rownames(metrics_df) <- NULL
  metrics_df <- metrics_df[order(metrics_df$method, -metrics_df$Bias_SDs), ]

  # Add overall by averaging
  overall <- aggregate(
    cbind(L1, L1_percent, js, JS_percent, KL_fwd, KL_rev, Bias_SDs) ~
      method,
    data = metrics_df,
    FUN = mean
  )
  overall$name <- "Overall"
  metrics_df <- rbind(metrics_df, overall)
  metrics_df$name <- factor(metrics_df$name, levels = c("Overall", parnames))
  # Average within name+method (collapses duplicates for Overall)
  metrics_df <- aggregate(
    cbind(L1, L1_percent, js, JS_percent, KL_fwd, KL_rev, Bias_SDs) ~
      name + method,
    data = metrics_df,
    FUN = mean
  )
  metrics_df$name <- factor(metrics_df$name, levels = c("Overall", parnames))

  # --- Plotting ---
  use_ggplot <- isTRUE(use_ggplot) &&
    requireNamespace("ggplot2", quietly = TRUE)

  if (use_ggplot) {
    # --- ggplot2 version ---
    # Plot of L1 and JS errors
    err_long <- reshape(
      metrics_df[, c("name", "method", "L1_percent", "JS_percent")],
      direction = "long",
      varying = list(c("L1_percent", "JS_percent")),
      v.names = "value",
      timevar = "metric",
      times = c("L1 Error", "Jensen-Shannon Error"),
      idvar = c("name", "method")
    )
    rownames(err_long) <- NULL
    err_long$name <- factor(
      err_long$name,
      levels = rev(levels(factor(metrics_df$name)))
    )

    p_errors <-
      ggplot2::ggplot(err_long) +
      ggplot2::geom_bar(
        ggplot2::aes(x = name, y = value, fill = method),
        width = 0.7,
        stat = "identity",
        position = ggplot2::position_dodge()
      ) +
      ggplot2::coord_flip() +
      ggplot2::scale_y_continuous(labels = function(x) {
        paste0(round(x * 100), "%")
      }) +
      ggplot2::scale_fill_manual(values = mycols) +
      ggplot2::facet_grid(. ~ metric, scales = "free_x") +
      ggplot2::theme_minimal() +
      ggplot2::theme(
        plot.margin = ggplot2::margin(t = 4, r = 8, b = 8, l = 8),
        legend.position = "top",
        legend.box.spacing = grid::unit(2, "pt"),
        legend.key.width = grid::unit(1, "cm")
      ) +
      ggplot2::labs(x = NULL, y = NULL, fill = NULL)

    p_compare <-
      ggplot2::ggplot() +
      ggplot2::geom_density(
        data = plot_df_blav,
        ggplot2::aes(value, fill = "MCMC"),
        col = NA,
        alpha = 0.38
      ) +
      ggplot2::geom_line(
        data = plot_df,
        ggplot2::aes(x, y, group = method, col = method),
        linewidth = 0.75
      ) +
      ggplot2::facet_wrap(~name, scales = "free", nrow = nrow, ncol = ncol) +
      ggplot2::scale_colour_manual(values = mycols) +
      ggplot2::scale_fill_manual(values = c("MCMC" = "#131516")) +
      ggplot2::theme_minimal() +
      ggplot2::theme(
        plot.margin = ggplot2::margin(t = 4, r = 8, b = 8, l = 8),
        legend.position = "top",
        legend.box.spacing = grid::unit(2, "pt"),
        legend.key.width = grid::unit(1, "cm")
      ) +
      ggplot2::labs(x = NULL, y = NULL, col = NULL, fill = NULL)

    if (!is.null(truth)) {
      if (length(truth) != length(parnames)) {
        cli_abort("Length of 'truth' must match number of parameters.")
      }

      truth_df <- data.frame(
        name = parnames,
        truth = as.numeric(truth)
      )

      p_compare <-
        p_compare +
        ggplot2::geom_vline(
          data = truth_df,
          ggplot2::aes(xintercept = truth),
          linetype = "dashed",
          color = "red",
          linewidth = 0.5
        )
    }

    if (isTRUE(show_error)) {
      # Pre-compute per-parameter effective right edge from MCMC KDE.
      # Use the rightmost x where density > 1% of peak to ignore sparse outliers.
      mcmc_q95 <- sapply(parnames, function(pnm) {
        vals <- plot_df_blav$value[plot_df_blav$name == pnm]
        d <- stats::density(vals)
        threshold <- 0.0025 * max(d$y)
        max(d$x[d$y >= threshold])
      })
      names(mcmc_q95) <- parnames

      # Summarise plot_df by name + method
      label_df <- do.call(
        rbind,
        lapply(
          split(plot_df, list(plot_df$name, plot_df$method), drop = TRUE),
          function(chunk) {
            pnm <- as.character(chunk$name[1])
            x_approx <- quantile(chunk$x, probs = 0.95)
            data.frame(
              name = chunk$name[1],
              method = chunk$method[1],
              x = max(x_approx, mcmc_q95[[pnm]]),
              panel_max_y = max(chunk$y),
              stringsAsFactors = FALSE
            )
          }
        )
      )
      rownames(label_df) <- NULL
      # Use a common reference height per panel (max across all methods),
      # then stagger each method's label vertically so they don't overlap.
      pmax_y <- tapply(label_df$panel_max_y, label_df$name, max)
      label_df$panel_max_y <- pmax_y[as.character(label_df$name)]
      label_df$method_rank <- match(label_df$method, inlav_names)
      label_df$y <- label_df$panel_max_y *
        (0.925 - 0.15 * (label_df$method_rank - 1))
      label_df$panel_max_y <- NULL
      label_df$method_rank <- NULL
      label_df <- merge(label_df, metrics_df, by = c("name", "method"))
      label_df$JS_percent <- {
        vals <- (1 - label_df$JS_percent) * 100
        ifelse(vals >= 99.95, "100%", sprintf("%.1f%%", vals))
      }

      p_compare <-
        p_compare +
        ggplot2::geom_text(
          data = label_df,
          ggplot2::aes(x, y, label = JS_percent, col = method),
          size = 3,
          hjust = 0.7,
          show.legend = FALSE
        )
    }

    return(list(
      p_compare = p_compare,
      p_errors = p_errors,
      metrics_df = metrics_df
    ))
  }

  # --- base R fallback ---
  # Use a temporary null device to build plots, then capture with recordPlot()
  n_params <- length(parnames)
  n_cols <- ceiling(sqrt(n_params))
  n_rows <- ceiling(n_params / n_cols)

  truth_vec <- NULL
  if (!is.null(truth)) {
    if (length(truth) != length(parnames)) {
      cli_abort("Length of 'truth' must match number of parameters.")
    }
    truth_vec <- setNames(as.numeric(truth), parnames)
  }

  # -- p_compare: MCMC density + INLAvaan density lines per parameter --
  # Reserve top row for a horizontal legend
  layout_mat <- matrix(
    seq_len(n_rows * n_cols),
    nrow = n_rows,
    ncol = n_cols,
    byrow = TRUE
  )
  layout_mat <- rbind(rep(n_rows * n_cols + 1, n_cols), layout_mat)
  layout(layout_mat, heights = c(0.8, rep(4, n_rows)))
  op <- par(mar = c(2, 2, 2, 1), oma = c(0, 0, 0, 0))

  for (pnm in parnames) {
    mcmc_vals <- plot_df_blav$value[plot_df_blav$name == pnm]
    d_mcmc <- stats::density(mcmc_vals)

    sub_df <- plot_df[plot_df$name == pnm, ]
    xlim <- range(c(d_mcmc$x, sub_df$x), na.rm = TRUE)
    ylim <- range(c(d_mcmc$y, sub_df$y), na.rm = TRUE)

    plot(
      d_mcmc,
      main = pnm,
      font.main = 1,
      xlab = "",
      ylab = "",
      bty = "n",
      col = NA,
      xlim = xlim,
      ylim = ylim,
      zero.line = FALSE
    )
    # grid(col = "lightgray", lty = "solid")
    polygon(d_mcmc$x, d_mcmc$y, col = adjustcolor("#131516", 0.25), border = NA)

    for (meth in names(mycols)) {
      mdf <- sub_df[sub_df$method == meth, ]
      if (nrow(mdf) > 0) {
        lines(mdf$x, mdf$y, col = mycols[meth], lwd = 1.5)
      }
    }

    if (!is.null(truth_vec)) {
      abline(v = truth_vec[pnm], lty = 2, col = "red", lwd = 0.8)
    }

    if (isTRUE(show_error)) {
      m_sub <- metrics_df[metrics_df$name == pnm, , drop = FALSE]
      for (i in seq_len(nrow(m_sub))) {
        val <- 100 * (1 - m_sub$JS_percent[i])
        lbl <- if (val >= 99.95) "100%" else sprintf("%.1f%%", val)
        mtext(
          lbl,
          side = 3,
          line = -1.2 - (i - 1) * 1,
          col = mycols[as.character(m_sub$method[i])],
          cex = 0.7,
          adj = 0.95
        )
      }
    }
  }

  # Fill any remaining empty panels
  remaining <- n_rows * n_cols - n_params
  for (i in seq_len(remaining)) {
    plot.new()
  }

  # Top legend panel
  par(mar = c(0, 0, 0, 0))
  plot.new()
  n_leg <- 1 + length(mycols)
  leg_labels <- c("MCMC", names(mycols))
  leg_pch <- c(22, rep(NA, length(mycols)))
  leg_pt_bg <- c(adjustcolor("#131516", 0.25), rep(NA, length(mycols)))
  leg_col <- c("gray40", mycols)
  leg_lwd <- c(NA, rep(1.5, length(mycols)))
  leg_lty <- c(NA, rep(1, length(mycols)))
  legend(
    "center",
    legend = leg_labels,
    pch = leg_pch,
    pt.bg = leg_pt_bg,
    pt.cex = 2,
    col = leg_col,
    lwd = leg_lwd,
    lty = leg_lty,
    horiz = TRUE,
    bty = "n",
    cex = 0.9,
    seg.len = 1.5
  )
  p_compare <- recordPlot()

  # -- p_errors: horizontal bar chart of L1 and JS errors --
  par(mfrow = c(1, 2), mar = c(4, 8, 3, 1), oma = c(0, 0, 2, 0))

  err_long <- metrics_df[, c("name", "method", "L1_percent", "JS_percent")]
  err_methods <- levels(factor(err_long$method))
  err_names <- rev(levels(factor(err_long$name)))
  n_methods <- length(err_methods)

  for (metric in c("L1_percent", "JS_percent")) {
    label <- if (metric == "L1_percent") "L1 Error" else "Jensen-Shannon Error"
    mat <- matrix(
      0,
      nrow = length(err_names),
      ncol = n_methods,
      dimnames = list(err_names, err_methods)
    )
    for (i in seq_len(nrow(err_long))) {
      r <- as.character(err_long$name[i])
      m <- as.character(err_long$method[i])
      if (r %in% err_names) mat[r, m] <- err_long[[metric]][i]
    }

    barplot(
      t(mat) * 100,
      beside = TRUE,
      horiz = TRUE,
      las = 1,
      col = mycols[err_methods],
      border = NA,
      main = label,
      xlab = "%",
      names.arg = err_names
    )
  }
  mtext("Error Metrics", outer = TRUE, cex = 1)
  legend(
    "topright",
    legend = err_methods,
    fill = mycols[err_methods],
    border = NA,
    bty = "n",
    cex = 0.8
  )
  p_errors <- recordPlot()

  par(op)

  list(
    p_compare = p_compare,
    p_errors = p_errors,
    metrics_df = metrics_df
  )
}
# nocov end

#' @keywords internal
#' @section Main features:
#' * [acfa()]: Approximate Confirmatory Factor Analysis.
#' * [asem()]: Approximate Structural Equation Modelling.
#' * [agrowth()]: Approximate Latent Growth Curve models.
#'
#' @section Model specifications:
#' Supports advanced 'lavaan' syntax features, including:
#' * Equality constraints
#' * Defined parameters (e.g., `:=` operator for indirect effects)
#' * Flexible prior specifications
#'
#' @section Methods for \code{INLAvaan} objects:
#' After fitting a model an [INLAvaan] object is returned. The following S4
#' methods are available. See [INLAvaan-class] for the class definition.
#'
#' * Summaries and parameter estimates: [summary()], [coef()], [vcov()], [standardisedsolution()]
#' * Fit assessment and model comparison: [fitmeasures()], [bfit_indices()], [compare()], [diagnostics()], [timing()]
#' * Posterior inference and simulation: [predict()], [sampling()], [simulate()]
#' * Visualisation: [plot()]

#' @section Online vignettes:
#' The [package website](https://inlavaan.haziqj.ml/) contains comprehensive examples covering:
#' * Confirmatory Factor Analysis (CFA)
#' * Structural Equation Models (SEM)
#' * Latent Growth Curve Models
#' * Multigroup and Invariance Testing
#' * Mediation Analysis
#'
#' @aliases INLAvaan
"_PACKAGE"

## usethis namespace: start
#' @import methods
#' @importFrom cli cli_abort
#' @importFrom cli cli_alert_info
#' @importFrom cli cli_alert_warning
#' @importFrom cli cli_inform
#' @importFrom cli cli_progress_bar
#' @importFrom cli cli_progress_done
#' @importFrom cli cli_progress_step
#' @importFrom cli cli_progress_update
#' @importFrom cli cli_rule
#' @importFrom cli cli_warn
#' @importFrom cli col_grey
#' @importFrom cli symbol
#' @importFrom graphics abline
#' @importFrom graphics barplot
#' @importFrom graphics grid
#' @importFrom graphics layout
#' @importFrom graphics legend
#' @importFrom graphics lines
#' @importFrom graphics mtext
#' @importFrom graphics par
#' @importFrom graphics plot.new
#' @importFrom graphics polygon
#' @importFrom grDevices adjustcolor
#' @importFrom grDevices recordPlot
#' @importFrom stats aggregate
#' @importFrom stats approx
#' @importFrom stats cov
#' @importFrom stats cov2cor
#' @importFrom stats dbeta
#' @importFrom stats density
#' @importFrom stats dgamma
#' @importFrom stats dnorm
#' @importFrom stats median
#' @importFrom stats nlminb
#' @importFrom stats optimize
#' @importFrom stats pnorm
#' @importFrom stats qnorm
#' @importFrom stats quantile
#' @importFrom stats reshape
#' @importFrom stats rnorm
#' @importFrom stats sd
#' @importFrom stats setNames
#' @importFrom stats splinefun
#' @importFrom stats uniroot
#' @importFrom stats var
#' @importFrom utils capture.output
#' @importFrom utils head
#' @importFrom utils packageVersion
#' @importFrom utils tail
## usethis namespace: end
NULL

call_next_lavaan_method <- function(object, ...) {
  class(object) <- "lavaan"
  callNextMethod()
}

# Loading blavaan (Suggests; used side-by-side with INLAvaan in compare_mcmc()
# workflows and in this package's own tests) overwrites lavaan's own
# coef,lavaan-method in the live session with a stale copy that calls the
# unexported lav_object_inspect_coef()/add.labels()/add.class() signature,
# even when a newer lavaan (with lav_inspect_coef()/add_labels()/add_class())
# is installed and was in effect before blavaan's namespace loaded. This is a
# cross-package method-table collision, not a version-drift issue on the
# INLAvaan object itself, so no amount of faithful fit0 recomposition avoids
# it: callNextMethod() would dispatch to whichever definition is currently
# registered for "lavaan", which blavaan can silently replace out from under
# us. Resolve and call the internal directly instead (positional args only,
# see lavaan-unexported.R), same as the log-likelihood/gradient call sites.
coef_inlavaan <- function(object, ...) {
  dots <- list(...)
  type <- dots$type %||% "free"
  lavaan___lav_inspect_coef(object, type, TRUE, TRUE)
}

#' @name coef
#' @rdname INLAvaan-class
#' @aliases coef,INLAvaan-method
#' @export
setMethod("coef", "INLAvaan", coef_inlavaan)

#' @name nobs
#' @rdname INLAvaan-class
#' @aliases nobs,INLAvaan-method
#' @export
setMethod("nobs", "INLAvaan", call_next_lavaan_method)

# All functions here compute summaries and also table of (x, f(x)) values for
# plotting of the jth component parameter's posterior marginal

## ----- Two-piece asymmetric Gaussian -----------------------------------------
# Product factor log-pdf
prodfac_lp <- function(z, sigma_asym) {
  res <- 0
  for (j in seq_len(nrow(sigma_asym))) {
    if (z[j] > 0) {
      res <- res + -0.5 * (z[j] / sqrt(sigma_asym[j, "sigma_plus"]))^2
    } else {
      res <- res + -0.5 * (z[j] / sqrt(sigma_asym[j, "sigma_minus"]))^2
    }
  }
  res
}

# Marginalised distribution
marg_lp <- function(tj, j, theta_star, Sigma_theta, sigma_asym) {
  # Canonical-ordered Cholesky whitening — permutation invariant
  pnames    <- colnames(Sigma_theta)
  cperm     <- if (!is.null(pnames)) order(pnames) else seq_len(nrow(Sigma_theta))
  iperm     <- order(cperm)
  R         <- chol(Sigma_theta[cperm, cperm]) # canonical upper Cholesky
  L_canon   <- t(R)                            # lower Cholesky in canonical order
  # Rows → original order; cols stay canonical z-order
  L         <- L_canon[iperm, ]
  # L^{-1} for whitening: solve L x = b via canonical Cholesky
  # L^{-1} = L_canon^{-1} P, which is forwardsolve(L_canon, P x)

  sapply(tj, function(thetaj) {
    # Conditional expectation
    theta_new <- rep(NA, length(theta_star))
    theta_new[-j] <- theta_star[-j] +
      Sigma_theta[-j, j] / Sigma_theta[j, j] * (thetaj - theta_star[j])
    theta_new[j] <- thetaj

    # Whiten via canonical Cholesky: z = L^{-1}(θ - θ*)
    diff_canon <- (theta_new - theta_star)[cperm]
    z_new <- forwardsolve(L_canon, diff_canon)
    prodfac_lp(z_new, sigma_asym[cperm, , drop = FALSE])
  })
}

post_marg_asymgaus <- function(
  j = 1,
  g = identity,
  g_prime = function(x) 1,
  ginv = identity,
  ginv_prime = function(x) 1,
  theta_star,
  Sigma_theta,
  sigma_asym
) {
  # Build the density by laying out some points and spline interpolation
  tt <- theta_star[j] + seq(-4, 4, length = 100) * sqrt(Sigma_theta[j, j])
  yy <- marg_lp(
    tt,
    j = j,
    theta_star = theta_star,
    Sigma_theta = Sigma_theta,
    sigma_asym = sigma_asym
  )
  yy <- yy - max(yy) # stabilise
  fj_lp <- stats::splinefun(tt, yy)
  fj <- function(par) exp(fj_lp(par)) # unnormalised

  # PDF transform x = ginv(theta)
  transform_density <- function(.fj, hinv, hinv_prime) {
    function(y) .fj(hinv(y)) * abs(hinv_prime(y))
  }
  fj_orig <- transform_density(.fj = fj, hinv = g, hinv_prime = g_prime)

  # Posterior mean and SD
  x <- ginv(tt)
  dx <- diff(x)
  fx <- fj_orig(x)
  fmid <- (utils::head(fx, -1) + utils::tail(fx, -1)) / 2
  ymid <- (utils::head(x, -1) + utils::tail(x, -1)) / 2
  C <- sum(fmid * dx)
  fx <- fx / C
  Ex <- sum(ymid * fmid * dx) / C
  Vx <- sum(ymid^2 * fmid * dx) / C - Ex^2
  SDx <- sqrt(Vx)

  # Posterior mode
  xmax <- stats::optimize(fj_orig, interval = range(x), maximum = TRUE)$maximum

  # Build CDF
  Fx <- c(0, cumsum(fmid * dx))
  Fx <- Fx / tail(Fx, 1)
  qfj_orig <- stats::splinefun(Fx, x, method = "monoH.FC")

  # Combine results
  res <- c(Ex, SDx, qfj_orig(c(0.025, 0.25, 0.5, 0.75, 0.975)), xmax)
  names(res) <- c("Mean", "SD", "2.5%", "25%", "50%", "75%", "97.5%", "Mode")

  list(
    summary = res,
    pdf_data = data.frame(x = x, y = fx)
  )
}

# Skew normal approximations
post_marg_skewnorm <- function(
  j = 1,
  g = identity,
  g_prime = function(x) 1,
  ginv = identity,
  ginv_prime = function(x) 1,
  theta_star,
  Sigma_theta,
  sn_params
) {
  # Skew normal parameters
  xi <- sn_params[j, "xi"]
  omega <- sn_params[j, "omega"]
  alpha <- sn_params[j, "alpha"]

  # Build the density by pdf transform
  tt <- theta_star[j] + seq(-4, 4, length = 100) * sqrt(Sigma_theta[j, j])
  fj <- function(par) dsnorm(par, xi, omega, alpha, log = FALSE)
  transform_density <- function(.fj, hinv, hinv_prime) {
    function(y) .fj(hinv(y)) * abs(hinv_prime(y))
  }
  fj_orig <- transform_density(.fj = fj, hinv = g, hinv_prime = g_prime)
  x <- ginv(tt)
  dx <- diff(x)
  fx <- fj_orig(x)
  fmid <- (utils::head(fx, -1) + utils::tail(fx, -1)) / 2
  ymid <- (utils::head(x, -1) + utils::tail(x, -1)) / 2
  C <- sum(fmid * dx)
  fx <- fx / C

  # Compute mean and variance using GH quadrature
  quad <- .gauss_hermite(61)
  nodes <- quad$nodes * sqrt(2)
  weights <- quad$weights / sqrt(pi)
  ginvz <- ginv(xi + omega * nodes)
  Ex <- 2 * sum(weights * ginvz * pnorm(alpha * nodes))
  Exsq <- 2 * sum(weights * (ginvz^2) * pnorm(alpha * nodes))
  Vx <- Exsq - Ex^2
  SDx <- sqrt(Vx)

  # Compute quantiles
  qq <- ginv(qsnorm_fast(
    c(0.025, 0.25, 0.5, 0.75, 0.975),
    xi = xi,
    omega = omega,
    alpha = alpha
  ))

  # Compute mode
  xmax <- optimize(fj_orig, interval = range(x), maximum = TRUE)$maximum

  # Combine results
  res <- c(Ex, SDx, qq, xmax)
  names(res) <- c("Mean", "SD", "2.5%", "25%", "50%", "75%", "97.5%", "Mode")

  list(
    summary = res,
    pdf_data = data.frame(x = x, y = fx)
  )
}

# Marginal Gaussian approximation
post_marg_marggaus <- function(
  j = 1,
  g = identity,
  g_prime = function(x) 1,
  ginv = identity,
  ginv_prime = function(x) 1,
  theta_star,
  Sigma_theta
) {
  # Posterior mean and SD in theta space
  thetaj_mean <- theta_star[j]
  thetaj_sd <- sqrt(Sigma_theta[j, j])

  # Transform to original space
  x_mean <- ginv(thetaj_mean)
  x_sd <- abs(ginv_prime(thetaj_mean)) * thetaj_sd

  # Compute quantiles
  qq <- ginv(qnorm(c(0.025, 0.25, 0.5, 0.75, 0.975), mean = thetaj_mean, sd = thetaj_sd))

  # Build PDF data
  tt <- theta_star[j] + seq(-4, 4, length = 100) * sqrt(Sigma_theta[j, j])
  fj <- function(par) dnorm(par, mean = thetaj_mean, sd = thetaj_sd)
  transform_density <- function(.fj, hinv, hinv_prime) {
    function(y) .fj(hinv(y)) * abs(hinv_prime(y))
  }
  fj_orig <- transform_density(.fj = fj, hinv = g, hinv_prime = g_prime)
  x <- ginv(tt)
  dx <- diff(x)
  fx <- fj_orig(x)
  fmid <- (utils::head(fx, -1) + utils::tail(fx, -1)) / 2
  C <- sum(fmid * dx)
  fx <- fx / C

  # Compute mode
  xmax <- optimize(fj_orig, interval = range(x), maximum = TRUE)$maximum

  # Combine results
  res <- c(x_mean, x_sd, qq, xmax)
  names(res) <- c("Mean", "SD", "2.5%", "25%", "50%", "75%", "97.5%", "Mode")

  list(
    summary = res,
    pdf_data = data.frame(x = x, y = fx)
  )
}

# Sampling approximation
# Accepts pre-computed x_samp (nsamp x p matrix) and summarises each parameter
post_marg_sampling <- function(x_samp) {
  apply(x_samp, 2, summarise_samples)
}

# Unexported lavaan internals used by INLAvaan. lavaan >= 0.7 renamed many
# of its internal functions (e.g. lav_model_gradient -> lav_model_grad), and
# resolving them at the top level would bake in whichever name exists at
# *install* time. Instead, the bindings below start as NULL and are resolved
# once per session in .onLoad() (see zzz.R): for each binding, the first
# alias found in the lavaan namespace wins. Call-time overhead is zero.
#
# NOTE: lavaan >= 0.7 also renamed arguments (dot.case -> snake_case), so
# call sites must avoid argument names that differ across versions; they
# rely on positional matching instead (the order is stable).

lavaan___lav_model_loglik <- NULL
lavaan___lav_model_objective <- NULL
lavaan___lav_model_gradient <- NULL
lavaan___lav_model_lambda <- NULL
lavaan___lav_model_veta <- NULL
lavaan___lav_model_eeta <- NULL
lavaan___lav_model_ey <- NULL
lavaan___lav_mvnorm_loglik_samplestats <- NULL
lavaan___lav_mvnorm_missing_loglik_samplestats <- NULL
lavaan___lav_mvnorm_cluster_implied22l <- NULL
lavaan___lav_model_delta <- NULL
lavaan___lav_mvnorm_scores_mu_vech_sigma <- NULL
lavaan___lav_mvnorm_cluster_loglik_samplestats_2l <- NULL
lavaan___lav_mvnorm_cluster_dlogl_2l_samplestats <- NULL
lavaan___lav_mvnorm_cluster_missing_loglik_samplestats_2l <- NULL
lavaan___lav_mvnorm_cluster_missing_dlogl_2l_samplestats <- NULL
lavaan___lav_data_missing_patterns <- NULL
lavaan___lav_inspect_coef <- NULL

# Aliases ordered new (lavaan >= 0.7) first, old (lavaan < 0.7) last, so on
# new lavaan we bind the real function rather than its deprecated shim.
lavaan_internal_aliases <- list(
  lavaan___lav_model_loglik = "lav_model_loglik",
  lavaan___lav_model_objective = "lav_model_objective",
  lavaan___lav_model_gradient = c("lav_model_grad", "lav_model_gradient"),
  lavaan___lav_model_lambda = c("lav_model_lambda", "computeLAMBDA"),
  lavaan___lav_model_veta = c("lav_model_veta", "computeVETA"),
  lavaan___lav_model_eeta = c("lav_model_eeta", "computeEETA"),
  lavaan___lav_model_ey = c("lav_model_ey", "computeEY"),
  lavaan___lav_mvnorm_loglik_samplestats = c(
    "lav_mvn_loglik_samp",
    "lav_mvnorm_loglik_samplestats"
  ),
  lavaan___lav_mvnorm_missing_loglik_samplestats = c(
    "lav_mvn_mi_loglik_samp",
    "lav_mvnorm_missing_loglik_samplestats"
  ),
  lavaan___lav_mvnorm_cluster_implied22l = c(
    "lav_mvn_cl_implied22l",
    "lav_mvnorm_cluster_implied22l"
  ),
  lavaan___lav_model_delta = c("lav_model_delta", "computeDelta"),
  lavaan___lav_mvnorm_scores_mu_vech_sigma = c(
    "lav_mvn_sc_mu_sigma",
    "lav_mvnorm_scores_mu_vech_sigma"
  ),
  lavaan___lav_mvnorm_cluster_loglik_samplestats_2l = c(
    "lav_mvn_cl_loglik_samp_2l",
    "lav_mvnorm_cluster_loglik_samplestats_2l"
  ),
  lavaan___lav_mvnorm_cluster_dlogl_2l_samplestats = c(
    "lav_mvn_cl_dlogl_2l_samp",
    "lav_mvnorm_cluster_dlogl_2l_samplestats"
  ),
  lavaan___lav_mvnorm_cluster_missing_loglik_samplestats_2l = c(
    "lav_mvn_cl_mi_loglik_samp_2l",
    "lav_mvnorm_cluster_missing_loglik_samplestats_2l"
  ),
  lavaan___lav_mvnorm_cluster_missing_dlogl_2l_samplestats = c(
    "lav_mvn_cl_mi_dlogl_2l_samp",
    "lav_mvnorm_cluster_missing_dlogl_2l_samplestats"
  ),
  lavaan___lav_data_missing_patterns = c(
    "lav_data_mi_patterns",
    "lav_data_missing_patterns"
  ),
  lavaan___lav_inspect_coef = c("lav_inspect_coef", "lav_object_inspect_coef")
)

resolve_lavaan_internals <- function(ns) {
  lav_ns <- asNamespace("lavaan")
  for (binding in names(lavaan_internal_aliases)) {
    aliases <- lavaan_internal_aliases[[binding]]
    found <- NULL
    for (alias in aliases) {
      if (exists(alias, envir = lav_ns, inherits = FALSE)) {
        found <- get(alias, envir = lav_ns, inherits = FALSE)
        break
      }
    }
    if (is.null(found)) {
      # nocov start
      stop(
        "INLAvaan needs the internal lavaan function ",
        paste(sQuote(aliases), collapse = " or "),
        ", but neither exists in lavaan ",
        as.character(utils::packageVersion("lavaan")),
        ". Please report this at https://github.com/haziqj/INLAvaan/issues.",
        call. = FALSE
      )
      # nocov end
    }
    assign(binding, found, envir = ns)
  }
  invisible(NULL)
}

partable_classify_sem_matrix <- function(
  lhs,
  op,
  rhs,
  ov.names,
  std.ov,
  std.lv
) {
  lhs_is_ov <- lhs %in% ov.names
  rhs_is_ov <- rhs %in% ov.names

  if (op == "=~") {
    return("lambda")
  }

  if (op == "~~") {
    if (lhs_is_ov & rhs_is_ov) {
      if (lhs == rhs) {
        return("theta_var")
      } else {
        if (isTRUE(std.ov)) {
          return("theta_cor")
        } else {
          return("theta_cov")
        }
      }
    } else {
      (!lhs_is_ov & !rhs_is_ov)
    }
    if (lhs == rhs) {
      return("psi_var")
    } else {
      if (isTRUE(std.lv)) {
        return("psi_cor")
      } else {
        return("psi_cov")
      }
    }
  }

  if (op == "~*~") {
    return("delta")
  }

  if (op == "~") {
    # if (!lhs_is_ov) {
    return("beta")
    # }
  }

  if (op == "~1") {
    if (lhs_is_ov) {
      return("nu")
    } else if (!lhs_is_ov) {
      return("alpha")
    }
  }

  if (op == "|") {
    return("tau")
  }
  if (op == ":=") {
    return("defined")
  }
  if (op %in% c("==", "<", ">")) {
    return("constraint")
  }
  if (op == "@") {
    return("fixed")
  }

  return(NA_character_)
}

partable_prior_from_row <- function(matrix, lhs, rhs, op, dp) {
  is_var <- grepl("_var", matrix)

  if (matrix == "nu") {
    return(dp[["nu"]])
  }
  if (matrix == "alpha") {
    return(dp[["alpha"]])
  }
  if (matrix == "lambda") {
    return(dp[["lambda"]])
  }
  if (matrix == "beta") {
    return(dp[["beta"]])
  }

  if (grepl("theta", matrix)) {
    return(if (is_var) dp[["theta"]] else dp[["rho"]])
  }
  if (grepl("psi", matrix)) {
    return(if (is_var) dp[["psi"]] else dp[["rho"]])
  }

  if (matrix == "tau") {
    return(dp[["tau"]])
  }

  return(NA_character_)
}

safe_tanh <- function(x, eps = 1e-6) {
  (1 - eps) * tanh(x)
}

partable_transform_funcs <- function(matrix) {
  g <- identity
  g_prime <- function(x) 1
  ginv <- identity
  ginv_prime <- function(x) 1
  ginv_prime2 <- function(x) 0

  if (grepl("theta_var|psi_var", matrix)) {
    g <- log
    g_prime <- function(x) 1 / x
    ginv <- exp
    ginv_prime <- exp
    ginv_prime2 <- exp
  }

  if (grepl("theta_cor|theta_cov|psi_cor|psi_cov", matrix)) {
    g <- atanh
    g_prime <- function(x) 1 / (1 - x^2)
    ginv <- tanh
    ginv_prime <- function(x) 1 - tanh(x)^2
    ginv_prime2 <- function(x) -2 * safe_tanh(x) * (1 - safe_tanh(x)^2)
  }

  return(list(
    g = g,
    g_prime = g_prime,
    ginv = ginv,
    ginv_prime = ginv_prime,
    ginv_prime2 = ginv_prime2
  ))
}

inlavaanify_partable <- function(
  pt,
  dp = priors_for(),
  lavdata,
  lavoptions
) {
  nlevels <- lavdata@nlevels
  is_multilvl <- nlevels > 1
  if (is_multilvl) {
    pt$group <- pt$level
  }
  ngroups <- max(pt$group)
  std_ov <- lavoptions$std.ov
  std_lv <- lavoptions$std.lv
  pt$mat <- NA

  for (g in c(0, seq_len(ngroups))) {
    # Identify stuff
    if (is_multilvl) {
      ov.names <- lavdata@ov.names[[1]]
    } else {
      ov.names <- if (g == 0) NULL else lavdata@ov.names[[g]]
    }
    pt$mat[pt$group == g] <- mapply(
      partable_classify_sem_matrix,
      lhs = pt$lhs[pt$group == g],
      op = pt$op[pt$group == g],
      rhs = pt$rhs[pt$group == g],
      MoreArgs = list(ov.names = ov.names, std.ov = std_ov, std.lv = std_lv),
      SIMPLIFY = TRUE
    )
  }
  pt$mat <- unlist(pt$mat)

  # Add priors
  # Note: Possible to add own non-standard priors, but the evaluation of
  # prior_logdens() will return an error.
  user_prior <- pt$prior
  pt$prior <- mapply(
    partable_prior_from_row,
    matrix = pt$mat,
    lhs = pt$lhs,
    rhs = pt$rhs,
    op = pt$op,
    MoreArgs = list(dp = dp),
    USE.NAMES = FALSE
  )
  pt$prior[pt$free == 0L | duplicated(pt$free)] <- NA_character_
  if (!is.null(user_prior)) {
    where_user_prior <- user_prior != ""
    pt$prior[where_user_prior] <- user_prior[where_user_prior]
  }

  # Add transformations to unrestricted parameter space
  tmp <- lapply(pt$mat, partable_transform_funcs)
  pt$g <- lapply(tmp, `[[`, "g")
  pt$g_prime <- lapply(tmp, `[[`, "g_prime")
  pt$ginv <- lapply(tmp, `[[`, "ginv")
  pt$ginv_prime <- lapply(tmp, `[[`, "ginv_prime")
  pt$ginv_prime2 <- lapply(tmp, `[[`, "ginv_prime2")

  # Compute starting values in unrestricted space
  pt$parstart <- mapply(
    function(fun, val) fun(val),
    pt$g,
    pt$start,
    USE.NAMES = FALSE
  )

  # Add names
  pt$names <- mapply(paste0, pt$lhs, pt$op, pt$rhs)
  where_label <- pt$label != ""
  pt$names[where_label] <- pt$label[where_label]
  if (ngroups > 1) {
    prefix <- ifelse(is_multilvl, ".l", ".g")
    pt$names <- paste0(pt$names, prefix, pt$group)
    pt$names <- gsub(".g1|.l1", "", pt$names)
  }
  # Remove .l0 from names of defined params
  pt$names[pt$op == ":="] <- gsub(".l0", "", pt$names[pt$op == ":="])

  # Remove pt$group if multilevel (interacts with lav_partable_labels())
  if (is_multilvl) {
    pt$group <- NULL
  }

  # FIXME: Perhaps add a 'inlavaan_partable' class to this object
  as.list(pt)
}

test_that("partable_classify_sem_matrix classifies SEM entries correctly", {
  ov <- c("y1", "y2")

  # Loadings
  expect_equal(
    partable_classify_sem_matrix(
      "eta1",
      "=~",
      "y1",
      ov.names = ov,
      std.ov = FALSE,
      std.lv = FALSE
    ),
    "lambda"
  )

  # Observed residuals: variance vs correlation / covariance
  expect_equal(
    partable_classify_sem_matrix(
      "y1",
      "~~",
      "y1",
      ov.names = ov,
      std.ov = FALSE,
      std.lv = FALSE
    ),
    "theta_var"
  )
  expect_equal(
    partable_classify_sem_matrix(
      "y1",
      "~~",
      "y2",
      ov.names = ov,
      std.ov = TRUE,
      std.lv = FALSE
    ),
    "theta_cor"
  )
  expect_equal(
    partable_classify_sem_matrix(
      "y1",
      "~~",
      "y2",
      ov.names = ov,
      std.ov = FALSE,
      std.lv = FALSE
    ),
    "theta_cov"
  )

  # Latent (co)variances
  expect_equal(
    partable_classify_sem_matrix(
      "eta1",
      "~~",
      "eta1",
      ov.names = ov,
      std.ov = FALSE,
      std.lv = FALSE
    ),
    "psi_var"
  )
  expect_equal(
    partable_classify_sem_matrix(
      "eta1",
      "~~",
      "eta2",
      ov.names = ov,
      std.ov = FALSE,
      std.lv = TRUE
    ),
    "psi_cor"
  )
  expect_equal(
    partable_classify_sem_matrix(
      "eta1",
      "~~",
      "eta2",
      ov.names = ov,
      std.ov = FALSE,
      std.lv = FALSE
    ),
    "psi_cov"
  )

  # Residual–latent covariance
  expect_equal(
    partable_classify_sem_matrix(
      "y1",
      "~*~",
      "eta1",
      ov.names = ov,
      std.ov = FALSE,
      std.lv = FALSE
    ),
    "delta"
  )

  # Regressions
  expect_equal(
    partable_classify_sem_matrix(
      "eta2",
      "~",
      "eta1",
      ov.names = ov,
      std.ov = FALSE,
      std.lv = FALSE
    ),
    "beta"
  )

  # Intercepts / means
  expect_equal(
    partable_classify_sem_matrix(
      "y1",
      "~1",
      "",
      ov.names = ov,
      std.ov = FALSE,
      std.lv = FALSE
    ),
    "nu"
  )
  expect_equal(
    partable_classify_sem_matrix(
      "eta1",
      "~1",
      "",
      ov.names = ov,
      std.ov = FALSE,
      std.lv = FALSE
    ),
    "alpha"
  )

  # Thresholds
  expect_equal(
    partable_classify_sem_matrix(
      "y1",
      "|",
      "t1",
      ov.names = ov,
      std.ov = FALSE,
      std.lv = FALSE
    ),
    "tau"
  )

  # Defined parameters, constraints, fixed
  expect_equal(
    partable_classify_sem_matrix(
      "p1",
      ":=",
      "y1 + y2",
      ov.names = ov,
      std.ov = FALSE,
      std.lv = FALSE
    ),
    "defined"
  )
  expect_equal(
    partable_classify_sem_matrix(
      "y1",
      "==",
      "y2",
      ov.names = ov,
      std.ov = FALSE,
      std.lv = FALSE
    ),
    "constraint"
  )
  expect_equal(
    partable_classify_sem_matrix(
      "y1",
      "@",
      "1",
      ov.names = ov,
      std.ov = FALSE,
      std.lv = FALSE
    ),
    "fixed"
  )
})

test_that("partable_prior_from_row maps to dp entries correctly", {
  dp <- list(
    nu = "nu_prior",
    alpha = "alpha_prior",
    lambda = "lambda_prior",
    beta = "beta_prior",
    theta = "theta_prior",
    psi = "psi_prior",
    rho = "rho_prior",
    tau = "tau_prior"
  )

  dummy <- list(lhs = "eta1", rhs = "eta2", op = "~~")

  expect_equal(
    partable_prior_from_row("nu", dummy$lhs, dummy$rhs, dummy$op, dp),
    "nu_prior"
  )
  expect_equal(
    partable_prior_from_row("alpha", dummy$lhs, dummy$rhs, dummy$op, dp),
    "alpha_prior"
  )
  expect_equal(
    partable_prior_from_row("lambda", dummy$lhs, dummy$rhs, dummy$op, dp),
    "lambda_prior"
  )
  expect_equal(
    partable_prior_from_row("beta", dummy$lhs, dummy$rhs, dummy$op, dp),
    "beta_prior"
  )

  # theta_*: variance vs correlation
  expect_equal(
    partable_prior_from_row("theta_var", dummy$lhs, dummy$rhs, dummy$op, dp),
    "theta_prior"
  )
  expect_equal(
    partable_prior_from_row("theta_cov", dummy$lhs, dummy$rhs, dummy$op, dp),
    "rho_prior"
  )

  # psi_*: variance vs correlation
  expect_equal(
    partable_prior_from_row("psi_var", dummy$lhs, dummy$rhs, dummy$op, dp),
    "psi_prior"
  )
  expect_equal(
    partable_prior_from_row("psi_cor", dummy$lhs, dummy$rhs, dummy$op, dp),
    "rho_prior"
  )

  expect_equal(
    partable_prior_from_row("tau", dummy$lhs, dummy$rhs, dummy$op, dp),
    "tau_prior"
  )

  expect_true(is.na(
    partable_prior_from_row("unknown", dummy$lhs, dummy$rhs, dummy$op, dp)
  ))
})

test_that("safe_tanh rescales tanh and remains within (-1, 1)", {
  x <- c(-10, -1, 0, 1, 10)
  y <- safe_tanh(x)
  expect_equal(y, (1 - 1e-6) * tanh(x))
  expect_true(all(y > -1 & y < 1))
})

test_that("partable_transform_funcs returns appropriate transforms", {
  # Identity case
  tf_id <- partable_transform_funcs("lambda")
  x <- c(-1, 0, 2)
  expect_equal(tf_id$g(x), x)
  expect_equal(tf_id$ginv(x), x)
  expect_equal(tf_id$g_prime(x), 1)
  expect_equal(tf_id$ginv_prime(x), 1)
  expect_equal(tf_id$ginv_prime2(x), 0)

  # Variance: log / exp mapping
  tf_var <- partable_transform_funcs("theta_var")
  x_var <- c(0.5, 1, 2)
  gx <- tf_var$g(x_var)
  expect_equal(tf_var$ginv(gx), x_var, tolerance = 1e-10)
  expect_equal(tf_var$g_prime(x_var), 1 / x_var)
  expect_equal(tf_var$ginv_prime(gx), exp(gx))
  expect_equal(tf_var$ginv_prime2(gx), exp(gx))

  # Correlation: atanh / tanh mapping
  tf_cor <- partable_transform_funcs("psi_cor")
  r <- c(-0.5, 0, 0.4)
  gr <- tf_cor$g(r)
  expect_equal(tf_cor$ginv(gr), tanh(gr))
  expect_equal(tf_cor$ginv(gr), r, tolerance = 1e-10)
  expect_equal(tf_cor$g_prime(r), 1 / (1 - r^2))
  expect_equal(tf_cor$ginv_prime(gr), 1 - tanh(gr)^2)
})

test_that("inlavaanify_partable integrates with lavaan / blavaan objects", {
  skip_if_not_installed("lavaan")
  skip_if_not_installed("blavaan")

  HS.model <- '
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
  '

  fit <- lavaan::cfa(HS.model, lavaan::HolzingerSwineford1939, std.lv = TRUE)
  pt <- lavaan::parTable(fit)

  res <- inlavaanify_partable(
    pt = pt,
    dp = blavaan::dpriors(),
    lavdata = fit@Data,
    lavoptions = fit@Options
  )

  # Basic structure
  expect_type(res, "list")
  expect_true(all(
    c("lhs", "rhs", "op", "mat", "prior", "g", "ginv") %in% names(res)
  ))

  # Check some classifications exist and look reasonable
  expect_true(any(res$mat == "lambda"))
  expect_true(any(res$mat == "theta_var"))
  expect_true(any(res$mat == "psi_var"))

  # Names should be non-empty and length matches number of rows
  expect_equal(length(res$names), nrow(pt))
  expect_true(all(nzchar(res$names)))
})

# Compute the volume-slope correction gamma1_j for marginal approximations.
#
# @param j Index of the parameter (1..m).
# @param method One of "shortcut", "shortcut_fd", "hessian", or "none".
# @param theta_star Mode of the joint posterior (length m).
# @param Vscan Scan directions matrix (m x m), column j = Sigma[,j]/sqrt(Sigma[j,j]).
# @param L Lower-triangular Cholesky factor: L L^T = Sigma_theta.
# @param joint_lp_grad Function(theta) returning the gradient of the joint log-posterior.
# @param delta_outer Step size for outer finite difference (rate of change of Hessian).
# @param delta_inner Step size for inner finite difference (forward-diff Hessian diag).
# @param m Number of parameters.
# @return Scalar gamma1_j correction value.
compute_gamma1j <- function(j, method, theta_star, Vscan, L,
                            joint_lp_grad, delta_outer, delta_inner, m) {
  if (method == "none") return(0)

  vj <- Vscan[, j]
  th_plus <- theta_star + vj * delta_outer
  neg_grad <- function(x) -1 * joint_lp_grad(x)

  if (method == "hessian") { # nocov start
    # Full Hessian at two shifted points (central outer FD).
    # Cost: 4m gradient evaluations per parameter.
    th_minus <- theta_star - vj * delta_outer
    Htheta_plus  <- fast_jacobian(neg_grad, th_plus)
    Htheta_minus <- fast_jacobian(neg_grad, th_minus)

    Hz_plus  <- diag(crossprod(L, Htheta_plus  %*% L))
    Hz_minus <- diag(crossprod(L, Htheta_minus %*% L))
    dH_dz <- (Hz_plus - Hz_minus) / (2 * delta_outer)

    vHv_plus  <- as.numeric(crossprod(vj, Htheta_plus  %*% vj))
    vHv_minus <- as.numeric(crossprod(vj, Htheta_minus %*% vj))
    d_vHv <- (vHv_plus - vHv_minus) / (2 * delta_outer)

    gamma1j <- -0.5 * sum(dH_dz) + 0.5 * d_vHv

  } else if (method == "shortcut") {
    # Central-difference shortcut (Q2): full z-trace + Schur correction.
    # Inner: central FD with h = 1e-5.  Outer: forward FD with delta_outer.
    # Cost: 2m + 2 gradient evaluations per parameter.
    h <- 1e-5
    trace_Hz1 <- 0
    for (kk in seq_len(m)) {
      Lk <- L[, kk]
      g_fwd <- neg_grad(th_plus + Lk * h)
      g_bwd <- neg_grad(th_plus - Lk * h)
      trace_Hz1 <- trace_Hz1 + sum(Lk * (g_fwd - g_bwd)) / (2 * h)
    }
    full_trace <- -0.5 * (trace_Hz1 - m) / delta_outer

    # Schur correction: v_j' H(theta*) v_j = 1 by construction
    g_fwd_vj <- neg_grad(th_plus + vj * h)
    g_bwd_vj <- neg_grad(th_plus - vj * h)
    vHv_1 <- as.numeric(crossprod(vj, g_fwd_vj - g_bwd_vj)) / (2 * h)
    d_vHv <- (vHv_1 - 1) / delta_outer

    gamma1j <- full_trace + 0.5 * d_vHv
  } # nocov end

  else if (method == "shortcut_fd") { # nocov start
    # Forward-difference shortcut (Q2): full z-trace + Schur correction.
    # Cost: m + 2 gradient evaluations per parameter.
    g0 <- neg_grad(th_plus)
    trace_Hz1 <- 0
    for (kk in seq_len(m)) {
      Lk <- L[, kk]
      g_fwd <- neg_grad(th_plus + Lk * delta_inner)
      trace_Hz1 <- trace_Hz1 + sum(Lk * (g_fwd - g0)) / delta_inner
    }
    full_trace <- -0.5 * (trace_Hz1 - m) / delta_outer

    # Schur correction: v_j' H(theta*) v_j = 1 by construction
    g_vj <- neg_grad(th_plus + vj * delta_inner)
    vHv_1 <- as.numeric(crossprod(vj, g_vj - g0)) / delta_inner
    d_vHv <- (vHv_1 - 1) / delta_outer

    gamma1j <- full_trace + 0.5 * d_vHv
  } # nocov end

  gamma1j
}

pars_to_x <- function(theta, pt) {
  # Convert unrestricted theta-side parameters to lavaan-side parameters x.
  # Always receive UNPACKED theta and returns PACKED theta.
  if (is.null(pt) | missing(pt)) { # nocov
    cli_abort("Parameter table 'pt' must be provided.")
  }

  is_multilvl <- "level" %in% names(pt)
  if (is_multilvl) {
    pt$group <- pt$level
  }
  nG <- max(pt$group)
  idxfree <- pt$free > 0
  pars <- pt$parstart
  pars[idxfree] <- theta
  npt <- length(pars)
  xx <- x <- mapply(function(f, z) f(z), pt$ginv, pars)
  sd1sd2 <- rep(1, npt)
  jcb_mat <- NULL
  thidx <- integer(npt)
  thidx[pt$free > 0] <- seq_len(sum(pt$free > 0))

  # Now deal with covariances
  for (g in seq_len(nG)) {
    idxcov <- which(grepl("cov", pt$mat) & pt$group == g)
    for (j in idxcov) {
      X1 <- pt$lhs[j]
      X2 <- pt$rhs[j]
      where_varX1 <- which(
        pt$lhs == X1 & pt$op == "~~" & pt$rhs == X1 & pt$group == g
      )
      where_varX2 <- which(
        pt$lhs == X2 & pt$op == "~~" & pt$rhs == X2 & pt$group == g
      )

      sd1 <- sqrt(x[where_varX1])
      sd2 <- sqrt(x[where_varX2])
      rho <- x[j]
      x[j] <- rho * sd1 * sd2

      thidx1 <- thidx[where_varX1]
      thidx2 <- thidx[where_varX2]
      thidx3 <- thidx[j]
      jcb_mat <- rbind(jcb_mat, c(thidx1, thidx3, 0.5 * rho * sd1 * sd2))
      jcb_mat <- rbind(jcb_mat, c(thidx2, thidx3, 0.5 * rho * sd1 * sd2))
      sd1sd2[j] <- sd1 * sd2
    }
  }
  jcb_mat <- jcb_mat[jcb_mat[, 1] != 0 & jcb_mat[, 2] != 0, ]

  out <- x[pt$free > 0L & !duplicated(pt$free)]
  attr(out, "xcor") <- xx[pt$free > 0L & !duplicated(pt$free)]
  attr(out, "sd1sd2") <- sd1sd2[idxfree]
  attr(out, "jcb_mat") <- jcb_mat
  out
}

# Both functions below modified by Haavard Rue, April 2021
#
# Repo is https://github.com/thomasluu/sncdfinv/
#
# Copyright 2015-2016 Thomas Luu
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#
# plog: Computation of the Lambert W-function by Halley's Method.
#
# Initial guesses based on:
#
# D.A. Barry, J.-Y. Parlange, L. Li, H. Prommer, C.J. Cunningham, and F.
# Stagnitti. Analytical approximations for real values of the Lambert
# W-function. Mathematics and Computers in Simulation, 53(1):95-103, 2000.
#
# D.A. Barry, J.-Y. Parlange, L. Li, H. Prommer, C.J. Cunningham, and F.
# Stagnitti. Erratum to analytical approximations for real values of the Lambert
# W-function. Mathematics and computers in simulation, 59(6):543-543, 2002.
#
# GMRFLib_sn_Pinv:
#
# Based on: Luu, T; (2016) Fast and accurate parallel computation of quantile
# functions for random number generation. Doctoral thesis, UCL (University
# College London). https://discovery.ucl.ac.uk/1482128/

#' Fast Approximation of Skew-Normal Quantile Function
#'
#' A fast approximation of skew-normal quantiles using the high-performance
#' approximation algorithm from the INLA GMRFLib C source, and originally by
#' Thomas Luu (see details for reference).
#'
#' @details This function implements a high-performance approximation for the
#'   skew-normal quantile function based on the algorithm described by Luu
#'   (2016). The method uses a domain decomposition strategy to achieve high
#'   accuracy (\eqn{< 10^{-7}} relative error) without iterative numerical
#'   inversion.
#'
#'   The domain is split into two regions:
#' \itemize{
#'   \item **Tail Regions:** For extreme probabilities where \eqn{\vert u \vert} is large, the quantile
#'   is approximated using the Lambert W-function, \eqn{W(z)}, solving \eqn{z = \Phi(q)} via
#'   asymptotic expansion:
#'   \deqn{q \approx \sqrt{2 W\left(\frac{1}{2\pi (1-p)^2}\right)}}
#'   \item **Central Region:** For the main body of the distribution, the function
#'   uses a high-order Taylor expansion of the inverse error function around a
#'   carefully selected expansion point $x_0$:
#'   \deqn{\Phi^{-1}(p) \approx \sum_{k=0}^5 c_k (z - x_0)^k}
#' }
#'
#'   This approach is significantly faster than standard numerical inversion
#'   (e.g., `uniroot`) while maintaining sufficient precision for most
#'   statistical applications.
#'
#' @references Luu, T. (2016). *Fast and accurate parallel computation of
#'   quantile functions for random number generation* (Doctoral thesis). UCL
#'   (University College London). \url{https://discovery.ucl.ac.uk/1482128/}

#' @param p Vector of probabilities.
#' @param xi Location parameter (numeric vector).
#' @param omega Scale parameter (numeric vector).
#' @param alpha Shape parameter (numeric vector).
#'
#' @return Vector of quantiles.
#'
#' @examples
#' qsnorm_fast(c(0.025, 0.5, 0.975))
#' qsnorm_fast(c(0.025, 0.5, 0.975), xi = 2, omega = 0.5, alpha = 1)
#'
#' @export
qsnorm_fast <- function(p, xi = 0, omega = 1, alpha = 0) {
  # Recycling inputs to match length of p
  n <- length(p)
  if (length(xi) < n) {
    xi <- rep(xi, length.out = n)
  }
  if (length(omega) < n) {
    omega <- rep(omega, length.out = n)
  }
  if (length(alpha) < n) {
    alpha <- rep(alpha, length.out = n)
  }

  # Standardized quantile calculation
  z <- .qsn_std_fast(p, alpha)

  # Scale and shift
  return(xi + omega * z)
}

# Internal standardized quantile function
.qsn_std_fast <- function(u, a) {
  # Constants
  SQRT2 <- 1.4142135623730950488
  INV_SQRT_PI <- 0.564189583547756286948
  CONST_TOL <- 1.6448536269514722 # qnorm(0.99) / sqrt(2)
  TOL <- 0.01

  n <- length(u)
  res <- numeric(n)

  # Handle special cases for u
  res[u == 0] <- -Inf
  res[u == 1] <- Inf

  # ---- Symmetry relation: Q(-alpha, u) = -Q(alpha, 1 - u) ----
  # Fold negative alpha to positive so all downstream code assumes a >= 0.
  flip <- (a < 0)
  u[flip] <- 1 - u[flip]
  a[flip] <- -a[flip]

  # Pre-processing for a
  # If a == 1, u = sqrt(u)
  idx_a1 <- (a == 1)
  u[idx_a1] <- sqrt(u[idx_a1])

  # Standard normal quantile (z in C code)
  z <- stats::qnorm(u)

  A <- a  # a >= 0 after fold

  # --- Region Checks ---
  # right_limit = erf(const_tol / A)
  # erf(x) = 2 * pnorm(x * sqrt(2)) - 1
  right_limit <- 2 * stats::pnorm((CONST_TOL / A) * SQRT2) - 1

  # Mask for calculations (skip 0/1/Inf boundaries and a=0,1 special cases)
  mask <- (u > 0) & (u < 1) & (a != 0) & (a != 1)

  if (!any(mask)) { # nocov start
    res[!mask] <- z[!mask]
    res[flip] <- -res[flip]
    return(res)
  } # nocov end

  # Process masked elements
  u_m <- u[mask]
  A_m <- A[mask]
  z_m <- z[mask]
  rl_m <- right_limit[mask]

  final_vals <- z_m # Default to z, will overwrite

  # 1. Tail Region (Using Lambert W approximation 'plog')
  # a >= 0, so only positive tail check needed
  idx_tail <- (u_m > rl_m)

  if (any(idx_tail)) {
    final_vals[idx_tail] <- stats::qnorm((1 + u_m[idx_tail]) / 2)
  }

  # 2. Main Expansion Region
  idx_main <- !idx_tail

  if (any(idx_main)) {
    u_sub <- u_m[idx_main]
    A_sub <- A_m[idx_main]
    z_sub <- z_m[idx_main]

    # Expansion point 'x'
    val_x <- 0.5 - 0.31830988618379067154 * atan(A_sub)
    x_pt <- stats::qnorm(val_x)

    # Precompute terms
    expon <- exp(-0.5 * x_pt^2)
    errfn <- 1.0
    efder <- expon * 0.79788456080286535588 * A_sub / errfn

    # Coefficients c0 - c5
    c0 <- 0
    c1 <- expon / errfn
    c2 <- -expon * (efder + errfn * x_pt) / (2 * errfn^2)

    term_c3_num <- 3 *
      efder^2 +
      errfn^2 * (-1 + x_pt^2) +
      expon^2 +
      efder * (3 * errfn * x_pt)
    c3 <- 0.16666666666666666667 * expon * term_c3_num / (errfn^3)

    term_c4_num <- 15 *
      efder^3 +
      errfn^3 * x_pt * (-3 + x_pt^2) +
      6 * errfn * expon^2 * x_pt +
      efder^2 * (18 * errfn * x_pt) +
      efder * (errfn^2 * (-4 + 7 * x_pt^2) + expon^2 * (7 - A_sub^2))
    c4 <- -0.041666666666666666667 * expon * term_c4_num / (errfn^4)

    term_c5_num <- 105 *
      efder^4 +
      errfn^4 * (3 - 6 * x_pt^2 + x_pt^4) +
      5 * errfn^2 * expon^2 * (-2 + 5 * x_pt^2) +
      expon^4 * 7 +
      15 * efder^3 * (10 * errfn * x_pt) +
      efder *
        (5 *
          errfn^3 *
          x_pt *
          (-5 + 3 * x_pt^2) +
          10 * errfn * expon^2 * x_pt * (7 - A_sub^2)) +
      5 *
        efder^2 *
        (3 * errfn^2 * (-2 + 5 * x_pt^2) + expon^2 * (-3 * (-4 + A_sub^2)))
    c5 <- 0.0083333333333333333333 * expon * term_c5_num / (errfn^5)

    # Check "Deep Tail" sub-condition within main block using plog
    h <- 0.75 * abs(TOL / c5)^0.2
    left_limit <- x_pt - h

    idx_deep <- (z_sub < left_limit)
    res_sub <- numeric(length(z_sub))

    # -- Deep Tail Calculation (using plog) --
    if (any(idx_deep)) {
      u_d <- u_sub[idx_deep]
      A_d <- A_sub[idx_deep]

      C_PLOG <- 6.2831853071795864769 # 2*pi

      # a >= 0 guaranteed: -sqrt(2 * plog(1 / (2pi * u * a)) / (1 + a^2))
      arg <- 1 / (C_PLOG * u_d * A_d)
      res_sub[idx_deep] <- -sqrt(2 * .plog(arg) / (1 + A_d^2))
    }

    # -- Taylor Expansion Calculation --
    if (any(!idx_deep)) {
      h_val <- z_sub[!idx_deep] - x_pt[!idx_deep]

      poly <- c0 +
        h_val *
          (c1[!idx_deep] +
            h_val *
              (c2[!idx_deep] +
                h_val *
                  (c3[!idx_deep] +
                    h_val * (c4[!idx_deep] + h_val * c5[!idx_deep]))))
      res_sub[!idx_deep] <- poly
    }

    final_vals[idx_main] <- res_sub
  }

  res[mask] <- final_vals

  # Apply symmetry flip for originally-negative alpha
  res[flip] <- -res[flip]

  return(res)
}

# Lambert W-function Approximation (Halley's Method)
.plog <- function(x) {
  # Handle x <= 0 (though log(0) handled inside)
  res <- numeric(length(x))

  # Branch 1: x > 0
  idx_pos <- (x > 0)
  if (any(idx_pos)) {
    xx <- x[idx_pos]
    w0 <- log(1.2 * xx / log(2.4 * xx / log1p(2.4 * xx)))
    res[idx_pos] <- .plog_iterate(w0, xx)
  }

  # Branch 2: x <= 0 (Approximation for other branch/neg inputs)
  # In the context of qsn, x is derived from 1/prob, so usually x > 0.
  # But copying C logic for completeness.
  idx_neg <- !idx_pos
  if (any(idx_neg)) { # nocov start
    # Check for 0 specifically
    is_zero <- (x == 0)
    # If exactly 0, res is 0 (already set)

    # If not zero but <= 0 (unlikely for 1/prob, but strictly following C)
    # The C code uses sqrt(1 + 2.718... * x). If x is negative this might NaN.
    # Assuming x is small positive or handled by caller.
    # We will compute for non-zero negatives if they exist.
    valid_neg <- idx_neg & !is_zero
    if (any(valid_neg)) {
      xx <- x[valid_neg]
      v <- 1.4142135623730950488 * sqrt(1 + 2.7182818284590452354 * xx)
      N2 <- 10.242640687119285146 + 1.9797586132081854940 * v
      N1 <- 0.29289321881345247560 * (1.4142135623730950488 + N2)
      w0 <- -1.0 + v * (N2 + v) / (N2 + v + N1 * v)
      res[valid_neg] <- .plog_iterate(w0, xx)
    }
  } # nocov end

  return(res)
}

.plog_iterate <- function(w0, x) {
  # Halley's method iterations
  # 3 iterations is typically sufficient for machine precision
  for (i in 1:3) {
    e <- exp(w0)
    f <- w0 * e - x
    # Prevent div by zero if w0 approx -1 or -2, though unlikely in this domain
    denom <- f * (2.0 + w0) - (e + e) * (1.0 + w0)^2
    w1 <- w0 + ((f + f) * (1.0 + w0)) / denom
    w0 <- w1
  }
  return(w0)
}

test_that("Correct qsn_fast results", {
  skip_if_not_installed("sn")
  expect_equal(
    sn::qsn(c(0.025, 0.5, 0.975), xi = 2.09, omega = 0.19, alpha = 0.96),
    qsnorm_fast(c(0.025, 0.5, 0.975), xi = 2.09, omega = 0.19, alpha = 0.96),
    tolerance = 1e-3
  )
})

test_that("qsnorm_fast is correct for negative alpha", {
  skip_if_not_installed("sn")
  probs <- c(0.01, 0.1, 0.25, 0.5, 0.75, 0.9, 0.99)

  # Test with negative alpha (previously broken)
  q_fast <- qsnorm_fast(probs, xi = -0.44, omega = 0.25, alpha = -2.0)
  q_sn   <- sn::qsn(probs, xi = -0.44, omega = 0.25, alpha = -2.0)
  expect_equal(q_fast, q_sn, tolerance = 1e-3)

  # Verify symmetry: Q(-alpha, u) == -Q(alpha, 1-u)
  q_pos <- qsnorm_fast(1 - probs, xi = 0, omega = 1, alpha = 2.0)
  q_neg <- qsnorm_fast(probs, xi = 0, omega = 1, alpha = -2.0)
  expect_equal(q_neg, -q_pos, tolerance = 1e-6)
})

test_that("qsnorm_fast is monotonically non-decreasing", {
  u <- seq(0.01, 0.99, length = 500)
  # Positive alpha
  q1 <- qsnorm_fast(u, xi = 0, omega = 1, alpha = 1.5)
  expect_true(all(diff(q1) >= -1e-10))
  # Negative alpha
  q2 <- qsnorm_fast(u, xi = 0, omega = 1, alpha = -1.5)
  expect_true(all(diff(q2) >= -1e-10))
})

# library(microbenchmark)
# microbenchmark(
#   sn = sn::qsn(c(0.025, 0.5, 0.975), xi = 2.09, omega = 0.19, alpha = 0.96),
#   INLAvaan = qsnorm_fast(
#     c(0.025, 0.5, 0.975),
#     xi = 2.09,
#     omega = 0.19,
#     alpha = 0.96
#   ),
#   times = 5000
# )

# lavaan >= 0.7 renamed many exported functions'/generics' arguments, mostly
# dot.case -> snake_case (e.g. fit.measures -> fit_measures, cov.std ->
# cov_std) and a few by letter-case (GLIST -> glist). INLAvaan keeps its
# historic dot.case names as its own stable convention and adapts at the
# lavaan boundary. Because every affected function/generic also has "...",
# passing the wrong spelling is not an error: the value silently falls into
# "..." and the caller's override is *dropped* rather than applied. So --
# exactly like the renamed function *names* in lavaan-unexported.R -- the
# spelling the installed lavaan actually uses must be resolved from the
# target's formals once per session in .onLoad() (see zzz.R), never baked in
# at build/install time. This one resolver serves every lavaan argument
# INLAvaan sets: the standardizedSolution()/parameterEstimates() calls below
# and the fitMeasures()/fitmeasures() methods in method-fitmeasures.R.
#
# Each target maps its INLAvaan-canonical (dot.case) argument names to the
# spellings to probe for, newest (lavaan >= 0.7) first and oldest last --
# mirroring the alias vectors in lavaan_internal_aliases.
lavaan_argname_aliases <- list(
  standardizedSolution = list(
    cov.std = c("cov_std", "cov.std"),
    remove.eq = c("remove_eq", "remove.eq"),
    remove.ineq = c("remove_ineq", "remove.ineq"),
    remove.def = c("remove_def", "remove.def"),
    GLIST = c("glist", "GLIST")
  ),
  parameterEstimates = list(
    remove.eq = c("remove_eq", "remove.eq"),
    remove.system.eq = c("remove_system_eq", "remove.system.eq"),
    remove.ineq = c("remove_ineq", "remove.ineq"),
    remove.def = c("remove_def", "remove.def")
  ),
  fitMeasures = list(
    fit.measures = c("fit_measures", "fit.measures"),
    baseline.model = c("baseline_model", "baseline.model")
  )
)

# Filled in by resolve_lavaan_argnames(): one named character vector per
# target above, mapping each INLAvaan-canonical name to the spelling the
# installed lavaan expects. Consumed by call_lavaan() and, for fitMeasures,
# by register_fitmeasures_methods().
lavaan_argnames <- NULL

# Formal names of a lavaan target, whether it is a plain function
# (standardizedSolution, parameterEstimates) or an S4 generic (fitMeasures).
lavaan_target_formals <- function(target) {
  gen <- methods::getGeneric(target, mustFind = FALSE)
  fun <- if (!is.null(gen)) gen else get(target, envir = asNamespace("lavaan"))
  names(formals(fun))
}

resolve_lavaan_argnames <- function(ns) {
  resolved <- lapply(names(lavaan_argname_aliases), function(target) {
    fnames <- lavaan_target_formals(target)
    vapply(
      lavaan_argname_aliases[[target]],
      function(cands) {
        hit <- cands[cands %in% fnames]
        if (length(hit)) hit[[1L]] else cands[[length(cands)]]
      },
      character(1)
    )
  })
  names(resolved) <- names(lavaan_argname_aliases)
  assign("lavaan_argnames", resolved, envir = ns)
  invisible(NULL)
}

# Call lavaan's `target` (standardizedSolution or parameterEstimates),
# renaming any INLAvaan-canonical dot.case argument in `...` to whichever
# spelling the installed lavaan expects. `get(target, ...)` fetches lavaan's
# own function, not INLAvaan's same-named wrapper.
call_lavaan <- function(target, ...) {
  args <- list(...)
  map <- lavaan_argnames[[target]]
  nm <- names(args)
  hit <- nm %in% names(map)
  nm[hit] <- map[nm[hit]]
  names(args) <- nm
  do.call(get(target, envir = asNamespace("lavaan")), args)
}

#' Simulate Datasets from the Generative Model
#'
#' Generate complete synthetic datasets from a fitted INLAvaan model. For each
#' simulation, a single parameter vector is drawn (from the posterior or prior),
#' and then `sample.nobs` observations are generated from the model-implied
#' distribution at that parameter value.
#'
#' @details
#' This function is designed for tasks that require **full replicate datasets**
#' from a single parameter draw, such as simulation-based calibration (SBC) and
#' posterior predictive p-values. It differs from [sampling()] which generates
#' one observation per parameter draw (useful for prior/posterior predictive
#' density overlays).
#'
#' For each simulation \eqn{s = 1, \ldots, S}:
#' 1. Draw \eqn{\boldsymbol\theta^{(s)}} from the posterior (or prior).
#' 2. Compute the model-implied covariance
#'    \eqn{\boldsymbol\Sigma(\boldsymbol\theta^{(s)})}.
#'    If it is not positive-definite, reject and redraw.
#' 3. Generate a dataset of `sample.nobs` rows from
#'    \eqn{N(\boldsymbol\mu(\boldsymbol\theta^{(s)}),\,
#'    \boldsymbol\Sigma(\boldsymbol\theta^{(s)}))}.
#'
#' Parameter draws reuse the same internal machinery as [sampling()]
#' (`sample_params_prior` / `sample_params_posterior`), so the prior
#' specification is consistent.
#'
#' @param object An object of class [INLAvaan].
#' @param nsim Number of replicate datasets to generate (default 1).
#' @param seed Optional random seed (passed to [set.seed()]).
#' @param sample.nobs Number of observations per dataset. Defaults to the
#'   sample size of the original data.
#' @param prior Logical. When `TRUE`, parameters are drawn from the prior;
#'   when `FALSE` (default), from the posterior.
#' @param samp_copula Logical. When `TRUE` (default) and `prior = FALSE`,
#'   posterior parameter draws use the copula method. Ignored when
#'   `prior = TRUE`.
#' @param silent Logical. When `TRUE`, suppresses the informational message
#'   about rejected non-PD draws. Default `FALSE`.
#' @param ... Additional arguments (currently unused).
#'
#' @returns A list of length `nsim`. Each element is a data frame with
#'   `sample.nobs` rows and two attributes:
#'   * `"truth"` — named numeric vector of lavaan-side (x-space, constrained)
#'     parameter values used to generate the dataset.
#'   * `"truth_theta"` — named numeric vector of the corresponding unconstrained
#'     (theta-space) parameter values.
#'
#' @seealso [sampling()] for single-observation draws from the predictive
#'   distribution (prior/posterior predictive checks).
#'
#' @example inst/examples/ex-simulate.R
#' @importFrom stats simulate rgamma
#' @name simulate
#' @rdname simulate
#' @aliases simulate,INLAvaan-method
#' @export
setMethod(
  "simulate",
  "INLAvaan",
  function(
    object,
    nsim = 1L,
    seed = NULL,
    sample.nobs = NULL,
    prior = FALSE,
    samp_copula = TRUE,
    silent = FALSE,
    ...
  ) {
    if (!is.null(seed)) {
      set.seed(seed)
    }
    nsim <- as.integer(nsim)

    int <- object@external$inlavaan_internal
    pt <- int$partable
    lavmodel <- int$lavmodel

    xnames <- pt$names[pt$free > 0 & !duplicated(pt$free)]
    n <- if (is.null(sample.nobs)) object@SampleStats@ntotal else sample.nobs

    # Draw a generous batch of parameters upfront
    oversample <- nsim * 5L
    if (isTRUE(prior)) {
      samp <- sample_params_prior(int, oversample)
    } else {
      samp <- sample_params_posterior(int, oversample, samp_copula)
    }
    colnames(samp$x_samp) <- xnames
    colnames(samp$theta_samp) <- xnames

    results <- vector("list", nsim)
    collected <- 0L
    idx <- 0L # index into pre-drawn batch
    max_attempts <- nsim * 20L
    attempts <- 0L

    while (collected < nsim && attempts < max_attempts) {
      idx <- idx + 1L
      attempts <- attempts + 1L

      # Refill batch if exhausted
      if (idx > nrow(samp$x_samp)) {
        # nocov start
        if (isTRUE(prior)) {
          samp <- sample_params_prior(int, oversample)
        } else {
          samp <- sample_params_posterior(int, oversample, samp_copula)
        }
        colnames(samp$x_samp) <- xnames
        colnames(samp$theta_samp) <- xnames
        idx <- 1L
      } # nocov end

      x_draw <- samp$x_samp[idx, ]
      theta_draw <- samp$theta_samp[idx, ]

      # Check PD of model-implied covariance
      lavmodel_x <- lavaan::lav_model_set_parameters(
        lavmodel,
        as.numeric(x_draw)
      )
      implied <- lavaan::lav_model_implied(lavmodel_x)

      all_pd <- TRUE
      for (g in seq_along(implied$cov)) {
        eigs <- eigen(
          implied$cov[[g]],
          symmetric = TRUE,
          only.values = TRUE
        )$values
        if (any(eigs < 1e-10)) {
          all_pd <- FALSE
          break
        }
      }
      if (!all_pd) {
        next
      }

      # Generate data via lavaan's simulateData using the partable
      pt_sim <- lavaan::lav_partable_complete(lavaan::partable(object))
      pt_sim$est[pt_sim$free > 0] <- as.numeric(x_draw)

      dat <- tryCatch(
        lavaan::simulateData(pt_sim, sample.nobs = n),
        error = function(e) NULL
      )
      if (is.null(dat)) {
        # nocov
        next
      }

      # Without a mean structure lavaan::simulateData() centres the columns at
      # zero (there are no intercepts in the partable). INLAvaan integrates the
      # saturated means out under a flat prior, so the posterior of the mean is
      # N(ybar, Sigma / n). Draw one such mean vector for this replicate (all
      # its rows share the same theta, hence the same mean) and shift the data
      # onto the observed scale, mirroring sampling(type = "observed").
      if (
        !isTRUE(lavmodel@meanstructure) &&
          !isTRUE(prior) &&
          lavmodel@ngroups == 1L
      ) {
        ybar <- colMeans(int$lavdata@X[[1L]], na.rm = TRUE)
        names(ybar) <- int$lavdata@ov.names[[1L]]
        mu <- ybar
        if (marginalised_means_active(lavmodel)) {
          Sig <- implied$cov[[1L]]
          ch <- tryCatch(chol(Sig), error = function(e) NULL) # nocov
          if (!is.null(ch)) {
            n_fit <- nrow(int$lavdata@X[[1L]])
            mu <- ybar +
              as.numeric(crossprod(ch, stats::rnorm(length(ybar)))) /
                sqrt(n_fit)
          }
        }
        for (v in intersect(names(mu), colnames(dat))) {
          if (is.numeric(dat[[v]])) {
            dat[[v]] <- dat[[v]] + mu[[v]]
          }
        }
      }

      collected <- collected + 1L
      attr(dat, "truth") <- x_draw
      attr(dat, "truth_theta") <- theta_draw
      results[[collected]] <- dat
    }

    rejected <- attempts - collected
    if (rejected > 0L && !isTRUE(silent)) {
      # nocov start
      rej_pct <- round(100 * rejected / attempts, 1)
      cli_inform(
        "simulate: {rejected} of {attempts} draw{?s} ({rej_pct}%) rejected (non-PD model-implied covariance)."
      )
    } # nocov end

    if (collected < nsim) {
      # nocov start
      cli_warn(c(
        "Rejection sampling fell short of the requested {nsim} simulations.",
        "i" = "Only {collected} of {nsim} obtained after {attempts} attempts.",
        "i" = "Consider using more informative priors."
      ))
      if (collected == 0L) {
        cli_abort("No valid draws obtained. Priors may be too vague.")
      }
      results <- results[seq_len(collected)]
    } # nocov end

    results
  }
)

get_SEM_param_matrix <- function(x, mat, lavmodel) {
  nG <- lavmodel@ngroups
  lavmodel_x <- lavaan::lav_model_set_parameters(lavmodel, x)

  GLIST <- Map(
    function(mat, dn) {
      rownames(mat) <- dn[[1]]
      colnames(mat) <- dn[[2]]
      mat
    },
    lavmodel_x@GLIST,
    lavmodel_x@dimNames
  )

  uniq_names <- unique(names(GLIST))
  k <- length(uniq_names)
  out <- vector("list", nG)
  for (g in seq_len(nG)) {
    idx <- ((g - 1) * k + 1):(g * k)
    out[[g]] <- GLIST[idx]
    names(out[[g]]) <- uniq_names
  }

  if (mat == "all" | mat == "GLIST") {
    return(out)
  } else {
    return(lapply(out, function(glist) glist[[mat]]))
  }
}

# For factor scores, there is the plugin marginal_method and sampling
# marginal_method.
#
# For plugin marginal_method, eta | y ~ N(mu(theta, y), V(theta)), where
# mu(theta,y) = E(eta | y,theta) = Phi Lambda Sigma^{-1} y
# V(theta) = Phi - Phi Lambda' Sigma^{-1} Lambda Phi
# Phi = (I - B)^{-1} Psi (I - B')^{-1}'
#
# For sampling marginal_method just sample from the above distribution.

# Helper: build data matrices from newdata, reusing metadata from lavdata
build_newdata <- function(newdata, lavdata) {
  nG <- lavdata@ngroups
  grp <- lavdata@group
  has_group <- length(grp) > 0L && nzchar(grp)

  if (has_group) {
    group_labels <- lavdata@group.label
    groups_in_data <- as.character(newdata[[grp]])
    new_X <- vector("list", nG)
    new_nobs <- vector("list", nG)
    for (g in seq_len(nG)) {
      rows_g <- which(groups_in_data == group_labels[g])
      ov_names_g <- lavdata@ov.names[[g]]
      new_X[[g]] <- as.matrix(newdata[rows_g, ov_names_g, drop = FALSE])
      new_nobs[[g]] <- length(rows_g)
    }
  } else {
    ov_names <- lavdata@ov.names[[1L]]
    new_X <- list(as.matrix(newdata[, ov_names, drop = FALSE]))
    new_nobs <- list(nrow(newdata))
  }

  list(
    X = new_X,
    ngroups = nG,
    group.label = if (has_group) lavdata@group.label else character(0),
    nobs = new_nobs
  )
}

# nocov start
# Helper: compute factor scores for multilevel blocks.
# Returns an n x nfac matrix of factor scores.
compute_fs_ml <- function(
  data_block,
  VETA_b,
  LAMBDA_b,
  Sigma_hat_b,
  Sigma_inv_b,
  EETA_b,
  EY_b
) {
  n <- nrow(data_block)
  nfac <- ncol(VETA_b)
  if (nfac == 0L) {
    return(matrix(0, n, nfac))
  }

  Yc <- t(t(data_block) - EY_b)
  FSC <- VETA_b %*% t(LAMBDA_b) %*% Sigma_inv_b
  t(FSC %*% t(Yc) + EETA_b)
}
# nocov end

# nocov start
# Impute missing values in multilevel data by sampling from the conditional
# posterior: y_mis | y_obs, theta ~ N(mu_cond, Sigma_cond).
# Returns the data matrix with NAs replaced by posterior draws.
impute_ml_data <- function(
  yg,
  lavimplied,
  g,
  nlevels,
  ov_names_block,
  lavdata
) {
  if (!any(is.na(yg))) {
    return(yg)
  }

  p <- ncol(yg)
  var_names <- colnames(yg)
  if (is.null(var_names)) {
    var_names <- lavdata@ov.names[[g]]
  }

  block_w <- (g - 1) * nlevels + 1
  block_b <- (g - 1) * nlevels + 2

  Sigma_w <- lavimplied$cov[[block_w]]
  Sigma_b <- lavimplied$cov[[block_b]]

  nw <- ov_names_block[[block_w]]
  if (length(nw) == nrow(Sigma_w)) {
    rownames(Sigma_w) <- colnames(Sigma_w) <- nw
  }
  nb <- ov_names_block[[block_b]]
  if (length(nb) == nrow(Sigma_b)) {
    rownames(Sigma_b) <- colnames(Sigma_b) <- nb
  }

  Sigma_y <- matrix(0, p, p, dimnames = list(var_names, var_names))

  vn_w <- rownames(Sigma_w)
  w_in_data <- match(vn_w, var_names)
  w_keep <- !is.na(w_in_data)
  if (any(w_keep)) {
    idx <- w_in_data[w_keep]
    Sigma_y[idx, idx] <- Sigma_y[idx, idx] +
      Sigma_w[w_keep, w_keep, drop = FALSE]
  }

  vn_b <- rownames(Sigma_b)
  b_in_data <- match(vn_b, var_names)
  b_keep <- !is.na(b_in_data)
  if (any(b_keep)) {
    idx <- b_in_data[b_keep]
    Sigma_y[idx, idx] <- Sigma_y[idx, idx] +
      Sigma_b[b_keep, b_keep, drop = FALSE]
  }

  mu_y <- rep(0, p)
  names(mu_y) <- var_names
  if (!is.null(lavimplied$mean)) {
    mu_b <- as.numeric(lavimplied$mean[[block_b]])
    names(mu_b) <- nb
    b_match <- intersect(var_names, nb)
    mu_y[match(b_match, var_names)] <- mu_b[b_match]
  }

  na_mat <- is.na(yg)
  patterns <- apply(na_mat, 1, function(r) paste(which(r), collapse = ","))
  unique_patterns <- unique(patterns[patterns != ""])

  for (pat in unique_patterns) {
    mis_idx <- as.integer(strsplit(pat, ",")[[1]])
    obs_idx <- setdiff(seq_len(p), mis_idx)
    case_rows <- which(patterns == pat)

    Sigma_oo <- Sigma_y[obs_idx, obs_idx, drop = FALSE]
    Sigma_mo <- Sigma_y[mis_idx, obs_idx, drop = FALSE]
    Sigma_mm <- Sigma_y[mis_idx, mis_idx, drop = FALSE]

    A <- Sigma_mo %*% solve(Sigma_oo)
    Sigma_cond <- Sigma_mm - A %*% t(Sigma_mo)
    Sigma_cond <- (Sigma_cond + t(Sigma_cond)) / 2
    chol_cond <- t(chol(Sigma_cond))

    n_mis <- length(mis_idx)
    n_cases <- length(case_rows)

    y_obs_c <- yg[case_rows, obs_idx, drop = FALSE] -
      matrix(
        mu_y[obs_idx],
        nrow = n_cases,
        ncol = length(obs_idx),
        byrow = TRUE
      )
    mu_cond <- matrix(
      mu_y[mis_idx],
      nrow = n_cases,
      ncol = n_mis,
      byrow = TRUE
    ) +
      y_obs_c %*% t(A)

    Z <- matrix(rnorm(n_cases * n_mis), nrow = n_mis, ncol = n_cases)
    yg[case_rows, mis_idx] <- mu_cond + t(chol_cond %*% Z)
  }
  yg
}
# nocov end

# nocov start
# Compute cluster random effects (BLUP) directly from (imputed) data,
# bypassing the precomputed YLp which may contain NAs.
compute_ml_ranef <- function(y_g, Lp, decomp) {
  nclusters <- Lp$nclusters[[2]]
  cluster_idx <- Lp$cluster.idx[[2]]
  ov_idx <- Lp$ov.idx[[1]]
  between_idx <- Lp$between.idx[[2]]
  nvar <- length(ov_idx)

  mu_y <- decomp$mu.w + decomp$mu.b
  MB.j <- matrix(0, nclusters, nvar)

  has_between <- length(between_idx) > 0L
  if (has_between) {
    sigma_1 <- cbind(decomp$sigma.yz, decomp$sigma.b)
    mu_all <- c(decomp$mu.z, mu_y)
  } else {
    sigma_1 <- decomp$sigma.b
    mu_all <- mu_y
  }

  for (cl in seq_len(nclusters)) {
    obs_in_cl <- which(cluster_idx == cl)
    nj <- length(obs_in_cl)
    ybar <- colMeans(y_g[obs_in_cl, ov_idx, drop = FALSE])

    if (has_between) {
      b_vals <- y_g[obs_in_cl[1L], between_idx, drop = TRUE]
      b_j <- c(b_vals, ybar)
      sigma_j <- decomp$sigma.w + nj * decomp$sigma.b
      omega_j <- rbind(
        cbind(decomp$sigma.zz, t(decomp$sigma.yz)),
        cbind(decomp$sigma.yz, (1 / nj) * sigma_j)
      )
    } else {
      b_j <- ybar
      omega_j <- (1 / nj) * (decomp$sigma.w + nj * decomp$sigma.b)
    }

    omega_j_inv <- solve(omega_j)
    MB.j[cl, ] <- as.numeric(
      decomp$mu.b + sigma_1 %*% omega_j_inv %*% (b_j - mu_all)
    )
  }
  MB.j
}
# nocov end

#' @exportS3Method predict inlavaan_internal
#' @keywords internal
predict.inlavaan_internal <- function(
  object,
  type = c("lv", "yhat", "ov", "ypred", "ydist", "ymis", "ovmis"),
  newdata = NULL,
  level = 1L,
  nsamp = 250,
  ymis_only = FALSE,
  summary = FALSE,
  ...
) {
  type <- match.arg(type)
  # Aliases: "ov"/"yhat" -> "yhat"; "ypred"/"ydist" -> "ypred";
  #          "ymis"/"ovmis" -> "ymis"
  if (type == "ov") {
    type <- "yhat"
  }
  if (type == "ydist") {
    type <- "ypred"
  }
  if (type == "ovmis") {
    type <- "ymis"
  }

  theta_star <- object$theta_star
  Sigma_theta <- object$Sigma_theta
  marginal_method <- object$marginal_method
  approx_data <- object$approx_data
  pt <- object$partable
  lavmodel <- object$lavmodel
  lavdata <- object$lavdata
  nlevels <- lavdata@nlevels

  # Error early: ymis does not support newdata
  if (type == "ymis" && !is.null(newdata)) {
    cli_abort("Type {.val ymis} does not support {.arg newdata}.")
  }

  # Multilevel restrictions
  if (is_multilevel(lavdata)) {
    # nocov start
    if (!is.null(newdata)) {
      cli_abort("{.arg newdata} is not supported for multilevel models.")
    }
    if (!level %in% c(1L, 2L)) {
      cli_abort("{.arg level} must be {.val 1} or {.val 2}.")
    }
  } # nocov end

  # Handle newdata: rebuild lavdata matrices
  if (!is.null(newdata)) {
    new_ld <- build_newdata(newdata, lavdata)
    y <- new_ld$X
    nG <- new_ld$ngroups
    group_labels <- new_ld$group.label
    nobs_out <- new_ld$nobs
  } else {
    y <- lavdata@X
    nG <- lavdata@ngroups
    group_labels <- lavdata@group.label
    nobs_out <- lavdata@nobs
  }

  # Saturated means for models without a mean structure: the implied mean
  # does not exist, but the conditioning kernels below still need one. The
  # sample means of the *fitted* data are the correct plug-in (they are
  # the saturated means the fit conditions on), also for newdata.
  ybar_fit <- lapply(lavdata@X, colMeans)

  samp <- sample_params(
    theta_star = theta_star,
    Sigma_theta = Sigma_theta,
    method = marginal_method,
    approx_data = approx_data,
    pt = pt,
    lavmodel = lavmodel,
    nsamp = nsamp
  )
  x_samp <- samp$x_samp

  # ---- type = "lv": Posterior draws of latent variable scores ----
  if (type == "lv") {
    if (is_multilevel(lavdata)) {
      # nocov start
      # ---- Multilevel path: use lavaan internals ----
      lavsamplestats <- object$lavsamplestats

      # Pre-compute per-block ov names for imputation
      ov_names_block <- vector("list", lavmodel@nblocks)
      for (b in seq_len(lavmodel@nblocks)) {
        g_b <- ceiling(b / nlevels)
        ov_all <- lavdata@ov.names[[g_b]]
        ov_names_block[[b]] <- unique(
          pt$lhs[
            pt$block == b &
              pt$op == "~~" &
              pt$lhs == pt$rhs &
              pt$lhs %in% ov_all
          ]
        )
      }

      # Helper: get LV names from the psi dimNames for a given block
      get_lv_names <- function(lavmodel_x, block) {
        nmat <- lavmodel_x@nmat
        mm <- seq_len(nmat[block]) + cumsum(c(0, nmat))[block]
        psi_pos <- which(names(lavmodel_x@GLIST[mm]) == "psi")
        lavmodel_x@dimNames[[mm[psi_pos]]][[1]]
      }

      sample_lv_ml <- function(xx) {
        lavmodel_x <- lavaan::lav_model_set_parameters(lavmodel, xx)
        lavimplied <- lavaan::lav_model_implied(lavmodel_x)

        LAMBDA <- lavaan___lav_model_lambda(lavmodel = lavmodel_x)
        VETA <- lavaan___lav_model_veta(lavmodel = lavmodel_x)
        EETA <- lavaan___lav_model_eeta(
          lavmodel = lavmodel_x,
          lavsamplestats = lavsamplestats
        )
        EY <- lavaan___lav_model_ey(
          lavmodel = lavmodel_x,
          lavsamplestats = lavsamplestats
        )
        Sigma.hat <- lavimplied$cov
        Sigma.inv <- lapply(Sigma.hat, ginv_base)

        out <- vector("list", nG)
        names(out) <- group_labels

        for (g in seq_len(nG)) {
          b <- (g - 1) * nlevels + level

          # Impute missing data from conditional posterior, then decompose
          y_g <- impute_ml_data(
            y[[g]],
            lavimplied,
            g,
            nlevels,
            ov_names_block,
            lavdata
          )

          Lp <- lavdata@Lp[[g]]
          group.idx <- (g - 1) * nlevels + seq_len(nlevels)
          implied.group <- lapply(lavimplied, function(x) x[group.idx])

          # positional: first argument is Lp in lavaan < 0.7 but lp in >= 0.7
          decomp <- lavaan___lav_mvnorm_cluster_implied22l(Lp, implied.group)
          MB.j <- compute_ml_ranef(y_g, Lp, decomp)

          ov.idx <- Lp$ov.idx

          if (level == 1L) {
            data.obs.g <- y_g[, ov.idx[[1]], drop = FALSE] -
              MB.j[Lp$cluster.idx[[2]], , drop = FALSE]
          } else {
            # level == 2L
            Data.B <- matrix(0, nrow = nrow(MB.j), ncol = ncol(y_g))
            Data.B[, ov.idx[[1]]] <- MB.j
            between.idx <- Lp$between.idx[[2 * g]]
            if (length(between.idx) > 0L) {
              unique_rows <- match(
                seq_len(Lp$nclusters[[2]]),
                Lp$cluster.idx[[2]]
              )
              Data.B[, between.idx] <-
                y_g[unique_rows, between.idx, drop = FALSE]
            }
            data.obs.g <- Data.B[, ov.idx[[2]], drop = FALSE]
          }

          VETA.g <- VETA[[b]]
          EETA.g <- EETA[[b]]
          LAMBDA.g <- LAMBDA[[b]]
          EY.g <- EY[[b]]
          Sigma.inv.g <- Sigma.inv[[b]]

          FS.g <- compute_fs_ml(
            data.obs.g,
            VETA.g,
            LAMBDA.g,
            Sigma.hat[[b]],
            Sigma.inv.g,
            EETA.g,
            EY.g
          )

          # Replace dummy LV columns with data (level 1 only, per lavaan)
          if (level == 1L) {
            if (length(lavmodel_x@ov.y.dummy.lv.idx[[b]]) > 0L) {
              FS.g[, lavmodel_x@ov.y.dummy.lv.idx[[b]]] <-
                data.obs.g[, lavmodel_x@ov.y.dummy.ov.idx[[b]], drop = FALSE]
            }
            if (length(lavmodel_x@ov.x.dummy.lv.idx[[b]]) > 0L) {
              FS.g[, lavmodel_x@ov.x.dummy.lv.idx[[b]]] <-
                data.obs.g[, lavmodel_x@ov.x.dummy.ov.idx[[b]], drop = FALSE]
            }
          }

          colnames(FS.g) <- get_lv_names(lavmodel_x, b)
          out[[g]] <- FS.g
        }

        if (nG == 1L) {
          out <- out[[1L]]
        } else {
          out <- do.call(
            rbind,
            Map(function(g, df) data.frame(group = g, df), names(out), out)
          )
        }
        rownames(out) <- NULL
        out
      }

      out <- vector("list", nsamp)
      cli_progress_bar(
        "Sampling latent variables (multilevel)",
        total = nsamp,
        clear = FALSE
      )
      for (i in seq_len(nsamp)) {
        out[[i]] <- sample_lv_ml(x_samp[i, ])
        cli_progress_update()
      }
      cli_progress_done()
    } else {
      # nocov end
      # ---- Single-level path: full posterior draw ----
      sample_lv <- function(xx) {
        GLIST <- get_SEM_param_matrix(xx, "all", lavmodel)
        out <- vector("list", nG)
        names(out) <- group_labels
        for (g in seq_len(nG)) {
          glist <- GLIST[[g]]
          Lambda <- glist$lambda
          Psi <- glist$psi
          Theta <- glist$theta
          B <- glist$beta
          alpha <- glist$alpha

          if (is.null(alpha)) {
            alpha <- 0
          }

          if (is.null(B)) {
            Phi <- Psi
            front <- Lambda
          } else {
            IminB_inv <- solve(diag(nrow(B)) - B)
            Phi <- IminB_inv %*% Psi %*% t(IminB_inv)
            front <- Lambda %*% IminB_inv
          }

          Sigmay_inv <- solve(front %*% Psi %*% t(front) + Theta)
          PhiLtSinv <- Phi %*% t(Lambda) %*% Sigmay_inv

          # E(eta | y) = alpha + Phi Lambda' Sigma^{-1} (y - mu_y): centre
          # by the implied mean, or the saturated means without one
          alpha_vec <- rep_len(as.numeric(alpha), ncol(front))
          mu_y <- if (!is.null(glist$nu)) {
            as.numeric(glist$nu + front %*% alpha_vec)
          } else {
            ybar_fit[[g]]
          }
          yc <- sweep(y[[g]], 2L, mu_y)
          mu_eta <- t(as.numeric(alpha) + PhiLtSinv %*% t(yc))

          V_eta <- Phi - PhiLtSinv %*% Lambda %*% Phi
          chol_V <- t(chol(V_eta))

          n_obs <- nrow(mu_eta)
          nlv <- ncol(mu_eta)
          Z <- matrix(rnorm(n_obs * nlv), nrow = nlv, ncol = n_obs)
          outg <- mu_eta + t(chol_V %*% Z)

          out[[g]] <- outg
        }

        if (nG == 1L) {
          colnames(out[[1L]]) <- colnames(Psi)
          out <- out[[1L]]
        } else {
          out <- do.call(
            rbind,
            Map(function(g, df) data.frame(group = g, df), names(out), out)
          )
          colnames(out)[-1] <- colnames(Psi)
        }
        rownames(out) <- NULL
        out
      }

      out <- vector("list", nsamp)
      cli_progress_bar(
        "Sampling latent variables",
        total = nsamp,
        clear = FALSE
      )
      for (i in seq_len(nsamp)) {
        out[[i]] <- sample_lv(x_samp[i, ])
        cli_progress_update()
      }
      cli_progress_done()
    }

    # ---- type = "yhat": Predicted means E(y | eta, theta) ----
    # ---- type = "ypred": Predicted values y = E(y|eta,theta) + eps ----
  } else if (type %in% c("yhat", "ypred")) {
    add_noise <- (type == "ypred")

    if (nlevels > 1L) {
      # nocov start
      # ---- Multilevel yhat/ypred ----
      lavsamplestats <- object$lavsamplestats

      # Pre-compute per-block ov names for imputation
      ov_names_block <- vector("list", lavmodel@nblocks)
      for (b in seq_len(lavmodel@nblocks)) {
        g_b <- ceiling(b / nlevels)
        ov_all <- lavdata@ov.names[[g_b]]
        ov_names_block[[b]] <- unique(
          pt$lhs[
            pt$block == b &
              pt$op == "~~" &
              pt$lhs == pt$rhs &
              pt$lhs %in% ov_all
          ]
        )
      }

      sample_yhat_ml <- function(xx) {
        lavmodel_x <- lavaan::lav_model_set_parameters(lavmodel, xx)
        lavimplied <- lavaan::lav_model_implied(lavmodel_x)

        LAMBDA <- lavaan___lav_model_lambda(lavmodel = lavmodel_x)
        VETA <- lavaan___lav_model_veta(lavmodel = lavmodel_x)
        EETA <- lavaan___lav_model_eeta(
          lavmodel = lavmodel_x,
          lavsamplestats = lavsamplestats
        )
        EY <- lavaan___lav_model_ey(
          lavmodel = lavmodel_x,
          lavsamplestats = lavsamplestats
        )
        Sigma.hat <- lavimplied$cov
        Sigma.inv <- lapply(Sigma.hat, ginv_base)

        nmat <- lavmodel_x@nmat

        out <- vector("list", nG)
        names(out) <- group_labels

        for (g in seq_len(nG)) {
          b_w <- (g - 1) * nlevels + 1 # within block
          b_b <- (g - 1) * nlevels + 2 # between block

          # Impute missing data from conditional posterior, then decompose
          y_g <- impute_ml_data(
            y[[g]],
            lavimplied,
            g,
            nlevels,
            ov_names_block,
            lavdata
          )

          Lp <- lavdata@Lp[[g]]
          group.idx <- (g - 1) * nlevels + seq_len(nlevels)
          implied.group <- lapply(lavimplied, function(x) x[group.idx])

          # positional: first argument is Lp in lavaan < 0.7 but lp in >= 0.7
          decomp <- lavaan___lav_mvnorm_cluster_implied22l(Lp, implied.group)
          MB.j <- compute_ml_ranef(y_g, Lp, decomp)

          ov.idx <- Lp$ov.idx
          n_obs <- nrow(y_g)
          p <- ncol(y_g)

          # --- Within factor scores ---
          data.w <- y_g[, ov.idx[[1]], drop = FALSE] -
            MB.j[Lp$cluster.idx[[2]], , drop = FALSE]
          eta.w <- compute_fs_ml(
            data.w,
            VETA[[b_w]],
            LAMBDA[[b_w]],
            Sigma.hat[[b_w]],
            Sigma.inv[[b_w]],
            EETA[[b_w]],
            EY[[b_w]]
          )
          if (length(lavmodel_x@ov.x.dummy.lv.idx[[b_w]]) > 0L) {
            eta.w[, lavmodel_x@ov.x.dummy.lv.idx[[b_w]]] <-
              data.w[, lavmodel_x@ov.x.dummy.ov.idx[[b_w]], drop = FALSE]
          }
          if (length(lavmodel_x@ov.y.dummy.lv.idx[[b_w]]) > 0L) {
            eta.w[, lavmodel_x@ov.y.dummy.lv.idx[[b_w]]] <-
              data.w[, lavmodel_x@ov.y.dummy.ov.idx[[b_w]], drop = FALSE]
          }

          # --- Between factor scores ---
          Data.B <- matrix(0, nrow = nrow(MB.j), ncol = p)
          Data.B[, ov.idx[[1]]] <- MB.j
          between.idx <- Lp$between.idx[[2 * g]]
          if (length(between.idx) > 0L) {
            unique_rows <- match(
              seq_len(Lp$nclusters[[2]]),
              Lp$cluster.idx[[2]]
            )
            Data.B[, between.idx] <-
              y_g[unique_rows, between.idx, drop = FALSE]
          }
          data.b <- Data.B[, ov.idx[[2]], drop = FALSE]
          eta.b <- compute_fs_ml(
            data.b,
            VETA[[b_b]],
            LAMBDA[[b_b]],
            Sigma.hat[[b_b]],
            Sigma.inv[[b_b]],
            EETA[[b_b]],
            EY[[b_b]]
          )

          # --- Predicted values: within + between ---
          eta_w_c <- sweep(eta.w, 2, EETA[[b_w]])
          yhat_w <- t(EY[[b_w]] + LAMBDA[[b_w]] %*% t(eta_w_c)) # n_obs x p_w

          eta_b_c <- sweep(eta.b, 2, EETA[[b_b]])
          yhat_b <- t(EY[[b_b]] + LAMBDA[[b_b]] %*% t(eta_b_c)) # n_clust x p_b

          yhat <- matrix(0, n_obs, p)
          yhat[, ov.idx[[1]]] <- yhat_w
          yhat[, ov.idx[[2]]] <- yhat[, ov.idx[[2]]] +
            yhat_b[Lp$cluster.idx[[2]], , drop = FALSE]

          # --- Residual noise for ypred ---
          if (add_noise) {
            # Endogenous variable indices per block (exclude dummy LVs)
            dummy_w <- c(
              lavmodel_x@ov.x.dummy.ov.idx[[b_w]],
              lavmodel_x@ov.y.dummy.ov.idx[[b_w]]
            )
            endo_w <- setdiff(seq_len(length(ov.idx[[1]])), dummy_w)

            dummy_b <- c(
              lavmodel_x@ov.x.dummy.ov.idx[[b_b]],
              lavmodel_x@ov.y.dummy.ov.idx[[b_b]]
            )
            endo_b <- setdiff(seq_len(length(ov.idx[[2]])), dummy_b)

            # Within residual noise (endogenous only)
            if (length(endo_w) > 0L) {
              mm_w <- seq_len(nmat[b_w]) + cumsum(c(0, nmat))[b_w]
              Theta_w <- lavmodel_x@GLIST[mm_w][["theta"]]
              Theta_w_sub <- Theta_w[endo_w, endo_w, drop = FALSE]
              chol_Tw <- t(chol(Theta_w_sub))
              n_ew <- length(endo_w)
              E_w <- matrix(rnorm(n_obs * n_ew), nrow = n_ew, ncol = n_obs)
              yhat[, ov.idx[[1]][endo_w]] <-
                yhat[, ov.idx[[1]][endo_w]] + t(chol_Tw %*% E_w)
            }

            # Between residual noise (endogenous, cluster-level, expanded)
            if (length(endo_b) > 0L) {
              mm_b <- seq_len(nmat[b_b]) + cumsum(c(0, nmat))[b_b]
              Theta_b <- lavmodel_x@GLIST[mm_b][["theta"]]
              Theta_b_sub <- Theta_b[endo_b, endo_b, drop = FALSE]
              n_clust <- Lp$nclusters[[2]]
              chol_Tb <- t(chol(Theta_b_sub))
              n_eb <- length(endo_b)
              E_b <- matrix(rnorm(n_clust * n_eb), nrow = n_eb, ncol = n_clust)
              eps_b <- t(chol_Tb %*% E_b)
              yhat[, ov.idx[[2]][endo_b]] <-
                yhat[, ov.idx[[2]][endo_b]] +
                eps_b[Lp$cluster.idx[[2]], , drop = FALSE]
            }
          }

          cn <- colnames(y[[g]])
          if (is.null(cn)) {
            cn <- lavdata@ov.names[[g]]
          }
          colnames(yhat) <- cn
          out[[g]] <- yhat
        }

        if (nG == 1L) {
          out <- out[[1L]]
        } else {
          out <- do.call(
            rbind,
            Map(function(g, df) data.frame(group = g, df), names(out), out)
          )
        }
        rownames(out) <- NULL
        out
      }

      msg <- if (add_noise) {
        "Sampling predicted values (multilevel)"
      } else {
        "Sampling fitted values (multilevel)"
      }
      out <- vector("list", nsamp)
      cli_progress_bar(msg, total = nsamp, clear = FALSE)
      for (i in seq_len(nsamp)) {
        out[[i]] <- sample_yhat_ml(x_samp[i, ])
        cli_progress_update()
      }
      cli_progress_done()
    } else {
      # nocov end
      # ---- Single-level yhat/ypred ----
      sample_yhat <- function(xx) {
        GLIST <- get_SEM_param_matrix(xx, "all", lavmodel)
        out <- vector("list", nG)
        names(out) <- group_labels

        for (g in seq_len(nG)) {
          glist <- GLIST[[g]]
          Lambda <- glist$lambda
          Psi <- glist$psi
          Theta <- glist$theta
          B <- glist$beta
          alpha <- glist$alpha
          nu <- glist$nu

          if (is.null(alpha)) {
            alpha <- rep(0, ncol(Lambda))
          }
          if (is.null(nu)) {
            nu <- rep(0, nrow(Lambda))
          }

          if (is.null(B)) {
            Phi <- Psi
            front <- Lambda
          } else {
            IminB_inv <- solve(diag(nrow(B)) - B)
            Phi <- IminB_inv %*% Psi %*% t(IminB_inv)
            front <- Lambda %*% IminB_inv
          }

          Sigmay_inv <- solve(front %*% Psi %*% t(front) + Theta)
          PhiLtSinv <- Phi %*% t(Lambda) %*% Sigmay_inv

          # Posterior draw of eta | y, theta, centring by the implied mean
          # (or the saturated means when no mean structure exists)
          alpha_vec <- rep_len(as.numeric(alpha), ncol(front))
          mu_y <- if (!is.null(glist$nu)) {
            as.numeric(glist$nu + front %*% alpha_vec)
          } else {
            ybar_fit[[g]]
          }
          yc <- sweep(y[[g]], 2L, mu_y)
          mu_eta <- t(as.numeric(alpha) + PhiLtSinv %*% t(yc))
          V_eta <- Phi - PhiLtSinv %*% Lambda %*% Phi
          chol_V <- t(chol(V_eta))
          n_obs <- nrow(mu_eta)
          nlv <- ncol(mu_eta)
          Z <- matrix(rnorm(n_obs * nlv), nrow = nlv, ncol = n_obs)
          eta_draw <- mu_eta + t(chol_V %*% Z)

          # yhat = mu_y + front (eta - alpha)
          nu_eff <- mu_y - as.numeric(front %*% alpha_vec)
          if (is.null(B)) {
            yhat <- sweep(tcrossprod(eta_draw, Lambda), 2, nu_eff, "+")
          } else {
            yhat <- sweep(
              tcrossprod(eta_draw %*% t(IminB_inv), Lambda),
              2,
              nu_eff,
              "+"
            )
          }

          # Add residual noise for ypred
          if (add_noise) {
            p <- ncol(yhat)
            chol_Theta <- t(chol(Theta))
            E <- matrix(rnorm(n_obs * p), nrow = p, ncol = n_obs)
            yhat <- yhat + t(chol_Theta %*% E)
          }

          out[[g]] <- yhat
        }

        if (nG == 1L) {
          cn <- colnames(y[[1L]])
          if (is.null(cn)) {
            cn <- lavdata@ov.names[[1L]]
          }
          colnames(out[[1L]]) <- cn
          out <- out[[1L]]
        } else {
          out <- do.call(
            rbind,
            Map(function(g, df) data.frame(group = g, df), names(out), out)
          )
          cn <- colnames(y[[1L]])
          if (is.null(cn)) {
            cn <- lavdata@ov.names[[1L]]
          }
          colnames(out)[-1] <- cn
        }
        rownames(out) <- NULL
        out
      }

      msg <- if (add_noise) {
        "Sampling predicted values"
      } else {
        "Sampling fitted values"
      }
      out <- vector("list", nsamp)
      cli_progress_bar(msg, total = nsamp, clear = FALSE)
      for (i in seq_len(nsamp)) {
        out[[i]] <- sample_yhat(x_samp[i, ])
        cli_progress_update()
      }
      cli_progress_done()
    }

    # ---- type = "ymis": Missing data imputation ----
  } else if (type == "ymis") {
    # nocov start
    # For each posterior draw of model parameters, compute the model-implied
    # covariance Sigma(theta) and mean mu(theta), then draw missing values
    # from their conditional distribution given observed values:
    #   y_mis | y_obs, theta ~ N(mu_cond, Sigma_cond)
    nlevels <- lavdata@nlevels

    # Pre-compute per-block ov names for naming model-implied matrices
    ov_names_block <- NULL
    if (nlevels > 1L) {
      # nocov start
      nblocks <- lavmodel@nblocks
      ov_names_block <- vector("list", nblocks)
      for (b in seq_len(nblocks)) {
        g_b <- ceiling(b / nlevels)
        ov_all <- lavdata@ov.names[[g_b]]
        ov_names_block[[b]] <- unique(
          pt$lhs[
            pt$block == b &
              pt$op == "~~" &
              pt$lhs == pt$rhs &
              pt$lhs %in% ov_all
          ]
        )
      }
    } # nocov end

    sample_ymis <- function(xx) {
      lavmodel_x <- lavaan::lav_model_set_parameters(lavmodel, xx)
      lavimplied <- lavaan::lav_model_implied(lavmodel_x)

      out <- vector("list", nG)
      names(out) <- group_labels

      for (g in seq_len(nG)) {
        yg <- y[[g]]
        p <- ncol(yg)
        n_obs <- nrow(yg)
        outg <- yg

        if (nlevels == 1L) {
          Sigma_y <- lavimplied$cov[[g]]
          mu_y <- if (!is.null(lavimplied$mean)) {
            as.numeric(lavimplied$mean[[g]])
          } else {
            # no mean structure: condition on the saturated (sample) means
            # (defensive: missing = "ML" forces a mean structure in lavaan,
            # so this branch is unreachable from a real fit)
            ybar_fit[[g]] # nocov
          }
        } else {
          # nocov start
          # Multilevel: marginal covariance = within + between
          block_w <- (g - 1) * nlevels + 1
          block_b <- (g - 1) * nlevels + 2

          Sigma_w <- lavimplied$cov[[block_w]]
          Sigma_b <- lavimplied$cov[[block_b]]

          # Name the model-implied matrices using per-block ov names
          if (!is.null(ov_names_block)) {
            nw <- ov_names_block[[block_w]]
            if (length(nw) == nrow(Sigma_w)) {
              rownames(Sigma_w) <- colnames(Sigma_w) <- nw
            }
            nb <- ov_names_block[[block_b]]
            if (length(nb) == nrow(Sigma_b)) {
              rownames(Sigma_b) <- colnames(Sigma_b) <- nb
            }
          }

          var_names <- colnames(yg)
          if (is.null(var_names)) {
            var_names <- lavdata@ov.names[[g]]
          }
          Sigma_y <- matrix(0, p, p, dimnames = list(var_names, var_names))

          vn_w <- rownames(Sigma_w)
          w_in_data <- match(vn_w, var_names)
          w_keep <- !is.na(w_in_data)
          if (any(w_keep)) {
            idx <- w_in_data[w_keep]
            Sigma_y[idx, idx] <- Sigma_y[idx, idx] +
              Sigma_w[w_keep, w_keep, drop = FALSE]
          }

          vn_b <- rownames(Sigma_b)
          b_in_data <- match(vn_b, var_names)
          b_keep <- !is.na(b_in_data)
          if (any(b_keep)) {
            idx <- b_in_data[b_keep]
            Sigma_y[idx, idx] <- Sigma_y[idx, idx] +
              Sigma_b[b_keep, b_keep, drop = FALSE]
          }

          mu_y <- rep(0, p)
          names(mu_y) <- var_names
          if (!is.null(lavimplied$mean)) {
            mu_b <- as.numeric(lavimplied$mean[[block_b]])
            names(mu_b) <- vn_b
            b_match <- intersect(var_names, vn_b)
            mu_y[match(b_match, var_names)] <- mu_b[b_match]
          }
        } # nocov end

        # Detect missing values
        na_mat <- is.na(yg)
        if (!any(na_mat)) {
          out[[g]] <- outg
          next
        }

        # Group cases by missing-data pattern for efficiency
        patterns <- apply(na_mat, 1, function(r) {
          paste(which(r), collapse = ",")
        })
        unique_patterns <- unique(patterns[patterns != ""])

        for (pat in unique_patterns) {
          mis_idx <- as.integer(strsplit(pat, ",")[[1]])
          obs_idx <- setdiff(seq_len(p), mis_idx)
          case_rows <- which(patterns == pat)

          Sigma_oo <- Sigma_y[obs_idx, obs_idx, drop = FALSE]
          Sigma_mo <- Sigma_y[mis_idx, obs_idx, drop = FALSE]
          Sigma_mm <- Sigma_y[mis_idx, mis_idx, drop = FALSE]

          A <- Sigma_mo %*% solve(Sigma_oo)
          Sigma_cond <- Sigma_mm - A %*% t(Sigma_mo)
          Sigma_cond <- (Sigma_cond + t(Sigma_cond)) / 2
          chol_cond <- t(chol(Sigma_cond))

          n_mis <- length(mis_idx)
          n_cases <- length(case_rows)

          y_obs_centred <- yg[case_rows, obs_idx, drop = FALSE] -
            matrix(
              mu_y[obs_idx],
              nrow = n_cases,
              ncol = length(obs_idx),
              byrow = TRUE
            )
          mu_cond <- matrix(
            mu_y[mis_idx],
            nrow = n_cases,
            ncol = n_mis,
            byrow = TRUE
          ) +
            y_obs_centred %*% t(A)

          Z <- matrix(rnorm(n_cases * n_mis), nrow = n_mis, ncol = n_cases)
          draws <- mu_cond + t(chol_cond %*% Z)
          outg[case_rows, mis_idx] <- draws
        }

        out[[g]] <- outg
      }

      if (ymis_only) {
        # Return only the imputed cells as a named vector: "varname[rowindex]"
        row_offset <- 0L
        imp_list <- vector("list", nG)
        for (g in seq_len(nG)) {
          yg_orig <- y[[g]]
          var_nms <- colnames(yg_orig)
          if (is.null(var_nms)) {
            var_nms <- lavdata@ov.names[[g]]
          }
          na_pos <- which(is.na(yg_orig), arr.ind = TRUE)
          na_pos <- na_pos[order(na_pos[, 1L], na_pos[, 2L]), , drop = FALSE]
          if (nrow(na_pos) == 0L) {
            imp_list[[g]] <- numeric(0L)
          } else {
            vals <- out[[g]][na_pos]
            nms <- paste0(
              var_nms[na_pos[, 2L]],
              "[",
              na_pos[, 1L] + row_offset,
              "]"
            )
            names(vals) <- nms
            imp_list[[g]] <- vals
          }
          row_offset <- row_offset + nrow(yg_orig)
        }
        return(do.call(c, imp_list))
      }

      if (nG == 1L) {
        cn <- colnames(y[[1L]])
        if (is.null(cn)) {
          cn <- lavdata@ov.names[[1L]]
        }
        colnames(out[[1L]]) <- cn
        out <- out[[1L]]
      } else {
        out <- do.call(
          rbind,
          Map(function(g, df) data.frame(group = g, df), names(out), out)
        )
      }
      rownames(out) <- NULL
      out
    }

    out <- vector("list", nsamp)
    cli_progress_bar("Imputing missing values", total = nsamp, clear = FALSE)
    for (i in seq_len(nsamp)) {
      out[[i]] <- sample_ymis(x_samp[i, ])
      cli_progress_update()
    }
    cli_progress_done()
  } # nocov end

  attr(out, "nobs") <- nobs_out
  attr(out, "type") <- type
  out <- structure(out, class = "predict.inlavaan_internal")
  if (isTRUE(summary)) {
    return(base::summary(out))
  }
  out
}

#' @exportS3Method print predict.inlavaan_internal
#' @keywords internal
print.predict.inlavaan_internal <- function(
  x,
  n = 10L,
  nd = 3L,
  ...
) {
  type <- attr(x, "type")
  cat("Predicted values from inlavaan model")
  if (!is.null(type)) {
    cat(sprintf(" (type = \"%s\")", type))
  }
  cat("\n")
  cat("Number of samples:", length(x), "\n")
  cat("First sample:\n")

  first <- x[[1L]]

  # Named vector (ymis_only output)
  if (is.numeric(first) && !is.matrix(first) && !is.data.frame(first)) {
    # nocov start
    n_total <- length(first)
    if (n_total > n) {
      print(first[seq_len(n)], digits = nd)
      cat(col_grey(paste0(
        "# ",
        symbol$info,
        " ",
        n_total - n,
        " more value",
        if (n_total - n == 1L) "" else "s",
        "\n"
      )))
      cat(col_grey(paste0(
        "# ",
        symbol$info,
        " Use `summary()` to see summary statistics\n"
      )))
    } else {
      print(first, digits = nd)
    }
    return(invisible(x))
  } # nocov end

  # Matrix / data frame output
  nr <- nrow(first)
  if (!is.null(nr) && nr > n) {
    print(as.data.frame(first[seq_len(n), , drop = FALSE]), digits = nd)
    cat(col_grey(paste0(
      "# ",
      symbol$info,
      " ",
      nr - n,
      " more row",
      if (nr - n == 1L) "" else "s",
      "\n"
    )))
    cat(col_grey(paste0(
      "# ",
      symbol$info,
      " Use `summary()` to see summary statistics\n"
    )))
  } else {
    print(as.data.frame(first), digits = nd)
  }
  invisible(x)
}

#' @exportS3Method summary predict.inlavaan_internal
#' @keywords internal
summary.predict.inlavaan_internal <- function(object, ...) {
  is_group <- FALSE
  if (!is.null(names(object[[1]])[1])) {
    is_group <- names(object[[1]])[1] == "group" &
      is.character(object[[1]][, 1])
  }

  if (is_group) {
    # Remove the group column, assuming it's always the first column
    group_id <- object[[1]][, 1]
    object <- lapply(object, function(df) as.matrix(df[-1]))
  } else {
    group_id <- NULL
  }
  arr <- simplify2array(object)

  Mean <- apply(arr, c(1, 2), mean)
  SD <- apply(arr, c(1, 2), sd)
  Q <- apply(arr, c(1, 2), quantile, probs = c(0.025, 0.5, 0.975))
  Mode <- apply(arr, c(1, 2), function(x) {
    d <- density(x)
    d$x[which.max(d$y)]
  })

  res <- list(
    group_id = group_id,
    Mean = Mean,
    SD = SD,
    `2.5%` = Q[1, , ],
    `50%` = Q[2, , ],
    `97.5%` = Q[3, , ],
    Mode = Mode
  )
  structure(res, class = "summary.predict.inlavaan_internal")
}

#' @exportS3Method print summary.predict.inlavaan_internal
#' @keywords internal
print.summary.predict.inlavaan_internal <- function(
  x,
  stat = "Mean",
  n = 10L,
  nd = 3L,
  ...
) {
  cat(paste0(stat, " of predicted values from inlavaan model\n\n"))
  mat <- x[[stat]]
  nr <- nrow(mat)
  if (!is.null(nr) && nr > n) {
    print(as.data.frame(mat[seq_len(n), , drop = FALSE]), digits = nd)
    cat(col_grey(paste0(
      "# ",
      symbol$info,
      " ",
      nr - n,
      " more row",
      if (nr - n == 1L) "" else "s",
      "\n"
    )))
  } else {
    print(as.data.frame(mat), digits = nd)
  }
}

# #' @exportS3Method plot predict.inlavaan_internal
# #' @keywords internal
# plot.predict.inlavaan_internal <- function(x, nrow = NULL, ncol = NULL, ...) {
#   summ <- summary(x)
#   nobs <- attr(x, "nobs")
#   nG <- length(nobs)
#   groups <- rep(seq_len(nG), times = nobs)
#
#   means <- summ$Mean
#   score <- rowSums(means)
#   ranks <- order(score, decreasing = TRUE)
#   summ <- lapply(summ, function(mat) cbind(group = groups, mat))
#
#   plot_df <-
#     bind_rows(lapply(summ, function(x) {
#       as.data.frame(x) |>
#         rownames_to_column("id")
#     }), .id = "statistic") |>
#     pivot_longer(-c(statistic, id, group), names_to = "var", values_to = "val") |>
#     pivot_wider(names_from = statistic, values_from = val) |>
#     mutate(
#       id = factor(id, levels = ranks),
#       group = factor(group, labels = attr(x, "group.label"))
#     )
#
#   if (nG > 1L) {
#     browser()
#     p <-
#       ggplot(plot_df) +
#       geom_pointrange(aes(x = id, y = Mean, ymin = `2.5%`, ymax = `97.5%`, col = group), size = 0)
#   } else {
#     p <-
#       ggplot(plot_df) +
#       geom_pointrange(aes(x = id, y = Mean, ymin = `2.5%`, ymax = `97.5%`), size = 0)
#   }
#
#   p +
#     facet_wrap(~ var, nrow = nrow, ncol = ncol) +
#     theme_minimal() +
#     theme(axis.text.x = element_text(angle = 90, vjust = 0.5, hjust = 1, size = 5)) +
#     labs(x = "Individual ID", y = "Value")
#
# }

#' Posterior Predictions for INLAvaan Models
#'
#' Compute posterior predictions from a fitted \code{INLAvaan} model,
#' including latent variable scores, predicted observed values, and imputed
#' missing data.
#'
#' @param object An object of class [INLAvaan].
#' @param type Character string specifying the type of prediction:
#'   \describe{
#'     \item{\code{"lv"}}{(default) Posterior draws of latent variable scores
#'       \eqn{\eta | y, \theta}.}
#'     \item{\code{"yhat"}, \code{"ov"}}{Predicted means for observed variables
#'       \eqn{E(y | \eta, \theta) = \nu + \Lambda \eta}; no residual noise.}
#'     \item{\code{"ypred"}, \code{"ydist"}}{Predicted observed values including
#'       residual noise \eqn{y = \nu + \Lambda \eta + \varepsilon},
#'       \eqn{\varepsilon \sim N(0, \Theta)}.}
#'     \item{\code{"ymis"}, \code{"ovmis"}}{Imputed values for missing
#'       observations, drawn from the conditional distribution
#'       \eqn{y_{mis} | y_{obs}, \theta}.}
#'   }
#' @param newdata An optional data frame of new observations. If supplied,
#'   predictions are computed for \code{newdata} rather than the original
#'   training data. Not supported for \code{type = "ymis"}.
#' @param level Integer; for \code{type = "lv"} in multilevel models, specifies
#'   whether level 1 or level 2 latent variables are desired (default \code{1L}).
#' @param nsamp Integer; number of posterior samples to use for prediction.
#'   Defaults to \code{1000}.
#' @param ymis_only Logical; only applies when \code{type = "ymis"}. When
#'   \code{TRUE}, returns only the imputed values as a named numeric vector per
#'   sample (names of the form \code{"varname[rowindex]"}, matching the blavaan
#'   convention). When \code{FALSE} (default), returns the full data matrix with
#'   missing values filled in.
#' @param summary Logical. When \code{TRUE}, collapse the posterior draws with
#'   \code{summary()} and return summary statistics (mean, SD, quantiles, mode)
#'   instead of the raw draws -- equivalent to calling
#'   \code{summary(predict(object, ...))} but without materialising the
#'   intermediate draws object. Default \code{FALSE}.
#' @param ... Currently unused.
#'
#' @returns A list of \code{nsamp} posterior draws, each a matrix (or data
#'   frame, for multiple groups) with rows corresponding to cases and columns
#'   to variables or latent factors. When \code{summary = TRUE}, instead
#'   returns a \code{summary.predict.inlavaan_internal} object with the
#'   posterior mean, SD, quantiles, and mode for each case/variable.
#'
#' @examples
#' \donttest{
#' HS.model <- "
#'   visual  =~ x1 + x2 + x3
#'   textual =~ x4 + x5 + x6
#'   speed   =~ x7 + x8 + x9
#' "
#' utils::data("HolzingerSwineford1939", package = "lavaan")
#' fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
#'             test = "none", verbose = FALSE)
#'
#' # Posterior latent variable scores
#' lv_scores <- predict(fit)
#' head(lv_scores)
#'
#' # Predicted observed variable means
#' yhat <- predict(fit, type = "yhat")
#' head(yhat)
#'
#' # Point estimates only, skipping the manual summary() step
#' predict(fit, type = "yhat", summary = TRUE)
#' }
#'
#' @seealso [sampling()], [simulate()], [summary()]
#'
#' @name predict
#' @rdname predict
#' @aliases predict,INLAvaan-method
#' @export
setMethod(
  "predict",
  "INLAvaan",
  function(
    object,
    type = c("lv", "yhat", "ov", "ypred", "ydist", "ymis", "ovmis"),
    newdata = NULL,
    level = 1L,
    nsamp = 1000,
    ymis_only = FALSE,
    summary = FALSE,
    ...
  ) {
    type <- match.arg(type)
    predict.inlavaan_internal(
      object@external$inlavaan_internal,
      type = type,
      newdata = newdata,
      level = level,
      nsamp = nsamp,
      ymis_only = ymis_only,
      summary = summary,
      ...
    )
  }
)

#' Variance-Covariance Matrix for INLAvaan Models
#'
#' Extract the posterior variance-covariance matrix of model parameters from a
#' fitted \code{INLAvaan} model.
#'
#' @param object An object of class [INLAvaan].
#' @param type Character. \code{"lavaan"} (default) returns the posterior
#'   covariance matrix of the model parameters computed from posterior
#'   samples (matching lavaan output). \code{"theta"} returns the Laplace
#'   approximation covariance in the internal parameterisation.
#' @param ... Currently unused.
#'
#' @returns A square numeric matrix.
#'
#' @seealso [summary()], [coef()], [standardisedsolution()]
#'
#' @examples
#' \donttest{
#' HS.model <- "
#'   visual  =~ x1 + x2 + x3
#'   textual =~ x4 + x5 + x6
#'   speed   =~ x7 + x8 + x9
#' "
#' utils::data("HolzingerSwineford1939", package = "lavaan")
#' fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
#'             test = "none", verbose = FALSE)
#'
#' # Default: posterior covariance of lavaan parameters
#' vcov(fit)
#'
#' # Internal parameterisation (Laplace approximation)
#' vcov(fit, type = "theta")
#' }
#'
#' @name vcov
#' @aliases vcov,INLAvaan-method
#' @export
setMethod("vcov", "INLAvaan", function(object, type = c("lavaan", "theta"), ...) {
  type <- match.arg(type)
  if (type == "lavaan") {
    object@vcov$vcov
  } else {
    object@vcov$vcov_theta
  }
})

# inla_sem_cached <- function(
#     cmd = c("graph", "Q", "mu", "initial", "log.norm.const", "log.prior", "quit"),
#     theta = NULL,
#     .debug = FALSE) {
#
#   # In the environment require
#   # - n (sample size)
#   # - p (no of items)
#   # - q (no of factors)
#   # - init (initial values)
#   # - partable
#
#   # CACHE
#   envir <- parent.env(environment())
#   if (!exists("INLAvaan_SEM_cache", envir = envir)) {
#
#     # FIXME: These are indices for free parameters theta. But sometimes, with
#     # fixed parameter values we still need the indices...
#
#     make_pd <- function(x) {
#       ed <- eigen(x, symmetric = TRUE, only.values = TRUE)
#       if (any (ed$values < 0)) {
#         ed <- eigen(x, symmetric = TRUE)
#         eval <- ed$values
#         evec <- ed$vectors
#         eval[eval < 0] <- .Machine$double.eps
#         out <- evec %*% diag(eval) %*% t(evec)
#       } else {
#         out <- x
#       }
#       out
#     }
#     assign("make_pd", make_pd, envir = envir)
#
#     # Indices of free parameters
#     idx_lam   <- partable$free[partable$mat == "lambda" & partable$free > 0]
#     idx_beta  <- partable$free[partable$mat == "beta" & partable$free > 0]
#     idx_theta <- partable$free[partable$mat == "theta" & partable$free > 0]
#     idx_rho   <- partable$free[partable$mat == "rho" & partable$free > 0]
#     idx_psi   <- partable$free[partable$mat == "psi" & partable$free > 0]
#     idx_lvrho <- partable$free[partable$mat == "lvrho" & partable$free > 0]
#     assign("idx_lam", idx_lam, envir = envir)
#     assign("idx_beta", idx_beta, envir = envir)
#     assign("idx_theta", idx_theta, envir = envir)
#     assign("idx_rho", idx_rho, envir = envir)
#     assign("idx_psi", idx_psi, envir = envir)
#     assign("idx_lvrho", idx_lvrho, envir = envir)
#
#     # Lambda matrix
#     Lam_df <- partable[partable$mat == "lambda", ]
#     Lambda <- matrix(0, nrow = max(Lam_df$row), ncol = max(Lam_df$col))
#     Lambda[cbind(Lam_df$row, Lam_df$col)] <- Lam_df$est
#     LAM_IDX <- as.matrix(Lam_df[Lam_df$free > 0, c("row", "col")])
#     assign("Lambda", Lambda, envir = envir)
#     assign("LAM_IDX", LAM_IDX, envir = envir)
#
#     # (I-B) matrix
#     B_df <- partable[partable$mat == "beta", ]
#     B_IDX <- as.matrix(B_df[B_df$free > 0, c("row", "col")])
#     IminB <- 1
#     if (length(idx_beta) > 0) {
#       IminB <- diag(q)
#       IminB[B_IDX] <- -B_df$est
#     }
#     assign("IminB", IminB, envir = envir)
#     assign("B_IDX", B_IDX, envir = envir)
#
#     # Rho and Theta matrix
#     Rho_df <- partable[partable$mat == "rho", ]
#     Theta <- diag(subset(partable, mat == "theta" & row == col)$start)
#     RHO_IDX <- cbind(Rho_df$row, Rho_df$col)
#     assign("RHO_IDX", RHO_IDX, envir = envir)
#     assign("Theta", Theta, envir = envir)
#
#     # LVRho and Psi matrix
#     LVRho_df <- partable[partable$mat == "lvrho", ]
#     Psi <- diag(subset(partable, mat == "psi" & row == col)$start)
#     LVRHO_IDX <- cbind(LVRho_df$row, LVRho_df$col)
#     assign("LVRHO_IDX", LVRHO_IDX, envir = envir)
#     assign("Psi", Psi, envir = envir)
#
#     assign("INLAvaan_SEM_cache", TRUE, envir = envir)
#   }
#
#   interpret.theta <- function() {
#
#     lambda <- theta[idx_lam]
#     beta <- theta[idx_beta]
#     sd_e <- sqrt(exp(theta[idx_theta]))  # sd_e = sd_e ^ 2 (item sd)
#     rho <- theta_to_rho(theta[idx_rho])
#     sd_z <- sqrt(exp(theta[idx_psi]))  # sd_z = sd_z ^ 2 (latent sd)
#     lvrho <- theta_to_rho(theta[idx_lvrho])
#
#     if (!is.null(sp)) sp$spin()
#
#     list(
#       lambda = lambda,
#       beta = beta,
#       sd_e = sd_e,
#       rho = rho,
#       sd_z = sd_z,
#       lvrho = lvrho
#     )
#   }
#
#   Q <- function(debug = .debug) {
#     params <- interpret.theta()
#
#     # The matrices are cached and updated here
#     Lambda[LAM_IDX] <- params$lambda
#     if (length(idx_beta) > 0) IminB[B_IDX] <- -params$beta
#
#     # Theta matrix
#     diag(Theta) <- params$sd_e ^ 2
#     if (length(idx_rho) > 0) {
#       I <- RHO_IDX[, 1]
#       J <- RHO_IDX[, 2]
#       values <- params$rho * params$sd_e[I] * params$sd_e[J]
#       Theta[cbind(I, J)] <- values
#       Theta[cbind(J, I)] <- values
#
#       Theta <- make_pd(Theta)  # force pd
#     }
#
#     # Psi matrix
#     diag(Psi) <- params$sd_z ^ 2
#     if (length(idx_lvrho) > 0) {
#       I <- LVRHO_IDX[, 1]
#       J <- LVRHO_IDX[, 2]
#       values <- params$lvrho * params$sd_z[I] * params$sd_z[J]
#       Psi[cbind(I, J)] <- values
#       Psi[cbind(J, I)] <- values
#
#       Psi <- make_pd(Psi)  # force pd
#     }
#
#     if (isTRUE(debug)) {
#       return(list(
#         Lambda = Lambda,
#         IminB = IminB,
#         Theta = Theta,
#         Psi = Psi
#       ))
#     }
#
#     # Return inverse Sigma
#     if (is.matrix(IminB)) {
#       front <- Lambda %*% solve(IminB)
#     } else {
#       front <- Lambda
#     }
#     Sigma <- front %*% tcrossprod(Psi, front) + Theta
#     Sigma <- Sigma + 1e-10 * diag(nrow(Sigma))
#     solve(Sigma)
#   }
#
#   mu <- function() { numeric(0) }
#
#   log.norm.const <- function() { numeric(0) }
#
#   log.prior = function() {
#     params <- interpret.theta()
#
#     # If sigma ~ gamma(shape,rate) then this is the logpdf transform of theta = log(sigma^2)
#     log_pdf_scale <- function(x, shape = 1, rate = 0.5) {
#       dgamma(sqrt(exp(x)), shape, rate, log = TRUE) + x
#     }
#
#     # If rho ~ beta(a,b) then this is the logpdf of theta = log(rho / (1 - rho))
#     log_pdf_rho <- function(x, shape1 = 1, shape2 = 1) {
#       pos_only <- FALSE
#       u <- 1 / (1 + exp(-x))
#       log_beta_density <- dbeta(u, shape1, shape2, log = TRUE)
#       if (pos_only) {
#         log_jacobian <- 0
#       } else {
#         log_jacobian <- -(log(2) + log(u) + log(1-u)) # log(Jacobian) = theta - 2 * log(1 + exp(theta))
#       }
#
#       log_beta_density + log_jacobian
#     }
#
#     # FIXME: Adjust priors in the future
#     res <-
#       sum(dnorm(params$lambda, mean = 0, sd = 10, log = TRUE)) +
#       sum(dnorm(params$beta, mean = 0, sd = 10, log = TRUE)) +
#       # Variances
#       # sum(dgamma(params$sd_e, shape = 1, rate = 5e-05, log = TRUE)) +
#       # sum(dbeta(params$rho, shape1 = 1, shape2 = 1, log = TRUE)) +
#       sum(log_pdf_scale(params$sd_e, shape = 1, rate = 0.5)) +
#       sum(log_pdf_rho(params$rho, shape1 = 1, shape2 = 1)) +
#       # sum(dgamma(params$sd_z, shape = 1, rate = 5e-05, log = TRUE)) +
#       # sum(dbeta(params$lvrho, shape1 = 1, shape2 = 1, log = TRUE)) +
#       sum(log_pdf_scale(params$sd_z, shape = 1, rate = 0.5)) +
#       sum(log_pdf_rho(params$lvrho, shape1 = 1, shape2 = 1))
#     res
#   }
#
#   initial = function() {
#     init
#   }
#
#   graph <- function() {
#     matrix(1, nrow = p, ncol = p)
#   }
#
#   quit = function() { return(invisible()) }
#
#   if (!length(theta)) theta = initial()
#   val = do.call(match.arg(cmd), args = list())
#   return (val)
# }

#' Approximate Leave-One-Out Cross-Validation for INLAvaan Models
#'
#' Computes leave-one-out (LOO) cross-validation for a fitted [INLAvaan]
#' model from a single fit, with no refitting and no sampling, via a Taylor
#' approximation of the case-deletion posterior around the Laplace summary.
#' Single-level models are scored per subject (leave-one-subject-out, LOSO);
#' two-level models are scored per cluster (leave-one-cluster-out, LOCO).
#'
#' @details
#' For a unit \eqn{u} (a subject for LOSO, a cluster for LOCO) with
#' log-likelihood contribution \eqn{\ell_u(\theta)}, score \eqn{s_u} and
#' Hessian \eqn{H_u} evaluated at the posterior summary
#' \eqn{(\theta^*, \Sigma)}, the log conditional predictive ordinate is
#' approximated to first and second order by
#' \deqn{\log \mathrm{CPO}_u^{(1)} = \ell_u - \tfrac{1}{2} s_u' \Sigma s_u,}
#' \deqn{\log \mathrm{CPO}_u^{(2)} = \ell_u
#'   - \tfrac{1}{2} s_u' (\Sigma^{-1} + H_u)^{-1} s_u
#'   + \tfrac{1}{2} \log |I + \Sigma H_u|.}
#' The reported `elpd_loo` is the sum of the second-order terms (first-order
#' when `second_order = FALSE`), with standard error
#' \eqn{\sqrt{n \, \mathrm{var}(\log \mathrm{CPO}_u)}} and
#' `looic` \eqn{= -2 \, \mathrm{elpd}}. The effective number of parameters is
#' \eqn{p_{\mathrm{loo}} = \sum_u (\mathrm{lpd}_u - \log \mathrm{CPO}_u)},
#' where \eqn{\mathrm{lpd}_u} is the analogous Taylor approximation of the
#' full-posterior pointwise log predictive density. Units whose second-order
#' curvature matrix is not positive definite fall back to first order for
#' that unit only (flagged in `per_unit$ok`).
#'
#' The type is resolved automatically: per-cluster (`"loco"`) when the model
#' was fitted with a `cluster` argument, per-subject (`"loso"`) otherwise.
#' For a two-level model these are the two estimands of Merkle, Furr &
#' Rabe-Hesketh (2019): the default per-cluster `"loco"` is the *marginal*
#' predictive (leave-one-cluster-out -- prediction for a *new* cluster),
#' while `type = "loso"` forces the *conditional* predictive
#' (leave-one-unit-out -- prediction for a new observation within an
#' *observed* cluster), where row \eqn{i} of cluster \eqn{j} contributes
#' \eqn{\ell_i = \ell_j(\mathrm{full}) - \ell_j(\mathrm{minus\ row\ } i)},
#' the conditional density of the row given the rest of its cluster. The two
#' answer different questions and are easily conflated, so the per-cluster
#' marginal is the default and `type = "loso"` warns. It works with and
#' without missing data, costs one cluster evaluation per row per Hessian
#' direction, and is best subset with `units`.
#'
#' **Multigroup models.** Groups are independent, so each unit is scored
#' against its own group's implied moments; without a mean structure the
#' exchangeability transformation applies per group, and cross-group
#' equality constraints (`group.equal`) flow through the packed parameter
#' space automatically. The per-unit results are stacked by group (a
#' `group` column records the membership), and units are identified by
#' *case number* -- the row number of the analysed dataset -- so a unit
#' keeps its identity across fits that assign or order groups differently
#' (e.g. a pooled fit versus a grouped fit of the same data, which
#' [compare()] pairs unit by unit). This makes
#' `compare(..., loo = TRUE)` the instrument of choice for the
#' measurement-invariance ladder: configural, metric, and scalar fits are
#' compared on a proper predictive scale with paired standard errors.
#'
#' Supplying `theta` and/or `Sigma` scores the model at an *arbitrary*
#' Gaussian posterior summary instead of the fit's own, without refitting.
#' This is the building block for refit-free model exploration: for example,
#' conditioning the encompassing model's summary on a parameter being zero
#' (a rank-one update of `theta` and `Sigma`) and scoring the result gives
#' the LOO of that submodel from a single fit. INLAvaan provides only this
#' evaluation API; search strategies are left to the user. A conditioned
#' `Sigma` may be singular; the computation automatically restricts to the
#' non-degenerate block, which is exact.
#'
#' Parallelism is strictly opt-in: the default `cores = NULL` runs serially,
#' and `cores > 1` parallelises the Hessian stage via forking (not available
#' on Windows).
#'
#' Calling `loo()` never modifies the fitted object. Under the default
#' `test = "standard"`, [inlavaan()] already computes and stores the full
#' LOO at fit time whenever the model is supported, has a mean structure,
#' and the predicted serial cost is within a 10-second budget (measured by
#' timing one score evaluation); `test = "loo"` forces the computation
#' regardless of the budget, and `fit <- add_loo(fit)` stores it post hoc.
#' A stored result is returned directly by `loo(fit)` when called with
#' default arguments, and is reused by [fitmeasures()] and [compare()]
#' without recomputation.
#'
#' **Exogenous covariates.** The flavour of the score follows the fitted
#' likelihood. Under `fixed.x = FALSE` the covariates are modelled jointly
#' and each unit is scored by the joint predictive density of its outcomes
#' *and* covariates (`flavour = "joint"`). Under `fixed.x = TRUE` (the
#' lavaan default) the fitted likelihood is the conditional one, and each
#' unit is scored by the predictive density of its outcomes *given* its
#' covariates (`flavour = "conditional"`); since the conditional likelihood
#' is exactly invariant to the frozen covariate moments, this involves no
#' additional approximation. The two flavours estimate different quantities
#' whose scales differ by the covariate predictive density, so a joint and a
#' conditional elpd must never be compared ([compare()] refuses
#' mixed-flavour comparisons). Conditional scores of models conditioning on
#' *different* covariate sets are comparable provided the outcome variables
#' match -- the natural setting for covariate selection. Both flavours
#' support any covariate placement: single-level covariates, and
#' cluster-level (between) and/or within-level covariates in two-level
#' models.
#'
#' **Missing data.** Fits estimated by full-information maximum likelihood
#' (`missing = "ml"`) are scored on the *observed-data* predictive: each
#' unit contributes the density of the entries it actually has, with its
#' full row (single-level) or whole cluster (two-level) removed from the
#' conditioning set. For single-level fits the casewise kernels operate on
#' each unit's observed subset, grouping rows by missing pattern, so a unit
#' with fewer observed entries contributes a smaller log-likelihood term
#' *and* a smaller score and thus self-weights in the elpd. Two-level fits
#' are scored per cluster (`"loco"`): each cluster contributes its
#' observed-data marginal likelihood, evaluated by lavaan's raw-data
#' cluster kernels (no per-cluster sufficient statistics are needed, since
#' leave-one-cluster-out deletes the whole cluster). All carry the same
#' missing-at-random assumption as the FIML fit itself. Because the score
#' is the observed-entry predictive, a [compare()] of two missing-data fits
#' is meaningful only when they share the same observed entries (the same
#' data *and* the same holes). The two-level conditional predictive
#' (`type = "loso"`) is available under missing data too, on the same
#' kernels.
#'
#' Supported models: continuous-indicator models fitted with the `ML`
#' estimator (including FIML, `missing = "ml"`, single- and two-level),
#' single-group or multigroup (multigroup two-level models are not
#' supported yet). If the `loo` package is attached it masks this generic,
#' but `loo(fit)` continues to dispatch correctly because the method is
#' registered by generic name.
#'
#' @param x A fitted [INLAvaan] object (or its `inlavaan_internal` list).
#' @param type Unit type: `"auto"` (default) resolves to `"loso"`
#'   (per-subject) for single-level models and `"loco"` (per-cluster,
#'   marginal predictive) for two-level models. `"loco"` cannot be forced
#'   on a model without clusters; `"loso"` on a two-level model scores the
#'   conditional (leave-one-unit-out) predictive instead (with a warning;
#'   see Details).
#' @param units Optional integer vector of unit indices to score; defaults
#'   to all units. For LOSO these are case numbers (row numbers of the
#'   analysed dataset, as recorded in the fit -- for multigroup fits the
#'   full results are stacked by group, but a unit is always addressed by
#'   its case number); for LOCO, cluster positions.
#' @param second_order Logical; compute the second-order correction
#'   (default `TRUE`). `FALSE` skips the Hessian stage entirely and reports
#'   first-order estimates.
#' @param theta,Sigma Optional posterior mean vector and covariance matrix
#'   (in the unconstrained parameter space, as stored in `theta_star` and
#'   `Sigma_theta`) at which to evaluate the LOO instead of the fit's own
#'   Laplace summary. See Details.
#' @param cores Number of cores for the Hessian stage. The default `NULL`
#'   runs serially; parallelism must be requested explicitly.
#' @param verbose Logical; print progress (default `FALSE`).
#' @param ... Not used.
#'
#' @returns An object of class `inlavaan_loo`: a list with elements
#'   \describe{
#'     \item{`per_unit`}{Data frame of pointwise results: `unit` (case
#'       number for LOSO, cluster position for LOCO), `group` (multigroup
#'       fits only), `nobs` (1 for LOSO, the cluster size for LOCO),
#'       `l_star` (unit log-likelihood at the summary), `score_norm`,
#'       `lpd_1`/`lpd_2` (pointwise log predictive density),
#'       `log_cpo_1`/`log_cpo_2` (pointwise LOO contributions), `det_term`,
#'       and `ok` (second-order success flag).}
#'     \item{`estimates`}{Matrix with rows `elpd_loo`, `p_loo`, `looic` and
#'       columns `Estimate`, `SE` (headline second-order values).}
#'     \item{`elpd_1`, `elpd_2`, `se_1`, `se_2`, `p_loo_1`, `p_loo_2`}{
#'       First- and second-order aggregates.}
#'     \item{`type`, `flavour`, `n_units`, `n_groups`, `n_ok`,
#'       `second_order`, `theta_overridden`}{Metadata; `flavour` records
#'       whether units were scored jointly with their covariates
#'       (`"joint"`) or conditionally on them (`"conditional"`, for
#'       `fixed.x` fits).}
#'   }
#'
#' @seealso [fitmeasures()], [compare()], [inlavaan()]
#'
#' @example inst/examples/ex-loo.R
#'
#' @export
loo <- function(x, ...) {
  UseMethod("loo")
}

#' @rdname loo
#' @exportS3Method loo INLAvaan
loo.INLAvaan <- function(
  x,
  type = c("auto", "loso", "loco"),
  units = NULL,
  second_order = TRUE,
  theta = NULL,
  Sigma = NULL,
  cores = NULL,
  verbose = FALSE,
  ...
) {
  loo.inlavaan_internal(
    x@external$inlavaan_internal,
    type = type,
    units = units,
    second_order = second_order,
    theta = theta,
    Sigma = Sigma,
    cores = cores,
    verbose = verbose
  )
}

#' @rdname loo
#' @exportS3Method loo inlavaan_internal
loo.inlavaan_internal <- function(
  x,
  type = c("auto", "loso", "loco"),
  units = NULL,
  second_order = TRUE,
  theta = NULL,
  Sigma = NULL,
  cores = NULL,
  verbose = FALSE,
  ...
) {
  type <- match.arg(type)

  # Reuse a stored result (computed at fit time via test = "loo", or with
  # add_loo()) when no argument deviates from the defaults
  all_defaults <- type == "auto" &&
    is.null(units) &&
    isTRUE(second_order) &&
    is.null(theta) &&
    is.null(Sigma)
  if (all_defaults && !is.null(x$loo)) {
    if (isTRUE(verbose)) {
      cli_alert_info("Returning the LOO result stored with the fit.")
    }
    return(x$loo)
  }

  inlav_loo(
    int = x,
    type = type,
    units = units,
    second_order = second_order,
    theta = theta,
    Sigma = Sigma,
    eff_cores = resolve_loo_cores(cores),
    verbose = verbose
  )
}

#' @rdname loo
#' @param object A fitted [INLAvaan] object.
#' @returns `add_loo()` returns a copy of `object` with the LOO result
#'   stored alongside the fit (the input object is unchanged); reassign it,
#'   e.g. `fit <- add_loo(fit)`. Only the default LOO is stored, so the
#'   stored result always matches `loo(fit)`.
#' @export
add_loo <- function(object, cores = NULL, verbose = FALSE) {
  if (!is_INLAvaan(object)) {
    cli_abort("{.arg object} must be a fitted {.cls INLAvaan} model.")
  }
  res <- loo(object, cores = cores, verbose = verbose)
  object@external$inlavaan_internal$loo <- res
  object
}

#' @exportS3Method print inlavaan_loo
print.inlavaan_loo <- function(x, ...) {
  label <- switch(
    x$type,
    loso = "leave-one-subject-out",
    loco = "leave-one-cluster-out"
  )
  unit_word <- switch(x$type, loso = "subject", loco = "cluster")
  order_lab <- if (x$second_order && x$n_ok > 0L) {
    "second-order"
  } else {
    "first-order"
  }
  cat("Taylor ", label, " cross-validation (INLAvaan)\n", sep = "")
  cat(
    "Computed from ",
    x$n_units,
    " ",
    unit_word,
    if (x$n_units != 1L) "s",
    if (!is.null(x$n_groups) && x$n_groups > 1L) {
      paste0(" in ", x$n_groups, " groups")
    },
    " (",
    order_lab,
    " Taylor approximation)\n",
    sep = ""
  )
  if (identical(x$flavour, "conditional")) {
    cat("Scored conditionally on the exogenous covariates (fixed.x fit)\n")
  }
  cat("\n")
  print(round(x$estimates, 1))
  if (x$second_order && x$n_ok < x$n_units) {
    cat(
      "\n",
      x$n_units - x$n_ok,
      " of ",
      x$n_units,
      " units fell back to first order (non-positive-definite curvature).\n",
      sep = ""
    )
  }
  if (isTRUE(x$theta_overridden)) {
    cat("\nEvaluated at a user-supplied (theta, Sigma) summary.\n")
  }
  invisible(x)
}

# Small deterministic subset of HolzingerSwineford1939 shared by every fit in
# this file, in place of the full 301-row dataset
set.seed(1)
dat <- lavaan::HolzingerSwineford1939[
  sample(nrow(lavaan::HolzingerSwineford1939), 40),
]

mod_null <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"
mod_full <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  visual ~~ textual
"
mod_speed <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"

# Shared by every test below that just needs *some* comparable pair of
# no-mean-structure, test = "none" fits (historically each re-fit these
# identically)
fit1 <- acfa(mod_null, dat, verbose = FALSE, nsamp = 3, test = "none")
fit2 <- acfa(mod_full, dat, verbose = FALSE, nsamp = 3, test = "none")

# Same, but with a mean structure (needed for loo = TRUE comparisons); shared
# across every loo-comparison test that doesn't need its own bespoke fit
fit1_ms <- acfa(
  mod_null,
  dat,
  meanstructure = TRUE,
  marginal_method = "marggaus",
  vb_correction = FALSE,
  verbose = FALSE,
  nsamp = 3,
  test = "none"
)
fit2_ms <- acfa(
  mod_full,
  dat,
  meanstructure = TRUE,
  marginal_method = "marggaus",
  vb_correction = FALSE,  
  verbose = FALSE,
  nsamp = 3,
  test = "none"
)

test_that("compare() returns compare.inlavaan_internal data.frame", {
  cmp <- compare(fit1, fit2)
  expect_s3_class(cmp, "compare.inlavaan_internal")
  expect_s3_class(cmp, "data.frame")
  expect_equal(nrow(cmp), 2)
  expect_true("npar" %in% names(cmp))
  expect_true("Marg.Loglik" %in% names(cmp))
  expect_true("logBF" %in% names(cmp))
  # sorted by descending marginal log-likelihood; the best model has logBF 0
  expect_equal(cmp$Marg.Loglik, sort(cmp$Marg.Loglik, decreasing = TRUE))
  expect_equal(cmp$logBF[1], 0)
})

test_that("compare() print runs without error", {
  cmp <- compare(fit1, fit2)
  expect_output(print(cmp), "Bayesian Model Comparison")
  expect_output(print(cmp), "marginal log-likelihood")
})

test_that("compare() with fit.measures appends extra columns", {
  cmp <- compare(fit1, fit2, fit.measures = "margloglik")
  expect_true("margloglik" %in% names(cmp))
  expect_output(print(cmp), "Baseline model")
})

test_that("compare() includes DIC/pD when test != 'none'", {
  fit1_std <- acfa(mod_null, dat, verbose = FALSE, nsamp = 3)
  fit2_std <- acfa(mod_full, dat, verbose = FALSE, nsamp = 3)
  cmp <- compare(fit1_std, fit2_std)
  expect_true("DIC" %in% names(cmp))
  expect_true("pD" %in% names(cmp))
})

test_that("compare.inlavaan_internal S3 method works", {
  int1 <- INLAvaan:::get_inlavaan_internal(fit1)
  int2 <- INLAvaan:::get_inlavaan_internal(fit2)
  cmp <- INLAvaan:::compare.inlavaan_internal(int1, int2)
  expect_s3_class(cmp, "compare.inlavaan_internal")
  expect_equal(nrow(cmp), 2)
})

test_that("compare() warns when mean-structure treatments differ", {
  fit_ms <- acfa(
    mod_null,
    dat,
    meanstructure = TRUE,
    verbose = FALSE,
    nsamp = 3,
    test = "none"
  )
  fit_nms <- acfa(
    mod_null,
    dat,
    meanstructure = FALSE,
    verbose = FALSE,
    nsamp = 3,
    test = "none"
  )
  expect_warning(compare(fit_ms, fit_nms), "mean structure")
  # ... but the same comparison under loo = TRUE is unaffected (leave-one-out
  # conditionals are proper under both treatments)
  expect_warning(
    cmp <- compare(fit_ms, fit_nms, loo = TRUE),
    "Interpret only the ELPD columns"
  )
  expect_true(all(is.finite(cmp$ELPD)))
})

test_that("compare() accepts more than two models via ...", {
  fit_speed <- acfa(mod_speed, dat, verbose = FALSE, nsamp = 3, test = "none")
  cmp <- compare(fit1, fit2, fit_speed)
  expect_equal(nrow(cmp), 3)
  expect_setequal(cmp$Model, c("fit1", "fit2", "fit_speed"))
})

test_that("compare(loo = TRUE) appends ELPD columns with paired SEs", {
  cmp <- compare(fit1_ms, fit2_ms, loo = TRUE)
  expect_true(all(
    c("ELPD", "SE", "p_loo", "elpd_diff", "se_diff") %in% names(cmp)
  ))
  # Sorted by descending ELPD; the best model has zero differences
  expect_equal(cmp$ELPD, sort(cmp$ELPD, decreasing = TRUE))
  expect_equal(cmp$elpd_diff[1], 0)
  expect_equal(cmp$se_diff[1], 0)
  expect_true(all(cmp$elpd_diff <= 0))
  expect_true(all(is.finite(cmp$se_diff)) && all(cmp$se_diff >= 0))
  # ELPD agrees with loo() on each fit
  expect_equal(
    sort(cmp$ELPD, decreasing = TRUE),
    sort(
      c(
        unname(loo(fit1_ms)$estimates["elpd_loo", "Estimate"]),
        unname(loo(fit2_ms)$estimates["elpd_loo", "Estimate"])
      ),
      decreasing = TRUE
    ),
    tolerance = 1e-3
  )
  expect_output(print(cmp), "paired differences")

  # Stored LOO results are reused
  cmp2 <- compare(add_loo(fit1_ms), add_loo(fit2_ms), loo = TRUE)
  expect_equal(cmp2$ELPD, cmp$ELPD)
})

test_that("compare(loo = TRUE) aborts for models on different data", {
  fit3 <- acfa(
    mod_null,
    dat[1:20, ],
    meanstructure = TRUE,
    verbose = FALSE,
    nsamp = 3,
    test = "none"
  )
  expect_error(compare(fit1_ms, fit3, loo = TRUE), "same data")
})

test_that("compare(loo = TRUE) aborts when the variable sets differ", {
  fit9 <- acfa(
    mod_speed,
    dat,
    meanstructure = TRUE,
    verbose = FALSE,
    nsamp = 3,
    test = "none"
  )
  expect_error(compare(fit1_ms, fit9, loo = TRUE), "same set of observed")
})

test_that("compare(loo = TRUE) aborts when conditional outcome sets differ", {
  # Both fixed.x = TRUE (conditional flavour), but the outcome variable sets
  # differ (covariate sets may differ under conditional scoring, but outcomes
  # must match) -- distinct from the joint-flavour "same set of observed
  # variables" case above
  modA <- "
    visual =~ x1 + x2 + x3
    visual ~ ageyr
  "
  modB <- "
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
    visual ~ ageyr
  "
  suppressWarnings(fitA <- asem(
    modA,
    dat,
    fixed.x = TRUE,
    meanstructure = TRUE,
    verbose = FALSE,
    nsamp = 3,
    test = "none"
  ))
  suppressWarnings(fitB <- asem(
    modB,
    dat,
    fixed.x = TRUE,
    meanstructure = TRUE,
    verbose = FALSE,
    nsamp = 3,
    test = "none"
  ))
  expect_error(
    suppressWarnings(compare(fitA, fitB, loo = TRUE)),
    "outcome variables"
  )
})

#' Draw Samples from the Generative Model
#'
#' Sample model parameters, latent variables, or observed variables from the
#' generative model underlying a fitted INLAvaan model. By default, parameters
#' are drawn from the **posterior** distribution; set `prior = TRUE` to draw
#' from the **prior** instead (useful for prior predictive checks).
#'
#' @details
#' Each row of the output corresponds to a **fresh parameter draw**: a new
#' \eqn{\boldsymbol\theta^{(s)}} is sampled and then propagated through the
#' generative chain to produce one latent vector and one observed vector. This
#' makes `sampling()` ideal for **prior and posterior predictive checks**
#' (e.g., density overlays, test statistic distributions).
#'
#' The generative chain is:
#' \deqn{\boldsymbol\theta^{(s)} \sim \pi(\boldsymbol\theta \mid \mathbf{y})}
#' \deqn{\boldsymbol\eta^{(s)} \sim N((\mathbf{I} - \mathbf{B})^{-1}\boldsymbol\alpha,\,\boldsymbol\Phi)}
#' \deqn{\mathbf{y}^{*(s)} \sim N(\boldsymbol\Lambda\boldsymbol\eta^{(s)} + \boldsymbol\nu,\,\boldsymbol\Theta)}
#'
#' If you need **complete replicate datasets** (many observations from a single
#' parameter draw) — for example, for simulation-based calibration (SBC) — use
#' [simulate()] instead.
#'
#' This is distinct from [predict()], which computes individual-specific
#' factor scores \eqn{\boldsymbol\eta \mid \mathbf{y},\boldsymbol\theta}
#' conditional on observed data.
#'
#' @param object An object of class [INLAvaan] (or `inlavaan_internal`).
#' @param type Character string specifying what to sample:
#'   \describe{
#'     \item{`"lavaan"`}{(Default) The lavaan-side (constrained) model
#'       parameters. Returns an `nsamp` by `npar` matrix.}
#'     \item{`"theta"`}{The INLAvaan-side unconstrained parameters.
#'       Returns an `nsamp` by `npar` matrix.}
#'     \item{`"latent"`}{Latent variables from the model-implied
#'       distribution. Returns an `nsamp` by `nlv` matrix (one draw per
#'       posterior sample, not tied to any individual).}
#'     \item{`"observed"`}{Observed variables generated from the full
#'       model. Returns an `nsamp` by `nobs_vars` matrix.}
#'     \item{`"implied"`}{Model-implied moments. Returns a length-`nsamp`
#'       list, each element a list with `cov` (model-implied covariance
#'       matrix) and, when `meanstructure = TRUE`, `mean` (model-implied
#'       mean vector). For multi-group models each element is itself a
#'       list of groups.}
#'     \item{`"all"`}{A named list with elements `lavaan`, `theta`,
#'       `latent`, `observed`, and `implied`.}
#'   }
#' @param nsamp Number of samples to draw.
#' @param samp_copula Logical. When `TRUE` (default), posterior parameter
#'   samples use the copula method with the fitted marginals. When `FALSE`,
#'   samples are drawn from the joint Gaussian (Laplace) approximation.
#'   Ignored when `prior = TRUE`.
#' @param prior Logical. When `TRUE`, parameters are drawn from the prior
#'   distribution and then propagated through the generative model. When
#'   `FALSE` (default), parameters come from the posterior.
#' @param silent Logical. When `TRUE`, suppresses the informational message
#'   about rejected non-PD draws during prior rejection sampling. Default
#'   `FALSE`.
#' @param ... Additional arguments (currently unused).
#'
#' @returns A matrix or named list, depending on `type`.
#'
#' @seealso [simulate()] for generating complete replicate datasets (e.g.,
#'   for SBC); [predict()] for individual-specific factor scores;
#'   [bfit_indices()] for Bayesian fit indices.
#'
#' @example inst/examples/ex-sampling.R
#' @export
setGeneric("sampling", function(object, ...) standardGeneric("sampling"))

#' @name sampling
#' @rdname sampling
#' @aliases sampling,INLAvaan-method
#' @export
setMethod(
  "sampling",
  "INLAvaan",
  function(
    object,
    type = c("lavaan", "theta", "latent", "observed", "implied", "all"),
    nsamp = 1000L,
    samp_copula = TRUE,
    prior = FALSE,
    silent = FALSE,
    ...
  ) {
    sampling_impl(
      object@external$inlavaan_internal,
      type = type,
      nsamp = nsamp,
      samp_copula = samp_copula,
      prior = prior,
      meanstructure = isTRUE(object@Options$meanstructure),
      silent = silent,
      ...
    )
  }
)

#' @exportS3Method sampling inlavaan_internal
sampling.inlavaan_internal <- function(
  object,
  type = c("lavaan", "theta", "latent", "observed", "implied", "all"),
  nsamp = 1000L,
  samp_copula = TRUE,
  prior = FALSE,
  silent = FALSE,
  ...
) {
  sampling_impl(
    object,
    type = type,
    nsamp = nsamp,
    samp_copula = samp_copula,
    prior = prior,
    silent = silent,
    ...
  )
}

# ---- Internal: draw parameter samples (theta/x) -----------------------------

sample_params_posterior <- function(int, nsamp, samp_copula) {
  method <- if (isTRUE(samp_copula)) int$marginal_method else "sampling"
  sample_params(
    theta_star = int$theta_star,
    Sigma_theta = int$Sigma_theta,
    method = method,
    approx_data = int$approx_data,
    pt = int$partable,
    lavmodel = int$lavmodel,
    nsamp = nsamp,
    R_star = int$R_star
  )
}

# Draw from the prior: independent draws per free parameter, map to x-space.
sample_params_prior <- function(int, nsamp) {
  pt <- int$partable
  lavmodel <- int$lavmodel

  # All unique free parameter indices in the partable
  PTFREEIDX <- which(pt$free > 0L & !duplicated(pt$free))
  m <- length(PTFREEIDX)

  # Draw in the interpretable/natural scale for each parameter
  x_natural <- matrix(NA_real_, nrow = nsamp, ncol = m)

  for (j in seq_len(m)) {
    prior_str <- pt$prior[PTFREEIDX[j]]

    if (is.na(prior_str) || prior_str == "") {
      # Fallback: wide normal for params without an explicit prior
      x_natural[, j] <- stats::rnorm(nsamp, 0, 10) # nocov
    } else if (grepl("^normal", prior_str)) {
      par <- as.numeric(strsplit(
        gsub("normal\\(|\\)", "", prior_str),
        ","
      )[[1]])
      x_natural[, j] <- stats::rnorm(nsamp, par[1], par[2])
    } else if (grepl("^gamma", prior_str)) {
      is_sd <- grepl("\\[sd\\]", prior_str)
      is_prec <- grepl("\\[prec\\]", prior_str)
      par <- as.numeric(strsplit(
        gsub("gamma\\(|\\)|\\[sd\\]|\\[prec\\]", "", prior_str),
        ","
      )[[1]])
      raw <- stats::rgamma(nsamp, shape = par[1], rate = par[2])
      if (is_sd) {
        x_natural[, j] <- raw^2 # SD → variance  # nocov
      } else if (is_prec) {
        x_natural[, j] <- 1 / raw # precision → variance  # nocov
      } else {
        x_natural[, j] <- raw
      }
    } else if (grepl("^beta", prior_str)) {
      # nocov start
      par <- as.numeric(strsplit(
        gsub("beta\\(|\\)", "", prior_str),
        ","
      )[[1]])
      x_natural[, j] <- stats::rbeta(nsamp, par[1], par[2]) * 2 - 1
    } else {
      x_natural[, j] <- stats::rnorm(nsamp, 0, 10)
    } # nocov end
  }

  # Map natural scale → unconstrained theta-space via g()
  theta_samp <- x_natural
  for (j in seq_len(m)) {
    theta_samp[, j] <- vapply(
      x_natural[, j],
      pt$g[[PTFREEIDX[j]]],
      numeric(1)
    )
  }

  # Apply equality constraints if present
  if (lavmodel@ceq.simple.only) {
    # nocov start
    K <- lavmodel@ceq.simple.K
    theta_samp <- t(apply(theta_samp, 1, function(p) as.numeric(K %*% p)))
  } # nocov end

  # Map theta → lavaan x-space (handles covariance = cor * sqrt(var1 * var2))
  x_samp <- t(apply(theta_samp, 1, pars_to_x, pt = pt))

  list(theta_samp = theta_samp, x_samp = x_samp)
}

# ---- Internal: generate eta from model-implied distribution ------------------

sample_latent_from_model <- function(x_row, lavmodel, strict = FALSE) {
  GLIST <- get_SEM_param_matrix(x_row, "all", lavmodel)
  nG <- lavmodel@ngroups
  eta_list <- vector("list", nG)

  for (g in seq_len(nG)) {
    glist <- GLIST[[g]]
    Psi <- glist$psi
    B <- glist$beta
    alpha <- glist$alpha

    IminB <- if (is.null(B)) diag(nrow(Psi)) else (diag(nrow(B)) - B)
    if (is.null(alpha)) {
      alpha <- rep(0, nrow(Psi))
    }
    IminB_inv <- solve(IminB)

    mu_eta <- as.numeric(IminB_inv %*% alpha)
    Phi <- IminB_inv %*% Psi %*% t(IminB_inv)
    chol_Phi <- if (strict) {
      t(chol(Phi)) # nocov - error propagates if non-PD
    } else {
      tryCatch(t(chol(Phi)), error = function(e) t(chol(make_pd(Phi))))
    }

    eta <- mu_eta + as.numeric(chol_Phi %*% stats::rnorm(length(mu_eta)))
    names(eta) <- colnames(Psi)
    eta_list[[g]] <- eta
  }

  if (nG == 1L) eta_list[[1L]] else eta_list
}

# ---- Internal: generate y from model given eta ------------------------------

sample_observed_from_model <- function(x_row, eta, lavmodel, strict = FALSE) {
  GLIST <- get_SEM_param_matrix(x_row, "all", lavmodel)
  nG <- lavmodel@ngroups
  y_list <- vector("list", nG)

  for (g in seq_len(nG)) {
    glist <- GLIST[[g]]
    Lambda <- glist$lambda
    Theta <- glist$theta
    nu <- glist$nu

    eta_g <- if (nG == 1L) eta else eta[[g]]
    if (is.null(nu)) {
      nu <- rep(0, nrow(Lambda))
    }

    mu_y <- as.numeric(Lambda %*% eta_g + nu)
    chol_Theta <- if (strict) {
      t(chol(Theta)) # nocov - error propagates if non-PD
    } else {
      tryCatch(t(chol(Theta)), error = function(e) t(chol(make_pd(Theta))))
    }
    y <- mu_y + as.numeric(chol_Theta %*% stats::rnorm(length(mu_y)))
    names(y) <- rownames(Lambda)
    y_list[[g]] <- y
  }

  if (nG == 1L) y_list[[1L]] else y_list
}

# ---- Internal: compute model-implied moments --------------------------------

compute_implied_moments <- function(x_row, lavmodel, meanstructure = FALSE) {
  GLIST <- get_SEM_param_matrix(x_row, "all", lavmodel)
  nG <- lavmodel@ngroups
  out_list <- vector("list", nG)

  for (g in seq_len(nG)) {
    glist <- GLIST[[g]]
    Lambda <- glist$lambda
    Psi <- glist$psi
    Theta <- glist$theta
    B <- glist$beta
    alpha <- glist$alpha
    nu <- glist$nu

    IminB <- if (is.null(B)) diag(nrow(Psi)) else (diag(nrow(B)) - B)
    IminB_inv <- solve(IminB)

    # Sigma_y = Lambda (I-B)^{-1} Psi [(I-B)^{-1}]' Lambda' + Theta
    front <- Lambda %*% IminB_inv
    Sigma_y <- front %*% Psi %*% t(front) + Theta
    rownames(Sigma_y) <- colnames(Sigma_y) <- rownames(Lambda)

    res <- list(cov = Sigma_y)

    if (meanstructure) {
      # nocov start
      if (is.null(alpha)) {
        alpha <- rep(0, nrow(Psi))
      }
      if (is.null(nu)) {
        nu <- rep(0, nrow(Lambda))
      }
      mu_y <- as.numeric(Lambda %*% IminB_inv %*% alpha + nu)
      names(mu_y) <- rownames(Lambda)
      res$mean <- mu_y
    } # nocov end

    out_list[[g]] <- res
  }

  if (nG == 1L) out_list[[1L]] else out_list # nocov (else = multigroup)
}

# ---- Internal: prior generative sampling with reject-and-redraw --------------
#
# When prior = TRUE and we need latent/observed draws, parameter vectors that
# produce non-positive-definite model-implied covariance matrices are rejected
# and redrawn.  This preserves the exact prior distribution rather than silently
# projecting non-PD matrices to PD space via make_pd().

sampling_prior_generative <- function(
  int,
  type,
  nsamp,
  meanstructure = FALSE,
  silent = FALSE
) {
  pt <- int$partable
  xnames <- pt$names[pt$free > 0 & !duplicated(pt$free)]
  lavmodel <- int$lavmodel
  nG <- lavmodel@ngroups

  # For 'implied' alone, no Cholesky decomposition is needed — just sample
  # parameters and compute the moments directly (no rejection required).
  if (type == "implied") {
    samp <- sample_params_prior(int, nsamp)
    colnames(samp$x_samp) <- xnames
    return(lapply(seq_len(nsamp), function(i) {
      compute_implied_moments(samp$x_samp[i, ], lavmodel, meanstructure)
    }))
  }

  need_obs <- type %in% c("observed", "all")
  need_implied <- type == "all"

  # Pre-compute dimensions from a single draw
  samp0 <- sample_params_prior(int, 1L)
  GLIST0 <- get_SEM_param_matrix(samp0$x_samp[1, ], "all", lavmodel)
  nlv <- ncol(GLIST0[[1]]$psi)
  nobs <- nrow(GLIST0[[1]]$lambda)
  lv_names <- colnames(GLIST0[[1]]$psi)
  ov_names <- rownames(GLIST0[[1]]$lambda)

  # Column names for output matrices
  if (nG == 1L) {
    eta_cn <- lv_names
    y_cn <- ov_names
  } else {
    eta_cn <- paste0(rep(lv_names, nG), ".g", rep(seq_len(nG), each = nlv)) # nocov
    y_cn <- paste0(rep(ov_names, nG), ".g", rep(seq_len(nG), each = nobs)) # nocov
  }

  # Pre-allocate storage
  npar <- length(xnames)
  x_mat <- matrix(NA_real_, nsamp, npar)
  theta_mat <- matrix(NA_real_, nsamp, npar)
  eta_mat <- matrix(NA_real_, nsamp, length(eta_cn))
  y_mat <- if (need_obs) matrix(NA_real_, nsamp, length(y_cn)) else NULL
  colnames(x_mat) <- xnames
  colnames(theta_mat) <- xnames
  colnames(eta_mat) <- eta_cn
  if (need_obs) {
    colnames(y_mat) <- y_cn
  }
  implied_list <- if (need_implied) vector("list", nsamp) else NULL

  max_attempts <- nsamp * 20L # tolerate up to ~95% rejection rate
  collected <- 0L
  attempts <- 0L

  while (collected < nsamp && attempts < max_attempts) {
    # Draw a batch of parameters (draw what we still need)
    batch_n <- nsamp - collected
    samp_batch <- sample_params_prior(int, batch_n)

    for (i in seq_len(batch_n)) {
      attempts <- attempts + 1L
      if (attempts > max_attempts) {
        # nocov
        break
      }

      x1 <- samp_batch$x_samp[i, ]

      # Try latent draw (strict = TRUE: no make_pd fallback)
      eta1 <- tryCatch(
        sample_latent_from_model(x1, lavmodel, strict = TRUE),
        error = function(e) NULL
      )
      if (is.null(eta1)) {
        # nocov
        next
      }

      # Try observed draw if needed
      if (need_obs) {
        y1 <- tryCatch(
          sample_observed_from_model(x1, eta1, lavmodel, strict = TRUE),
          error = function(e) NULL
        )
        if (is.null(y1)) next # nocov
      }

      # Valid draw -- store it
      collected <- collected + 1L
      x_mat[collected, ] <- x1
      theta_mat[collected, ] <- samp_batch$theta_samp[i, ]

      if (nG == 1L) {
        eta_mat[collected, ] <- eta1
        if (need_obs) y_mat[collected, ] <- y1
      } else {
        # nocov start
        eta_mat[collected, ] <- unlist(eta1)
        if (need_obs) y_mat[collected, ] <- unlist(y1)
      } # nocov end

      if (need_implied) {
        implied_list[[collected]] <- compute_implied_moments(
          x1,
          lavmodel,
          meanstructure
        )
      }

      if (collected >= nsamp) break
    }
  }

  rejected <- attempts - collected
  if (rejected > 0L && !isTRUE(silent)) {
    # nocov start
    rej_pct <- round(100 * rejected / attempts, 1)
    cli_inform(
      "Prior sampling: {rejected} of {attempts} draw{?s} ({rej_pct}%) rejected (non-PD model-implied covariance)."
    )
  } # nocov end

  if (collected < nsamp) {
    # nocov start
    cli_warn(c(
      "Prior rejection sampling fell short of the requested sample size.",
      "i" = "Only {collected} of {nsamp} samples obtained after {attempts} attempts.",
      "i" = "Consider using more informative priors."
    ))
    if (collected == 0L) {
      cli_abort("No valid prior draws obtained. Priors may be too vague.")
    }
    # Trim to valid rows
    x_mat <- x_mat[seq_len(collected), , drop = FALSE]
    theta_mat <- theta_mat[seq_len(collected), , drop = FALSE]
    eta_mat <- eta_mat[seq_len(collected), , drop = FALSE]
    if (need_obs) {
      y_mat <- y_mat[seq_len(collected), , drop = FALSE]
    }
    if (need_implied) implied_list <- implied_list[seq_len(collected)]
  } # nocov end

  if (type == "latent") {
    return(eta_mat)
  }
  if (type == "observed") {
    return(y_mat)
  }

  # type == "all"
  list(
    lavaan = x_mat,
    theta = theta_mat,
    latent = eta_mat,
    observed = y_mat,
    implied = implied_list
  )
}

# ---- Main workhorse ----------------------------------------------------------

sampling_impl <- function(
  int,
  type = c("lavaan", "theta", "latent", "observed", "implied", "all"),
  nsamp = 1000L,
  samp_copula = TRUE,
  prior = FALSE,
  meanstructure = FALSE,
  silent = FALSE,
  ...
) {
  type <- match.arg(type)

  # For prior sampling with generative draws, use reject-and-redraw to preserve
  # the exact prior (no silent PD projection).
  if (isTRUE(prior) && type %in% c("latent", "observed", "implied", "all")) {
    return(sampling_prior_generative(
      int,
      type,
      nsamp,
      meanstructure,
      silent = silent
    ))
  }

  # Step 1: draw parameters
  if (isTRUE(prior)) {
    samp <- sample_params_prior(int, nsamp)
  } else {
    samp <- sample_params_posterior(int, nsamp, samp_copula)
  }

  pt <- int$partable
  xnames <- pt$names[pt$free > 0 & !duplicated(pt$free)]
  lavmodel <- int$lavmodel

  colnames(samp$x_samp) <- xnames
  colnames(samp$theta_samp) <- xnames

  # Early return for parameter-only types
  if (type == "lavaan") {
    return(samp$x_samp)
  }
  if (type == "theta") {
    return(samp$theta_samp)
  }

  # Compute model-implied moments if requested
  if (type == "implied" || type == "all") {
    implied_list <- lapply(seq_len(nsamp), function(i) {
      compute_implied_moments(samp$x_samp[i, ], lavmodel, meanstructure)
    })
    if (type == "implied") return(implied_list)
  }

  # Pre-compute dimensions from the first draw
  nG <- lavmodel@ngroups
  GLIST0 <- get_SEM_param_matrix(samp$x_samp[1, ], "all", lavmodel)
  nlv <- ncol(GLIST0[[1]]$psi)
  nobs <- nrow(GLIST0[[1]]$lambda)
  lv_names <- colnames(GLIST0[[1]]$psi)
  ov_names <- rownames(GLIST0[[1]]$lambda)

  # Step 2: draw latent variables from model-implied distribution
  if (nG == 1L) {
    # matrix(..., byrow = TRUE) rather than t(vapply()): with a single latent
    # variable vapply() returns a length-nsamp vector and t() would produce a
    # 1 x nsamp row matrix, corrupting the sample/variable orientation.
    eta_mat <- matrix(
      vapply(
        seq_len(nsamp),
        function(i) {
          sample_latent_from_model(samp$x_samp[i, ], lavmodel)
        },
        numeric(nlv)
      ),
      nrow = nsamp,
      ncol = nlv,
      byrow = TRUE
    )
    colnames(eta_mat) <- lv_names
  } else {
    # nocov start
    eta_list <- lapply(seq_len(nsamp), function(i) {
      sample_latent_from_model(samp$x_samp[i, ], lavmodel)
    })
    eta_mat <- do.call(
      rbind,
      lapply(eta_list, function(el) {
        unlist(el)
      })
    )
    colnames(eta_mat) <- paste0(
      rep(lv_names, nG),
      ".g",
      rep(seq_len(nG), each = nlv)
    )
  } # nocov end

  if (type == "latent") {
    return(eta_mat)
  }

  # Step 3: draw observed variables from model given eta
  if (nG == 1L) {
    # matrix(..., byrow = TRUE) rather than t(vapply()): guards the single
    # observed variable case the same way as the latent draws above.
    y_mat <- matrix(
      vapply(
        seq_len(nsamp),
        function(i) {
          sample_observed_from_model(samp$x_samp[i, ], eta_mat[i, ], lavmodel)
        },
        numeric(nobs)
      ),
      nrow = nsamp,
      ncol = nobs,
      byrow = TRUE
    )
    colnames(y_mat) <- ov_names
  } else {
    # nocov start
    y_mat <- t(vapply(
      seq_len(nsamp),
      function(i) {
        eta_per_group <- split(eta_mat[i, ], rep(seq_len(nG), each = nlv))
        eta_per_group <- lapply(eta_per_group, unname)
        unlist(sample_observed_from_model(
          samp$x_samp[i, ],
          eta_per_group,
          lavmodel
        ))
      },
      numeric(nobs * nG)
    ))
    colnames(y_mat) <- paste0(
      rep(ov_names, nG),
      ".g",
      rep(seq_len(nG), each = nobs)
    )
  } # nocov end

  # Without a mean structure, sample_observed_from_model() centres the
  # draws at zero (nu does not exist); add the saturated (sample) means of
  # the fitted data so replicates live on the data scale
  if (!isTRUE(lavmodel@meanstructure) && !isTRUE(prior) && nG == 1L) {
    y_mat <- sweep(y_mat, 2L, colMeans(int$lavdata@X[[1L]], na.rm = TRUE), "+")
    # under the marginalised likelihood the saturated means have posterior
    # N(ybar, Sigma/n); propagate that uncertainty into each replicate, as
    # estimated-nu draws do automatically when a mean structure exists
    if (marginalised_means_active(lavmodel)) {
      n_fit <- nrow(int$lavdata@X[[1L]])
      for (i in seq_len(nrow(y_mat))) {
        Sg <- compute_implied_moments(samp$x_samp[i, ], lavmodel)$cov
        ch <- tryCatch(chol(Sg), error = function(e) NULL) # nocov
        if (!is.null(ch)) {
          y_mat[i, ] <- y_mat[i, ] +
            as.numeric(crossprod(ch, rnorm(ncol(y_mat)))) / sqrt(n_fit)
        }
      }
    }
  }

  if (type == "observed") {
    return(y_mat)
  }

  # type == "all"
  list(
    lavaan = samp$x_samp,
    theta = samp$theta_samp,
    latent = eta_mat,
    observed = y_mat,
    implied = implied_list
  )
}

#' Log-Likelihood for INLAvaan Models
#'
#' Extract a log-likelihood-flavoured summary from a fitted \code{INLAvaan}
#' model. Two distinct quantities are available, deliberately not conflated:
#' the Bayesian marginal log-likelihood (the default) and the classical
#' log-likelihood evaluated at the posterior mean.
#'
#' @param object An object of class [INLAvaan].
#' @param type Character. \code{"marginal"} (default) returns the
#'   Laplace-approximated marginal log-likelihood (log evidence), the same
#'   quantity [compare()] uses for Bayes factors. \code{"plugin"} returns the
#'   classical log-likelihood evaluated at the posterior mean point estimate,
#'   with \code{df}/\code{nobs} attributes and class \code{"logLik"} so it
#'   supports \code{\link[stats]{AIC}}/\code{\link[stats]{BIC}} at the point
#'   estimate. Requires the model to have been fitted with
#'   \code{test != "none"}.
#' @param ... Currently unused.
#' @param k Numeric penalty per parameter passed to the (disabled) `AIC()`
#'   method; see [stats::AIC()]. Defaults to 2.
#'
#' @returns For \code{type = "marginal"}, a length-one numeric of class
#'   \code{inlavaan_logLik} that prints with a note on its interpretation.
#'   For \code{type = "plugin"}, a standard \code{"logLik"} object.
#'
#' @details
#' The marginal log-likelihood already integrates over the (Laplace-
#' approximated) posterior, so it is not on the same scale as a classical
#' log-likelihood and should not be passed to \code{AIC()}/\code{BIC()} --
#' doing so would double-penalise model complexity that the evidence has
#' already accounted for. Use [compare()] to compare models via Bayes
#' factors, DIC, or LOO/WAIC. The plug-in variant exists for users who
#' specifically want a point-estimate-based classical comparison.
#'
#' \code{AIC()}/\code{BIC()} on an \code{INLAvaan} fit are themselves
#' disabled (mirroring [anova()]): both are large-sample asymptotic
#' approximations to quantities INLAvaan already computes directly --
#' \code{AIC} approximates predictive accuracy, which [loo()]/[waic()]
#' already estimate more rigorously; \code{BIC} approximates \eqn{-2} times
#' the log marginal likelihood, which \code{logLik()} already returns
#' directly (up to the Laplace approximation). Point-estimate AIC/BIC remain
#' available for reporting-convention purposes via
#' \code{AIC(logLik(object, type = "plugin"))} /
#' \code{BIC(logLik(object, type = "plugin"))}.
#'
#' @seealso [deviance()], [compare()], [loo()], [waic()]
#'
#' @examples
#' \donttest{
#' HS.model <- "
#'   visual  =~ x1 + x2 + x3
#'   textual =~ x4 + x5 + x6
#'   speed   =~ x7 + x8 + x9
#' "
#' utils::data("HolzingerSwineford1939", package = "lavaan")
#' fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
#'             test = "standard", verbose = FALSE)
#'
#' # Marginal log-likelihood (log evidence)
#' logLik(fit)
#'
#' # Classical log-likelihood at the posterior mean, AIC/BIC-compatible
#' ll <- logLik(fit, type = "plugin")
#' AIC(ll)
#' }
#'
#' @importFrom stats logLik
#' @name logLik
#' @aliases logLik,INLAvaan-method
#' @export
setMethod("logLik", "INLAvaan", function(object, type = c("marginal", "plugin"), ...) {
  type <- match.arg(type)
  int <- get_inlavaan_internal(object)

  if (type == "marginal") {
    return(structure(int$mloglik, class = "inlavaan_logLik"))
  }

  if (is.null(int$DIC)) {
    cli_abort(
      "{.code type = \"plugin\"} requires DIC components. Refit with
       {.code test != \"none\"}."
    )
  }
  val <- -int$DIC$Dhat / 2
  attr(val, "df") <- length(coef(object))
  attr(val, "nobs") <- nobs(object)
  class(val) <- "logLik"
  val
})

#' @rdname logLik
#' @aliases AIC,INLAvaan-method
#' @export
setMethod("AIC", "INLAvaan", function(object, ..., k = 2) {
  cli_abort(
    "{.fn AIC} at the posterior mean is a large-sample asymptotic proxy for
     predictive accuracy, which {.fn loo} and {.fn waic} already estimate
     more rigorously for {.cls INLAvaan} fits. For the point-estimate AIC
     (e.g. for reporting conventions), use
     {.code AIC(logLik(object, type = \"plugin\"))}."
  )
})

#' @rdname logLik
#' @aliases BIC,INLAvaan-method
#' @export
setMethod("BIC", "INLAvaan", function(object, ...) {
  cli_abort(
    "{.fn BIC} at the posterior mean is a large-sample asymptotic proxy for
     -2 * log(marginal likelihood), which {.fn logLik} already returns
     directly for {.cls INLAvaan} fits. For the point-estimate BIC (e.g. for
     reporting conventions), use
     {.code BIC(logLik(object, type = \"plugin\"))}."
  )
})

#' @exportS3Method print inlavaan_logLik
print.inlavaan_logLik <- function(x, digits = 3L, ...) {
  cat(sprintf("'log Lik.' %s (marginal)\n", format(round(unclass(x), digits))))
  cat(col_grey(paste0(
    "# ", symbol$info,
    " Laplace-approximated log evidence -- not comparable to classical\n"
  )))
  cat(col_grey(paste0(
    "# ", symbol$info,
    " logLik()/AIC()/BIC(). See `compare()` for Bayes-factor comparison.\n"
  )))
  invisible(x)
}

#' Fit an Approximate Bayesian Latent Variable Model
#'
#' This function fits a Bayesian latent variable model by approximating the
#' posterior distributions of the model parameters using various methods,
#' including skew-normal, asymmetric Gaussian, marginal Gaussian, or
#' sampling-based approaches. It leverages the lavaan package for model
#' specification and estimation.
#'
#' @inheritParams lavaan::lavaan
#' @inheritParams lavaan::simulateData
#' @inheritParams blavaan::blavaan
#' @param model.type The lavaan entry point used to fit `model`: `"cfa"`,
#'   `"sem"`, or `"growth"` (matching lavaan's model-specific wrapper
#'   functions), or `"lavaan"` for the general-purpose interface. Set
#'   automatically by [acfa()], [asem()], and [agrowth()]; documented
#'   explicitly here because lavaan >= 0.7-1 renamed the corresponding
#'   `simulateData()` argument to `model_type`, so it can no longer be
#'   inherited from there.
#'
#' @param dp Default prior distributions for the different types of model
#'   parameters; a named character vector as returned by [priors_for()].
#' @param test Character indicating which post-estimation quantities to
#'   compute. Defaults to "standard": posterior fit indices (PPP and DIC),
#'   plus -- for models supported by the casewise machinery and fitted with
#'   a mean structure -- the WAIC (reusing the fit's posterior draws, when
#'   `nsamp >= 100`) and a full leave-one-out cross-validation whenever its
#'   predicted serial cost is within a 10-second budget; both are stored
#'   with the fit (see [loo()] and [waic()]). "none" skips all of these.
#'   Include "loo" (e.g. `test = c("standard", "loo")`, or `test = "loo"`
#'   alone) to force the full LOO regardless of the budget.
#' @param vb_correction Logical indicating whether to apply a variational Bayes
#'   correction for the posterior mean vector of estimates. Defaults to `TRUE`.
#' @param marginal_method The method for approximating the marginal posterior
#'   distributions. Options include `"skewnorm"` (skew-normal), `"asymgaus"`
#'   (two-piece asymmetric Gaussian), `"marggaus"` (marginalising the Laplace
#'   approximation), and `"sampling"` (sampling from the joint Laplace
#'   approximation).
#' @param marginal_correction Which type of correction to use when fitting the
#'   skew-normal or two-piece Gaussian marginals. `"hessian"` computes the full
#'   `"shortcut"` (default) computes only diagonals via central differences
#'   (full z-trace plus Schur complement correction), `"shortcut_fd"` is the
#'   same formula using forward differences (roughly half the cost, less
#'   accurate), `"hessian"` computes the full Hessian-based correction (slow),
#'   and `"none"` (or `FALSE`) applies no correction.
#' @param nsamp The number of samples to draw for all sampling-based approaches
#'   (including posterior sampling for model fit indices).
#' @param samp_copula Logical. When `TRUE` (default), posterior samples are
#'   drawn using the copula method with the fitted marginals (e.g. skew-normal
#'   or asymmetric Gaussian), with NORTA correlation adjustment. When `FALSE`,
#'   samples are drawn from the Gaussian (Laplace) approximation. Only re
#' @param sn_fit_ngrid Number of grid points to lay out per dimension when
#'   fitting the skew-normal marginals. A finer grid gives a better fit at the
#'   cost of more joint-log-posterior evaluations. Defaults to `21`.
#' @param sn_fit_logthresh The log-threshold for fitting the skew-normal. Points
#'   with log-posterior drop below this threshold (relative to the maximum) will
#'   be excluded from the fit. Defaults to `-6`.
#' @param sn_fit_temp Temperature parameter for fitting the skew-normal.
#'   Defaults to `1` (weights are the density values themselves). If `NA`, the
#'   temperature is included as an additional optimisation parameter.
#' @param sn_fit_sample Logical. When `TRUE` (default), a parametric skew-normal
#'   is fitted to the posterior samples for covariance and defined parameters.
#'   When `FALSE`, these are summarised using kernel density estimation instead.
#' @param control A list of control parameters for the optimiser.
#' @param verbose Logical indicating whether to print progress messages.
#' @param debug Logical indicating whether to return debug information.
#' @param add_priors Logical indicating whether to include prior densities in
#'   the posterior computation.
#' @param optim_method The optimisation method to use for finding the posterior
#'   mode. Options include `"nlminb"` (default), `"ucminf"`, and `"optim"`
#'   (BFGS).
#' @param numerical_grad Logical indicating whether to use numerical gradients
#'   for the optimisation. Defaults to `FALSE` to use analytical gradients.
#' @param start Optional numeric vector of starting values for the optimiser,
#'   given as a full vector of free parameters in the internal (unconstrained)
#'   parameterisation. Mainly for internal use by [update()], which warm-starts
#'   mode-finding from a previous fit's posterior mode; supplying a hand-built
#'   vector requires knowledge of the internal parameter ordering. Its length
#'   must equal the number of free parameters or an error is raised.
#' @param cores Integer or `NULL`. Number of cores for parallel marginal
#'   fitting. When `NULL` (default), serial execution is used unless the number
#'   of free parameters exceeds 120, in which case parallelisation is enabled
#'   automatically using all available physical cores. Set to `1L` to force
#'   serial execution. If `cores > 1`, marginal fits are distributed across
#'   cores using [parallel::mclapply()] (fork-based; no parallelism on Windows).
#' @param ... Additional arguments to be passed to the [lavaan] model fitting
#'   function.
#'
#' @seealso Typically, users will interact with the specific latent variable
#'   model functions instead, including [acfa()], [asem()], and [agrowth()].
#'
#' @example inst/examples/ex-inlavaan.R
#'
#' @return An S4 object of class `INLAvaan` which is a subclass of the
#'   [lavaan-class] class.
#' @export
inlavaan <- function(
  model,
  data,
  model.type = "sem",
  dp = priors_for(),
  test = "standard",
  vb_correction = TRUE,
  marginal_method = c("skewnorm", "asymgaus", "marggaus", "sampling"),
  marginal_correction = c("shortcut", "shortcut_fd", "hessian", "none"),
  nsamp = 1000,
  samp_copula = TRUE,
  sn_fit_ngrid = 21,
  sn_fit_logthresh = -6,
  sn_fit_temp = 1,
  sn_fit_sample = TRUE,
  control = list(),
  verbose = TRUE,
  debug = FALSE,
  add_priors = TRUE,
  optim_method = c("nlminb", "ucminf", "optim"),
  numerical_grad = FALSE,
  start = NULL,
  cores = NULL,
  ...
) {
  mc <- match.call()
  mc$start <- NULL # warm start is transient; keep it out of the recorded call
  start.time0 <- proc.time()[3]
  timing <- list(start.time = start.time0)

  ## ----- Check arguments -----------------------------------------------------
  if (!is.null(cores)) {
    # nocov start
    cores <- as.integer(cores)
    if (is.na(cores) || cores < 1L) cores <- 1L
  } # nocov end
  marginal_method <- match.arg(marginal_method)
  if (isFALSE(marginal_correction)) {
    marginal_correction <- "none"
  } else {
    marginal_correction <- match.arg(marginal_correction)
  }
  optim_method <- match.arg(optim_method)
  if (isTRUE(debug)) {
    verbose <- TRUE
  }
  # "loo" is INLAvaan-specific: strip it before `test` reaches lavaan
  do_loo <- "loo" %in% test
  test <- setdiff(test, "loo")
  if (length(test) == 0L) {
    test <- "none"
  }

  lavargs <- list(...)
  lavargs$model <- model
  lavargs$data <- data
  lavargs$ceq.simple <- TRUE # FIXME: Force ceq.simple rather than eq.constraints
  lavargs$verbose <- FALSE # FIXME: Need some quiet mode maybe
  lavargs$do.fit <- FALSE
  lavargs$parser <- "old" # To get priors parsed
  lavargs$test <- test

  if ("estimator" %in% names(lavargs)) {
    if (!(lavargs$estimator %in% c("ML", "PML"))) {
      # nocov
      cli_abort("Only 'ML' and 'PML' estimators are supported currently.")
    }
  }

  # Two-level models cannot drop the mean structure: the between-level
  # statistics are the cluster means, and the marginalised treatment of
  # covariance-only analyses is single-level (and multigroup) only. lavaan
  # force-enables meanstructure for multilevel data anyway; if the user
  # *explicitly* asked for FALSE, say so rather than silently complying.
  if (isFALSE(lavargs$meanstructure) && !is.null(lavargs$cluster)) {
    cli_warn(c(
      "Two-level models require a mean structure; fitting with
       {.code meanstructure = TRUE}.",
      "i" = "See {.code vignette(\"meanstructure\", package = \"INLAvaan\")}
       for how INLAvaan treats mean structures."
    ))
    lavargs$meanstructure <- TRUE
  }

  ## ----- Initialise lavaan object --------------------------------------------
  fit0 <- do.call(get(model.type, envir = asNamespace("lavaan")), lavargs)
  if (length(fit0@Data@ordered) > 0) {
    # Redo automatically with PML if ordinal data
    lavargs$estimator <- "PML"
    lavargs$parameterization <- "theta"
    lavargs$test <- "none"
    fit0 <- do.call(get(model.type, envir = asNamespace("lavaan")), lavargs)
  }
  lavmodel <- fit0@Model
  lavsamplestats <- fit0@SampleStats
  lavdata <- fit0@Data
  lavoptions <- fit0@Options
  lavpartable <- fit0@ParTable
  lavcache <- fit0@Cache
  n <- fit0@SampleStats@ntotal
  ceq.simple <- lavmodel@ceq.simple.only
  ceq.K <- lavmodel@ceq.simple.K # used to pack params/grads

  # lavaan < 0.7-1.2707 computes a slightly inexact two-level FIML gradient
  # for cases fully missing on the within-level variables: it keeps such
  # cases but its gradient kernel mishandles the zero-observed pattern, so
  # the optimiser uses a slightly wrong gradient (fixed upstream in lavaan
  # PR #581). loo()/waic() are unaffected (the cluster kernels drop these
  # rows), but the fitted estimates may be mildly off, so warn on affected
  # lavaan versions only. Drop the version gate and this block once the
  # minimum supported lavaan carries the fix.
  if (
    is_multilevel(lavdata) &&
      isTRUE(lavsamplestats@missing.flag) &&
      utils::packageVersion("lavaan") < "0.7.1.2707"
  ) {
    Xw <- lavdata@X[[1L]]
    bidx <- lavdata@Lp[[1L]]$between.idx[[2L]]
    wcols <- if (length(bidx) > 0L) {
      seq_len(ncol(Xw))[-bidx]
    } else {
      seq_len(ncol(Xw))
    }
    n_fm <- sum(rowSums(!is.na(Xw[, wcols, drop = FALSE])) == 0L)
    if (n_fm > 0L) {
      cli_warn(c(
        "{n_fm} case{?s} {?is/are} fully missing on the within-level
         variables.",
        "i" = "This version of lavaan computes a slightly inexact two-level
         FIML gradient for such cases, so the estimates may be mildly
         affected; update lavaan or remove these cases. {.fn loo} and
         {.fn waic} handle them correctly."
      ))
    }
  }

  # Partable and check for equality constraints
  pt <- inlavaanify_partable(lavpartable, dp, lavdata, lavoptions)
  PTFREEIDX <- which(pt$free > 0L)
  if (isTRUE(ceq.simple)) {
    # Note: Always work in the reduced space
    PTFREEIDX <- which(pt$free > 0L & !duplicated(pt$free))
  }
  m <- length(PTFREEIDX)
  parnames <- pt$names[PTFREEIDX]

  # Cache partable for prior logdens and grad
  prior_cache <- prepare_priors_for_optim(pt)

  # Saturated-means fast path (see saturated_mean_idx): along the free
  # intercept axes the posterior is exactly Gaussian and block-diagonal at
  # the mode, so the Hessian block is analytic and the marginal scans are
  # redundant for those coordinates.
  fastpath <- saturated_mean_idx(
    pt,
    lavmodel,
    lavsamplestats,
    lavdata,
    ceq.simple
  )
  fp_idx <- if (is.null(fastpath)) integer(0) else fastpath$idx

  ## ----- Prep work for approximation -----------------------------------------
  joint_lp <- function(pars) {
    if (isTRUE(ceq.simple)) {
      pars_unpacked <- as.numeric(ceq.K %*% pars)
      x <- pars_to_x(pars_unpacked, pt)
    } else {
      x <- pars_to_x(pars, pt)
    }
    ll <- inlav_model_loglik(
      x,
      lavmodel,
      lavsamplestats,
      lavdata,
      lavoptions,
      lavcache
    )
    pld <- 0
    if (isTRUE(add_priors)) {
      # Always take in packed version
      # pld <- prior_logdens(pars, pt)
      pld <- prior_logdens_vectorized(pars, prior_cache, debug = FALSE)
    }
    as.numeric(ll + pld)
  }

  joint_lp_grad <- function(pars) {
    # First, the likelihood gradient
    if (isTRUE(ceq.simple)) {
      pars_unpacked <- as.numeric(ceq.K %*% pars)
      x <- pars_to_x(pars_unpacked, pt)
      jcb <- mapply(
        function(f, x) f(x),
        pt$ginv_prime[pt$free > 0],
        pars_unpacked
      )
    } else {
      x <- pars_to_x(pars, pt)
      jcb <- mapply(function(f, x) f(x), pt$ginv_prime[pt$free > 0], pars)
    }
    gll <- inlav_model_grad(x, lavmodel, lavsamplestats, lavdata, lavcache)

    # Jacobian adjustment: d/dθ log p(y|x(θ)) = d/dx log p(y|x) * dx/dθ
    jcb <- diag(jcb, length(jcb))

    # This is the extra jacobian adjustment for covariances, since dx/dθ affects
    # more than one place if covariance exists
    sd1sd2 <- attr(x, "sd1sd2")
    jcb <- jcb * sd1sd2 # this adjusts the correlation parameters (diagonals)
    jcb_mat <- attr(x, "jcb_mat")

    if (!is.null(jcb_mat)) {
      for (k in seq_len(nrow(jcb_mat))) {
        i <- jcb_mat[k, 1]
        j <- jcb_mat[k, 2]
        jcb[i, j] <- jcb_mat[k, 3]
      }
    }
    gll_th <- as.numeric(jcb %*% gll) # this adjusts the cov parameters, if any
    if (isTRUE(ceq.simple)) {
      gll_th <- as.numeric(gll_th %*% ceq.K)
    } # Repack

    # Next, the prior gradient
    glp_th <- 0
    if (isTRUE(add_priors)) {
      # Always take in packed version
      # glp_th <- prior_grad(pars, pt)
      glp_th <- prior_grad_vectorized(pars, prior_cache)
    }

    as.numeric(gll_th + glp_th)
  }

  timing <- add_timing(timing, "init")

  ## ----- Start optimisation --------------------------------------------------
  if (isTRUE(verbose)) {
    optim_stage <- "Mode finding and Hessian computation"
    cli_progress_step(
      "{optim_stage}.",
      msg_done = "Posterior mode and Hessian."
    )
  }

  ob <- function(x) -1 * joint_lp(x)
  gr <- if (isTRUE(numerical_grad)) NULL else function(x) -1 * joint_lp_grad(x)
  parstart <- pt$parstart[PTFREEIDX]

  # Warm start: `start` is a full vector of free parameters in the internal
  # (unconstrained) parameterisation, e.g. the posterior mode of an earlier
  # fit reused by `update()`. Only valid when the parameter structure matches.
  if (!is.null(start)) {
    if (length(start) != length(parstart)) {
      cli_abort(c(
        "{.arg start} has length {length(start)} but the model has
         {length(parstart)} free parameter{?s}.",
        "i" = "{.arg start} must be a full vector of free parameters in the
               internal (unconstrained) parameterisation."
      ))
    }
    parstart <- as.numeric(start)
  }

  if (optim_method == "nlminb") {
    opt <- nlminb(
      start = parstart,
      objective = ob,
      gradient = gr,
      control = control
    )
    theta_star <- opt$par
    if (isTRUE(verbose)) {
      optim_stage <- "Computing the Hessian"
      cli_progress_update()
    }
    if (isTRUE(numerical_grad)) {
      H_neg <- fast_hessian(ob, theta_star)
    } else if (length(fp_idx)) {
      # assemble in blocks: finite differences over the covariance columns
      # only; the intercept block is analytic (n Sigma^{-1} restricted to
      # the free-intercept variables, plus the prior precision) and the
      # cross block is exactly zero at the mode
      cc <- setdiff(seq_len(m), fp_idx)
      H_neg <- fast_jacobian(
        function(x) -1 * joint_lp_grad(x),
        theta_star,
        cols = cc
      )
      H_neg[fp_idx, ] <- 0
      Sg_hat <- lavaan::lav_model_implied(
        lavaan::lav_model_set_parameters(lavmodel, pars_to_x(theta_star, pt))
      )$cov[[1L]]
      sp <- fastpath$sigma_pos
      H_neg[fp_idx, fp_idx] <- n *
        chol2inv(chol(Sg_hat))[sp, sp, drop = FALSE] +
        diag(fastpath$prec, length(fp_idx))
    } else {
      # H_neg <- numDeriv::jacobian(function(x) -1 * joint_lp_grad(x), theta_star)
      H_neg <- fast_jacobian(function(x) -1 * joint_lp_grad(x), theta_star)
    }
  } else if (optim_method == "ucminf") {
    # nocov start
    if (!requireNamespace("ucminf", quietly = TRUE)) {
      cli_abort(
        "The `ucminf` package is required for this optimization method. Please install it using `install.packages('ucminf')`."
      )
    }

    opt <- ucminf::ucminf(
      par = parstart,
      fn = ob,
      gr = gr,
      control = list(),
      hessian = 1
    )
    theta_star <- opt$par
    H_neg <- opt$hessian
  } else {
    # nocov end
    opt <- stats::optim(
      par = parstart,
      fn = ob,
      gr = gr,
      method = "BFGS",
      hessian = TRUE,
      control = list()
    )
    theta_star <- opt$par
    H_neg <- opt$hessian
  }
  # Cholesky-factorise the precision (neg. Hessian), then derive covariance
  # via triangular backsolve. We first sort parameters into a canonical order
  # (by name) so results don't depend on the latent-variable ordering in the
  # model specification string.
  H_sym <- 0.5 * (H_neg + t(H_neg))
  canon_perm <- order(parnames)
  inv_perm <- order(canon_perm)
  H_canon <- H_sym[canon_perm, canon_perm]
  R_prec <- chol(H_canon) # upper Cholesky of canonical precision
  L_canon <- backsolve(R_prec, diag(m)) # L_c L_c^T = Sigma_canon (upper tri)
  L <- L_canon[inv_perm, ] # rows back to original param order
  Sigma_theta <- tcrossprod(L) # reconstruct covariance
  dimnames(Sigma_theta) <- list(parnames, parnames)
  lp_max <- joint_lp(theta_star) # before correction

  Vscan <- sweep(Sigma_theta, 2, sqrt(diag(Sigma_theta)), "/")

  # Derivatives at optima
  opt$dx <- fast_grad(function(x) -1 * joint_lp(x), theta_star) # fd grad
  opt$dx_analytic <- -1 * joint_lp_grad(theta_star) # analytic grad
  if (isTRUE(debug)) {
    tab <- data.frame(
      analytic = round(opt$dx_analytic, 6),
      fd = round(opt$dx, 6),
      diff = round(opt$dx_analytic - opt$dx, 6),
      row.names = parnames
    )
    cli::cli_rule(left = "{.strong Gradient check at posterior mode}")
    print(tab)
    cli::cli_rule()
  }

  timing <- add_timing(timing, "optim")

  ## ----- VB correction -------------------------------------------------------
  vb_opt <- vb_shift <- vb_kld <- vb_kld_global <- n_qmc <- NA
  if (isTRUE(vb_correction)) {
    if (isTRUE(verbose)) {
      cli_progress_step(
        "Performing VB correction.",
        msg_done = "VB correction; mean |\U03B4| = {formatC(mean(abs(vb_shift) / sqrt(diag(Sigma_theta))),
                    format = 'f', digits = 3)}\U03C3."
      )
    }

    # QMC noise (scrambled Sobol); scale n with dimension
    n_qmc <- min(100L, max(30L, m + 20L))
    us <- sobol_owen(n = n_qmc, d = m)
    zs <- rbind(0, qnorm(us) %*% t(L)) # Add 0 to "lock at" mode

    vb_ob <- function(delta, mu0, Z) {
      mu_new <- mu0 + as.numeric(L %*% delta)
      ns <- nrow(Z)
      lp_total <- 0
      for (b in seq_len(ns)) {
        thetab <- mu_new + Z[b, , drop = TRUE]
        lp_total <- lp_total + joint_lp(thetab)
      }
      -1 * (lp_total / ns)
    }

    vb_gr <- function(eta, mu0, Z) {
      mu_new <- mu0 + as.numeric(L %*% eta)
      ns <- nrow(Z)
      lpgrad_total <- numeric(length(mu0))
      for (b in seq_len(ns)) {
        thetab <- mu_new + Z[b, , drop = TRUE]
        lpgrad_total <- lpgrad_total + joint_lp_grad(thetab)
      }
      lpgrad_avg <- -1 * lpgrad_total / ns
      as.numeric(t(L) %*% lpgrad_avg)
    }

    vb_opt <- nlminb(
      start = rep(0, m),
      objective = vb_ob,
      gradient = vb_gr,
      mu0 = theta_star,
      Z = zs,
      control = list(rel.tol = 1e-4)
    )

    vb_shift <- as.numeric(L %*% vb_opt$par)
    vb_kld <- (vb_shift)^2 / (2 * diag(Sigma_theta))
    vb_kld_global <- lp_max + vb_opt$objective
  }

  vb <- list(
    opt = vb_opt,
    n_qmc = n_qmc,
    correction = vb_shift,
    kld = vb_kld,
    kld_global = vb_kld_global
  )
  timing <- add_timing(timing, "vb")

  ## ----- Info at optima ------------------------------------------------------
  theta_star_vbc <- theta_star
  if (isTRUE(vb_correction)) {
    theta_star_vbc <- theta_star + vb_shift
  }
  if (ceq.simple) {
    theta_star_trans <- pars_to_x(as.numeric(ceq.K %*% theta_star_vbc), pt)
  } else {
    theta_star_trans <- pars_to_x(theta_star_vbc, pt)
  }

  # Marginal log-likelihood (for BF comparison)
  # log det(Sigma) = -2 sum(log(diag(R_prec))) from the precision Cholesky
  mloglik <- lp_max + (m / 2) * log(2 * pi) - sum(log(diag(R_prec)))
  if (isTRUE(vb_correction)) {
    mloglik <- mloglik - vb_kld_global
  }
  timing <- add_timing(timing, "loglik")

  ## ----- Marginal approximations ---------------------------------------------
  if (isTRUE(verbose)) {
    cli_progress_done()
  }

  # pars_list <- setNames(as.list(1:m), paste0("pars[", 1:m, "]"))
  pars_list <- setNames(as.list(1:m), parnames)
  visual_debug <- NULL

  # When asymgaus or skewnorm marginals, we need the correction factor gamma1
  if (marginal_method %in% c("asymgaus", "skewnorm")) {
    # Step size for finite difference / central difference
    delta_outer <- 0.01 # for rate of change of Hessian (3rd deriv)
    delta_inner <- 0.001 # for rate of change of gradients (2nd deriv)

    get_gamma1 <- function(.j) {
      compute_gamma1j(
        j = .j,
        method = marginal_correction,
        theta_star = theta_star,
        Vscan = Vscan,
        L = L,
        joint_lp_grad = joint_lp_grad,
        delta_outer = delta_outer,
        delta_inner = delta_inner,
        m = m
      )
    }
  }

  if (marginal_method == "sampling") {
    approx_data <- NULL
  } else {
    # --- Resolve effective core count for marginal fitting ------------------
    if (is.null(cores)) {
      # Auto: serial for small m, parallel for large m
      if (m > 120L) {
        eff_cores <- parallel::detectCores(logical = FALSE)
        if (is.na(eff_cores) || eff_cores < 2L) eff_cores <- 1L
      } else {
        eff_cores <- 1L
      }
    } else {
      eff_cores <- cores
    }
    if (eff_cores > 1L && .Platform$OS.type == "windows") {
      # nocov start
      cli_alert_warning(
        "Parallel marginal fitting uses forking and is not available on
        Windows. Falling back to serial."
      )
      eff_cores <- 1L
    } # nocov end
    eff_cores <- min(eff_cores, m)

    if (marginal_method == "asymgaus") {
      obtain_approx_data <- function(j) {
        # Gauge the drop in joint_lp in whitened Z space
        k <- 2
        gamma1j <- get_gamma1(j)
        dplus <- max(
          0.01,
          lp_max - joint_lp(theta_star + Vscan[, j] * k) + gamma1j * k
        )
        dminus <- max(
          0.01,
          lp_max - joint_lp(theta_star - Vscan[, j] * k) + gamma1j * k
        )
        c(
          sigma_plus = sqrt(k^2 / (2 * dplus)),
          sigma_minus = sqrt(k^2 / (2 * dminus))
        )
      }

      approx_data <- run_parallel_or_serial(
        m = m,
        FUN = obtain_approx_data,
        cores = eff_cores,
        verbose = verbose,
        msg_serial = "Calibrating {j}/{m} asymmetric Gaussian{?s}.",
        msg_parallel = "Calibrating {done}/{m} asymmetric Gaussians ({cores}\U00D7).",
        msg_done = "Calibrate {m}/{m} asymmetric Gaussian{?s}."
      )
      approx_data <- do.call(what = "rbind", approx_data)

      post_marg <- function(j, g, g_prime, ginv, ginv_prime) {
        post_marg_asymgaus(
          j = j,
          g = g,
          g_prime = g_prime,
          ginv = ginv,
          ginv_prime = ginv_prime,
          theta_star = theta_star_vbc,
          Sigma_theta = Sigma_theta,
          sigma_asym = approx_data
        )
      }
    } else if (marginal_method == "skewnorm") {
      obtain_approx_data <- function(j) {
        if (j %in% fp_idx) {
          # saturated-means fast path: this axis is exactly Gaussian, and
          # the scan would reproduce the Laplace marginal to numerical
          # precision -- emit it directly
          return(list(
            fit = c(
              xi = theta_star[j],
              omega = sqrt(Sigma_theta[j, j]),
              alpha = 0,
              logC = 0,
              k = 0,
              rmse = 0,
              nmad = 0,
              gamma1 = 0
            ),
            visual_debug = NULL
          ))
        }
        z <- seq(-4, 4, length = sn_fit_ngrid)
        yync <- yy <- numeric(length(z))
        gamma1j <- get_gamma1(j)

        for (k in seq_along(z)) {
          yync[k] <- joint_lp(theta_star + Vscan[, j] * z[k])
          yy[k] <- yync[k] + gamma1j * z[k]
        }

        fit_sn <- fit_skew_normal(
          x = z,
          y = yy - max(yy),
          threshold_log_drop = sn_fit_logthresh,
          temp = sn_fit_temp
        )

        vd <- data.frame(
          x = z,
          Original = exp(yync - max(yync)),
          Corrected = exp(yy - max(yy)),
          SN_Fit = dsnorm(
            x = z,
            xi = fit_sn$xi,
            omega = fit_sn$omega,
            alpha = fit_sn$alpha,
            logC = fit_sn$logC
          )
        )

        # Adjust back to theta space
        fit_sn$xi <- theta_star[j] + fit_sn$xi * sqrt(Sigma_theta[j, j])
        fit_sn$omega <- fit_sn$omega * sqrt(Sigma_theta[j, j])

        list(fit = c(unlist(fit_sn), gamma1 = gamma1j), visual_debug = vd)
      }

      all_results <- run_parallel_or_serial(
        m = m,
        FUN = obtain_approx_data,
        cores = eff_cores,
        verbose = verbose,
        msg_serial = "Fitting {j}/{m} skew-normal marginal{?s}.",
        msg_parallel = "Fitting {done}/{m} skew-normal marginals ({cores}\U00D7).",
        msg_done = "Fit {m}/{m} skew-normal marginal{?s}."
      )

      approx_data <- do.call(what = "rbind", lapply(all_results, `[[`, "fit"))
      rownames(approx_data) <- parnames
      visual_debug <- lapply(all_results, `[[`, "visual_debug")
      names(visual_debug) <- parnames

      post_marg <- function(j, g, g_prime, ginv, ginv_prime) {
        post_marg_skewnorm(
          j = j,
          g = g,
          g_prime = g_prime,
          ginv = ginv,
          ginv_prime = ginv_prime,
          theta_star = theta_star_vbc,
          Sigma_theta = Sigma_theta,
          sn_params = approx_data
        )
      }
    } else if (marginal_method == "marggaus") {
      approx_data <- NULL

      post_marg <- function(j, g, g_prime, ginv, ginv_prime) {
        post_marg_marggaus(
          j = j,
          g = g,
          g_prime = g_prime,
          ginv = ginv,
          ginv_prime = ginv_prime,
          theta_star = theta_star_vbc,
          Sigma_theta = Sigma_theta
        )
      }
    }

    # Compute posterior marginals ----------------------------------------------
    postmargres <- Map(
      f = post_marg,
      j = seq_len(m),
      g = pt$g[PTFREEIDX],
      g_prime = pt$g_prime[PTFREEIDX],
      ginv = pt$ginv[PTFREEIDX],
      ginv_prime = pt$ginv_prime[PTFREEIDX]
    )
  }
  timing <- add_timing(timing, "marginals")

  ## ----- NORTA adjustment for SN copula sampling ----------------------------
  R_star <- NULL
  if (marginal_method == "skewnorm" && isTRUE(samp_copula)) {
    if (isTRUE(verbose)) {
      cli_progress_step(
        "Adjusting copula correlations (NORTA).",
        msg_done = "Adjust copula correlations (NORTA)."
      )
    }
    R_star <- norta_adjust_R(cov2cor(Sigma_theta), approx_data)
  }
  timing <- add_timing(timing, "norta")

  ## ----- Draw posterior samples (once) ---------------------------------------
  # Draw-based summaries: covariances, defined (:=) and delta (~*~) parameters,
  # or (for the pure sampling method) every marginal
  needs_draw_summaries <-
    marginal_method == "sampling" ||
    sum(pt$free > 0 & grepl("cov", pt$mat)) > 0 ||
    any(pt$op == ":=") ||
    any(pt$op == "~*~")
  has_extra_samp_work <- needs_draw_summaries || test != "none"
  samp_env <- NULL
  if (isTRUE(verbose)) {
    samp_stage <- if (has_extra_samp_work) {
      "Posterior sampling and summarising"
    } else {
      "Drawing posterior samples"
    }
    # Rewritten at the end of the block with an inventory of what the draws
    # were used for
    samp_done <- paste0(samp_stage, ".")
    samp_env <- environment()
    cli_progress_step(
      "{samp_stage}.",
      msg_done = "{samp_done}",
      spinner = TRUE,
      .envir = samp_env
    )
  }
  samp <- sample_params(
    theta_star = theta_star_vbc,
    Sigma_theta = Sigma_theta,
    method = if (isTRUE(samp_copula)) marginal_method else "sampling",
    approx_data = approx_data,
    pt = pt,
    lavmodel = lavmodel,
    nsamp = nsamp,
    R_star = R_star
  )
  theta_samp <- samp$theta_samp
  x_samp <- samp$x_samp
  vcov_x <- cov(x_samp)
  dimnames(vcov_x) <- list(parnames, parnames)
  timing <- add_timing(timing, "sampling")

  if (marginal_method == "sampling") {
    postmargres <- post_marg_sampling(x_samp)
  }

  summ <- do.call(
    "rbind",
    Map(
      f = function(x, y) {
        out <- t(data.frame(x$summary))
        row.names(out) <- y
        out
      },
      x = postmargres,
      y = parnames
    )
  )
  summ <- cbind(
    summ,
    kld = vb$kld,
    vb_shift_sigma = vb$correction / sqrt(diag(Sigma_theta))
  )

  pdf_data <- lapply(postmargres, function(x) x$pdf_data)
  names(pdf_data) <- parnames

  coefs <- summ[, "Mean"]
  names(coefs) <- parnames

  summ <- as.data.frame(summ)
  summ$Prior <- pt$prior[PTFREEIDX]

  ## ----- Sampling for covariances and defined params -------------------------
  if (sum(pt$free > 0 & grepl("cov", pt$mat)) > 0) {
    if (marginal_method == "sampling") {
      # Already covered by post_marg_sampling above
    } else {
      if (marginal_method == "skewnorm" && isTRUE(sn_fit_sample)) {
        samp_cov <- sample_covariances_fit_sn(x_samp, pt)
        sn_rows <- do.call(rbind, lapply(samp_cov, `[[`, "sn_params"))
        approx_data <- rbind(approx_data, sn_rows)
      } else {
        samp_cov <- sample_covariances(x_samp, pt)
      }

      for (cov_name in names(samp_cov)) {
        tmp_new_summ <- samp_cov[[cov_name]]$summary
        summ[cov_name, names(tmp_new_summ)] <- tmp_new_summ
        # keep the coefficient vector on the covariance scale too; the
        # per-parameter marginal mean is on the correlation (tanh) scale
        coefs[cov_name] <- tmp_new_summ[["Mean"]]
        pdf_data[[cov_name]] <- samp_cov[[cov_name]]$pdf_data
      }
    }
  }
  timing <- add_timing(timing, "covariances")

  # Defined parameters
  if (any(pt$op == ":=")) {
    if (marginal_method == "skewnorm" && isTRUE(sn_fit_sample)) {
      # nocov start
      defpars <- get_defpars_fit_sn(x_samp, pt)
      sn_rows <- do.call(rbind, lapply(defpars, `[[`, "sn_params"))
      approx_data <- rbind(approx_data, sn_rows)
    } else {
      # nocov end
      defpars <- get_defpars(x_samp, pt)
    }

    for (def_name in names(defpars)) {
      tmp_new_summ <- defpars[[def_name]]$summary
      summ[def_name, names(tmp_new_summ)] <- tmp_new_summ
      pdf_data[[def_name]] <- defpars[[def_name]]$pdf_data
    }
  }
  timing <- add_timing(timing, "definedpars")

  # For binary and ordinal data, sample the deltas
  if (any(pt$op == "~*~")) {
    deltapars <- get_thetaparamerization_deltas(x_samp, lavmodel)
    names(deltapars) <- pt$names[which(pt$op == "~*~")]

    for (delta_name in names(deltapars)) {
      tmp_new_summ <- deltapars[[delta_name]]$summary
      summ[delta_name, names(tmp_new_summ)] <- tmp_new_summ
      pdf_data[[delta_name]] <- deltapars[[delta_name]]$pdf_data
    }
  }
  timing <- add_timing(timing, "deltapars")

  ## ----- Compute ppp and dic -------------------------------------------------
  if (test != "none") {
    if (isTRUE(verbose)) {
      samp_stage <- "Computing fit indices (PPP/DIC)"
      cli_progress_update(.envir = samp_env)
    }
    ppp <- get_ppp(
      x_samp = x_samp,
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavpartable = lavpartable,
      cli_env = samp_env
    )
    dic_list <- get_dic(
      x_samp = x_samp,
      theta_star = theta_star_vbc,
      pt = pt,
      lavmodel = lavmodel,
      loglik = function(x) {
        inlav_model_loglik(
          x,
          lavmodel,
          lavsamplestats,
          lavdata,
          lavoptions,
          lavcache
        )
      },
      cli_env = samp_env
    )
  } else {
    ppp <- dic_list <- NULL
  }
  timing <- add_timing(timing, "test")

  ## ----- Fit-time LOO and WAIC -------------------------------------------------
  # Minimal internal view of the fit for the casewise machinery
  int_fit <- list(
    partable = pt,
    lavmodel = lavmodel,
    lavdata = lavdata,
    lavsamplestats = lavsamplestats,
    theta_star = as.numeric(theta_star_vbc),
    Sigma_theta = Sigma_theta,
    marginal_method = marginal_method,
    approx_data = approx_data,
    nsamp = nsamp,
    R_star = R_star
  )
  # The default path (test = "standard") computes LOO/WAIC only for models
  # the casewise kernels support, quietly, and (for LOO) only when the
  # predicted serial cost fits a 10 s budget. An explicit "loo" in `test`
  # always computes the full LOO.
  casewise_ok <- tryCatch(
    {
      suppressWarnings(check_loo_model(int_fit))
      TRUE
    },
    error = function(e) FALSE
  )

  loo_res <- NULL
  if (isTRUE(do_loo) || (test != "none" && casewise_ok)) {
    if (isTRUE(verbose)) {
      samp_stage <- "Computing Taylor LOO"
      cli_progress_update(.envir = samp_env)
    }
    loo_res <- tryCatch(
      inlav_loo(
        int = int_fit,
        eff_cores = resolve_loo_cores(cores),
        verbose = FALSE,
        max_seconds = if (isTRUE(do_loo)) Inf else 10
      ),
      inlavaan_loo_budget = function(e) {
        if (isTRUE(verbose)) {
          cli_alert_info(
            "Skipping fit-time LOO (predicted cost exceeds 10 s); compute it
             post hoc with {.fn loo} or {.fn add_loo}."
          )
        }
        NULL
      },
      error = function(e) {
        if (isTRUE(do_loo)) {
          cli_warn(c(
            "Skipping the fit-time LOO computation.",
            "x" = conditionMessage(e)
          ))
        }
        NULL
      }
    )
    timing <- add_timing(timing, "loo")
  }

  # WAIC from the draws already produced above, so only the casewise pass is
  # paid; skipped for small nsamp, where p_waic estimates are meaningless.
  # The reliability warning is left to print/explicit waic() calls.
  waic_res <- NULL
  if (test != "none" && casewise_ok && nsamp >= 100L) {
    if (isTRUE(verbose)) {
      samp_stage <- "Computing WAIC"
      cli_progress_update(.envir = samp_env)
    }
    waic_res <- tryCatch(
      suppressWarnings(
        waic_from_draws(
          int_fit,
          x_samp,
          eff_cores = resolve_loo_cores(cores)
        )
      ),
      error = function(e) NULL
    )
    timing <- add_timing(timing, "waic")
  }

  if (isTRUE(verbose)) {
    # Close the sampling step with an overview; the specific fit measures
    # computed are listed on a separate info line below
    fit_measures <- c(
      if (!is.null(ppp)) c("PPP", "DIC"),
      if (!is.null(loo_res)) "LOO",
      if (!is.null(waic_res)) "WAIC"
    )
    samp_done <- if (needs_draw_summaries || length(fit_measures)) {
      paste0("Summarise ", nsamp, " posterior draws.")
    } else {
      paste0("Draw ", nsamp, " posterior samples.")
    }
    cli_progress_done(.envir = samp_env)
    if (length(fit_measures)) {
      cli_alert_info(
        paste0("Fit measures: ", paste(fit_measures, collapse = ", "), ".")
      )
    }
  }

  ## ----- Output --------------------------------------------------------------
  out <- list(
    coefficients = coefs,
    mloglik = mloglik,
    DIC = dic_list,
    summary = summ,
    ppp = ppp,
    loo = loo_res,
    waic = waic_res,
    optim_method = optim_method,
    marginal_method = marginal_method,
    theta_star_novbc = as.numeric(theta_star),
    theta_star = as.numeric(theta_star_vbc),
    Sigma_theta = Sigma_theta,
    R_star = R_star,
    vcov_x = vcov_x,
    theta_star_trans = theta_star_trans,
    approx_data = approx_data,
    nsamp = nsamp,
    pdf_data = pdf_data,
    partable = pt,
    lavmodel = lavmodel,
    lavsamplestats = lavsamplestats,
    lavdata = lavdata,
    opt = opt,
    timing = timing[-1], # remove start.time
    visual_debug = visual_debug,
    vb = vb,
    call = mc,
    version = as.character(utils::packageVersion("INLAvaan"))
  )
  class(out) <- "inlavaan_internal"

  # Warn (once, consolidated) if the convergence/approximation diagnostics
  # look off; see warn_fit_diagnostics() for the checks and thresholds
  warn_fit_diagnostics(out)

  if (isTRUE(debug)) {
    return(out)
  } else {
    out <- create_lav_from_inlavaan_internal(fit0, out)
    return(new("INLAvaan", out))
  }
}

#' Fit an Approximate Bayesian Confirmatory Factor Analysis Model
#'
#' Fit an Approximate Bayesian Confirmatory Factor Analysis Model
#'
#' The [acfa()] function is a wrapper for the more general [inlavaan()]
#' function, using the following default arguments:
#'   - `int.ov.free = TRUE`
#'   - `int.lv.free = FALSE`
#'   - `auto.fix.first = TRUE` (unless `std.lv = TRUE`)
#'   - `auto.fix.single = TRUE`
#'   - `auto.var = TRUE`
#'   - `auto.cov.lv.x = TRUE`
#'   - `auto.efa = TRUE`
#'   - `auto.th = TRUE`
#'   - `auto.delta = TRUE`
#'   - `auto.cov.y = TRUE`
#'
#' For further information regarding these arguments, please refer to the
#' [lavaan::lavOptions()] documentation.
#'
#' @inherit inlavaan params return seealso
#' @example inst/examples/ex-cfa.R
#' @export
acfa <- function(
  model,
  data,
  dp = priors_for(),
  test = "standard",
  vb_correction = TRUE,
  marginal_method = c("skewnorm", "asymgaus", "marggaus", "sampling"),
  marginal_correction = c("shortcut", "shortcut_fd", "hessian", "none"),
  nsamp = 1000,
  samp_copula = TRUE,
  sn_fit_ngrid = 21,
  sn_fit_logthresh = -6,
  sn_fit_temp = 1,
  sn_fit_sample = TRUE,
  control = list(),
  verbose = TRUE,
  debug = FALSE,
  add_priors = TRUE,
  optim_method = c("nlminb", "ucminf", "optim"),
  numerical_grad = FALSE,
  cores = NULL,
  ...
) {
  sc <- sys.call()
  sc[["model.type"]] <- quote("cfa")
  sc[[1L]] <- quote(INLAvaan::inlavaan)
  eval(sc, parent.frame())
}

#' Fit an Approximate Bayesian Structural Equation Model
#'
#' Fit an Approximate Bayesian Structural Equation Model
#'
#' The [asem()] function is a wrapper for the more general [inlavaan()]
#' function, using the following default arguments:
#'   - `int.ov.free = TRUE`
#'   - `int.lv.free = FALSE`
#'   - `auto.fix.first = TRUE` (unless `std.lv = TRUE`)
#'   - `auto.fix.single = TRUE`
#'   - `auto.var = TRUE`
#'   - `auto.cov.lv.x = TRUE`
#'   - `auto.efa = TRUE`
#'   - `auto.th = TRUE`
#'   - `auto.delta = TRUE`
#'   - `auto.cov.y = TRUE`
#'
#' For further information regarding these arguments, please refer to the
#' [lavaan::lavOptions()] documentation.
#'
#' @inherit inlavaan params return seealso
#' @example inst/examples/ex-sem.R
#' @export
asem <- function(
  model,
  data,
  dp = priors_for(),
  test = "standard",
  vb_correction = TRUE,
  marginal_method = c("skewnorm", "asymgaus", "marggaus", "sampling"),
  marginal_correction = c("shortcut", "shortcut_fd", "hessian", "none"),
  nsamp = 1000,
  samp_copula = TRUE,
  sn_fit_ngrid = 21,
  sn_fit_logthresh = -6,
  sn_fit_temp = 1,
  sn_fit_sample = TRUE,
  control = list(),
  verbose = TRUE,
  debug = FALSE,
  add_priors = TRUE,
  optim_method = c("nlminb", "ucminf", "optim"),
  numerical_grad = FALSE,
  cores = NULL,
  ...
) {
  sc <- sys.call()
  sc[["model.type"]] <- quote("sem")
  sc[[1L]] <- quote(INLAvaan::inlavaan)
  eval(sc, parent.frame())
}

#' Fit an Approximate Bayesian Growth Curve Model
#'
#' Fit an Approximate Bayesian Growth Curve Model
#'
#' The [asem()] function is a wrapper for the more general [inlavaan()]
#' function, using the following default arguments:
#'   - `meanstructure = TRUE`
#'   - `int.ov.free = FALSE`
#'   - `int.lv.free = TRUE`
#'   - `auto.fix.first = TRUE` (unless `std.lv = TRUE`)
#'   - `auto.fix.single = TRUE`
#'   - `auto.var = TRUE`
#'   - `auto.cov.lv.x = TRUE`
#'   - `auto.efa = TRUE`
#'   - `auto.th = TRUE`
#'   - `auto.delta = TRUE`
#'   - `auto.cov.y = TRUE`
#'
#' @inherit inlavaan params return seealso
#' @example inst/examples/ex-growth.R
#' @export
agrowth <- function(
  model,
  data,
  dp = priors_for(),
  test = "standard",
  vb_correction = TRUE,
  marginal_method = c("skewnorm", "asymgaus", "marggaus", "sampling"),
  marginal_correction = c("shortcut", "shortcut_fd", "hessian", "none"),
  nsamp = 1000,
  samp_copula = TRUE,
  sn_fit_ngrid = 21,
  sn_fit_logthresh = -6,
  sn_fit_temp = 1,
  sn_fit_sample = TRUE,
  control = list(),
  verbose = TRUE,
  debug = FALSE,
  add_priors = TRUE,
  optim_method = c("nlminb", "ucminf", "optim"),
  numerical_grad = FALSE,
  cores = NULL,
  ...
) {
  sc <- sys.call()
  sc[["model.type"]] <- quote("growth")
  sc[[1L]] <- quote(INLAvaan::inlavaan)
  eval(sc, parent.frame())
}

testthat::skip_on_ci()
testthat::skip_on_cran()

gen_bivariate_data <- function(n, corr = 0.5) {
  Sigma_true <- matrix(c(1, corr, corr, 1), nrow = 2)
  U <- chol(Sigma_true)
  z_raw <- matrix(rnorm(n * ncol(Sigma_true)), nrow = n)
  x <- z_raw %*% U
  # x <- mvtnorm::rmvnorm(n, sigma = Sigma_true)
  data.frame(x1 = x[, 1], x2 = x[, 2])
}

mod <- "
    eta1 =~ 1*x1
    eta2 =~ 1*x2
  "

test_that("Comparison to MCMC", {
  suppressPackageStartupMessages(library(blavaan))
  the_prior <- "gamma(10,10)[sd]"

  dat100 <- gen_bivariate_data(100)
  fit100 <- asem(
    mod,
    dat100,
    verbose = FALSE,
    dp = blavaan::dpriors(psi = the_prior)
  )
  tmp <- suppressWarnings(capture.output(
    fit100_blav <<- bsem(
      mod,
      dat100,
      n.chains = 1,
      sample = 2000,
      dp = blavaan::dpriors(psi = the_prior)
    )
  ))
  expect_equal(coef(fit100), coef(fit100_blav), tolerance = 0.1)

  # fit2500 <- asem(mod, dat2500, verbose = FALSE)
  # tmp <- capture.output(fit2500_blav <<- bsem(mod, dat2500))
  # expect_equal(coef(fit2500), coef(fit2500_blav), tolerance = 0.1)
})

################################################################################
## CHECK AGAINST MCMC ##########################################################
################################################################################
testthat::skip_if_not(interactive())
testthat::skip_if_not_installed("rstan")
testthat::skip_if_not_installed("future")
library(rstan)
future::plan("multisession", workers = future::availableCores() - 2)
n <- 250

stan_mod <- "
data {
  int<lower=1> n;
  matrix[n, 2] x;
}

parameters {
  vector<lower=0>[2] sigma;
  real<lower=-1,upper=1> rho;
}

transformed parameters {
  cov_matrix[2] Sigma;
  Sigma[1,1] = square(sigma[1]);
  Sigma[2,2] = square(sigma[2]);
  Sigma[1,2] = rho * sigma[1] * sigma[2];
  Sigma[2,1] = Sigma[1,2];
  vector[2] sigma2;
  sigma2[1] = square(sigma[1]);
  sigma2[2] = square(sigma[2]);
}

model {
  sigma2 ~ gamma(1, 1);
  rho ~ uniform(-1, 1);
  for (i in 1:n)
    x[i] ~ multi_normal(rep_vector(0, 2), Sigma);
}
"
dat <- gen_bivariate_data(n)

stan_data <- list(
  n = nrow(dat),
  x = as.matrix(dat)
)

fit_stan <- stan(
  model_code = stan_mod,
  data = stan_data,
  cores = 3,
  iter = 10000
)
draws <- as.data.frame(rstan::extract(fit_stan, permuted = TRUE))
unrestricted_params <- c("Sigma.1.1", "Sigma.2.2", "Sigma.1.2")

fit_inlv <- asem(
  mod,
  dat,
  debug = TRUE,
  dp = blavaan::dpriors(psi = "gamma(1,1)")
  # sn_fit_temp = 1,
  # sn_fit_logthresh = -1
)

plot_df_stan <-
  draws |>
  select(all_of(unrestricted_params)) |>
  pivot_longer(everything()) |>
  mutate(
    name = factor(name, levels = unrestricted_params),
    name = factor(
      names(coef(fit_inlv))[match(name, unrestricted_params)],
      levels = names(coef(fit_inlv))
    )
  )

plot_df <-
  list(
    skewnorm = fit_inlv$pdf_data
  ) |>
  map(function(plot_df_list) {
    plot_df <-
      imap(plot_df_list, \(x, idx) mutate(x, name = idx)) |>
      bind_rows() |>
      mutate(name = factor(name, levels = names(coef(fit_inlv))))
  }) |>
  bind_rows(.id = "method") |>
  mutate(
    method = factor(
      method,
      levels = c("skewnorm", "asymgaus", "marggaus", "sampling")
    )
  )

ggplot() +
  geom_density(
    data = plot_df_stan,
    aes(value, fill = "MCMC"),
    col = NA,
    alpha = 0.38
  ) +
  geom_line(
    data = plot_df,
    aes(x, y, group = method, col = method),
    linewidth = 0.75
  ) +
  facet_wrap(~name, scales = "free") +
  scale_colour_manual(values = c("skewnorm" = "#00A6AA")) +
  scale_fill_manual(values = c("MCMC" = "#131516")) +
  theme_minimal() +
  theme(
    plot.margin = margin(t = 4, r = 8, b = 8, l = 8),
    legend.position = "top",
    legend.box.spacing = unit(2, "pt"),
    legend.key.width = unit(1, "cm")
  ) +
  labs(x = NULL, y = NULL, col = NULL, fill = NULL)

test_that("vcov and fitMeasures", {
  data("PoliticalDemocracy", package = "lavaan")
  model <- "
    # Latent variable definitions
    ind60 =~ x1 + x2 + x3
    dem60 =~ y1 + y2 + y3 + y4
    dem65 =~ y5 + y6 + y7 + y8

    # Latent regressions
    dem60 ~ ind60
    dem65 ~ ind60 + dem60

    # Residual correlations
    y1 ~~ y5
    y2 ~~ y4 + y6
    y3 ~~ y7
    y4 ~~ y8
    y6 ~~ y8
  "
  fit <- asem(
    model,
    PoliticalDemocracy,
    verbose = FALSE,
    marginal_method = "marggaus",
    marginal_correction = "none",
    nsamp = 2
  )

  expect_message(out <- capture.output(print(vcov(fit))))
  expect_no_error(out <- capture.output(fitmeasures(fit)))
})

test_that("Missing data", {
  data("PoliticalDemocracy", package = "lavaan")
  model <- "
    # Latent variable definitions
    ind60 =~ x1 + x2 + x3
    dem60 =~ y1 + y2 + y3 + y4
    dem65 =~ y5 + y6 + y7 + y8

    # Latent regressions
    dem60 ~ ind60
    dem65 ~ ind60 + dem60

    # Residual correlations
    y1 ~~ y5
    y2 ~~ y4 + y6
    y3 ~~ y7
    y4 ~~ y8
    y6 ~~ y8
  "

  # FIXME
  # ECM: Any possibility of a "full information" approach where you skip over
  # the NAs?
  set.seed(9619)
  mis <- matrix(
    rbinom(prod(dim(PoliticalDemocracy)), 1, .9),
    nrow(PoliticalDemocracy),
    ncol(PoliticalDemocracy)
  )
  pd <- PoliticalDemocracy * mis
  pd[pd == 0] <- NA

  fitm <- asem(
    model,
    pd,
    verbose = FALSE,
    marginal_method = "marggaus",
    marginal_correction = "none",
    missing = "ML",
    test = "none",
    nsamp = 2
  )

  expect_equal(fitm@Data@missing, "ml")
  # expect_equal(fitm@Data@nobs[[1]], nrow(pd[complete.cases(pd), ]))
})

test_that("Approximate constraints", {
  testthat::skip() # FIXME

  # ECM: Approximate constraints could be interesting
  HS.model <- "
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
    speed   =~ x7 + x8 + x9
  "

  expect_error({
    fit3 <- acfa(
      HS.model,
      data = lavaan::HolzingerSwineford1939,
      group = "school",
      group.equal = c("intercepts", "loadings"),
      wiggle = "intercepts",
      wiggle.sd = .1,
      verbose = FALSE
    )
  })
})

test_that("4-group model with a bunch of constraints and predict", {
  testthat::skip() # FIXME

  data(StereotypeThreat, package = "psychotools")
  StereotypeThreat <- transform(
    StereotypeThreat,
    group = interaction(ethnicity, condition)
  )
  StereotypeThreat <- StereotypeThreat[order(StereotypeThreat$group), ]

  mod <- "
    f1 =~ abstract + verbal + c(l1, l1, l1, l4)*numerical

    f1 ~  c(maj, min1, maj, min2)*1 + c(NA, 0, NA, 0)*1
    abstract ~ c(ar1, ar2, ar3, ar3)*1
    numerical ~ c(na1,na1,na1,na4)*1

    numerical ~~ c(e1, e1, e1, e4)*numerical
    f1 ~~ c(v1.maj, v1.min, v1.maj, v1.min)*f1
  "

  fit <- acfa(
    mod,
    StereotypeThreat,
    group = "group",
    group.equal = c("loadings", "residuals", "intercepts"),
    verbose = FALSE,
    # marginal_method = "marggaus",
    marginal_correction = "none",
    test = "none",
    nsamp = 2
  )
  prd <- predict(fit, nsamp = 3)
  expect_no_error(out <- capture.output(print(prd)))

  # lavPredict(fit) # ECM: I believe these are coming from lavaan
  # will not create inlavPredict(), just depend on predict
})


test_that("Constraint on correlations", {
  # ECM: This is a place where I think blavaan will constrain correlations
  # instead of covariances. That might not be the right thing to do, but
  # equality constraints on covariances are already weird.

  HS.modelc <- "
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
    speed   =~ x7 + x8 + x9

    x1 ~~ c(a, b)*x4
    x2 ~~ c(a, b)*x5
  "

  fit <- acfa(
    HS.modelc,
    data = lavaan::HolzingerSwineford1939,
    group = "school",
    dp = blavaan::dpriors(lambda = "normal(1,.5)"),
    verbose = FALSE,
    marginal_method = "marggaus",
    marginal_correction = "none",
    test = "none",
    nsamp = 2
  )

  expect_no_error(out <- capture.output(summary(fit)))
  expect_s4_class(fit, "INLAvaan")

  # Check it runs more than 1 group
  expect_true(length(coef(fit)) > 30)
  expect_equal(fit@Data@ngroups, 2L)
})

test_that("Standardised LVs with equality constraints", {
  # ECM: MCMC can choke on sign indeterminacy
  mod <- "
    visual  =~ c(l1,l1)*x1 + c(l2,l2)*x2 + c(l3,l3)*x3
    textual =~ c(l4,l4)*x4 + c(l5,l5)*x5 + c(l6,l6)*x6
    visual ~~ c(1, NA)*visual
    textual ~~ c(1, NA)*textual
  "

  fit <- acfa(
    mod,
    data = lavaan::HolzingerSwineford1939,
    group = "school",
    std.lv = TRUE,
    meanstructure = TRUE,
    verbose = FALSE,
    # marginal_method = "marggaus",
    marginal_correction = "none",
    test = "none",
    nsamp = 2
  )

  expect_no_error(out <- capture.output(summary(fit)))
  expect_s4_class(fit, "INLAvaan")
})

test_that("Growth models", {
  # ECM: Also see the examples at https://osf.io/4bpmq
  data("Demo.growth", package = "lavaan")
  colnames(Demo.growth)[1:4] <- c("X1", "X2", "X3", "X4")

  # Latent Change Score (LCS) specification with constant change
  constantX.syntax <- "
    # Latent variables
    lX1 =~ 1*X1
    lX2 =~ 1*X2
    lX3 =~ 1*X3
    lX4 =~ 1*X4

    # Autoregressions
    lX2 ~ 1*lX1
    lX3 ~ 1*lX2
    lX4 ~ 1*lX3

    # Change (constant)
    dX1 =~ 1*lX2
    dX2 =~ 1*lX3
    dX3 =~ 1*lX4

    # Random intercept and slope
    intX   =~ 1*lX1
    slopeX =~ 1*dX1 + 1*dX2 + 1*dX3

    # Residuals equal
    X1 ~~ rsdX*X1
    X2 ~~ rsdX*X2
    X3 ~~ rsdX*X3
    X4 ~~ rsdX*X4

    # Manifest means fixed to 0
    X1 ~ 0*1
    X2 ~ 0*1
    X3 ~ 0*1
    X4 ~ 0*1

    # Auto proportions
    dX1 ~ 0*lX1
    dX2 ~ 0*lX2
    dX3 ~ 0*lX3

    # Estimate slope and intercept means
    slopeX ~ 1
    intX ~ 1

    # Estimate latent variances and covariance
    slopeX ~~ slopeX
    intX ~~ intX
    slopeX ~~ intX

    # Means and variances of latent variables fixed to 0
    lX1 ~ 0*1
    lX2 ~ 0*1
    lX3 ~ 0*1
    lX4 ~ 0*1

    dX1 ~ 0*1
    dX2 ~ 0*1
    dX3 ~ 0*1

    lX1 ~~ 0*lX1
    lX2 ~~ 0*lX2
    lX3 ~~ 0*lX3
    lX4 ~~ 0*lX4

    dX1 ~~ 0*dX1
    dX2 ~~ 0*dX2
    dX3 ~~ 0*dX3
  "
  fit <- inlavaan(constantX.syntax, Demo.growth, verbose = FALSE)

  expect_no_error(out <- capture.output(summary(fit)))
  expect_s4_class(fit, "INLAvaan")
})

test_that("Non-recursive (cyclic) model", {
  set.seed(1234)
  pop_model <- "
    dv1 ~ 0.3*dv3
    dv2 ~ 0.5*dv1
    dv3 ~ 0.7*dv2
  "
  dat <- lavaan::simulateData(pop_model, sample.nobs = 500)
  fit <- asem(
    model = "dv1 ~ dv3; dv2 ~ dv1; dv3 ~ dv2",
    data = dat,
    dp = blavaan::dpriors(beta = "normal(.5,.5)"),
    verbose = FALSE
  )

  expect_no_error(out <- capture.output(summary(fit)))
  expect_s4_class(fit, "INLAvaan")
})

## Validation: Permutation invariance of marginal posteriors
## Verifies that reordering latent variables in the model syntax produces
## identical parameter estimates (up to numerical tolerance).

dat <- lavaan::HolzingerSwineford1939

mod1 <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
mod2 <- "
  speed   =~ x7 + x8 + x9
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"

## --- shortcut (Q2, Schur correction) ----------------------------------------
fit1 <- acfa(mod1, dat, std.lv = TRUE)
fit2 <- acfa(mod2, dat, std.lv = TRUE)

coef1 <- coef(fit1)
class(coef1) <- "numeric"
parnames <- names(coef1)

# lavaan may flip the order of covariance names
parnames[parnames == "visual~~speed"]   <- "speed~~visual"
parnames[parnames == "textual~~speed"]  <- "speed~~textual"
coef2 <- coef(fit2)[parnames]
class(coef2) <- "numeric"

cat("=== shortcut (Q2) ===\n")
cat(sprintf("Max |diff|: %.2e\n", max(abs(coef1 - coef2))))
tinytest::expect_equal(coef1, coef2, tolerance = 1e-4)

## --- shortcut_fd (Q2, forward difference) ------------------------------------
fit3 <- acfa(mod1, dat, std.lv = TRUE, marginal_correction = "shortcut_fd")
fit4 <- acfa(mod2, dat, std.lv = TRUE, marginal_correction = "shortcut_fd")

coef3 <- coef(fit3)
class(coef3) <- "numeric"
coef4 <- coef(fit4)[parnames]
class(coef4) <- "numeric"

cat("\n=== shortcut_fd (Q2, forward diff) ===\n")
cat(sprintf("Max |diff|: %.2e\n", max(abs(coef3 - coef4))))
tinytest::expect_equal(coef3, coef4, tolerance = 1e-4)

## --- Compare shortcut vs shortcut_fd -----------------------------------------
cat("\n=== shortcut vs shortcut_fd (same model) ===\n")
cat(sprintf("Max |diff|: %.2e\n", max(abs(coef1 - coef3))))

dat <- lavaan::HolzingerSwineford1939
mod <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"
fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none", 
            marginal_method = "marggaus", vb_correction = FALSE)

test_that("anova() on a single INLAvaan fit redirects to compare()", {
  expect_error(anova(fit), "compare")
})

test_that("anova() on two INLAvaan fits redirects to compare()", {
  expect_error(anova(fit, fit), "compare")
})

test_that("plain lavaan fits are unaffected by the INLAvaan anova() override", {
  fit_lav <- lavaan::cfa(mod, dat)
  expect_no_error(anova(fit_lav))
})

dat <- lavaan::HolzingerSwineford1939
mod <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"

test_that("vcov() returns a matrix for INLAvaan objects", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  vc <- suppressMessages(vcov(fit))
  expect_true(is.matrix(vc))
  expect_equal(nrow(vc), length(coef(fit)))
  expect_equal(ncol(vc), length(coef(fit)))
})

test_that("vcov(type = 'theta') returns Laplace covariance", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  vt <- vcov(fit, type = "theta")
  expect_true(is.matrix(vt))
  expect_true(isSymmetric(vt))
})

test_that("Method: skewnorm", {
  set.seed(1234)
  X <- rnorm(100)
  M <- 0.5 * X + rnorm(100)
  Y <- 0.7 * M + rnorm(100)

  dat <- data.frame(X = X, Y = Y, M = M)
  mod <- "
    # Direct effect
    Y ~ c*X

    # Mediators
    M ~ a*X
    Y ~ b*M

    # Indirect effect (a*b)
    ab := a*b

    # Total effect
    total := c + (a*b)
  "

  fit_lav <- lavaan::sem(mod, dat)
  expect_no_error({
    fit <- asem(
      mod,
      dat,
      verbose = FALSE
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
  expect_equal(coef(fit), coef(fit_lav), tolerance = 0.1)
  expect_equal(fit@optim$dx, rep(0, length(coef(fit))), tolerance = 1e-3)
})

dat <- lavaan::HolzingerSwineford1939
mod <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"

# Fit once, reuse (fast defaults)
fit <- acfa(
  mod,
  dat,
  verbose = FALSE,
  nsamp = 5,
  vb_correction = FALSE,
  test = "none",
  marginal_method = "marggaus"
)

test_that("simulate() returns list of correct length", {
  sims <- simulate(fit, nsim = 3)
  expect_true(is.list(sims))
  expect_length(sims, 3)
})

test_that("simulate() each element is a data frame with truth attr", {
  sims <- simulate(fit, nsim = 2)
  for (s in sims) {
    expect_true(is.data.frame(s))
    truth <- attr(s, "truth")
    expect_true(is.numeric(truth))
    expect_false(is.null(names(truth)))
    truth_theta <- attr(s, "truth_theta")
    expect_true(is.numeric(truth_theta))
    expect_false(is.null(names(truth_theta)))
    expect_equal(length(truth_theta), length(truth))
  }
})

test_that("simulate() data has correct dimensions (default nobs)", {
  sims <- simulate(fit, nsim = 1)
  expect_equal(nrow(sims[[1]]), nrow(dat))
  expect_equal(ncol(sims[[1]]), 6) # x1..x6
})

test_that("simulate() data has correct dimensions (custom nobs)", {
  sims <- simulate(fit, nsim = 1, sample.nobs = 50)
  expect_equal(nrow(sims[[1]]), 50)
})

test_that("simulate() truth attr names match coef()", {
  sims <- simulate(fit, nsim = 1)
  expect_equal(names(attr(sims[[1]], "truth")), names(coef(fit)))
})

test_that("simulate() with prior = TRUE works", {
  sims <- simulate(fit, nsim = 3, prior = TRUE)
  expect_length(sims, 3)
  for (s in sims) {
    expect_true(is.data.frame(s))
    expect_equal(nrow(s), nrow(dat))
  }
})

test_that("simulate() with seed is reproducible", {
  s1 <- simulate(fit, nsim = 1, seed = 42)
  s2 <- simulate(fit, nsim = 1, seed = 42)
  expect_equal(s1[[1]], s2[[1]])
  expect_equal(attr(s1[[1]], "truth"), attr(s2[[1]], "truth"))
  expect_equal(attr(s1[[1]], "truth_theta"), attr(s2[[1]], "truth_theta"))
})

test_that("simulate() nsim = 1 returns length-1 list", {
  sims <- simulate(fit, nsim = 1)
  expect_length(sims, 1)
})

test_that("simulate() with samp_copula = FALSE works", {
  sims <- simulate(fit, nsim = 2, samp_copula = FALSE)
  expect_length(sims, 2)
  expect_true(is.data.frame(sims[[1]]))
})

# Fixtures for the mean-structure behaviour tests below
vars <- paste0("x", 1:6)
ybar <- colMeans(dat[, vars])
fit_ms <- acfa(
  mod,
  dat,
  meanstructure = FALSE,
  verbose = FALSE,
  nsamp = 20,
  vb_correction = FALSE,
  test = "none",
  marginal_method = "marggaus"
)
fit_true <- acfa(
  mod,
  dat,
  meanstructure = TRUE,
  verbose = FALSE,
  nsamp = 20,
  vb_correction = FALSE,
  test = "none",
  marginal_method = "marggaus"
)

sim_col_means <- function(sims) {
  rowMeans(vapply(sims, function(d) colMeans(d[, vars]), numeric(length(vars))))
}

test_that("simulate() replicates live on the data scale without a mean structure", {
  # Regression test: simulateData() centres columns at zero when there is no
  # mean structure; the saturated means (integrated out under a flat prior)
  # must be added back so replicates share the observed location.
  sims <- simulate(fit_ms, nsim = 40, silent = TRUE)
  expect_lt(max(abs(sim_col_means(sims) - ybar)), 0.5)
})

test_that("simulate() does not double-shift when a mean structure is present", {
  # Guards the !meanstructure gate: with an explicit mean structure the
  # intercepts already place the data on scale, so no extra shift is applied.
  sims <- simulate(fit_true, nsim = 40, silent = TRUE)
  expect_lt(max(abs(sim_col_means(sims) - ybar)), 0.5)
})

test_that("simulate() propagates saturated-mean uncertainty N(ybar, Sigma/n)", {
  # A large sample.nobs shrinks within-replicate sampling noise (~0.02) so the
  # between-replicate spread reflects the mean-uncertainty draw (~0.07); a point
  # ybar shift (no uncertainty term) would leave spread near 0.02.
  set.seed(22)
  sims <- simulate(fit_ms, nsim = 60, sample.nobs = 4000, silent = TRUE)
  x1_means <- vapply(sims, function(d) mean(d$x1), numeric(1))
  expect_gt(sd(x1_means), 0.03)
})

test_that("simulate() prior predictive replicates stay centred at zero", {
  # The mean shift is posterior-only (flat prior on the means is improper);
  # prior = TRUE draws must remain on the latent, zero-centred scale. A wrongly
  # applied shift would land near ybar (>= 2.19); prior sampling noise stays
  # well under 1 (seed pinned for determinism).
  set.seed(99)
  sims <- simulate(fit_ms, nsim = 40, prior = TRUE, silent = TRUE)
  expect_lt(max(abs(sim_col_means(sims))), 1)
})

## ----- Shared fit (complete data) --------------------------------------------
mod_ml <- "
  level: 1
    fw =~ y1 + y2 + y3
    fw ~ x1 + x2 + x3
  level: 2
    fb =~ y1 + y2 + y3
    fb ~ w1 + w2
"
fit_ml <- asem(
  mod_ml,
  lavaan::Demo.twolevel,
  cluster = "cluster",
  verbose = FALSE,
  test = "none",
  marginal_correction = "none",
  vb_correction = FALSE,
  nsamp = 3
)

test_that("Multilevel: fit and summary", {
  fit_lav <- lavaan::sem(mod_ml, lavaan::Demo.twolevel, cluster = "cluster")

  expect_s4_class(fit_ml, "INLAvaan")
  expect_no_error(capture.output(summary(fit_ml)))
  expect_equal(coef(fit_ml), coef(fit_lav), tolerance = 0.1)
  expect_equal(fit_ml@optim$dx, rep(0, length(coef(fit_ml))), tolerance = 1e-2)
})

test_that("Multilevel predict lv", {
  nsamp <- 5

  # Level 1
  pred1 <- predict(fit_ml, type = "lv", level = 1L, nsamp = nsamp)
  expect_length(pred1, nsamp)
  m1 <- pred1[[1]]
  expect_equal(nrow(m1), 2500)
  expect_true("fw" %in% colnames(m1))
  expect_true(ncol(m1) >= 1)

  # Level 2
  pred2 <- predict(fit_ml, type = "lv", level = 2L, nsamp = nsamp)
  expect_length(pred2, nsamp)
  m2 <- pred2[[1]]
  expect_equal(nrow(m2), 200)
  expect_true("fb" %in% colnames(m2))
  expect_true(ncol(m2) >= 1)
})

test_that("Multilevel predict yhat and ypred", {
  nsamp <- 5

  # yhat
  pred_yhat <- predict(fit_ml, type = "yhat", nsamp = nsamp)
  expect_length(pred_yhat, nsamp)
  m <- pred_yhat[[1]]
  expect_equal(nrow(m), 2500)
  expect_equal(ncol(m), 8)
  expect_true(all(
    c("y1", "y2", "y3", "x1", "x2", "x3", "w1", "w2") %in% colnames(m)
  ))
  expect_false(any(is.na(m)))

  # ypred
  pred_ypred <- predict(fit_ml, type = "ypred", nsamp = nsamp)
  expect_length(pred_ypred, nsamp)
  m2 <- pred_ypred[[1]]
  expect_equal(nrow(m2), 2500)
  expect_equal(ncol(m2), 8)
  expect_false(any(is.na(m2)))
})

test_that("Multilevel predict errors for unsupported options", {
  expect_error(
    predict(fit_ml, type = "lv", newdata = lavaan::Demo.twolevel, nsamp = 3),
    "not supported for multilevel"
  )
})

## ----- Missing data (FIML) fit -----------------------------------------------
test_that("Multilevel predict ymis works", {
  dat <- lavaan::Demo.twolevel
  dat <- dat[dat$cluster <= 10, ]
  set.seed(123)
  dat$y1[sample(nrow(dat), 10)] <- NA

  fit_miss <- asem(
    mod_ml,
    dat,
    cluster = "cluster",
    missing = "ML",
    test = "none",
    marginal_correction = "none",
    vb_correction = FALSE,
    verbose = FALSE,
    nsamp = 3
  )

  nsamp <- 5

  # Full imputed dataset (default)
  pred <- predict(fit_miss, type = "ymis", nsamp = nsamp)
  expect_length(pred, nsamp)
  expect_false(any(is.na(pred[[1]])))
  expect_equal(ncol(pred[[1]]), 8L) # 8 model variables (y1-y3, x1-x3, w1-w2)

  # ymis_only: named vector of just the imputed cells
  pred_only <- predict(fit_miss, type = "ymis", nsamp = nsamp, ymis_only = TRUE)
  expect_length(pred_only, nsamp)
  v <- pred_only[[1]]
  expect_true(is.numeric(v))
  expect_length(v, 10L) # exactly the 10 NAs we injected
  expect_true(all(grepl("^y1\\[", names(v)))) # all from y1
})

test_that("two-level FIML warns about fully-missing-within cases", {
  # interim warning, only on lavaan versions with the gradient bug (fixed
  # upstream in lavaan PR #581); skip when the installed lavaan carries the
  # fix, where the warning correctly does not fire
  skip_if(
    utils::packageVersion("lavaan") >= "0.7.1.2707",
    "lavaan two-level FIML gradient bug is patched"
  )
  d <- lavaan::Demo.twolevel[, c("y1", "y2", "y3", "cluster")]
  d[1, c("y1", "y2", "y3")] <- NA
  m <- "
    level: 1
      fw =~ y1 + y2 + y3
    level: 2
      fb =~ y1 + y2 + y3
  "
  # capture all warnings (lavaan also emits its own "empty cases" note) and
  # assert ours is among them
  w <- testthat::capture_warnings(
    asem(
      m,
      d,
      cluster = "cluster",
      missing = "ml",
      verbose = FALSE,
      test = "none",
      nsamp = 3,
      vb_correction = FALSE,
      marginal_method = "marggaus",
      marginal_correction = "none"
    )
  )
  expect_true(any(grepl("fully missing on the within", w)))
})

testthat::skip()
# Simulate binary data
set.seed(141)
n <- 100
truval <- c(0.8, 0.7, 0.6, 0.5, 0.4, -1.43, -0.55, -0.13, -0.72, -1.13)
dat <- lavaan::simulateData(
  "eta =~ 0.8*y1 + 0.7*5y2 + 0.6*y3 + 0.5*y4 + 0.4*y5
   y1 | -1.43*t1
   y2 | -0.55*t1
   y3 | -0.13*t1
   y4 | -0.72*t1
   y5 | -1.13*t1",
  ordered = TRUE,
  sample.nobs = n
)
mod <- "eta  =~ y1 + y2 + y3 + y4 + y5"

test_that("Method: skewnorm", {
  expect_no_error({
    fit <- acfa(mod, dat, ordered = TRUE, verbose = FALSE, nsamp = 3)
  })
  expect_no_error(out <- capture.output(summary(fit)))
  expect_no_error(out <- plot(fit))

  expect_s4_class(fit, "INLAvaan")
  expect_equal(fit@optim$dx, rep(0, length(coef(fit))), tolerance = 1e-3)
})

dat <- lavaan::HolzingerSwineford1939
mod <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"

test_that("residuals() returns observed-minus-fitted moments for INLAvaan objects", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  rs <- residuals(fit)
  expect_type(rs, "list")
  expect_equal(rs$type, "raw")
  ov <- lavaan::lavNames(fit, "ov")
  expect_equal(dim(rs$cov), c(length(ov), length(ov)))
  expect_true(isSymmetric(unclass(rs$cov)))
})

test_that("residuals() matches observed minus fitted()", {
  fit <- acfa(mod, dat, meanstructure = TRUE, verbose = FALSE,
              nsamp = 3, test = "none")
  ov <- lavaan::lavNames(fit, "ov")
  obs_cov <- lavaan::lavInspect(fit, "sampstat")$cov
  implied <- fitted(fit)$cov
  expect_equal(unclass(residuals(fit)$cov), unclass(obs_cov) - unclass(implied),
               tolerance = 1e-8)
})

test_that("residuals() supports the lavaan 'type' options", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  # lavaan canonicalises the reported $type; "cor" -> "cor.bollen"
  expected <- c(raw = "raw", cor = "cor.bollen", cor.bollen = "cor.bollen",
                cor.bentler = "cor.bentler", normalized = "normalized",
                standardized = "standardized")
  for (ty in names(expected)) {
    rs <- residuals(fit, type = ty)
    expect_equal(rs$type, unname(expected[ty]))
  }
})

test_that("residuals(type = 'casewise') returns observed-minus-fitted per case", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  cw <- residuals(fit, type = "casewise")
  expect_true(is.matrix(cw))
  expect_equal(nrow(cw), nrow(dat))
  expect_equal(ncol(cw), length(lavaan::lavNames(fit, "ov")))
})

test_that("resid() matches residuals()", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  expect_equal(resid(fit), residuals(fit))
})

dat <- lavaan::HolzingerSwineford1939
mod <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"

test_that("fitted() returns model-implied moments for INLAvaan objects", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  fv <- fitted(fit)
  expect_type(fv, "list")
  expect_true("cov" %in% names(fv))
  ov <- lavaan::lavNames(fit, "ov")
  expect_equal(dim(fv$cov), c(length(ov), length(ov)))
  expect_true(isSymmetric(unclass(fv$cov)))
})

test_that("fitted() includes the mean vector with a mean structure", {
  fit <- acfa(mod, dat, meanstructure = TRUE, verbose = FALSE,
              nsamp = 3, test = "none")
  fv <- fitted(fit)
  expect_true("mean" %in% names(fv))
  expect_length(fv$mean, length(lavaan::lavNames(fit, "ov")))
})

test_that("fitted(type = 'ov') returns casewise predicted values", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  cw <- fitted(fit, type = "ov")
  expect_true(is.matrix(cw))
  expect_equal(nrow(cw), nrow(dat))
  expect_equal(ncol(cw), length(lavaan::lavNames(fit, "ov")))
})

test_that("fitted.values() matches fitted()", {
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  expect_equal(fitted.values(fit), fitted(fit))
})

dat <- lavaan::HolzingerSwineford1939
mod <- "
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
  "
NSAMP <- 3

test_that("Method: nlminb with numerical grad", {
  expect_no_error({
    fit <- asem(
      mod,
      dat,
      numerical_grad = TRUE,
      verbose = FALSE,
      nsamp = NSAMP
    )
  })

  expect_s4_class(fit, "INLAvaan")
  expect_equal(fit@optim$dx, rep(0, length(coef(fit))), tolerance = 1e-3)
})

test_that("Method: ucminf", {
  expect_no_error({
    fit <- asem(
      mod,
      dat,
      optim_method = "ucminf",
      verbose = FALSE,
      nsamp = NSAMP
    )
  })

  expect_s4_class(fit, "INLAvaan")
  expect_equal(fit@optim$dx, rep(0, length(coef(fit))), tolerance = 1e-3)
})

test_that("Method: optim", {
  expect_no_error({
    fit <- asem(
      mod,
      dat,
      optim_method = "optim",
      verbose = FALSE,
      nsamp = NSAMP
    )
  })

  expect_s4_class(fit, "INLAvaan")
  expect_equal(fit@optim$dx, rep(0, length(coef(fit))), tolerance = 1e-2)
})

mod <- "
  # intercept and slope with fixed coefficients
  i =~ 1*t1 + 1*t2 + 1*t3 + 1*t4
  s =~ 0*t1 + 1*t2 + 2*t3 + 3*t4

  # regressions
  i ~ x1 + x2
  s ~ x1 + x2

  # time-varying covariates
  t1 ~ c1
  t2 ~ c2
  t3 ~ c3
  t4 ~ c4
"
dat <- lavaan::Demo.growth
fit_lav <- lavaan::cfa(mod, dat)
NSAMP <- 3

test_that("Method: marggaus (fast)", {
  expect_no_error({
    fit <- agrowth(
      mod,
      dat,
      marginal_method = "marggaus",
      vb_correction = FALSE,
      test = "none",
      verbose = FALSE,
      nsamp = NSAMP
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
})

test_that("Method: skewnorm", {
  expect_no_error({
    fit <- agrowth(
      mod,
      dat,
      marginal_method = "skewnorm",
      vb_correction = FALSE,
      test = "none",
      verbose = FALSE,
      nsamp = NSAMP
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
  expect_equal(fit@optim$dx, rep(0, length(coef(fit))), tolerance = 1e-3)
})

test_that("Method: asymgaus", {
  expect_no_error({
    fit <- agrowth(
      mod,
      dat,
      marginal_method = "asymgaus",
      vb_correction = FALSE,
      test = "none",
      verbose = FALSE,
      nsamp = NSAMP
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
})

test_that("Method: sampling", {
  expect_no_error({
    fit <- agrowth(
      mod,
      dat,
      marginal_method = "sampling",
      vb_correction = FALSE,
      test = "none",
      verbose = FALSE,
      nsamp = NSAMP
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
})

test_that("Gradients are correct (Finite Difference Check)", {
  suppressMessages(
    tmp <- capture.output(fit <- agrowth(mod, dat, test = "none", debug = TRUE))
  )
  test_df <- read.table(text = tmp, skip = 1)[, -1]
  colnames(test_df) <- c("fd", "analytic", "diff")

  expect_equal(
    as.numeric(test_df$fd),
    as.numeric(test_df$diff),
    tolerance = 1e-3
  )
  expect_equal(
    as.numeric(test_df$diff),
    rep(0, nrow(test_df)),
    tolerance = 1e-3
  )
})

dat     <- lavaan::HolzingerSwineford1939
sem_dat <- lavaan::PoliticalDemocracy
NSAMP   <- 3

mod <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"
sem_mod <- "
  ind60 =~ x1 + x2 + x3
  dem60 =~ y1 + y2 + y3 + y4
  dem60 ~ ind60
"

# Fit once; reused across all tests below (fast defaults)
fit_cfa <- acfa(mod, dat, verbose = FALSE, nsamp = NSAMP,
                vb_correction = FALSE, test = "none",
                marginal_method = "marggaus")
fit_mg  <- acfa(mod, dat, verbose = FALSE, nsamp = NSAMP, group = "school",
                vb_correction = FALSE, test = "none",
                marginal_method = "marggaus")
fit_sem <- asem(sem_mod, sem_dat, verbose = FALSE, nsamp = NSAMP,
                vb_correction = FALSE, test = "none",
                marginal_method = "marggaus")

# ---- CFA: type = "lv" (default) -----------------------------------------

test_that("type = 'lv' returns correctly-shaped latent predictions", {
  prd  <- predict(fit_cfa, nsamp = NSAMP)
  summ <- summary(prd)
  expect_no_error(capture.output(print(prd)))
  expect_no_error(capture.output(print(summ)))
  expect_equal(length(prd),     NSAMP)
  expect_equal(nrow(summ$Mean), nrow(dat))
})

# ---- CFA: type = "yhat" / "ov" and "ypred" / "ydist" -------------------

test_that("type = 'yhat' and alias 'ov' return n x p fitted-mean matrices", {
  prd_yhat <- predict(fit_cfa, type = "yhat", nsamp = NSAMP)
  prd_ov   <- predict(fit_cfa, type = "ov",   nsamp = NSAMP)
  expect_equal(length(prd_yhat),    NSAMP)
  expect_equal(nrow(prd_yhat[[1]]), nrow(dat))
  expect_equal(ncol(prd_yhat[[1]]), 6L)
  expect_equal(dim(prd_ov[[1]]),    dim(prd_yhat[[1]]))
})

test_that("type = 'ypred' and alias 'ydist' return n x p predicted matrices", {
  prd_ypred <- predict(fit_cfa, type = "ypred", nsamp = NSAMP)
  prd_ydist <- predict(fit_cfa, type = "ydist", nsamp = NSAMP)
  expect_equal(length(prd_ypred),    NSAMP)
  expect_equal(nrow(prd_ypred[[1]]), nrow(dat))
  expect_equal(ncol(prd_ypred[[1]]), 6L)
  expect_equal(dim(prd_ydist[[1]]),  dim(prd_ypred[[1]]))
})

# ---- SEM (B-matrix path): lv, yhat, ypred --------------------------------

test_that("SEM predict covers lv, yhat, and ypred types", {
  prd_lv    <- predict(fit_sem, nsamp = NSAMP)
  prd_yhat  <- predict(fit_sem, type = "yhat",  nsamp = NSAMP)
  prd_ypred <- predict(fit_sem, type = "ypred", nsamp = NSAMP)
  n <- nrow(sem_dat)
  expect_equal(nrow(summary(prd_lv)$Mean), n)
  expect_equal(nrow(prd_yhat[[1]]),         n)
  expect_equal(ncol(prd_yhat[[1]]),         7L)
  expect_equal(nrow(prd_ypred[[1]]),        n)
})

# ---- newdata -------------------------------------------------------------

test_that("newdata works for lv, yhat, and ypred types", {
  newdat <- dat[1:5, ]
  expect_equal(nrow(predict(fit_cfa, type = "lv",    newdata = newdat, nsamp = NSAMP)[[1]]), 5L)
  expect_equal(nrow(predict(fit_cfa, type = "yhat",  newdata = newdat, nsamp = NSAMP)[[1]]), 5L)
  expect_equal(nrow(predict(fit_cfa, type = "ypred", newdata = newdat, nsamp = NSAMP)[[1]]), 5L)
})

test_that("type = 'ymis' with newdata throws an error", {
  expect_error(
    predict(fit_cfa, type = "ymis", newdata = dat[1:5, ], nsamp = NSAMP),
    "ymis.*newdata"
  )
})

# ---- Multigroup: lv, yhat, newdata ---------------------------------------

test_that("multigroup predict works for lv, yhat, and newdata", {
  prd      <- predict(fit_mg, nsamp = NSAMP)
  prd_yhat <- predict(fit_mg, type = "yhat", nsamp = NSAMP)
  newdat   <- rbind(
    dat[dat$school == "Pasteur",     ][1:3, ],
    dat[dat$school == "Grant-White", ][1:2, ]
  )
  prd_new  <- predict(fit_mg, type = "lv", newdata = newdat, nsamp = NSAMP)
  expect_equal(nrow(summary(prd)$Mean),      nrow(dat))
  expect_equal(nrow(summary(prd_yhat)$Mean), nrow(dat))
  expect_equal(nrow(summary(prd_new)$Mean),  5L)
})

# ---- summary = TRUE shortcut ---------------------------------------------

test_that("predict(summary = TRUE) matches summary(predict(...))", {
  set.seed(1)
  a <- predict(fit_cfa, type = "yhat", nsamp = NSAMP, summary = TRUE)
  set.seed(1)
  b <- summary(predict(fit_cfa, type = "yhat", nsamp = NSAMP))
  expect_s3_class(a, "summary.predict.inlavaan_internal")
  expect_equal(a, b)
})

test_that("predict() default (summary = FALSE) still returns raw draws", {
  prd <- predict(fit_cfa, type = "yhat", nsamp = NSAMP)
  expect_s3_class(prd, "predict.inlavaan_internal")
  expect_equal(length(prd), NSAMP)
})

test_that("Factor scores are centred on the implied/saturated means", {
  # Regression test: predict() used to condition on raw y instead of
  # y - mu_y, offsetting every factor score by Phi Lambda' Sigma^{-1} mu_y
  # (several sd on uncentred data), under both meanstructure settings.
  fit50 <- acfa(mod, dat, verbose = FALSE, nsamp = 50,
                vb_correction = FALSE, test = "none")
  draws <- unclass(predict(fit50, type = "lv"))
  fs <- Reduce(`+`, draws) / length(draws)
  fs_lav <- lavaan::lavPredict(lavaan::cfa(mod, dat))
  for (k in seq_len(ncol(fs_lav))) {
    expect_lt(max(abs(fs[, k] - fs_lav[, k])) / sd(fs_lav[, k]), 0.6)
  }
})

mod <- "
  visual  =~ x1 + a*x2 + b*x3
  textual =~ x4 + x5 + x6
  a == b
"
dat <- lavaan::HolzingerSwineford1939
fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")

test_that("summary(standardized = TRUE) works with equality constraints", {
  expect_no_error(
    out <- capture.output(summary(fit, standardized = TRUE, nsamp = 5))
  )
  expect_true(any(grepl("Std.lv", out)))
  expect_true(any(grepl("Std.all", out)))
})

test_that("summary(estimates = FALSE) prints the header only", {
  expect_no_error(out <- capture.output(summary(fit, estimates = FALSE)))
  expect_true(any(grepl("Number of model parameters", out)))
  expect_false(any(grepl("Parameter Estimates", out)))
})

test_that("summary(rsquare = TRUE) prints the R-square section", {
  expect_no_warning(out <- capture.output(summary(fit, rsquare = TRUE)))
  expect_true(any(grepl("R-Square", out)))
})

test_that("summary accepts the British spelling standardised = TRUE", {
  expect_no_error(
    out <- capture.output(summary(fit, standardised = TRUE, nsamp = 5))
  )
  expect_true(any(grepl("Std.lv", out)))
  expect_true(any(grepl("Std.all", out)))
})

test_that("summary(ci = FALSE) omits the credible interval columns", {
  out <- capture.output(summary(fit, ci = FALSE))
  expect_false(any(grepl("2.5%", out, fixed = TRUE)))
  out_ci <- capture.output(summary(fit))
  expect_true(any(grepl("2.5%", out_ci, fixed = TRUE)))
  expect_true(any(grepl("97.5%", out_ci, fixed = TRUE)))
})

test_that("summary(rsquare = TRUE, standardized = TRUE) rows stay aligned", {
  expect_no_error(
    out <- capture.output(
      summary(fit, rsquare = TRUE, standardized = TRUE, nsamp = 5)
    )
  )
  expect_true(any(grepl("R-Square", out)))
  expect_true(any(grepl("Std.all", out)))
})

dat <- lavaan::HolzingerSwineford1939
mod <- "
    visual  =~ x1 + x2 # + x3
    textual =~ x4 + x5 # + x6
    # speed   =~ x7 + x8 + x9
  "
NSAMP <- 3
STDLV <- TRUE

test_that("Multigroup fitting and testing", {
  # Configural invariance
  expect_no_error({
    fit1 <- acfa(
      mod,
      dat,
      verbose = FALSE,
      test = "none",
      std.lv = STDLV,
      group = "school"
    )
  })
  expect_no_error(out <- capture.output(summary(fit1)))
  expect_equal(fit1@optim$dx, rep(0, length(fit1@optim$dx)), tolerance = 1e-3)

  # Weak invariance
  expect_no_error({
    fit2 <- acfa(
      mod,
      dat,
      verbose = FALSE,
      test = "none",
      std.lv = STDLV,
      group = "school",
      group.equal = "loadings"
    )
  })
  expect_no_error(out <- capture.output(summary(fit2)))
  expect_equal(fit2@optim$dx, rep(0, length(fit2@optim$dx)), tolerance = 1e-3)

  # Weak invariance
  expect_no_error({
    fit3 <- acfa(
      mod,
      dat,
      verbose = FALSE,
      test = "none",
      std.lv = STDLV,
      group = "school",
      group.equal = c("intercepts", "loadings")
    )
  })
  expect_no_error(out <- capture.output(summary(fit3)))
  expect_equal(fit3@optim$dx, rep(0, length(fit3@optim$dx)), tolerance = 1e-3)

  # Comparison
  expect_no_error({
    cp <- compare(fit1, fit2, fit3)
    out <- capture.output(print(cp))
  })
})

test_that("Gradients are correct (Finite Difference Check)", {
  suppressWarnings(suppressMessages(
    tmp <- capture.output(fit <- acfa(mod, dat, test = "none", debug = TRUE))
  ))
  test_df <- read.table(text = tmp, skip = 1)[, -1]
  colnames(test_df) <- c("fd", "analytic", "diff")

  expect_equal(
    as.numeric(test_df$fd),
    as.numeric(test_df$diff),
    tolerance = 1e-3
  )
  expect_equal(
    as.numeric(test_df$diff),
    rep(0, nrow(test_df)),
    tolerance = 1e-3
  )
})

test_that("Plot method works", {
  dat <- lavaan::HolzingerSwineford1939
  mod <- "
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
  "
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  pdf(file = NULL)
  on.exit(dev.off(), add = TRUE)
  expect_silent(plt <- plot(fit))
  expect_silent(plt <- plot(fit, truth = coef(fit)))
  # expect_equal(length(plt@layers), length(coef(fit)))
})

test_that("Plot with use_ggplot = TRUE returns ggplot object", {
  dat <- lavaan::HolzingerSwineford1939
  mod <- "
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
  "
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  pdf(file = NULL)
  on.exit(dev.off(), add = TRUE)

  plt <- plot(fit, use_ggplot = TRUE)
  expect_s3_class(plt, "ggplot")

  plt_truth <- plot(fit, truth = coef(fit), use_ggplot = TRUE)
  expect_s3_class(plt_truth, "ggplot")
})

test_that("Plot with use_ggplot = FALSE uses base graphics", {
  dat <- lavaan::HolzingerSwineford1939
  mod <- "
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
  "
  fit <- acfa(mod, dat, verbose = FALSE, nsamp = 3, test = "none")
  pdf(file = NULL)
  on.exit(dev.off(), add = TRUE)

  res <- plot(fit, use_ggplot = FALSE)
  expect_null(res)

  res_truth <- plot(fit, truth = coef(fit), use_ggplot = FALSE)
  expect_null(res_truth)
})

mod <- "
  # intercept with coefficients fixed to 1
  i =~  1*Day0 + 1*Day1 + 1*Day2 + 1*Day3 + 1*Day4 +
        1*Day5 + 1*Day6 + 1*Day7 + 1*Day8 + 1*Day9

  # slope with coefficients fixed to 0:9 (number of days)
  s =~  0*Day0 + 1*Day1 + 2*Day2 + 3*Day3 + 4*Day4 +
        5*Day5 + 6*Day6 + 7*Day7 + 8*Day8 + 9*Day9

  i ~~ i
  i ~ 1

  s ~~ s
  s ~ 1

  i ~~ s

  # fix intercepts
  Day0 ~ 0*1
  Day1 ~ 0*1
  Day2 ~ 0*1
  Day3 ~ 0*1
  Day4 ~ 0*1
  Day5 ~ 0*1
  Day6 ~ 0*1
  Day7 ~ 0*1
  Day8 ~ 0*1
  Day9 ~ 0*1

  # apply equality constraints
  Day0 ~~ v*Day0
  Day1 ~~ v*Day1
  Day2 ~~ v*Day2
  Day3 ~~ v*Day3
  Day4 ~~ v*Day4
  Day5 ~~ v*Day5
  Day6 ~~ v*Day6
  Day7 ~~ v*Day7
  Day8 ~~ v*Day8
  Day9 ~~ v*Day9
  "
dat <- reshape(
  lme4::sleepstudy,
  timevar = "Days",
  idvar = "Subject",
  direction = "wide"
)
names(dat) <- sub("^Reaction\\.(.*)$", "Day\\1", names(dat))
fit_lav <- lavaan::growth(mod, dat)
NSAMP <- 3

test_that("Method: skewnorm", {
  expect_no_error({
    fit <- agrowth(
      mod,
      dat,
      marginal_method = "skewnorm",
      verbose = FALSE,
      nsamp = NSAMP
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
  gr_at_opt <- fit@optim$dx
  gt_at_opt <- as.numeric(fit@Model@ceq.simple.K %*% gr_at_opt)
  expect_equal(gt_at_opt, rep(0, length(coef(fit))), tolerance = 1e-3)
})

test_that("Method: asymgaus", {
  expect_no_error({
    fit <- agrowth(
      mod,
      dat,
      marginal_method = "asymgaus",
      verbose = FALSE,
      nsamp = NSAMP
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
})

test_that("Method: marggaus", {
  expect_no_error({
    fit <- agrowth(
      mod,
      dat,
      marginal_method = "marggaus",
      verbose = FALSE,
      nsamp = NSAMP
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
})

test_that("Method: sampling", {
  expect_no_error({
    fit <- agrowth(
      mod,
      dat,
      marginal_method = "sampling",
      verbose = FALSE,
      nsamp = NSAMP
    )
  })
  expect_no_error(out <- capture.output(summary(fit)))

  expect_s4_class(fit, "INLAvaan")
})

test_that("Gradients are correct (Finite Difference Check)", {
  suppressMessages(
    tmp <- capture.output(fit <- agrowth(mod, dat, test = "none", debug = TRUE))
  )
  test_df <- read.table(text = tmp, skip = 1)[, -1]
  colnames(test_df) <- c("fd", "analytic", "diff")

  expect_equal(
    as.numeric(test_df$fd),
    as.numeric(test_df$diff),
    tolerance = 1e-3
  )
  expect_equal(
    as.numeric(test_df$diff),
    rep(0, nrow(test_df)),
    tolerance = 1e-3
  )
})

#' Fit a skew normal distribution to log-density evaluations
#'
#' @details This skew normal fitting function uses a weighted least squares
#'   approach to fit the log-density evaluations provided in `y` at points `x`.
#'   The weights are set to be the density evaluations raised to the power of
#'   the temperature parameter `k`. This has somewhat an interpretation of
#'   finding the skew normal fit that minimises the Kullback-Leibler divergence
#'   from the true density to it.
#'
#'   In R-INLA, the C code implementation from which this was translated from
#'   can be found
#'   [here](https://github.com/hrue/r-inla/blob/b63eb379f69d6553f965b360f1b88877cfef20d1/gmrflib/fit-sn.c).
#'
#'
#' @param x A numeric vector of points where the density is evaluated.
#' @param y A numeric vector of log-density evaluations at points x.
#' @param threshold_log_drop A negative numeric value indicating the log-density
#'   drop threshold below which points are ignored in the fitting. Default is
#'   -6.
#' @param temp A numeric value for the temperature parameter k. If NA (default),
#'   it is included in the optimisation.
#'
#' @returns A list with fitted parameters:
#'   - `xi`: location parameter
#'   - `omega`: scale parameter
#'   - `alpha`: shape parameter
#'   - `logC`: log-normalization constant
#'   - `k`: temperature parameter
#'   - `rsq`: R-squared of the fit
#'
#'  Note that `logC` and `k` are not used when fitting from a sample.
#'
#' @example inst/examples/ex-skewnorm-fit.R
#' @export
fit_skew_normal <- function(x, y, threshold_log_drop = -6, temp = NA) {
  # NOTE: y is the density evaluations at x on the log scale, i.e. log f(x).
  # y should ideally be normalized so max(y) = 0 for numerical stability
  if (max(y) > 0) { # nocov start
    cli_warn(
      "In {.fn fit_skew_normal}, log density {.arg y} should be normalized so that max(y) = 0 for numerical stability. Automatically normalizing now."
    )
    y <- y - max(y)
  } # nocov end
  if (threshold_log_drop >= 0) { # nocov start
    cli_abort(
      "In {.fn fit_skew_normal}, {.arg threshold_log_drop} must be negative."
    )
  } # nocov end
  is_est_k <- is.na(temp) | is.null(temp)

  # Integration weights (used for calculating moments to get inits)
  if (FALSE) {
    w_int <- rep(1, length(x))
  } else {
    # Trapezoidal rule weights
    w_int <- numeric(length(x))
    w_int[1] <- 0.5 * (x[2] - x[1])
    w_int[-1] <- 0.5 * (c(diff(x[-1]), 0) + c(0, diff(x[-length(x)])))
    # w_int     <- w_int / sum(w_int)
  }

  # # Fitting (importance) weights for objective
  # w_fit <- exp(10 * y)
  # w_fit[y < threshold_log_drop] <- 0  # the "clean" KLD approach
  # w_fit <- w_fit / sum(w_fit)  # normalise for stability

  # Initial moment-based estimates
  p <- exp(y)
  m0 <- sum(w_int * p)
  m1 <- sum(w_int * p * x)
  m2 <- sum(w_int * p * x^2)

  # Protect against degenerate m0 (e.g. if all y are very small)
  if (m0 < .Machine$double.eps) { # nocov
    m0 <- 1
  }

  mean_hat <- m1 / m0
  var_hat <- m2 / m0 - mean_hat^2
  sd_hat <- sqrt(max(.Machine$double.eps, var_hat))
  m3 <- sum(w_int * p * (x - mean_hat)^3)

  skew_hat <- (m3 / m0) / (sd_hat^3)
  skew_hat <- max(min(skew_hat, 0.9), -0.9)
  # Note: Theoretical limit of skewness is 0.9952717

  f_skew_to_delta <- function(delta, target_skew) {
    # gamma1(delta) from Wikipedia:
    # γ1 = ((4−π)/2) * ((δ*√(2/π))^3) / (1 − 2 δ^2/π)^(3/2)
    # δ = α / √(1 + α^2)
    num <- (4 - pi) / 2 * (delta * sqrt(2 / pi))^3
    den <- (1 - 2 * delta^2 / pi)^(3 / 2)
    num / den - target_skew
  }

  # Solve for delta in (−1,1)
  root <- tryCatch(
    {
      stats::uniroot(
        f_skew_to_delta,
        interval = c(-0.995, 0.995),
        target_skew = skew_hat
      )$root
    },
    error = function(e) 0
  ) # fallback to normal if uniroot fails
  alpha_init <- root / sqrt(1 - root^2)
  omega_init <- sd_hat / sqrt(1 - 2 * root^2 / pi)
  xi_init <- mean_hat - omega_init * root * sqrt(2 / pi)

  # Optimisation functions
  ob <- function(param, x, y) {
    mu <- param[1]
    lsinv <- param[2]
    a <- param[3]
    logC <- param[4]
    logk <- if (is_est_k) param[5] else log(temp)
    # print(exp(logk))

    w <- exp(exp(logk) * y)
    w[y < threshold_log_drop] <- 0
    w <- w / sum(w)

    # log skew-normal density up to constant
    xx <- (x - mu) * exp(lsinv)
    logdens <- log(2) +
      stats::dnorm(xx, log = TRUE) +
      stats::pnorm(a * xx, log.p = TRUE) +
      lsinv +
      logC
    resid <- y - logdens
    sum(w * resid^2) # Weighted Sum of Squares
  }

  gr <- function(param, x, y) {
    mu <- param[1]
    lsinv <- param[2]
    a <- param[3]
    logC <- param[4]
    s <- exp(lsinv)
    xx <- s * (x - mu)
    u <- a * xx
    R <- mills_ratio(u)

    logk <- if (is_est_k) param[5] else log(temp)
    w <- exp(exp(logk) * y)
    w[y < threshold_log_drop] <- 0
    w <- w / sum(w)

    # First derivatives of log-density L wrt parameters
    L_mu <- s * xx - a * s * R
    L_lsinv <- -xx^2 + a * xx * R + 1
    L_a <- xx * R
    L_logC <- 1

    L <- log(2) +
      stats::dnorm(xx, log = TRUE) +
      stats::pnorm(u, log.p = TRUE) +
      lsinv +
      logC
    r <- y - L

    # Weighted Gradients
    g1 <- -2 * sum(w * r * L_mu)
    g2 <- -2 * sum(w * r * L_lsinv)
    g3 <- -2 * sum(w * r * L_a)
    g4 <- -2 * sum(w * r * L_logC)
    g5 <- sum(y * w * r^2) * exp(logk)

    if (is_est_k) {
      return(c(g1, g2, g3, g4, g5))
    } else {
      return(c(g1, g2, g3, g4))
    }
  }

  hs <- function(param, x, y) {
    w <- exp(temp * y)
    w[y < threshold_log_drop] <- 0
    w <- w / sum(w)

    mu <- param[1]
    lsinv <- param[2]
    a <- param[3]
    logC <- param[4]
    s <- exp(lsinv)
    xx <- s * (x - mu)
    u <- a * xx
    R <- mills_ratio(u)
    Rprime <- -R * (u + R) # derivative of Mills ratio

    # First derivatives
    L_mu <- s * xx - a * s * R
    L_lsinv <- -xx^2 + a * xx * R + 1
    L_a <- xx * R
    L_logC <- 1

    # Second derivatives
    L_mumu <- -s^2 + a^2 * s^2 * Rprime
    L_mu_lsinv <- 2 * s * xx - a * s * R - a^2 * s * xx * Rprime
    L_mu_a <- -s * R - a * s * xx * Rprime
    L_mu_logC <- 0

    L_lsinv_lsinv <- -2 * xx^2 + a * xx * R + a^2 * xx^2 * Rprime
    L_lsinv_a <- xx * R + a * xx^2 * Rprime
    L_lsinv_logC <- 0

    L_aa <- xx^2 * Rprime
    L_a_logC <- 0

    L <- log(2) +
      stats::dnorm(xx, log = TRUE) +
      stats::pnorm(u, log.p = TRUE) +
      lsinv +
      logC
    r <- y - L

    # Weighted Hessian: H = 2 * sum w * (gradL gradL^T - r * HessL)
    H11 <- 2 * sum(w * (L_mu * L_mu - r * L_mumu))
    H12 <- 2 * sum(w * (L_mu * L_lsinv - r * L_mu_lsinv))
    H13 <- 2 * sum(w * (L_mu * L_a - r * L_mu_a))
    H14 <- 2 * sum(w * (L_mu * L_logC - r * L_mu_logC))

    H22 <- 2 * sum(w * (L_lsinv * L_lsinv - r * L_lsinv_lsinv))
    H23 <- 2 * sum(w * (L_lsinv * L_a - r * L_lsinv_a))
    H24 <- 2 * sum(w * (L_lsinv * L_logC - r * L_lsinv_logC))

    H33 <- 2 * sum(w * (L_a * L_a - r * L_aa))
    H34 <- 2 * sum(w * (L_a * L_logC - r * L_a_logC))

    H44 <- 2 * sum(w * (L_logC * L_logC))

    # fmt: skip
    matrix(c(H11, H12, H13, H14,
             H12, H22, H23, H24,
             H13, H23, H33, H34,
             H14, H24, H34, H44), nrow = 4, byrow = TRUE)
  }
  if (is_est_k) {
    hs <- NULL
  }

  # Optimise skew normal parameters
  st <- if (is_est_k) {
    c(xi_init, log(1 / omega_init), alpha_init, 0, log(10))
  } else {
    c(xi_init, log(1 / omega_init), alpha_init, 0)
  }
  fit <- stats::nlminb(
    st,
    ob,
    gr,
    hs,
    x = x,
    y = y
    # w = w_fit
  )
  xi_hat <- fit$par[1]
  omega_hat <- exp(-fit$par[2])
  alpha_hat <- fit$par[3]
  logC_hat <- fit$par[4]
  k_hat <- if (is_est_k) exp(fit$par[5]) else temp

  # Calculate RMSE and R2 (unweighted)
  expy <- exp(y)
  fity <- dsnorm(x, xi_hat, omega_hat, alpha_hat, logC_hat)
  rss <- sum((expy - fity)^2)
  tss <- sum(expy^2) # total sum of squares relative to 0
  rmse <- rss / length(expy)
  rsq <- 1 - (rss / tss)

  # Normalised max abs deviation (NMAD)
  nmad <- max(abs(expy - fity)) / max(expy)

  list(
    xi = xi_hat,
    omega = omega_hat,
    alpha = alpha_hat,
    logC = logC_hat,
    k = k_hat,
    rmse = rmse,
    nmad = nmad
  )
}

mills_ratio <- function(u) {
  logphi <- stats::dnorm(u, log = TRUE)
  logPhi <- stats::pnorm(u, log.p = TRUE)
  # guard against -Inf in logPhi; cap to avoid exp overflow
  logPhi <- pmax(logPhi, -745) # ~exp(-745) ~ 5e-324
  exp(logphi - logPhi)
}


#' The Skew Normal Distribution
#'
#' Density for the skew normal distribution with location `xi`, scale `omega`,
#' and shape `alpha`.
#'
#' @param x Vector of quantiles.
#' @param xi Location parameter.
#' @param omega Scale parameter.
#' @param alpha Shape parameter.
#' @param logC Log-normalization constant.
#' @param log Logical; if TRUE, returns the log density.
#'
#' @returns A numeric vector of (log) density values.
#' @references https://en.wikipedia.org/wiki/Skew_normal_distribution
#'
#' @export
#' @examples
#' x <- seq(-2, 5, length.out = 100)
#' y <- dsnorm(x, xi = 0, omega = 1, alpha = 5)
#' plot(x, y, type = "l", main = "Skew Normal Density")
dsnorm <- function(x, xi, omega, alpha, logC = 0, log = FALSE) {
  xx <- (x - xi) / omega
  if (isTRUE(log)) {
    log(2) +
      stats::dnorm(xx, log = TRUE) +
      stats::pnorm(alpha * xx, log.p = TRUE) -
      log(omega) +
      logC
  } else {
    2 / omega * stats::dnorm(xx) * stats::pnorm(alpha * xx) * exp(logC)
  }
}

# snorm_EX_VarX <- function(xi, omega, alpha) {
#   delta <- alpha / sqrt(1 + alpha^2)
#   EX <- xi + omega * delta * sqrt(2 / pi)
#   VarX <- omega^2 * (1 - (2 * delta^2) / pi)
#   list(EX = EX, VarX = VarX)
# }

#' Fit a skew normal distribution to a sample
#'
#' @details Uses maximum likelihood estimation to fit a skew normal distribution
#' to the provided numeric vector `x`.
#'
#' @param x A numeric vector of sample data.
#'
#' @inherit fit_skew_normal return
#' @export
#'
#' @examples
#' x <- rnorm(100, mean = 5, sd = 1)
#' unlist(fit_skew_normal_samp(x))
fit_skew_normal_samp <- function(x) {
  # Starting values
  xi0 <- stats::median(x)
  omega0 <- log(stats::sd(x)) # log-scale param
  alpha0 <- 0
  start <- c(xi0, omega0, alpha0)

  opt <- nlminb(
    start = c(xi0, omega0, alpha0),
    objective = function(par) {
      -1 *
        sum(dsnorm(
          x,
          xi = par[1],
          omega = exp(par[2]),
          alpha = par[3],
          log = TRUE
        ))
    }
  )

  xi_hat <- opt$par[1]
  omega_hat <- exp(opt$par[2])
  alpha_hat <- opt$par[3]

  list(
    xi = xi_hat,
    omega = omega_hat,
    alpha = alpha_hat
  )
}

test_that("SN fit fixed temp", {
  x_grid <- seq(-2, 5, length.out = 21)
  y_log <- sapply(x_grid, function(x) {
    dsnorm(x, xi = 1, omega = 1, alpha = 5, log = TRUE)
  })
  y_log <- y_log - max(y_log) # normalise to have maximum at zero

  res <- fit_skew_normal(x_grid, y_log, temp = 1)
  expect_equal(res$xi, 1)
  expect_equal(res$omega, 1)
  expect_equal(res$alpha, 5)
})

test_that("SN fit variable temp", {
  x_grid <- seq(-2, 5, length.out = 31)
  y_log <- sapply(x_grid, function(x) {
    dsnorm(x, xi = 1, omega = 1, alpha = 5, log = TRUE)
  })
  y_log <- y_log - max(y_log) # normalise to have maximum at zero

  res <- fit_skew_normal(x_grid, y_log, temp = NA)
  expect_equal(res$xi, 1)
  expect_equal(res$omega, 1)
  expect_equal(res$alpha, 5)
})

create_lav_from_inlavaan_internal <- function(fit0, fit_inlv) {
  ## ----- Update Model and implied slots --------------------------------------
  x <- fit_inlv$coefficients
  fit0@Model <- lavaan::lav_model_set_parameters(fit0@Model, x)
  # fit0@Model@estimator <- "BAYES"
  fit0@implied <- lavaan::lav_model_implied(fit0@Model)

  ## ----- Update ParTable slot ------------------------------------------------

  # # Find Theta matrix (residuals) and Sigmay (implied covariance matrix)
  # thetadiag <- diag(fit0@Model@GLIST$theta)
  # Sigmay <- fit0@implied$cov[[1]]  # FIXME: Group 1 only
  SD <- fit_inlv$summary[, "SD"] # free only

  pt <- fit_inlv$partable
  pt$est[pt$free > 0] <- x[pt$free[pt$free > 0]]
  pt$se[pt$free > 0] <- SD[pt$free[pt$free > 0]]

  pt$par <- pt$parstart
  pt$par[pt$free > 0] <- fit_inlv$theta_star[pt$free[pt$free > 0]]

  # Update defined parameters
  if (any(pt$op == ":=")) {
    pt_def_rows <- which(pt$op == ":=")
    def_names <- pt$names[pt_def_rows]

    pt$est[pt_def_rows] <- fit_inlv$summary[def_names, "Mean"]
    pt$se[pt_def_rows] <- fit_inlv$summary[def_names, "SD"]
  }

  # Put the delta scales back (theta parameterisation)
  if (any(pt$op == "~*~")) {
    delta_rows <- which(pt$op == "~*~")
    delta_names <- pt$names[delta_rows]

    pt$est[delta_rows] <- fit_inlv$summary[delta_names, "Mean"]
    pt$se[delta_rows] <- fit_inlv$summary[delta_names, "SD"]
  }

  # Put the diag theta values in the pt (ONLY IF ORDINAL)
  # ov_names <- fit0@Data@ov.names[[1]]  # FIXME: Group 1 only
  # pt$est[pt$lhs %in% ov_names &
  #          pt$rhs %in% ov_names &
  #          pt$lhs == pt$rhs &
  #          pt$op == "~~"] <- thetadiag

  # Manually change the slack column
  # slack_values <- as.vector(fit0@Model@con.jac %*% x - fit0@Model@ceq.rhs)
  # pt$est[pt$op == "=="] <- slack_values

  # Flatten functions list
  pt$g <- sapply(pt$g, as_fun_string)
  pt$g_prime <- sapply(pt$g_prime, as_fun_string)
  pt$ginv <- sapply(pt$ginv, as_fun_string)
  pt$ginv_prime <- sapply(pt$ginv_prime, as_fun_string)
  pt$ginv_prime2 <- sapply(pt$ginv_prime2, as_fun_string)

  fit0@ParTable <- pt

  ## ----- Update Options slot -------------------------------------------------
  optim_method <- fit_inlv$optim_method
  if (optim_method == "optim") {
    optim_method <- "BFGS"
  }
  fit0@Options$optim.method <- fit_inlv$optim_method
  fit0@Options$do.fit <- TRUE
  fit0@Options$estimator <- "Bayes"
  fit0@Options$se <- "standard"

  ## ----- Update Fit slot -----------------------------------------------------
  fit0@Fit@x <- x
  # fit0@Fit@TH[[1]] <- tau  # FIXME: Group 1
  fit0@Fit@est <- pt$est
  fit0@Fit@se <- pt$se
  fit0@Fit@start <- pt$start

  if (fit_inlv$optim_method == "nlminb") {
    fit0@Fit@iterations <- as.integer(fit_inlv$opt$iterations)
    fit0@Fit@converged <- fit_inlv$opt$convergence == 0L
    fit0@Fit@fx <- fit_inlv$opt$objective
  } else if (fit_inlv$optim_method == "optim") {
    fit0@Fit@iterations <- as.integer(fit_inlv$opt$counts["function"])
    fit0@Fit@converged <- fit_inlv$opt$convergence == 0L
    fit0@Fit@fx <- fit_inlv$opt$value
  } else if (fit_inlv$optim_method == "ucminf") {
    fit0@Fit@iterations <- as.integer(fit_inlv$opt$info["neval"])
    fit0@Fit@converged <- fit_inlv$opt$convergence == 1L
    fit0@Fit@fx <- fit_inlv$opt$value
  }
  fit0@Fit@Sigma.hat <- fit0@implied

  fit0@Fit@test <- list(
    mloglik = list(
      test = "mloglik",
      stat = fit_inlv$mloglik,
      stat.group = fit_inlv$mloglik, # FIXME
      df = NA,
      refdistr = NA,
      pvalue = NA
    )
  )
  if (!is.null(fit_inlv$ppp)) {
    fit0@Fit@test <- c(
      fit0@Fit@test,
      list(
        ppp = list(
          test = "ppp",
          stat = fit_inlv$ppp,
          stat.group = fit_inlv$ppp, # FIXME
          df = NA,
          refdistr = NA,
          pvalue = NA
        )
      )
    )
  }

  ## ----- Update optim slot ---------------------------------------------------
  fit0@optim$x <- fit_inlv$opt$par
  fit0@optim$dx <- fit_inlv$opt$dx
  fit0@optim$npar <- length(x)

  fit0@optim$fx <- fit0@Fit@fx
  fit0@optim$fx.group <- fit0@Fit@fx.group # FIXME: What's this?
  fit0@optim$iterations <- fit0@Fit@iterations
  fit0@optim$converged <- fit0@Fit@converged

  ## ----- Update loglik slot --------------------------------------------------
  # if (fit_inlv$method == "ucminf") {
  #   fit0@loglik$loglik <- fit_inlv$numFit$value
  # } else if (fit_inlv$method == "SA") {
  #   fit0@loglik$loglik <- fit_inlv$stoFit@nll
  # }
  # fit0@loglik$estimator <- "ML"  # FIXME: to turn off warning for now
  # if (!is.null(vars)) {
  #   fit0@loglik$AIC <- get_AIC(NLL = fit0@loglik$loglik, INVH = vars$invH,
  #                              J = vars$J)
  #   fit0@loglik$BIC <- get_BIC(NLL = fit0@loglik$loglik, INVH = vars$invH,
  #                              J = vars$J, N = n)
  # }
  fit0@loglik <- list()

  ## ----- Update vcov slot ----------------------------------------------------
  fit0@vcov <- list(
    vcov = fit_inlv$vcov_x,          # Sample-based vcov (lavaan parameterisation)
    vcov_theta = fit_inlv$Sigma_theta # Laplace vcov (theta parameterisation)
  )

  ## ----- Update test slot ----------------------------------------------------
  fit0@test <- fit0@Fit@test

  ## ----- Change baseline slot ------------------------------------------------
  fit0@baseline <- list()

  ## ----- Change timing slot --------------------------------------------------
  t0 <- fit0@timing
  t1 <- fit_inlv$timing
  nms <- union(names(t0), names(t1))
  fit0@timing <- lapply(nms, function(n) {
    (if (n %in% names(t0)) t0[[n]] else 0) +
      (if (n %in% names(t1)) t1[[n]] else 0)
  })
  names(fit0@timing) <- nms
  fit0@timing$total <- NULL
  fit0@timing$total <- sum(unlist(fit0@timing))

  ## ----- Return --------------------------------------------------------------
  fit0@external <- list(
    inlavaan_internal = fit_inlv #[!grepl("lav", names(fit_inlv))]
  )
  fit0
}

testthat::skip()
test_that("All exported functions have @return and @examples", {
  # --- 1. Get Exported Functions (Text Parsing) ---
  if (!file.exists("NAMESPACE")) {
    stop("No NAMESPACE file found. Are you in the package root?")
  }
  ns_lines <- readLines("NAMESPACE")
  export_lines <- grep("^export\\(", ns_lines, value = TRUE)
  exported_funcs <- gsub("^export\\(|\\)$", "", export_lines)
  exported_funcs <- unlist(strsplit(exported_funcs, ","))
  exported_funcs <- gsub('"', '', trimws(exported_funcs))
  exported_funcs <- exported_funcs[exported_funcs != ""] # Remove empties

  cli::cli_alert_info("Found {length(exported_funcs)} exported functions.")

  # --- 2. Check Documentation ---
  man_files <- list.files("man", pattern = "\\.Rd$", full.names = TRUE)

  results <- data.frame(
    Function = character(),
    Has_Return = logical(),
    Has_Examples = logical(),
    stringsAsFactors = FALSE
  )

  for (func in exported_funcs) {
    # Find the .Rd file for this function
    rd_file <- NA
    for (f in man_files) {
      rd_content <- tryCatch(tools::parse_Rd(f), error = function(e) NULL)
      if (is.null(rd_content)) {
        next
      }

      # Extract aliases manually
      aliases <- unlist(lapply(rd_content, function(x) {
        if (!is.null(attr(x, "Rd_tag")) && attr(x, "Rd_tag") == "\\alias") {
          return(as.character(x))
        }
      }))

      if (func %in% aliases) {
        rd_file <- f
        break
      }
    }

    if (is.na(rd_file)) {
      results <- rbind(
        results,
        data.frame(Function = func, Has_Return = FALSE, Has_Examples = FALSE)
      )
      next
    }

    # --- FIX: Extract tags manually using attributes ---
    rd_data <- tools::parse_Rd(rd_file)
    # Get the tag for every top-level element in the Rd file
    tags <- sapply(rd_data, attr, "Rd_tag")

    has_return <- "\\value" %in% tags
    has_examples <- "\\examples" %in% tags

    results <- rbind(
      results,
      data.frame(
        Function = func,
        Has_Return = has_return,
        Has_Examples = has_examples
      )
    )
  }

  # --- 3. Report ---
  missing_docs <- results[!results$Has_Return | !results$Has_Examples, ]
  if (nrow(missing_docs) > 0) {
    cli::cli_h1("Documentation Check")
    cli::cli_alert_warning(
      "Found {nrow(missing_docs)} functions with missing tags:"
    )

    cli::cli_ul() # Start a list

    for (i in 1:nrow(missing_docs)) {
      row <- missing_docs[i, ]

      # Identify exactly which parts are missing
      missing_parts <- character()
      if (!row$Has_Return) {
        missing_parts <- c(missing_parts, "@return")
      }
      if (!row$Has_Examples) {
        missing_parts <- c(missing_parts, "@examples")
      }

      # Format the specific line
      # {.fn name} formats it as a function code style
      # {.val name} formats the missing tags nicely
      cli::cli_li(
        "Function {.fn {row$Function}} is missing: {.val {missing_parts}}"
      )
    }

    cli::cli_end() # End the list
  } else {
    cli::cli_alert_success(
      "All exported functions have {.strong @return} and {.strong @examples}!"
    )
  }

  expect_true(
    nrow(missing_docs) == 0,
    info = paste(
      "Some exported functions are missing @return or @examples tags. See details below."
    )
  )
})

---
title: 'Mediation Analysis'
# descriotion:
vignette: >
  %\VignetteIndexEntry{Mediation Analysis}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
# format:
#   html:
#     mermaid-format: png
knitr:
  opts_chunk: 
    collapse: true
    comment: '#>'
bibliography: refs.bib
---

Mediation analysis [@yuan2009bayesian] allows researchers to investigate the mechanism by which an independent variable ($X$) influences a dependent variable ($Y$). 
Rather than just asking "Does X affect Y?", mediation asks "Does X affect Y through an intermediate variable M?"

Common examples include:

 - **Psychology:** Does a therapy ($X$) reduce anxiety ($M$), which in turn improves sleep quality ($Y$)?
- **Medicine:** Does a new drug ($X$) lower blood pressure ($M$), thereby decreasing the risk of heart attack ($Y$)?

In this vignette, we demonstrate how to estimate a simple mediation model using `{INLAvaan}`.
We will fit a standard three-variable mediation model:

```{mermaid}
%%| fig-align: center
graph LR
    X((X)) -->|a| M((M))
    M -->|b| Y((Y))
    X -->|c| Y
```

- $a$: The effect of $X$ on $M$.
- $b$: The effect of $M$ on $Y$.
- $c$: The direct effect of $X$ on $Y$.
- $a \times b$: The indirect effect (the mediation effect).

In a mediation model, the *Total Effect* represents the overall impact of $X$ on $Y$, ignoring the specific pathway. 
It answers the question: 
"If I change $X$, how much does $Y$ change in _total_, regardless of whether it goes through $M$ or not?".

## Data Simulation 

To verify that `{INLAvaan}` recovers the correct parameters, we simulate data where the "truth" is known. 
The logic is as follows: Generate...

1. $X$ normally; 
2. $M$ dependent on $X$ with a coefficient of 0.5; and
3. $Y$ dependent only on $M$ with a coefficient of 0.7.

Critically, we do not add $X$ to the generation of $Y$. 
This means the true direct effect ($c$) is 0, and the relationship is fully mediated.
We expect our model to estimate $a \approx 0.5$, $b \approx 0.7$, and the indirect effect $ab \approx 0.35$. The direct effect $c$ should be close to zero.

```{r}
set.seed(11)
n <- 100  # sample size

# 1. Predictor
X <- rnorm(n)

# 2. Mediator (Path a = 0.5)
M <- 0.5 * X + rnorm(n)

# 3. Outcome (Path b = 0.7, Path c = 0)
Y <- 0.7 * M + rnorm(n) 

dat <- data.frame(X = X, Y = Y, M = M)
```

## Model Specification and Fit

The standard `lavaan` syntax for a mediation model is straightforward (note the use of the `:=` operator to define the indirect effect as a new parameter.):

```{r}
mod <- "
  # Direct effect (path c)
  Y ~ c*X

  # Mediator paths (path a and b)
  M ~ a*X
  Y ~ b*M

  # Define Indirect effect (a*b)
  ab := a*b

  # Define Total effect
  total := c + (a*b)
"
```

The model is fit using `asem()`.
The `meanstructure = TRUE` argument is supplied to estimate intercepts for the variables.

```{r}
library(INLAvaan)
fit <- asem(mod, dat, meanstructure = TRUE)
```

The user may wish to specify different prior distributions for the parameters.
See the relevant section in the [Get started](https://inlavaan.haziqj.ml/articles/INLAvaan.html#setting-priors) vignetted for further details.

## Results 

The summary output provides the posterior mean, standard deviation, and 95% credible intervals 
for all paths.

```{r}
summary(fit)
```

Looking at the Regressions and Defined Parameters sections of the output:

```{r}
#| include: false
summ <- get_inlavaan_internal(fit)$summary
fmt <- function(x) sprintf("%.3f", x)

a     <- fmt(summ["a", "Mean"])
b     <- fmt(summ["b", "Mean"])
c     <- fmt(summ["c", "Mean"])
c_lo  <- fmt(summ["c", "2.5%"])
c_hi  <- fmt(summ["c", "97.5%"])
ab    <- fmt(summ["ab", "Mean"])
ab_lo <- fmt(summ["ab", "2.5%"])
ab_hi <- fmt(summ["ab", "97.5%"])
tot   <- fmt(summ["total", "Mean"])
```


- Both intercepts are non-significant, since we simulated data with true means of zero.
- Path $a$ (`M ~ X`) estimated at `r a` (true value 0.5).
- Path $b$ (`Y ~ M`) estimated at `r b` (true value 0.7).
- Path $c$ (`Y ~ X`) estimated at `r c`. The 95% Credible Interval [`r c_lo`, `r c_hi`] includes zero, correctly identifying that there is no direct effect.
- Indirect Effect $ab$ estimated at `r ab` (true value 0.35). The interval [`r ab_lo`, `r ab_hi`] does not cross zero, indicating significant mediation.
- Total Effect estimated at `r tot`.
    * This is the sum of the direct and indirect effects ($c + ab$). 
    * It tells us that a 1-unit increase in $X$ leads to a total increase of roughly `r tot` in $Y$.
    * **Note:** In this simulation, even though the *direct* effect is non-significant (close to zero), the *total* effect is significant because the mechanism via $M$ is strong. This illustrates a "full mediation" scenario: $X$ affects $Y$, but *only* because of $M$.

## References

---
title: 'Get started'
description: > 
  A vignette introducing the INLAvaan package for Bayesian latent variable modeling using INLA.
vignette: >
  %\VignetteIndexEntry{Introduction to INLAvaan}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
knitr:
  opts_chunk: 
    collapse: true
    comment: '#>'
bibliography: refs.bib
toc: true
---

```{r}
#| label: setup
#| include: false
library(INLAvaan)
library(lavaan)
```


## Introduction

SEMs are ubiquitous in the social sciences, psychology, ecology, and other fields. 
The `{INLAvaan}` package [@jamil2026implementation] provides a user-friendly interface for fitting Bayesian SEMs using Integrated Nested Laplace Approximations [INLA, @rue2009approximate], based on a bespoke approximate Bayesian inference framework for SEMs [@jamil2026approximate].
This vignette will guide you through the basics of using `{INLAvaan}` to fit a simple SEM.
Before we begin make sure you have installed the INLAvaan package from GitHub by running the commands below.

```{r}
#| label: install-instructions
#| eval: false
# Load all libraries for this example
library(INLAvaan)
library(tidyverse)
library(lavaan)
```


## Motivating example

To motivate the use of SEMs, consider the introductory example in @song2012basic: 
*Does poorer glycemic control lead to greater severity of kidney disease?*
We observe three indicators of glycemic control ($y_1$, $y_2$, $y_3$) and three indicators of kidney disease severity ($y_4$, $y_5$, $y_6$).

| | **Indicator** | **Description** | **Unit** |
|---|---|---|---|
| $y_1$ | HbA1c | 3-month avg. blood glucose | % |
| $y_2$ | FPG | Fasting plasma glucose | mmol/L |
| $y_3$ | Insulin | Fasting insulin level | μU/mL |
| $y_4$ | PCr | Plasma creatinine | μmol/L |
| $y_5$ | ACR | Albumin–creatinine ratio | mg/g |
| $y_6$ | BUN | Blood urea nitrogen | mmol/L |

Rather than fitting separate regression models for each indicator, SEM allows us to model the relationship between the latent constructs themselves, providing a clearer and more coherent representation of the underlying processes.
The hypothesised SEM is illustrated by the figure below:

![](https://haziqj.ml/sembias-gradsem/index_files/figure-revealjs/unnamed-chunk-4-1.png){fig-align="center" width="75%"}

## Data

For the two-factor SEM we described above, it is easy to simulate some data using the `{lavaan}` package to do so.
The code is below:

```{r}
#| label: simulate-data
pop_mod <- "
  eta1 =~ 1*y1 + 0.8*y2 + 0.6*y3
  eta2 =~ 1*y4 + 0.8*y5 + 0.6*y6
  eta2 ~ 0.3*eta1
  
  # Variances
  y1 ~~ 0.5*y1
  y2 ~~ 0.5*y2
  y3 ~~ 0.5*y3
  y4 ~~ 0.5*y4
  y5 ~~ 0.5*y5
  y6 ~~ 0.5*y6
  eta1 ~~ 1*eta1
  eta2 ~~ 1*eta2
"
set.seed(123)
dat <- lavaan::simulateData(pop_mod, sample.nobs = 1000)
str(dat)
```

From the code above, note the true values of the parameters, including the factor loadings $\Lambda$, regression coefficient $\beta$ between the two latent variables, as well as the residual and latent variances $\Theta$ and $\Psi$ respectively.

## Model fit

Now that we have simulated some data, we can fit the SEM using INLAvaan.
The model syntax is similar to that of `{lavaan}`, making it easy to specify the model.
For further details on the model syntax, refer to the [lavaan website](https://lavaan.ugent.be/tutorial/syntax1.html).
`{INLAvaan}` provides mirror functions for the main model fitting functions in `{lavaan}`:

- `acfa()` mirrors `lavaan::cfa()` for confirmatory factor analysis; and
- `asem()` mirrors `lavaan::sem()` for structural equation models;
- `agrowth()` mirrors `lavaan::growth()` for latent growth curve models.

The code to fit the SEM model is below:

```{r}
#| label: fit-model
mod <- "
  eta1 =~ y1 + y2 + y3
  eta2 =~ y4 + y5 + y6
  eta2 ~ eta1
"
fit <- asem(mod, dat)
```

As with `lavaan::sem()`, this model is fitted without a mean structure by default (`meanstructure = FALSE`).
Rather than profiling the means out as lavaan does, `{INLAvaan}` assigns the saturated means flat priors and marginalises them analytically, which keeps the likelihood a proper Bayesian object.
Intercepts can instead be modelled explicitly with `meanstructure = TRUE`.
See the [mean structures article](https://inlavaan.haziqj.ml/articles/meanstructure.html) for how the two treatments relate and when model comparisons across them are meaningful.

`{INLAvaan}` computes an approximation to the posterior density by way of a Laplace approximation [@tierney1989fully; @jamil2026approximate].
The joint mode and the Hessian needs to be computed, which gives a Gaussian distribution for the joint posterior of the parameters.
The default method for optimisation is `stats::nlminb()`, but other optimisers can be used by specifying `optim_method = "ucminf"` for the `{ucminf}` package or `optim_method = "optim"` to call the `stats::optim()` function with method `"BFGS"`.

From this, marginal posterior distributions for each parameter can be obtained by one of several ways, including 1) Skew normal fitting [`marginal_method = "skewnorm"`, the default method, see @chiuchiolo2023joint]; 2) Two-piece asymmetric Gaussian fitting [`marginal_method = "asymgaus"`, see @martins2013bayesian]; 3) Direct marginalisation of the joint Gaussian posterior (`marginal_method = "marggaus"`); and 4) Sampling from the joint Gaussian posterior (`marginal_method = "sampling"`).

Once the marginal posterior distributions have been obtained, we can further use these to compute any derived quantities of interest via copula sampling.
The posterior predictive p-values [@gelman1996posterior] and Deviance Information Criterion [DIC, @spiegelhalter2002bayesian] are computed this way.
Under the default `test = "standard"`, leave-one-out cross-validation and WAIC results (see the model comparison section below) are also computed at fit time and stored with the fit, whenever the model supports them and the additional cost is small.
Often, the posterior sampling takes longer than the model fitting itself, so the number of samples can be controlled via the `nsamp` argument (default is `nsamp = 1000`), or the post-fitting computations can be skipped altogether (`test = "none"`).

## Methods

The resulting object is of class `INLAvaan`, a subclass of `lavaan` objects.

```{r}
#| label: results-str
str(fit, 1)
fit
```

As a result, most of the methods that work for `lavaan` objects will also work for `INLAvaan` objects.
The most common ones are probably `coef()` and `summary()`.

```{r}
#| label: results-coef-summary
# Inspect coefficients
coef(fit)

# Summary of results
summary(fit)
```

It's possible to request posterior medians and modes in the summary output by specifying `postmedian = TRUE` or `postmode = TRUE` in the `summary()` function.

### Predictions

Predicted values for the latent variables can be obtained using the `predict()` function.
This is done by sampling from the posterior distributions of the latent variables given the observed data.
The function also supports predictions for observed variables (e.g. `type = "ov"`) and missing data imputation, respecting multilevel structure if present.

```{r}
#| label: results-pred
eta_preds <- predict(fit, nsamp = 100)
length(eta_preds)
head(eta_preds[[1]])
```

This is an S3 object with a summary method that provides posterior means and credible intervals for the latent variables.
Alternatively, the user is welcome to perform their own summary statistics on the list of posterior samples returned by `predict()`.

```{r}
#| label: results-pred-summary
summ_eta <- summary(eta_preds)
str(summ_eta)
head(summ_eta$Mean)
```

### Predictive checks

The generative side of the model is exposed by `sampling()`, which draws parameters, latent variables, or observed variables from the posterior (or prior) generative SEM, and by `simulate()`, which produces complete replicate data sets from the fitted model.

```{r}
#| label: results-simulate
yrep <- simulate(fit, nsim = 1, seed = 1)
head(yrep[[1]])
```

Comparing such replicates with the observed data is the basis of prior and posterior predictive checking.
The [predictive checks article](https://inlavaan.haziqj.ml/articles/predictive-checks.html) walks through that workflow, and the [sampling article](https://inlavaan.haziqj.ml/articles/sampling.html) explains the machinery behind it.

### Fit measures

Global fit measures are collected by `fitmeasures()`: the PPP and DIC mentioned earlier, Bayesian analogues of the classical fit indices (BRMSEA, BGammaHat, and related indices), and the LOO and WAIC measures of the model comparison section below when these are stored with the fit.

```{r}
#| label: results-fitmeasures
fitmeasures(fit)
```

Definitions and worked examples are in the [Bayesian fit indices article](https://inlavaan.haziqj.ml/articles/fit-indices.html).

### Diagnostics

The `diagnostics()` function reports convergence and approximation-quality metrics for the fitted model.
Global diagnostics (`type = "global"`) check whether the optimiser converged, and quantify how well the skew-normal marginals match the joint posterior (via KL divergence and NMAD).
Per-parameter diagnostics (`type = "param"`) provide gradient norms and KL contributions for each free parameter, which is useful for identifying any problematic parameters.

```{r}
#| label: results-diagnostics
diagnostics(fit)
```

The `timing()` function reports how long each computation stage took, which can help identify bottlenecks when scaling to larger models.

```{r}
#| label: results-timing
timing(fit)
```

### Plot

A simple plot method is provided to view the marginal posterior distributions of the parameters.
The vertical lines indicate the posterior mode.

```{r}
#| label: results-plot
plot(fit)
```

## Model comparison

In addition to several global fit indices (i.e. PPP, DIC), it is possible to compare models by way of Bayes factors using the `compare()` function. 
This function takes two `INLAvaan` objects and computes the Bayes factor using the Laplace approximations to the marginal likelihoods.

```{r}
mod2 <- "
  # A model with uncorrelated factors
  eta1 =~ y1 + y2 + y3
  eta2 =~ y4 + y5 + y6
  eta1 ~~ 0*eta2
"
fit2 <- asem(mod2, dat)
compare(fit, fit2)
```

As a note, there have been several criticisms of the use of Bayes factors for model comparison, particularly in the context of SEMs [@tendeiro2019review; @schad2023workflow].
Leave-one-out cross-validation [LOO, @vehtari2017practical] and the widely applicable information criterion [WAIC, @watanabe2010asymptotic] are popular alternatives which compare models on out-of-sample predictive accuracy instead; see @merkle2019bayesian for their use with latent variable models.
Both are implemented in `{INLAvaan}`.
The function `loo()` requires neither refitting nor sampling, computing the statistic using a Taylor approximation of the case-deletion posterior, while `waic()` reuses the fit's own posterior draws.

```{r}
#| label: results-loo
loo(fit)
```

To compare models on this predictive scale, pass `loo = TRUE` to `compare()`.
Models are sorted by expected log predictive density (ELPD), with paired standard errors for the ELPD differences.

```{r}
#| label: results-compare-loo
compare(fit, fit2, loo = TRUE)
```

See the [cross-validation article](https://inlavaan.haziqj.ml/articles/loo.html) for the methodology and more advanced usage, including multigroup and two-level models, missing data, and scoring submodels without refitting.


## Setting priors

The `{INLAvaan}` package uses the same prior specification syntax as `{blavaan}` [@merkle2018blavaan;@merkle2021efficient], as detailed [here](https://blavaan.org/articles/prior.html).
Essentially, there are two ways to set priors for model parameters:
1) Globally for all parameters of a certain type (e.g., all factor loadings, all regression coefficients, etc.); and
2) Individually for specific parameters in the model syntax.

The default global priors are similar to those from `{blavaan}`:

```{r}
#| label: priors-default
priors_for()  # similar to blavaan::dpriors()
```

Note that, `{INLAvaan}` uses the separation strategy for variance matrices, and consequently places priors on **correlations** instead of *covariances*.
If, instead we wished to set global priors, say a gamma distribution on **variances** instead of **standard deviations** (default), then we would do the following:

```{r}
#| label: priors-global
DP <- priors_for(theta = "gamma(1,1)", psi = "gamma(1,1)")
DP
## fit <- asem(mod, dat, dpriors = DP)  # not run
```

To set individual priors for specific parameters, we can do so in the model syntax itself.
For instance, to set a normal prior with mean 1 and standard deviation 3 for the factor loading of `y3` on `eta1`, and a normal prior with mean 0 and standard deviation 0.5 for the regression coefficient from `eta1` to `eta2`, we would specify the model as follows:

```{r}
mod <- "
  eta1 =~ y1 + y2 + prior('normal(1,3)')*y3
  eta2 =~ y4 + y5 + y6
  eta2 ~ prior('normal(0,.5)')*eta1
"
## fit <- asem(mod, dat)  # not run
```

## Dependency on R-INLA

Dependency on R-INLA has been temporarily removed for the current version of `{INLAvaan}` (>= 0.2.0).
For a wide class LVMs and SEMs where the latent variables are unstructured and independent, the current implementation is sufficient.
However, future versions of `{INLAvaan}` will re-introduce dependency on R-INLA to allow for more complex latent structures, such as spatial and temporal dependencies.

## References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skew-normal.R
\name{dsnorm}
\alias{dsnorm}
\title{The Skew Normal Distribution}
\usage{
dsnorm(x, xi, omega, alpha, logC = 0, log = FALSE)
}
\arguments{
\item{x}{Vector of quantiles.}

\item{xi}{Location parameter.}

\item{omega}{Scale parameter.}

\item{alpha}{Shape parameter.}

\item{logC}{Log-normalization constant.}

\item{log}{Logical; if TRUE, returns the log density.}
}
\value{
A numeric vector of (log) density values.
}
\description{
Density for the skew normal distribution with location \code{xi}, scale \code{omega},
and shape \code{alpha}.
}
\examples{
x <- seq(-2, 5, length.out = 100)
y <- dsnorm(x, xi = 0, omega = 1, alpha = 5)
plot(x, y, type = "l", main = "Skew Normal Density")
}
\references{
https://en.wikipedia.org/wiki/Skew_normal_distribution
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{dbeta_box}
\alias{dbeta_box}
\title{Density of a Beta distribution on a bounded interval}
\usage{
dbeta_box(x, shape1, shape2, a, b, log = FALSE)
}
\arguments{
\item{x}{A numeric vector of quantiles.}

\item{shape1, shape2}{non-negative parameters of the Beta distribution.}

\item{a}{The lower bound of the interval.}

\item{b}{The upper bound of the interval.}

\item{log}{Logical; if TRUE, probabilities p are given as log(p).}
}
\value{
A numeric vector of density values.
}
\description{
Density of a Beta distribution on a bounded interval
}
\examples{
# Beta(2,5) on (0,100)
x <- seq(0, 100, length.out = 100)
y <- dbeta_box(x, shape1 = 2, shape2 = 5, a = 0, b = 100)
plot(x, y, type = "l", main = "Beta(2,5) on (0,100)")

# Beta(1,1) i.e. uniform on (-1, 1)
x <- seq(-1, 1, length.out = 100)
y <- dbeta_box(x, shape1 = 1, shape2 = 1, a = -1, b = 1)
plot(x, y, type = "l", main = "Beta(1,1) on (-1,1)")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-waic.R
\name{waic}
\alias{waic}
\alias{waic.INLAvaan}
\alias{waic.inlavaan_internal}
\title{Widely Applicable Information Criterion for INLAvaan Models}
\usage{
waic(x, ...)

\method{waic}{INLAvaan}(
  x,
  type = c("auto", "loso", "loco"),
  units = NULL,
  nsamp = NULL,
  cores = NULL,
  verbose = FALSE,
  ...
)

\method{waic}{inlavaan_internal}(
  x,
  type = c("auto", "loso", "loco"),
  units = NULL,
  nsamp = NULL,
  cores = NULL,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{x}{A fitted \link{INLAvaan} object (or its \code{inlavaan_internal} list).}

\item{...}{Not used.}

\item{type}{Unit type: \code{"auto"} (default) resolves to per-subject for
single-level models and per-cluster (marginal WAIC) for two-level
models. \code{"loso"} on a two-level model scores the conditional
(leave-one-unit-out) WAIC instead (with a warning; see Details);
\code{"loco"} cannot be forced on a model without clusters.}

\item{units}{Optional integer vector of unit indices to score; defaults
to all units.}

\item{nsamp}{Number of posterior draws. Defaults to the \code{nsamp} used when
fitting the model.}

\item{cores}{Number of cores for evaluating draws. The default \code{NULL}
runs serially; parallelism must be requested explicitly.}

\item{verbose}{Logical; print progress (default \code{FALSE}).}
}
\value{
An object of class \code{inlavaan_waic}: a list with \code{per_unit}
(pointwise \code{lpd}, \code{p_waic}, \code{elpd_waic}, with the same \code{unit}/\code{group}
identification as \code{\link[=loo]{loo()}}), \code{estimates} (matrix with rows
\code{elpd_waic}, \code{p_waic}, \code{waic} and columns \code{Estimate}, \code{SE}), \code{type},
\code{n_units}, \code{n_groups}, and \code{nsamp}.
}
\description{
Computes the WAIC of a fitted \link{INLAvaan} model from unit log-likelihoods
evaluated over posterior draws. Single-level models are scored per
subject; two-level models are scored per cluster by default, matching the
units used by \code{\link[=loo]{loo()}}. For a two-level model \code{type = "loso"} instead
scores the \emph{conditional} (leave-one-unit-out) WAIC; see Details.
}
\details{
For each posterior draw \eqn{\theta_s} (drawn with the same copula
sampler used for the fit's posterior summaries) and unit \eqn{u}, the
log-likelihood \eqn{\log p(y_u \mid \theta_s)} is evaluated; then
\eqn{\mathrm{lpd}_u = \log \tfrac{1}{S}\sum_s p(y_u \mid \theta_s)},
\eqn{p_{\mathrm{waic},u} = \mathrm{var}_s \log p(y_u \mid \theta_s)}, and
\eqn{\mathrm{elpd}_{\mathrm{waic}} = \sum_u (\mathrm{lpd}_u -
p_{\mathrm{waic},u})} with \eqn{\mathrm{WAIC} = -2\,
\mathrm{elpd}_{\mathrm{waic}}}. Unlike \code{\link[=loo]{loo()}}, this is a sampling-based
estimate: results vary with the random draws, and units with
\eqn{p_{\mathrm{waic},u} > 0.4} trigger a reliability warning (also
annotated when printing). The same model restrictions as \code{\link[=loo]{loo()}} apply,
and so does the flavour rule: fits with \code{fixed.x = TRUE} are scored
conditionally on the exogenous covariates, fits with \code{fixed.x = FALSE}
jointly (see \code{\link[=loo]{loo()}}).

\strong{Marginal vs conditional WAIC (two-level models).} The default
per-cluster scoring is the \emph{marginal} WAIC, which corresponds to
leave-one-cluster-out cross-validation -- prediction for a \emph{new} cluster.
Setting \code{type = "loso"} scores the \emph{conditional} WAIC, corresponding to
leave-one-unit-out -- prediction for a new observation within an
\emph{observed} cluster (each row contributes the conditional density of its
observed entries given the rest of its cluster). The two answer different
questions and are easily conflated (Merkle, Furr & Rabe-Hesketh, 2019);
the per-cluster marginal is the usual model-comparison target, so it is
the default, and \code{type = "loso"} warns. This matches \code{loo(type = "loso")}
-- the two compute the same estimand by sampling and by Taylor expansion.

Under the default \code{test = "standard"}, \code{\link[=inlavaan]{inlavaan()}} computes the WAIC at
fit time by reusing the posterior draws the fit already produced (when
the model is supported and \code{nsamp >= 100}), and
stores it with the fit: \code{waic(fit)} then returns the stored result when
called with default arguments, and \code{\link[=fitmeasures]{fitmeasures()}} reports \code{waic},
\code{p_waic}, \code{se_waic} as part of \code{"all"} for free. If the \code{loo} package is
attached it masks this generic, but dispatch on INLAvaan objects
continues to work.
}
\examples{
\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, meanstructure = TRUE)
waic(fit)
}

}
\seealso{
\code{\link[=loo]{loo()}}, \code{\link[=fitmeasures]{fitmeasures()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skew-normal.R
\name{fit_skew_normal}
\alias{fit_skew_normal}
\title{Fit a skew normal distribution to log-density evaluations}
\usage{
fit_skew_normal(x, y, threshold_log_drop = -6, temp = NA)
}
\arguments{
\item{x}{A numeric vector of points where the density is evaluated.}

\item{y}{A numeric vector of log-density evaluations at points x.}

\item{threshold_log_drop}{A negative numeric value indicating the log-density
drop threshold below which points are ignored in the fitting. Default is
-6.}

\item{temp}{A numeric value for the temperature parameter k. If NA (default),
it is included in the optimisation.}
}
\value{
A list with fitted parameters:
\itemize{
\item \code{xi}: location parameter
\item \code{omega}: scale parameter
\item \code{alpha}: shape parameter
\item \code{logC}: log-normalization constant
\item \code{k}: temperature parameter
\item \code{rsq}: R-squared of the fit
}

Note that \code{logC} and \code{k} are not used when fitting from a sample.
}
\description{
Fit a skew normal distribution to log-density evaluations
}
\details{
This skew normal fitting function uses a weighted least squares
approach to fit the log-density evaluations provided in \code{y} at points \code{x}.
The weights are set to be the density evaluations raised to the power of
the temperature parameter \code{k}. This has somewhat an interpretation of
finding the skew normal fit that minimises the Kullback-Leibler divergence
from the true density to it.

In R-INLA, the C code implementation from which this was translated from
can be found
\href{https://github.com/hrue/r-inla/blob/b63eb379f69d6553f965b360f1b88877cfef20d1/gmrflib/fit-sn.c}{here}.
}
\examples{
# Fit a SN curve to gamma log-density
logdens <- function(x) dgamma(x, shape = 3, rate = 1, log = TRUE)

x_grid <- seq(0.1, 8, length.out = 21)
y_log  <- sapply(x_grid, logdens)
y_log  <- y_log - max(y_log)  # normalise to have maximum at zero

res <- fit_skew_normal(x_grid, y_log, temp = 10)
unlist(res)

# Compare truth vs skew-normal approximation
x_fine <- seq(0.1, 8, length.out = 200)
y_true <- exp(logdens(x_fine))
y_sn   <- dsnorm(x_fine, xi = res$xi, omega = res$omega, alpha = res$alpha)

plot(x_fine, y_true, type = "n", xlab = "x", ylab = "Density", bty = "n")
polygon(c(x_fine, rev(x_fine)), c(y_true, rep(0, length(x_fine))),
        col = adjustcolor("#131516", 0.25), border = NA)
lines(x_fine, y_sn, col = "#00A6AA", lwd = 2)
legend("topright", legend = c("Truth", "SN Approx."),
       fill = c(adjustcolor("#131516", 0.25), NA), border = NA,
       col = c(NA, "#00A6AA"), lwd = c(NA, 2), lty = c(NA, 1),
       bty = "n")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-plot.R
\name{plot}
\alias{plot}
\alias{plot,INLAvaan,ANY-method}
\title{Plot an INLAvaan Object}
\usage{
\S4method{plot}{INLAvaan,ANY}(
  x,
  y,
  type = c("marg_pdf", "sn_fit", "sn_fit_log"),
  params = "all",
  nrow = NULL,
  ncol = NULL,
  use_ggplot = TRUE,
  points = FALSE,
  ...
)
}
\arguments{
\item{x}{An object of class \link{INLAvaan}.}

\item{y}{Not used.}

\item{type}{Character. One of \code{"marg_pdf"} (default; posterior marginal
densities), \code{"sn_fit"} (skew-normal fit diagnostic on natural scale),
or \code{"sn_fit_log"} (same on log scale).}

\item{params}{Character vector of parameter names to plot, or \code{"all"}
(default) to plot all free parameters.}

\item{nrow, ncol}{Integer. Number of rows/columns for the facet grid when
\code{use_ggplot = TRUE}. If \code{NULL} (default), layout is chosen
automatically.}

\item{use_ggplot}{Logical. When \code{TRUE} (default) and \pkg{ggplot2} is
available, a \pkg{ggplot2} object is returned. Set to \code{FALSE} for a
base-R plot.}

\item{points}{Logical. When \code{TRUE}, individual grid points are overlaid
on the curves. Defaults to \code{FALSE}.}

\item{...}{Additional arguments (currently unused).}
}
\description{
Generates diagnostic plots for a fitted \code{INLAvaan} model.
}
\examples{
\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
            test = "none", verbose = FALSE)

# Posterior marginal densities (default)
plot(fit)

# Skew-normal fit diagnostic for a single parameter
plot(fit, type = "sn_fit", params = "visual=~x1")
}

}
\seealso{
\code{\link[=diagnostics]{diagnostics()}}, \code{\link[=summary]{summary()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors_for.R
\name{priors_for}
\alias{priors_for}
\title{Specify priors for a SEM}
\usage{
priors_for(...)
}
\arguments{
\item{...}{Named arguments specifying prior distributions for lavaan
parameter types.}
}
\value{
A named character vector of prior specifications, where names
correspond to lavaan parameter types (e.g., "lambda", "beta", "theta",
etc.) and values are character strings specifying the prior distribution
(e.g., \code{"normal(0,10)"}, \code{"gamma(1,0.5)[sd]"},
\code{"gamma(1,1)[prec]"}, etc.).
}
\description{
Specify priors for a SEM, similar to how \code{\link[blavaan:dpriors]{blavaan::dpriors()}} works.
}
\details{
This function provides a convenient way to specify prior distributions for
different types of parameters in a structural equation model (SEM). It uses a
registry of default priors for common lavaan parameter types (e.g., loadings,
regressions, residuals, etc.) and allows users to override these defaults by
passing named arguments.

The parameter names, and default settings, are:
\itemize{
\item \code{nu = "normal(0,32)"}: Observed variable intercepts
\item \code{alpha = "normal(0,10)"}: Latent variable intercepts
\item \code{lambda = "normal(0,10)"}: Factor loadings
\item \code{beta = "normal(0,10)"}: Regression coefficients
\item \code{theta = "gamma(1,.5)[sd]"}: Residual precisions
\item \code{psi = "gamma(1,.5)[sd]"}: Latent variable precisions
\item \code{rho = "beta(1,1)"}: Correlations (both latent and observed)
\item \code{tau = "normal(0,1.5)"}: Thresholds for ordinal variables
}

Note that the normal distributions are parameterised using standard
deviations, and not variances. For example, \code{normal(0,10)} means a
normal distribution with mean 0 and standard deviation 10 (not variance 10).
}
\section{Scale qualifiers}{

For variance parameters (\code{theta}, \code{psi}), the prior distribution
can be placed on a transformed scale by appending a qualifier:
\itemize{
\item \code{[sd]}: Prior is on the standard deviation \eqn{\sigma}.
Example: \code{"gamma(1,0.5)[sd]"} places a Gamma(1, 0.5) prior on
\eqn{\sigma = \sqrt{\text{variance}}}.
\item \code{[prec]}: Prior is on the precision \eqn{\tau = 1/\sigma^2}.
Example: \code{"gamma(1,1)[prec]"} places a Gamma(1, 1) prior on
\eqn{\tau = 1/\text{variance}}. This is the parameterisation used by
blavaan and corresponds to an Inverse-Gamma prior on the variance.
}
The necessary Jacobian adjustment is applied automatically in both cases.
}

\examples{
priors_for(nu = "normal(0,10)", lambda = "normal(0,1)", rho = "beta(3,3)")

# Precision-scale prior for residual variances (blavaan-style)
priors_for(theta = "gamma(1,1)[prec]")
}
\seealso{
\code{\link[=inlavaan]{inlavaan()}}, \code{\link[=acfa]{acfa()}}, \code{\link[=asem]{asem()}}, \code{\link[=agrowth]{agrowth()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_inlavaan_internal}
\alias{get_inlavaan_internal}
\title{Extract the Internal INLAvaan Object}
\usage{
get_inlavaan_internal(object, what)
}
\arguments{
\item{object}{An object of class \link{INLAvaan}.}

\item{what}{Character. Name of the element to extract from the internal
list. If missing, the entire list is returned. Common elements include
\code{"coefficients"}, \code{"summary"}, \code{"Sigma_theta"}, \code{"vcov_x"},
\code{"theta_star"}, \code{"approx_data"}, \code{"pdf_data"}, \code{"partable"},
\code{"marginal_method"}, \code{"nsamp"}, \code{"mloglik"}, \code{"DIC"}, \code{"ppp"},
\code{"vb"}, \code{"opt"}, \code{"timing"}, \code{"visual_debug"}.}
}
\value{
The full \code{inlavaan_internal} list, or the named element when
\code{what} is supplied.
}
\description{
Returns the \code{inlavaan_internal} list stored inside a fitted \link{INLAvaan}
object, optionally extracting a single named element.
}
\examples{
\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
            test = "none", verbose = FALSE)

# Full internal object
int <- get_inlavaan_internal(fit)
names(int)

# Extract a specific element
get_inlavaan_internal(fit, "coefficients")
}

}
\seealso{
\link[=INLAvaan-class]{INLAvaan}, \code{\link[=diagnostics]{diagnostics()}}, \code{\link[=timing]{timing()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/INLAvaan-package.R
\docType{package}
\name{INLAvaan-package}
\alias{INLAvaan-package}
\alias{INLAvaan}
\title{INLAvaan: Approximate Bayesian Latent Variable Analysis}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Implements approximate Bayesian inference for Structural Equation Models (SEM) using a custom adaptation of the Integrated Nested Laplace Approximation (Rue et al., 2009) \doi{10.1111/j.1467-9868.2008.00700.x} as described in Jamil and Rue (2026a) \doi{10.48550/arXiv.2603.25690}. Provides a computationally efficient alternative to Markov Chain Monte Carlo (MCMC) for Bayesian estimation, allowing users to fit latent variable models using the 'lavaan' syntax. See also the companion paper on implementation and workflows, Jamil and Rue (2026b) \doi{10.48550/arXiv.2604.00671}.
}
\section{Main features}{

\itemize{
\item \code{\link[=acfa]{acfa()}}: Approximate Confirmatory Factor Analysis.
\item \code{\link[=asem]{asem()}}: Approximate Structural Equation Modelling.
\item \code{\link[=agrowth]{agrowth()}}: Approximate Latent Growth Curve models.
}
}

\section{Model specifications}{

Supports advanced 'lavaan' syntax features, including:
\itemize{
\item Equality constraints
\item Defined parameters (e.g., \verb{:=} operator for indirect effects)
\item Flexible prior specifications
}
}

\section{Methods for \code{INLAvaan} objects}{

After fitting a model an \link{INLAvaan} object is returned. The following S4
methods are available. See \link[=INLAvaan-class]{INLAvaan} for the class definition.
\itemize{
\item Summaries and parameter estimates: \code{\link[=summary]{summary()}}, \code{\link[=coef]{coef()}}, \code{\link[=vcov]{vcov()}}, \code{\link[=standardisedsolution]{standardisedsolution()}}
\item Fit assessment and model comparison: \code{\link[=fitmeasures]{fitmeasures()}}, \code{\link[=bfit_indices]{bfit_indices()}}, \code{\link[=compare]{compare()}}, \code{\link[=diagnostics]{diagnostics()}}, \code{\link[=timing]{timing()}}
\item Posterior inference and simulation: \code{\link[=predict]{predict()}}, \code{\link[=sampling]{sampling()}}, \code{\link[=simulate]{simulate()}}
\item Visualisation: \code{\link[=plot]{plot()}}
}
}

\section{Online vignettes}{

The \href{https://inlavaan.haziqj.ml/}{package website} contains comprehensive examples covering:
\itemize{
\item Confirmatory Factor Analysis (CFA)
\item Structural Equation Models (SEM)
\item Latent Growth Curve Models
\item Multigroup and Invariance Testing
\item Mediation Analysis
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://inlavaan.haziqj.ml/}
  \item \url{https://github.com/haziqj/INLAvaan}
  \item Report bugs at \url{https://github.com/haziqj/INLAvaan/issues}
}

}
\author{
\strong{Maintainer}: Haziq Jamil \email{haziq.jamil@gmail.com} (\href{https://orcid.org/0000-0003-3298-1010}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Haziq Jamil \email{haziq.jamil@gmail.com} (\href{https://orcid.org/0000-0003-3298-1010}{ORCID}) [copyright holder]
}

Other contributors:
\itemize{
  \item Håvard Rue (\href{https://orcid.org/0000-0002-0222-1881}{ORCID}) (Statistical and computational methodology) [contributor]
  \item Alvin Bong (Initial site build) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-predict.R
\name{predict}
\alias{predict}
\alias{predict,INLAvaan-method}
\title{Posterior Predictions for INLAvaan Models}
\usage{
\S4method{predict}{INLAvaan}(
  object,
  type = c("lv", "yhat", "ov", "ypred", "ydist", "ymis", "ovmis"),
  newdata = NULL,
  level = 1L,
  nsamp = 1000,
  ymis_only = FALSE,
  summary = FALSE,
  ...
)
}
\arguments{
\item{object}{An object of class \link{INLAvaan}.}

\item{type}{Character string specifying the type of prediction:
\describe{
\item{\code{"lv"}}{(default) Posterior draws of latent variable scores
\eqn{\eta | y, \theta}.}
\item{\code{"yhat"}, \code{"ov"}}{Predicted means for observed variables
\eqn{E(y | \eta, \theta) = \nu + \Lambda \eta}; no residual noise.}
\item{\code{"ypred"}, \code{"ydist"}}{Predicted observed values including
residual noise \eqn{y = \nu + \Lambda \eta + \varepsilon},
\eqn{\varepsilon \sim N(0, \Theta)}.}
\item{\code{"ymis"}, \code{"ovmis"}}{Imputed values for missing
observations, drawn from the conditional distribution
\eqn{y_{mis} | y_{obs}, \theta}.}
}}

\item{newdata}{An optional data frame of new observations. If supplied,
predictions are computed for \code{newdata} rather than the original
training data. Not supported for \code{type = "ymis"}.}

\item{level}{Integer; for \code{type = "lv"} in multilevel models, specifies
whether level 1 or level 2 latent variables are desired (default \code{1L}).}

\item{nsamp}{Integer; number of posterior samples to use for prediction.
Defaults to \code{1000}.}

\item{ymis_only}{Logical; only applies when \code{type = "ymis"}. When
\code{TRUE}, returns only the imputed values as a named numeric vector per
sample (names of the form \code{"varname[rowindex]"}, matching the blavaan
convention). When \code{FALSE} (default), returns the full data matrix with
missing values filled in.}

\item{summary}{Logical. When \code{TRUE}, collapse the posterior draws with
\code{summary()} and return summary statistics (mean, SD, quantiles, mode)
instead of the raw draws -- equivalent to calling
\code{summary(predict(object, ...))} but without materialising the
intermediate draws object. Default \code{FALSE}.}

\item{...}{Currently unused.}
}
\value{
A list of \code{nsamp} posterior draws, each a matrix (or data
frame, for multiple groups) with rows corresponding to cases and columns
to variables or latent factors. When \code{summary = TRUE}, instead
returns a \code{summary.predict.inlavaan_internal} object with the
posterior mean, SD, quantiles, and mode for each case/variable.
}
\description{
Compute posterior predictions from a fitted \code{INLAvaan} model,
including latent variable scores, predicted observed values, and imputed
missing data.
}
\examples{
\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
            test = "none", verbose = FALSE)

# Posterior latent variable scores
lv_scores <- predict(fit)
head(lv_scores)

# Predicted observed variable means
yhat <- predict(fit, type = "yhat")
head(yhat)

# Point estimates only, skipping the manual summary() step
predict(fit, type = "yhat", summary = TRUE)
}

}
\seealso{
\code{\link[=sampling]{sampling()}}, \code{\link[=simulate]{simulate()}}, \code{\link[=summary]{summary()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skew-normal.R
\name{fit_skew_normal_samp}
\alias{fit_skew_normal_samp}
\title{Fit a skew normal distribution to a sample}
\usage{
fit_skew_normal_samp(x)
}
\arguments{
\item{x}{A numeric vector of sample data.}
}
\value{
A list with fitted parameters:
\itemize{
\item \code{xi}: location parameter
\item \code{omega}: scale parameter
\item \code{alpha}: shape parameter
\item \code{logC}: log-normalization constant
\item \code{k}: temperature parameter
\item \code{rsq}: R-squared of the fit
}

Note that \code{logC} and \code{k} are not used when fitting from a sample.
}
\description{
Fit a skew normal distribution to a sample
}
\details{
Uses maximum likelihood estimation to fit a skew normal distribution
to the provided numeric vector \code{x}.
}
\examples{
x <- rnorm(100, mean = 5, sd = 1)
unlist(fit_skew_normal_samp(x))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-update.R
\name{update}
\alias{update}
\alias{update,INLAvaan-method}
\title{Update and re-fit an INLAvaan model}
\usage{
\S4method{update}{INLAvaan}(object, model, add, ..., evaluate = TRUE)
}
\arguments{
\item{object}{An object of class \link{INLAvaan}.}

\item{model}{Optional replacement model, in the same form accepted by
\code{\link[=inlavaan]{inlavaan()}} (lavaan model syntax or a parameter table).}

\item{add}{Optional lavaan syntax appended to the original model. Requires
the original model to have been specified as a syntax string, and is
ignored when \code{model} is supplied.}

\item{...}{Further arguments passed to \code{\link[=inlavaan]{inlavaan()}}, overriding the values
in the original call (e.g. \code{dp}, \code{marginal_method}, \code{nsamp}, \code{data}).}

\item{evaluate}{Logical. If \code{TRUE} (default) the updated call is evaluated
and the re-fitted \link{INLAvaan} object is returned; if \code{FALSE} the updated,
unevaluated call is returned for inspection.}
}
\value{
A re-fitted \link{INLAvaan} object, or the updated call when
\code{evaluate = FALSE}.
}
\description{
Re-fit an \link{INLAvaan} model with modified arguments, in the style of
\code{\link[stats:update]{stats::update()}}. This is convenient for prior-sensitivity analyses
(vary \code{dp} or the \code{prior()} modifiers), iterative respecification (swap the
\code{model} or \code{add} syntax), and swapping approximation settings (e.g.
\code{marginal_method}, \code{nsamp}) without retyping the whole call.
}
\details{
The method edits the recorded \code{\link[=inlavaan]{inlavaan()}} call, replacing \code{model} (or
extending it via \code{add}) and overriding any argument supplied through \code{...},
then re-evaluates it. Because \link{INLAvaan} extends the \pkg{lavaan} class, a
dedicated method is required: without it, \code{update()} would dispatch to
\pkg{lavaan}'s method and silently return a frequentist \pkg{lavaan} fit
instead of a Bayesian one.

When the parameter structure is preserved (no change to \code{model}, \code{add}, or
\code{model.type}), the previous posterior mode is passed as the optimiser's
starting value (a warm start), so mode-finding is typically much faster than
a cold fit. This makes prior sweeps particularly cheap. The warm start only
changes the optimiser's starting point, not the fitted result. Supply
\code{start} explicitly to override it.

Fits produced by INLAvaan versions that did not record their call cannot be
updated; re-fit once with \code{\link[=inlavaan]{inlavaan()}} (or \code{\link[=acfa]{acfa()}}, \code{\link[=asem]{asem()}}, \code{\link[=agrowth]{agrowth()}})
and \code{update()} will work on the new object.
}
\examples{
\dontrun{
model <- "visual =~ x1 + x2 + x3"
fit <- acfa(model, data = lavaan::HolzingerSwineford1939)

# Prior sensitivity: refit under a tighter loading prior (warm-started)
fit2 <- update(fit, dp = priors_for(lambda = "normal(0,1)"))

# Respecify: add a residual covariance
fit3 <- update(fit, add = "x1 ~~ x2")
compare(fit, fit3)

# Swap approximation settings only
fit4 <- update(fit, marginal_method = "sampling", nsamp = 4000)
}
}
\seealso{
\code{\link[=inlavaan]{inlavaan()}}, \code{\link[=compare]{compare()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qsnorm_fast.R
\name{qsnorm_fast}
\alias{qsnorm_fast}
\title{Fast Approximation of Skew-Normal Quantile Function}
\usage{
qsnorm_fast(p, xi = 0, omega = 1, alpha = 0)
}
\arguments{
\item{p}{Vector of probabilities.}

\item{xi}{Location parameter (numeric vector).}

\item{omega}{Scale parameter (numeric vector).}

\item{alpha}{Shape parameter (numeric vector).}
}
\value{
Vector of quantiles.
}
\description{
A fast approximation of skew-normal quantiles using the high-performance
approximation algorithm from the INLA GMRFLib C source, and originally by
Thomas Luu (see details for reference).
}
\details{
This function implements a high-performance approximation for the
skew-normal quantile function based on the algorithm described by Luu
(2016). The method uses a domain decomposition strategy to achieve high
accuracy (\eqn{< 10^{-7}} relative error) without iterative numerical
inversion.

The domain is split into two regions:
\itemize{
\item \strong{Tail Regions:} For extreme probabilities where \eqn{\vert u \vert} is large, the quantile
is approximated using the Lambert W-function, \eqn{W(z)}, solving \eqn{z = \Phi(q)} via
asymptotic expansion:
\deqn{q \approx \sqrt{2 W\left(\frac{1}{2\pi (1-p)^2}\right)}}
\item \strong{Central Region:} For the main body of the distribution, the function
uses a high-order Taylor expansion of the inverse error function around a
carefully selected expansion point $x_0$:
\deqn{\Phi^{-1}(p) \approx \sum_{k=0}^5 c_k (z - x_0)^k}
}

This approach is significantly faster than standard numerical inversion
(e.g., \code{uniroot}) while maintaining sufficient precision for most
statistical applications.
}
\examples{
qsnorm_fast(c(0.025, 0.5, 0.975))
qsnorm_fast(c(0.025, 0.5, 0.975), xi = 2, omega = 0.5, alpha = 1)

}
\references{
Luu, T. (2016). \emph{Fast and accurate parallel computation of
quantile functions for random number generation} (Doctoral thesis). UCL
(University College London). \url{https://discovery.ucl.ac.uk/1482128/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-loo.R
\name{loo}
\alias{loo}
\alias{loo.INLAvaan}
\alias{loo.inlavaan_internal}
\alias{add_loo}
\title{Approximate Leave-One-Out Cross-Validation for INLAvaan Models}
\usage{
loo(x, ...)

\method{loo}{INLAvaan}(
  x,
  type = c("auto", "loso", "loco"),
  units = NULL,
  second_order = TRUE,
  theta = NULL,
  Sigma = NULL,
  cores = NULL,
  verbose = FALSE,
  ...
)

\method{loo}{inlavaan_internal}(
  x,
  type = c("auto", "loso", "loco"),
  units = NULL,
  second_order = TRUE,
  theta = NULL,
  Sigma = NULL,
  cores = NULL,
  verbose = FALSE,
  ...
)

add_loo(object, cores = NULL, verbose = FALSE)
}
\arguments{
\item{x}{A fitted \link{INLAvaan} object (or its \code{inlavaan_internal} list).}

\item{...}{Not used.}

\item{type}{Unit type: \code{"auto"} (default) resolves to \code{"loso"}
(per-subject) for single-level models and \code{"loco"} (per-cluster,
marginal predictive) for two-level models. \code{"loco"} cannot be forced
on a model without clusters; \code{"loso"} on a two-level model scores the
conditional (leave-one-unit-out) predictive instead (with a warning;
see Details).}

\item{units}{Optional integer vector of unit indices to score; defaults
to all units. For LOSO these are case numbers (row numbers of the
analysed dataset, as recorded in the fit -- for multigroup fits the
full results are stacked by group, but a unit is always addressed by
its case number); for LOCO, cluster positions.}

\item{second_order}{Logical; compute the second-order correction
(default \code{TRUE}). \code{FALSE} skips the Hessian stage entirely and reports
first-order estimates.}

\item{theta, Sigma}{Optional posterior mean vector and covariance matrix
(in the unconstrained parameter space, as stored in \code{theta_star} and
\code{Sigma_theta}) at which to evaluate the LOO instead of the fit's own
Laplace summary. See Details.}

\item{cores}{Number of cores for the Hessian stage. The default \code{NULL}
runs serially; parallelism must be requested explicitly.}

\item{verbose}{Logical; print progress (default \code{FALSE}).}

\item{object}{A fitted \link{INLAvaan} object.}
}
\value{
An object of class \code{inlavaan_loo}: a list with elements
\describe{
\item{\code{per_unit}}{Data frame of pointwise results: \code{unit} (case
number for LOSO, cluster position for LOCO), \code{group} (multigroup
fits only), \code{nobs} (1 for LOSO, the cluster size for LOCO),
\code{l_star} (unit log-likelihood at the summary), \code{score_norm},
\code{lpd_1}/\code{lpd_2} (pointwise log predictive density),
\code{log_cpo_1}/\code{log_cpo_2} (pointwise LOO contributions), \code{det_term},
and \code{ok} (second-order success flag).}
\item{\code{estimates}}{Matrix with rows \code{elpd_loo}, \code{p_loo}, \code{looic} and
columns \code{Estimate}, \code{SE} (headline second-order values).}
\item{\code{elpd_1}, \code{elpd_2}, \code{se_1}, \code{se_2}, \code{p_loo_1}, \code{p_loo_2}}{
First- and second-order aggregates.}
\item{\code{type}, \code{flavour}, \code{n_units}, \code{n_groups}, \code{n_ok},
\code{second_order}, \code{theta_overridden}}{Metadata; \code{flavour} records
whether units were scored jointly with their covariates
(\code{"joint"}) or conditionally on them (\code{"conditional"}, for
\code{fixed.x} fits).}
}

\code{add_loo()} returns a copy of \code{object} with the LOO result
stored alongside the fit (the input object is unchanged); reassign it,
e.g. \code{fit <- add_loo(fit)}. Only the default LOO is stored, so the
stored result always matches \code{loo(fit)}.
}
\description{
Computes leave-one-out (LOO) cross-validation for a fitted \link{INLAvaan}
model from a single fit, with no refitting and no sampling, via a Taylor
approximation of the case-deletion posterior around the Laplace summary.
Single-level models are scored per subject (leave-one-subject-out, LOSO);
two-level models are scored per cluster (leave-one-cluster-out, LOCO).
}
\details{
For a unit \eqn{u} (a subject for LOSO, a cluster for LOCO) with
log-likelihood contribution \eqn{\ell_u(\theta)}, score \eqn{s_u} and
Hessian \eqn{H_u} evaluated at the posterior summary
\eqn{(\theta^*, \Sigma)}, the log conditional predictive ordinate is
approximated to first and second order by
\deqn{\log \mathrm{CPO}_u^{(1)} = \ell_u - \tfrac{1}{2} s_u' \Sigma s_u,}
\deqn{\log \mathrm{CPO}_u^{(2)} = \ell_u
  - \tfrac{1}{2} s_u' (\Sigma^{-1} + H_u)^{-1} s_u
  + \tfrac{1}{2} \log |I + \Sigma H_u|.}
The reported \code{elpd_loo} is the sum of the second-order terms (first-order
when \code{second_order = FALSE}), with standard error
\eqn{\sqrt{n \, \mathrm{var}(\log \mathrm{CPO}_u)}} and
\code{looic} \eqn{= -2 \, \mathrm{elpd}}. The effective number of parameters is
\eqn{p_{\mathrm{loo}} = \sum_u (\mathrm{lpd}_u - \log \mathrm{CPO}_u)},
where \eqn{\mathrm{lpd}_u} is the analogous Taylor approximation of the
full-posterior pointwise log predictive density. Units whose second-order
curvature matrix is not positive definite fall back to first order for
that unit only (flagged in \code{per_unit$ok}).

The type is resolved automatically: per-cluster (\code{"loco"}) when the model
was fitted with a \code{cluster} argument, per-subject (\code{"loso"}) otherwise.
For a two-level model these are the two estimands of Merkle, Furr &
Rabe-Hesketh (2019): the default per-cluster \code{"loco"} is the \emph{marginal}
predictive (leave-one-cluster-out -- prediction for a \emph{new} cluster),
while \code{type = "loso"} forces the \emph{conditional} predictive
(leave-one-unit-out -- prediction for a new observation within an
\emph{observed} cluster), where row \eqn{i} of cluster \eqn{j} contributes
\eqn{\ell_i = \ell_j(\mathrm{full}) - \ell_j(\mathrm{minus\ row\ } i)},
the conditional density of the row given the rest of its cluster. The two
answer different questions and are easily conflated, so the per-cluster
marginal is the default and \code{type = "loso"} warns. It works with and
without missing data, costs one cluster evaluation per row per Hessian
direction, and is best subset with \code{units}.

\strong{Multigroup models.} Groups are independent, so each unit is scored
against its own group's implied moments; without a mean structure the
exchangeability transformation applies per group, and cross-group
equality constraints (\code{group.equal}) flow through the packed parameter
space automatically. The per-unit results are stacked by group (a
\code{group} column records the membership), and units are identified by
\emph{case number} -- the row number of the analysed dataset -- so a unit
keeps its identity across fits that assign or order groups differently
(e.g. a pooled fit versus a grouped fit of the same data, which
\code{\link[=compare]{compare()}} pairs unit by unit). This makes
\code{compare(..., loo = TRUE)} the instrument of choice for the
measurement-invariance ladder: configural, metric, and scalar fits are
compared on a proper predictive scale with paired standard errors.

Supplying \code{theta} and/or \code{Sigma} scores the model at an \emph{arbitrary}
Gaussian posterior summary instead of the fit's own, without refitting.
This is the building block for refit-free model exploration: for example,
conditioning the encompassing model's summary on a parameter being zero
(a rank-one update of \code{theta} and \code{Sigma}) and scoring the result gives
the LOO of that submodel from a single fit. INLAvaan provides only this
evaluation API; search strategies are left to the user. A conditioned
\code{Sigma} may be singular; the computation automatically restricts to the
non-degenerate block, which is exact.

Parallelism is strictly opt-in: the default \code{cores = NULL} runs serially,
and \code{cores > 1} parallelises the Hessian stage via forking (not available
on Windows).

Calling \code{loo()} never modifies the fitted object. Under the default
\code{test = "standard"}, \code{\link[=inlavaan]{inlavaan()}} already computes and stores the full
LOO at fit time whenever the model is supported, has a mean structure,
and the predicted serial cost is within a 10-second budget (measured by
timing one score evaluation); \code{test = "loo"} forces the computation
regardless of the budget, and \code{fit <- add_loo(fit)} stores it post hoc.
A stored result is returned directly by \code{loo(fit)} when called with
default arguments, and is reused by \code{\link[=fitmeasures]{fitmeasures()}} and \code{\link[=compare]{compare()}}
without recomputation.

\strong{Exogenous covariates.} The flavour of the score follows the fitted
likelihood. Under \code{fixed.x = FALSE} the covariates are modelled jointly
and each unit is scored by the joint predictive density of its outcomes
\emph{and} covariates (\code{flavour = "joint"}). Under \code{fixed.x = TRUE} (the
lavaan default) the fitted likelihood is the conditional one, and each
unit is scored by the predictive density of its outcomes \emph{given} its
covariates (\code{flavour = "conditional"}); since the conditional likelihood
is exactly invariant to the frozen covariate moments, this involves no
additional approximation. The two flavours estimate different quantities
whose scales differ by the covariate predictive density, so a joint and a
conditional elpd must never be compared (\code{\link[=compare]{compare()}} refuses
mixed-flavour comparisons). Conditional scores of models conditioning on
\emph{different} covariate sets are comparable provided the outcome variables
match -- the natural setting for covariate selection. Both flavours
support any covariate placement: single-level covariates, and
cluster-level (between) and/or within-level covariates in two-level
models.

\strong{Missing data.} Fits estimated by full-information maximum likelihood
(\code{missing = "ml"}) are scored on the \emph{observed-data} predictive: each
unit contributes the density of the entries it actually has, with its
full row (single-level) or whole cluster (two-level) removed from the
conditioning set. For single-level fits the casewise kernels operate on
each unit's observed subset, grouping rows by missing pattern, so a unit
with fewer observed entries contributes a smaller log-likelihood term
\emph{and} a smaller score and thus self-weights in the elpd. Two-level fits
are scored per cluster (\code{"loco"}): each cluster contributes its
observed-data marginal likelihood, evaluated by lavaan's raw-data
cluster kernels (no per-cluster sufficient statistics are needed, since
leave-one-cluster-out deletes the whole cluster). All carry the same
missing-at-random assumption as the FIML fit itself. Because the score
is the observed-entry predictive, a \code{\link[=compare]{compare()}} of two missing-data fits
is meaningful only when they share the same observed entries (the same
data \emph{and} the same holes). The two-level conditional predictive
(\code{type = "loso"}) is available under missing data too, on the same
kernels.

Supported models: continuous-indicator models fitted with the \code{ML}
estimator (including FIML, \code{missing = "ml"}, single- and two-level),
single-group or multigroup (multigroup two-level models are not
supported yet). If the \code{loo} package is attached it masks this generic,
but \code{loo(fit)} continues to dispatch correctly because the method is
registered by generic name.
}
\examples{
\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, meanstructure = TRUE)

# Leave-one-subject-out (LOSO) from the single fit -- no refitting
res <- loo(fit)
res
head(res$per_unit)

# Score a submodel without refitting: condition the Laplace summary on the
# visual ~~ speed covariance being zero, then evaluate at that summary
int <- get_inlavaan_internal(fit)
theta <- int$theta_star
Sigma <- int$Sigma_theta
p <- which(names(coef(fit)) == "visual~~speed")
theta_c <- theta - Sigma[, p] * (theta[p] / Sigma[p, p])
Sigma_c <- Sigma - tcrossprod(Sigma[, p]) / Sigma[p, p]
loo(fit, theta = theta_c, Sigma = Sigma_c)

# Two-level models are scored per cluster (LOCO) automatically
utils::data("Demo.twolevel", package = "lavaan")
model2l <- "
  level: 1
    fw =~ y1 + y2 + y3
    fw ~ x1 + x2 + x3
  level: 2
    fb =~ y1 + y2 + y3
    fb ~ w1 + w2
"
fit2l <- asem(model2l, Demo.twolevel, cluster = "cluster",
              meanstructure = TRUE, fixed.x = FALSE)
loo(fit2l)
}
}
\seealso{
\code{\link[=fitmeasures]{fitmeasures()}}, \code{\link[=compare]{compare()}}, \code{\link[=inlavaan]{inlavaan()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-fitted.R
\name{fitted}
\alias{fitted}
\alias{fitted,INLAvaan-method}
\alias{fitted.values,INLAvaan-method}
\title{Model-Implied Moments for INLAvaan Models}
\usage{
\S4method{fitted}{INLAvaan}(object, type = "moments", labels = TRUE, ...)

\S4method{fitted.values}{INLAvaan}(object, type = "moments", labels = TRUE, ...)
}
\arguments{
\item{object}{An object of class \link{INLAvaan}.}

\item{type}{Character. \code{"moments"} (default) returns the model-implied
variance-covariance matrix and, when relevant, the mean vector (plus
thresholds for ordinal data). \code{"casewise"} (aliases \code{"obs"},
\code{"ov"}) returns the model-predicted values for each observation.}

\item{labels}{Logical. Attach variable names to the output. Default
\code{TRUE}.}

\item{...}{Currently unused.}
}
\value{
For \code{type = "moments"}, a list (or list of lists, for
multiple groups) with elements such as \code{cov}, \code{mean}, and
\code{th}. For \code{type = "casewise"}, a numeric matrix of predicted
observed-variable values.
}
\description{
Extract the model-implied (fitted) sample statistics from a fitted
\code{INLAvaan} model. As in \pkg{lavaan} and \pkg{blavaan}, the moments are
the model-implied covariance matrix (and mean vector, when a mean structure
is present) evaluated at the parameter estimates -- here the posterior means.
}
\details{
This delegates to \pkg{lavaan}'s own \code{fitted()} machinery, so the return
structure matches lavaan exactly. Because INLAvaan stores the posterior means
as the point estimates of the fitted object, the implied moments are the
posterior-mean model-implied moments (mirroring \pkg{blavaan}).
}
\examples{
\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
            test = "none", verbose = FALSE)

# Model-implied covariance matrix (posterior means)
fitted(fit)

# Casewise model-predicted observed values
head(fitted(fit, type = "ov"))
}

}
\seealso{
\code{\link[=predict]{predict()}}, \code{\link[=coef]{coef()}}, \link[lavaan:fitMeasures]{fitMeasures()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-simulate.R
\name{simulate}
\alias{simulate}
\alias{simulate,INLAvaan-method}
\title{Simulate Datasets from the Generative Model}
\usage{
\S4method{simulate}{INLAvaan}(
  object,
  nsim = 1L,
  seed = NULL,
  sample.nobs = NULL,
  prior = FALSE,
  samp_copula = TRUE,
  silent = FALSE,
  ...
)
}
\arguments{
\item{object}{An object of class \link{INLAvaan}.}

\item{nsim}{Number of replicate datasets to generate (default 1).}

\item{seed}{Optional random seed (passed to \code{\link[=set.seed]{set.seed()}}).}

\item{sample.nobs}{Number of observations per dataset. Defaults to the
sample size of the original data.}

\item{prior}{Logical. When \code{TRUE}, parameters are drawn from the prior;
when \code{FALSE} (default), from the posterior.}

\item{samp_copula}{Logical. When \code{TRUE} (default) and \code{prior = FALSE},
posterior parameter draws use the copula method. Ignored when
\code{prior = TRUE}.}

\item{silent}{Logical. When \code{TRUE}, suppresses the informational message
about rejected non-PD draws. Default \code{FALSE}.}

\item{...}{Additional arguments (currently unused).}
}
\value{
A list of length \code{nsim}. Each element is a data frame with
\code{sample.nobs} rows and two attributes:
\itemize{
\item \code{"truth"} — named numeric vector of lavaan-side (x-space, constrained)
parameter values used to generate the dataset.
\item \code{"truth_theta"} — named numeric vector of the corresponding unconstrained
(theta-space) parameter values.
}
}
\description{
Generate complete synthetic datasets from a fitted INLAvaan model. For each
simulation, a single parameter vector is drawn (from the posterior or prior),
and then \code{sample.nobs} observations are generated from the model-implied
distribution at that parameter value.
}
\details{
This function is designed for tasks that require \strong{full replicate datasets}
from a single parameter draw, such as simulation-based calibration (SBC) and
posterior predictive p-values. It differs from \code{\link[=sampling]{sampling()}} which generates
one observation per parameter draw (useful for prior/posterior predictive
density overlays).

For each simulation \eqn{s = 1, \ldots, S}:
\enumerate{
\item Draw \eqn{\boldsymbol\theta^{(s)}} from the posterior (or prior).
\item Compute the model-implied covariance
\eqn{\boldsymbol\Sigma(\boldsymbol\theta^{(s)})}.
If it is not positive-definite, reject and redraw.
\item Generate a dataset of \code{sample.nobs} rows from
\eqn{N(\boldsymbol\mu(\boldsymbol\theta^{(s)}),\,
   \boldsymbol\Sigma(\boldsymbol\theta^{(s)}))}.
}

Parameter draws reuse the same internal machinery as \code{\link[=sampling]{sampling()}}
(\code{sample_params_prior} / \code{sample_params_posterior}), so the prior
specification is consistent.
}
\examples{
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa("visual =~ x1 + x2 + x3", HolzingerSwineford1939)

# Simulate one replicate dataset from the posterior
sims <- simulate(fit, nsim = 1)
head(sims[[1]])                    # data frame
attr(sims[[1]], "truth")           # true lavaan-side (x-space) parameters
attr(sims[[1]], "truth_theta")     # corresponding unconstrained (theta-space) parameters

# Simulate from the prior (e.g., for SBC)
sims_prior <- simulate(fit, nsim = 5, prior = TRUE)
lapply(sims_prior, nrow)
}
\seealso{
\code{\link[=sampling]{sampling()}} for single-observation draws from the predictive
distribution (prior/posterior predictive checks).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-deviance.R
\name{deviance}
\alias{deviance}
\alias{deviance.INLAvaan}
\title{Deviance for INLAvaan Models}
\usage{
\method{deviance}{INLAvaan}(object, type = c("mean", "plugin"), ...)
}
\arguments{
\item{object}{An object of class \link{INLAvaan}.}

\item{type}{Character. \code{"mean"} (default) returns the posterior mean
deviance \eqn{\bar{D} = E[-2\log p(y \mid \theta)]}, averaged over
posterior draws. \code{"plugin"} returns the deviance evaluated at the
posterior mean point estimate, \eqn{\hat{D} = -2\log p(y \mid \hat\theta)}
(matching \code{-2 * logLik(object, type = "plugin")}). Both require the
model to have been fitted with \code{test != "none"}.}

\item{...}{Currently unused.}
}
\value{
A length-one numeric of class \code{inlavaan_deviance}, with the
effective number of parameters (\code{pD}) and \code{DIC} attached as
attributes.
}
\description{
Extract the (Bayesian) deviance of a fitted \code{INLAvaan} model. Unlike
\pkg{lavaan}, which has no \code{deviance()} method, this follows the
BUGS/JAGS/Stan convention: "deviance" is \eqn{-2} times the log-likelihood,
summarised over the posterior.
}
\details{
\eqn{\bar{D}} and \eqn{\hat{D}} are the two ingredients of the Deviance
Information Criterion, \eqn{DIC = \bar{D} + p_D} where
\eqn{p_D = \bar{D} - \hat{D}} is the effective number of parameters. Use
\code{\link[=compare]{compare()}} to compare models by DIC (or Bayes factors, or LOO/WAIC)
rather than comparing raw deviances directly.
}
\examples{
\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
            test = "standard", verbose = FALSE)

deviance(fit)
attr(deviance(fit), "DIC")
}

}
\seealso{
\code{\link[=logLik]{logLik()}}, \code{\link[=compare]{compare()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-fitmeasures.R
\name{bfit_indices}
\alias{bfit_indices}
\alias{summary.bfit_indices}
\alias{print.bfit_indices}
\title{Bayesian Fit Indices}
\usage{
bfit_indices(
  object,
  baseline.model = NULL,
  rescale = c("devM", "MCMC"),
  nsamp = NULL,
  samp_copula = TRUE
)

\method{summary}{bfit_indices}(object, ...)

\method{print}{bfit_indices}(x, ...)
}
\arguments{
\item{object}{An object of class \link{INLAvaan}.}

\item{baseline.model}{An optional \link{INLAvaan} object representing the
baseline (null) model. Required for incremental fit indices (BCFI, BTLI,
BNFI).}

\item{rescale}{Character string controlling how the Bayesian chi-square
is rescaled. \code{"devM"} (default) subtracts pD from the deviance at each
sample. \code{"MCMC"} uses the classical chi-square and classical df at each
sample.}

\item{nsamp}{Number of posterior samples to draw. Defaults to the value
used when fitting the model.}

\item{samp_copula}{Logical. When \code{TRUE} (default), posterior samples are
drawn using the copula method with the fitted marginals. When \code{FALSE},
samples are drawn from the Gaussian (Laplace) approximation.}

\item{...}{Additional arguments passed to methods.}

\item{x}{An object of class \code{bfit_indices} (for \code{print}).}
}
\value{
An S3 object of class \code{"bfit_indices"} containing:
\describe{
\item{\code{indices}}{Named list of numeric vectors (one per posterior sample)
for each computed fit index.}
\item{\code{details}}{List with \code{chisq} (per-sample deviance), \code{df}, \code{pD},
\code{rescale}, and \code{nsamp}.}
}
Use \code{\link[=summary]{summary()}} to obtain a table of posterior summaries (Mean, SD,
quantiles, Mode) for each index.
}
\description{
Compute posterior distributions of Bayesian fit indices for an INLAvaan
model, analogous to \code{\link[blavaan:blavFitIndices]{blavaan::blavFitIndices()}}.
}
\examples{
\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
            verbose = FALSE)

# Absolute fit indices
bf <- bfit_indices(fit)
bf
summary(bf)
}

}
\seealso{
\code{\link[lavaan:fitMeasures]{lavaan::fitMeasures()}}, \code{\link[blavaan:blavFitIndices]{blavaan::blavFitIndices()}},
\code{\link[=fitmeasures]{fitmeasures()}}, \code{\link[=compare]{compare()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R, R/method-anova.R
\name{compare}
\alias{compare}
\alias{compare,INLAvaan-method}
\alias{anova}
\alias{anova,INLAvaan-method}
\title{Compare Bayesian Models Fitted with INLAvaan}
\usage{
compare(x, y, ..., fit.measures = NULL, loo = FALSE)

\S4method{compare}{INLAvaan}(x, y, ..., fit.measures = NULL, loo = FALSE)

\S4method{anova}{INLAvaan}(object, ...)
}
\arguments{
\item{x}{An \link{INLAvaan} (or \code{inlavaan_internal}) object used as the
\strong{baseline} (null) model. It is included in the comparison table and
passed to \link[lavaan:fitMeasures]{fitMeasures()} for incremental indices.}

\item{y, ...}{One or more \link{INLAvaan} (or \code{inlavaan_internal}) objects to
compare against the baseline.}

\item{fit.measures}{Character vector of additional fit-measure names to
include (e.g. \code{"BRMSEA"}, \code{"BCFI"}). Use \code{"all"} to include every
measure returned by \link[lavaan:fitMeasures]{fitMeasures()}. The default
(\code{NULL}) shows only the core comparison statistics.}

\item{loo}{Logical; if \code{TRUE}, compare models by leave-one-out
cross-validation with paired standard errors (see Details). Defaults to
\code{FALSE}.}

\item{object}{An \link{INLAvaan} object (the \code{anova()} method, which is disabled
and redirects to \code{compare()}).}
}
\value{
A data frame of class \code{compare.inlavaan_internal} containing model
fit statistics, sorted by descending marginal log-likelihood (or by
descending ELPD when \code{loo = TRUE}).
}
\description{
Compare two or more Bayesian SEM fitted with INLAvaan, reporting model-fit
statistics and (optionally) fit indices side by side.
}
\details{
The first argument \code{x} serves as the \strong{baseline} (null) model.
All models (including the baseline) appear in the comparison table. The
baseline is also passed to \link[lavaan:fitMeasures]{fitMeasures()} when
incremental fit indices (BCFI, BTLI, BNFI) are requested via
\code{fit.measures}.

The default table always includes:
\itemize{
\item \strong{npar}: Number of free parameters.
\item \strong{Marg.Loglik}: Approximated marginal log-likelihood.
\item \strong{logBF}: Natural-log Bayes factor relative to the best model.
\item \strong{DIC} / \strong{pD}: Deviance Information Criterion and effective number
of parameters (when \code{test != "none"} was used during fitting).
}

Set \code{fit.measures} to a character vector of measure names (anything
returned by \link[lavaan:fitMeasures]{fitMeasures()}) to append extra columns.
Use \code{fit.measures = "all"} to include every available measure.

Set \code{loo = TRUE} to compare models by leave-one-out cross-validation
(see \code{\link[=loo]{loo()}}). This appends \strong{ELPD} / \strong{SE} (the second-order Taylor
expected log predictive density and its standard error), \strong{p_loo}, and,
against the best-ELPD model, the difference \strong{elpd_diff} with its
\emph{paired} standard error \strong{se_diff} computed from the pointwise
contributions (the appropriate uncertainty for nested or same-data
comparisons). The table is then sorted by descending ELPD. All models
must be fitted to the same data with matching units; units are paired
by id rather than by row order, so fits that stack groups differently
-- a pooled fit against a multigroup fit, or multigroup fits with
different group orderings -- still pair up unit by unit. For
missing-data (FIML) fits, "the same data" also means the same observed
entries: each unit is scored on the entries it has, so comparisons
require identical missingness patterns across models. All models
must also share the
score flavour (see \code{\link[=loo]{loo()}}): mixing fits with modelled covariates
(\code{fixed.x = FALSE}, joint scores) and fixed covariates
(\code{fixed.x = TRUE}, conditional scores) is refused. Joint scores
additionally require identical variable sets across models, while
conditional scores require only matching outcome variables -- covariate
sets may differ, which is the covariate-selection setting. Stored LOO
results (\code{test = "loo"} or \code{\link[=add_loo]{add_loo()}}) are reused.

\code{anova()} is disabled for \code{INLAvaan} fits -- there is no direct Bayesian
analogue of the classical likelihood-ratio test -- and points here instead.
}
\examples{
\donttest{
# Model comparison on multigroup analysis (measurement invariance)
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")

# Configural invariance
fit1 <- acfa(HS.model, data = HolzingerSwineford1939, group = "school")

# Weak invariance
fit2 <- acfa(
  HS.model,
  data = HolzingerSwineford1939,
  group = "school",
  group.equal = "loadings"
)

# Strong invariance
fit3 <- acfa(
  HS.model,
  data = HolzingerSwineford1939,
  group = "school",
  group.equal = c("intercepts", "loadings")
)

# Compare models (fit1 = configural = baseline, always first argument)
compare(fit1, fit2, fit3)

# With extra fit measures
compare(fit1, fit2, fit.measures = c("BRMSEA", "BMc"))

# With incremental indices (baseline = fit1, passed to fitMeasures())
compare(fit1, fit2, fit3, fit.measures = c("BCFI", "BTLI"))
}
}
\references{
\url{https://lavaan.ugent.be/tutorial/groups.html}
}
\seealso{
\code{\link[=fitmeasures]{fitmeasures()}}, \code{\link[=bfit_indices]{bfit_indices()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{is_same_function}
\alias{is_same_function}
\title{Helper function to check if two functions are the same}
\usage{
is_same_function(f, g)
}
\arguments{
\item{f, g}{Functions to compare.}
}
\value{
Logical.
}
\description{
Helper function to check if two functions are the same
}
\examples{
f1 <- function(x) { x^2 + 1 }
f2 <- function(x) { x^2 + 1 }
is_same_function(f1, f2)  # TRUE
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-diagnostics.R
\name{diagnostics}
\alias{diagnostics}
\alias{diagnostics,INLAvaan-method}
\title{Convergence and Approximation Diagnostics for INLAvaan Models}
\usage{
diagnostics(object, ...)

\S4method{diagnostics}{INLAvaan}(object, type = c("global", "param"), ...)
}
\arguments{
\item{object}{An object of class \link{INLAvaan}.}

\item{...}{Currently unused.}

\item{type}{Character. \code{"global"} (default) returns a named numeric
vector of scalar diagnostics. \code{"param"} returns a data frame with one
row per free parameter containing per-parameter diagnostics.}
}
\value{
For \code{type = "global"}, a named numeric vector (class
\code{"diagnostics.INLAvaan"}). For \code{type = "param"}, a data frame
(class \code{c("diagnostics.INLAvaan.param", "data.frame")}).
}
\description{
Extract convergence and approximation-quality diagnostics from a fitted
\code{INLAvaan} model.
}
\details{
\strong{Global diagnostics} (\code{type = "global"}):
\describe{
\item{\code{npar}}{Number of free parameters.}
\item{\code{nsamp}}{Number of posterior samples drawn.}
\item{\code{converged}}{1 if the optimiser converged, 0 otherwise.}
\item{\code{iterations}}{Number of optimiser iterations.}
\item{\code{grad_inf}}{L-infinity norm of the analytic gradient at the mode
(max |grad|). Should be ~0 at convergence.}
\item{\code{grad_inf_rel}}{Relative L-infinity norm of the analytic gradient
(max |grad| / (|par| + 1e-6)).}
\item{\code{grad_l2}}{L2 (Euclidean) norm of the analytic gradient at the mode.}
\item{\code{mode_shift_max}}{Maximum, across parameters, of the Newton
step at the reported mode in posterior-SD units
(max |\eqn{\Sigma_\theta} grad| / se). Unlike the raw gradient norms
this is scale-free: it estimates how far the reported mode sits from
the true posterior mode relative to the posterior uncertainty. Should
be ~0 at convergence.}
\item{\code{hess_cond}}{Condition number of the Hessian (precision matrix)
computed from \eqn{\Sigma_\theta}. Large values indicate near-singularity.}
\item{\code{vb_kld_global}}{Global KL divergence from the VB mean correction
(NA if VB correction was not applied).}
\item{\code{vb_applied}}{1 if VB correction was applied, 0 otherwise.}
\item{\code{kld_max}}{Maximum per-parameter KL divergence from the VB correction.}
\item{\code{kld_mean}}{Mean per-parameter KL divergence.}
\item{\code{nmad_max}}{Maximum normalised max-absolute-deviation across
marginals (skew-normal method only; NA otherwise).}
\item{\code{nmad_mean}}{Mean NMAD across marginals.}
}

\strong{Per-parameter diagnostics} (\code{type = "param"}):
A data frame with columns:
\describe{
\item{\code{param}}{Parameter name.}
\item{\code{grad}}{Analytic gradient of the negative log-posterior at the
mode. Should be ~0 at convergence.}
\item{\code{grad_num}}{Numerical (finite-difference) gradient at the mode.
Should agree with \code{grad}; large discrepancies indicate a bug in the
analytic gradient.}
\item{\code{grad_diff}}{Difference \code{grad_num - grad}: should be ~0.}
\item{\code{grad_abs}}{Absolute analytic gradient.}
\item{\code{grad_rel}}{Relative analytic gradient |grad| / (|par| + 1e-6).}
\item{\code{mode_shift_sigma}}{Newton step at the reported mode in
posterior-SD units. Should be ~0 at convergence.}
\item{\code{kld}}{Per-parameter KL divergence from the VB correction.}
\item{\code{vb_shift}}{VB correction shift (in original scale).}
\item{\code{vb_shift_sigma}}{VB shift in units of posterior SD.}
\item{\code{nmad}}{Normalised max-absolute-deviation of the skew-normal fit
(NA when not using the skewnorm method).}
}

\strong{Fit-time warnings}: \code{\link[=inlavaan]{inlavaan()}} runs these checks once at the end
of every fit and emits a single consolidated warning (condition class
\code{"inlavaan_diagnostics_warning"}) when any of them look off: the
optimiser did not converge, \code{mode_shift_max} exceeds 0.1,
any marginal has NMAD above 0.1, the VB correction shifted a posterior
mean by more than 1 posterior SD, or the Hessian condition number exceeds
1e8. A healthy fit stays silent. Silence the check with
\code{suppressWarnings()}, or selectively by handling the condition class.
}
\examples{
\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
            test = "none", verbose = FALSE)

# Global convergence summary
diagnostics(fit)

# Per-parameter table
diagnostics(fit, type = "param")
}

}
\seealso{
\code{\link[=timing]{timing()}}, \code{\link[=fitmeasures]{fitmeasures()}}, \code{\link[=plot]{plot()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-timing.R
\name{timing}
\alias{timing}
\alias{timing,INLAvaan-method}
\title{Timing Information for INLAvaan Models}
\usage{
timing(object, ...)

\S4method{timing}{INLAvaan}(object, what = "total", ...)
}
\arguments{
\item{object}{An object of class \link{INLAvaan}.}

\item{...}{Currently unused.}

\item{what}{Character vector of timing segment names to return, or
\code{"all"} to return every segment. Defaults to \code{"total"}.
Available segments (depending on model options): \code{"init"},
\code{"optim"}, \code{"vb"}, \code{"loglik"}, \code{"marginals"},
\code{"norta"}, \code{"sampling"}, \code{"covariances"},
\code{"definedpars"}, \code{"deltapars"}, \code{"test"}, \code{"total"}.}
}
\value{
A named numeric vector (class \code{c("timing.INLAvaan",
  "numeric")}) of elapsed times in seconds. Printing formats short
durations as seconds, longer ones as minutes or hours.
}
\description{
Extract wall-clock timings for individual computation stages of a fitted
\code{INLAvaan} model.
}
\examples{
\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
            test = "none", verbose = FALSE)

# Total elapsed time
timing(fit)

# All stages
timing(fit, what = "all")

# Specific stages
timing(fit, what = c("optim", "marginals"))
}

}
\seealso{
\code{\link[=diagnostics]{diagnostics()}}, \code{\link[=summary]{summary()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-vcov.R
\name{vcov}
\alias{vcov}
\alias{vcov,INLAvaan-method}
\title{Variance-Covariance Matrix for INLAvaan Models}
\usage{
\S4method{vcov}{INLAvaan}(object, type = c("lavaan", "theta"), ...)
}
\arguments{
\item{object}{An object of class \link{INLAvaan}.}

\item{type}{Character. \code{"lavaan"} (default) returns the posterior
covariance matrix of the model parameters computed from posterior
samples (matching lavaan output). \code{"theta"} returns the Laplace
approximation covariance in the internal parameterisation.}

\item{...}{Currently unused.}
}
\value{
A square numeric matrix.
}
\description{
Extract the posterior variance-covariance matrix of model parameters from a
fitted \code{INLAvaan} model.
}
\examples{
\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
            test = "none", verbose = FALSE)

# Default: posterior covariance of lavaan parameters
vcov(fit)

# Internal parameterisation (Laplace approximation)
vcov(fit, type = "theta")
}

}
\seealso{
\code{\link[=summary]{summary()}}, \code{\link[=coef]{coef()}}, \code{\link[=standardisedsolution]{standardisedsolution()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inlavaan.R
\name{acfa}
\alias{acfa}
\title{Fit an Approximate Bayesian Confirmatory Factor Analysis Model}
\usage{
acfa(
  model,
  data,
  dp = priors_for(),
  test = "standard",
  vb_correction = TRUE,
  marginal_method = c("skewnorm", "asymgaus", "marggaus", "sampling"),
  marginal_correction = c("shortcut", "shortcut_fd", "hessian", "none"),
  nsamp = 1000,
  samp_copula = TRUE,
  sn_fit_ngrid = 21,
  sn_fit_logthresh = -6,
  sn_fit_temp = 1,
  sn_fit_sample = TRUE,
  control = list(),
  verbose = TRUE,
  debug = FALSE,
  add_priors = TRUE,
  optim_method = c("nlminb", "ucminf", "optim"),
  numerical_grad = FALSE,
  cores = NULL,
  ...
)
}
\arguments{
\item{model}{A description of the user-specified model. Typically, the model
    is described using the lavaan model syntax. See 
    \code{\link[lavaan]{model.syntax}} for more information. Alternatively, a
    parameter table (e.g., the output of the \code{lavParTable()} function) is also
    accepted.}

\item{data}{An optional data frame containing the observed variables used in
    the model. If some variables are declared as ordered factors, lavaan will
    treat them as ordinal variables.}

\item{dp}{Default prior distributions for the different types of model
parameters; a named character vector as returned by \code{\link[=priors_for]{priors_for()}}.}

\item{test}{Character indicating which post-estimation quantities to
compute. Defaults to "standard": posterior fit indices (PPP and DIC),
plus -- for models supported by the casewise machinery and fitted with
a mean structure -- the WAIC (reusing the fit's posterior draws, when
\code{nsamp >= 100}) and a full leave-one-out cross-validation whenever its
predicted serial cost is within a 10-second budget; both are stored
with the fit (see \code{\link[=loo]{loo()}} and \code{\link[=waic]{waic()}}). "none" skips all of these.
Include "loo" (e.g. \code{test = c("standard", "loo")}, or \code{test = "loo"}
alone) to force the full LOO regardless of the budget.}

\item{vb_correction}{Logical indicating whether to apply a variational Bayes
correction for the posterior mean vector of estimates. Defaults to \code{TRUE}.}

\item{marginal_method}{The method for approximating the marginal posterior
distributions. Options include \code{"skewnorm"} (skew-normal), \code{"asymgaus"}
(two-piece asymmetric Gaussian), \code{"marggaus"} (marginalising the Laplace
approximation), and \code{"sampling"} (sampling from the joint Laplace
approximation).}

\item{marginal_correction}{Which type of correction to use when fitting the
skew-normal or two-piece Gaussian marginals. \code{"hessian"} computes the full
\code{"shortcut"} (default) computes only diagonals via central differences
(full z-trace plus Schur complement correction), \code{"shortcut_fd"} is the
same formula using forward differences (roughly half the cost, less
accurate), \code{"hessian"} computes the full Hessian-based correction (slow),
and \code{"none"} (or \code{FALSE}) applies no correction.}

\item{nsamp}{The number of samples to draw for all sampling-based approaches
(including posterior sampling for model fit indices).}

\item{samp_copula}{Logical. When \code{TRUE} (default), posterior samples are
drawn using the copula method with the fitted marginals (e.g. skew-normal
or asymmetric Gaussian), with NORTA correlation adjustment. When \code{FALSE},
samples are drawn from the Gaussian (Laplace) approximation. Only re}

\item{sn_fit_ngrid}{Number of grid points to lay out per dimension when
fitting the skew-normal marginals. A finer grid gives a better fit at the
cost of more joint-log-posterior evaluations. Defaults to \code{21}.}

\item{sn_fit_logthresh}{The log-threshold for fitting the skew-normal. Points
with log-posterior drop below this threshold (relative to the maximum) will
be excluded from the fit. Defaults to \code{-6}.}

\item{sn_fit_temp}{Temperature parameter for fitting the skew-normal.
Defaults to \code{1} (weights are the density values themselves). If \code{NA}, the
temperature is included as an additional optimisation parameter.}

\item{sn_fit_sample}{Logical. When \code{TRUE} (default), a parametric skew-normal
is fitted to the posterior samples for covariance and defined parameters.
When \code{FALSE}, these are summarised using kernel density estimation instead.}

\item{control}{A list of control parameters for the optimiser.}

\item{verbose}{Logical indicating whether to print progress messages.}

\item{debug}{Logical indicating whether to return debug information.}

\item{add_priors}{Logical indicating whether to include prior densities in
the posterior computation.}

\item{optim_method}{The optimisation method to use for finding the posterior
mode. Options include \code{"nlminb"} (default), \code{"ucminf"}, and \code{"optim"}
(BFGS).}

\item{numerical_grad}{Logical indicating whether to use numerical gradients
for the optimisation. Defaults to \code{FALSE} to use analytical gradients.}

\item{cores}{Integer or \code{NULL}. Number of cores for parallel marginal
fitting. When \code{NULL} (default), serial execution is used unless the number
of free parameters exceeds 120, in which case parallelisation is enabled
automatically using all available physical cores. Set to \code{1L} to force
serial execution. If \code{cores > 1}, marginal fits are distributed across
cores using \code{\link[parallel:mclapply]{parallel::mclapply()}} (fork-based; no parallelism on Windows).}

\item{...}{Additional arguments to be passed to the \link[lavaan:lavaan]{lavaan} model fitting
function.}
}
\value{
An S4 object of class \code{INLAvaan} which is a subclass of the
\link[lavaan:lavaan-class]{lavaan} class.
}
\description{
Fit an Approximate Bayesian Confirmatory Factor Analysis Model
}
\details{
The \code{\link[=acfa]{acfa()}} function is a wrapper for the more general \code{\link[=inlavaan]{inlavaan()}}
function, using the following default arguments:
\itemize{
\item \code{int.ov.free = TRUE}
\item \code{int.lv.free = FALSE}
\item \code{auto.fix.first = TRUE} (unless \code{std.lv = TRUE})
\item \code{auto.fix.single = TRUE}
\item \code{auto.var = TRUE}
\item \code{auto.cov.lv.x = TRUE}
\item \code{auto.efa = TRUE}
\item \code{auto.th = TRUE}
\item \code{auto.delta = TRUE}
\item \code{auto.cov.y = TRUE}
}

For further information regarding these arguments, please refer to the
\code{\link[lavaan:lavOptions]{lavaan::lavOptions()}} documentation.
}
\examples{
# The famous Holzinger and Swineford (1939) example
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")

# Fit a CFA model with standardised latent variables
fit <- acfa(HS.model, data = HolzingerSwineford1939, std.lv = TRUE, nsamp = 100)
summary(fit)
}
\seealso{
Typically, users will interact with the specific latent variable
model functions instead, including \code{\link[=acfa]{acfa()}}, \code{\link[=asem]{asem()}}, and \code{\link[=agrowth]{agrowth()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-fitmeasures.R
\name{fitmeasures}
\alias{fitmeasures}
\alias{fitMeasures,INLAvaan-method}
\alias{fitmeasures,INLAvaan-method}
\title{Fit Measures for a Latent Variable Model estimated using INLA}
\arguments{
\item{object}{An object of class \link{INLAvaan}.}

\item{fit.measures}{If \code{"all"}, all fit measures available will be returned. If
only a single or a few fit measures are specified by name, only those are
computed and returned. The LOO measures \code{"elpd_loo"}, \code{"se_loo"},
\code{"p_loo"} and \code{"looic"} (see \code{\link[=loo]{loo()}}) are included in \code{"all"} only when
a LOO result is stored with the fit (\code{test = "loo"} in \code{\link[=inlavaan]{inlavaan()}} or
\code{\link[=add_loo]{add_loo()}}); otherwise they are computed on demand when requested by
name, and recomputed on every call -- store the result with
\code{fit <- add_loo(fit)} (or call \code{\link[=loo]{loo()}} directly) for repeated access.}

\item{baseline.model}{An optional \link{INLAvaan} object representing the
baseline (null) model. Required for incremental fit indices (BCFI, BTLI,
BNFI). Must have been fitted with \code{test != "none"}.}

\item{h1.model}{Ignored (included for compatibility with the lavaan
generic).}

\item{fm.args}{Ignored (included for compatibility with the lavaan
generic).}

\item{output}{Ignored (included for compatibility with the lavaan
generic).}

\item{...}{Additional arguments. Currently supports:
\describe{
\item{\code{rescale}}{Character string controlling how the Bayesian chi-square
is computed, following \code{blavaan::blavFitIndices()}. Options are \code{"devM"}
(default) which uses the deviance rescaled by \code{pD} from DIC, or \code{"MCMC"}
which uses the classical chi-square (\code{(N-1) * F_ML}) and classical
degrees of freedom (\code{p - npar}) at each posterior sample.}
}}
}
\value{
A named numeric vector of fit measures.
}
\description{
Fit Measures for a Latent Variable Model estimated using INLA
}
\examples{
\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
            verbose = FALSE)

# All available fit measures
fitMeasures(fit)

# Specific measures
fitMeasures(fit, c("npar", "DIC", "pD", "ppp"))
}

}
\seealso{
\code{\link[=bfit_indices]{bfit_indices()}}, \code{\link[=compare]{compare()}}, \code{\link[=diagnostics]{diagnostics()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/standardisedsolution.R
\name{standardisedsolution}
\alias{standardisedsolution}
\alias{standardisedSolution}
\alias{standardizedsolution}
\alias{standardizedSolution}
\title{Standardised solution of a latent variable model}
\usage{
standardisedsolution(
  object,
  type = "std.all",
  se = TRUE,
  ci = TRUE,
  level = 0.95,
  postmedian = FALSE,
  postmode = FALSE,
  cov.std = TRUE,
  remove.eq = TRUE,
  remove.ineq = TRUE,
  remove.def = FALSE,
  nsamp = 250,
  ...
)

standardisedSolution(
  object,
  type = "std.all",
  se = TRUE,
  ci = TRUE,
  level = 0.95,
  postmedian = FALSE,
  postmode = FALSE,
  cov.std = TRUE,
  remove.eq = TRUE,
  remove.ineq = TRUE,
  remove.def = FALSE,
  nsamp = 250,
  ...
)

standardizedsolution(
  object,
  type = "std.all",
  se = TRUE,
  ci = TRUE,
  level = 0.95,
  postmedian = FALSE,
  postmode = FALSE,
  cov.std = TRUE,
  remove.eq = TRUE,
  remove.ineq = TRUE,
  remove.def = FALSE,
  nsamp = 250,
  ...
)

standardizedSolution(
  object,
  type = "std.all",
  se = TRUE,
  ci = TRUE,
  level = 0.95,
  postmedian = FALSE,
  postmode = FALSE,
  cov.std = TRUE,
  remove.eq = TRUE,
  remove.ineq = TRUE,
  remove.def = FALSE,
  nsamp = 250,
  ...
)
}
\arguments{
\item{object}{An object of class \link{INLAvaan}.}

\item{type}{If \code{"std.lv"}, the standardized estimates are based
on the variances of the (continuous) latent variables only.
If \code{"std.all"}, the standardized estimates are based
on the variances of both (continuous) observed and latent variables.
If \code{"std.nox"}, the standardized estimates are based
on the variances of both (continuous) observed and latent variables,
but not the variances of exogenous covariates.
Alternatively, \code{type} may be a vector of (observed) variable names
(for example \code{type = c("x1", "x2")}); in that case only the parameters
involving these variables are standardized (the other observed variables are
left unstandardized). This is a generalization of \code{"std.nox"}, where the
(observed) exogenous \code{x} variables are the ones left unstandardized.}

\item{se}{Logical. If TRUE, standard errors for the standardized parameters
will be computed, together with a z-statistic and a p-value.}

\item{ci}{If \code{TRUE}, confidence intervals are added to
the output.}

\item{level}{The confidence level required.}

\item{postmedian}{Logical; if TRUE, include posterior median in estimates.}

\item{postmode}{Logical; if TRUE, include posterior mode in estimates.}

\item{cov.std}{Logical. If \code{TRUE}, the (residual) observed covariances
are scaled by the square root of the \code{Theta} diagonal elements, and the
(residual) latent covariances are scaled by the square root of the
\code{Psi} diagonal elements. If \code{FALSE}, the (residual) observed
covariances are scaled by the square root of the diagonal elements of
the model-implied observed covariance matrix, and the (residual)
latent covariances are scaled similarly using the model-implied
covariance matrix of the latent variables. Documented explicitly here
(rather than inherited) because lavaan >= 0.7-1 renamed this and the
next three arguments to snake_case.}

\item{remove.eq}{Logical. If \code{TRUE}, filter the output by removing all
rows containing equality constraints, if any.}

\item{remove.ineq}{Logical. If \code{TRUE}, filter the output by removing all
rows containing inequality constraints, if any.}

\item{remove.def}{Logical. If \code{TRUE}, filter the output by removing all
rows containing parameter definitions, if any.}

\item{nsamp}{The number of samples to draw from the approximate posterior
distribution for the calculation of standardised estimates.}

\item{...}{Additional arguments sent to \code{lavaan::standardizedSolution()}.}
}
\value{
A \code{data.frame} containing standardised model parameters.
}
\description{
Standardised solution of a latent variable model
}
\examples{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
"
utils::data("HolzingerSwineford1939", package = "lavaan")

# Fit a CFA model with standardised latent variables
fit <- acfa(
  HS.model,
  data = HolzingerSwineford1939,
  test = "none",
  nsamp = 10,
  vb_correction = FALSE,
  verbose = FALSE
)
standardisedsolution(fit, nsamp = 10, se = FALSE, ci = FALSE)
}
\seealso{
\code{\link[=summary]{summary()}}, \code{\link[=coef]{coef()}}, \code{\link[=vcov]{vcov()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-residuals.R
\name{residuals}
\alias{residuals}
\alias{residuals,INLAvaan-method}
\alias{resid,INLAvaan-method}
\title{Model Residuals for INLAvaan Models}
\usage{
\S4method{residuals}{INLAvaan}(object, type = "raw", labels = TRUE, ...)

\S4method{resid}{INLAvaan}(object, type = "raw", ...)
}
\arguments{
\item{object}{An object of class \link{INLAvaan}.}

\item{type}{Character. \code{"raw"} (default) returns the unscaled
difference between the observed and model-implied covariance matrix (and
mean vector, when a mean structure is present). \code{"cor"} (or
\code{"cor.bollen"}) first rescales both matrices to a correlation
matrix. \code{"cor.bentler"} rescales both by the observed variances
(the basis of the SRMR). \code{"normalized"} and \code{"standardized"}
divide the raw residuals by their asymptotic standard errors.
\code{"casewise"} (aliases \code{"case"}, \code{"obs"},
\code{"observations"}, \code{"ov"}) returns observed-minus-fitted values
for each observation.}

\item{labels}{Logical. Attach variable names to the output. Default
\code{TRUE}.}

\item{...}{Currently unused.}
}
\value{
For moment-based \code{type}s, a list with elements
\code{type}, \code{cov}, and (when relevant) \code{mean}. For
\code{type = "casewise"}, a numeric matrix of observed-minus-fitted
values.
}
\description{
Extract the difference between the observed and model-implied (fitted)
sample statistics from a fitted \code{INLAvaan} model. As in \pkg{lavaan}
and \pkg{blavaan}, residuals are computed at the parameter estimates --
here the posterior means -- not as a posterior distribution over residuals.
}
\details{
This delegates to \pkg{lavaan}'s own \code{residuals()} machinery, so the
return structure matches lavaan exactly. Because INLAvaan stores the
posterior means as the point estimates of the fitted object, the residuals
are the observed statistics minus the posterior-mean model-implied
statistics (mirroring \pkg{blavaan}, which likewise inherits lavaan's
\code{residuals()} without overriding it).
}
\examples{
\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
            test = "none", verbose = FALSE)

# Raw residual covariance matrix (posterior means)
residuals(fit)

# SRMR-basis residuals
residuals(fit, type = "cor.bentler")

# Casewise observed-minus-fitted values
head(residuals(fit, type = "casewise"))
}

}
\seealso{
\code{\link[=fitted]{fitted()}}, \code{\link[=predict]{predict()}}, \link[lavaan:fitMeasures]{fitMeasures()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-loglik.R
\name{logLik}
\alias{logLik}
\alias{logLik,INLAvaan-method}
\alias{AIC,INLAvaan-method}
\alias{BIC,INLAvaan-method}
\title{Log-Likelihood for INLAvaan Models}
\usage{
\S4method{logLik}{INLAvaan}(object, type = c("marginal", "plugin"), ...)

\S4method{AIC}{INLAvaan}(object, ..., k = 2)

\S4method{BIC}{INLAvaan}(object, ...)
}
\arguments{
\item{object}{An object of class \link{INLAvaan}.}

\item{type}{Character. \code{"marginal"} (default) returns the
Laplace-approximated marginal log-likelihood (log evidence), the same
quantity \code{\link[=compare]{compare()}} uses for Bayes factors. \code{"plugin"} returns the
classical log-likelihood evaluated at the posterior mean point estimate,
with \code{df}/\code{nobs} attributes and class \code{"logLik"} so it
supports \code{\link[stats]{AIC}}/\code{\link[stats]{BIC}} at the point
estimate. Requires the model to have been fitted with
\code{test != "none"}.}

\item{...}{Currently unused.}

\item{k}{Numeric penalty per parameter passed to the (disabled) \code{AIC()}
method; see \code{\link[stats:AIC]{stats::AIC()}}. Defaults to 2.}
}
\value{
For \code{type = "marginal"}, a length-one numeric of class
\code{inlavaan_logLik} that prints with a note on its interpretation.
For \code{type = "plugin"}, a standard \code{"logLik"} object.
}
\description{
Extract a log-likelihood-flavoured summary from a fitted \code{INLAvaan}
model. Two distinct quantities are available, deliberately not conflated:
the Bayesian marginal log-likelihood (the default) and the classical
log-likelihood evaluated at the posterior mean.
}
\details{
The marginal log-likelihood already integrates over the (Laplace-
approximated) posterior, so it is not on the same scale as a classical
log-likelihood and should not be passed to \code{AIC()}/\code{BIC()} --
doing so would double-penalise model complexity that the evidence has
already accounted for. Use \code{\link[=compare]{compare()}} to compare models via Bayes
factors, DIC, or LOO/WAIC. The plug-in variant exists for users who
specifically want a point-estimate-based classical comparison.

\code{AIC()}/\code{BIC()} on an \code{INLAvaan} fit are themselves
disabled (mirroring \code{\link[=anova]{anova()}}): both are large-sample asymptotic
approximations to quantities INLAvaan already computes directly --
\code{AIC} approximates predictive accuracy, which \code{\link[=loo]{loo()}}/\code{\link[=waic]{waic()}}
already estimate more rigorously; \code{BIC} approximates \eqn{-2} times
the log marginal likelihood, which \code{logLik()} already returns
directly (up to the Laplace approximation). Point-estimate AIC/BIC remain
available for reporting-convention purposes via
\code{AIC(logLik(object, type = "plugin"))} /
\code{BIC(logLik(object, type = "plugin"))}.
}
\examples{
\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
            test = "standard", verbose = FALSE)

# Marginal log-likelihood (log evidence)
logLik(fit)

# Classical log-likelihood at the posterior mean, AIC/BIC-compatible
ll <- logLik(fit, type = "plugin")
AIC(ll)
}

}
\seealso{
\code{\link[=deviance]{deviance()}}, \code{\link[=compare]{compare()}}, \code{\link[=loo]{loo()}}, \code{\link[=waic]{waic()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-sampling.R
\name{sampling}
\alias{sampling}
\alias{sampling,INLAvaan-method}
\title{Draw Samples from the Generative Model}
\usage{
sampling(object, ...)

\S4method{sampling}{INLAvaan}(
  object,
  type = c("lavaan", "theta", "latent", "observed", "implied", "all"),
  nsamp = 1000L,
  samp_copula = TRUE,
  prior = FALSE,
  silent = FALSE,
  ...
)
}
\arguments{
\item{object}{An object of class \link{INLAvaan} (or \code{inlavaan_internal}).}

\item{...}{Additional arguments (currently unused).}

\item{type}{Character string specifying what to sample:
\describe{
\item{\code{"lavaan"}}{(Default) The lavaan-side (constrained) model
parameters. Returns an \code{nsamp} by \code{npar} matrix.}
\item{\code{"theta"}}{The INLAvaan-side unconstrained parameters.
Returns an \code{nsamp} by \code{npar} matrix.}
\item{\code{"latent"}}{Latent variables from the model-implied
distribution. Returns an \code{nsamp} by \code{nlv} matrix (one draw per
posterior sample, not tied to any individual).}
\item{\code{"observed"}}{Observed variables generated from the full
model. Returns an \code{nsamp} by \code{nobs_vars} matrix.}
\item{\code{"implied"}}{Model-implied moments. Returns a length-\code{nsamp}
list, each element a list with \code{cov} (model-implied covariance
matrix) and, when \code{meanstructure = TRUE}, \code{mean} (model-implied
mean vector). For multi-group models each element is itself a
list of groups.}
\item{\code{"all"}}{A named list with elements \code{lavaan}, \code{theta},
\code{latent}, \code{observed}, and \code{implied}.}
}}

\item{nsamp}{Number of samples to draw.}

\item{samp_copula}{Logical. When \code{TRUE} (default), posterior parameter
samples use the copula method with the fitted marginals. When \code{FALSE},
samples are drawn from the joint Gaussian (Laplace) approximation.
Ignored when \code{prior = TRUE}.}

\item{prior}{Logical. When \code{TRUE}, parameters are drawn from the prior
distribution and then propagated through the generative model. When
\code{FALSE} (default), parameters come from the posterior.}

\item{silent}{Logical. When \code{TRUE}, suppresses the informational message
about rejected non-PD draws during prior rejection sampling. Default
\code{FALSE}.}
}
\value{
A matrix or named list, depending on \code{type}.
}
\description{
Sample model parameters, latent variables, or observed variables from the
generative model underlying a fitted INLAvaan model. By default, parameters
are drawn from the \strong{posterior} distribution; set \code{prior = TRUE} to draw
from the \strong{prior} instead (useful for prior predictive checks).
}
\details{
Each row of the output corresponds to a \strong{fresh parameter draw}: a new
\eqn{\boldsymbol\theta^{(s)}} is sampled and then propagated through the
generative chain to produce one latent vector and one observed vector. This
makes \code{sampling()} ideal for \strong{prior and posterior predictive checks}
(e.g., density overlays, test statistic distributions).

The generative chain is:
\deqn{\boldsymbol\theta^{(s)} \sim \pi(\boldsymbol\theta \mid \mathbf{y})}
\deqn{\boldsymbol\eta^{(s)} \sim N((\mathbf{I} - \mathbf{B})^{-1}\boldsymbol\alpha,\,\boldsymbol\Phi)}
\deqn{\mathbf{y}^{*(s)} \sim N(\boldsymbol\Lambda\boldsymbol\eta^{(s)} + \boldsymbol\nu,\,\boldsymbol\Theta)}

If you need \strong{complete replicate datasets} (many observations from a single
parameter draw) — for example, for simulation-based calibration (SBC) — use
\code{\link[=simulate]{simulate()}} instead.

This is distinct from \code{\link[=predict]{predict()}}, which computes individual-specific
factor scores \eqn{\boldsymbol\eta \mid \mathbf{y},\boldsymbol\theta}
conditional on observed data.
}
\examples{
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa("visual =~ x1 + x2 + x3", HolzingerSwineford1939)

# Posterior samples of lavaan-side parameters
samps <- sampling(fit, nsamp = 500)
head(samps)

# Compare copula vs Gaussian sampling
s_cop <- sampling(fit, nsamp = 500, samp_copula = TRUE)
s_gaus <- sampling(fit, nsamp = 500, samp_copula = FALSE)

# Prior predictive samples
y_prior <- sampling(fit, type = "observed", nsamp = 500, prior = TRUE)
}
\seealso{
\code{\link[=simulate]{simulate()}} for generating complete replicate datasets (e.g.,
for SBC); \code{\link[=predict]{predict()}} for individual-specific factor scores;
\code{\link[=bfit_indices]{bfit_indices()}} for Bayesian fit indices.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/INLAvaan-class.R, R/method-others.R,
%   R/method-show.R, R/method-summary.R
\docType{class}
\name{INLAvaan-class}
\alias{INLAvaan-class}
\alias{coef}
\alias{coef,INLAvaan-method}
\alias{nobs}
\alias{nobs,INLAvaan-method}
\alias{show}
\alias{show,INLAvaan-method}
\alias{summary}
\alias{summary,INLAvaan-method}
\title{Class For Representing a (Fitted) Latent Variable Model}
\usage{
\S4method{coef}{INLAvaan}(object, ...)

\S4method{nobs}{INLAvaan}(object, ...)

\S4method{show}{INLAvaan}(object)

\S4method{summary}{INLAvaan}(
  object,
  header = TRUE,
  fit.measures = TRUE,
  estimates = TRUE,
  ci = TRUE,
  standardized = FALSE,
  standardised = standardized,
  rsquare = FALSE,
  postmedian = FALSE,
  postmode = FALSE,
  nmad = TRUE,
  kld = FALSE,
  vb_shift = FALSE,
  priors = TRUE,
  nd = 3L,
  ...
)
}
\arguments{
\item{object}{An object of class \code{INLAvaan}.}

\item{...}{Additional arguments passed to methods.}

\item{header}{Logical; if TRUE, print model fit information header.}

\item{fit.measures}{Logical; if TRUE, print fit measures (DIC and PPP).}

\item{estimates}{Logical; if TRUE, print parameter estimates table.}

\item{ci}{Logical; if TRUE (default), include 95\% credible intervals
(2.5\% and 97.5\% posterior quantiles) in the estimates table.}

\item{standardized}{Logical; if TRUE, include standardized estimates.}

\item{standardised}{Alias of \code{standardized}; either spelling is accepted
(the alias wins if both are supplied).}

\item{rsquare}{Logical; if TRUE, include R-square values for the observed
dependent variables, computed at the posterior mean point estimates.}

\item{postmedian}{Logical; if TRUE, include posterior median in estimates.}

\item{postmode}{Logical; if TRUE, include posterior mode in estimates.}

\item{nmad}{Logical; if TRUE (default), include the NMAD column for
skew-normal marginal fit quality.}

\item{kld}{Logical; if FALSE (default), omit the per-parameter KLD column.
Set to TRUE to show it.}

\item{vb_shift}{Logical; if FALSE (default), omit the VB shift column
(shift in units of posterior SD). Set to TRUE to show it.}

\item{priors}{Logical; if TRUE, include prior information in estimates.}

\item{nd}{Integer; number of decimal places to print for numeric values.}
}
\description{
This is a class that extends the \link[lavaan:lavaan-class]{lavaan} class. Several S4 methods are
available.
}
\section{Slots}{

\describe{
\item{\code{external}}{A list containing an \code{inlavaan_internal} object.}
}}

\examples{
\donttest{
HS.model <- "
  visual  =~ x1 + x2 + x3
  textual =~ x4 + x5 + x6
  speed   =~ x7 + x8 + x9
"
utils::data("HolzingerSwineford1939", package = "lavaan")
fit <- acfa(HS.model, HolzingerSwineford1939, std.lv = TRUE, nsamp = 100,
            test = "none", verbose = FALSE)

# Print basic info
fit

# Detailed summary
summary(fit)

# Extract coefficients
coef(fit)
}

}
\seealso{
\link[lavaan:lavaan-class]{lavaan}, \code{\link[=inlavaan]{inlavaan()}}, \code{\link[=acfa]{acfa()}}, \code{\link[=asem]{asem()}}, \code{\link[=agrowth]{agrowth()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{as_fun_string}
\alias{as_fun_string}
\title{Convert function to single string}
\usage{
as_fun_string(f)
}
\arguments{
\item{f}{Function to convert.}
}
\value{
A single character vector representing the function.
}
\description{
Convert function to single string
}
\examples{
f <- function(x) { x^2 + 1 }
as_fun_string(f)
}
